/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans.
 *
 * CosyBeans is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.application;

/**
 * A specialization of the <code>CosyComponent</code> interface that declares
 * additional functionality for installing this component into the manager and
 * container. Plugins must:
 * <ul>
 *  <li>
 *   Declare a public, default no-arg constructor. The initialization of the 
 *   plugin must not occur in the constructor.
 *  </li>
 *  <li>
 *   Implement the <code>installPlugIn</code> method declared in this interface.
 *   The <code>PlugInManager</code> design contract guarantees, that the plugin
 *   will be added to the <code>CosyPanel</code> container <b>before</b> this 
 *   method is invoked. Therefore, the plugin can use its <code>getCosyPanelParent</code>
 *   reference in this method.
 *  </li>
 * </ul>
 * <p>
 * Plugins are identified by their class type, i.e. the <code>Class</code> object
 * used during the <code>installPlugIn</code> call on the manager is used to 
 * instantiate as well as lookup the plugin (after it has been installed). Therefore, 
 * the class object <b>must not</b> represent an abstract class, an interface or any
 * other non-instantiable object.
 * </p>
 * <p>
 * The manager must, when it is destroying, destroy the plugin before it removes it
 * from its cosy panel parent. During destruction, the plugin must release its
 * resources, but handle all subsequent invocations of its methods gracefully (either
 * by doing NOP, or by returning uninitialized values). This is intended to 
 * prevent other plugins that depend on this one to receive exceptions when they
 * invoke methods on this as part of their destroy. 
 * </p>
 * <p>
 * During installation, this plugin may request the plugin manager to install
 * other plugins. The manager will throw an exception if cyclic referencing is
 * discovered. The manager may keep the list of dependencies, but does not need to.
 * </p>
 * <p>
 * If this plugin depends on other plugins, it should, before calling their methods,
 * check if the plugins are still valid (not destroyed). Even better, instad of 
 * holding fixed references to the plugins, this plugin should invoke 
 * <code>getPlugIn</code> on the manager each time it wants to invoke a method, always
 * checking for <code>null</code> return value. Another option is for the plugin to
 * become a <code>CosyPanelListener</code> of its panel and stop executing methods 
 * on other plugins after it receives <code>destroying</code> event.
 * </p>
 * 
 */
public interface PlugIn {
	
	/**
	 * Installs the plugin. It is the responsibility of the manager to detect
	 * the cyclic references, add this plugin to the parent cosy panel and its
	 * containment list. Implement this method to provide installation for
	 * the plugin; it can reference the <code>getCosyPanelParent</code>.
	 * 
	 * @param manager the reference to the manager installing this plugin
	 * 
	 * @exception PlugInException when the installation fails for whatever reason
	 *            (invalid preconditions, cyclic reference etc).
	 */
	public void installPlugIn(PlugInManager manager) throws PlugInException;
	
	/**
	 * Destroys the plugin and all resources associated with this plugin should
	 * be released.
	 * 
	 */
	public void destroy();
}

