/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of Java-Common.
 *
 * Java-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Java-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Java-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.application.state.impl;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Stack;

import javax.xml.transform.TransformerFactory;
import javax.xml.transform.sax.SAXResult;
import javax.xml.transform.sax.SAXSource;
import javax.xml.transform.stream.StreamResult;

import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.DTDHandler;
import org.xml.sax.EntityResolver;
import org.xml.sax.ErrorHandler;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXNotRecognizedException;
import org.xml.sax.SAXNotSupportedException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;

import com.cosylab.application.state.State;
import com.cosylab.application.state.StateFactory;
import com.cosylab.application.state.StateStorage;

public class NewXMLStateStorage extends DefaultStateStorage {
    private static final String ROOT_EL="states";
    private static final String STATE_EL="state";
    
    private static abstract class DefaultReader implements XMLReader {
        ContentHandler ch;
        HashMap<String, Boolean> features;
        {
            features=new HashMap<>();
            features.put("http://xml.org/sax/features/namespaces", Boolean.TRUE);
            features.put("http://xml.org/sax/features/namespace-prefixes", Boolean.FALSE);
        }
        
        HashMap<String,Object> properties;
        
        public ContentHandler getContentHandler() {
            return null;
        }
        public DTDHandler getDTDHandler() {
            return null;
        }
        public EntityResolver getEntityResolver() {
            return null;
        }
        public ErrorHandler getErrorHandler() {
            return null;
        }
        public boolean getFeature(String name) throws SAXNotRecognizedException, SAXNotSupportedException {
            return features.get(name);
        }
        public Object getProperty(String name) throws SAXNotRecognizedException, SAXNotSupportedException {
            if (properties==null) throw new SAXNotRecognizedException(name+" is not in the properties map");
            return properties.get(name);
        }
        public void parse(String systemId) throws IOException, SAXException {
            parse(new InputSource(systemId));
        }
        public void setContentHandler(ContentHandler handler) {
            ch=handler;
        }
        public void setDTDHandler(DTDHandler handler) {
        }
        public void setEntityResolver(EntityResolver resolver) {
        }
        public void setErrorHandler(ErrorHandler handler) {
        }
        public void setFeature(String name, boolean value) throws SAXNotRecognizedException, SAXNotSupportedException {
            features.put(name, value);
        }
        public void setProperty(String name, Object value) throws SAXNotRecognizedException, SAXNotSupportedException {
            if (properties==null) properties=new HashMap<>();
            properties.put(name, value);
        }
    }

    private abstract static class AttrsImpl implements Attributes {
        public int getIndex(String uri, String localName) {return getIndex(localName);}
        public String getLocalName(int index) {return getQName(index);}
        public String getType(String qName) {return getType(getIndex(qName));}
        public String getType(String uri, String localName) {return getType(localName);}
        public String getURI(int index) {return "";}
        public String getValue(String qName) {return getValue(getIndex(qName));}
        public String getValue(String uri, String localName) {return getValue(localName);}
        public String getType(int index) { return (index<0 || index>=getLength())?null:"CDATA"; }
    }
    
    private static class MyAttributes extends AttrsImpl {
        private String[] atrNames=new String[0];
        private String[] atrVals=new String[0];
        
        public MyAttributes() {
        }
        
        public MyAttributes(String[] atrNames, String[] atrVals){
            this.atrNames=atrNames;
            this.atrVals=atrVals;
        }
        
        public int getIndex(String qName) {
            return atrNames.length;
        }

        public int getLength() {
            return atrVals.length;
        }

        public String getQName(int index) {
            return atrNames[index];
        }

        public String getValue(int index) {
            return atrVals[index];
        }
    }
    
    private static class MyReader extends DefaultReader {
        List<State> states;
        public MyReader(List<State> states) {
            this.states=states;
        }
        public void parse(InputSource input) throws SAXException {
            ch.startDocument();
            ch.startElement("", ROOT_EL, ROOT_EL, new MyAttributes());
            
            for (int i = 0; i < states.size(); i++) {
                State s = states.get(i);
                storeState(STATE_EL, s);
            }            
            
            ch.endElement("", ROOT_EL, ROOT_EL);
            ch.endDocument();
        }
        
        private void storeState(String name, State s) throws SAXException{
            Iterator<String> iter = s.keySet().iterator();
            ArrayList<String> atrNames=new ArrayList<>();
            ArrayList<String> atrVals=new ArrayList<>();
            ArrayList<String> stNames=new ArrayList<>();
            ArrayList<Object> stVals=new ArrayList<>();
            while (iter.hasNext()) {
                String key = iter.next();
                Object value = s.getObject(key);        
                if (value instanceof State) {
                    stNames.add(key);
                    stVals.add(value);
                } else {
                    atrNames.add(key);
                    atrVals.add(String.valueOf(value));
                }
            }
            String[] arAtrNames=atrNames.toArray(new String[atrNames.size()]);
            String[] arAtrVals=atrVals.toArray(new String[atrVals.size()]);
            ch.startElement("", name, name, new MyAttributes(arAtrNames,arAtrVals));
            for (int i = 0; i < stNames.size(); i++) {
                storeState(stNames.get(i), (State)stVals.get(i));
            }
            ch.endElement("", name, name);
        }
    }
    
    private static class MyContentHandler extends DefaultHandler {
        private List<State> output;
        private Stack<State> current;
        public MyContentHandler(List<State> output) {
            this.output=output;
            this.current=new Stack<>();
        }
        @Override
        public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException {
            if (current.isEmpty() && ROOT_EL.equals(localName)) return;
            State newSt;
            if (current.isEmpty()) {
                newSt=StateFactory.createState();
                output.add(newSt);
            } else {
                newSt=(current.peek()).createState(localName);
            }
            for (int i = 0; i < attributes.getLength(); i++) {
                newSt.putString(attributes.getLocalName(i), attributes.getValue(i));
            }
            current.push(newSt);
        }
        @Override
        public void endElement(String uri, String localName, String qName) throws SAXException {
            if (current.isEmpty() && ROOT_EL.equals(localName)) return;
            current.pop();
        }
    }
    
    /**
     * DOCUMENT ME!
     *
     * @param is DOCUMENT ME!
     *
     * @return DOCUMENT ME!
     *
     * @throws IOException DOCUMENT ME!
     */
    public static final State[] loadStates(InputSource is)
    	throws IOException
    {
        ArrayList<State> retList = new ArrayList<>();
    	loadStates(is, retList);
    
    	State[] ret = new State[retList.size()];
    	retList.toArray(ret);
    
    	return ret;
    }

    private static final void loadStates(InputSource is, List<State> retStates)
    	throws IOException
    {
    	try {
    		if (is == null || (is.getByteStream() == null && is.getCharacterStream()==null)) {
    			return;
    		}
    		SAXSource src=new SAXSource(is);
            SAXResult res=new SAXResult(new MyContentHandler(retStates));
            TransformerFactory.newInstance().newTransformer().transform(src, res);
    	} catch (Exception e) {
    		IOException exc = new IOException("Unable to read state from "
    			    + is.getSystemId() + ": " + e.getMessage());
    		exc.initCause(e);
    		throw exc;
    	}
    }

    private static final void storeStates(List<State> outStates, OutputStream os)
    	throws IOException
    {
        try {
            SAXSource src=new SAXSource(new MyReader(outStates), new InputSource());
            StreamResult res=new StreamResult(os);
            TransformerFactory.newInstance().newTransformer().transform(src, res);

    		os.flush();
    		os.close();
    	} catch (Exception e) {
    		e.printStackTrace();
    
    		IOException exc = new IOException("Unable to store state! "
    			    + e.getMessage());
    		exc.initCause(e);
    		throw exc;
    	}
    }

    /**
     * DOCUMENT ME!
     *
     * @param outStates DOCUMENT ME!
     * @param os DOCUMENT ME!
     *
     * @throws IOException DOCUMENT ME!
     */
    public static final void storeStates(State[] outStates, OutputStream os)
    	throws IOException
    {
    	storeStates(Arrays.asList(outStates), os);
    }

    /**
    		             *
    		             */
    public NewXMLStateStorage()
    {
    	super();
    }

    /**
     * Creates a new XMLStateStorage object.
     *
     * @param ss DOCUMENT ME!
     */
    public NewXMLStateStorage(StateStorage ss)
    {
    	super(ss);
    }

    /**
     * DOCUMENT ME!
     *
     * @param is DOCUMENT ME!
     *
     * @throws IOException DOCUMENT ME!
     */
    public void load(InputSource is) throws IOException
    {
    	loadStates(is, states);
    }

    /**
     * DOCUMENT ME!
     *
     * @param is DOCUMENT ME!
     *
     * @throws IOException DOCUMENT ME!
     */
    public void load(InputStream is) throws IOException
    {
    	load(new InputSource(is));
    }

    /* (non-Javadoc)
     * @see com.cosylab.application.state.StateStorage#load(java.io.InputStream)
     */
    @Override
    public void load(String filePath, String applicationName)
    	throws IOException
    {
    	if (delegate != null) {
    		delegate.load(filePath, applicationName);
    		addAll(delegate.getStates());
    
    		return;
    	}
    
    	// get the InputStream if file exists
    	InputStream is = getInputStream(filePath, applicationName + ".xml");
    
    	if (is != null) {
    		try {
    			load(is);
    		} catch (IOException e) {
    			IOException ex = new IOException("Failed to load from file '"
    				    + filePath + "/" + applicationName + ".xml'");
    			ex.initCause(e);
    			throw ex;
    		}
    	}
    }

    /* (non-Javadoc)
     * @see com.cosylab.application.state.StateStorage#store(java.io.FileInputStream, java.lang.String)
     */
    public void store(OutputStream os) throws IOException
    {
    	storeStates(states, os);
    }

    /* (non-Javadoc)
     * @see com.cosylab.application.state.StateStorage#store(java.io.OutputStream)
     */
    @Override
    public void store(String filePath, String applicationName)
    	throws IOException
    {
    	OutputStream os = getOutputStream(filePath, applicationName + ".xml");
    	store(os);
    }

}
