package com.cosylab.cosydist;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.StringWriter;
import java.util.regex.Pattern;

import com.cosylab.cosydist.Logger;

public class LicenseHeaderExecutor {

	/**
	 * Pattern that is used to tell, whether a line is a comment (or part of
	 * it), or not.
	 */
	private final static Pattern commentPattern = Pattern
			.compile("[ \t]*\\*.*|[ \t]*/\\*\\*.*|[ \t]*/\\*.*|[ \t]*\\*/[ \t]*|[ \t]*|[ \t]*//.*");
	private final static String lineSeparator = System
			.getProperty("line.separator");

	/**
	 * Attaches the license to all the java files in given source directories
	 * that don't already include license at the beginning of the file.
	 * 
	 * @param sourceDirectories
	 *            array of source directories
	 * @param licenseFile
	 *            path to the file with license
	 * @param log
	 *            logger for displaying results
	 */
	public static void attachLicense(String[] sourceDirectories,
			File licenseFile, Logger log) {

		attachLicense(sourceDirectories, licenseFile, log, true, false, false);
	}

	/**
	 * Attaches the license to all the java files in given source directories
	 * that don't already include license at the beginning of the file.<br />
	 * You can choose if you want to see, what happens to files, or not.<br />
	 * You can also choose if you want to see, what files are skipped (non .java
	 * files), or not.
	 * 
	 * @param sourceDirectories
	 *            array of source directories
	 * @param licenseFile
	 *            path to the file with license
	 * @param log
	 *            logger for displaying results
	 * @param displayProcessed
	 *            true if you want to display processed files, false otherwise
	 * @param displaySkipped
	 *            true if you want to display skipped files, false otherwise
	 * @param displayNonJava
	 *            true if you want to display files that don't have .java ending, false otherwise
	 */
	public static void attachLicense(String[] sourceDirectories,
			File licenseFile, Logger log, boolean displayProcessed,
			boolean displaySkipped, boolean displayNonJava) {
		
		if (sourceDirectories == null) {
			log.info("You have to define <sourceDirectory> tag.");
			return;
		}
		if (licenseFile == null) {
			log.info("You have to define <licenseFile> tag.");
			return;
		}

		String license = readLicense(licenseFile, log, displayProcessed);
		if (license == null) {
			return;
		}
		
		/**
		 * Counts the files, this plugin changes.
		 * <ul>
		 * <li>stats[0] = how many files don't have .java ending</li>
		 * <li>stats[1] = how many files have correct license</li>
		 * <li>stats[2] = how many files did this plugin change</li>
		 * </ul>
		 */
		int[] stats = new int[3];

		File sourceDirectoryFile;
		for (int i = 0; i < sourceDirectories.length; i++) {
			sourceDirectoryFile = new File(sourceDirectories[i]);

			if (displayProcessed)
				System.out.println(license);
			attach(sourceDirectoryFile, license, 0, log, displayProcessed, displaySkipped, displayNonJava, stats);
		}
		
		// display results
		log.info((stats[1] + stats[2])
				+ " java file" + (stats[1] + stats[2] == 1 ? "" : "s") + " checked");
		log.info("license added to " + stats[2] + " file" + (stats[2] == 1 ? "" : "s"));
	}

	/**
	 * Reads license file and puts the license into a String, which will be
	 * later attached to java files.
	 * 
	 * @param licenseFile
	 *            file, where license is
	 * @param log
	 *            logger for displaying results
	 * @param displayProcessed
	 *            true if you want to display processed files, false otherwise
	 */
	private static String readLicense(File licenseFile, Logger log, boolean displayProcessed) {
		String currentLicense = null;
		if (displayProcessed)
			log.info("licenseFile = " + licenseFile);
		StringWriter sw = new StringWriter();
		String currentLine = null;
		try {
			BufferedReader b = new BufferedReader(new FileReader(licenseFile));
			int lineNumber = 0;
			while (b.ready()) {
				currentLine = b.readLine();
				lineNumber++;
				if (currentLine == null || !commentPattern.matcher(currentLine).matches()){
					log.severe("License is false. License must be writtren in java comment form.");
					log.severe("Line nr. " + lineNumber + " is not properly commented: " + currentLine);
					return null;
				}
				sw.write(currentLine);
				sw.write(lineSeparator);
			}
			currentLicense = sw.toString().trim();
			if ((currentLicense.contains("/**") || currentLicense.contains("/*")) && !currentLicense.contains("*/")){
				log.severe("License is false. License must end with \"/*\".");
				return null;
			}
		} catch (FileNotFoundException e) {
			log.severe("License can not be found! "
					+ licenseFile.getAbsolutePath());
		} catch (IOException e) {
			System.err.println("License can not be read!");
		}
		return currentLicense;
	}

	/**
	 * Attaches the license to all .java files in folder.
	 * 
	 * @param folder
	 *            where to attach the license
	 * @param license
	 *            to be attached
	 * @param level
	 *            how deep in folder tree are we
	 * @param log
	 *            logger for displaying results
	 * @param displayProcessed
	 *            true if you want to display processed files, false otherwise
	 * @param displaySkipped
	 *            true if you want to display skipped files, false otherwise
	 * @param displayNonJava
	 *            true if you want to display files that don't have .java ending, false otherwise
	 * @param stats
	 *            array that counts statistics of files that are changed
	 */
	private static void attach(File folder, String license, int level,
			Logger log, boolean displayProcessed, boolean displaySkipped, boolean displayNonJava, int[] stats) {
		File[] files = folder.listFiles();
		String space = "";
		for (int i = 0; i < level * 2; i++) {
			space += " ";
		}

		for (int i = 0; i < files.length; i++) {
			if (files[i].isDirectory()) {
				if (displayProcessed) {
					//log.info(space + "<" + files[i].getName() + ">");
				}
				attach(files[i], license, level + 1, log, displayProcessed, displaySkipped, displayNonJava, stats);
			} else {
				if (!files[i].getName().toLowerCase().endsWith(".java")) {
					stats[0]++;
					if (displayNonJava) {
						//log.info(space + files[i].getName() + " < skipped");
					}
					continue;
				}
				writeLicensedFile(files[i], license, log, space, displayProcessed, displaySkipped, stats);
			}
		}
	}

	/**
	 * Writes the license to the beginning of the file.
	 * 
	 * @param writeToFile
	 *            file that license will be added to
	 * @param license
	 *            to be attached
	 * @param log
	 *            logger for displaying results
	 * @param space
	 *            for displaying log
	 * @param displayProcessed
	 *            true if you want to display processed files, false otherwise
	 * @param displaySkipped
	 *            true if you want to display skipped files, false otherwise
	 * @param stats
	 *            array that counts statistics of files that are changed
	 */
	private static void writeLicensedFile(File writeToFile, String license,
			Logger log, String space, boolean displayProcessed, boolean displaySkipped, int[] stats) {
		try {
			// read file
			BufferedReader b = new BufferedReader(new FileReader(writeToFile));

			StringWriter currentFileStringWriter = new StringWriter(), currentFilesLicenseStringWriter = new StringWriter();

			String line = null;
			boolean firstComparison = true;
			while (b.ready()) {
				line = b.readLine();
				if (commentPattern.matcher(line).matches() && firstComparison) {
					currentFilesLicenseStringWriter.write(line);
					currentFilesLicenseStringWriter.write(lineSeparator);
				} else {
					if (firstComparison) {
						firstComparison = false;
						String currentFilesLicense = currentFilesLicenseStringWriter
								.toString();
						if (currentFilesLicense.contains(license)) {
							stats[1]++;
							if (displaySkipped)
								log.info(space + writeToFile.getName()
										+ " < license is ok");
							b.close();
							return;
						}
					}
					currentFileStringWriter.write(line);
					currentFileStringWriter.write(lineSeparator);
				}
			}
			b.close();

			String currentFile = currentFileStringWriter.toString();

			// check if file doesn't contain correct license
			if (currentFile != null) {
				BufferedWriter bw = new BufferedWriter(new FileWriter(
						writeToFile));

				// first writes the license into the file
				bw.write(license);
				bw.newLine();
				bw.newLine();
				bw.write(currentFile);

				bw.close();
				stats[2]++;
				if (displayProcessed)
					log.info(space + writeToFile.getName()
							+ " < license changed");
			}
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
}
