/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans.
 *
 * CosyBeans is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui;

import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.beans.PropertyVetoException;
import java.util.Map;

import javax.swing.AbstractAction;
import javax.swing.JFrame;

import com.cosylab.gui.adapters.Converter;
import com.cosylab.gui.components.LabelledGauger;
import com.cosylab.gui.components.customizer.AbstractCustomizerPanel;
import com.cosylab.gui.components.range2.TrimValuePolicy;
import com.cosylab.gui.components.util.PopupManager;
import com.cosylab.gui.displayers.ConvertibleDisplayer;
import com.cosylab.gui.displayers.DataConsumer;
import com.cosylab.gui.displayers.DataSource;
import com.cosylab.gui.displayers.DataState;
import com.cosylab.gui.displayers.DisplayerUtilities;
import com.cosylab.gui.displayers.DoubleConsumer;
import com.cosylab.gui.displayers.DoubleConsumerMulticaster;
import com.cosylab.gui.displayers.DoubleDisplayer;
import com.cosylab.util.CommonException;

/**
 * Gauger displayer.
 *
 * @version 1.0
 */
public class GaugerDisplayer extends LabelledGauger
	implements DoubleDisplayer, ConvertibleDisplayer 
{
	private static final long serialVersionUID = 1L;
	private int suspendCount = 0;
	private double value;
	private DataState dataState = new DataState(DataState.UNDEFINED);
	private DataSource dataSource;  //  @jve:decl-index=0:
	private Converter converter;  //  @jve:decl-index=0:
	private AbstractCustomizerPanel customizer;
	private InfoDialog infoDialog;
	private PopupManager popupManager;
	
	/** 
	 * Creates new gauger displayer. It defaults to initial values as specified
	 * in the gauger component.
	 */
	public GaugerDisplayer()
	{
		super();
		initialize();
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#setCharacteristics(java.util.Map)
	 */
	public void setCharacteristics(Map characteristics)
	{
		if (characteristics == null) {
			throw new NullPointerException("characteristics");
		}

		DisplayerUtilities.setCharacteristics(characteristics, this);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getDataConsumer(java.lang.Class)
	 */
	public DataConsumer getDataConsumer(Class type)
	{
		if (type == DoubleConsumer.class) {
			return this;
		}

		return DoubleConsumerMulticaster.createDataConsumer(type, this);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.CommonDisplayer#getDataState()
	 */
	public DataState getDataState()
	{
		return dataState;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getDefaultDataConsumer()
	 */
	public DataConsumer getDefaultDataConsumer()
	{
		return this;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getSupportedCharacteristics()
	 */
	public String[] getSupportedCharacteristics()
	{
		return DisplayerUtilities.COMMON_NUMERIC_DISPLAYER_CHARACTERISTICS;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getSupportedConsumerTypes()
	 */
	public Class[] getSupportedConsumerTypes()
	{
		return DoubleConsumerMulticaster.PREFERED_CONSUMER_TYPES;
	}

	/**
	 * @see DoubleDisplayer#setValue(double)
	 */
	@Override
    public void setValue(double value)
	{
		if (this.value == value) {
			return;
		}

		double oldValue = this.value;
		this.value = value;
		firePropertyChange("value", oldValue, value);

		if (!isSuspended()) {
				super.setValue(value);
		}
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.CommonDisplayer#cleanup()
	 */
	public void cleanup()
	{
		internalCleanup();
		updateDataState(new DataState(DataState.NOT_INITIALIZED));
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.core.CosyComponent#destroy()
	 */
	public void destroy()
	{
		suspend();
		cleanup();
		dataState = null;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.CommonDisplayer#resume()
	 */
	public void resume()
	{
		if (suspendCount > 0) {
			suspendCount--;
		}
		
		if (suspendCount==0) {
			setEnabled(true);
			super.setValue(value);
		}
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.CommonDisplayer#suspend()
	 */
	public void suspend()
	{
		suspendCount++;
		setEnabled(false);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#updateDataState(com.cosylab.gui.displayers.DataState)
	 */
	public void updateDataState(DataState state)
	{
		DataState old = dataState;
		dataState = state;
		firePropertyChange(DATA_STATE, old, dataState);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DoubleConsumer#updateValue(long, double)
	 */
	public void updateValue(long timestamp, double value)
		throws CommonException
	{
		setValue(value);
	}

	private void initialize()
	{
		
		internalCleanup();
	}

	private void internalCleanup() 
	{
		setTitle("Gauger");
		setUnits(null);
		setUnitsVisible(true);
		setFormat("%3.2f");
		setResizable(true);
		setTitleVisible(true);
		setLogarithmicScale(false); 
		setValuePolicy(new TrimValuePolicy());
		setPopupEnabled(true);
	}
	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.CommonDisplayer#isSuspended()
	 */
	public boolean isSuspended()
	{
		return (suspendCount > 0);
	}

	@Override
    public double getValue()
	{
		return value;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.Displayer#getDataSource()
	 */
	public DataSource getDataSource()
	{
		return dataSource;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.Displayer#setDataSource(com.cosylab.gui.displayers.DataSource)
	 */
	public void setDataSource(DataSource dataSource)
		throws PropertyVetoException
	{
		DisplayerUtilities.prepareNewDataSource(dataSource,this);
		
		DataSource old= this.dataSource;
		this.dataSource = dataSource;
		
		firePropertyChange(DATA_SOURCE,old,dataSource);
	}

	
	/**
	 * Returns the converter.
	 * @return Returns the converter.
	 */
	public Converter getConverter() {
		return converter;
	}

	
	/**
	 * Sets new converter.
	 * @param converter The converter to set.
	 * @throws PropertyVetoException if set fails
	 */
	public void setConverter(Converter converter) throws PropertyVetoException {
		if (this.converter != null && this.converter.equals(converter) ||
				(this.converter == null && converter == null)) return;
		DisplayerUtilities.prepareNewConverter(converter,this);
		
		Converter old= this.converter;
		this.converter = converter;
		
		firePropertyChange(CONVERTER_PROPERTY,old,this.converter);    
	}

	@Override
	public AbstractCustomizerPanel getCustomizer() {
		if (customizer == null) {
			customizer = AbstractCustomizerPanel.findCustomizer(this);
		}
   
		return customizer;
	}
	
	public InfoDialog getInfoDialog() {
		if (infoDialog == null) {
			infoDialog = new InfoDialog(this);
		}
		return infoDialog;
	}
	
	@Override
    public PopupManager getPopupManager()
	{
		if (popupManager == null) {
			popupManager = super.getPopupManager();
			popupManager.addAction(new AbstractAction("Info...") {
	
				private static final long serialVersionUID = 1L;

					public void actionPerformed(ActionEvent e)
					{
						getInfoDialog().setVisible(true);
					}
				});
		}
		
		return popupManager;
	}

	/**
	 * Run test applet.
	 *
	 * @param args command line parameters
	 */
	public static void main(String[] args)
	{
		JFrame dialog = new JFrame();

		dialog.getContentPane().setLayout(new GridBagLayout());

		GridBagConstraints c = new GridBagConstraints();
		c.fill = GridBagConstraints.BOTH;
		c.weightx = 1.0;
		c.weighty = 1.0;

		//dialog.setModal(true);

		final GaugerDisplayer gd = new GaugerDisplayer();
		gd.setPopupEnabled(false);
		
		try {
			gd.setConverter(new com.cosylab.gui.adapters.MultiplierConverter(2));
		} catch (PropertyVetoException e) {
			e.printStackTrace();
		}
		dialog.setSize(100, 400);

		dialog.getContentPane().add(gd, c);
		dialog.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		gd.setValue(1.8);

//		Timer t = new Timer();
//		t.schedule(new TimerTask(){
//			int i = 0;
//			public void run() {
//				gd.setValue(i);
//				i = (i + 1)%20;
//			}
//			
//		}, 500, 500);
		
		//Timer t = new Timer();
		//t.schedule(new TimerTask() {
		//	public void run() {
		//		if (gd.isSuspended()) {
		//			gd.resume();
		//		} else {
		//			gd.suspend();
		//		}
		//		Debug.out("Gauger#isSuspended="+gd.isSuspended());
		//	}
		//},500,500);

		//dialog.show();
		dialog.setVisible(true);
		
		
		//System.exit(0);

		/*javax.swing.JDialog dialog = new javax.swing.JDialog();

		dialog.getContentPane().setLayout(new GridBagLayout());

		GridBagConstraints c = new GridBagConstraints();
		c.fill = GridBagConstraints.BOTH;
		c.weightx = 1.0;
		c.weighty = 1.0;

		dialog.setModal(true);

		GaugerDisplayer gd = new GaugerDisplayer();

		dialog.setSize(400, 400);

		dialog.getContentPane().add(gd, c);

		gd.gauger.setValue(1.8);

		dialog.show();

		System.exit(0);
		*/
	}	
}
/* __oOo__ */
