/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans.
 *
 * CosyBeans is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.adapters;

import java.beans.PropertyVetoException;
import java.util.Map;

import com.cosylab.gui.displayers.DataConsumer;
import com.cosylab.gui.displayers.DataState;


/**
 * 
 * <code>ConverterChain</code> combines several converter into a single one.
 * Single converters can be added to this chain. The value will then be converted
 * by each of this single converters resulting in an arbitrary function applied
 * to the input value.
 * 
 * @author Igor Kriznar (igor.kriznarATcosylab.com)
 *
 */
public class ConverterChain implements Converter {
	
	private static final long serialVersionUID = 1785528428091352421L;
	private Converter[] converters;
	private Converter first;
	private Converter last;
	
	/**
	 * Creates new instance of ConverterChain.
	 */
	public ConverterChain() throws PropertyVetoException {
		this(new Converter[]{new IdentityConverter()});
	}

	/**
	 * Creates new instance of ConverterChain.
	 */
	public ConverterChain(Converter[] converters) throws PropertyVetoException {
		initialize(converters);
	}

	/**
	 * @param converters
	 * @throws PropertyVetoException
	 */
	private void initialize(Converter[] converters) throws PropertyVetoException {
		if (converters == null || converters.length ==0) {
			converters = new Converter[]{new IdentityConverter()};
		}
		this.converters=converters;
		first= converters[0];
		last= converters[converters.length-1];
		for (int i = 1; i < converters.length; i++) {
			Converter c1 = converters[i-1];
			Converter c2 = converters[i];
			c1.addConsumer(c2);
		}
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataSource#addConsumer(com.cosylab.gui.displayers.DataConsumer)
	 */
	public void addConsumer(DataConsumer consumer) throws PropertyVetoException {
		last.addConsumer(consumer);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataSource#removeConsumer(com.cosylab.gui.displayers.DataConsumer)
	 */
	public void removeConsumer(DataConsumer consumer) {
		last.removeConsumer(consumer);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataSource#getConsumers()
	 */
	public DataConsumer[] getConsumers() {
		return last.getConsumers();
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataSource#getAcceptableConsumerTypes()
	 */
	public Class<DataConsumer>[] getAcceptableConsumerTypes() {
		return last.getAcceptableConsumerTypes();
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataSource#removeAllConsumers()
	 */
	public void removeAllConsumers() {
		last.removeAllConsumers();
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getDataConsumer(java.lang.Class)
	 */
	public <D extends DataConsumer>D getDataConsumer(Class<D> type) {
		return first.getDataConsumer(type);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getDefaultDataConsumer()
	 */
	public DataConsumer getDefaultDataConsumer() {
		return first.getDefaultDataConsumer();
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#updateDataState(com.cosylab.gui.displayers.DataState)
	 */
	public void updateDataState(DataState state) {
		first.updateDataState(state);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#setCharacteristics(java.util.Map)
	 */
	public void setCharacteristics(Map characteristics) {
		first.setCharacteristics(characteristics);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getName()
	 */
	public String getName() {
		return first.getName();
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getSupportedCharacteristics()
	 */
	public String[] getSupportedCharacteristics() {
		return first.getSupportedCharacteristics();
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getSupportedConsumerTypes()
	 */
	public Class<DataConsumer>[] getSupportedConsumerTypes() {
		return first.getSupportedConsumerTypes();
	}

	/**
	 * Returns the last converter in the chain.
	 * 
	 * @return
	 */
	public Converter getLast() {
		return last;
	}

	/**
	 * Returns the first converter in the chain.
	 * 
	 * @return
	 */
	public Converter getFirst() {
		return first;
	}

	/**
	 * Returns the converter under the specified index.
	 * 
	 * @param index index of the requested converter
	 * @return the converter
	 */
	public Converter get(int index) {
		return converters[index];
	}

	/**
	 * Returns the number of all converters in the chain.
	 * 
	 * @return
	 */
	public int size() {
		return converters.length;
	}

	/**
	 * Returns all converters in the chain.
	 * 
	 * @return
	 */
	public Converter[] getConverters() {
		return converters;
	}

	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
    public String toString() {
		StringBuffer buffer = new StringBuffer();
		for (Converter c : converters) {
			buffer.append(c.toString() + ";");
		}
		return buffer.substring(0, buffer.length()-1);
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Object#clone()
	 */
	@Override
	public Object clone() throws CloneNotSupportedException {
		try {
			ConverterChain c = new ConverterChain();
			Converter[] cc= new Converter[converters.length];
			for (int i=0; i<cc.length; i++) {
				cc[i]=(Converter)converters[i].clone();
			}
			try {
				c.initialize(cc);
			} catch (PropertyVetoException e) {
				e.printStackTrace();
			}
			return c;
		} catch (PropertyVetoException e) {
			throw new CloneNotSupportedException();
		}
	}
}

