/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans.
 *
 * CosyBeans is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.adapters;

import java.util.Map;

import com.cosylab.gui.displayers.AsynchronousNumberConsumer;
import com.cosylab.gui.displayers.AsynchronousObjectConsumer;
import com.cosylab.gui.displayers.AsynchronousStringConsumer;
import com.cosylab.gui.displayers.DataConsumer;
import com.cosylab.gui.displayers.DataState;
import com.cosylab.gui.displayers.NonblockingNumberConsumer;
import com.cosylab.gui.displayers.NonblockingObjectConsumer;
import com.cosylab.gui.displayers.NonblockingStringConsumer;
import com.cosylab.gui.displayers.SynchronousNumberConsumer;
import com.cosylab.gui.displayers.SynchronousObjectConsumer;
import com.cosylab.gui.displayers.SynchronousStringConsumer;
import com.cosylab.gui.displayers.UpdateRequest;
import com.cosylab.util.CommonException;


/**
 * <code>FeedbackConsumerSupport</code> is support class fo all implementators
 * who want to implement consumer, which receives requests for data change
 * from data consumers and displayers.
 *
 * @author <a href="mailto:igor.kriznar@cosylab.com">Igor Kriznar</a>
 * @version $Id: FeedbackConsumerSupport.java,v 1.4 2008-04-22 12:31:02 jbobnar Exp $
 *
 * @see com.cosylab.gui.displayers.CommonDisplayer
 * @since Apr 7, 2004.
 */
public class FeedbackConsumerSupport implements AsynchronousNumberConsumer,
	AsynchronousObjectConsumer, AsynchronousStringConsumer,
	SynchronousNumberConsumer, SynchronousObjectConsumer,
	SynchronousStringConsumer, NonblockingNumberConsumer,
	NonblockingObjectConsumer, NonblockingStringConsumer
{
	protected String name = "FeedbackConsumerSupport";
	private static final Class[] supported = {
		AsynchronousNumberConsumer.class, AsynchronousObjectConsumer.class,
		AsynchronousStringConsumer.class, SynchronousNumberConsumer.class,
		SynchronousObjectConsumer.class, SynchronousStringConsumer.class,
		NonblockingNumberConsumer.class, NonblockingObjectConsumer.class,
		NonblockingStringConsumer.class
	};

	/**
	 * Creates new instance.
	 */
	public FeedbackConsumerSupport()
	{
		super();
	}

	/**
	 * Forwards call to <code>updateRequest()</code> and replays to request
	 * when update finishes.
	 *
	 * @see com.cosylab.gui.displayers.AsynchronousStringConsumer#updateAcynchronously(com.cosylab.gui.displayers.UpdateRequest)
	 */
	public void updateAcynchronously(UpdateRequest request)
		throws CommonException
	{
		try {
			Object o = request.getNewTargetValue();

			if (o instanceof Number) {
				updateRequest(request.getTimestamp(), (Number)o);
			} else if (o instanceof String) {
				updateRequest(request.getTimestamp(), (String)o);
			} else {
				updateRequest(request.getTimestamp(), o);
			}

			request.reply(o, true, null);
		} catch (Exception e) {
			request.reply(request.getNewTargetValue(), false, e);
		}
	}

	/**
	 * Forwards call to <code>updateRequest()</code>.
	 *
	 * @see com.cosylab.gui.displayers.SynchronousNumberConsumer#updateSynchronously(long,
	 *      java.lang.Number)
	 */
	public void updateSynchronously(long timestamp, Number value)
		throws CommonException
	{
		updateRequest(timestamp, value);
	}

	/**
	 * Forwards call to <code>updateRequest()</code>.
	 *
	 * @see com.cosylab.gui.displayers.SynchronousObjectConsumer#updateSynchronously(long,
	 *      java.lang.Object)
	 */
	public void updateSynchronously(long timestamp, Object value)
		throws CommonException
	{
		updateRequest(timestamp, value);
	}

	/**
	 * Forwards call to <code>updateRequest()</code>.
	 *
	 * @see com.cosylab.gui.displayers.SynchronousStringConsumer#updateSynchronously(long,
	 *      java.lang.String)
	 */
	public void updateSynchronously(long timestamp, String value)
		throws CommonException
	{
		updateRequest(timestamp, value);
	}

	/**
	 * Forwards call to <code>updateRequest()</code>.
	 *
	 * @see com.cosylab.gui.displayers.NonblockingNumberConsumer#updateNonblocking(java.lang.Number)
	 */
	public void updateNonblocking(Number value)
	{
		updateRequest(System.currentTimeMillis(), value);
	}

	/**
	 * Forwards call to <code>updateRequest()</code>.
	 *
	 * @see com.cosylab.gui.displayers.NonblockingObjectConsumer#updateNonblocking(java.lang.Object)
	 */
	public void updateNonblocking(Object value)
	{
		updateRequest(System.currentTimeMillis(), value);
	}

	/**
	 * Forwards call to <code>updateRequest()</code>.
	 *
	 * @see com.cosylab.gui.displayers.NonblockingStringConsumer#updateNonblocking(java.lang.String)
	 */
	public void updateNonblocking(String value)
	{
		updateRequest(System.currentTimeMillis(), value);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getDataConsumer(java.lang.Class)
	 */
	public DataConsumer getDataConsumer(Class type)
	{
		if (type.isAssignableFrom(this.getClass())) {
			return this;
		}

		return null;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getDefaultDataConsumer()
	 */
	public DataConsumer getDefaultDataConsumer()
	{
		return this;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#updateDataState(com.cosylab.gui.displayers.DataState)
	 */
	public void updateDataState(DataState state)
	{
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#setCharacteristics(java.util.Map)
	 */
	public void setCharacteristics(Map characteristics)
	{
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getName()
	 */
	public String getName()
	{
		return name;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getSupportedCharacteristics()
	 */
	public String[] getSupportedCharacteristics()
	{
		return null;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getSupportedConsumerTypes()
	 */
	public Class[] getSupportedConsumerTypes()
	{
		return supported;
	}

	/**
	 * This method is called whenever any of number updates is called.
	 *
	 * @param timestamp
	 * @param value
	 */
	protected void updateRequest(long timestamp, Number value)
	{
	}

	/**
	 * This method is called whenever any of Object updates is called.
	 *
	 * @param timestamp
	 * @param value
	 */
	protected void updateRequest(long timestamp, Object value)
	{
	}

	/**
	 * This method is called whenever any of String updates is called.
	 *
	 * @param timestamp
	 * @param value
	 */
	protected void updateRequest(long timestamp, String value)
	{
	}
}

/* __oOo__ */
