/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components;

import java.util.ArrayList;


/**
 * Convenience implementation of <code>ProgressMonitor</code>. Contains
 * methods for firing events.
 *
 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
 * @version $id$
 */
public abstract class AbstractProgressTask extends Thread
	implements ProgressMonitor
{
	private final ArrayList<ProgressListener> listeners = new ArrayList<>();

	/**
	 * AbstractProgressTask constructor comment.
	 */
	public AbstractProgressTask()
	{
		super();
	}

	/**
	 * Adds listener. Listeners will be notified when task reports progress.
	 *
	 * @param listener to add.
	 */
	public void addProgressListener(ProgressListener listener)
	{
		listeners.add(listener);
	}

	/**
	 * Notifies listeners that current progress has changed. There is no
	 * prescribed frequency with which these events should be fired.
	 *
	 * @param status Description of current status.
	 * @param current progress, should be less than total
	 * @param total ammount.
	 */
	protected void fireProgressEvent(String status, int current, int total)
	{
		yield();

		ProgressEvent e = new ProgressEvent(this, status, current, total);

		for (int i = 0; i < listeners.size(); i++) {
			listeners.get(i).progress(e);
		}

		yield();
	}

	/**
	 * Notifies listeners that a task has been completed.
	 */
	protected void fireTaskComplete()
	{
		yield();

		ProgressEvent e = new ProgressEvent(this, getStatus(), getCurrent(),
			    getTotal());

		for (int i = 0; i < listeners.size(); i++) {
			listeners.get(i).taskComplete(e);
		}

		yield();
	}

	/**
	 * Notifies listeners that a task has been interrupted during execution.
	 *
	 * @param reason Description of reason for interruption.
	 */
	protected void fireTaskInterrupted(String reason)
	{
		yield();

		ProgressEvent e = new ProgressEvent(this, reason, 0, -1);

		for (int i = 0; i < listeners.size(); i++) {
			listeners.get(i).taskInterruped(e);
		}

		yield();
	}

	/**
	 * Notify listeners that task has started execution.
	 */
	protected void fireTaskStarted()
	{
		yield();

		ProgressEvent e = new ProgressEvent(this, getStatus(), getCurrent(),
			    getTotal());

		for (int i = 0; i < listeners.size(); i++) {
			listeners.get(i).taskStarted(e);
		}

		yield();
	}

	/**
	 * Force listeners to update immediately.
	 */
	protected void immediateUpdate()
	{
		fireProgressEvent(getStatus(), getCurrent(), getTotal());
	}

	/**
	 * Removes listener.
	 *
	 * @param listener to remove.
	 */
	public void removeProgressListener(ProgressListener listener)
	{
		listeners.remove(listener);
	}
}

/* __oOo__ */
