/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import javax.swing.JComponent;
import javax.swing.JFrame;

import com.cosylab.application.state.State;
import com.cosylab.events.SetListener;
import com.cosylab.gui.components.customizer.AbstractCustomizerPanel;
import com.cosylab.gui.components.util.PopupManageable;
import com.cosylab.gui.components.wheelswitch.AbstractWheelswitchFormatter;
import com.cosylab.gui.components.wheelswitch.WheelswitchFormatter;


/**
 * <code>LabelledWheelswitch</code> is extended <code>Wheelswitch</code> with
 * title label and unit label in same style.
 *
 * @author <a href="mailto:igor.kriznar@cosylab.com">Igor Kriznar</a>
 * @author <a href="mailto:jernej.kamenik@cosylab.com">Jernej Kamenik</a>
 * @version $id$
 */
public class LabelledWheelswitch extends AbstractNumericDisplayerPanel implements PopupManageable
{
	private static final long serialVersionUID = 1L;
	/** Value property name. */
	public static final String VALUE = Wheelswitch.VALUE;
	private Wheelswitch number;
	private AbstractCustomizerPanel customizer;
	/**
	 * Creates a new LabelledWheelswitch object.
	 */
	public LabelledWheelswitch()
	{
		this(0.0, "3.2f", null, null);
	}

	/**
	 * Creates a new LabelledWheelswitch object.
	 *
	 * @param value initial double value
	 */
	public LabelledWheelswitch(double value)
	{
		this(value, "3.2f" , null, null);
	}

	/**
	 * Creates a new LabelledWheelswitch object.
	 *
	 * @param format initial number (C-style) format
	 */
	public LabelledWheelswitch(String format)
	{
		this(0.0, format, null, null);
	}

	/**
	 * Creates a new LabelledWheelswitch object.
	 *
	 * @param value initial double value
	 * @param format initial number (C-style) format
	 * @param title initial title
	 * @param units initial units
	 */
	public LabelledWheelswitch(double value, String format, String title,
	    String units)
	{
		super();

		setValue(value);
		setFormat(format);
		setUnits(units);
		setTitle(title);

		setUnitsVisible(true);
		setTitleVisible(true);
		setLayoutType(Layout.HORIZONTAL);
		setResizable(true);
		setEnhanced(true);
		addMouseListener(getPopupManager().getMouseHook());	
		//Ike's FR by jbobnar RT#41405
		getTitleComponent().addMouseListener(getWheelswitch().handler);
	}
		
	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.components.AbstractDisplayerPanel#getCustomizer()
	 */
	@Override
	public AbstractCustomizerPanel getCustomizer()
	{
		if (customizer == null) {
			customizer = AbstractCustomizerPanel.findCustomizer(this);
		}

		return customizer;
	}
			

	/**
	 * Sets the editability of the wheelswitch.
	 *
	 * @param newEditable
	 */
	public void setEditable(boolean newEditable)
	{
		if (newEditable == isEditable()) return;
		getWheelswitch().setEditable(newEditable);
		//has to fire event EVERY time!
		firePropertyChange("editable",!newEditable,newEditable);
	}

	/**
	 * Returns whether the wheelswitch can be edited  by the user.
	 *
	 * @return boolean
	 */
	public boolean isEditable()
	{
		return getWheelswitch().isEditable();
	}

	/**
	 * Sets the maximum allowed value.
	 *
	 * @param newValue
	 *
	 * @see WheelswitchFormatter#setMaximum(double)
	 */
	public void setMaximum(double newValue)
	{
		double oldValue = getMaximum();
		if (newValue == oldValue) return;
		getWheelswitch().setGraphMax(newValue);
		super.setMaximumValue(new Double(getWheelswitch().getGraphMax()));
		newValue = getMaximum();
		firePropertyChange("maximum", oldValue, newValue);
	}

	/**
	 * Returns the maximum alowed value.
	 *
	 * @return double
	 *
	 * @see WheelswitchFormatter#getMaximum()
	 */
	public double getMaximum()
	{
		return getWheelswitch().getGraphMax();
	}

	/**
	 * Sets the minimum allowed value.
	 *
	 * @param newValue
	 *
	 * @see WheelswitchFormatter#setMinimum(double)
	 */
	public void setMinimum(double newValue)
	{
		double oldValue = getMinimum();
		if (newValue == oldValue) return;
		getWheelswitch().setGraphMin(newValue);
		super.setMinimumValue(new Double(getWheelswitch().getGraphMin()));
		newValue = getMinimum();
		firePropertyChange("minimum", oldValue, newValue);
	}

	/**
	 * Returns the minimum alowed value.
	 *
	 * @return double
	 *
	 * @see WheelswitchFormatter#getMaximum()
	 */
	public double getMinimum()
	{
		return getWheelswitch().getGraphMin();
	}

	/**
	 * Sets the maximum and minimum allowed values.
	 *
	 * @param max
	 * @param min
	 *
	 * @see com.cosylab.gui.components.Wheelswitch#setMaxMin(double, double)
	 */
	public void setMaxMin(double max, double min)
	{
		setMaximum(max);
		setMinimum(min);
	}

	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.components.AbstractNumericDisplayerPanel#setState(com.cosylab.application.state.State)
	 */
	@Override
	public void setState(State state)
	{
		super.setState(state);

		setTiltingEnabled(state.getBoolean("tiltingEnabled", isTiltingEnabled()));
		setEditable(state.getBoolean("editable", isEditable()));
		setMaximum(state.getDouble("maximum", getMaximum()));
		setMinimum(state.getDouble("minimum", getMinimum()));
	}

	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.components.AbstractNumericDisplayerPanel#getState()
	 */
	@Override
	public State getState()
	{
		State state = super.getState();

		state.putDouble("maximum", getMaximum());
		state.putDouble("minimum", getMinimum());
		state.putBoolean("editable", isEditable());
		state.putBoolean("tiltingEnabled", isTiltingEnabled());

		return state;
	}

	/**
	 * Enables/disables tilting of the displayer. Displayer tilts when out of bounds.
	 *
	 * @param b true if tilting is enabled
	 * @see Wheelswitch#setTiltingEnabled(boolean);
	 */
	public void setTiltingEnabled(boolean b)
	{	
		if (isTiltingEnabled() == b) return; 
		getWheelswitch().setTiltingEnabled(b);
		firePropertyChange("tiltingEnabled", !b, b);
	}

	/**
	 * Returns true if tilting is enabled.
	 *
	 * @return true if tilting is enabled
	 */
	public boolean isTiltingEnabled()
	{
		return getWheelswitch().isTiltingEnabled();
	}

	/**
	 * Sets the value and displays it in the wheelswitch. The method may  also
	 * change the current digit selection if neccessary in order to point to
	 * the same decimal digit of the displayed value.
	 *
	 * @param newValue
	 *
	 * @see WheelswitchFormatter#setValue(double)
	 */
	public void setValue(double newValue)
	{	
		double oldValue = getValue();
		getWheelswitch().setValue(newValue);
		firePropertyChange("value", oldValue, newValue);
	}
	
	/**
	 * Sets the animated property.
	 * 
	 * @param animated
	 * @see Wheelswitch#setAnimated(boolean)
	 */
	public void setAnimated(boolean animated) {
		if (isAnimated() == animated) return;
		getWheelswitch().setAnimated(animated);
		firePropertyChange("animated", !animated, animated);
	}
	
	/**
	 * Returns weather the labelled wheelswitch is animated.
	 * 
	 * @return animated property
	 * @see Wheelswitch#isAnimated()
	 */
	public boolean isAnimated() {
		return getWheelswitch().isAnimated();
	}

	/**
	 * Returns the value displayed by the <code>Wheelswitch</code> and stored by
	 * the <code>formatter</code>.
	 *
	 * @return double
	 *
	 * @see WheelswitchFormatter#getValue()
	 */
	public double getValue()
	{
		return getWheelswitch().getValue();
	}

	/**
	 * Adds set listener to the <code>NumberField</code>.
	 *
	 * @see NumberField#addSetListener(SetListener)
	 */
	public void addSetListener(SetListener l)
	{
		getWheelswitch().addSetListener(l);
	}

	/**
	 * Removes set listener from the <code>NumberField</code>.
	 *
	 * @see NumberField#removeSetListener(SetListener)
	 */
	public void removeSetListener(SetListener l)
	{
		getWheelswitch().removeSetListener(l);
	}

	protected Wheelswitch getWheelswitch()
	{
		if (number == null) {
			number = new Wheelswitch();
			
			//			number.addPropertyChangeListener(new PropertyChangeListener() {
			//					/**
			//					 * @see java.beans.PropertyChangeListener#propertyChange(PropertyChangeEvent)
			//					 */
			//					public void propertyChange(PropertyChangeEvent evt)
			//					{
			//						firePropertyChange(evt.getPropertyName(),
			//						    evt.getOldValue(), evt.getNewValue());
			//					}
			//				});
			number.addMouseListener(getPopupManager().getMouseHook());
			number.setUnitSeparate(true);
			number.setPopupEnabled(false);
			//Ike's FR by jbobnar RT#41405
			addMouseListener(number.handler);
			getUnitsComponent().addMouseListener(number.handler);
		}

		return number;
	}

	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.components.AbstractDisplayerPanel#getValueComponent()
	 */
	@Override
	protected JComponent getValueComponent()
	{
		return getWheelswitch();
	}
	
	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.components.AbstractNumericDisplayerPanel#internalSetEnabled()
	 */
	@Override
	protected void internalSetEnabled()
	{
		super.internalSetEnabled();
		getWheelswitch().setEnabled(isEnabled());
	}

	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.components.AbstractNumericDisplayerPanel#internalSetEnhanced()
	 */
	@Override
	protected void internalSetEnhanced()
	{
		super.internalSetEnhanced();
		getWheelswitch().setEnhanced(isEnhanced());
	}

	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.components.AbstractNumericDisplayerPanel#internalSetFormat()
	 */
	@Override
	protected void internalSetFormat()
	{
		super.internalSetFormat();

		String tFormat = getFormat();

		if (tFormat == null) {
			tFormat = NumberField.createDefaultFormat(Double.class);
		}

		getWheelswitch().setFormat(WheelswitchFormatter.transformFormat(tFormat));
	}

	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.components.AbstractNumericDisplayerPanel#internalSetResizable()
	 */
	@Override
	protected void internalSetResizable()
	{
		super.internalSetResizable();
		getWheelswitch().revalidate();
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.components.AbstractNumericDisplayerPanel#setUnitsVisible(boolean)
	 */
	@Override
	public void setUnitsVisible(boolean b) {
		if (b) getWheelswitch().setUnit(getUnits()); 
		else getWheelswitch().setUnit(null);
		super.setUnitsVisible(b);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.components.AbstractNumericDisplayerPanel#setUnits(java.lang.String)
	 */
	@Override
	public void setUnits(String value) {
		if (isUnitsVisible()) {
			getWheelswitch().setUnit(value);
		}
		super.setUnits(value);
	}
	/* (non-Javadoc)
	 * @see com.cosylab.gui.components.AbstractNumericDisplayerPanel#layoutValueAndTitleAndUnits()
	 */
	@Override
	protected void layoutValueAndTitleAndUnits() {
		super.layoutValueAndTitle();
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.components.AbstractNumericDisplayerPanel#layoutValueAndTitleAndUnitsAndBounds()
	 */
	@Override
	protected void layoutValueAndTitleAndUnitsAndBounds() {
		super.layoutValueAndTitleAndBounds();
	}
	
	/* (non-Javadoc)
	 * @see com.cosylab.gui.components.AbstractNumericDisplayerPanel#layoutValueAndUnits()
	 */
	@Override
	protected void layoutValueAndUnits() {
		super.layoutValue();
	}
	
	/* (non-Javadoc)
	 * @see com.cosylab.gui.components.AbstractNumericDisplayerPanel#layoutValueAndUnitsAndBounds()
	 */
	@Override
	protected void layoutValueAndUnitsAndBounds() {
		super.layoutValueAndBounds();
	}
	
	/* (non-Javadoc)
	 * @see javax.swing.JComponent#setBackground(java.awt.Color)
	 */
	@Override
	public void setBackground(Color bg) {
		getWheelswitch().setBackground(bg);
		super.setBackground(bg);
	}
	/**
	 * Set the formatter fot the value.
	 * 
	 * @param formatter new formatter
	 * @see Wheelswitch#setFormatter(AbstractWheelswitchFormatter)
	 */
	public void setFormatter(AbstractWheelswitchFormatter formatter) {
		AbstractWheelswitchFormatter oldValue = getFormatter();
		getWheelswitch().setFormatter(formatter);
		firePropertyChange("formatter", oldValue, formatter);
	}
	/**
	 * Returns the formatter employed by the displayer.
	 * 
	 * @see Wheelswitch#getFormatter()
	 */
	public AbstractWheelswitchFormatter getFormatter() {
		return getWheelswitch().getFormatter();
	}
	
	/* (non-Javadoc)
     * @see com.cosylab.gui.components.AbstractNumericDisplayerPanel#setTitleMaximumFontSize(int)
     */
	@Override
    public void setTitleMaximumFontSize(int max) {
        getWheelswitch().setUnitsMaximumFontSize(max);
        super.setTitleMaximumFontSize(max);
    }
    
    /* (non-Javadoc)
     * @see com.cosylab.gui.components.AbstractNumericDisplayerPanel#setTitleMinimumFontSize(int)
     */
	@Override
    public void setTitleMinimumFontSize(int min) {
        getWheelswitch().setUnitsMinimumFontSize(min);        
        super.setTitleMinimumFontSize(min);
    }
    
    /**
     * Set <code>true</code> if digits should take up all space, set 
     * <code>false</code> if place for not visible digits should be reserved
     * (there is no resizing of digits and labels when 9 changes to 10 etc.).
     * 
     * @param bool boolean value
     */
    public void setDigitsTakeUpAllSpace(boolean bool) {
    	if(getWheelswitch().getDigitsTakeUpAllSpace() == bool)
    		return;
    	
    	firePropertyChange("digitsTakeUpAllSpace", getWheelswitch().getDigitsTakeUpAllSpace(), bool);
    	
    	getWheelswitch().setDigitsTakeUpAllSpace(bool);
    }
    
    /**
     * <code>true</code> if digits take up all space,
     *  otherwise <code>false</code>
     * @return
     */
    public boolean getDigitsTakeUpAllSpace() {
    	return getWheelswitch().getDigitsTakeUpAllSpace();
    }
    
    /**
     * If {@link #setDigitsTakeUpAllSpace(boolean)} is set to <code>false</code>
     * the place for <code>numberOfAllDigits</code> is reserved. Not needed 
     * additional digits are not shown.
     * 
     * @param numberOfAllDigits number of all digits
     */
    public void setNumberOfAllDigits(int numberOfAllDigits) {
    	if(getWheelswitch().getNumberOfAllDigits() == numberOfAllDigits)
    		return;
    	
    	firePropertyChange("numberOfAllDigits", getWheelswitch().getNumberOfAllDigits(), numberOfAllDigits);
    	getWheelswitch().setNumberOfAllDigits(numberOfAllDigits);
    }
    
    /**
     * @return The number of reserved places for digits.
     */
    public int getNumberOfAllDigits() {
    	return(getWheelswitch().getNumberOfAllDigits());
    }
    
    /*
     * (non-Javadoc)
     * @see com.cosylab.gui.components.AbstractDisplayerPanel#setPopupEnabled(boolean)
     */
    @Override
    public void setPopupEnabled(boolean enabled) {
    	super.setPopupEnabled(enabled);
    	getWheelswitch().setPopupEnabled(enabled);
    }
    
    /**
     * Sets the digit that will be selected whenever wheelswitch determines that it needs
     * to automatically select one of the digits. If the value is -1, it will determine the
     * most appropriate digit automatically. 
     * 
     * @param forcedSelectedDigit the digit that will be forcefully selected
     */
	public void setForcedSelectedDigit(int forcedSelectedDigit) {
		if (getForcedSelectedDigit() == forcedSelectedDigit) return;
		int oldValue = getForcedSelectedDigit();
		getWheelswitch().setForcedSelectedDigit(forcedSelectedDigit);
		firePropertyChange("forcedSelectedDigit",oldValue,forcedSelectedDigit);
	}
	
	/**
	 * @return the digit that will be forcefully selected whenever the wheelswitch gains
	 * 			focus or in another similar situation that requires the wheelswitch to automatically
	 * 			select the digit
	 */
	public int getForcedSelectedDigit() {
		return getWheelswitch().getForcedSelectedDigit();
	}
	
	/**
	 * Sets the value precision, which is the minimal change that can occur of the wheelswitch.
	 * If set to Double.NaN, there are no restrictions. If the precision is set to 0.2, whenever
	 * a button will be pressed to increase or decrease the value, the value will change for at 
	 * least 0.2.
	 * 
	 * @param valuePrecision the value precision.
	 */
	public void setValuePrecision(double valuePrecision) {
		double oldValue = getValuePrecision();
		if (oldValue == valuePrecision) return;
		getWheelswitch().setValuePrecision(valuePrecision);
		firePropertyChange("valuePrecision",oldValue,valuePrecision);
	}
	
	/**
	 * @return the value precision, which is the minimal value change that is allowed to occur
	 */
	public double getValuePrecision() {
		return getWheelswitch().getValuePrecision();
	}

    /**
	 * Runs simple test applet.
	 *
	 * @param args String[]
	 */
	public static void main(String[] args)
	{
		LabelledWheelswitch ws = new LabelledWheelswitch(2.449, "%2.1f",
			    "Current", "A");
		ws.setDigitsTakeUpAllSpace(false);
		ws.setNumberOfAllDigits(5);
		JFrame frame = new JFrame("Wheelwsitch Testing Applet");
		frame.getContentPane().setLayout(new BorderLayout(10, 10));
		frame.getContentPane().add(ws);
		frame.setSize(300, 100);
		frame.addWindowListener(new WindowAdapter() {
			@Override
				public void windowClosing(WindowEvent e)
				{
					System.exit(0);
				}
			});
		frame.setVisible(true);
	}
}

/* __oOo__ */
