/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components;

import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.JLabel;
import javax.swing.JProgressBar;
import javax.swing.Timer;

import com.cosylab.application.AutoScalable;


/**
 * Simple progress bar dialog for displaying <code>ProgressMonitor</code>
 * enabled tasks. Implementation takes care to prevent GUI freezes during
 * extensive tasks.
 *
 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
 * @version $id$
 */
@AutoScalable
public class ProgressDialog extends javax.swing.JDialog
{
	private static final long serialVersionUID = 1L;
	private JLabel statusLabel;
	private JProgressBar progressBar;
	private AbstractProgressTask task;
	private Timer timer;

	/**
	 * Timer that will update GUI twice a second if regardless of events
	 * generated from task. This ensures consistent GUI updates even if task
	 * does not report progress frequently.
	 */
	protected class TimerListener implements ActionListener
	{
		/**
		 * Updates GUI.
		 *
		 * @param e event.
		 */
		public void actionPerformed(ActionEvent e)
		{
			updateState();
		}
	}

	/**
	 * Listens for status updates from thread implementing ProgressMonitor and
	 * updates interface accordingly.
	 */
	protected class StateListener implements ProgressListener
	{
		/**
		 * Updates GUI with new progress value.
		 *
		 * @param e event.
		 */
		public void progress(ProgressEvent e)
		{
			updateState();
		}

		/**
		 * Hides the dialog.
		 *
		 * @param e event.
		 */
		public void taskComplete(ProgressEvent e)
		{
			updateState();
			setVisible(false);
		}

		/**
		 * Hides the dialog.
		 *
		 * @param e event.
		 */
		public void taskInterruped(ProgressEvent e)
		{
			setStatus(e.getStatus());
			setVisible(false);
		}

		/**
		 * No action, since dialog is expected to be already visible.
		 *
		 * @param e event.
		 */
		public void taskStarted(ProgressEvent e)
		{
			//			show();
		}
	}

	/**
	 * ProgressDialog constructor comment.
	 *
	 * @param task this dialog will monitor.
	 */
	public ProgressDialog(AbstractProgressTask task)
	{
		super();
		this.task = task;
		initialize();
	}

	private GridBagConstraints createConstraints(int x, int y)
	{
		GridBagConstraints gc = new GridBagConstraints();
		gc.gridx = x;
		gc.gridy = y;
		gc.fill = GridBagConstraints.HORIZONTAL;
		gc.weightx = 1.0;
		gc.weighty = 0.0;
		gc.anchor = GridBagConstraints.WEST;
		gc.insets = new java.awt.Insets(4, 4, 4, 4);

		return gc;
	}

	private void initialize()
	{
		setSize(350, 100);
		getContentPane().setLayout(new GridBagLayout());

		statusLabel = new JLabel();
		getContentPane().add(statusLabel, createConstraints(0, 0));

		progressBar = new JProgressBar();
		getContentPane().add(progressBar, createConstraints(0, 1));
	}

	/**
	 * Starts the task associated with this component.
	 */
	public void performTask()
	{
		TimerListener tl = new TimerListener();
		ProgressListener pl = new StateListener();

		timer = new Timer(500, tl);
		task.addProgressListener(pl);

		task.setPriority(Thread.MIN_PRIORITY);
		task.start();
		timer.start();

		setVisible(true);
	}

	protected void setProgress(int value)
	{
		progressBar.setValue(value);
	}

	protected void setProgressBounds(int min, int max)
	{
		if (min >= max) {
			return;
		}

		progressBar.setMinimum(min);
		progressBar.setMaximum(max);
		progressBar.setValue(min);
		progressBar.repaint();
	}

	protected void setStatus(String status)
	{
		statusLabel.setText(status);
	}

	protected void updateState()
	{
		ProgressDialog owner = ProgressDialog.this;

		int current = task.getCurrent();
		int total = task.getTotal();

		owner.progressBar.setVisible(total > -1);

		owner.setProgressBounds(0, total);
		owner.setProgress(current);

		setStatus(owner.task.getStatus());
	}
}

/* __oOo__ */
