/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Graphics;
import java.awt.Insets;
import java.awt.LayoutManager;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.util.Vector;

import javax.swing.CellRendererPane;
import javax.swing.ComboBoxModel;
import javax.swing.DefaultComboBoxModel;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.ListCellRenderer;
import javax.swing.plaf.ComboBoxUI;
import javax.swing.plaf.basic.BasicComboBoxRenderer;
import javax.swing.plaf.basic.BasicComboBoxUI;
import javax.swing.plaf.basic.BasicComboPopup;
import javax.swing.plaf.metal.MetalBorders;

import com.cosylab.gui.components.util.ColorHelper;


/**
 * A simple implementation of a combo box that provides several display styles
 * through different javax.swing.plaf.ComboBoxUI implementations. These can be
 * set through setStyle(int) method and are enumerated as: SIMPLE_STYLE
 * displays no button in the comboBox. Items are selected from a drop down
 * list activated on mouseclick. DROP_DOWN_STYLE displays an arrowed button on
 * the right side of the combo box that activates the drop down list.
 * SWITCH_STYLE_POPUP_ENABLED positions two buttons on each side of the combo
 * box that can be used to switch between items of the combo box, while the
 * complete list of items can still be accesed by mouseclick on the selected
 * item cell. SWITCH_STYLE_POPUP_DISABLED has the same features as
 * SWITCH_STYLE_POPUP_ENABLED but the drop-down list is no longer available to
 * the user. The method setUI(javax.swing.plaf.ComboBoxUI) has been overrriden
 * to accept only SimpleComboBoxUI instances. In addition SimpleComboBox
 * accepts custom renderer to be set to rendere only the selected item cell
 * differently from all the other items in the drop-down list. This renderer
 * can be set through setSelectedItemRenderer(ListCellRenderer). If null, the
 * default renderer is used.
 *
 * @author <a href="mailto:jernej.kamenik@cosylab.com">Jernej Kamenik</a>
 * @version $id$
 */
public class SimpleComboBox extends JComboBox
{

    private static final long serialVersionUID = -7775055879270581688L;

	/** @deprecated 
	 *  @use {@link Style#SIMPLE_STYLE} 
	 */
	@Deprecated
    public static int SIMPLE_STYLE = 0;

	/** @deprecated
	 *  @use {@link Style#DROP_DOWN_STYLE} 
	 */
	@Deprecated
    public static int DROP_DOWN_STYLE = 3;

	/** @deprecated
	 *  @use {@link Style#SWITCH_STYLE_POPUP_ENABLED} 
	 */
	@Deprecated
    public static int SWITCH_STYLE_POPUP_ENABLED = 1;

	/** @deprecated
	 *  @use {@link Style#SWITCH_STYLE_POPUP_DISABLED} 
	 */
	@Deprecated
    public static int SWITCH_STYLE_POPUP_DISABLED = 2;

    public enum Style {
    	SIMPLE_STYLE,
    	SWITCH_STYLE_POPUP_ENABLED,
    	SWITCH_STYLE_POPUP_DISABLED,
    	DROP_DOWN_STYLE
    }

	private ListCellRenderer selectedItemRenderer = null;
	private int style = SIMPLE_STYLE;

    private String HIDE_POPUP_KEY;

    private Style styleE;

	/**
	 * Constructor for SimpleComboBox.
	 *
	 * @param aModel
	 *
	 * @see javax.swing.JComboBox#JComboBox(ComboBoxModel)
	 */
	public SimpleComboBox(ComboBoxModel aModel)
	{
		super(aModel);
		init();
	}

	/**
	 * Constructor for SimpleComboBox.
	 *
	 * @param items
	 *
	 * @see javax.swing.JComboBox#JComboBox(Object[])
	 */
	public SimpleComboBox(Object[] items)
	{
		super(items);
		init();
	}

	/**
	 * Constructor for SimpleComboBox.
	 *
	 * @param items
	 *
	 * @see javax.swing.JComboBox#JComboBox(Vector)
	 */
	public SimpleComboBox(Vector items)
	{
		super(items);
		init();
	}

	/**
	 * Constructor for SimpleComboBox.
	 *
	 * @see javax.swing.JComboBox#JComboBox()
	 */
	public SimpleComboBox()
	{
		super();
		init();
	}

	private void init()
	{
	    //stupid thing needed to get the BasicComboBoxUI.HIDE_POPUP_KEY reference
        super.setUI(new BasicComboBoxUI());
        HIDE_POPUP_KEY = String.valueOf(getClientProperty("doNotCancelPopup"));
        
		setUI(new SimpleComboBoxUI());
		setBorder(new MetalBorders.Flush3DBorder());
	}

	/**
	 * This method has been overriden to enforce usage of SimpleComboBoxUI.
	 *
	 * @param ui ComboBoxUI
	 *
	 * @see javax.swing.JComboBox#setUI(ComboBoxUI)
	 */
	@Override
    public void setUI(ComboBoxUI ui)
	{
	    
		if (ui instanceof SimpleComboBoxUI) {
			super.setUI(ui);
		}
	}

	/**
	 * Returns the selectedItemRenderer.
	 *
	 * @return ListCellRenderer
	 */
	public ListCellRenderer getSelectedItemRenderer()
	{
		return selectedItemRenderer;
	}
	
	@Override
	public void setEnabled(boolean b) {
		super.setEnabled(b);
		ComboBoxUI ui = getUI();
		if (ui instanceof SwitchComboBoxUI) {
			((SwitchComboBoxUI)ui).getLeftButton().setEnabled(b);
			((SwitchComboBoxUI)ui).getRightButton().setEnabled(b);
		} else if (ui instanceof DropComboBoxUI) {
			((DropComboBoxUI)ui).getDropButton().setEnabled(b);
		}
	}

	/**
	 * Sets the selectedItemRenderer.
	 *
	 * @param newSelectedItemRenderer The selectedItemRenderer to set. If null,
	 *        the selected item is rendered with the common list cell renderer
	 *        of the combo box.
	 *
	 * @see #setRenderer(ListCellRenderer)
	 */
	public void setSelectedItemRenderer(
	    ListCellRenderer newSelectedItemRenderer)
	{
		ListCellRenderer oldSelectedItemRenderer = selectedItemRenderer;
		selectedItemRenderer = newSelectedItemRenderer;
		firePropertyChange("selectedItemRenderer", oldSelectedItemRenderer,
		    newSelectedItemRenderer);
	}

	/**
	 * Returns the style.
	 *
	 * @return int
	 */
	public int getStyle()
	{
		return style;
	}

    public void setStyle(Style newStyleE) {
        if (styleE == newStyleE) {
            return;
        }

        if (newStyleE == Style.SIMPLE_STYLE) {
            setUI(new SimpleComboBoxUI());
        } else if (newStyleE == Style.DROP_DOWN_STYLE) {
            setUI(new DropComboBoxUI());
        } else if (newStyleE == Style.SWITCH_STYLE_POPUP_ENABLED) {
            SwitchComboBoxUI ui = new SwitchComboBoxUI();
            setUI(ui);
            ui.setPopupEnabled(true);
        } else if (newStyleE == Style.SWITCH_STYLE_POPUP_DISABLED) {
            SwitchComboBoxUI ui = new SwitchComboBoxUI();
            setUI(ui);
            ui.setPopupEnabled(false);
        } else {
            throw (new IllegalArgumentException());
        }

        Style oldStyleE = styleE;
        styleE = newStyleE;
        firePropertyChange("style", oldStyleE, newStyleE);
    }
    
	/**
	 * Sets the style. SIMPLE_STYLE only displays the selected item. When
	 * clicked the drop down list dialog shows up. SWITCH_STYLE_POPUP_DISABLED
	 * displays the selected item together with two arrowed buttons on each
	 * style to switch between adjacent items in the list. No popup is
	 * displayed. SWITCH_STYLE_POPUP_ENABLED is the same as
	 * SWITCH_STYLE_POPUP_DISABLED only that at mouse click the drop down list
	 * of items is displayed. DROP_DOWN_STYLE displays the SimpleComboBox in
	 * conventional way. The selected item is displyed next to a downwards
	 * pointing arrow button which brings up the drop down list of all items.
	 *
     * @deprecated use {@link #setStyle(Style)}
	 * @param newStyle int the style to be used by this SimpleComboBox
	 */
	@Deprecated
    public void setStyle(int newStyle)
	{
        setStyle(Style.values()[newStyle]);
	}

	/**
	 * For testing and demonstration purposes only.
	 *
	 * @param args String[]
	 */
	public static void main(String[] args)
	{
		JFrame frame = new JFrame("SimpleComboBox Demo");
		final SimpleComboBox box = new SimpleComboBox();
		DefaultComboBoxModel model = new DefaultComboBoxModel();

		for (int i = 1; i <= 50; i++) {
			model.addElement("item " + i);
		}

		box.setModel(model);

		final JComboBox box1 = new JComboBox();
		DefaultComboBoxModel model1 = new DefaultComboBoxModel();
        for (Style st : Style.values()) {
            model1.addElement(st);            
        }
		box1.setModel(model1);
		box1.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e)
				{
					box.setStyle(((Style)box1.getSelectedItem()));
				}
			});

		frame.getContentPane().setLayout(new FlowLayout());
		frame.getContentPane().add(box);
		frame.getContentPane().add(box1);
		frame.setSize(500, 200);
		frame.addWindowListener(new WindowAdapter() {
				@Override
                public void windowClosing(WindowEvent e)
				{
					System.exit(0);
				}
			});
		frame.setVisible(true);
	}

	/**
	 * The simplest implementation od ComboBoxUI displays no button in the
	 * comboBox. Items are selected from a drop down list activated on mouse
	 * click.
	 *
	 * @author <a href="mailto:jernej.kamenik@cosylab.com">Jernej Kamenik</a>
	 * @version $id$
	 */
	public static class SimpleComboBoxUI extends ComboBoxUI
	{
		protected SimpleComboBox comboBox = null;
		protected BasicComboPopup popup = null;
		protected CellRendererPane cellRenderingPane = null;
		protected ItemListener itemListener;
        
		/**
		 * Constructor for SimpleComboBoxUI.
		 */
		public SimpleComboBoxUI()
		{
			super();
		}

		/**
		 * Delegates to <code>comboBox</code> implementation.
		 *
		 * @param c JComboBox must be <code>comboBox</code>
		 *
		 * @return boolean
		 *
		 * @see javax.swing.plaf.ComboBoxUI#isFocusTraversable(JComboBox)
		 */
		@Override
        public boolean isFocusTraversable(JComboBox c)
		{
			//assert (c == comboBox);

			return !comboBox.isEditable();
		}

		/**
		 * Delegates to <code>comboBox</code> implementation.
		 *
		 * @param c JComboBox must be <code>comboBox</code>
		 *
		 * @return boolean
		 *
		 * @see javax.swing.plaf.ComboBoxUI#isPopupVisible(JComboBox)
		 */
		@Override
        public boolean isPopupVisible(JComboBox c)
		{
			//assert (c == comboBox);

			if (popup == null) {
				return false;
			}

			return popup.isVisible();
		}

		/**
		 * Delegates to <code>comboBox</code> implementation.
		 *
		 * @param c JComboBox must be <code>comboBox</code>
		 * @param v boolean
		 *
		 * @see javax.swing.plaf.ComboBoxUI#setPopupVisible(JComboBox, boolean)
		 */
        @Override
        public void setPopupVisible( JComboBox c, boolean v ) {
            if ( v ) {
                popup.show();
            } else {
                popup.hide();
            }
        }

		/**
		 * Installs this UI to the supplied SimpleComboBox
		 *
		 * @param c JComponent must be an instance of SimpleComboBox
		 *
		 * @see javax.swing.plaf.ComponentUI#installUI(JComponent)
		 */
		@Override
        public void installUI(JComponent c)
		{
			//assert (c instanceof SimpleComboBox);

			comboBox = (SimpleComboBox)c;
			popup = new BasicComboPopup(comboBox) {
				private static final long serialVersionUID = -5379054665405013304L;

				/* (non-Javadoc)
				 * @see javax.swing.plaf.basic.BasicComboPopup#createKeyListener()
				 */
				@Override
                protected KeyListener createKeyListener() {
					return new KeyAdapter() {
					@Override
                    public void keyPressed(KeyEvent e) {
						if (e.getKeyCode()==KeyEvent.VK_UP || e.getKeyCode()==KeyEvent.VK_RIGHT) getList().setSelectedIndex(list.getSelectedIndex()-1);
						if (e.getKeyCode()==KeyEvent.VK_DOWN || e.getKeyCode()==KeyEvent.VK_LEFT) getList().setSelectedIndex(list.getSelectedIndex()+1);
						if (e.getKeyCode()==KeyEvent.VK_ENTER) {
							comboBox.setSelectedIndex(getList().getSelectedIndex());
							hide();
						}
						if (e.getKeyCode()==KeyEvent.VK_ESCAPE) hide();
						getList().ensureIndexIsVisible(getList().getSelectedIndex());
					}
				};
                
				}
			};
			comboBox.addKeyListener(popup.getKeyListener());
			cellRenderingPane = new CellRendererPane();
			itemListener = new ItemListener() {
						public void itemStateChanged(ItemEvent e)
						{
							comboBox.revalidate();
							comboBox.repaint();
						}
					};

			comboBox.addItemListener(itemListener);
			comboBox.addMouseListener(popup.getMouseListener());
			comboBox.addMouseMotionListener(popup.getMouseMotionListener());
			comboBox.setLayout(new ComboBoxLayoutManager());
			comboBox.add(cellRenderingPane, BorderLayout.CENTER);
			                                   
            comboBox.putClientProperty("doNotCancelPopup", comboBox.HIDE_POPUP_KEY);

        }

		/**
		 * Uninstalls this UI from the supplied SimpleComboBox
		 *
		 * @param c JComponent must be <code>comboBox</code>
		 *
		 * @see javax.swing.plaf.ComponentUI#uninstallUI(JComponent)
		 */
		@Override
        public void uninstallUI(JComponent c)
		{
			//assert (c == comboBox);
			super.uninstallUI(c);
			comboBox.removeItemListener(itemListener);
			comboBox.removeMouseListener(popup.getMouseListener());
			comboBox.removeMouseMotionListener(popup.getMouseMotionListener());
			comboBox.setLayout(null);
			comboBox.remove(cellRenderingPane);
			comboBox = null;
			popup = null;
			cellRenderingPane = null;
			itemListener = null;
		}

		/**
		 * Returns the maximum allowed layout size.
		 *
		 * @param c JComponent must be <code>comboBox</code>
		 *
		 * @return Dimension layout size.
		 *
		 * @see javax.swing.plaf.ComponentUI#getMaximumSize(JComponent)
		 */
		@Override
        public Dimension getMaximumSize(JComponent c)
		{
			//assert (c == comboBox);

			return new Dimension(Short.MAX_VALUE, Short.MAX_VALUE);
		}

		/**
		 * Returns the minimum allowed layout size.
		 *
		 * @param c JComponent must be <code>comboBox</code>
		 *
		 * @return Dimension layout size.
		 *
		 * @see javax.swing.plaf.ComponentUI#getMinimumSize(JComponent)
		 */
		@Override
        public Dimension getMinimumSize(JComponent c)
		{
			//assert (c == comboBox);

			return getPreferredSize(c);
		}

		/**
		 * Returns the preferred layout size.
		 *
		 * @param c JComponent must be <code>comboBox</code>
		 *
		 * @return Dimension layout size.
		 *
		 * @see javax.swing.plaf.ComponentUI#getPreferredSize(JComponent)
		 * @see SimpleComboBoxUI#calculatePreferredSize()
		 */
		@Override
        public Dimension getPreferredSize(JComponent c)
		{
			//assert (c == comboBox);

			return calculatePreferredSize();
		}

		/**
		 * This method has been overriden to implement correct combo box
		 * painting.
		 *
		 * @param g Graphics
		 * @param jc JComponent must be <code>comboBox</code>
		 *
		 * @see javax.swing.plaf.ComponentUI#paint(Graphics, JComponent)
		 */
		@Override
        public void paint(Graphics g, JComponent jc)
		{
			//assert (jc == comboBox);

			if (comboBox.getRenderer() == null) {
				comboBox.setRenderer(new BasicComboBoxRenderer.UIResource());
			}

			Component c = null;

			if (comboBox.getSelectedItemRenderer() == null) {
				c = comboBox.getRenderer().getListCellRendererComponent(popup
					    .getList(), comboBox.getSelectedItem(),
					    comboBox.getSelectedIndex(), true, comboBox.hasFocus());
			} else {
				c = comboBox.getSelectedItemRenderer()
					.getListCellRendererComponent(popup.getList(),
					    comboBox.getSelectedItem(),
					    comboBox.getSelectedIndex(), true, comboBox.hasFocus());
			}

			Rectangle bounds = calculateRendererCellRectangle();

			// Fix for 4238829: should lay out the JPanel.
			boolean shouldValidate = false;

			if (c instanceof JPanel) {
				shouldValidate = true;
			}

			cellRenderingPane.paintComponent(g, c, comboBox, bounds.x,
			    bounds.y, bounds.width, bounds.height, shouldValidate);
		}

		/*
		 * @return Rectangle
		 * @see com.cosylab.gui.components.SimpleComboBox.SimpleComboBoxUI#calculateRendererCellRectangle(SimpleComboBox)
		 */
		protected Rectangle calculateRendererCellRectangle()
		{
			int width = comboBox.getWidth();
			int height = comboBox.getHeight();
			Insets insets = comboBox.getInsets();

			return new Rectangle(insets.left, insets.top,
			    width - (insets.left + insets.right),
			    height - (insets.top + insets.bottom));
		}

		/*
		 * @see com.cosylab.gui.components.SimpleComboBox.SimpleComboBoxUI#layoutComboBox(SimpleComboBox)
		 */
		protected void calculateLayout()
		{
			//cellRenderingPane.setBounds(calculateRendererCellRectangle());
		}

		/*
		 * @return Dimension
		 * @see com.cosylab.gui.components.SimpleComboBox.SimpleComboBoxUI#calculatePreferredSize(SimpleComboBox)
		 */
		protected Dimension calculatePreferredSize()
		{
			if (comboBox.getRenderer() == null) {
				comboBox.setRenderer(new BasicComboBoxRenderer.UIResource());
			}

			Component c = null;
			int width = 0;
			int height = 0;

			for (int i = 0; i < popup.getList().getModel().getSize(); i++) {
				if (comboBox.getSelectedItemRenderer() == null) {
					c = comboBox.getRenderer().getListCellRendererComponent(popup
						    .getList(), comboBox.getItemAt(i), i, true,
						    comboBox.hasFocus());
				} else {
					c = comboBox.getSelectedItemRenderer()
						.getListCellRendererComponent(popup.getList(),
						    comboBox.getItemAt(i), i, true, comboBox.hasFocus());
				}

				width = Math.max(c.getPreferredSize().width, width);
				height = Math.max(c.getPreferredSize().height, height);
			}

			Insets insets = comboBox.getInsets();

            int buttonSize = ArrowButton.PREFERRED_SIZE;//height - (insets.top + insets.bottom);

            Dimension retVal = new Dimension(buttonSize+insets.left + insets.right + width,
				    insets.top + insets.bottom + height);

			return retVal;
		}

		/*
		 * Layout manager for SimpleComboBox
		     * @author Jernej Kamenik
		     * @version $id$
		     */
		protected class ComboBoxLayoutManager implements LayoutManager
		{
			/**
			 * Empty implementation.
			 *
			 * @param name String
			 * @param comp Component
			 *
			 * @see java.awt.LayoutManager#addLayoutComponent(String,
			 *      Component)
			 */
			public void addLayoutComponent(String name, Component comp)
			{
			}

			/**
			 * Empty implementation.
			 *
			 * @param comp Component
			 *
			 * @see java.awt.LayoutManager#removeLayoutComponent(Component)
			 */
			public void removeLayoutComponent(Component comp)
			{
			}

			/**
			 * Delegates to <code>parent</code> implementation.
			 *
			 * @param parent Container
			 *
			 * @return Dimension
			 *
			 * @see java.awt.LayoutManager#preferredLayoutSize(Container)
			 */
			public Dimension preferredLayoutSize(Container parent)
			{
				return parent.getPreferredSize();
			}

			/**
			 * Delegates to <code>parent</code> implementation.
			 *
			 * @param parent Container
			 *
			 * @return Dimension
			 *
			 * @see java.awt.LayoutManager#minimumLayoutSize(Container)
			 */
			public Dimension minimumLayoutSize(Container parent)
			{
				return parent.getMinimumSize();
			}

			/**
			 * Delegates to <code>parent</code> implementation.
			 *
			 * @param parent Container
			 *
			 * @see java.awt.LayoutManager#layoutContainer(Container)
			 * @see SimpleComboBoxUI#calculateLayout()
			 */
			public void layoutContainer(Container parent)
			{
				((SimpleComboBoxUI)((SimpleComboBox)parent).getUI())
				.calculateLayout();
			}
		}
	}

	/**
	 * 
	 * @author <a href="mailto:jernej.kamenik@cosylab.com">Jernej Kamenik</a>
	 * @version $id$ This extension of SimpleComboBoxUI positions two buttons on each side of the combo box to be used for switching between items in the drop-down list.
	 */
	public static class SwitchComboBoxUI extends SimpleComboBoxUI
	{
		protected ArrowButton leftButton = null;
		protected ArrowButton rightButton = null;
		private boolean popupEnabled = true;
		private boolean mustSetPopup = false;

		/*
		 * @return ArrowButton
		 */
		protected SimpleButton getLeftButton()
		{
			if (leftButton == null) {
				leftButton = new ArrowButton(ArrowButton.Orientation.LEFT);
				leftButton.addActionListener(new ActionListener() {
						public void actionPerformed(ActionEvent e)
						{
							if (comboBox.getSelectedIndex() > 0) {
								comboBox.setSelectedIndex(comboBox
								    .getSelectedIndex() - 1);
							}
						}
					});
				leftButton.setBackground(ColorHelper.getControl());
				leftButton.setBorder(null);
				leftButton.setPressedBorder(null);
                leftButton.putClientProperty("doNotCancelPopup", comboBox.HIDE_POPUP_KEY);
			}

			return leftButton;
		}

		/*
		 * @return ArrowButton
		 */
		protected SimpleButton getRightButton()
		{
			if (rightButton == null) {
				rightButton = new ArrowButton(ArrowButton.Orientation.RIGHT);
				rightButton.addActionListener(new ActionListener() {
						public void actionPerformed(ActionEvent e)
						{
							if (comboBox.getSelectedIndex() < (comboBox
							    .getItemCount() - 1)) {
								comboBox.setSelectedIndex(comboBox
								    .getSelectedIndex() + 1);
							}
						}
					});
				rightButton.setBackground(ColorHelper.getControl());
				rightButton.setBorder(null);
				rightButton.setPressedBorder(null);
                rightButton.putClientProperty("doNotCancelPopup", comboBox.HIDE_POPUP_KEY);
			}

			return rightButton;
		}

		/*
		 * @return Dimension
		 * @see com.cosylab.gui.components.SimpleComboBox.SimpleComboBoxUI#calculatePreferredSize()
		 */
		@Override
        protected Dimension calculatePreferredSize()
		{
			Dimension retVal = super.calculatePreferredSize();
			retVal.width += ArrowButton.PREFERRED_SIZE;//(getRightButton().getPreferredSize().width + getLeftButton().getPreferredSize().width);

			return retVal;
		}

		/*
		 * @return Rectangle
		 * @see com.cosylab.gui.components.SimpleComboBox.SimpleComboBoxUI#calculateRendererCellRectangle()
		 */
		@Override
        protected Rectangle calculateRendererCellRectangle()
		{
			Rectangle retVal = super.calculateRendererCellRectangle();
			retVal.x += getRightButton().getBounds().width;
			retVal.width -= (getRightButton().getBounds().width
			+ getLeftButton().getBounds().width);

			return retVal;
		}

		/*
		 * @see com.cosylab.gui.components.SimpleComboBox.SimpleComboBoxUI#calculateLayout()
		 */
		@Override
        protected void calculateLayout()
		{
			int width = comboBox.getWidth();
			int height = comboBox.getHeight();
			Insets insets = comboBox.getInsets();

			int buttonSize = height - (insets.top + insets.bottom);

			getLeftButton().setBounds(insets.left, insets.top, buttonSize,
			    buttonSize);
			getRightButton().setBounds(width - (insets.right + buttonSize),
			    insets.top, buttonSize, buttonSize);
		}

		/**
		 * Returns wether the popup list dialog is enabled.
		 *
		 * @return boolean
		 */
		public boolean isPopupEnabled()
		{
			return popupEnabled;
		}

		/**
		 * Sets the popupEnabled. When true, the user can access the drop-down
		 * list by clicking on the selected item cell.
		 *
		 * @param newPopupEnabled The popupEnabled to set
		 */
		public void setPopupEnabled(boolean newPopupEnabled)
		{
			if (popupEnabled == newPopupEnabled) {
				return;
			}

			popupEnabled = newPopupEnabled;

			if (comboBox == null) {
				mustSetPopup = true;

				return;
			} else {
				mustSetPopup = false;
			}

			if (popupEnabled) {
				comboBox.addMouseListener(popup.getMouseListener());
				comboBox.addMouseMotionListener(popup.getMouseMotionListener());
			} else {
				comboBox.removeMouseListener(popup.getMouseListener());
				comboBox.removeMouseMotionListener(popup.getMouseMotionListener());
			}
		}

		/**
		 * This method has been overriden to implement uninstallation of the
		 * two additional buttons.
		 *
		 * @param c JComponent
		 *
		 * @see javax.swing.plaf.ComponentUI#uninstallUI(JComponent)
		 */
		@Override
        public void uninstallUI(JComponent c)
		{
			comboBox.remove(getRightButton());
			comboBox.remove(getLeftButton());
			rightButton = null;
			leftButton = null;
			super.uninstallUI(c);
		}

		/**
		 * This method has been overriden to implement installation of the two
		 * additional buttons.
		 *
		 * @param c JComponent
		 *
		 * @see javax.swing.plaf.ComponentUI#installUI(JComponent)
		 */
		@Override
        public void installUI(JComponent c)
		{
			super.installUI(c);
			comboBox.add(getRightButton(), BorderLayout.EAST);
			comboBox.add(getLeftButton(), BorderLayout.WEST);

			if (mustSetPopup) {
				setPopupEnabled(popupEnabled);
			}
		}
	}

	/**
	 * This extension of SimpleComboBoxUI uses the standard right-side arrow
	 * button to activate the drop-down list.
	 *
	 * @author <a href="mailto:jernej.kamenik@cosylab.com">Jernej Kamenik</a>
	 * @version $id$
	 */
	public static class DropComboBoxUI extends SimpleComboBoxUI
	{
		protected ArrowButton dropButton = null;

		/*
		 * @return ArrowButton
		 */
		protected SimpleButton getDropButton()
		{
			if (dropButton == null) {
				dropButton = new ArrowButton(ArrowButton.Orientation.DOWN);
				dropButton.addMouseListener(popup.getMouseListener());
				dropButton.addMouseMotionListener(popup.getMouseMotionListener());
				dropButton.setBackground(ColorHelper.getControl());
				dropButton.setBorder(null);
				dropButton.setPressedBorder(null);
                dropButton.putClientProperty("doNotCancelPopup", comboBox.HIDE_POPUP_KEY);
			}

			return dropButton;
		}

		/*
		 * @return Dimension
		 * @see com.cosylab.gui.components.SimpleComboBox.SimpleComboBoxUI#calculatePreferredSize(SimpleComboBox)
		 */
		@Override
        protected Dimension calculatePreferredSize()
		{
			Dimension retVal = super.calculatePreferredSize();
			//retVal.width += getDropButton().getPreferredSize().width;

			return retVal;
		}

		/*
		 * @return Rectangle
		 * @see com.cosylab.gui.components.SimpleComboBox.SimpleComboBoxUI#calculateRendererCellRectangle(SimpleComboBox)
		 */
		@Override
        protected Rectangle calculateRendererCellRectangle()
		{
			Rectangle retVal = super.calculateRendererCellRectangle();
			retVal.width -= getDropButton().getBounds().width;

			return retVal;
		}

		/*
		 * @see com.cosylab.gui.components.SimpleComboBox.SimpleComboBoxUI#layoutComboBox(SimpleComboBox)
		 */
		@Override
        protected void calculateLayout()
		{
			int width = comboBox.getWidth();
			int height = comboBox.getHeight();
			Insets insets = comboBox.getInsets();

			int buttonSize = height - (insets.top + insets.bottom);

			getDropButton().setBounds(width - (insets.right + buttonSize),
			    insets.top, buttonSize, buttonSize);
		}

		/**
		 * This method has been overriden to implement the uninstallation of
		 * the additional button.
		 *
		 * @param c JComponent
		 *
		 * @see javax.swing.plaf.ComponentUI#uninstallUI(JComponent)
		 */
		@Override
        public void uninstallUI(JComponent c)
		{
			comboBox.remove(getDropButton());
			dropButton = null;
			super.uninstallUI(c);
		}

		/**
		 * @see javax.swing.plaf.ComponentUI#installUI(JComponent)
		 */
		@Override
        public void installUI(JComponent c)
		{
			super.installUI(c);
			comboBox.add(getDropButton(), BorderLayout.EAST);
		}
	}
}

/* __oOo__ */
