/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components.gauger;

/**
 * Simple class to generate perlin noise signal.
 *
 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
 * @version $id$
 */
public class NoiseMaker
{
	/**
	 * Pseudo-random generator function. Returns random value at given time.
	 * This function will return pseudo-random values for increasing time, but
	 * always same value for same time.
	 *
	 * @param x time.
	 *
	 * @return Random value at given time.
	 */
	public static final double noise(long x)
	{
		x ^= (x << 13);

		return (1.0
		- ((x * (x * x * 15731 + 789221) + 1376312589) & 0x7fffffff) / 1073741824.0);
	}

	/**
	 * Returns gaussian averaged noise value sampled over three time units.
	 *
	 * @param x time offset.
	 *
	 * @return Smoothed noise value.
	 */
	public static final double smoothedNoise(final long x)
	{
		return noise(x) * 0.5 + 0.25 * (noise(x - 1) + noise(x + 1));
	}

	/**
	 * Returns cosine interpolated noise value between two integral time units.
	 *
	 * @param x time.
	 *
	 * @return Interpolated noise value at time x between floor(x) and ceil(x)
	 */
	public static final double interpolatedNoise(final double x)
	{
		final long intX = (long)x;

		double n1 = noise(intX);
		double n2 = noise(intX + 1);

		double v1 = (noise(intX - 1) + n2) * 0.25 + n1 * 0.5;
		double v2 = (noise(intX + 2) + n1) * 0.25 + n2 * 0.5;

		return v1 + (v2 - v1) * (1 - Math.cos((x - intX) * Math.PI)) * 0.5;
	}

	/**
	 * Returns pseudo random noise signal at given value. For a given parameter
	 * x the return value will always be the same, but for increasing values,
	 * return value will be seemingly random smoothed noise.
	 *
	 * @param x Time at which to sample the noise.
	 *
	 * @return Value of signal.
	 */
	public static final double perlinNoise(final double x)
	{
		double total = 0;
		double n = 6;
		double frequency = 1;

		for(int i = 0; i < n; i++) {
			total += interpolatedNoise(x * frequency);

			frequency *= 2;
		}

		return total / n;
	}

	/**
	 * Simple test method.
	 *
	 * @param args Not used.
	 */
	public static void main(String[] args)
	{
		double min = Double.MAX_VALUE;
		double max = Double.MIN_VALUE;
		double c = 0;

		for(int i = 0; i < 10000000; i++) {
			c = perlinNoise(i);

			if(c < min) {
				min = c;
				System.out.println("Minimum = " + min + ", Maximum = " + max);
			}

			if(c > max) {
				max = c;
				System.out.println("Minimum = " + min + ", Maximum = " + max);
			}
		}
	}
}

/* __oOo__ */
