/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components.gauger;

import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.geom.Point2D;

import java.util.ArrayList;


/**
 * Abstract utility class that provides the mathematic foundation for drawing
 * the Gauger scales. Scale is defined in its own coordinate system with the
 * axis (u, v). In this system the extents of the scale are defined to be  (0,
 * 0) and (1, 1), meaning that any point on this scale can be specified  with
 * the (u, v) parameter value between 0 and 1. This class is the
 * implementation of this transformation and the component cartesian
 * coordinate system.
 *
 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
 * @version $id$
 */
public abstract class ScaleTransform
{
	private Point2D tempPoint = new Point2D.Double();
	private Point labelPosition = new Point();
	private boolean polar = false;

	/**
	 * Abstract base class for scale segments.
	 *
	 * @author $Author: jbobnar $
	 * @version $Revision: 1.4 $
	 */
	public class ScaleSegment
	{
	}

	/**
	 * Class defining linear scale segment.
	 *
	 * @author $Author: jbobnar $
	 * @version $Revision: 1.4 $
	 */
	public class LineSegment extends ScaleSegment
	{
		/** Starting point of line */
		public Point2D lineStart;

		/** Ending point of line */
		public Point2D lineEnd;

		/**
		 * Creates a new LineSegment object.
		 *
		 * @param xs X coordinate of starting point.
		 * @param ys Y coordinate of starting point.
		 * @param xe X coordinate of ending point.
		 * @param ye Y coordinate of ending point.
		 */
		public LineSegment(double xs, double ys, double xe, double ye)
		{
			lineStart = new Point2D.Double(xs, ys);
			lineEnd = new Point2D.Double(xe, ye);
		}
	}

	/**
	 * Definition of arc scale segment.
	 *
	 * @author $Author: jbobnar $
	 * @version $Revision: 1.4 $
	 */
	public class ArcSegment extends ScaleSegment
	{
		/** Center point of arc */
		public Point2D center;

		/** Outer radius of arc */
		public double rOuter;

		/** Inner radius of arc */
		public double rInner;

		/** Starting angle of arc */
		public double startAngle;

		/** Included angle of arc */
		public double arcAngle;

		/**
		 * Creates a new ArcSegment object.
		 *
		 * @param c Center point
		 * @param r inner radius
		 * @param R Outer radius
		 * @param start Starting angle
		 * @param angle Included angle
		 */
		public ArcSegment(Point2D c, double r, double R, double start,
		    double angle)
		{
			center = c;
			rOuter = R;
			rInner = r;
			startAngle = start;
			arcAngle = angle;
		}
	}

	/**
	 * Defines rectangular scale segment.
	 *
	 * @author $Author: jbobnar $
	 * @version $Revision: 1.4 $
	 */
	public class RectangleSegment extends ScaleSegment
	{
		/** Coordinates of rectangle */
		public Rectangle rect;

		/**
		 * Creates a new RectangleSegment object.
		 *
		 * @param rect Coordinates of rectangle.
		 */
		public RectangleSegment(Rectangle rect)
		{
			this.rect = rect;
		}
	}

	/** Segments comprising this scale. */
	private ArrayList<ScaleSegment> segments = null;

	/**
	 * Creates a new ScaleTransform object.
	 */
	public ScaleTransform()
	{
		super();
		segments = new ArrayList<>();
	}

	/**
	 * Sets whether this transform is polar. A polar scale will be rendered in
	 * polar coordinates rather then cartesian. Lines will be rendered as arcs
	 * and text will be rotated.
	 *
	 * @param value True if polar.
	 */
	public void setPolar(boolean value)
	{
		polar = value;
	}

	/**
	 * Returns whether this scale is polar.
	 *
	 * @return True if polar.
	 *
	 * @see setPolar(boolean)
	 */
	public boolean isPolar()
	{
		return polar;
	}

	/**
	 * Sets the component parameters for this transformation.
	 *
	 * @param w width of the component that will display this transformation
	 * @param h height of the component that will display this transformation
	 * @param marginX not used
	 * @param marginY not used
	 * @param tickOffset space between the edge of the scale and the first tick
	 */
	public abstract void setParameters(int w, int h, int marginX, int marginY,
	    int tickOffset);

	/**
	 * Maps the point specified with (u, v) coordinates in the scale space to
	 * cartesian space with coordinates (x, y). (u, v) values must be in
	 * [0..1]  range, the (x, y) points will be in (0..w-1, 0..h-1) range.
	 *
	 * @param scaleSpace point in scale space
	 * @param cartesianSpace point in cartesian space
	 */
	public abstract void mapUVtoXY(Point2D scaleSpace, Point2D cartesianSpace);

	/**
	 * Maps the point specified with (u, v) coordinates in the scale space to
	 * cartesian space with coordinates (x, y). (u, v) values must be in
	 * [0..1] range, the (x, y) points will be in (0..w-1, 0..h-1) range.
	 *
	 * @param u coordinate in scale space
	 * @param v coordinate in scale space
	 * @param cartesianSpace point in cartesian space
	 */
	public void mapUVtoXY(double u, double v, Point2D cartesianSpace)
	{
		tempPoint.setLocation(u, v);
		mapUVtoXY(tempPoint, cartesianSpace);
	}

	/**
	 * Maps point specified with (u, v) coordinates in scale space to
	 * cartesian space with coordinates (x, y)
	 *
	 * @param u coordinate in scale space.
	 * @param v coordinate in scale space.
	 *
	 * @return Point2D cartesianSpace point.
	 *
	 * @see mapUVtoXY(u, v, Point2D)
	 */
	public Point2D mapUVtoXY(double u, double v)
	{
		Point2D point = new Point2D.Double();
		mapUVtoXY(u, v, point);

		return point;
	}

	/**
	 * Returns angle between horizontal and specified relative position in
	 * scale space. This angle is defined as angle between X axis and line
	 * from point (x, 0.0) to point (x, 100.0). This is the angle at which
	 * vertical straight line in polar mode is rotated in cartesian space.
	 *
	 * @param x Relative point in scale space in range 0.0 to 1.0.
	 *
	 * @return double angle in radians
	 */
	public final double getAngle(double x)
	{
		final Point2D p1 = mapUVtoXY(x, 0);
		final Point2D p2 = mapUVtoXY(x, 100000.0);

		final double dx = (p2.getX() - p1.getX());
		final double dy = (p2.getY() - p1.getY());

		double phi = (dx == 0.0) ? 0 : -Math.atan(dy / dx) - Math.PI * 0.5;

		if(dx < 0) {
			phi += Math.PI;
		}

		return phi;
	}

	/**
	 * Returns the width of scale in pixels at the specified location v in the
	 * scale space. Although the scale for v is only defined in the range
	 * [0..1], this value can generaly return even values outside this range,
	 * although this is not guaranteed.
	 *
	 * @return Width of scale in pixels.
	 */
	public abstract double scaleWidth(double v);

	/**
	 * Returns the height of scale in pixels at the specified location u in the
	 * scale space. Although the scale for u is only defined in the range
	 * [0..1], this value can generaly return even values outside this range,
	 * although this is not guaranteed.
	 *
	 * @return Height of scale in pixels.
	 */
	public abstract double scaleHeight(double u);

	/**
	 * Returns the point where the label should be positioned. This point
	 * represents geometric center of the text. There are no limits to the
	 * size  of the text written around this position.
	 *
	 * @return Center point for label.
	 */
	public Point getLabelPosition()
	{
		return labelPosition;
	}

	/**
	 * Returns the number of segments describing this scale. Even if the value
	 * is zero, this transform may still provide space transformation
	 * functionality.
	 *
	 * @return Number of segments.
	 */
	public int getSegmentCount()
	{
		return segments.size();
	}

	/**
	 * Adds segment to this transform.
	 * 
	 * @param segment to add.
	 */
	protected void addSegment(ScaleSegment segment)
	{
		segments.add(segment);
	}

	/**
	 * Returns scale segment at specified index.
	 *
	 * @param i Index of segment.
	 *
	 * @return Scale segment.
	 */
	public ScaleSegment getSegment(int i)
	{
		return segments.get(i);
	}

	/**
	 * Sets position of value label. The position is center point of label.
	 * 
	 * @param x coordinate of label center.
	 * @param y coordinate of label center.
	 */
	protected void setLabelPosition(int x, int y)
	{
		labelPosition.x = x;
		labelPosition.y = y;
	}

	/**
	 * Measues the width of tick label.
	 *
	 * @param g Graphic context.
	 * @param x Relative position of tick.
	 * @param text Tick label.
	 *
	 * @return Width of label in pixels.
	 */
	public abstract int measureTick(Graphics g, double x, String text);
}

/* __oOo__ */
