/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components.ledder;
import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.geom.Rectangle2D;

import javax.swing.ImageIcon;

/**
 * This is the default icon used when displaying leds. This class is 
 * instantiated when led icon could not be loaded from the resources. It 
 * displays filled circle of appropriate color.
 * 
 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
 * @version $id$
 * 
 * @see com.cosylab.gui.components.ledder.Led
 * @see com.cosylab.gui.components.ledder.LedIconFactory
 */
public class DefaultLedIcon extends ImageIcon {

    private static final long serialVersionUID = 1659727989748852974L;

    // Parameters defining rendering of the led.
	private LedPaintParameters parameters;

	// Paint implementation used to render the led.
	private LedPaint paint;
	
	protected int size=14;

	/**
	 * Default constructor for DefaultLedIcon. If the color parameter is null,
	 * color will be set to <code>java.awt.Color.BLACK</code>
	 * 
	 * @param color Color
	 */
	protected DefaultLedIcon(Color color) {
		super();

		if (color == null)
			color = Color.BLACK;

		parameters = createDefaultParameters();
		
		getParameters().color = color;
	}

	/**
	 * Returns default parameters to be used for rendering of leds.
	 * 
	 * @return LedPaintParameters
	 */
	protected LedPaintParameters createDefaultParameters() {
		LedPaintParameters lpp = new LedPaintParameters();
		
		//lpp.centerX = 7;
		//lpp.centerY = 7;
		lpp.radius = 6;
		lpp.highlightFactor = 0.85;
		lpp.highlightFocus = 3.0;
		lpp.shadowIntensity = 1.0;
		lpp.borderSize = 0.05;
		lpp.lightX = -0.20;
		lpp.lightY = -0.50;
		lpp.antialiasing = 3;
		
		return lpp;
	}

	/**
	 * Returns LedPaint object that renders the icon.
	 * 
	 * @return LedPaint
	 */
	private final LedPaint getPaint() {
		if (paint==null) {
			paint = new LedPaint(getParameters());
		}
		return paint;
	}

	/**
	 * Returns LedPaintParameters that are initialized in the constructor.
	 * @return LedPaintParameters
	 */
	private final LedPaintParameters getParameters() {
		return parameters;
	}

	/**
	 * Returns the height of this icon. This value is set to 16.
	 * 
	 * @return int
	 * @see javax.swing.Icon#getIconHeight()
	 */
	@Override
    public final int getIconHeight() {
		return size;
	}

	/**
	 * Returns the width of this icon. This value is set to 16.
	 * 
	 * @return int
	 * @see javax.swing.Icon#getIconWidth()
	 */
	@Override
    public final int getIconWidth() {
		return size;
	}
	
	/**
	 * Renders the icon.
	 * 
	 * @param component Component
	 * @param g Graphics
	 * @param x int
	 * @param y int
	 * @see javax.swing.Icon#paintIcon(Component, Graphics, int, int)
	 */
	@Override
    public synchronized void paintIcon(Component component, Graphics g, int x, int y) {
		Graphics2D g2 = (Graphics2D) g;
		getPaint().setCenter(x+parameters.radius+2,y+parameters.radius+2);
		g2.setPaint(getPaint());
		g2.fill(new Rectangle2D.Double(x, y, getIconWidth(), getIconHeight()));

	}
}
