/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components.ledder;
import java.awt.Color;
import java.awt.Font;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;

import javax.swing.Icon;
import javax.swing.JLabel;
import javax.swing.SwingConstants;
import javax.swing.SwingUtilities;

/**
 * Visual representation of LED status. A LED is represented by icon indicating 
 * current status and textual description. Status is defined by binary value of 
 * true or false and each of the two states has user defined color associated.
 * <p>
 * This implementation attempts to create the icons using the LedFactory class.
 * If icon for the desired color cannot be found, simple color representation
 * will be used.
 * 
 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
 * @version $id$

 * @see com.cosylab.gui.components.ledder.LedIconFactory
 * @see com.cosylab.gui.components.ledder.DefaultLedIcon
 */
public class Led extends JLabel {

    private static final long serialVersionUID = 6869641345996051678L;
    // Icon representing the OFF state.
	private Icon iconWhenOff = null;
	// Icon representing the ON state.
	private Icon iconWhenOn = null;

	// Flag indicating current state of the led.
	private boolean on = false;
	
	private CustomBorder border;
	private boolean editable = false;

	
	
	/**
	 * Default constructor for Led component. 
	 * Creates a green/red led icon and "No description" label.
	 */
	public Led() {
		this(Color.GREEN, Color.RED);
		
	}

	/**
	 * Switches between editable and read only mode. When editable, the led
	 * responds to mouse clicks.
	 * 
	 * @param editable
	 */
	public void setEditable(boolean editable){
	    this.editable=editable;
	    if (editable) {
	    	setBorder(getCustomBorder());
	    } else {
	    	setBorder(null);
	    }
	}
	/**
	 * Constructs new Led with user defined colors. 
	 * Description will be set to "No description".
	 * 
	 * @param colorWhenOn Color
	 * @param colorWhenOff Color
	 */
	public Led(Color colorWhenOn, Color colorWhenOff) {
	    
		this("< No description >", colorWhenOn, colorWhenOff);
	
	}

	/**
	 * Constructs new Led with user defined description and colors.
	 * 
	 * @param description String
	 * @param colorWhenOn Color
	 * @param colorWhenOff Color
	 */
	public Led(String description, Color colorWhenOn, Color colorWhenOff) {
		super();

		setInheritsPopupMenu(true);
		
		setFont(getFont().deriveFont(Font.PLAIN));

		setText(description);
		setVerticalAlignment(SwingConstants.CENTER);
		setVerticalTextPosition(SwingConstants.CENTER);

		iconWhenOff = LedIconFactory.createIcon(colorWhenOff);
		iconWhenOn = LedIconFactory.createIcon(colorWhenOn);

		updateIcon();
		setOpaque(false);		
	    
		addMouseListener(new MouseListener() {
		        
	        public void mouseClicked(MouseEvent e) {
	        	
	      
	        		
	        }

            public void mouseEntered(MouseEvent e) {
            	if(editable){
                	getCustomBorder().setVisible(true);
                	repaint();
                 }
             }

            public void mouseExited(MouseEvent e) {
            	if(editable){
                	getCustomBorder().setVisible(false);
                	repaint();
                 }
            }

            public void mousePressed(MouseEvent e) {
            	if(SwingUtilities.isLeftMouseButton(e)){
	        		if(editable){
	        			if(isOn()){
	        				setOn(false);
	                
	        			}else {
	        				setOn(true);
	        			}     
	        		
	        		}
            	}
            }

            public void mouseReleased(MouseEvent e) {
                // TODO Auto-generated method stub
                
            }

            
          	});
		
		
	}

	/**
	 * Returns current state of the led, true or false. Visual representation 
	 * of this state depends on the specified colors.
	 * 
	 * @return boolean
	 */
	public boolean isOn() {
		return on;
	}

	/**
	 * Sets the state of the led. Visual representation of this state depends 
	 * on the specified colors.
	 * 
	 * @param value boolean
	 */
	public void setOn(boolean value) {
		on = value;
		updateIcon();
	}

	/**
	 * Sets the apropriate icon based on the state of the led.
	 */
	private void updateIcon() {
		setIcon(on ? iconWhenOn : iconWhenOff);
	}

	/**
	 * Sets the text of led and the tool tip to be displayed when the user
	 * holds the mouse above the led.
	 * 
	 * @param s String
	 */
	@Override
	public void setText(String s) {
		super.setText(s);
		setToolTipText(s);
	}

	/**
	 * Simple test method.
	 * 
	 * @param args String[]
	 */
	public static void main(String[] args) {
		javax.swing.JDialog dialog = new javax.swing.JDialog();
		dialog.setModal(true);
		dialog.setSize(200, 200);

		dialog.getContentPane().setLayout(new java.awt.FlowLayout());

		Led l = new Led();
		l.setText("Transmit");
		dialog.getContentPane().add(l);

		l = new Led(Color.YELLOW, Color.BLUE);
		l.setText("Receive");
		dialog.getContentPane().add(l);

		l = new Led("Status", Color.RED, Color.GREEN);
		l.setOn(false);
		dialog.getContentPane().add(l);

		l = new Led(Color.blue, Color.blue);
		l.setText("On");
		dialog.getContentPane().add(l);

		l = new Led(Color.GREEN, Color.RED);
		l.setText("Off");
		dialog.getContentPane().add(l);

		l = new Led(Color.yellow, Color.blue);
		l.setText("Power");
		dialog.getContentPane().add(l);

		dialog.setVisible(true);

		System.exit(0);
	}

	/**
	 * @return Returns the border.
	 */
	private CustomBorder getCustomBorder() {
		if (border == null) {
			border = new CustomBorder();
			border.setVisible(false);
		}

		return border;
	}
	
	/**
	 * Returns true if this led is editable. When editable, the led responds to
	 * the mouse clicks.
	 * @return
	 */
	public boolean isEditable() {
		return editable;
	}
	
	
}

