/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components.ledder;

import java.awt.PaintContext;
import java.awt.image.ColorModel;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;


/**
 * Renderer for Leds. Implements ray tracing with specular highlighting and
 * antialiasing.
 *
 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
 * @version $id$
 */
public class LedPaintContext implements PaintContext
{
	private LedRenderer lr;
	private int aliasFactor = 1;
	private int[] offset;
	private int lastOffsetW = 0;
	private int lastOffsetF = 0;

	/**
	 * Creates a new LedPaintContext object.
	 *
	 * @param x horizontal size.
	 * @param y vertical size.
	 * @param parameters defining the context.
	 */
	public LedPaintContext(int x, int y, LedPaintParameters parameters)
	{
		aliasFactor = parameters.antialiasing;
		lr = new LedRenderer(aliasFactor * x, aliasFactor * y,
			    aliasFactor * parameters.radius, parameters);
	}

	/**
	 * Destroys this context.
	 */
	public void dispose()
	{
		offset = null;
		lr = null;
	}

	/**
	 * Returns color model used by this context:
	 * <code>ColorModel.getRGBdefault()</code>.
	 *
	 * @return RGB Color model.
	 */
	public ColorModel getColorModel()
	{
		return ColorModel.getRGBdefault();
	}

	private int[] subsample(final int[] pixels, final int w, final int h,
	    final int f)
	{
		int[] result = new int[4 * w * h];
		int oR = 0;
		int c = f * f;
		int acc;
		int base = 0;
		int baseInc = 4 * (f - 1);

		for (int y = 0; y < h; y++) {
			base = c * y * w << 2;

			for (int x = 0; x < w; x++) {
				for (int i = 0; i < 4; i++) {
					acc = 0;

					for (int n = 0; n < c; n++) {
						acc += pixels[base + offset[n]];
					}

					result[oR++] = acc / c;
					base++;
				}

				base += baseInc;
			}
		}

		return result;
	}

	/**
	 * Creates array for antialiasing.
	 *
	 * @param factor by which to scale the array.
	 * @param width original width of the array.
	 */
	public void createOffsetArray(final int factor, final int width)
	{
		if ((lastOffsetW == width) && (lastOffsetF == factor)) {
			return;
		}

		lastOffsetW = width;
		lastOffsetF = factor;

		offset = new int[factor * factor];

		int c = 0;

		for (int m = 0; m < factor; m++) {
			for (int n = 0; n < factor; n++) {
				offset[c++] = 4 * (m * factor * width + n);
			}
		}
	}

	/**
	 * Creates buffer, renders the led and antialiases the result.
	 *
	 * @param x offset.
	 * @param y offset.
	 * @param w width of area to render.
	 * @param h height of area to render.
	 *
	 * @return Raster of rendered pixels.
	 */
	public Raster getRaster(int x, int y, int w, int h)
	{
		WritableRaster raster = getColorModel().createCompatibleWritableRaster(w,
			    h);

		int[] data = lr.renderSection(aliasFactor * x, aliasFactor * y,
			    aliasFactor * w, aliasFactor * h);

		createOffsetArray(aliasFactor, w);

		int[] result = (aliasFactor == 1) ? data
			: subsample(data, w, h, aliasFactor);

		raster.setPixels(0, 0, w, h, result);

		return raster;
	}
}

/* __oOo__ */
