/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components.range2;

/**
 * Implementation of logarithmic range.
 *
 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
 * @version $id$
 */
public class LogarithmicRange implements Range
{
	
	private TickCalculator defaultTickCalculator;
	/**
	 * Returns absolute value. Function is only guaranteed to return for values
	 * larger than 0.
	 *
	 * @param relative value in 0.0 to 1.0 range.
	 *
	 * @return absolute value.
	 *
	 * @see com.cosylab.gui.components.range.Range#toAbsolute(double)
	 */
	public double toAbsolute(double relative, RangedValue val)
	{
		double logMin = Math.log(val.getMinimum());
		double logMax = Math.log(val.getMaximum());

		return Math.exp(logMin + relative * (logMax - logMin));
	}

	/**
	 * Returns relative value in 0.0 to 1.0 range inclusively.
	 *
	 * @param absolute value, must be larger than 0.0.
	 *
	 * @return value in 0.0 to 1.0 range.
	 *
	 * @see com.cosylab.gui.components.range.Range#toRelative(double)
	 */
	public double toRelative(double absolute, RangedValue val)
	{
		double logMin = Math.log(val.getMinimum());
		double logMax = Math.log(val.getMaximum());

		double range = logMax - logMin;

		if (range == 0.0) {
			return 0.5;
		}

		return (Math.log(absolute) - logMin) / range;
	}

	/** Constant defining smallest number that can be defined within range. */
	public static final double SMALLEST = 1e-8;

	/**
	 * Checks whether value is positive. If value is smaller than value defined
	 * by <code>SMALLEST</code> constant, this constant is returned. This
	 * prevents setting of values that cannot be representated on logarithmic
	 * scale.
	 *
	 * @param value to check.
	 *
	 * @return same as value parameter if valid, <code>SMALLEST</code>
	 *         otherwise.
	 */
	public final double validate(double value)
	{
		return (value > SMALLEST) ? value : SMALLEST;
	}

	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.components.range2.Range#getDefaultTickCalculator()
	 */
	public TickCalculator getDefaultTickCalculator() {
		if (defaultTickCalculator == null) {
			defaultTickCalculator = new LogarithmicTickCalculator(3, false);
		}
		return defaultTickCalculator;
	}
}

/* __oOo__ */
