/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components.range2;

/**
 * This class represents adjustable policies for <code>RangedValue</code>.
 * Whenever a change any of the values (minimum, maximum or value) is made in
 * <code>RangedValue</code> validate method will be called giving the
 * implementor of this interface the chance to adjust the values before they
 * are set and property change events invoked.<br>
 * Note: Regardles of the changes proposed by validate method,
 * <code>RangedValue</code> will always ensure the proper relation between the
 * values: values must always form increasing order (minimum, value,
 * maximum).
 *
 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
 * @version $id$
 */
public abstract class RangedValuePolicy
{
	RangedValuePolicy peer;
	
	/**
	 * Passes the value triplet <code>params</code> that may be modified by
	 * validate method. Regardles of the changes made, all three values will
	 * still be made to comply with the basic definition of <code>
	 * RangedValue</code>.
	 * 
	 * <p>Before or after this policy has been executed it must call protected method validatePeerPolicy(RangedValue).
	 *
	 * @param params Triplet of values to validate.
	 * 
	 * @return if the validation require the change of value or bounds of the RangedValue
	 * 		a new RangedValue with appropriate values should be returned
	 */
	public abstract RangedValue validate(RangedValue params);
	
	/**
	 * Inserts provided polici as first policy in chain, here next peer policy is this policy.
	 * @param p new first policy in chain
	 * @return returns new first policy in chain, the p parameter.
	 */
	public RangedValuePolicy insertPeerPolicy(RangedValuePolicy p) {
		p.peer=this;
		return p;
	}

	/**
	 * Adds provided policy to the end of the chin
	 * @param p
	 */
	public void addPeerPolicy(RangedValuePolicy p) {
		if (p == this) return;
		if (peer!=null) {
			peer.addPeerPolicy(p);
		} else {
			peer=p;
		}
	}
	
	/**
	 * Returns the peer policy.
	 * 
	 * @return
	 */
	public RangedValuePolicy getPeerPolicy() {
		return peer;
	}
	
	/**
	 * Removes provided policy from the chain, and returnes first policy in chain.
	 * @param p policy to be removed from chain
	 * @return returns new first policy in chain, this or peer policy
	 */
	public RangedValuePolicy removePeerPolicy(RangedValuePolicy p) {
		if (p==this) {
			p.peer=null;
			return peer;
		}
		if (peer != null) {
			peer=peer.removePeerPolicy(p);
		}
		return this;
	}
	
	/**
	 * Removes provided policy from the chain, and returnes first policy in chain.
	 * @param p policy to be removed from chain
	 * @return returns new first policy in chain, this or peer policy
	 */
	public RangedValuePolicy removePeerPolicyByType(Class<? extends RangedValuePolicy> c) {
		if (c.isAssignableFrom(this.getClass())) {
			peer=null;
			return peer;
		}
		if (peer != null) {
			peer=peer.removePeerPolicyByType(c);
		}
		return this;
	}

	protected RangedValue validatePeerPolicy(RangedValue p) {
		if (peer !=null) {
			return peer.validate(p);
		}
		return p;
	}
	
}

/* __oOo__ */
