/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components.range2;

/**
 * Implementation of <code>RangedValuePolicy</code> that resizes the range
 * based on the value parameter.
 * 
 * <p>
 * Whenever the value is set below minimum, minimum will be changed to the the
 * value, decreased by factor  (maximum - minimum). Same rule applies for
 * maximum.
 * </p>
 * This policy is value dominant i.e., bounds will be adjusted to ensure that
 * value is always valid. If either bound is set incorrectly, it will be 
 * adjusted to conform with this condition.
 * <p>
 * Rule for adjusting bounds is as follows. Whenever value is either below
 * minimum ore above maximum, the corresponding bound will be adjusted using
 * value - range * factor or value + range * factor for minimum and maximum
 * respectively, where range is maximum - minimum.
 * 
 * 
 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
 * @version $id$
 */
public class RescalingValuePolicy extends RangedValuePolicy
{
	private double rescaleFactor;

	/**
	 * Default constructor for RescalingValuePolicy. Initializes rescaling
	 * factor to 0.25.
	 */
	public RescalingValuePolicy()
	{
		this(0.25);
	}

	/**
	 * Constructor for RescalingValuePolicy. Allows custom rescaling factor.
	 *
	 * @param factor by which the range will be resized.
	 */
	public RescalingValuePolicy(double factor)
	{
		super();
		setFactor(factor);
	}

	/**
	 * Adjusts bounds as specified in class JavaDoc.
	 *
	 * @param params RangedValueHolder
	 *
	 * @see com.cosylab.gui.components.range.RangedValuePolicy#validate(RangedValueHolder)
	 */
	@Override
	public RangedValue validate(RangedValue params)
	{
		double value = params.getValue();
		double range = params.getMaximum() - params.getMinimum();

		if(value < params.getMinimum()) {
			params = params.deriveWithMin(value - range * getFactor());

			return validatePeerPolicy(params);
		}

		if(value > params.getMaximum()) {
			params = params.deriveWithMax(value + range * getFactor());

			return validatePeerPolicy(params);
		}
		
		return validatePeerPolicy(params);
	}

	/**
	 * Returns rescaling factor.
	 *
	 * @return factor.
	 */
	public double getFactor()
	{
		return rescaleFactor;
	}

	/**
	 * Sets new rescaling factor. Factor set will not be less that 0.
	 *
	 * @param factor new rescaling factor.
	 */
	public void setFactor(double factor)
	{
		rescaleFactor = Math.max(0.0, factor);
	}
}

/* __oOo__ */
