/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components.range2;

import java.awt.CardLayout;
import java.awt.Color;
import java.awt.FontMetrics;
import java.awt.Graphics;

import javax.swing.JComponent;
import javax.swing.JFrame;

import com.cosylab.util.PrintfFormat;


/**
 * Simple demonstration of how to use classes in range package.
 *
 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
 * @version $id$
 */
public class SimpleDemoDisplayer extends JComponent implements TickParameters
{
	private static final long serialVersionUID = 1L;

	private class Updater implements RangedValueListener
	{
		/**
		 * Called whenever any of the parameters change.
		 * 
		 * @see com.cosylab.gui.components.range.RangedValueListener#valueChanged(com.cosylab.gui.components.range.RangedValueEvent)
		 */
		public void valueChanged(RangedValueEvent event)
		{
			repaint();
		}
	}

	private RangedValueController rv;
	private PrintfFormat pf = new PrintfFormat("%3.2f");

	/**
	 * Creates a new SimpleDemoDisplayer object.
	 */
	public SimpleDemoDisplayer()
	{
		rv = new RangedValueController();
		rv.setRange(new LogarithmicRange());
		rv.setValue(0, 10, 2);

		rv.addRangedValueListener(new Updater());

		rv.setPolicy(new RescalingValuePolicy());
	}

	/**
	 * Sets new minimum.
	 *
	 * @param min value.
	 */
	public void setMinimum(double min)
	{
		rv.setValue(min, rv.getMaximum(), rv.getValue());
	}

	/**
	 * Sets new maximum.
	 *
	 * @param max value.
	 */
	public void setMaximum(double max)
	{
		rv.setValue(rv.getMinimum(), max, rv.getValue());
	}

	/**
	 * Sets new value.
	 *
	 * @param val new value.
	 */
	public void setValue(double val)
	{
		rv.setValue(rv.getMinimum(), rv.getMaximum(), val);
	}

	/**
	 * Returns minimum value displayed.
	 *
	 * @return double minimum value.
	 */
	public double getMinimum()
	{
		return rv.getMinimum();
	}

	/**
	 * Returns maximum value displayed.
	 *
	 * @return double maximum value.
	 */
	public double getMaximum()
	{
		return rv.getMaximum();
	}

	/**
	 * Returns current value displayed.
	 *
	 * @return double current value.
	 */
	public double getValue()
	{
		return rv.getValue();
	}

	/**
	 * Code that renders the component. Paints background, ticks and value.
	 *
	 * @param g graphics.
	 */
	@Override
	public void paintComponent(Graphics g)
	{
		super.paintComponent(g);

		g.setColor(Color.BLACK);
		g.fillRect(0, 0, getWidth(), getHeight());

		g.setColor(Color.WHITE);
		g.drawRect(0, 0, getWidth() - 1, getHeight() - 1);

		Tick[] ticks = rv.calculateTicks(getWidth(),this);

		for (int i = 0; i < ticks.length; i++) {
			Tick t = ticks[i];
			if (t==null) continue;
			double x = t.proportional * getWidth();

			if (t.major) {
				g.drawString(t.text,
				    (int)x - g.getFontMetrics().stringWidth(t.text) / 2,
				    getHeight() / 4 - 2);
				g.drawLine((int)x, getHeight(), (int)x, getHeight() / 4);
			} else {
				g.drawLine((int)x, getHeight(), (int)x, getHeight() / 2);
			}
		}

		g.setColor(Color.RED);

		double x = rv.getRelativeValue() * getWidth();
		g.drawLine((int)x, getHeight(), (int)x, 0);
	}	

	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.components.range2.TickParameters#measureTick(double, java.lang.String)
	 */
	public int measureTick(double position, String text)
	{
		FontMetrics fm = getFontMetrics(getFont());

		return fm.stringWidth(text);
	}
	
	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.components.range2.TickParameters#formatNumber(double)
	 */
	public String formatNumber(double value) {
		return pf.sprintf(value);
	}
	
	/**
	 * Test method.
	 *
	 * @param args command-line arguments.
	 */
	public static void main(String[] args)
	{
		JFrame jf = new JFrame();
		jf.setSize(200, 80);
		jf.getContentPane().setLayout(new CardLayout());

		SimpleDemoDisplayer sdd = new SimpleDemoDisplayer();
		sdd.setMinimum(0.2);

		//sdd.setMaximum(90.312524);
		sdd.setMaximum(2);
		jf.getContentPane().add(sdd, "SimpleDemoDisplayer");

		jf.setVisible(true);

		while (true) {
			sdd.setValue(Math.random() * 9);

			try {
				Thread.sleep(100);
			} catch (InterruptedException e) {
			}
		}
	}
}

/* __oOo__ */
