/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components.table.renderers;

import com.cosylab.gui.components.table.cells.TableCell;
import com.cosylab.gui.components.util.ColorHelper;
import com.cosylab.gui.components.util.PaintHelper;

import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics;

import java.util.GregorianCalendar;

import javax.swing.JTable;


/**
 * Default renderer used for all <code>PropertyCell</code> objects. Uses color
 * coding to indicate state of cell, but does not render value.
 *
 * @author <a href="mailto:igor.kriznar@cosylab.com">Igor Kriznar</a>
 * @version $id$
 */
public class DefaultTableCellRenderer
	extends javax.swing.table.DefaultTableCellRenderer
{
	private static final long serialVersionUID = 1L;

	/**
	 * Default constructor.
	 */
	public DefaultTableCellRenderer()
	{
		this(false,true);
	}
	/**
	 * Default constructor.
	 */
	public DefaultTableCellRenderer(boolean decorateBorder, boolean decorateBackground)
	{
		super();
		setOpaque(true);
		this.decorateBorder=decorateBorder;
		this.decorateBackground=decorateBackground;
	}

	protected boolean timeout;
	protected boolean error;
	protected boolean alarm;
	protected boolean active;
	protected String severity;
	protected boolean decorateBorder=false;
	protected boolean decorateBackground=true;

	/**
	 * Returns cell renderer component. This implementation is specific to
	 * <code>PropertyCell</code> objects. Color coding will be used to
	 * indicate states. If cell indicates timeout, background will be rendered
	 * gray, if value is out of bounds as defined by <code>BoundedCell</code>,
	 * background will be colored red.
	 *
	 * @param table JTable
	 * @param cell Object
	 * @param isSelected boolean
	 * @param hasFocus boolean
	 * @param row int
	 * @param column int
	 *
	 * @return Component
	 *
	 * @see javax.swing.table.TableCellRenderer#getTableCellRendererComponent(JTable,
	 *      Object, boolean, boolean, int, int)
	 */
	@Override
    public Component getTableCellRendererComponent(JTable table, Object cell,
	    boolean isSelected, boolean hasFocus, int row, int column)
	{
		Component comp = super.getTableCellRendererComponent(table, cell,
			    isSelected, hasFocus, row, column);

		// Workaround for ugly default behaviour of DefaultTableCellRenderer
		if (isSelected) {
			comp.setBackground(table.getSelectionBackground());
		} else {
			comp.setBackground(table.getBackground());
		}

		if (cell instanceof TableCell) {
			TableCell tc = (TableCell)cell;

			if (tc.isOutOfBounds()) {
				comp.setBackground(ColorHelper.getWarning());
			}

			timeout = tc.isTimeout();
			error = tc.isError();
			active = !tc.isSuspended();
			alarm = tc.isAlarm();
			severity = (String)tc.getCharacteristic(TableCell.SEVERITY);

			String t = tc.getTooltip();
			setToolTipText(t);
			
			if (decorateBackground) {
				if (error || timeout) {
					setBackground(ColorHelper.getAlarmInvalidBackground());
				}

				if (alarm) {
					if (severity == null) {
						setBackground(Color.WHITE);
					} else if (severity.equals("MAJOR_ALARM")) {
						setBackground(ColorHelper.getAlarmMajorBackground());
					} else if (severity.equals("MINOR_ALARM")) {
						setBackground(ColorHelper.getAlarmMinorBackground());
					} else if (severity.equals("INVALID_ALARM")) {
						setBackground(ColorHelper.getAlarmInvalidBackground());
					}
				}

			}
		}

		return comp;
	}

	/* (non-Javadoc)
	 * @see javax.swing.JComponent#paintComponent(java.awt.Graphics)
	 */
	@Override
    protected void paintComponent(Graphics g)
	{
		super.paintComponent(g);

		paintDecorations(g);
	}

	protected void paintSuperComponent(Graphics g)
	{
		super.paintComponent(g);
	}

	protected void paintDecorations(Graphics g)
	{
		if (!active) {
			PaintHelper.paintDisabled(g, 0, 0, getWidth(), getHeight());
		}
		
		if (!decorateBorder) {
			return;
		}

		int size = getHeight();
		int y = 0;
		int x = (getWidth() - size) / 2;

		if (error && !timeout) {
			PaintHelper.paintAlarm(g, x, y, size);
		}

		if (!error && timeout) {
			PaintHelper.paintTimeout(g, x, y, size, new GregorianCalendar(1900,0,0,3,0).getTime());
		}

		if (error && timeout) {
			x = (getWidth() - 2 * size) / 2;
			PaintHelper.paintAlarm(g, x, y, size);
			x += size;
			PaintHelper.paintTimeout(g, x, y, size, new GregorianCalendar(1900,0,0,3,0).getTime());
		}

		if (alarm) {
			if (severity == null) {
				PaintHelper.paintRectangle(g, 0, 0, getWidth(), getHeight(),
				    ColorHelper.getAlarm(), 2);
			} else if (severity.equals("MAJOR_ALARM")) {
				PaintHelper.paintRectangle(g, 0, 0, getWidth(), getHeight(),
				    ColorHelper.getAlarmMajor(), 2);
			} else if (severity.equals("MINOR_ALARM")) {
				PaintHelper.paintRectangle(g, 0, 0, getWidth(), getHeight(),
				    ColorHelper.getAlarmMinor(), 2);
			} else if (severity.equals("INVALID_ALARM")) {
				PaintHelper.paintRectangle(g, 0, 0, getWidth(), getHeight(),
				    ColorHelper.getAlarmInvalid(), 2);
			}
		}
	}
	public boolean isDecorateBackground() {
		return decorateBackground;
	}
	public void setDecorateBackground(boolean decorateBackground) {
		this.decorateBackground = decorateBackground;
	}
	public boolean isDecorateBorder() {
		return decorateBorder;
	}
	public void setDecorateBorder(boolean decorateBorder) {
		this.decorateBorder = decorateBorder;
	}
}

/* __oOo__ */
