/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components.util;

import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import javax.swing.AbstractButton;
import javax.swing.Action;
import javax.swing.ButtonGroup;
import javax.swing.Icon;
import javax.swing.JButton;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JComponent;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.JRadioButtonMenuItem;
import javax.swing.JSeparator;
import javax.swing.JToggleButton;
import javax.swing.SwingConstants;


/**
 * Convenience class with set of utilities to help work with Action instances.
 * Supports enhencments to Action framework by Abeans.
 * 
 * <p>
 * If ActionList contains null action, than separator menu item is used.
 * </p>
 *
 * @author <a href="mailto:miha.kadunc@cosylab.com">Miha Kadunc</a>
 * @author <a href="mailto:igor.kriznar@cosylab.com">Igor Kriznar</a>
 * @version $id$
 */
public final class Actions
{
	/** Action value key for large icon image. */
	public static final String LARGE_ICON = "LargeIcon";

	/** Action, which is used as placeholder for separator menu item. */
	public static final Action SEPARATOR = new Action() {
			public boolean isEnabled()
			{
				return false;
			}

			public void setEnabled(boolean b)
			{
			}

			public void addPropertyChangeListener(
			    PropertyChangeListener listener)
			{
			}

			public void removePropertyChangeListener(
			    PropertyChangeListener listener)
			{
			}

			public Object getValue(String key)
			{
				if (ActionList.SEPARATOR.equals(key)) {
					return Boolean.TRUE;
				}

				return null;
			}

			public void putValue(String key, Object value)
			{
			}

			public void actionPerformed(ActionEvent e)
			{
			}
		};

	/**
	 * Creates a button using the provided action.
	 *
	 * @param action 
	 *
	 * @return new button with assigned action
	 */
	public static AbstractButton createButton(final Action action)
	{
		if (action instanceof ToggleAction) {
			final JToggleButton btn = new JToggleButton(action);
			btn.setSelected(((ToggleAction)action).isSelected());

			if (action.getValue(ToggleAction.SMALL_SELECTED_ICON) != null) {
				btn.setSelectedIcon((Icon)action.getValue(
				        ToggleAction.SMALL_SELECTED_ICON));
			}

			action.addPropertyChangeListener(new PropertyChangeListener() {
					public void propertyChange(PropertyChangeEvent evt)
					{
						if ("selected".equals(evt.getPropertyName())) {
							btn.setSelected(((Boolean)evt.getNewValue())
							    .booleanValue());
						} else if ("visible".equals(evt.getPropertyName())) {
							btn.setVisible(((Boolean)evt.getNewValue())
							    .booleanValue());
						}
					}
				});
            btn.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    boolean actSel=((ToggleAction)action).isSelected();
                    if (actSel!=btn.isSelected()) {
                        btn.setSelected(actSel);
                    }
                }
            });

			return btn;
		} else if (action instanceof ActionList) {
			JButton butt = createButton((ActionList)action);
			butt.setAction(action);

			return butt;
		} else {
			return new JButton(action);
		}
	}

	/**
	 * Constructs a button from the provided ActionList.
	 *
	 * @param info
	 *
	 * @return button with assigned action
	 */
	public static JButton createButton(ActionList info)
	{
		JButton actionButton = new JButton();

		if (info.getValue(Action.NAME) != null) {
			actionButton.setText((String)info.getValue(Action.NAME));
		}

		if (info.getValue(Action.SHORT_DESCRIPTION) != null) {
			actionButton.setToolTipText((String)info.getValue(
			        Action.SHORT_DESCRIPTION));
		}

		if (info.getValue(Action.SMALL_ICON) != null) {
			actionButton.setIcon((Icon)info.getValue(Action.SMALL_ICON));
		}

		Object actions = info.getValue(ActionList.ACTIONS);

		if (actions instanceof Action[]) {
			actionButton.addActionListener(new ListenerForMenu(
			        (Action[])actions));
		}

		return actionButton;
	}

	/**
	 * Creates new <code>JMenuItem</code>. If action is <code>null</code> or
	 * has <code>SEPARATOR</code>value <code>true</code>, <code>null</code> is
	 * returned.
	 *
	 * @param action action for which menu item should be created
	 *
	 * @return <code>JMenuItem</code> or <code>null</code> if action is
	 *         <code>null</code> or has <code>SEPARATOR</code>value
	 *         <code>true</code>
	 */
	public static JMenuItem createMenuItem(Action action)
	{
		return createMenuItem(action, false);
	}

	/**
	 * Creates new <code>JMenuItem</code> or <code>JSeparator</code>, depending
	 * on provided action. If action is <code>null</code> or has
	 * <code>SEPARATOR</code>value <code>true</code>, then is reckognized as
	 * separator and instance of <code>JSeparator</code> with provided
	 * orientation is returned.
	 * 
	 * <p>
	 * Same as <code>createMenuComponent(Action action, boolean radio, int
	 * orientation)</code> with <code>false</code> for radio and
	 * <code>SwingConstants.HORIZONTAL</code> for orientation.
	 * </p>
	 *
	 * @param action action for which menu item should be created
	 *
	 * @return <code>JMenuItem</code> or <code>JSeparator</code> if action is
	 *         <code>null</code> or has <code>SEPARATOR</code>value
	 *         <code>true</code>
	 */
	public static JComponent createMenuComponent(Action action)
	{
		return createMenuComponent(action, false, SwingConstants.HORIZONTAL);
	}

	/**
	 * Creates new <code>JMenuItem</code> or <code>JSeparator</code>, depending
	 * on provided action. If action is <code>null</code> or has
	 * <code>SEPARATOR</code>value <code>true</code>, then is reckognized as
	 * separator and instance of <code>JSeparator</code> with provided
	 * orientation is returned.
	 *
	 * @param action action for which menu item should be created
	 * @param radio if <code>true</code> then <code>JRadioButtonMenuItem</code>
	 *        is created
	 * @param orientation orientation used to
	 *
	 * @return <code>JMenuItem</code> or <code>JSeparator</code> if action is
	 *         <code>null</code> or has <code>SEPARATOR</code>value
	 *         <code>true</code>
	 */
	public static JComponent createMenuComponent(Action action, boolean radio,
	    int orientation)
	{
		JComponent comp = createMenuItem(action, radio);

		if (comp == null) {
			comp = new JSeparator(orientation);
		}

		return comp;
	}

	/**
	 * Creates new <code>JMenuItem</code>. If action is <code>null</code> or
	 * has <code>SEPARATOR</code>value <code>true</code>, <code>null</code> is
	 * returned.
	 *
	 * @param action action for which menu item should be created
	 * @param radio if <code>true</code> then <code>JRadioButtonMenuItem</code>
	 *        is created
	 *
	 * @return <code>JMenuItem</code> or <code>null</code> if action is
	 *         <code>null</code> or has <code>SEPARATOR</code>value
	 *         <code>true</code>
	 */
	public static JMenuItem createMenuItem(Action action, boolean radio)
	{
		if (isSeparator(action)) {
			return null;
		}

		if (action instanceof ToggleAction) {
			if (radio) {
				final JRadioButtonMenuItem itm = new JRadioButtonMenuItem(action);
				itm.setIcon(null);
				itm.setSelected(((ToggleAction)action).isSelected());
				action.addPropertyChangeListener(new PropertyChangeListener() {
						public void propertyChange(PropertyChangeEvent evt)
						{
							if ("selected".equals(evt.getPropertyName())) {
								itm.setSelected(((Boolean)evt.getNewValue())
								    .booleanValue());
							}
						}
					});

				return itm;
			}

			final JCheckBoxMenuItem itm = new JCheckBoxMenuItem(action);
			itm.setIcon(null);
			itm.setSelected(((ToggleAction)action).isSelected());
			action.addPropertyChangeListener(new PropertyChangeListener() {
					public void propertyChange(PropertyChangeEvent evt)
					{
						if ("selected".equals(evt.getPropertyName())) {
							itm.setSelected(((Boolean)evt.getNewValue())
							    .booleanValue());
						}
					}
				});

			return itm;
		} else if (action instanceof ActionList) {
			return createMenuItem((ActionList)action);
		} else {
			if (radio) {
				JRadioButtonMenuItem itm = new JRadioButtonMenuItem(action);

				return itm;
			}

			JMenuItem itm = new JMenuItem(action);

			return itm;
		}
	}

	/**
	 * Returns <code>true</code> if action should be mapped to separator menu
	 * item.
	 *
	 * @param action
	 *
	 * @return <code>true</code> if action should be mapped to separator menu
	 *         item
	 */
	public static boolean isSeparator(Action action)
	{
		return action == null
		|| Boolean.TRUE.equals(action.getValue(ActionList.SEPARATOR));
	}

	/**
	 * Creates a popup menu with the given actions.
	 *
	 * @param info actions to be included in the popup
	 *
	 * @return popup menu with assigned actions
	 */
	public static JPopupMenu createPopupMenu(ActionList info)
	{
		JPopupMenu theMenu = new JPopupMenu();

		if (info.getValue(Action.SHORT_DESCRIPTION) != null) {
			theMenu.setToolTipText((String)info.getValue(
			        Action.SHORT_DESCRIPTION));
		}

		if (info.getValue(Action.SHORT_DESCRIPTION) != null) {
			theMenu.setToolTipText((String)info.getValue(
			        Action.SHORT_DESCRIPTION));
		}

		Action[] act = info.toArray();
		JComponent currentItem = null;
		boolean isThereIcon = false;

		for (int i = 0; i < act.length; i++) {
			if (isSeparator(act[i])) {
				theMenu.add(new JPopupMenu.Separator());
			} else {
				currentItem = createMenuItem(act[i],
					    info.getValue(ActionList.MENU_CONTEXT) == ActionList.MENU_CONTEXT_RADIO);
				theMenu.add(currentItem);

				if (!isThereIcon && currentItem instanceof JMenuItem
				    && ((JMenuItem)currentItem).getIcon() != null) {
					isThereIcon = true;
				}
			}
		}

		// Workaround for GUI bug
		if (isThereIcon
		    && info.getValue(ActionList.MENU_CONTEXT) != ActionList.MENU_CONTEXT_RADIO) {
			Component[] c = theMenu.getComponents();

			for (int i = 0; i < c.length; i++) {
				if (c[i] instanceof JMenuItem) {
					currentItem = (JMenuItem)c[i];

					if (currentItem instanceof JMenuItem
					    && ((JMenuItem)currentItem).getIcon() == null) {
						((JMenuItem)currentItem).setIcon(IconHelper.createIcon(
						        "icons/general/Blank16.gif"));
					}
				}
			}
		}

		if (info.getValue(ActionList.MENU_CONTEXT) == ActionList.MENU_CONTEXT_RADIO) {
			Component[] c = theMenu.getComponents();
			ButtonGroup bg = new ButtonGroup();

			for (int i = 0; i < c.length; i++) {
				bg.add((AbstractButton)c[i]);
			}

			info.putValue("buttonGroup", bg);
		}

		return theMenu;
	}

	/**
	 * Creates a menu item with the given action.
	 *
	 * @param actions 
	 *
	 * @return menu item with assigned action
	 */
	public static JMenuItem createMenuItem(ActionList actions)
	{
		return createMenuItem(actions,
		    actions.getValue(ActionList.MENU_CONTEXT) == ActionList.MENU_CONTEXT_RADIO);
	}

	/**
	 * Creates a menu item with the given action.
	 *
	 * @param actions 
	 * @param radio if true the menu item will be a radio button
	 *
	 * @return menu item with assigned action
	 */
	public static JMenuItem createMenuItem(ActionList actions, boolean radio)
	{
		return addToMenu(new JMenu(), actions, radio, SwingConstants.HORIZONTAL);
	}

	/**
	 * Constructs a menu with the given actions as menu items.
	 *
	 * @param actions menu items actions
	 *
	 * @return menu with assigned actions
	 */
	public static JMenu createMenu(ActionList actions)
	{
		return addToMenu(new JMenu(), actions, false, SwingConstants.HORIZONTAL);
	}

	/**
	 * Constructs a menu with given actions as menu items.
	 *
	 * @param actions menu items actions
	 * @param orientation 
	 *
	 * @return menu with assigned actions
	 */
	public static JMenu createMenu(ActionList actions, int orientation)
	{
		return addToMenu(new JMenu(), actions, false, orientation);
	}

	/**
	 * Constructs a menu bar with the give actions.
	 *
	 * @param actions 
	 *
	 * @return menu bar with assigned action
	 */
	public static JMenuBar createMenuBar(ActionList actions)
	{
		final JMenuBar bar = new JMenuBar();
		actions.addPropertyChangeListener(new PropertyChangeListener() {
				public void propertyChange(PropertyChangeEvent evt)
				{
					if (evt.getPropertyName() == ActionList.ACTIONS) {
						updateMenuBar((Action[])evt.getNewValue(), bar);
					}
				}
			});

		return bar;
	}

	private static void updateMenuBar(Action[] a, JMenuBar bar)
	{
		bar.removeAll();

		for (int i = 0; i < a.length; i++) {
			if (a[i] instanceof ActionList) {
				bar.add(createMenu((ActionList)a[i], SwingConstants.VERTICAL));
			} else {
				bar.add(createMenuComponent(a[i], false, SwingConstants.VERTICAL));
			}
		}
	}

	private static JMenu addToMenu(JMenu theMenu, ActionList info,
	    boolean radio, int orientation)
	{
		// No text in toolbars
		if (info.getValue(Action.NAME) != null) {
			theMenu.setText((String)info.getValue(Action.NAME));
		}

		if (info.getValue(Action.SHORT_DESCRIPTION) != null) {
			theMenu.setToolTipText((String)info.getValue(
			        Action.SHORT_DESCRIPTION));
		}

		if (info.getValue(Action.SMALL_ICON) != null) {
			theMenu.setIcon((Icon)info.getValue(Action.SMALL_ICON));
		}

		Action[] act = info.toArray();

		for (int i = 0; i < act.length; i++) {
			theMenu.add(createMenuComponent(act[i], radio, orientation));
		}

		if (radio) {
			Component[] c = theMenu.getComponents();
			ButtonGroup bg = new ButtonGroup();

			for (int i = 0; i < c.length; i++) {
				bg.add((AbstractButton)c[i]);
			}

			info.putValue("buttonGroup", bg);
		}

		return theMenu;
	}

	/**
	 * DOCUMENT ME!
	 *
	 * @author $Author: jbobnar $
	 * @version $Revision: 1.13 $
	 */
	public static class ListenerForMenu implements ActionListener
	{
		private Action[] actions = null;
		private JPopupMenu popup = null;

		/**
		 * Creates a new ListenerForMenu object.
		 *
		 * @param subActions DOCUMENT ME!
		 */
		public ListenerForMenu(Action[] subActions)
		{
			actions = subActions;
		}

		/**
		 * DOCUMENT ME!
		 *
		 * @return DOCUMENT ME!
		 */
		public JPopupMenu getPopup()
		{
			if (popup == null) {
				popup = new JPopupMenu();

				for (int i = 0; i < actions.length; i++) {
					if (isSeparator(actions[i])) {
						popup.add(new JPopupMenu.Separator());
					} else {
						popup.add(createMenuItem(actions[i]));
					}
				}
			}

			return popup;
		}

		/**
		 * DOCUMENT ME!
		 *
		 * @param e DOCUMENT ME!
		 */
		public void actionPerformed(ActionEvent e)
		{
			Component invoker = (Component)e.getSource();
			getPopup().show(invoker, 0, invoker.getHeight());
		}
	}
}

/* __oOo__ */
