/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components.util;

import java.awt.Image;
import java.awt.Toolkit;
import java.awt.image.ImageProducer;

import java.io.IOException;

import java.net.URL;

import java.security.AccessController;
import java.security.PrivilegedAction;

import java.util.HashMap;


/**
 * ImageHelper acts as an interface to the all images currently in classpath.
 * ImageHelper is a singleton class that maintains internal cache to all
 * accessed images. This class may be used to minimize the memory consumption
 * where application uses static images. If images are changed independently,
 * this class should not be used.
 *
 * @author <a href="mailto:jernej.kamenik@cosylab.com">Jernej Kamenik</a>
 * @version $id$
 *
 */
public class ImageHelper {
    /** cache for images. */
    private static final HashMap<String, Image> cache = new HashMap<>();

    /**
     * Protected constructor for creation of singleton class.
     */
    private ImageHelper() {
        super();
    }

    /**
     * Returns image from the cache. If image cannot be found in cache, return
     * value will be <code>null</code>.
     *
     * @param name String
     * @return Image instance of an image or null if not found
     */
    private static Image getCachedImage(String name) {
        if (cache.containsKey(name)) {
            return cache.get(name);
        }

        return null;
    }

    /**
     * Puts image in the cache. If an image with the specified name has already been stored,
     * current image will be overwritten.
     *
     * @param name
     * @param image
     */
    private static synchronized void setCachedImage(String name, Image image) {
        cache.put(name, image);
    }

    /**
     * Returns instance of the image by loading it using the <code>
     * ClassLoader</code> instance provided. images are cached. If an image with
     * the same name is requested several times, <code>createImage</code> will
     * return the same instance each time. If images will be manipulated
     * individually, do not use this method to load images. If image is not found
     * in the classpath, return value will be null.
     *
     * @param name String
     * @param cl ClassLoader Instance of class loader to use
     * @return image Instance of image or null if image could not be found.
     */
    public static Image createimage(String name, ClassLoader cl) {
        Image image = getCachedImage(name);

        if (image == null) {
            image = loadResource(name, cl);

            if (image != null) {
                setCachedImage(name, image);
            }
        }

        return image;
    }

    /**
     * Returns instance of the image. images are cached. If an image with the same
     * name is requested several times, <code>createImage</code> will return the
     * same instance each time. If images will be manipulated individually, do
     * not use this method to load images. If image is not found in the classpath,
     * return value will be null.
     *
     * @param name String
     * @return image Instance of image or null if image could not be found.
     */
    public static Image createImage(String name) {
        return createimage(name, ImageHelper.class.getClassLoader());
    }

    /**
     * This is a utility method to help in loading image images.
     * It takes the name of a resource file and loads an image object
     * from that file.  Typically images will be GIFs.
     * <p>
     * @param resourceName  A pathname relative to the directory
     *                holding the class loader specified.  For example,
     *                "wombat.gif".
     * @param c ClassLoader.
     * @return  an image object.  May be null if the load failed.
     */
    public static Image loadResource(final String resourceName,
        final ClassLoader c) {
        try {
            ImageProducer ip = (ImageProducer) AccessController.doPrivileged(new PrivilegedAction<Object>() {
                        public Object run() {
                            URL url;

                            if ((url = c.getResource(resourceName)) == null) {
                                return null;
                            } else {
                                try {
                                    return url.getContent();
                                } catch (IOException ioe) {
                                    return null;
                                }
                            }
                        }
                    });

            if (ip == null) {
                return null;
            }

            Toolkit tk = Toolkit.getDefaultToolkit();

            return tk.createImage(ip);
        } catch (Exception ex) {
            return null;
        }
    }
}
