/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components.wheelswitch;



/**
 * Extension of the abstract wheelswitch formatter for formatting numeric strings representing
 * angular values in format of ddddmmss, where d are angular degrees, m are angular minutes in 
 * the range of 0-59 and s are angular seconds also in the range of 0-59.
 * The number of digits representing m and s is fixed to 2 per m and 2 per s, while d are assumed 
 * to take all the remaining digits up front.
 * In the format string passed to the formatter all the decimal digits are ingored and only the
 * leading (integer) digits are counted when determining the right formatted value string.
 * If the format string does not contain enough integer digits to display the whole number, new digits
 * are automatically added to the value string.
 *
 * @author <a href="mailto:jernej.kamenik@cosylab.com">Jernej Kamenik</a>
 * @version $id$
 */
public class RadialWheelswitchFormatter extends AbstractWheelswitchFormatter
{
	/**
	 * Constructs the PlainWheelswitchFormatter and sets the format  string.
	 *
	 * @param newFormatString
	 *
	 * @see #setFormat(String)
	 */
	public RadialWheelswitchFormatter(String newFormatString)
	{
		super(newFormatString);
	}

	/**
	 * Constructs the PlainWheelswitchFormatter with no format string.
	 *
	 * @see #setFormat(String)
	 */
	public RadialWheelswitchFormatter()
	{
		super();
	}

	/* (non-Javadoc)
     * @see com.cosylab.gui.components.wheelswitch.AbstractWheelswitchFormatter#setFormat(java.lang.String)
     */
	@Override
    public void setFormat(String newFormatString)
            throws IllegalArgumentException {
        if (newFormatString!=null && newFormatString.indexOf('E')>-1) throw new IllegalArgumentException("Exponential format not supported by this formatter.");
        super.setFormat(newFormatString);
    }

    private String parseString(String newValueString, boolean pad) {
        double newValue = Double.parseDouble(newValueString);
        
		//decreasing of digits has to be handled differently from increasing
		boolean reverse = false;
		if (Math.abs(value)>Math.abs(newValue)) {
		    reverse = true;
		}
		
		//determine the sign and trim the string if necessary
		int sign = 0;
		if (formatString.startsWith("+")) {
		    sign = 1;
		}
		if (newValueString.startsWith("-")) {
		    sign = -1;
		    newValueString = newValueString.substring(1);
		}
		if (newValueString.startsWith("+")) {
		    newValueString = newValueString.substring(1);
		}
		

		//the effective string length is up to the decimal digits
		int decStrLength = newValueString.length();
		int dotIndex = newValueString.indexOf('.');
		if (dotIndex>-1) decStrLength = dotIndex;
		
		//determine secs value string
		String secString = "0";
		if (decStrLength==1) secString = newValueString;
		if (decStrLength>1) secString = newValueString.substring(decStrLength-2);
		//determine mins value string
		String minString = "0";
		if (decStrLength==3) minString = newValueString.substring(0,decStrLength-2);
		if (decStrLength>3) minString = newValueString.substring(decStrLength-4,decStrLength-2);
		//determine degs value string
		String degString = "0";
		if (decStrLength>4) degString = newValueString.substring(0,decStrLength-4);
		
		//determine secs value
		double secValue = Double.parseDouble(secString);
		//determine secs overflow
		int secOff = (int)secValue/60;
		//correct the overflow
		if (secOff>=1) {
		    //if going backwards, there really is no overflow
		    //but the value has to be reduced by the difference between
		    //the decimal and the angular system
		    if (reverse) {
		        secValue -= 40;
		        secOff = 0;
		    }
		    else secValue %= 60;
		}
		//create new secs string (integer valued)
		secString = String.valueOf(secValue);
		//complete if too short
		//the effective string length is up to the decimal digits
		decStrLength = secString.length();
		dotIndex = secString.indexOf('.');
		if (dotIndex>-1) decStrLength = dotIndex;
		while (decStrLength<2) {
		    secString = "0"+secString;
		    decStrLength++;
		}
		
		//the same for mins only account for the secs overflow
		int minValue = Integer.parseInt(minString)+secOff;
		int minOff = minValue/60;
		if (minOff>0) {
		    if (reverse) {
		        minValue -= 40;
		        minOff = 0;
		    }
		    else minValue %= 60;
		}
		minString = String.valueOf(minValue);
		while (minString.length()<2) minString = "0"+minString;
		
		//the same for degs
		double degValue = Double.parseDouble(degString)+minOff;
		degString = String.valueOf((int)degValue);
		
		//assebmle the complete string
		newValueString = degString+minString+secString;		
				
		if (pad) newValueString = padString(newValueString);
		
		//add back the sign specified by the value or the format
		if (sign>0) newValueString = "+" + newValueString;
		if (sign<0) newValueString = "-" + newValueString;

		return newValueString;
    }
    
	/*
	 * correct string to adhere to format
     * @param newValueString
     */
    private String padString(String newValueString) {
		//first determine the number of digits specified by the format
        int numFormatInteger = formatString.indexOf('.');
		int numFormatDecimal = 0;
		if (numFormatInteger<0) numFormatInteger = formatString.length();
		else numFormatDecimal = formatString.length()-numFormatInteger-1;
		if (formatString.startsWith("+")) {
			numFormatInteger -= 1;
		}

		//then determine the number of actual digits currently in the string
		int numInteger = newValueString.indexOf('.');
		int numDecimal = 0;
		if (numInteger<0) numInteger = newValueString.length();
		else numDecimal = newValueString.length()-numInteger-1;
		
		//pad too short strings in front
		while (numInteger<numFormatInteger) {
		    newValueString = "0"+newValueString;
		    numInteger++;
		}
		//trim too long sring of the front zeros
		while (numInteger>numFormatInteger && newValueString.startsWith("0")) {
		    newValueString = newValueString.substring(1);
		    numInteger--;
		}
		
		//pad too short decimals at the back
		if (formatString.indexOf('.')>-1 && newValueString.indexOf('.')<0) newValueString += ".";
		while (numDecimal<numFormatDecimal) {
		    newValueString += "0";
		    numDecimal++;
		}
		//trim too long decimals of the last digits
		//all the rounding should already have been done beforehand, so one should always just trim zeros here
		while (numDecimal>numFormatDecimal) {
		    newValueString = newValueString.substring(0,newValueString.length()-1);
		    numDecimal--;
		}

		return newValueString;
    }

    @Override
    protected void internalSetString(String newValueString) {
		//System.out.println("value recieved = "+newValueString);

		//first parse the new value string only to determine the new value
		//no padding of digits neccessary
		newValueString = parseString(newValueString,false);
		//System.out.println("value stored = "+newValueString);
		value = Double.parseDouble(newValueString);
		
		//then parse again after triming unneccessary decimal digits
		newValueString = formatter.sprintf(value);
		//this is neccessary when no decimals are present
		while (newValueString.startsWith(" ")) newValueString = newValueString.substring(1);
		//System.out.println("value trimed = "+newValueString);
		//this time make proper padding of zero digits
		valueString = parseString(newValueString,true);		
		//System.out.println("value string = "+valueString);
	}
}

/* __oOo__ */
