/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans.
 *
 * CosyBeans is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.displayers;

import java.beans.PropertyVetoException;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;


/**
 * <code>DataSourceSupport</code> is conveniencee implementation base for
 * <code>DataSource</code> interface.
 *
 * @author <a href="mailto:igor.kriznar@cosylab.com">Igor Kriznar</a>
 * @version $Id: DataSourceSupport.java,v 1.11 2008-04-22 12:31:02 jbobnar Exp $
 *
 * @since Nov 24, 2003.
 */
public class DataSourceSupport implements DataSource, Serializable
{
	private static final long serialVersionUID = -7615919185970956618L;
	protected Class<DataConsumer>[] types;
	private transient List<DataConsumer> consumers;

	/**
	 * Constructs support with defined supported types.
	 *
	 * @param supportedTypes types of DataConsumer that this source will
	 *        support
	 */
	public DataSourceSupport(Class<DataConsumer>[] supportedTypes)
	{
		consumers = Collections.synchronizedList(new ArrayList<>());
		types = supportedTypes;
		for (int i = 0; i < supportedTypes.length; i++) {
			if (!DataConsumer.class.isAssignableFrom(supportedTypes[i])) {
				throw new IllegalArgumentException("Parameter type '"
				    + supportedTypes[i].getName() + "' is not DataConsumer.");
			}
		}

	}

	/**
	 * This support implementation does not perform any type check on consumer,
	 * it just adds it to list of consumers.
	 *
	 * @see com.cosylab.gui.displayers.DataSource#addConsumer(com.cosylab.gui.displayers.DataConsumer)
	 */
	public void addConsumer(DataConsumer consumer) throws PropertyVetoException
	{
		consumers.add(consumer);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataSource#removeConsumer(com.cosylab.gui.displayers.DataConsumer)
	 */
	public void removeConsumer(DataConsumer consumer)
	{
		if (consumers == null) return;
		consumers.remove(consumer);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataSource#getConsumers()
	 */
	public DataConsumer[] getConsumers()
	{
		if (consumers == null) return new DataConsumer[0];
		return consumers.toArray(new DataConsumer[consumers.size()]);
	}

	/**
	 * Clears all consumers from this DataSource.
	 */
	public void removeAllConsumers()
	{
		if (consumers != null) {
			consumers.clear();
		}
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataSource#getSupportedConsumersTypes()
	 */
	public Class<DataConsumer>[] getAcceptableConsumerTypes()
	{
		return types;
	}

	/**
	 * Form all contained consumers extracts supported characteristics.
	 *
	 * @return characteristic names from all contained consumers
	 */
	public String[] extractSupportedCharacteristics()
	{
		// recast consumers to new type
		DataConsumer[] c = getConsumers();
		Set<String> names = new HashSet<>();

		for (int i = 0; i < c.length; i++) {
			String[] s = c[i].getSupportedCharacteristics();

			if (s != null) {
				for (int j = 0; j < s.length; j++) {
					names.add(s[j]);
				}
			} else {
				return null;
			}
		}
		return names.toArray(new String[names.size()]);
	}

	/**
	 * Releases all consumers and goes into initial state.
	 */
	public void clear()
	{
		consumers.clear();
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Object#clone()
	 */
	@SuppressWarnings("unchecked")
	@Override
	protected Object clone() throws CloneNotSupportedException {
		DataSourceSupport c= (DataSourceSupport)super.clone();
		if (types!=null) {
			c.types= new Class[types.length];
			System.arraycopy(types,0,c.types,0,types.length);
		}
		c.consumers= new ArrayList<>();
		return c;
	}
}

/* __oOo__ */
