/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans.
 *
 * CosyBeans is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.displayers;

import java.io.Serializable;
import java.text.SimpleDateFormat;
import java.util.Date;


/**
 * <code>DataState</code> describes quality of dinamic value.
 *
 * @author <a href="mailto:igor.kriznar@cosylab.com">Igor Kriznar</a>
 * @version $Id: DataState.java,v 1.8 2008-04-22 12:31:02 jbobnar Exp $
 *
 * @since Nov 24, 2003.
 */
public class DataState implements Serializable
{
	private static final long serialVersionUID = -7654199155702253489L;

	private static SimpleDateFormat dateTimeFormatter = new SimpleDateFormat(
		    "yyyy-MM-dd HH:mm:ss.SSS");

	/**
	 * Type identifying this data quality state. This DataState objec defines
	 * several usefull states, but there is not limitation to define new
	 * identifiers in contract data source and consumer implementation.
	 */
	public final String type;

	/** Timestamp of data quality */
	public final long timestamp;

	/** Detailed description of this data quality representativ. */
	public final String description;
	
	/** Secondary type which can be used for the state of the write connection if separate from the monitor */
	public final String secondaryType;
	
	/** Detailed description of the secondary data quality */
	public final String secondaryDescription;
	
	/** Timestamp of the secondary data quality change */
	public final long secondaryTimestamp;

	/** State when dynamic value source receives data with timelag. */
	public static final String TIMELAG = "Timelag";

	/** State when dynamic value data is in timeout. */
	public static final String TIMEOUT = "Timeout";

	/** State when warning condition was rased for the dynamic value. */
	public static final String WARNING = "Warning";

	/** State when alarm condition was rased for the dynamic value. */
	public static final String ALARM = "Alarm";

	/** State when error condition was rased for the dynamic value. */
	public static final String ERROR = "Error";

	/** State for normal conditions. */
	public static final String NORMAL = "Normal";

	/** State for not initialized source of dynamic data. */
	public static final String NOT_INITIALIZED = "Not Initialized";

	/** State for connected source of dynamic data. */
	public static final String CONNECTED = "Connected";

	/** State for disconnected source of dynamic data. */
	public static final String DISCONNECTED = "Disconnected";

	/** State when data quality can not be determined. */
	public static final String UNDEFINED = "Undefined";

	/** State when dynamic data source ahs been suspended. */
	public static final String SUSPENDED = "Suspended";

	/**
     * Creates a new DataState object.
     *
     * @param type the string identifier for type.
     * @param timestamp the time of data quality change
     * @param description detailed description of this data quality state
     * @param secondaryType the secondary type, which can describe the state of the write link 
     *          if separate from monitor
     * @param secondaryTimestamp the time of the secondary data quality change
     * @param secondaryDescription the description of the secondary data quality state
     */
    public DataState(String type, long timestamp, String description, 
            String secondaryType, long secondaryTimestamp, String secondaryDescription)
    {
        this.type = type;
        this.timestamp = timestamp;
        this.description = description;
        this.secondaryType = secondaryType;
        this.secondaryDescription = secondaryDescription;
        this.secondaryTimestamp = secondaryTimestamp;
    }
	
	/**
	 * Creates a new DataState object.
	 *
	 * @param type the string identifier for type.
	 * @param timestamp the time of data qyality change
	 * @param description detailed description of this data quality state
	 */
	public DataState(String type, long timestamp, String description)
	{
		this(type,timestamp,description,UNDEFINED,System.currentTimeMillis(),null);
	}

	/**
	 * Creates a new DataState object. The timestamp is set to time of object
	 * creation.
	 *
	 * @param type the string identifier for type.
	 * @param description detailed description of this data quality state
	 */
	public DataState(String type, String description)
	{
		this(type, System.currentTimeMillis(), description);
	}

	/**
	 * Creates a new DateState object. The timestamp is set to time of object
	 * creation.
	 *
	 * @param type the string identifier for type.
	 */
	public DataState(String type)
	{
		this(type, System.currentTimeMillis(), null);
	}

	/**
	 * Returns the type identifyer of this data quality state.
	 *
	 * @return the type identifyer of this data quality state
	 */
	@Override
    public String toString()
	{
		StringBuffer sb = new StringBuffer();
		sb.append(type);
		sb.append(" = { ");
		sb.append(dateTimeFormatter.format(new Date(timestamp)));

		if (description != null) {
			sb.append(" \"");
			sb.append(description);
			sb.append("\"");
		}

		sb.append(" }");

		return sb.toString();
	}
	
	/**
	 * Returns the type of this DataState.
	 * 
	 * @return
	 */
	public String getType() {
	    return type;
	}
}

/* __oOo__ */
