/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans.
 *
 * CosyBeans is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.displayers;

import java.beans.PropertyVetoException;

/**
 * <code>Displayer</code> interface describes GUI component, which is able to
 * dispaly on screen single dynamic data source.
 * 
 * <p>
 * If displayer GUI commponent is editable, this means can receive input from
 * the end user, then such displayer will want to send this value change to the
 * datasource back. In this case the displayer must implement also
 * <code>DataSource</code> interface and accept consumers of type
 * <code>NonblockingNumberConsumer</code> at least or other
 * <code>Nonblocking&lt;TYPPE&gt;Consumer</code>,
 * <code>Synchronous&lt;TYPPE&gt;Consumer</code> and
 * <code>Asynchronous&lt;TYPPE&gt;Consumer</code> consumers. Displayer then
 * sends user data change request to this consumers.
 * </p>
 * 
 * @author <a href="mailto:igor.kriznar@cosylab.com">Igor Kriznar</a>
 * @version $Id: Displayer.java,v 1.9 2008-04-22 12:31:02 jbobnar Exp $
 * 
 * @since Nov 24, 2003.
 */
public interface Displayer extends DataConsumer, CommonDisplayer {
	/** Java Bean property name for DataSource property. */
	public static final String DATA_SOURCE = "dataSource";

	/**
     * Characteristic name for editable displayer JavaBean property. Displayers
     * which support value editing <b>MUST</b> implement also setter
     * <code>setEditable(booelan)</code> and support this characteristic name.
     * 
     * @see #isEditable()
     */
	public static final String C_EDITABLE = "editable";

	/**
     * <p>
     * Returns <code>true</code> if this displayer can interact with user and
     * fire value updates requested by user. If displayer is editable, mus
     * implement <code>DataSource</code> interface.
     * </p>
     * 
     * <p>
     * Displayers, which does not support value editing will return allways
     * <code>false</code>. Displayers which support value editing <b>MUST</b>
     * implement also setter <code>setEditable(booelan)</code>.
     * </p>
     * 
     * @return <code>true</code> if this displayer can receive value change
     *         requestes by user
     */
	public boolean isEditable();

	/**
     * Returns the title of this displayer. Title migth be used to identfy which
     * particular remote entity this is displayed.
     * 
     * @return the title of this displayer
     */
	public String getTitle();

	/**
     * Sets the title of this displayer. Title migth be used to identfy which
     * particular remote entity this is displayed. Setting to <code>null</code>
     * will hide title.
     * 
     * @param title
     *        new title of this displayer, can be <code>null</code>
     */
	public void setTitle(String title);

	/**
     * Sets data source and registeres this displayer as data consumer. This is
     * convenience method for integration with VCE development tools, you can
     * connect displayers directly to data sources without using this method.
     * 
     * @param dataSource
     *        new data source of this displayer
     * 
     * @throws PropertyVetoException
     *         DOCUMENT ME!
     */
	public void setDataSource(DataSource dataSource)
	        throws PropertyVetoException;

	/**
     * Returns data source. This is convenience method for integration with VCE
     * development tools. If <code>null</code> returned does not mean that
     * this displayer is not connected to data source, it only means that
     * <code>setDataSource()</code> method was not used.
     * 
     * @return the data source for this displayer, if set by setter
     */
	public DataSource getDataSource();

	/* 
	 * Below is try how to support DnD in Displayer. Ike is not sure what 
	 * to do with it at the moment.
	 */

	/*
     * Set the URI of the data source being displayed by this displayer. The
     * URIs are resoved to data sources ({@link DataSource}) through the
     * {@link DataSourceResolver} interface.
     * 
     * @param uri
     *        The URI of the data source.
     */
	//public void setDataSourceURI(String uri);
	 
	/*
     * Get the URI of the data source currently being displayed by this
     * displayer.
     * 
     * @return The URI of the data source.
     */
	//public String getDataSourceURI();

	/*
     * Set the settings of the displayer. The settings define how the displayer
     * renders the values from a data source (e.g., minimum and maximum values,
     * color-coding, etc.).
     * 
     * @param settings
     *        Settings to apply to the displayer.
     */
	//public void setDisplayerSettings(DisplayerSettings settings);

	/*
     * Fill-in the settings currently applied to the displayer. Note that this
     * method does not reset the {@link DisplayerSettings} object passed as the
     * parameter, but just sets/overrides the settings already there.
     * 
     * @param settings
     *        The settings object to fill-in. If <code>null</code>, a new
     *        {@link DisplayerSettings} object is created and returned.
     * @return The settings object that was filled-in with this displayer's
     *         settings.
     */
	//public DisplayerSettings getDisplayerSettings(DisplayerSettings settings);
}

/* __oOo__ */
