/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans.
 *
 * CosyBeans is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.displayers;

import java.beans.PropertyVetoException;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.Map;

import javax.swing.JComponent;

import com.cosylab.gui.adapters.Converter;
import com.cosylab.gui.util.UserSettingsProtection;
import com.cosylab.util.BitCondition;

/**
 * <code>DisplayerUtilities</code> contains set of usefull methods and fields
 * which help to implement various <code>Displayer</code> interfaces.
 *
 * @author <a href="mailto:igor.kriznar@cosylab.com">Igor Kriznar</a>
 * @version $Id: DisplayerUtilities.java,v 1.19 2008-04-22 12:31:02 jbobnar Exp $
 *
 * @since Dec 3, 2003.
 */
public final class DisplayerUtilities
{
	private final static boolean debug = false;

	/**
	 * List of commonly supporter characteristics for numeric displayers
	 * (double, long).
	 */
	public static final String[] COMMON_NUMERIC_DISPLAYER_CHARACTERISTICS = {
			CommonDisplayer.C_MINIMUM, CommonDisplayer.C_MAXIMUM,
			CommonDisplayer.C_FORMAT, CommonDisplayer.C_UNITS,
			CommonDisplayer.C_DISPLAY_NAME, Displayer.C_EDITABLE
		};

	/**
	 * List of commonly supported Java Bean properties for numeric displayers
	 * (double, long).
	 */
	public static final String[] COMMON_NUMERIC_DISPLAYER_PROPERTIES= {
		CommonDisplayer.C_MINIMUM, CommonDisplayer.C_MAXIMUM,
		CommonDisplayer.C_FORMAT, CommonDisplayer.C_UNITS,
		"title", Displayer.C_EDITABLE
	};

	/**
	 * List of commonly supporter characteristics for numeric displayers
	 * (double, long).
	 */
	public static final String[] PATTERN_DISPLAYER_CHARACTERISTICS = {
		CommonDisplayer.C_DISPLAY_NAME,
		CommonDisplayer.C_BIT_DESCRIPTIONS,
		CommonDisplayer.C_BIT_MASK,
		CommonDisplayer.C_CONDITION_WHEN_CLEARED,
		CommonDisplayer.C_CONDITION_WHEN_SET
		};

	/** List of commonly supporter characteristics for object displayer. */
	public static final String[] COMMON_OBJECT_DISPLAYER_CHARACTERISTICS = {  };

	/**
	 * Combines two arrays of characteristics names in to one.
	 *
	 * @param char1 array to be combined
	 * @param char2 array to be combined
	 *
	 * @return conbined array of characteristic names
	 */
	public static final String[] combineCharacteristics(String[] char1,
	    String[] char2)
	{
		ArrayList<String> l = new ArrayList<>((char1 != null ? char1.length : 0)
			    + (char2 != null ? char2.length : 0));

		if (char1 != null) {
			for (int i = 0; i < char1.length; i++) {
				l.add(char1[i]);
			}
		}

		if (char2 != null) {
			for (int i = 0; i < char2.length; i++) {
				l.add(char2[i]);
			}
		}

		return l.toArray(new String[l.size()]);
	}

	/**
	 * Combines arrays of characteristics names with singe characteristic name
	 * in to one.
	 *
	 * @param char1 array to be combined
	 * @param charName characteeristic name to be combined
	 *
	 * @return conbined array of characteristic names
	 */
	public static final String[] combineCharacteristics(String[] char1,
	    String charName)
	{
		ArrayList<String> l = new ArrayList<>((char1 != null ? char1.length : 0)
			    + (charName != null ? 1 : 0));

		if (char1 != null) {
			for (int i = 0; i < char1.length; i++) {
				l.add(char1[i]);
			}
		}

		if (charName != null) {
			l.add(charName);
		}

		return l.toArray(new String[l.size()]);
	}

	private DisplayerUtilities()
	{
		super();
	}

	/**
	 * Sets characteristics from <code>Map</code> to the displayer.
	 *
	 * @param characteristics the Map with characteristics
	 * @param disp the displayer to  which characteristics will be set
	 * @throws IllegalAccessException 
	 * @throws InvocationTargetException 
	 * @throws NoSuchMethodException 
	 */
	public static final void setCharacteristics(Map characteristics,
	    DoubleDisplayer disp)
	{
		if (debug) {
			System.out.println("[DispUtilities::setChar] setting to '" + disp
			    + "' characteristics " + characteristics);
		}

		disp.suspend();

		Double min = 0.0;
		Double max = 1.0;
	
		Object oMin = characteristics.get(CommonDisplayer.C_MINIMUM);

		if (oMin != null) {
			min = (((Number)oMin).doubleValue());
		}

		Object oMax = characteristics.get(CommonDisplayer.C_MAXIMUM);

		if (oMax != null) {
			max = (((Number)oMax).doubleValue());
		}

		if (oMax != null && max<disp.getMinimum()) {
			if (oMin != null) {
				try {
					UserSettingsProtection.setUnprotected((JComponent)disp,CommonDisplayer.C_MINIMUM,min);
				} catch (Exception e) {
					e.printStackTrace();
				}
				disp.setValue(min);
			} else {
				try {
					UserSettingsProtection.setUnprotected((JComponent)disp,CommonDisplayer.C_MINIMUM,max);
				} catch (Exception e) {
					e.printStackTrace();
				}
				disp.setValue(max);
			}
				try {
					UserSettingsProtection.setUnprotected((JComponent)disp,CommonDisplayer.C_MAXIMUM,max);
				} catch (Exception e) {
					e.printStackTrace();
				}
		} else if (oMin != null && min>disp.getMaximum()) {
			if (oMax != null) {
				try {
					UserSettingsProtection.setUnprotected((JComponent)disp,CommonDisplayer.C_MAXIMUM,max);
				} catch (Exception e) {
					e.printStackTrace();
				}
			} else {
				try {
					UserSettingsProtection.setUnprotected((JComponent)disp,CommonDisplayer.C_MAXIMUM,min);
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
			disp.setValue(min);
			try {
				UserSettingsProtection.setUnprotected((JComponent)disp,CommonDisplayer.C_MINIMUM,min);
			} catch (Exception e) {
				e.printStackTrace();
			}
		} else {
			if (oMin != null) {
				try {
					UserSettingsProtection.setUnprotected((JComponent)disp,CommonDisplayer.C_MINIMUM,min);
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
			if (oMax != null) {
				try {
					UserSettingsProtection.setUnprotected((JComponent)disp,CommonDisplayer.C_MAXIMUM,max);
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
		}
		
		Object o = characteristics.get(CommonDisplayer.C_FORMAT);

		if (o != null) {
			try {
				UserSettingsProtection.setUnprotected((JComponent)disp,CommonDisplayer.C_FORMAT,o.toString());
			} catch (Exception e) {
				e.printStackTrace();
			}
		}

		o = characteristics.get(CommonDisplayer.C_UNITS);

		if (o != null) {
			try {
				UserSettingsProtection.setUnprotected((JComponent)disp,CommonDisplayer.C_UNITS,o.toString());
			} catch (Exception e) {
				e.printStackTrace();
			}
		}

		o = characteristics.get(CommonDisplayer.C_DISPLAY_NAME);

		if (o != null) {
			try {
				UserSettingsProtection.setUnprotected((JComponent)disp,"title",o.toString());
			} catch (Exception e) {
				e.printStackTrace();
			}
		}

		/* Displayer officially does not support set but by JavaBeans contract
		 * some Displayers migth support also setter.*/
		o = characteristics.get(Displayer.C_EDITABLE);

		if (o != null) {
			try {
				UserSettingsProtection.setUnprotected((JComponent)disp,"editable",o);
			} catch (NoSuchMethodException ex) {
				// do nothing
			} catch (IllegalAccessException ex) {
				ex.printStackTrace();
			} catch (InvocationTargetException ex) {
				ex.printStackTrace();
			}
		}

		disp.resume();

		if (debug) {
			System.out.println("[DispUtilities::setChar] test: min= "
			    + disp.getMinimum() + " max= " + disp.getMaximum() + " title= "
			    + disp.getTitle());
		}
	}

	/**
	 * Sets characteristics from <code>Map</code> to the displayer.
	 *
	 * @param characteristics the Map with characteristics
	 * @param disp the displayer to  which characteristics will be set
	 * @throws IllegalAccessException 
	 * @throws InvocationTargetException 
	 * @throws NoSuchMethodException 
	 */
	public static final void setCharacteristics(Map characteristics,
	    LongDisplayer disp)
	{
		if (debug) {
			System.out.println("[DispUtilities::setChar] setting to '" + disp
			    + "' characteristics " + characteristics);
		}

		disp.suspend();

		long min = 0;
		long max = 1;
		
		Object oMin = characteristics.get(CommonDisplayer.C_MINIMUM);

		if (oMin != null) {
			min = (((Number)oMin).longValue());
		}

		Object oMax = characteristics.get(CommonDisplayer.C_MAXIMUM);

		if (oMax != null) {
			max = (((Number)oMax).longValue());
		}

		if (oMax != null && max<disp.getMinimum()) {
			if (oMin != null) {
				try {
					UserSettingsProtection.setUnprotected((JComponent)disp,CommonDisplayer.C_MINIMUM,min);
				} catch (Exception e) {
					e.printStackTrace();
				}
				disp.setValue(min);
			} else {
				try {
					UserSettingsProtection.setUnprotected((JComponent)disp,CommonDisplayer.C_MINIMUM,max);
				} catch (Exception e) {
					e.printStackTrace();
				}
				disp.setValue(max);
			}
			try {
				UserSettingsProtection.setUnprotected((JComponent)disp,CommonDisplayer.C_MAXIMUM,max);
			} catch (Exception e) {
				e.printStackTrace();
			}
		} else if (oMin != null && min>disp.getMaximum()) {
			if (oMax != null) {
				try {
					UserSettingsProtection.setUnprotected((JComponent)disp,CommonDisplayer.C_MAXIMUM,max);
				} catch (Exception e) {
					e.printStackTrace();
				}
			} else {
				try {
					UserSettingsProtection.setUnprotected((JComponent)disp,CommonDisplayer.C_MAXIMUM,min);
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
			disp.setValue(min);
			try {
				UserSettingsProtection.setUnprotected((JComponent)disp,CommonDisplayer.C_MINIMUM,min);
			} catch (Exception e) {
				e.printStackTrace();
			}
		} else {
			if (oMin != null) {
				try {
					UserSettingsProtection.setUnprotected((JComponent)disp,CommonDisplayer.C_MINIMUM,min);
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
			if (oMax != null) {
				try {
					UserSettingsProtection.setUnprotected((JComponent)disp,CommonDisplayer.C_MAXIMUM,max);
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
		}
		
		Object o = characteristics.get(CommonDisplayer.C_FORMAT);

		if (o != null) {
			try {
				UserSettingsProtection.setUnprotected((JComponent)disp,CommonDisplayer.C_FORMAT,o.toString());
			} catch (Exception e) {
				e.printStackTrace();
			}
		}

		o = characteristics.get(CommonDisplayer.C_UNITS);

		if (o != null) {
			try {
				UserSettingsProtection.setUnprotected((JComponent)disp,CommonDisplayer.C_UNITS,o.toString());
			} catch (Exception e) {
				e.printStackTrace();
			}
		}

		o = characteristics.get(CommonDisplayer.C_DISPLAY_NAME);

		if (o != null) {
			try {
				UserSettingsProtection.setUnprotected((JComponent)disp,"title",o.toString());
			} catch (Exception e) {
				e.printStackTrace();
			}
		}

		/* Displayer officially does not support set but by JavaBeans contract
		 * some Displayers migth support also setter.*/
		o = characteristics.get(Displayer.C_EDITABLE);

		if (o != null) {
			try {
				UserSettingsProtection.setUnprotected((JComponent)disp,"editable",o);
			} catch (NoSuchMethodException ex) {
				// do nothing
			} catch (IllegalAccessException ex) {
				ex.printStackTrace();
			} catch (InvocationTargetException ex) {
				ex.printStackTrace();
			}
		}

		disp.resume();

		if (debug) {
			System.out.println("[DispUtilities::setChar] test: min= "
			    + disp.getMinimum() + " max= " + disp.getMaximum() + " title= "
			    + disp.getTitle());
		}
	}

	/**
	 * Sets characteristics from <code>Map</code> to the displayer.
	 *
	 * @param characteristics the Map with characteristics
	 * @param disp the displayer to  which characteristics will be set
	 * @throws IllegalAccessException 
	 * @throws InvocationTargetException 
	 * @throws NoSuchMethodException 
	 */
	public static final void setCharacteristics(Map characteristics,
	    PatternDisplayer disp)
	{
		if (debug) {
			System.out.println("[DispUtilities::setChar] setting to '" + disp
			    + "' characteristics " + characteristics);
		}

		disp.suspend();

		Object o = characteristics.get(CommonDisplayer.C_FORMAT);

		if (o != null) {
			try {
				UserSettingsProtection.setUnprotected((JComponent)disp,CommonDisplayer.C_FORMAT,o.toString());
			} catch (Exception e) {
				e.printStackTrace();
			}
		}

		o = characteristics.get(CommonDisplayer.C_UNITS);

		if (o != null) {
			try {
				UserSettingsProtection.setUnprotected((JComponent)disp,CommonDisplayer.C_UNITS,o.toString());
			} catch (Exception e) {
				e.printStackTrace();
			}
		}

		o = characteristics.get(CommonDisplayer.C_DISPLAY_NAME);

		if (o != null) {
			try {
				UserSettingsProtection.setUnprotected((JComponent)disp,"title",o.toString());
			} catch (Exception e) {
				e.printStackTrace();
			}
		}

		/* Displayer officially does not support set but by JavaBeans contract
		 * some Displayers migth support also setter.*/
		o = characteristics.get(Displayer.C_EDITABLE);

		if (o != null) {
			try {
				UserSettingsProtection.setUnprotected((JComponent)disp,"editable",o);
			} catch (NoSuchMethodException ex) {
				// do nothing
			} catch (IllegalAccessException ex) {
				ex.printStackTrace();
			} catch (InvocationTargetException ex) {
				ex.printStackTrace();
			}
		}

		o = characteristics.get(CommonDisplayer.C_BIT_MASK);

		if (o != null) {
			try {
				UserSettingsProtection.setUnprotected((JComponent)disp,"bitMask",(BitSet)o);
			} catch (Exception e) {
				e.printStackTrace();
			}
		}

		o = characteristics.get(CommonDisplayer.C_CONDITION_WHEN_CLEARED);

		if (o != null) {
			try {
				UserSettingsProtection.setUnprotected((JComponent)disp,"conditionWhenCleared",(BitCondition[])o);
			} catch (Exception e) {
				e.printStackTrace();
			}
		}

		o = characteristics.get(CommonDisplayer.C_CONDITION_WHEN_SET);

		if (o != null) {
			try {
				UserSettingsProtection.setUnprotected((JComponent)disp,"conditionWhenSet",(BitCondition[])o);
			} catch (Exception e) {
				e.printStackTrace();
			}
		}

		o = characteristics.get(CommonDisplayer.C_BIT_DESCRIPTIONS);

		if (o != null) {
			try {
				UserSettingsProtection.setUnprotected((JComponent)disp,"bitDescriptions",(String[])o);
			} catch (Exception e) {
				e.printStackTrace();
			}
		}

		disp.resume();
	}

	/**
	 * Trys to transform double format string to long format string.
	 *
	 * @param format double prinf format string
	 *
	 * @return possibly safer long compatible format string
	 */
	public static String toLongFormat(String format)
	{
		if (format != null && (format.indexOf('f') > -1)) {
			format = format.replace('f', 'd');

			int dot = format.indexOf('.');

			if (dot > -1) {
				format = format.substring(0, dot)
					+ format.substring(format.indexOf('d'));
			}
		}

		return format;
	}
	
	/**
	 * This method can be used from <code>setConverter()</code> method from convertible displayer, which
	 * is passed as parameter. This method must be called before new converter was 
	 * stored in local field and after this method new converter must be stored in local field. 
	 * 
	 * @param converter new converter to be set to displayer
	 * @param disp displayer which receives new converter
	 * @throws PropertyVetoException 
	 */
	public static void prepareNewConverter(Converter converter, ConvertibleDisplayer disp) throws PropertyVetoException {
		reattachDataTransport(disp.getConverter(), converter, disp.getDataSource(), disp.getDataSource(), disp);
	}

	/**
	 * @param oldConverter
	 * @param newConverter
	 * @param oldDataSource
	 * @param displayer
	 * @throws PropertyVetoException
	 */
	private static void reattachDataTransport(Converter oldConverter, Converter newConverter, DataSource oldDataSource, DataSource newDataSource, Displayer displayer) throws PropertyVetoException {
		// Unregister old datasource
		if (oldDataSource!=null) {
			oldDataSource.removeConsumer(displayer);
			if (oldConverter!=null) {
				oldDataSource.removeConsumer(oldConverter);
			}
		}
		// Unregister old converter
		if (oldConverter!=null) {
			oldConverter.removeConsumer(displayer);
		}

		// register new converter
		if (newDataSource!=null) {
			if (newConverter!=null) {
				// insert new converter betwean displayer and datasource 
				newConverter.addConsumer(displayer);
				newDataSource.addConsumer(newConverter);
			} else {
				// there is no new converter, reattach dispalyer to datasource
				newDataSource.addConsumer(displayer);
			}
		}
	}
	
	/**
	 * This method can be used from <code>setDataSource()</code> method from displayer, which
	 * is passed as parameter. This method must be called before new datasource was 
	 * stored in local field and after this method new datasource must be stored in local field. 
	 * 
	 * @param dataSource new datasource to be set to displayer
	 * @param disp displayer which receives new converter
	 * @throws PropertyVetoException 
	 */
	public static void prepareNewDataSource(DataSource dataSource, ConvertibleDisplayer disp) throws PropertyVetoException {
		reattachDataTransport(disp.getConverter(), disp.getConverter(), disp.getDataSource(), dataSource, disp);
	}

	/**
	 * This method can be used from <code>setDataSource()</code> method from displayer, which
	 * is passed as parameter. This method must be called before new datasource was 
	 * stored in local field and after this method new datasource must be stored in local field. 
	 * 
	 * @param dataSource new datasource to be set to displayer
	 * @param disp displayer which receives new converter
	 * @throws PropertyVetoException 
	 */
	public static void prepareNewDataSource(DataSource dataSource, Displayer disp) throws PropertyVetoException {
		reattachDataTransport(null, null, disp.getDataSource(), dataSource, disp);
	}
	
	/**
	 * Extracts single value from an array. If index is over array size, then returns 
	 * value from index position closest to requested index.
	 * 
	 * @param index an index of value to return
	 * @param array source array
	 * @return value at specified index if valud or closes valid index
	 */
	public static double extract(int index, double[] array) {
		if (index<0) {
			return array[0];
		}
		if (index>=array.length) {
			return array[array.length-1];
		}
		return array[index];
	}
	
	/**
	 * Extracts single value from an array. If index is over array size, then returns 
	 * value from index position closest to requested index.
	 * 
	 * @param index an index of value to return
	 * @param array source array
	 * @return value at specified index if valud or closes valid index
	 */
	public static String extract(int index, String[] array) {
		if (index<0) {
			return array[0];
		}
		if (index>array.length) {
			return array[array.length];
		}
		return array[index];
	}
	
	/**
	 * Converts the array to string. This method is used displayer to transform
	 * a sequence to string.
	 * 
	 * @param value
	 * @return
	 */
	public static String toString(Object[] value) {
		StringBuilder s= new StringBuilder(16*value.length);
		s.append('[');
		if (value.length>0) {
			s.append(value[0]);
		}
		for (int i = 1; i < value.length; i++) {
			s.append(',');
			s.append(value[i]);
		}
		s.append(']');
		return s.toString();
	}

}

/* __oOo__ */
