/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans.
 *
 * CosyBeans is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.displayers;

/**
 * A specialization of the <code>Displayer</code> interface for Java primitive
 * type <code>double</code>.
 * 
 */
public interface DoubleDisplayer extends Displayer, DoubleConsumer
{
	/**
	 * Sets the value displayed by the displayer. This method may be invoked
	 * only by the adapter delivering a new value from the modeling layer.
	 *
	 * @param value new value to be displayed
	 */
	void setValue(double value);

	/**
	 * Gets the value displayed by the displayer. This method can be invoked by
	 * any object and will probably most often be invoked by the GUI rendition
	 * code.
	 *
	 * @return double currently displayed value
	 */
	public double getValue();

	/**
	 * Returns the lower limit for the value used to render the display. This
	 * lower limit must be lower or equal to the actual limit that the value
	 * can take. If may only be lower (and not equal) in order to round the
	 * minimum to some easily depictable value.
	 *
	 * @return double the lower limit for the value rendition
	 */
	public double getMinimum();

	/**
	 * Returns the upper limit for the value used to render the display. This
	 * upper limit must be greater or equal to the actual upper limit that
	 * the value can take. It may only be greater (and not equal) in order to
	 * round the maximum to some easiliy depictable value.
	 *
	 * @return double the upper limit for the value rendition
	 */
	public double getMaximum();

	/**
	 * Returns the physical units of this value represented by a string.
	 *
	 * @return String the units of the value, to be rendered, if possible, near
	 *         the actual value
	 */
	public String getUnits();

	/**
	 * Returns the C-style format string that is used to format the
	 * <code>double</code> value into text, if the displayer must perform such
	 * rendition.
	 *
	 * @return String printf style format specification
	 */
	public String getFormat();

	/**
	 * Sets the minimum value used for displaying in range.
	 *
	 * @param double new displaying minimum
	 *
	 * @see #getMinimum
	 */
	public void setMinimum(double value);

	/**
	 * Sets the maximum value used for displaying in range. 
	 *
	 * @param double new displaying maximum
	 *
	 * @see #getGraphMax
	 */
	public void setMaximum(double value);

	/**
	 * Sets the physical units for this value. Manly called by the adapter or
	 * user during  design time.
	 *
	 * @param String units for the value
	 *
	 * @see #getUnits
	 */
	public void setUnits(String value);

	/**
	 * Sets the C printf style format specification used to render double value
	 * into a String. Mainly called by the adapter or user during design time.
	 *
	 * @param String C-style format specification
	 *
	 * @see #getFormat
	 */
	public void setFormat(String value);
}

/* __oOo__ */
