/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans.
 *
 * CosyBeans is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.displayers;

import java.util.Map;

import com.cosylab.util.CommonException;


/**
 * <code>DoubleSeqConsumerMulticaster</code> casts different data consumers to
 * DoubleSeqConsumer. ObjectSeqConsumer cast must receive <code>Number</code>
 * instances. Delegate must be set, before any other method can be used.
 *
 * @author <a href="mailto:igor.kriznar@cosylab.com">Igor Kriznar</a>
 * @version $Id: DoubleSeqConsumerMulticaster.java,v 1.5 2008-04-22 12:31:02 jbobnar Exp $
 *
 * @since Dec 4, 2003.
 */
public class DoubleSeqConsumerMulticaster implements DoubleSeqConsumer,
	LongSeqConsumer, StringSeqConsumer, ObjectSeqConsumer
{
	/** Types of data consumers, that are supported by this multicaster. */
	public static final Class[] SUPPORTED_CONSUMER_TYPES = {
		DoubleSeqConsumer.class, LongSeqConsumer.class, ObjectSeqConsumer.class,
		StringSeqConsumer.class
	};

	/**
	 * Types of data consumers, that should be prefered by DoubleSeqConsumer
	 * parent.
	 */
	public static final Class[] PREFERED_CONSUMER_TYPES = {
		DoubleSeqConsumer.class, LongSeqConsumer.class
	};
	DoubleSeqConsumer delegate;

	/**
	 * If multicaster supports type, new multicaster is created, otherwise
	 * <code>null</code>is returned.
	 *
	 * @param type the requested consumer type
	 * @param delegate the consumer to which all call sare delegated
	 *
	 * @return new multicaster if type is supported, otherwise
	 *         <code>null</code>
	 */
	public static final DoubleSeqConsumer createDataConsumer(Class type,
	    DoubleSeqConsumer delegate)
	{
		if (type == DoubleSeqConsumer.class) {
			return delegate;
		}

		if (type == LongSeqConsumer.class || type == StringSeqConsumer.class
		    || type == ObjectSeqConsumer.class) {
			return new DoubleSeqConsumerMulticaster(delegate);
		}

		return null;
	}

	/**
	 * Creates a new DoubleConsumerMulticaster object.
	 *
	 * @param delegate the consumer to which all call sare delegated
	 */
	public DoubleSeqConsumerMulticaster(DoubleSeqConsumer delegate)
	{
		this.delegate = delegate;
	}

	/**
	 * Creates a new DoubleConsumerMulticaster object.
	 */
	public DoubleSeqConsumerMulticaster()
	{
	}

	/**
	 * Returns the consumer to which all call sare delegated.
	 *
	 * @return the consumer to which all call sare delegated
	 */
	public DoubleSeqConsumer getDelegate()
	{
		return delegate;
	}

	/**
	 * Sets the consumer to which all call sare delegated.
	 *
	 * @param delegate the consumer to which all calls are delegated
	 */
	public void setDelegate(DoubleSeqConsumer delegate)
	{
		this.delegate = delegate;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DoubleSeqConsumer#updateValue(long, double)
	 */
	public void updateValue(long timestamp, double[] value)
		throws CommonException
	{
		delegate.updateValue(timestamp, value);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.LongSeqConsumer#updateValue(long, long)
	 */
	public void updateValue(long timestamp, long[] value)
		throws CommonException
	{
		double[] d = new double[value.length];

		for (int i = 0; i < d.length; i++) {
			d[i] = value[i];
		}

		delegate.updateValue(timestamp, d);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getDefaultDataConsumer()
	 */
	public DataConsumer getDefaultDataConsumer()
	{
		return delegate.getDefaultDataConsumer();
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getName()
	 */
	public String getName()
	{
		return delegate.getName();
	}

	/* (non-Javadoc)
	 * @see DataConsumer#setCharacteristics(Map)
	 */
	public void setCharacteristics(Map characteristics)
	{
		delegate.setCharacteristics(characteristics);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#updateDataState(com.cosylab.gui.displayers.DataState)
	 */
	public void updateDataState(DataState state)
	{
		delegate.updateDataState(state);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.StringSeqConsumer#updateValue(long, java.lang.String)
	 */
	public void updateValue(long timestamp, String[] value)
		throws CommonException
	{
		double[] d = new double[value.length];

		for (int i = 0; i < d.length; i++) {
			d[i] = Double.parseDouble(value[i]);
		}

		delegate.updateValue(timestamp, d);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.ObjectSeqConsumer#updateValue(long, java.lang.Object)
	 */
	public void updateValue(long timestamp, Object[] value)
		throws CommonException
	{
		double[] d = new double[value.length];

		for (int i = 0; i < d.length; i++) {
			d[i] = ((Number)value[i]).doubleValue();
		}

		delegate.updateValue(timestamp, d);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getSupportedCharacteristics()
	 */
	public String[] getSupportedCharacteristics()
	{
		return delegate.getSupportedCharacteristics();
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getSupportedConsumerTypes()
	 */
	@SuppressWarnings("unchecked")
	public Class[] getSupportedConsumerTypes()
	{
		return PREFERED_CONSUMER_TYPES;
	}

	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getDataConsumer(java.lang.Class)
	 */
	public <D extends DataConsumer> D getDataConsumer(Class<D> type) {
		return delegate.getDataConsumer(type);
	}
}

/* __oOo__ */
