/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans.
 *
 * CosyBeans is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.displayers;

import java.util.BitSet;

import com.cosylab.util.BitCondition;


/**
 * A specialization of <code>Displayer</code> interface for patterns. A pattern
 * is Java primitive <code>long</code> type that is interpreted in its binary
 * form as a bit-pattern. In order to distinguish significant and
 * insignificant  bits, a bit-mask known as resolution is used. The total
 * number of bits set to 1  in the bitmask determine the number of bits this
 * displayer has to render.  In order to render these bits properly, the
 * adapter delivers the following data items:
 * 
 * <ul>
 * <li>
 * <b>String[] bitDescriptions</b>. For each bit in the resolution bit mask
 * which is set to 1, this array contains a description of maximum one line
 * length.
 * </li>
 * <li>
 * <b>Condition[] conditionWhenSet</b>. For each bit in the resolution     bit
 * mask, which is set to 1, this array contains hints to the     displayer
 * about how the bit should be rendered if it is set to 1.
 * </li>
 * <li>
 * <b>Condition[] conditionWhenCleared</b>. For each bit in the resolution bit
 * mask, which is set to 1, this array contains hints to the displayer about
 * how the bit should be rendered if the bit is set to 0.
 * </li>
 * <li>
 * <b>String format</b>. The C-style formatting string that specifies how the
 * whole value should be rendered. It may, for example, direct the displayer
 * to display the value in binary or hex mode.
 * </li>
 * </ul>
 * 
 *
 * @author <a href="mailto:igor.kriznar@cosylab.com">Igor Kriznar</a>
 * @version $id$
 */
public interface PatternDisplayer extends Displayer, PatternConsumer
{
	/**
	 * Returns the value currently displayed by the displayer. The value is
	 * interpreted as a bit-field.
	 *
	 * @return BitSet bit value
	 */
	public BitSet getValue();

	/**
	 * Sets the value to be displayed by the displayer. This method is invoked
	 * by adapter. The value is interpreted as a bit field as  described in
	 * the class documentation.
	 *
	 * @param value the value to be displayed.
	 */
	public void setValue(BitSet value);

	/**
	 * Returns an array of Strings, equal in length to the total number of
	 * significant bits, determined in the <code>resolution</code>. For each
	 * significant bit, this array contains a (maximum) one line description.
	 *
	 * @return String[] an array of descriptions
	 */
	public String[] getBitDescriptions();

	/**
	 * Sets the array of descriptions to be used by the displayer. This method
	 * will be usually called by the adapter or user during design time.
	 *
	 * @param value the new array of bit descriptions
	 *
	 * @see #getBitDescriptions
	 */
	public void setBitDescriptions(String[] value);

	/**
	 * Returns the array of <code>Condition</code> objects that carry
	 * information on how each significant bit should be rendered when its
	 * value is 0.
	 *
	 * @return BitCondition[] an array of rendering hints for cleared bits
	 */
	public BitCondition[] getConditionWhenCleared();

	/**
	 * Sets the rendering hints for the cleared bits. This method will be
	 * usually called by the adapter or during design time.
	 *
	 * @param BitCondition[] an array of rendering hints for cleared bits.
	 *
	 * @see #getConditionWhenCleared
	 */
	public void setConditionWhenCleared(BitCondition[] value);

	/**
	 * Returns an array of <code>Condition</code> objects that carry
	 * information on how each significant bit should be rendered when its
	 * value is 1.
	 *
	 * @return Condition[] an array of rendering hints for set bits
	 */
	public BitCondition[] getConditionWhenSet();

	/**
	 * Sets the array of rendering hitns for set bits. This method will be
	 * usually called by the adapter or user during design time.
	 *
	 * @param value the rendering hints
	 *
	 * @see #getConditionWhenSet
	 */
	public void setConditionWhenSet(BitCondition[] value);

	/**
	 * Returns the C-style format specification used when the <code>long</code>
	 * bit field must be output to string format.
	 *
	 * @return String C-style printf format specification
	 */
	public String getFormat();

	/**
	 * Sets the format specification used by this displayer to render its value
	 * into one string.
	 *
	 * @param String a new C-style format specification
	 *
	 * @see #getFormat
	 */
	public void setFormat(String value);

	/**
	 * Returns the resolution bit mask determining which bits in the
	 * <code>value</code> are significant. The total number of ones in the
	 * binary representation of the bit-mask is the number of significant
	 * bits. The value to be rendered is obtained by doing a bitwise and on
	 * the resolution and the value.
	 *
	 * @return BitSet the bit-mask for this displayer
	 */
	public BitSet getBitMask();

	/**
	 * Sets the bit-mask for this displayer.
	 *
	 * @param BitSet the new bit-mask for this displayer
	 *
	 * @see #getBitMask
	 */
	public void setBitMask(BitSet value);
}

/* __oOo__ */
