/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans.
 *
 * CosyBeans is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.displayers;

import java.util.EventObject;


/**
 * <p>
 * This update request is sent to <code>NumericAsynchonousConsumer</code> to
 * inform it that the source value has changed. The consumer take appropriate
 * action, which in the case of  Abeans adapter includes notifying the
 * modeling layer about the change in target value.
 * </p>
 */
public class UpdateRequest extends EventObject
{
	private static final long serialVersionUID = 1L;
	protected Object targetValue = null;
	protected Object newTargetValue = null;
	protected long timestamp = 0;
	protected long requestID = 0;
	protected static long requestCount = 0;

	/**
	 * Constructs a new instance of this event.
	 *
	 * @param source the listener which will receive replay for update request
	 */
	public UpdateRequest(UpdateRequestListener source)
	{
		super(source);
	}

	/**
	 * Constructs a new instance of this event. The event will contain unique
	 * serial number and other data that  allow the event source and target to
	 * track the request. This imposes additional performance overhead.
	 *
	 * @param source setter dispatching the event
	 * @param targetValue Object rendering of the target value
	 *
	 * @throws NullPointerException if target value is <code>null</code> 
	 */
	public UpdateRequest(UpdateRequestListener source, Object targetValue)
	{
		super(source);

		if (targetValue == null) {
			throw new NullPointerException("targetValue");
		}

		this.targetValue = targetValue;
		requestID = requestCount;
		requestCount++;
		timestamp = System.currentTimeMillis();
	}

	/**
	 * Returns the timestamp (as in <code>System.currentTimeMillis</code> when
	 * the event was created. Will return 0 if the event was created with
	 * single-paramter constructor.
	 *
	 * @return long timestamp
	 */
	public long getTimestamp()
	{
		return timestamp;
	}

	/**
	 * Returns the request ID associated with this request. IDs are unique
	 * within the JVM session. Will return 0 if the event was created  with
	 * single-parameter constructor.
	 *
	 * @return long UID
	 */
	public long getRequestID()
	{
		return requestID;
	}

	/**
	 * Signals the source <code>UpdateRequestListener</code> that the request
	 * has been processed. <b>This  method may be invoked only by the
	 * asynchronous data consumer.</b> This is a callback and needs to be
	 * called for each asynchronous update request. The event will delegate
	 * the call to the source <code>UpdateRequestListener</code>.
	 *
	 * @param newTargetValue the new value of the dynamic value after the set
	 *        operation. If success indicates <code>true</code>, this value
	 *        should be the same as the requested value for setting.
	 *        Otherwise,  this value may be something else (for instance, if
	 *        setter wants to set above maximum, the actual value set may be
	 *        maximum allowed limit).
	 * @param success <code>true</code> if update process proceeded without
	 *        problems, as defined by the asynchonous data consumer
	 * @param exception the exception thrown during unsuccesful update
	 */
	public void reply(Object newTargetValue, boolean success,
	    Exception exception)
	{
		// allow null values for newTargetValue if the remote call fails
		UpdateRequestListener source = (UpdateRequestListener)getSource();
		this.newTargetValue = newTargetValue;
		source.replay(this, success, exception);
	}

	/**
	 * Returns the value that was actually set as confirmed by the data
	 * consumer. This value may differ from the <code>targetValue</code>
	 * requested by the update requestor because of the possible exceptional
	 * conditions in the  modeling layer. This method will return
	 * <code>null</code> if the event was created with the single-parameter
	 * constructor. Note: it is the responsibility of the adapter to
	 * synchronize the target value of the setter with the confirmed value
	 * after the set by calling <code>setTargetValue</code> on the setter.
	 *
	 * @return Object an object rendering of the confirmed target value or
	 *         <code>null</code> if single parameter constructor was used.
	 */
	public Object getNewTargetValue()
	{
		return newTargetValue;
	}

	/**
	 * Returns the target value. Target value is the value requested by update
	 * requestor to be set in the modeling layer by the asynchronous data
	 * consumer. This method will return <code>null</code> if the event was
	 * instantiated with the single parameter constructor.
	 *
	 * @return Object an object rendering of the target value or
	 *         <code>null</code> if single parameter constructor was used.
	 */
	public Object getTargetValue()
	{
		return targetValue;
	}
}

/* __oOo__ */
