/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.property.editors;

import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;

import javax.swing.JCheckBox;



/**
 * JBooleanCheckBox is a component which serves as an editor for boolean
 * values. It displays a checkBox and a string representation of the value,
 * i.e. either "true" or "false". It can also display an arbitrary user text
 * which labels the checkBox.
 *
 * @author <a href="mailto:miha.kadunc@cosylab.com">Miha Kadunc</a>
 * @version $id$
 */
public class BooleanEditor extends JCheckBox implements PropertyEditor
{
	/**
     * 
     */
    private static final long serialVersionUID = -1376023177867694693L;
    private String description=null;
	/**
	 * Constructor for JBooleanCheckBox. Constructs a check box with default
	 * value set to false.
	 */
	public BooleanEditor()
	{
		this(false);
	}

	/**
	 * Constructor for JBooleanCheckBox. Constructs a check box with default
	 * value set to <code>value</code>.
	 *
	 * @param value the value to set.
	 */
	public BooleanEditor(boolean value)
	{
		super();
		setSelected(value);
		super.setText(Boolean.toString(value));
		init();
	}

	/**
	 * Constructor for JBooleanCheckBox. Constructs a check box with default
	 * value set to <code>value</code>.
	 *
	 * @param description desription of edited value
	 * @param value the value to set.
	 */
	public BooleanEditor(String description, boolean value)
	{
		this(value);
		setDescription(description);
	}

	/**
	 * returns a <code>Boolean</code> object with the value set. To retrieve a
	 * primitive <code>boolean</code>, use isSelected() method
	 *
	 * @return Object a Boolean containing the selected value.
	 */
	public Object getPropertyValue()
	{
		return new Boolean(isSelected());
	}

	/**
	 * If and only if the text is "true" this checkbox is selected.
	 *
	 * @param text the text to be set as boolean
	 *
	 * @see Boolean#getBoolean(java.lang.String)
	 * @see #setValue(boolean)
	 */
	@Override
    public void setText(String text)
	{
		if ("true".equalsIgnoreCase(text)) {
			setSelected(true);
		} else if ("false".equalsIgnoreCase(text)) {
			setSelected(false);
		}
		super.setText(text);
	}

	/**
	 * DOCUMENT ME!
	 *
	 * @param sel the value to be set
	 *
	 * @see setValue
	 */
	@Override
    public void setSelected(boolean sel)
	{
		boolean oldValue = isSelected();

		if (oldValue != sel) {
			super.setSelected(sel);
			//super.setText(Boolean.toString(sel));
			firePropertyChange(PROPERTY_VALUE_NAME, oldValue, sel);
			firePropertyChange("value", oldValue, sel);
		}
	}


	private void init()
	{
		addItemListener(new ItemListener() {
			public void itemStateChanged(ItemEvent e) {
				if (description==null) {
					BooleanEditor.super.setText(Boolean.toString(isSelected()));
				}
				firePropertyChange(PROPERTY_VALUE_NAME,!isSelected(),isSelected());
			}
		});
	}

	/**
	 * Sets the displayed value. Should be java.lang.Boolean
	 *
	 * @param value the value to be set.
	 *
	 * @return DOCUMENT ME!
	 *
	 * @see ValueEditor#setValue(Object)
	 */
	public boolean setPropertyValue(Object value)
	{
		if (value == null) {
			setSelected(false);
		} else if (value instanceof Boolean) {
			setSelected(((Boolean)value).booleanValue());

			return true;
		}

		return false;
	}

	/**
	 * @see com.cosylab.gui.property.editors.PropertyEditor#getDescription()
	 */
	public String getDescription()
	{
		return description;
	}

	/**
	 * @see com.cosylab.gui.property.editors.PropertyEditor#getEditor()
	 */
	public PropertyEditor getEditor()
	{
		return this;
	}

	/**
	 * @see com.cosylab.gui.property.editors.PropertyEditor#setDescription(String)
	 */
	public void setDescription(String description)
	{
		this.description= description; 
		setText(description);
	}
}

/* __oOo__ */
