/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans.
 *
 * CosyBeans is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.util;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.ListSelectionModel;

import com.cosylab.application.AutoScalable;
import com.cosylab.gui.displayers.DisplayerParameters;

/**
 * This class offers a dialog where a single object can be selected from a list 
 * of DisplayerParameters.
 * 
 * @author tkusterle
 *
 */
@AutoScalable
public class DisplayerParametersSelectorDialog extends JDialog {
	
	private static final long serialVersionUID = 1L;
	private static final String TITLE = "Select parameters";
	private static final String PROMPT = "Parameters:";
	
	private String prompt;
	private int selected;
	
	private JLabel textLabel;
	private JPanel labelPanel;
	private JButton okButton;
	private JButton cancelButton;
	private JPanel buttonPanel;
	private JList list;
	private JScrollPane listPane;
	private JPanel contentPanel;

	/**
	 * Constructs a new dialog that belongs to the given parent component.
	 * 
	 * @param comp parent component
	 */
	public DisplayerParametersSelectorDialog(Component comp) {
		this(comp, TITLE, PROMPT);
	}
	
	/**
	 * Constructs a new dialog that belongs to the supplied parent component
	 * and shows the given title and title of the array data among which
	 * the user is allowed to choose.
	 * 
	 * @param comp the parent component
	 * @param title the title of the dialog
	 * @param prompt title of the array data
	 */
	public DisplayerParametersSelectorDialog(Component comp, String title, String prompt) {
		super(JOptionPane.getFrameForComponent(comp), title, true);
		this.prompt = prompt;
		initialize();
	}
	
	private void initialize() {
		this.setLayout(new GridBagLayout());
		this.add(getContentPanel(), new GridBagConstraints(0,0,1,1,1,1,GridBagConstraints.CENTER, GridBagConstraints.BOTH, new Insets(11,11,11,11),1,1));
		this.setSize(300, 250);
		this.addWindowListener(new WindowCloser());
	}
	
	private JPanel getContentPanel() {
		if (contentPanel == null) {
			contentPanel = new JPanel();
			contentPanel.setLayout(new BorderLayout());
			contentPanel.add(getLabelPanel(), BorderLayout.NORTH);
			contentPanel.add(getListPane(), BorderLayout.CENTER);
			contentPanel.add(getButtonPanel(), BorderLayout.SOUTH);
		}
		return contentPanel;
	}
	
	private JPanel getLabelPanel() {
		if (labelPanel == null) {
			labelPanel = new JPanel();
			labelPanel.add(getTextLabel());
		}
		return labelPanel;
	}
	
	private JLabel getTextLabel() {
		if (textLabel == null) {
			textLabel = new JLabel(prompt);
		}
		return textLabel;
	}
	
	private JScrollPane getListPane() {
		if (listPane == null) {
			listPane = new JScrollPane(getList());
		}
		return listPane;
	}
	
	private JList getList() {
		if (list == null) {
			list = new JList();
			list.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		}
		return list;
	}
	
	private JPanel getButtonPanel() {
		if (buttonPanel == null) {
			buttonPanel = new JPanel();
			buttonPanel.add(getOkButton());
			buttonPanel.add(getCancelButton());
		}
		return buttonPanel;
	}
	
	private JButton getOkButton() {
		if (okButton == null) {
			okButton = new JButton("OK");
			okButton.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					selected = getList().getSelectedIndex();
					closeDialog();
				}
			});
		}
		return okButton;
	}
	
	private JButton getCancelButton() {
		if (cancelButton == null) {
			cancelButton = new JButton("Cancel");
			cancelButton.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					selected = -1;
					closeDialog();
				}
			});
		}
		return cancelButton;
	}
	
	private void closeDialog() {
		this.setVisible(false);
	}
	
	/**
	 * Shows a dialog relative to the specified component and shows the supplied
	 * parameters in a list.
	 * 
	 * @param c parent component
	 * @param paramArray list of parameters
	 * @return index of selected parameters
	 */
	public int showSelectionDialog(Component c, DisplayerParameters[] paramArray) {
		return showSelectionDialog(c, toStringArray(paramArray));
	}

	/**
	 * Shows a dialog relative to the specified component and shows the supplied
	 * parameters in a list
	 * 
	 * @param c parent component
	 * @param paramArray list of parameters' names
	 * 
	 * @return index of the selected parameters
	 */
	public int showSelectionDialog(Component c, String[] paramArray) {
		if (paramArray.length == 1) return 0;
		getList().setListData(paramArray);
		setLocationRelativeTo(c);
		this.setVisible(true);
		return selected;
	}
	
	private class WindowCloser extends WindowAdapter {

		/* (non-Javadoc)
		 * @see java.awt.event.WindowAdapter#windowClosing(java.awt.event.WindowEvent)
		 */
		@Override
		public void windowClosing(WindowEvent e) {
			selected = -1;
			super.windowClosed(e);
		}
		
	}
	
	private String[] toStringArray(DisplayerParameters[] dp) {
		String[] array = new String[dp.length];
		for (int i = 0; i < dp.length; i++) {
			array[i] = dp[i].getName();
		}
		return array;
	}
}
