/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans.
 *
 * CosyBeans is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.util;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.lang.reflect.InvocationTargetException;
import java.util.HashMap;
import java.util.Map;

import javax.swing.JComponent;

import com.cosylab.introspection.BeanIntrospector;


/**
 * This class implements pattern, which protects Java bean properties from undesigred changes. 
 * For example user setting are protected from data source characteristics override.
 * 
 * <p>
 * Setting properties to bean with this class should have minimum performance overhead if bean is not protected.
 * </p>
 * 
 * <p>
 * Warning! This class is not synchronized. It expects that only one thread will access GUI Java Beans
 * at a time. Otherwise precautions must be taken.
 * </p>
 * 
 * @author Igor Kriznar (igor.kriznarATcosylab.com)
 *
 */
public final class UserSettingsProtection implements PropertyChangeListener {
	
	static final Map<JComponent,UserSettingsProtection> protections= new HashMap<>(10);
	private JComponent bean;
	private Map<String,Boolean> prot= new HashMap<>(5);
	
	/**
	 * Creates new instance of UserSettingsProtection.
	 */
	private UserSettingsProtection(JComponent bean) {
		this.bean=bean;
	}
	
	/**
	 * Sets up or down property protection for provided JavaBean and properties.
	 * @param bean Java Bean
	 * @param properties the properties to be protected
	 * @param protection if <code>true</code> properties are proteccted, othervise protection is removed
	 */
	public static void setProtection(JComponent bean, String[] properties, boolean protection) {
		UserSettingsProtection p= protections.get(bean);
		if (protection) {
			if (p==null) {
				p= new UserSettingsProtection(bean);
				protections.put(bean,p);
			}
			p.setProtection(properties,protection);
		} else {
			if (p!=null) {
				p.setProtection(properties,protection);
			}
		}
	}

	public void clearProtections() {
		for (String property : prot.keySet()) {
			bean.removePropertyChangeListener(property,this);
		}
		prot.clear();
	}

	public void setProtection(String[] properties, boolean protection) {
		for (int i = 0; i < properties.length; i++) {
			prot.put(properties[i],protection);
			if (protection) {
				bean.addPropertyChangeListener(properties[i],this);
			} else {
				bean.removePropertyChangeListener(properties[i],this);
			}
		}
	}

	/**
	 * Sets to a bean a Java Bean property with new value only if property is not protected.
	 * @param bean a bean on which property is to be set
	 * @param property a property name
	 * @param value new property value to be set
	 * @return <code>true</code> if value was set, otherwise <code>false</code> (property was protected)
	 * @throws NoSuchMethodException
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 */
	public static boolean setUnprotected(JComponent bean, String property, Object value) throws NoSuchMethodException, InvocationTargetException, IllegalAccessException {
		if (property==null || value== null) {
			return false;
		}
		UserSettingsProtection p= protections.get(bean);
		if (p!=null) {
			return p.setUnprotected(property,value);
		}
		BeanIntrospector.setPropertyValue(bean,property,value);
		return true;
	}
	
	/**
	 * Sets to a bean a Java Bean property with new value only if property is not protected.
	 * @param property a property name
	 * @param value new property value to be set
	 * @return <code>true</code> if value was set, otherwise <code>false</code> (property was protected)
	 * @throws NoSuchMethodException
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 */
	public boolean setUnprotected(String property, Object value) throws NoSuchMethodException, InvocationTargetException, IllegalAccessException {
		if (property==null || value== null) {
			return false;
		}
		Boolean b= prot.get(property);
		if (b!=null && b.booleanValue()) {
			return false;
		}
		if (b!=null) {
			bean.removePropertyChangeListener(property,this);
			BeanIntrospector.setPropertyValue(bean,property,value);
			bean.addPropertyChangeListener(property,this);
		} else {
			BeanIntrospector.setPropertyValue(bean,property,value);
		}
		return true;
	}

	/* (non-Javadoc)
	 * @see java.beans.PropertyChangeListener#propertyChange(java.beans.PropertyChangeEvent)
	 */
	public void propertyChange(PropertyChangeEvent evt) {
		String p= evt.getPropertyName();
		if (prot.containsKey(p)) prot.put(p,true);
	}
	
}
