/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of Java-Common.
 *
 * Java-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Java-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Java-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.util;

import java.util.Enumeration;
import java.util.Iterator;


/**
 * A simple implementation of both the <code>java.util.Enumeration</code> and
 * <code>java.util.Iterator</code> that iterates over the array of type
 * <code>Object[]</code>. Removal of elements is not permitted. The iteration
 * is not thread safe.
 *
 * @author <a href="mailto:gasper.tkacik@cosylab.com">Gasper Tkacik</a>
 * @version $id$
 */
public class ArrayEnumeration implements Enumeration, Iterator
{
	/** The array storing the data over which this iterator will iterate. */
	protected Object[] array = null;
	private int ix = 0;

	/**
	 * Creates a new instance of array enumeration, given the array over which
	 * this instance will iterate.
	 *
	 * @param array array of iteration, non-<code>null</code>
	 */
	public ArrayEnumeration(Object[] array)
	{
		assert (array != null);
		this.array = array;
	}

	/**
	 * Returns <code>true</code> iff calling <code>nextElement()</code> will
	 * return a valid value.
	 *
	 * @return <code>true</code> if there are more elements to be iterated over
	 *         in the array
	 */
	public boolean hasMoreElements()
	{
        return ix != array.length;
	}

	/**
	 * Returns the next element in the array. If this method is called after
	 * the end of the array has been reached (when
	 * <code>hasMoreElements()</code> returns <code>false</code>, an array
	 * index out of bounds exception will be thrown.
	 *
	 * @return the next object in the array, can be <code>null</code>
	 */
	public Object nextElement()
	{
		Object retVal = array[ix];
		ix++;

		return retVal;
	}

	/**
	 * A shorthand for <code>hasMoreElements()</code>
	 *
	 * @return <code>true</code> iff this array has more elements to be
	 *         iterated over
	 */
	public boolean hasNext()
	{
		return hasMoreElements();
	}

	/**
	 * A shorthand for <code>nextElement()</code>.
	 *
	 * @return the next element in the array
	 */
	public Object next()
	{
		return nextElement();
	}

	/**
	 * Not implemented.
	 *
	 * @throws UnsupportedOperationException always
	 */
	@Override
    public void remove()
	{
		throw new UnsupportedOperationException();
	}
}

/* __oOo__ */
