/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of Java-Common.
 *
 * Java-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Java-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Java-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.util;

import java.util.Map;

import javax.swing.Action;


/**
 * An interface common to exceptions in projects that depend on  Java-Common.
 * This is an interface because it can be implemented by both checked
 * exceptions (extending <code>java.lang.Exception</code>)  and unchecked
 * exceptions (extending <code>java.lang.RuntimeException</code>), although
 * the later use will probably be not so common. Methods of this interface
 * provide additional information, besides that already contained in Java
 * exceptions (such as message, stack trace elements), that can be used to
 * describe the exception, turn it into a log, or submit it to the support
 * center. By defining a help ID the exceptions can be tied to JavaHelp system
 * as well.
 *
 * @see java.lang.Throwable
 */
public interface CommonThrowable
{
	/**
	 * Call this method if the exception is caught. Provides additional
	 * debugging info.
	 *
	 * @param method method name without parenthesis of prefixed class name
	 */
	void caughtIn(Object instance, String method);

	/**
	 * Returns the help identifier for this exception. The help identifier can
	 * be provided by the instantiator of the exception to indicate the
	 * specific problem for which the exception is being thrown. The help for
	 * this problem can then be looked up into the database by help
	 * applications. This is indended for problems that are common enough to
	 * have dedicated  help pages.
	 *
	 * @return String help ID used by JavaHelp to look up help for this
	 *         exception
	 */
	String getHelpID();

	/**
	 * Returns the currently logged in user. This is the user returned by the
	 * JVM <b>unless</b> the instance throwing the exception has more specific
	 * user information (such as local login).
	 *
	 * @return String user name
	 */
	String getUsername();

	/**
	 * Return the host this exception was generated on, if the JVM has access
	 * to such data.
	 *
	 * @return host name
	 */
	String getHost();

	/**
	 * Return the data about the instance which raised the exception. Normally,
	 * the exception carries only class data, not instance data. In case of
	 * libraries where many models of the same class are instantiated, having
	 * a stack trace does not help with problems that are tied to the instance
	 * and not to the class (for example remote communication with an instance
	 * whose device server crashed). If the object raising an exception does
	 * not implement <code>Identifiable</code> interface, it should return the
	 * <code>Identifiable</code> of its logically most "proximate"
	 * <code>Identifiable</code> instance. Note that while instances of
	 * <code>CoreException</code> must provide at least the proximate
	 * <code>Identifiable</code>, instances of <code>AssertionFailed</code>
	 * may return <code>null</code> in this method.
	 *
	 * @return Object instance data
	 */
	Object getInstance();

	/**
	 * When additional info about the exception is provided via
	 * <code>caughtIn(Object, String)</code> method, this method should  be
	 * called to obtain the name of the method that caught the exception
	 * together with the name of its declaring class. May return
	 * <code>null</code>, if this information is not available.
	 *
	 * @return String Name of the method that caught the exception.
	 */
	String getCaughtIn();

	/**
	 * Returns the message carried by exception implementing this interface.
	 *
	 * @return String message of the exception
	 */
	String getMessage();

	/**
	 * Return the parent exception that caused this exception to be thrown.
	 *
	 * @return parent exception
	 */
	Throwable getParent();

	/**
	 * Returns the thread of execution in which the constructor of this
	 * exception was invoked. It is presupposed that exception is thrown in
	 * the same thread in which it was instantiated.
	 *
	 * @return thread in which this exception was instantiated
	 */
	Thread getThread();

	/**
	 * Returns the timestamp of the instantiation of this exception. Value is
	 * in the format returned by Java <code>System.currentTimeMillis()</code>.
	 *
	 * @return exception creation timestamp
	 */
	long getTimestamp();

	/**
	 * Returns a map that can hold any key-value pairs that contain internal
	 * state of the system or specifically the <code>Identifiable</code>
	 * instance raising this exception. Such data may help in debugging.
	 *
	 * @return additional data
	 */
	Map getValues();

	/**
	 * Sets a value with a given key in the map of values.
	 *
	 * @param key the key under which the value will be stored in the map
	 * @param value the value to store
	 */
	void putValue(String key, Object value);

	/**
	 * Returns the value stored under a given key in the map of values.
	 *
	 * @param key the key to look up
	 *
	 * @return Object the value returned or <code>null</code> if no value is
	 *         stored under this key
	 */
	Object getValue(String key);

	/**
	 * Returns a list of suggested actions that can be added to the GUI.  For
	 * example, an exception could be sent to the system log,  or a report
	 * could be sent to some Web address, or it could be saved to a file etc.
	 * Such procedures should be coded as actions returned by this instance.
	 * Note that the instantiation of the actions can be delayed until this
	 * method is invoked (lazy creation).
	 *
	 * @return Action[] array of actions that can be performed on this
	 *         exception
	 */
	Action[] getActions();
}

/* __oOo__ */
