/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of Java-Common.
 *
 * Java-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Java-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Java-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.util;

/**
 * A simple object pool. This pool is created with an initial capacity and a class
 * (must be a subclass of <code>PoolableObject</code>) of the objects it contains.
 * The pool contains at most <code>capacity</code> of resurrected objects, which it
 * clears (by invoking <code>PoolableObject.clear</code>) before returning them to the
 * user. The user can request a new instance. If found in the pool, it is cleared and
 * returned. Otherwise a new instance creation is attempted with the default constructor.
 * If it fails, <code>null</code> is returned.
 */

public class ObjectPool
{
	private ObjectFactory objectFactory = null;
	private PoolableObject[] freeObjects = null;
	private int ix = -1;
	/** 
	 * Creates a new object pool with an initial capacity and the type of the pooled objects.
	 * ObjectPool will try to use DefaultObjecFactory with the type of the pooled objects.
	 * 
	 * @param objectType a type descending from <code>PoolableObject</code>
	 * @param capacity the initial capacity of the object pool
	 */
	public ObjectPool(Class objectType, int capacity)
	{
		this(new DefaultObjectFactory(objectType),capacity);
	}
	/** 
	 * Creates a new object pool with an initial capacity and ObjectFactory for creating new instances of pooled objects.
	 * 
	 * @param objectFactory ObjectFactory for creating new instances of pooled objects
	 * @param capacity the initial capacity of the object pool
	 */
	public ObjectPool(ObjectFactory objectFactory,int capacity)
	{
		if (objectFactory == null) throw new NullPointerException("objectFactory");
		if (capacity < 1) throw new IllegalArgumentException("Capacity cannot be less then 1.");
		this.objectFactory = objectFactory;
		freeObjects = new PoolableObject[capacity];
	}
	
	/**
	 * Produces a new instance. Returns an object from the pool, if it is available.
	 * Otherwise tries to create new instance with <code>ObjectFactory</code>
	 * 
	 * @return PoolableObject a new cleared object ready for use or <code>null</code>
	 *          if the new type construction failed
	 */
	public synchronized PoolableObject newInstance()
	{
		PoolableObject o=null;
		if (ix < 0)
		{
//			instantiated++;
//			checkCacheSize();
			o=objectFactory.newInstance();
			if (o==null) return null;
		}
		else
		{
//			cached++;
//			checkCacheSize();
			o = freeObjects[ix];
			ix--;
		}
		o.setObjectPool(this);
		return o;
	}
	
	/**
	 * Must be invoked from the <code>finalize</code> method of the 
	 * <code>PoolableObject</code>s to return them to the object pool.
	 * This method will add the object to the pool if there is any room
	 * left, otherwise the reference and the object will be discarded.
	 * 
	 * @param p the object being returned to the pool
	 */
	public synchronized void returnToPool(PoolableObject p)
	{
		if (p == null) throw new NullPointerException("p");
		if (ix >= freeObjects.length - 1) return;
		ix++;
		freeObjects[ix] = p;
		p.clear();
	}
	
	/**
	 * Contains a test sample implementation of a poolable object and the use
	 * of thread pool.
	 * 
	 * @param args ignored.
	 */
	public static void main(String[] args)
	{
		class TestObj implements PoolableObject
		{
			private int value = 1;
			private double v1 = 5;
			private String t = null;
			
			private ObjectPool pool = null;
			
			public TestObj()
			{
			}
			
			@Override
            public void finalize()
			{
				if (pool != null) pool.returnToPool(this);
			}
			public void clear()
			{
				value = 1;
			}
			public void setValue(int value)
			{
				this.value = value;
				t = String.valueOf(value);
				v1 = value;
			}
			
			public PoolableObject createInstance(){
				return new TestObj();
			}
			
			public void setObjectPool(ObjectPool pool)
			{
				this.pool = pool;
			}

		}
		class TestFactory implements ObjectFactory{

			/* (non-Javadoc)
			 * @see com.cosylab.util.ObjectFactory#newInstance()
			 */
			public PoolableObject newInstance() {
				return new TestObj();
			}
		}
		
		ObjectPool pool = new ObjectPool(new TestFactory(), 10000);
		try
		{
			Thread.sleep(1000);
		} catch (InterruptedException ie)
		{
		}
		long time1 = System.currentTimeMillis();
		for (int i = 1; i < 1000000; i++)
		{
			TestObj obj = (TestObj)pool.newInstance();/*null;*/
			if (obj == null) 
			{
				obj = new TestObj();
				obj.setObjectPool(pool);
			}
			obj.setValue(i);
		}
		System.out.println("Total time: " + (System.currentTimeMillis() - time1));
	}
}
