package com.cosylab.util;

import java.io.Serializable;
import java.util.List;

/**
 * 
 * <code>PrimitiveDoubleList</code> is a sort of list, which holds double
 * values as primitive types. doubles can be added to this container and the
 * size of the container will adjust according to the number of elements.
 * This class does not allow editing of the values (such as removing, inserting etc).
 * The only functionality is adding and transforming the values into a double array.
 * <p>
 * The class should be used as a replacement for the {@link List} in those
 * cases where primitive types are read and would use of a list present
 * a performance issues.
 * </p>
 *
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 *
 */
public class PrimitiveDoubleList implements Serializable {

	private static final long serialVersionUID = 5949526354022944289L;
	private double[] data;
	private int size = 0;
	
	/**
	 * Constructs a new PrimitiveLongList. 
	 */
	public PrimitiveDoubleList() {
		this(100);
	}
	
	/**
	 * Constructs a new PrimitiveLongList with the given initial capacity.
	 * Once the size of the list reaches the capacity, the list is automatically expanded.
	 * 
	 * @param initialCapacity the initial capacity of the list
	 */
	public PrimitiveDoubleList(int initialCapacity) {
		data = new double[initialCapacity];
	}
	
	/**
	 * Adds a value to the end of this list.
	 * 
	 * @param val the value
	 */
	public void add(double val) {
		ensureCapacity(size+1);
		data[size] = val;
		size++;
	}
	
	/**
	 * Returns the value under the given index (starting with 0).
	 *  
	 * @param index the index
	 * @return the value at index
	 */
	public double get(int index) {
		return data[index];
	}
	
	/**
	 * Ensures that the list has the provided capacity.
	 * 
	 * @param size the capacity
	 */
	private void ensureCapacity(int size) {
		if (data.length > size) return;
		double[] d = new double[(int)(data.length*1.5)+1];
		System.arraycopy(data,0,d,0,data.length);
		data = d;
	}
	
	/**
	 * Returns the size of this list.
	 * 
	 * @return the size
	 */
	public int size() {
		return size;
	}
	
	/**
	 * Returns the array of all values in this list, which has the length equal
	 * to {@link #size()}.
	 * 
	 * @return the array of values
	 */
	public double[] toArray() {
		double[] d = new double[size];
		System.arraycopy(data,0,d,0,size);
		return d;
	}
	
	/**
	 * Clears the data in this array.
	 */
	public void clear() {
		size = 0;
	}
	
	/**
	 * Checks if the given value is already in this list. Methods returns true if the
	 * value is in the list or false if it is not.
	 * @param valuethe value to check agains the values in this list
	 * @return true if this list contains the value
	 */
	public boolean contains(double value) {
		for (int i = 0; i < size; i++) {
			if (Double.compare(data[i], value) == 0) return true;
		}
		return false;
	}
}
