package com.cosylab.util;

import java.io.Serializable;
import java.util.List;

/**
 * 
 * <code>PrimitiveIntList</code> is a sort of list, which holds integer
 * values as primitive types. ints can be added to this container and the
 * size of the container will adjust according to the number of elements.
 * This class does not allow editing of the values (such as removing, inserting etc).
 * The only functionality is adding and transforming the values into an int array.
 * <p>
 * The class should be used as a replacement for the {@link List} in those
 * cases where primitive types are read and would use of a {@link List} implementation
 * present a performance issues.
 * </p>
 *
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 *
 */
public class PrimitiveIntList implements Serializable {

	private static final long serialVersionUID = 7095456004999464113L;
	private int[] data;
	private int size = 0;
	
	/**
	 * Constructs a new PrimitiveIntList.
	 */
	public PrimitiveIntList() {
		this(100);
	}
	
	/**
	 * Constructs a new PrimitiveIntList with the given initial capacity.
	 * Once the size of the list reaches the capacity, the list is automatically expanded.
	 * 
	 * @param initialCapacity the initial capacity of the list
	 */
	public PrimitiveIntList(int initialCapacity) {
		data = new int[initialCapacity];
	}
	
	/**
	 * Adds a value to the end of this list.
	 * 
	 * @param val the value
	 */
	public void add(int val) {
		ensureCapacity(size+1);
		data[size] = val;
		size++;
	}
	
	/**
	 * Ensures that the list has the provided capacity.
	 * 
	 * @param size the capacity
	 */
	private void ensureCapacity(int size) {
		if (data.length > size) return;
		int[] d = new int[(int)(data.length*1.25)];
		System.arraycopy(data,0,d,0,data.length);
		data = d;
	}
	
	/**
	 * Returns the size of this list.
	 * 
	 * @return the size
	 */
	public int size() {
		return size;
	}
	
	/**
	 * Returns the array of all values in this list, which has the length equal
	 * to {@link #size()}.
	 * 
	 * @return the array of values
	 */
	public int[] toArray() {
		int[] d = new int[size];
		System.arraycopy(data,0,d,0,size);
		return d;
	}
	
	/**
	 * Clears the data in this list.
	 */
	public void clear() {
		size = 0;
	}
	
	/**
	 * Checks if this list contains the given integer number and returns true if it does or false if it does not.
	 * 
	 * @param val the value to find
	 * @return true if the value was found or false if it was not
	 */
	public boolean contains(int val) {
	    for (int i = 0; i < size; i++) {
	        if (data[i] == val) return true;
	    }
	    return false;
	}
}
