package com.cosylab.util;

import java.io.Serializable;
import java.util.List;

/**
 * 
 * <code>PrimitiveLongList</code> is a sort of list, which holds long
 * values as primitive types. longs can be added to this container and the
 * size of the container will adjust according to the number of elements.
 * This class does not allow editing of the values (such as removing, inserting etc).
 * The only functionality is adding and transforming the values into a long array.
 * <p>
 * The class should be used as a replacement for the {@link List} in those
 * cases where primitive types are read and would use of a {@link List} implementation
 * present a performance issues.
 * </p>
 *
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 *
 */
public class PrimitiveLongList implements Serializable {

	private static final long serialVersionUID = 7095456004999464113L;
	private long[] data;
	private int size = 0;
	
	/**
	 * Constructs a new PrimitiveLongList.
	 */
	public PrimitiveLongList() {
		this(100);
	}
	
	/**
	 * Constructs a new PrimitiveLongList with the given initial capacity.
	 * Once the size of the list reaches the capacity, the list is automatically expanded.
	 * 
	 * @param initialCapacity the initial capacity of the list
	 */
	public PrimitiveLongList(int initialCapacity) {
		data = new long[initialCapacity];
	}
	
	/**
	 * Adds a value to the end of this list.
	 * 
	 * @param val the value
	 */
	public void add(long val) {
		ensureCapacity(size+1);
		data[size] = val;
		size++;
	}
	
	/**
	 * Ensures that the list has the provided capacity.
	 * 
	 * @param size the capacity
	 */
	private void ensureCapacity(int size) {
		if (data.length > size) return;
		long[] d = new long[(int)(data.length*1.25)];
		System.arraycopy(data,0,d,0,data.length);
		data = d;
	}
	
	/**
	 * Returns the size of this list.
	 * 
	 * @return the size
	 */
	public int size() {
		return size;
	}
	
	/**
	 * Returns the array of all values in this list, which has the length equal
	 * to {@link #size()}.
	 * 
	 * @return the array of values
	 */
	public long[] toArray() {
		long[] d = new long[size];
		System.arraycopy(data,0,d,0,size);
		return d;
	}
	
	/**
	 * Clears the data in this list.
	 */
	public void clear() {
		size = 0;
	}
}
