/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of Java-Common.
 *
 * Java-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Java-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Java-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.util;

/**
 * This is a gadget for storing values of prmitive <code>double</code> type
 * with <code>java.lang.ArrayList</code>-like functionality and
 * considerably better performance.
 * 
 * Woohoo!
 * 
 * @author <a href="mailto:gasper.pajor@cosylab.com">Gasper Pajor</a>
 * @version $id$
 * 
 */
public class ResizableDoubleList {

    private double[] bufferArray;
    private double[] compactArray;
    
    // Relevant array is always between foi and fei (foi <= i < fei)
    // If foi == fei the relevant array is empty.
    
    private int foi; // first occupied index
    private int fei; // first empty index

	/**
	 * Constructor for ResizableDoubleList. Calls <code>this(10)</code> constructor.
	 */
	public ResizableDoubleList() {
		this(10);
	}
	/**
	 * Constructor for ResizableDoubleList.
	 * @param initialBufferSize the initial size of buffer array into which the values are stored.
	 */
    public ResizableDoubleList(int initialBufferSize) {
        super();
        bufferArray = new double[initialBufferSize];
        foi = 0;
        fei = 0;
    }
    
	/**
	 * Adds a single value to the end of the array.
	 * @param d value to add
	 */
    public void add(double d){
        
        assureAddition(1);        
        bufferArray[fei]=d;
        fei++;
        
    }
    
	/**
	 * Adds array of doubles to the end of the array.
	 * @param da array to add
	 */
    public void add(double[] da){
        
        assureAddition(da.length);
        System.arraycopy(da,0,bufferArray,fei,da.length);
        fei=fei+da.length;
        
    }
    
	/**
	 * Adds a single value at specified index. All values that have index equal to
	 * or larger than specified index increase their indexes by one.
	 * @param d value to add
	 * @param index the index at which the value will be inserted
	 */
    public void add(double d, int index){
        
        assureAddition(1);
        if (checkIndex(index)) {
            
            double[] tempArray = new double[fei-index];
            System.arraycopy(bufferArray, index, tempArray, 0, tempArray.length);
            bufferArray[index] = d;
            System.arraycopy(tempArray,0,bufferArray, index+1, tempArray.length);
            fei++;
                        
        } else {
            throw new IndexOutOfBoundsException("Cannot add to specified index. Index: "+index+", Size: "+size()+".");
        } 
    }

    
	/**
	 * Adds an array of doubles at specified index. All values that have index equal to
	 * or larger than specified index increase their indexes by the lenght of the added array.
	 * @param da array to add
	 * @param index the index at which the array will be inserted
	 */
    public void add(double[] da, int index){
        
        assureAddition(da.length);
        if (checkIndex(index)) {
            double[] tempArray = new double[fei-index];
            // values stored to tempArray for appending after the insertion of new array
            System.arraycopy(bufferArray, index, tempArray, 0, tempArray.length);
            // new array (da) is copied into bufferArray
            System.arraycopy(da,0,bufferArray,index, da.length);
            // appending tempArray
            System.arraycopy(tempArray,0,bufferArray,index+da.length,tempArray.length);
            fei = fei + da.length;
        } else {
            throw new IndexOutOfBoundsException("Cannot add to specified index. Index: "+index+", Size: "+size()+".");            
        }
    }
    
	/**
	 * Removes the specified number of values from (including) specified index onwards.
	 * @param index the index of first value to remove
	 * @param lenght the number of values to remove
	 */
    public void remove(int index, int lenght){
        if (lenght == 0) return;
        if (checkIndex(index) && checkIndex(index+lenght-1)) {
            if (index == 0){
                foi = foi + lenght;
            } else if (index+lenght == fei) {
                fei = index;
            } else {
                double[] tempArray = new double[fei-(foi+index+lenght)];
                // copies what is after the removed part into temp array
                System.arraycopy(bufferArray, foi+index+lenght, tempArray, 0, tempArray.length);
                // copies that back, shifted for lenght
                System.arraycopy(tempArray, 0, bufferArray, foi+index, tempArray.length);
                fei = fei - lenght;
            }
            downSize();
        } else throw new IndexOutOfBoundsException("Cannot remove from index "+index+" to index "+(index+lenght-1) +". Size: "+size()+".");            
    }


	/**
	 * Checks if the buffer array is to large and trimms it to prevent memory leak.
	 */
	private void downSize() {
        if (size()<0.25*bufferArray.length && bufferArray.length > 20) {
            double[] tempArray = new double[(int)(0.5*bufferArray.length+1)];
            System.arraycopy(bufferArray, foi, tempArray,0,size());
            bufferArray = tempArray;
            fei = fei - foi;
            foi = 0;            
        }
    }

    
    
	/**
	 * Gets the size of the array.
	 * @return int the size of the array
	 */
    public int size(){
    
        return fei-foi;
    } 
    
	/**
	 * Gets the whole array
	 * @return double[] the array
	 */
    public double[] getArray(){
        
        if ((compactArray==null) || (compactArray.length!=(fei-foi))) compactArray = new double[fei-foi];
        System.arraycopy(bufferArray,foi, compactArray, 0, compactArray.length);

        
        return compactArray;
    }

	/**
	 * Gets the sub-array of specified lenght from the specified index
	 * @param index the index
	 * @param lenght the lenght
	 * @return double[] the sub-array
	 */
    public double[] getArray(int index, int lenght){
        if (lenght == 0) return new double[0];
        if (checkIndex(index) && checkIndex(index+lenght-1)) {
            if ((compactArray==null) || (compactArray.length!=(lenght))) compactArray = new double[lenght];
            System.arraycopy(bufferArray,foi+index, compactArray, 0, compactArray.length);
            return compactArray;       
        }
        else throw new IndexOutOfBoundsException("Cannot get sub-array from index "+index+" to index "+(index+lenght-1) +". Size: "+size()+".");            
    }

	/**
	 * Gets sub-array from specified index onwards.
	 * @param index the index
	 * @return double[] the sub-array
	 */
    public double[] getArray(int index){
        if (checkIndex(index)) {
            if ((compactArray==null) || (compactArray.length!=(fei-index))) compactArray = new double[fei-index]; 
            System.arraycopy(bufferArray, index, compactArray, 0, compactArray.length);
            return compactArray;
        }
        else throw new IndexOutOfBoundsException("Cannot get sub-array from index "+index+" onwards. Size: "+size()+".");
    }
    
	/**
	 * Gets the value at specified index.
	 * @param index the index
	 * @return double the value at index
	 */
    public double get(int index){
        if (checkIndex(index)){
        	return bufferArray[foi+index];
        }
        else throw new IndexOutOfBoundsException("Cannot get value from index "+index+". Size: "+size());
    }

	/**
	 * Sets the value at specified index.
	 * @param index the index
	 * @param value the new value
	 */
    public void set(int index, double value){
        if (checkIndex(index)){
            bufferArray[foi+index] = value;
        }
        else throw new IndexOutOfBoundsException("Cannot set value at index "+index+". Size: "+size()+".");
    }
    
	/**
	 * Enlarges the array, if it is to small for adding new elements.
	 * @param i the number of new elements
	 */
    private void assureAddition(int i) {
        if (i < 0) return;
        int freeSpacesEnd = bufferArray.length - 1 - fei;
        if (i < freeSpacesEnd) return;
        else {
            double[] tempArray;
            if (2*size()+i > bufferArray.length) tempArray = new double[2*size()+i];
            else tempArray = bufferArray;
            System.arraycopy(bufferArray, foi, tempArray,0,size());
            bufferArray = tempArray;
            fei = fei - foi;
            foi = 0;            
        } 
    }
    
	/**
	 * Validates the index.
	 * @param index the index to validate
	 * @return boolean true if the index is valid, false otherwise
	 */
    private boolean checkIndex(int index) {
        if (foi+index<fei /*|| (fei==foi && foi==index)*/) return true;
        else  return false;
    }

	/**
	 * This method is for testing and debuging purposes only.
	 * It returns the buffered array from within which the array or relevant values is stored.
	 * There should be no need to call this method unless you want to check the working of this class.
	 * @return double[] the buffer array
	 */
    public double[] getBufferArray(){
        return bufferArray;
    }
/**
 * 
 * Run test applet.
 * @param args command line parameters
 */
    public static void main(String[] args){
    
        //ArrayList a = new ArrayList();
        //a.add(3, "Lala");    
        ResizableDoubleList list = new ResizableDoubleList(10);
        list.displayIndexes();
        list.displayRelevant();

        list.add(0);
        list.add(1);
        list.add(2);
        list.add(3);
        list.add(4);
        list.add(5);
        list.add(6);
        list.add(7);
        list.add(8);
        list.add(9);                                
        list.displayRelevant();
        
        list.remove(0,3);
        list.displayRelevant();
        list.remove(0,3);
        list.displayRelevant();
        list.remove(0,3);
        list.displayRelevant();
        list.remove(0,3);
        list.displayRelevant();

/*
        list.add(new double[]{4,6});
        list.displayRelevant();
        
        list.add(5,5);
        list.displayRelevant();
        
        list.add(0);
        list.displayRelevant();
        list.add(1);
        list.displayRelevant();
        list.add(2);
        list.displayRelevant();
        list.add(3);
        list.displayRelevant();

        list.add(new double[]{4,6});
        list.displayRelevant();
        
        list.add(5,5);
        list.displayRelevant();
        
        list.add(new double[]{2.1,2.2,2.3},3);
        list.displayRelevant();
        
        list.remove(0,2);
        list.displayRelevant();

        list.remove(0,2);
        list.displayRelevant();
        
        list.remove(0,2);
        list.displayRelevant();        
        
        
        
        
        
        
/*        for (int i =0; i <list.size(); i++) System.out.println("Index: "+ i +" Value:" + list.get(i));
            
        list.remove(1,2);
        System.out.println("\nRemoving 2 at index 1\n");

        for (int i =0; i <list.size(); i++) System.out.println("Index: "+ i +" Value:" + list.get(i));        
        
        list.remove(3,4);

        System.out.println("\nRemoving 4 at index 3\n");

        for (int i =0; i <list.size(); i++) System.out.println("Index: "+ i +" Value:" + list.get(i));        

        System.out.println("\n"+(int)(10 / 3));
*/
    }
    
    /**
	 * This is merely a debug method.
	 */
    public void displayRelevant(){
    
    System.out.print("\n[");
    for (int i = 0; i < foi; i++){
        System.out.print("... ");
    }
    for (int i = foi; i < fei; i++){
        System.out.print(get(i-foi)+" ");
       
    }
    for (int i = fei; i < bufferArray.length; i++){
        System.out.print("... ");
    }
    System.out.print("] FOI: "+foi+" FEI: "+fei);
    }
	/**
	 * This is merely a debug method.
	 */
    public void displayIndexes(){
    
    System.out.print("\n ");
    for (int i = 0; i < bufferArray.length; i++) {
        if (i < 10) System.out.print("  "+i+" ");
        else if (i<100) System.out.print(" "+i+" ");
        else System.out.print(""+i+" ");
    }

    System.out.print(" ");
    }    
    
    

}
