/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of Java-Common.
 *
 * Java-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Java-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Java-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.util;

import java.util.Iterator;
import java.util.LinkedList;

/**
 * Iterator for tree-like structures.
 * 
 * @author Klemen Zagar, Cosylab
 */
public abstract class TreeIterator<T, U> implements Iterator<T> {
	private final LinkedList<U> queue = new LinkedList<>();

	private final Order order;

	private Iterator<T> currentIterator = null;

	private U currentNode = null;

	public enum Order {
		BreadthFirst,
		DepthFirst
	}

	/**
	 * Construct a tree iterator over a nested iterator structure.
	 * 
	 * @param root
	 *        Iterator through root-level elements.
	 * @param order
	 *        Order of iteration over children.
	 */
	public TreeIterator(U root, Order order) {
		this.queue.add(root);
		this.order = order;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.util.Iterator#hasNext()
	 */
	public boolean hasNext() {
		return getCurrentIterator().hasNext();
	}

	/**
	 * @return
	 */
	private Iterator<T> getCurrentIterator() {
		if (this.currentIterator != null) {
			if (this.currentIterator.hasNext()) {
				return this.currentIterator;
			}
			this.currentIterator = null;
		}
		do {
			do {
				if (this.queue.isEmpty()) {
					return new NoElementsIterator<>();
				}
				this.currentNode = this.queue.removeFirst();
				if (this.currentNode != null) {
					this.currentIterator = getIterator(this.currentNode);
				}
			} while (this.currentIterator == null);

			Iterator<U> iter = getChildren(this.currentNode);
			while (iter.hasNext()) {
				if (this.order == Order.BreadthFirst) {
					this.queue.add(iter.next());
				} else {
					this.queue.push(iter.next());
				}
			}
		} while (!this.currentIterator.hasNext());

		return this.currentIterator;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.util.Iterator#next()
	 */
	public T next() {
		return getCurrentIterator().next();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.util.Iterator#remove()
	 */
	@Override
    public void remove() {
		getCurrentIterator().remove();
	}

	protected abstract Iterator<U> getChildren(U obj);

	protected abstract Iterator<T> getIterator(U obj);
}

/* __oOo__ */
