/*
 * Copyright (c) 2006 Stiftung Deutsches Elektronen-Synchroton, Member of the Helmholtz Association, (DESY), HAMBURG,
 * GERMANY. THIS SOFTWARE IS PROVIDED UNDER THIS LICENSE ON AN "../AS IS" BASIS. WITHOUT WARRANTY OF ANY KIND, EXPRESSED
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR PARTICULAR PURPOSE AND
 * NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. SHOULD THE SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE USER
 * ASSUMES THE COST OF ANY NECESSARY SERVICING, REPAIR OR CORRECTION. THIS DISCLAIMER OF WARRANTY CONSTITUTES AN
 * ESSENTIAL PART OF THIS LICENSE. NO USE OF ANY SOFTWARE IS AUTHORIZED HEREUNDER EXCEPT UNDER THIS DISCLAIMER. DESY HAS
 * NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE FULL LICENSE SPECIFYING
 * FOR THE SOFTWARE THE REDISTRIBUTION, MODIFICATION, USAGE AND OTHER RIGHTS AND OBLIGATIONS IS INCLUDED WITH THE
 * DISTRIBUTION OF THIS PROJECT IN THE FILE LICENSE.HTML. IF THE LICENSE IS NOT INCLUDED YOU MAY FIND A COPY AT
 * HTTP://WWW.DESY.DE/LEGAL/LICENSE.HTM
 */
package de.desy.acop.displayers;

import java.awt.AlphaComposite;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Image;
import java.awt.Insets;
import java.awt.Window;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.HierarchyEvent;
import java.awt.event.HierarchyListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.beans.Beans;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintStream;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.function.Supplier;
import java.util.stream.Collectors;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.BorderFactory;
import javax.swing.ButtonGroup;
import javax.swing.DefaultListCellRenderer;
import javax.swing.DefaultListModel;
import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JRadioButton;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.JTabbedPane;
import javax.swing.JTextArea;
import javax.swing.SwingConstants;
import javax.swing.SwingUtilities;
import javax.swing.Timer;
import javax.swing.TransferHandler;
import javax.swing.border.EtchedBorder;

import com.cosylab.application.AutoScalable;
import com.cosylab.gui.components.util.IconHelper;
import com.cosylab.gui.components.util.RunnerHelper;
import com.cosylab.gui.plugins.VitragePlugIn;

import de.desy.acop.about.ApplicationInfo;
import de.desy.acop.displayers.selector.AcopComboBoxModel;
import de.desy.acop.displayers.tarantula.TarantulaPanel;
import de.desy.acop.displayers.tools.AcopInfoDialog;
import de.desy.acop.launcher.Launcher;
import de.desy.acop.launcher.Utilities;
import de.desy.acop.print.logbook.PrintCallback;
import de.desy.acop.print.logbook.PrintUtilities;
import de.desy.acop.print.logbook.PrinterType;
import de.desy.acop.transport.ConnectionParameters;
import de.desy.tine.addrUtils.FECInfo;
import de.desy.tine.client.TLink;
import de.desy.tine.client.TLinkFactory;
import de.desy.tine.console.TConsole;
import de.desy.tine.queryUtils.ServerQuery;
import de.desy.tine.queryUtils.TQuery;
import de.desy.tine.server.logger.MsgLog;
import de.desy.tine.server.logger.MsgLog.MsgEntry;

/**
 * <code>AcopSpider</code> is an error tracking monitor for Acop components. The component can be installed into any
 * container and will be presented as a small spider icon.
 * <p>
 * Upon constructing this component it will start monitoring all TINE connections. When there is an error in any of the
 * connections the component will change color from green to red and will start blinking. Clicking on the icon will open
 * a dialog with description of the error. If there is an AcopDisplayer in the same window as the spider is and that
 * particular error is connected with one of the displayers, the spider will change the visual appearance of the
 * displayer to notify the user what the error refers to.
 * </p>
 * AcopSpider can also be used for logging the exceptions that occurs during the operation (see
 * {@link #addError(Throwable)}).
 * </p>
 *
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 */
public class AcopSpider extends JLabel {

    /**
     * <code>SpiderDialog</code> is a dialog popup which appears when user clicks on the spider icon. The spider dialog
     * lists all active links in the system and presents the exception panel if the spider is used as an exception
     * reporter.
     *
     * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
     */
    @AutoScalable
    protected class SpiderDialog extends JDialog {

        class ListRenderer extends DefaultListCellRenderer {

            private static final long serialVersionUID = -6412407684875722175L;

            @Override
            public Component getListCellRendererComponent(JList<?> list, Object value, int index, boolean isSelected,
                    boolean cellHasFocus) {
                String name = "";
                if (value instanceof TLink) {
                    try {
                        TLink link = (TLink)value;
                        StringBuilder sb = new StringBuilder();
                        sb.append(formatter.format(new Date(link.getLastTimeStamp())));
                        sb.append("      ");
                        sb.append(link.getLastError());
                        sb.append("      ");
                        sb.append(link.getFullDeviceName());
                        sb.append('/');
                        sb.append(link.getProperty());
                        name = sb.toString();
                    } catch (Exception e) {
                        addError(e);
                    }
                }
                return super.getListCellRendererComponent(list,name,index,isSelected,cellHasFocus);
            }
        }

        SimpleDateFormat formatter = new SimpleDateFormat("EEE dd.MM.yyyy HH:mm:ss.SSS z");
        private static final long serialVersionUID = 2063792587625913979L;
        private JList<TLink> linkList;
        private AcopComboBoxModel<TLink> listModel;
        private JTextArea fullLinkList;
        private Map<TLink,String> linkDescriptions = new HashMap<>();
        private JTabbedPane tabPane;
        private JPanel contentPanel;
        private JPanel buttonPanel;
        private JPanel linkPanel;
        private JPanel fullLinkPanel;
        private JPanel messagePanel;
        private JSplitPane messageSplitPane;
        private JPanel exceptionModePanel;
        private JTextArea exceptionArea;
        private JList<MsgEntry> messageList;
        private JTextArea messageStackArea;
        private DefaultListModel<MsgEntry> messageModel;
        private JButton printButton;
        private JButton customButton;
        private JButton closeButton;
        private JCheckBox debugCheck;
        private JLabel debugLevelLabel;
        private JRadioButton debugLevel1;
        private JRadioButton debugLevel2;
        private ArrayList<Throwable> errors = new ArrayList<>();
        private InfoDialog infoDialog;
        private JPopupMenu popup;
        private JMenuItem infoDialogMI;
        private TarantulaPanel tarantula;
        private long lastMessageTime = 0;
        private PrintUtilities printUtilities;
        private Supplier<String> printContextSupplier;
        private PrintStream throwableStream = new PrintStream(new OutputStream() {

            @Override
            public void write(int b) throws IOException {
                getStackArea().append(new String(new int[]{b},0,1));
            }
        });

        /**
         * Sets this dialog visible.
         */
        @Override
        public void setVisible(boolean b) {
            super.setVisible(b);
            if (b) {
                update(true);
            }
            if (Beans.isDesignTime()) {
                getTabPane().setSelectedIndex(1);
            }
        }

        /**
         * Constructs a new SpiderDialog.
         */
        public SpiderDialog(Window owner) {
            super(owner);
            initialize();
        }

        private void initialize() {
            setSize(700,500);
            setContentPane(getContentPanel());
            setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
            setTitle("Tine Status Viewer");
        }

        private JPanel getContentPanel() {
            if (contentPanel == null) {
                contentPanel = new JPanel(new GridBagLayout());
                contentPanel.add(getTabPane(),new GridBagConstraints(0,0,1,1,1,1,GridBagConstraints.CENTER,
                        GridBagConstraints.BOTH,new Insets(3,3,3,3),0,0));
                contentPanel.add(getButtonPanel(),new GridBagConstraints(0,1,1,1,1,0,GridBagConstraints.CENTER,
                        GridBagConstraints.BOTH,new Insets(3,3,5,3),0,0));
            }
            return contentPanel;
        }

        private JTabbedPane getTabPane() {
            if (tabPane == null) {
                tabPane = new JTabbedPane();
                tabPane.add("Active Links",getLinkPanel());
                tabPane.add("All Links",getFullLinkPanel());
                tabPane.add("Tarantula",getTarantulaPanel());
                tabPane.add("Messages",getMessagePanel());
                tabPane.add("Exceptions",getExceptionModePanel());
                tabPane.setEnabledAt(4,false);
            }
            return tabPane;
        }

        private TarantulaPanel getTarantulaPanel() {
            if (tarantula == null) {
                tarantula = new TarantulaPanel();
            }
            return tarantula;
        }

        private JTextArea getExceptionArea() {
            if (exceptionArea == null) {
                exceptionArea = new JTextArea();
                exceptionArea.setWrapStyleWord(true);
                exceptionArea.setEditable(false);
            }
            return exceptionArea;
        }

        private JList<MsgEntry> getMessagesArea() {
            if (messageList == null) {
                messageList = new JList<>();
                messageModel = new DefaultListModel<>();
                messageList.setModel(messageModel);
                messageList.setCellRenderer(new DefaultListCellRenderer() {

                    private static final long serialVersionUID = 1L;

                    @Override
                    public Component getListCellRendererComponent(JList<?> list, Object value, int index,
                            boolean isSelected, boolean cellHasFocus) {
                        JLabel label = (JLabel)super.getListCellRendererComponent(list,value,index,isSelected,
                                cellHasFocus);
                        if (value != null && value instanceof MsgEntry) {
                            MsgEntry entry = (MsgEntry)value;
                            StringBuilder sb = new StringBuilder();
                            sb.append(formatter.format(new Date(entry.getTime())));
                            sb.append("      [");
                            sb.append(entry.getCode());
                            sb.append("] ");
                            sb.append(entry.getMsg());
                            if (entry.getExecption() != null) {
                                sb.append(" (");
                                sb.append(entry.getExecption().getMessage());
                                sb.append(")");
                                label.setIcon(IconHelper.createIcon("icons/general/bug.png"));
                            } else {
                                label.setIcon(null);
                            }
                            String text = sb.toString();
                            label.setText(text);
                            label.setToolTipText(text);
                            label.setHorizontalTextPosition(SwingConstants.LEFT);
                        }
                        return label;
                    }
                });
                messageList.addListSelectionListener(e -> {
                    MsgEntry msg = messageList.getSelectedValue();
                    if (msg == null) {
                        getStackArea().setText("");
                    } else {
                        Throwable t = msg.getExecption();
                        if (t != null) {
                            t.printStackTrace(throwableStream);
                        } else {
                            getStackArea().setText("");
                        }
                    }
                });
            }
            return messageList;
        }

        private JTextArea getStackArea() {
            if (messageStackArea == null) {
                messageStackArea = new JTextArea();
                messageStackArea.setWrapStyleWord(true);
                messageStackArea.setLineWrap(false);
                messageStackArea.setEditable(false);
            }
            return messageStackArea;
        }

        private JPanel getExceptionModePanel() {
            if (exceptionModePanel == null) {
                exceptionModePanel = new JPanel(new BorderLayout());
                JScrollPane areaScroll = new JScrollPane(getExceptionArea(),JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED,
                        JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
                exceptionModePanel.add(areaScroll,BorderLayout.CENTER);
                exceptionModePanel.add(createClearRefreshButtons(),BorderLayout.SOUTH);
            }
            return exceptionModePanel;
        }

        private JList<TLink> getLinkList() {
            if (linkList == null) {
                linkList = createList();
                listModel = (AcopComboBoxModel<TLink>)linkList.getModel();
            }
            return linkList;
        }

        private JTextArea getFullLinkList() {
            if (fullLinkList == null) {
                fullLinkList = new JTextArea();
                fullLinkList.setEditable(false);
                fullLinkList.setLineWrap(false);
            }
            return fullLinkList;
        }

        private JList<TLink> createList() {
            final JList<TLink> list = new JList<>(new AcopComboBoxModel<>(TLink.class));
            list.setCellRenderer(new ListRenderer());
            list.addMouseMotionListener(new MouseMotionAdapter() {
                @Override
                public void mouseDragged(MouseEvent e) {
                    list.getTransferHandler().exportAsDrag(list,e,TransferHandler.COPY);
                }
            });
            list.addMouseListener(new MouseAdapter() {

                @Override
                public void mouseReleased(MouseEvent e) {
                    if (SwingUtilities.isRightMouseButton(e) || e.isMetaDown()) {
                        getPopup().show(list,e.getX(),e.getY());
                    }
                }
                
                @Override
                public void mouseClicked(MouseEvent e) {
                    if (SwingUtilities.isLeftMouseButton(e) && e.getClickCount() > 1) {
                        TLink link = list.getSelectedValue();
                        if (link != null) {
                            getInfoDialog().show(toConnectionParameters(link));
                        }
                    }
                }
            });
            list.setTransferHandler(new TransferHandler() {

                private static final long serialVersionUID = 1658117801619157639L;

                @Override
                public boolean importData(JComponent comp, Transferable t) {
                    return false;
                }

                @Override
                protected Transferable createTransferable(JComponent c) {
                    return new Transferable() {

                        public Object getTransferData(DataFlavor flavor)
                                throws UnsupportedFlavorException, IOException {
                            if (flavor.isFlavorTextType()) {
                                Object objects = list.getSelectedValue();
                                return toConnectionParameters((TLink)objects).getRemoteName();
                            }
                            throw new UnsupportedFlavorException(flavor);
                        }

                        public boolean isDataFlavorSupported(DataFlavor flavor) {
                            return flavor.isFlavorTextType();
                        }

                        public DataFlavor[] getTransferDataFlavors() {
                            return new DataFlavor[]{DataFlavor.stringFlavor};
                        }
                    };
                }

                @Override
                public boolean canImport(JComponent comp, DataFlavor[] transferFlavors) {
                    return false;
                }

                @Override
                public int getSourceActions(JComponent c) {
                    return TransferHandler.COPY;
                }
            });
            return list;
        }

        private JPanel createClearRefreshButtons() {
            JButton clearButton = new JButton("Clear");
            clearButton.addActionListener(e ->  {
                synchronized (messageModel) {
                    listModel.setElements(new TLink[0]);
                    getFullLinkList().setText("");
                    linkDescriptions.clear();
                    listedLinks.clear();
                    messageModel.clear();
                }
                clearErrors();
                getExceptionArea().setText("");
                setError(false);
            });
            JPanel panel = new JPanel();
            panel.add(clearButton,new GridBagConstraints(1,0,1,1,0,0,GridBagConstraints.CENTER,GridBagConstraints.NONE,
                    new Insets(2,2,2,2),0,0));
            JButton refreshButton = new JButton("Refresh");
            refreshButton.addActionListener(e -> {
                lastMessageTime = 0;
                messageModel.clear();
                update(true);
            });
            panel.add(refreshButton,new GridBagConstraints(2,0,1,1,0,0,GridBagConstraints.CENTER,
                    GridBagConstraints.NONE,new Insets(2,2,2,2),0,0));
            return panel;
        }

        private JPanel getLinkPanel() {
            if (linkPanel == null) {
                linkPanel = new JPanel(new BorderLayout());
                JScrollPane listScroll = new JScrollPane(getLinkList());
                linkPanel.add(listScroll,BorderLayout.CENTER);
                linkPanel.add(createClearRefreshButtons(),BorderLayout.SOUTH);
            }
            return linkPanel;
        }

        private JPanel getFullLinkPanel() {
            if (fullLinkPanel == null) {
                fullLinkPanel = new JPanel(new BorderLayout());
                JScrollPane listScroll = new JScrollPane(getFullLinkList());
                fullLinkPanel.add(listScroll,BorderLayout.CENTER);
                fullLinkPanel.add(createClearRefreshButtons(),BorderLayout.SOUTH);
            }
            return fullLinkPanel;
        }

        private JPanel getMessagePanel() {
            if (messagePanel == null) {
                messagePanel = new JPanel(new BorderLayout());
                messageSplitPane = new JSplitPane(JSplitPane.VERTICAL_SPLIT);
                JScrollPane areaScroll = new JScrollPane(getMessagesArea());
                areaScroll.setMinimumSize(new Dimension(0,0));
                messageSplitPane.setTopComponent(areaScroll);
                JScrollPane stackScroll = new JScrollPane(getStackArea());
                stackScroll.setMinimumSize(new Dimension(0,0));
                messageSplitPane.setBottomComponent(stackScroll);
                messageSplitPane.setResizeWeight(0.5);
                SwingUtilities.invokeLater(() -> messageSplitPane.setDividerLocation(0.5));
                messagePanel.add(messageSplitPane,BorderLayout.CENTER);
                messagePanel.add(createClearRefreshButtons(),BorderLayout.SOUTH);
            }
            return messagePanel;
        }

        private JPanel getButtonPanel() {
            if (buttonPanel == null) {
                buttonPanel = new JPanel(new GridBagLayout());
                closeButton = new JButton("Close");
                closeButton.addActionListener(e -> dispose());
                buttonPanel.add(closeButton,new GridBagConstraints(0,0,1,1,0,0,GridBagConstraints.CENTER,
                        GridBagConstraints.NONE,new Insets(2,2,2,2),0,0));
                debugCheck = new JCheckBox("Debug");
                debugCheck.setSelected(false);
                debugCheck.addActionListener(e -> {
                    if (debugCheck.isSelected())
                        TLinkFactory.setOutputDebugLevel(getDebugLevel());
                    else
                        TLinkFactory.setOutputDebugLevel(0);
                });
                TConsole.getInstance().addWindowListener(new WindowAdapter() {

                    @Override
                    public void windowClosing(WindowEvent e) {
                        debugCheck.setSelected(false);
                    }
                });
                buttonPanel.add(debugCheck,new GridBagConstraints(3,0,1,1,0,0,GridBagConstraints.CENTER,
                        GridBagConstraints.NONE,new Insets(2,2,2,2),0,0));
                debugLevelLabel = new JLabel("Debug level:");
                buttonPanel.add(debugLevelLabel,new GridBagConstraints(4,0,1,1,0,0,GridBagConstraints.CENTER,
                        GridBagConstraints.NONE,new Insets(2,2,2,2),0,0));
                ButtonGroup debugGroup = new ButtonGroup();
                debugLevel1 = new JRadioButton("1");
                debugLevel1.addItemListener(e -> {
                    if (debugLevel1.isSelected() && debugCheck.isSelected()) {
                        TLinkFactory.setOutputDebugLevel(getDebugLevel());
                    }
                });
                debugGroup.add(debugLevel1);
                debugLevel2 = new JRadioButton("2");
                debugLevel2.addItemListener(e -> {
                    if (debugLevel2.isSelected() && debugCheck.isSelected()) {
                        TLinkFactory.setOutputDebugLevel(getDebugLevel());
                    }
                });
                debugGroup.add(debugLevel2);
                debugLevel1.setSelected(true);
                buttonPanel.add(debugLevel1,new GridBagConstraints(5,0,1,1,0,0,GridBagConstraints.CENTER,
                        GridBagConstraints.NONE,new Insets(2,2,2,2),0,0));
                buttonPanel.add(debugLevel2,new GridBagConstraints(6,0,1,1,0,0,GridBagConstraints.CENTER,
                        GridBagConstraints.NONE,new Insets(2,2,2,2),0,0));
                customButton = new JButton(customAction);
                buttonPanel.add(customButton,new GridBagConstraints(7,0,1,1,0,0,GridBagConstraints.CENTER,
                        GridBagConstraints.NONE,new Insets(2,2,2,2),0,0));
                customButton.setVisible(customAction != null);
                printButton = new JButton("Print");
                buttonPanel.add(printButton,new GridBagConstraints(8,0,1,1,0,0,GridBagConstraints.CENTER,
                        GridBagConstraints.NONE,new Insets(2,2,2,2),0,0));
                printButton.addActionListener(e -> {
                    if (printContextSupplier == null) {
                        getPrintUtilities().print(PrinterType.LOGBOOK,null,false);
                    } else {
                        getPrintUtilities().print(PrinterType.LOGBOOK,printContextSupplier.get(),false);
                    }
                });
            }
            return buttonPanel;
        }
        
        private PrintUtilities getPrintUtilities() {
            if (printUtilities == null) {
                printUtilities = new PrintUtilities(getContentPanel(),new PrintCallback() {
                    public void finished() {
                        //ignore
                    }
                    public ApplicationInfo getApplicationInfo(String... data) {
                        return new ApplicationInfo(new HashMap<>());
                    }
                    public void log(String message, int level) {
                        //ignore
                    }
                    public void reportException(Throwable e, String message) {
                        addError(e);
                    }
                },true);
            }
            return printUtilities;
        }
                
        public void setPrintContextSupplier(Supplier<String> printContextSupplier) {
            this.printContextSupplier = printContextSupplier;
        }

        private ConnectionParameters toConnectionParameters(TLink link) {
            String remoteName = link.getFullDeviceName() + "/" + link.getProperty();
            if (remoteName.startsWith("/")) {
                remoteName = "TINE" + remoteName;
            } else if (!remoteName.startsWith("TINE")) {
                remoteName = "TINE/" + remoteName;
            }
            return new ConnectionParameters(remoteName);
        }

        private JPopupMenu getPopup() {
            if (popup == null) {
                popup = new JPopupMenu();
                infoDialogMI = new JMenuItem("Info...");
                infoDialogMI.addActionListener(new ActionListener() {

                    public void actionPerformed(ActionEvent e) {
                        TLink link = null;
                        if (getTabPane().getSelectedIndex() == 0) {
                            link = getLinkList().getSelectedValue();
                        }
                        if (link == null) return;
                        getInfoDialog().show(toConnectionParameters(link));
                        //						getInfoDialog().show(new ConnectionParameters("TINE/test/test/test/test", AccessMode.POLL, 1000));
                    }
                });
                popup.add(infoDialogMI);
            }
            return popup;
        }

        private int getDebugLevel() {
            if (debugLevel1.isSelected())
                return 1;
            else if (debugLevel2.isSelected())
                return 2;
            else
                return 0;
        }

        /**
         * Updates the spider dialog.
         */
        public void update(final boolean allLinks) {
            if (isShowing()) {
                updateMessages();
                try {
                    Runnable r = () -> {
                        if (allLinks) {
                            getFullLinkList().setText(
                                    Arrays.stream(TLinkFactory.getLinksInLinkTable()).collect(Collectors.joining("\n")));
                        }
                        listedLinks.clear();
                        TLink[] links = TLinkFactory.getInstance().getActiveLinks();
                        listModel.setElements(links);
                        getLinkList().repaint();
                        boolean errorExists = false;
                        if (links != null) {
                            for (TLink l : links) {
                                if (l != null) {
                                    listedLinks.add(l);
                                    if (l.getLinkStatus() != 0) {
                                        errorExists = true;
                                    }
                                }
                            }
                        }
                        synchronized (errors) {
                            getExceptionArea().setText("");
                            StringBuilder sb;
                            for (Throwable e : errors) {
                                sb = new StringBuilder("Error: \n");
                                sb.append(e.toString() + "\n");
                                for (StackTraceElement st : e.getStackTrace()) {
                                    sb.append(st.toString() + "\n");
                                }
                                getExceptionArea().append(sb.toString() + " \n \n");
                            }
                            errorExists |= errors.size() > 0;
                            setError(errorExists);
                        }
                    };
                    if (SwingUtilities.isEventDispatchThread()) {
                        SwingUtilities.invokeLater(r);
                    } else {
                        SwingUtilities.invokeAndWait(r);
                    }
                } catch (Exception e) {
                    //ignore
                }
            } else {
                TLink[] links = TLinkFactory.getInstance().getActiveLinks();
                boolean errorExists = false;
                if (links != null) {
                    for (TLink l : links) {
                        if (l != null) {
                            if (l.getLinkStatus() != 0) {
                                errorExists = true;
                                break;
                            }
                        }
                    }
                }
                final boolean error = errorExists || !errors.isEmpty();
                SwingUtilities.invokeLater(() -> setError(error));
            }
        }

        public void updateMessages() {
            SwingUtilities.invokeLater(new Runnable() {

                @Override
                public void run() {
                    MsgEntry[] entries;
                    if (lastMessageTime == 0) {
                        entries = MsgLog.getMessages(512);
                    } else {
                        entries = MsgLog.getMessages(lastMessageTime);
                    }
                    lastMessageTime = System.currentTimeMillis();
                    synchronized (messageModel) {
                        if (entries != null) {
                            for (int i = entries.length - 1; i > -1; i--) {
                                messageModel.add(0,entries[i]);
                            }
                        }
                        if (messageModel.size() > maxMessagesNumber) {
                            messageModel.removeRange(maxMessagesNumber,messageModel.size() - 1);
                        }
                    }
                }
            });
        }

        /**
         * The spider can be used as an exception reporter. Use this method to report an exception.
         *
         * @param e
         */
        public void addError(Throwable e) {
            synchronized (errors) {
                errors.add(e);
                getTabPane().setEnabledAt(4,true);
            }
        }

        /**
         * Clear all exceptions in the exception panel.
         */
        public void clearErrors() {
            synchronized (errors) {
                errors.clear();
                getTabPane().setEnabledAt(4,false);
            }
        }

        /**
         * Returns the InfoDialog which shows the fec information about the selected link.
         *
         * @return the info dialog
         */
        protected InfoDialog getInfoDialog() {
            if (infoDialog == null) {
                infoDialog = new InfoDialog(this);
            }
            return infoDialog;
        }
    }

    /**
     * <code>InfoDialog</code> presents the fec info
     *
     * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
     */
    protected class InfoDialog extends JDialog {

        private static final long serialVersionUID = 1L;
        private SpiderDebugPanel debugPanel;
        private JButton closeButton;
        private JPanel contentPanel;

        public InfoDialog(Window owner) {
            super(owner);
            initialize();
        }

        private void initialize() {
            setSize(400,400);
            setTitle("Info");
            setLocationRelativeTo(AcopSpider.this.getSpiderDialog());
            setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
            setContentPane(getContentPanel());
        }

        private JPanel getContentPanel() {
            if (contentPanel == null) {
                contentPanel = new JPanel(new GridBagLayout());
                contentPanel.add(getDebugPanel(),new GridBagConstraints(0,0,1,1,1,1,GridBagConstraints.CENTER,
                        GridBagConstraints.BOTH,new Insets(5,5,5,5),0,0));
                contentPanel.add(getCloseButton(),new GridBagConstraints(0,1,1,1,0,0,GridBagConstraints.EAST,
                        GridBagConstraints.NONE,new Insets(0,5,5,5),0,0));
            }
            return contentPanel;
        }

        private JButton getCloseButton() {
            if (closeButton == null) {
                closeButton = new JButton("Close");
                closeButton.addActionListener(new ActionListener() {

                    public void actionPerformed(ActionEvent e) {
                        InfoDialog.this.dispose();
                    }
                });
            }
            return closeButton;
        }

        private SpiderDebugPanel getDebugPanel() {
            if (debugPanel == null) {
                debugPanel = new SpiderDebugPanel();
                debugPanel.setShowDebug(false);
            }
            return debugPanel;
        }

        public void show(ConnectionParameters parameters) {
            getDebugPanel().setConnectionParameters(parameters);
            setVisible(true);
            toFront();
        }
    }

    /**
     * <code>AcopDebugPanel</code> shows debug data for certain connection point. This panel can be used in combination
     * with <code>AcopInfoDialog</code>.
     *
     * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
     * @see AcopInfoDialog
     */
    protected class SpiderDebugPanel extends JPanel {

        private static final long serialVersionUID = -1597262900075362868L;
        private JTextArea textArea;
        private ConnectionParameters connectionParameters;
        private JCheckBox debugCheckBox;
        private boolean showDebug = true;
        private HashMap<String,ServerQuery[]> serverQueryCache = new HashMap<>();

        /**
         * Constructs new AcopDebugPanel.
         */
        public SpiderDebugPanel() {
            super();
            initialize();
        }

        private void initialize() {
            this.setLayout(new BorderLayout());
            textArea = new JTextArea();
            textArea.setBorder(BorderFactory.createTitledBorder("Fec Information"));
            this.add(textArea,BorderLayout.CENTER);
            JPanel southPanel = new JPanel(new GridBagLayout());
            debugCheckBox = new JCheckBox("Debug");
            debugCheckBox.addActionListener(new ActionListener() {

                public void actionPerformed(ActionEvent e) {
                    TLinkFactory.getInstance().setDebugLevel(debugCheckBox.isSelected() ? 2 : 0);
                }
            });
            southPanel.add(debugCheckBox,new GridBagConstraints(0,0,1,1,1,1,GridBagConstraints.WEST,
                    GridBagConstraints.NONE,new Insets(0,0,0,0),0,0));
            this.add(southPanel,BorderLayout.NORTH);
        }

        /**
         * Sets the ConnectionParameters, which specify the connection point that debug data is shown for.
         *
         * @param param the remote connection point
         */
        public void setConnectionParameters(ConnectionParameters param) {
            if (this.connectionParameters != null && this.connectionParameters.equals(param)) return;
            this.connectionParameters = param;
            if (connectionParameters != null) {
                textArea.setText("Updating ... Please wait.");
                new Thread(new Runnable() {

                    public void run() {
                        String srv = connectionParameters.getDeviceGroup();
                        String con = connectionParameters.getDeviceContext();
                        StringBuffer s = new StringBuffer();
                        s.append("Device Server(s): ");
                        String[] deviceServers = new String[0];
                        ServerQuery[] query = serverQueryCache.get(con);
                        if (query == null) {
                            query = TQuery.getDeviceServersEx(con,"ALL","ALL");
                            serverQueryCache.put(con,query);
                        }
                        if (query != null) {
                            ServerQuery fecQuery = null;
                            for (int i = 0; i < query.length; i++) {
                                if (query[i].getName().equalsIgnoreCase(srv)) {
                                    fecQuery = query[i];
                                    break;
                                }
                            }
                            if (fecQuery != null) {
                                deviceServers = TQuery.getDeviceServers(con,fecQuery.getXref());
                                for (int i = 0; i < deviceServers.length - 1; i++) {
                                    s.append(deviceServers[i]);
                                    s.append(", ");
                                }
                                if (deviceServers.length > 0) {
                                    s.append(deviceServers[deviceServers.length - 1]);
                                }
                                FECInfo fecInfo = TQuery.getServerInformation(fecQuery.getXref());
                                if (fecInfo != null) {
                                    s.append("\nResponsible: ");
                                    s.append(fecInfo.getResp());
                                    s.append("\nLocation: ");
                                    s.append(fecInfo.getLoc());
                                    s.append("\nOS: ");
                                    s.append(fecInfo.getOs());
                                    s.append("\nDescription: ");
                                    s.append(fecInfo.getDescription());
                                    s.append("\n");
                                } else {
                                    s.append("\nNo FEC info.\n");
                                }
                            } else {
                                s.append(srv);
                                s.append("\nNo FEC info.\n");
                            }
                        }
                        s.append(TQuery.getModuleAddressInfo(srv,con));
                        s.append("\n");
                        s.append(TQuery.getTineVersion(con,srv));
                        s.append("\n");
                        s.append(TQuery.getAppVersion(con,srv));
                        s.append("\n");
                        s.append(TQuery.getAppDate(con,srv));
                        s.append("\n");
                        textArea.setText(s.substring(0));
                    }
                }).start();
            } else {
                textArea.setText("");
            }
        }

        /**
         * Returns the ConnectionParameters that specify the remote connection point, which is being debugged.
         *
         * @return
         */
        public ConnectionParameters getConnectionParameters() {
            return connectionParameters;
        }

        /**
         * Sets the debug level for TINE TLinkFactory. It always uses level 2.
         *
         * @param debug new debug level (0 debug off)
         * @see TLinkFactory#setDebugLevel(int)
         */
        public void setDebug(int debug) {
            if (debug > 0) {
                debugCheckBox.setSelected(true);
            } else {
                debugCheckBox.setSelected(false);
            }
        }

        /**
         * Returns the debug level (0 - debug is off, 2 - debug is on).
         *
         * @return the debug level
         */
        public int isDebug() {
            return debugCheckBox.isSelected() ? 2 : 0;
        }

        /**
         * Sets the visibility of the debug checkbox.
         *
         * @param showDebug true if the checkbox should be visible, false if not
         */
        public void setShowDebug(boolean showDebug) {
            if (this.showDebug == showDebug) return;
            this.showDebug = showDebug;
            debugCheckBox.setVisible(showDebug);
            firePropertyChange("showDebug",!showDebug,showDebug);
        }

        /**
         * Returns true if the debug check box is visible.
         *
         * @return the debug check box visibile
         */
        public boolean isShowDebug() {
            return showDebug;
        }
    }

    private static final long serialVersionUID = -3394462753800819240L;
    private static final Icon SPIDER_ICON = IconHelper.createIcon("acop/spider.gif");
    private SpiderDialog spiderDialog;
    private boolean isError = false;
    private boolean autoResize = true;
    private VitragePlugIn vitrage;
    private boolean paintOverlayDecorations = true;
    private Timer timer;
    private boolean paintColor = true;
    private Thread linkListenerThread;
    private Action customAction;
    private List<TLink> listedLinks = new ArrayList<>();
    private int maxMessagesNumber = 5000;
    
    /**
     * Constructs a new AcopSpider.
     */
    public AcopSpider() {
        super();
        getSpiderDialog();
        initialize();
    }

    private void initialize() {
        setOpaque(true);
        setHorizontalAlignment(SwingConstants.CENTER);
        setVerticalAlignment(SwingConstants.CENTER);
        setIgnoreRepaint(true);
        setBackground(Color.WHITE);
        setBorder(BorderFactory.createEtchedBorder(EtchedBorder.RAISED));
        setIcon(SPIDER_ICON);
        addComponentListener(new ComponentAdapter() {

            @Override
            public void componentResized(ComponentEvent e) {
                if (isAutoResize()) rescaleIconToFit();
            }
        });
        addMouseListener(new MouseAdapter() {

            @Override
            public void mouseClicked(MouseEvent e) {
                if (getRootPane() != null) {
                    getSpiderDialog().setLocationRelativeTo(getRootPane());
                }
                getSpiderDialog().setVisible(true);
            }
        });
        addHierarchyListener(new HierarchyListener() {

            public void hierarchyChanged(HierarchyEvent e) {
                if (getRootPane() != null) {
                    initializeDefaultCustomAction();
                }
                if (vitrage == null && !Beans.isDesignTime()) {
                    if (getRootPane() != null) {
                        vitrage = new VitragePlugIn(getRootPane());
                        vitrage.setVitrageEnabled(isPaintOverlayDecorations());
                    }
                }
                if (getRootPane() != null && getRootPane().getParent() != null
                        && !getRootPane().getParent().isShowing()) {
                    if (linkListenerThread != null) {
                        linkListenerThread.interrupt();
                        linkListenerThread = null;
                    }
                    return;
                }
                if (linkListenerThread == null) {
                    //it is much more performance efficient to use a stand alone
                    //thread and pause it with sleep than using a regular timer.
                    linkListenerThread = new Thread(new Runnable() {

                        public void run() {
                            TLink[] links;
                            boolean errorExists = false;
                            try {
                                while (true) {
                                    links = TLinkFactory.getInstance().getActiveLinks();
                                    if (!isError() && links != null) {
                                        int linkCount = 0;
                                        for (TLink l : links) {
                                            if (l != null) {
                                                if (l.getLinkStatus() != 0) {
                                                    getSpiderDialog().update(false);
                                                    linkCount = -1;
                                                    break;
                                                }
                                                linkCount++;
                                            }
                                        }
                                        if (spiderDialog != null && (linkCount != -1 && spiderDialog.listModel != null
                                                && linkCount != spiderDialog.listModel.getSize())) {
                                            getSpiderDialog().update(false);
                                        }
                                    } else if (isError()) {
                                        errorExists = false;
                                        if (links != null) {
                                            for (TLink l : links) {
                                                if (l != null && l.getLinkStatus() != 0) {
                                                    errorExists = true;
                                                    break;
                                                }
                                            }
                                        }
                                        if (!errorExists) {
                                            getSpiderDialog().update(false);
                                        }
                                    }
                                    if (Thread.interrupted()) break;
                                    Thread.sleep(1000);
                                }
                            } catch (InterruptedException e) {
                                //ignore
                            }
                        }
                    },"Acop Spider");
                    linkListenerThread.setDaemon(true);
                    linkListenerThread.start();
                }
            }
        });
        setMaximumSize(new Dimension(80,80));
        setMinimumSize(new Dimension(16,16));
        setPreferredSize(new Dimension(30,30));
    }

    /**
     * Initializes the default custom action. This action will start a webstart application defined in the
     * acop/acop.properties file. The application is provided all active links as JVM parameters identified by -C flag.
     */
    protected void initializeDefaultCustomAction() {
        setCustomAction(new AbstractAction("History") {

            private static final long serialVersionUID = 1L;

            public void actionPerformed(ActionEvent e) {
                final Cursor prevCursor = getSpiderDialog().getCursor();
                getSpiderDialog().setCursor(new Cursor(Cursor.WAIT_CURSOR));
                final TLink[] links = listedLinks.toArray(new TLink[listedLinks.size()]);
                new Thread(new Runnable() {

                    public void run() {
                        try {
                            final String[] args = new String[links.length];
                            for (int i = 0; i < args.length; i++) {
                                args[i] = "-C\"" + links[i].getFullDeviceName() + "/" + links[i].getProperty() + "\"";
                            }
                            String customAction = Utilities.getAcopProperty(Utilities.SPIDER_CUSTOMACTION);
                            Launcher.runWebstartApplication(customAction,args,false);
                        } catch (Exception e) {
                            addError(e);
                        } finally {
                            getSpiderDialog().setCursor(prevCursor);
                        }
                    }
                }).start();
            }
        });
    }

    /*
     * (non-Javadoc)
     * @see javax.swing.JLabel#setIcon(javax.swing.Icon)
     */
    @Override
    public void setIcon(Icon icon) {
        int height = getHeight() > 0 ? (int)(getHeight() * 0.7) : 16;
        int width = getWidth() > 0 ? (int)(getWidth() * 0.7) : 16;
        if (icon instanceof ImageIcon) {
            if (width <= 0) width = 1;
            icon = new ImageIcon(((ImageIcon)icon).getImage().getScaledInstance(width,-1,Image.SCALE_FAST));
            if (icon.getIconHeight() > height) {
                if (height <= 0) height = 1;
                icon = new ImageIcon(((ImageIcon)icon).getImage().getScaledInstance(-1,height,Image.SCALE_FAST));
            }
        }
        super.setIcon(icon);
    }

    /*
     * (non-Javadoc)
     * @see javax.swing.JComponent#paintComponent(java.awt.Graphics)
     */
    @Override
    protected void paintComponent(Graphics g) {
        super.paintComponent(g);
        ((Graphics2D)g).setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_ATOP,0.5f));
        if (isError)
            if (paintColor) {
                g.setColor(Color.RED.brighter());
            } else {
                g.setColor(Color.GRAY.brighter());
            }
        else
            g.setColor(Color.GREEN.darker());
        g.fillRect(0,0,getWidth(),getHeight());
    }

    /**
     * Puts the spider into the error mode. When in the error mode the spider will be red and will blink. If not in the
     * error mode the spider will be green. In general this method should not be called from the outside except from the
     * spider hierarchy classes if spider handles other types of processes beside monitoring the active links.
     *
     * @param isError a flag indicating if spider should be in error mode
     */
    protected void setError(boolean isError) {
        if (this.isError == isError) return;
        this.isError = isError;
        if (isError) {
            getTimer().start();
        } else {
            getTimer().stop();
            repaint();
        }
    }

    /**
     * Returns true if AcopSpider is in error mode. The component is in error mode when there is a problem with one of
     * the TINE connections or there has been an exception reported. When in error mode spider will blink in red color.
     * If not in error mode, the component will be green.
     *
     * @return true if spider is error mode
     * @see #addError(Throwable)
     */
    public boolean isError() {
        return this.isError;
    }

    /**
     * Returns the spider dialog which enables inspection of the active links in the tine link factory.
     *
     * @return the spider dialog
     */
    protected SpiderDialog getSpiderDialog() {
        if (spiderDialog == null) {
            spiderDialog = new SpiderDialog(SwingUtilities.getWindowAncestor(this));
        }
        return spiderDialog;
    }

    /**
     * Returns true if the spider will paint the overlay decorations across the displayers which are connected with a
     * particular error.
     *
     * @return true if decorations are painted on the displayers
     */
    public boolean isPaintOverlayDecorations() {
        return paintOverlayDecorations;
    }

    /**
     * Enables/disables painting of overlay decorations on the displayers.
     *
     * @param useVitrage
     * @see #isPaintOverlayDecorations()
     */
    public void setPaintOverlayDecorations(boolean useVitrage) {
        boolean old = isPaintOverlayDecorations();
        if (old == useVitrage) return;
        paintOverlayDecorations = useVitrage;
        if (vitrage != null) vitrage.setVitrageEnabled(useVitrage);
        firePropertyChange("paintOverlayDecorations",old,useVitrage);
    }

    /**
     * Returns true if icon resizes automatically when the component is resized.
     *
     * @return true if automatical resizing is on
     */
    public boolean isAutoResize() {
        return autoResize;
    }

    /**
     * Sets whether icon should be resized automatically when component is resized.
     *
     * @param autoResize
     */
    public void setAutoResize(boolean autoResize) {
        this.autoResize = autoResize;
    }

    /**
     * Rescales the image icon to fit the current size of the displayer.
     */
    public void rescaleIconToFit() {
        setIcon(SPIDER_ICON);
    }

    /**
     * Timer which handles the blinking of the spider.
     *
     * @return
     */
    private Timer getTimer() {
        if (timer == null) {
            timer = new Timer(700,new ActionListener() {

                public void actionPerformed(ActionEvent e) {
                    paintColor = !paintColor;
                    AcopSpider.this.repaint();
                }
            });
        }
        return timer;
    }

    /**
     * Adds an exception and prints its stack trace in the exception panel. The spider will automatically switch to
     * error mode when this method is called.
     *
     * @param e reported exception
     * @see #isError()
     */
    public void addError(Throwable e) {
        getSpiderDialog().addError(e);
        setError(true);
    }

    /**
     * Sets the custom action for the custom button. If the action is non null the custom button will become visible in
     * the information dialog and will execute the given action when pressed. If null action is set, the button will be
     * invisible. Default value for the custom action is start of the webstart application defined in the
     * acop/acop.properties file.
     *
     * @param action the action to be performed on click
     */
    public void setCustomAction(Action action) {
        if (customAction == action) return;
        Action oldValue = customAction;
        customAction = action;
        getSpiderDialog().customButton.setAction(customAction);
        getSpiderDialog().customButton.setVisible(customAction != null);
        firePropertyChange("customAction",oldValue,customAction);
    }

    /**
     * Returns the currently set custom action.
     *
     * @return the custom action
     */
    public Action getCustomAction() {
        return customAction;
    }

    /**
     * Sets the maximum number of messages displayed in the messages list. When the number is exceeded the oldest
     * messages are purged from the list.
     *
     * @param maxMessagesNumber maximum number of messages
     */
    public void setMaxMessagesNumber(int maxMessagesNumber) {
        if (this.maxMessagesNumber == maxMessagesNumber) return;
        int oldValue = this.maxMessagesNumber;
        this.maxMessagesNumber = maxMessagesNumber;
        firePropertyChange("maxMessagesNumber",oldValue,maxMessagesNumber);
    }

    /**
     * @return the maximum number of messages displayed in the message list
     */
    public int getMaxMessagesNumber() {
        return maxMessagesNumber;
    }
        
    /**
     * Set the print context supplier which should provide the context to print to its logbook when the print button is 
     * pressed.
     * 
     * @param printContextSupplier the supplier should provide the selected context
     */
    public void setPrintContextSupplier(Supplier<String> printContextSupplier) {
        getSpiderDialog().setPrintContextSupplier(printContextSupplier);
    }

    //	/**
    //	 * Returns an array of all active links listed in the spider dialog.
    //	 * @return the array of active links
    //	 */
    //  commented out because of a palette problem in NetBeans
    //	public TLink[] getLinks() {
    //		return listedLinks.toArray(new TLink[listedLinks.size()]);
    //	}
    public static void main(String[] args) {
        final AcopSpider bean = new AcopSpider();
        RunnerHelper.runComponent(bean,300,300);
        SwingUtilities.invokeLater(new Runnable() {

            @Override
            public void run() {
                Exception ex = new Exception(new Exception());
                bean.addError(ex);
                //				ex.printStackTrace(bean.getSpiderDialog().throwableStream);
            }
        });
        new Thread(new Runnable() {

            public void run() {
                while (true) {
                    bean.setError(!bean.isError());
                    try {
                        Thread.sleep(5000);
                    } catch (InterruptedException e) {
                        e.printStackTrace();
                    }
                }
            }
        }).start();
        System.out.println(bean.getRootPane());
    }
}
