/*
 * Copyright (c) 2006 Stiftung Deutsches Elektronen-Synchroton,
 * Member of the Helmholtz Association, (DESY), HAMBURG, GERMANY.
 *
 * THIS SOFTWARE IS PROVIDED UNDER THIS LICENSE ON AN "../AS IS" BASIS.
 * WITHOUT WARRANTY OF ANY KIND, EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR PARTICULAR PURPOSE AND
 * NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE. SHOULD THE SOFTWARE PROVE DEFECTIVE
 * IN ANY RESPECT, THE USER ASSUMES THE COST OF ANY NECESSARY SERVICING, REPAIR OR
 * CORRECTION. THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF THIS LICENSE.
 * NO USE OF ANY SOFTWARE IS AUTHORIZED HEREUNDER EXCEPT UNDER THIS DISCLAIMER.
 * DESY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS,
 * OR MODIFICATIONS.
 * THE FULL LICENSE SPECIFYING FOR THE SOFTWARE THE REDISTRIBUTION, MODIFICATION,
 * USAGE AND OTHER RIGHTS AND OBLIGATIONS IS INCLUDED WITH THE DISTRIBUTION OF THIS
 * PROJECT IN THE FILE LICENSE.HTML. IF THE LICENSE IS NOT INCLUDED YOU MAY FIND A COPY
 * AT HTTP://WWW.DESY.DE/LEGAL/LICENSE.HTM
 */

package de.desy.acop.displayers.chart;

import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.util.Arrays;

import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JPanel;

import com.cosylab.gui.util.GUIUtilities;

import de.desy.acop.chart.AcopDisplayMode;

/**
 * <code>ChartOptionPanel</code> is a visual component for customizing
 * Acop chart options. The panel is independent of any chart objects
 * and only shows checkboxes and comboboxes for particular properties.
 * Any change made to this properties triggers a <code>PropertyChangeEvent</code>
 * which notifies the registered listeners.
 * <p>
 * The properties which can be changed using this panel include:
 * <ul>
 * 	<li>text printing</li>
 * 	<li>chubby lines</li>
 *  <li>line style - dots, line etc.</li>
 *  <li>extra digits</li>
 *  <li>show grid</li>
 *  <li>best scale</li>
 * </ul>
 *
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 *
 */
public class ChartOptionsPanel extends JPanel {

	private static final long serialVersionUID = 1L;

	private static final String DEFAULT_LINE_MODE = "Default";

	private JCheckBox textCheckBox;
	private JCheckBox cursorMarkerCheckBox;
	private JComboBox<Object> lineDotsCombo;
	private JLabel lineDotsLabel;
	private JCheckBox extraDigitCheckBox;
	private JCheckBox showGridCheckBox;
	private JCheckBox useBestScaleCheckBox;

	private boolean textSelected;
	private AcopDisplayMode lineStyle;
	private boolean cursorMarkerVisible;
	private boolean extraDigits;
	private boolean showGrid;
	private boolean useBestScale;

	private boolean allowDefaultLineMode = false;
	private boolean updateInProgress = false;

	protected Dimension comboDim = new Dimension(110, 21);

	public ChartOptionsPanel() {
		super();
		initialize();
	}

	private void initialize() {
		setLayout(new GridBagLayout());

		textCheckBox = new JCheckBox("Text");
		textCheckBox.addItemListener(new ItemListener(){
			public void itemStateChanged(ItemEvent e) {
				setTextSelected(textCheckBox.isSelected());
			}
		});
		textCheckBox.setToolTipText("Enables Text Markers");

		extraDigitCheckBox = new JCheckBox("Extra Digits");
		extraDigitCheckBox.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent e) {
				setExtraDigits(extraDigitCheckBox.isSelected());
			}
		});
		extraDigitCheckBox.setToolTipText("Shows Extra Digits Where Applicable");

		cursorMarkerCheckBox = new JCheckBox("Cursor Marker");
		cursorMarkerCheckBox.addItemListener(new ItemListener(){
			public void itemStateChanged(ItemEvent e) {
				setCursorMarkerVisible(cursorMarkerCheckBox.isSelected());
			}
		});
		cursorMarkerCheckBox.setToolTipText("Toggles Curosr Marker Visibility");

		showGridCheckBox = new JCheckBox("Show Grid");
		showGridCheckBox.addItemListener(new ItemListener(){
			public void itemStateChanged(ItemEvent e) {
				setShowGrid(showGridCheckBox.isSelected());
			}
		});
		showGridCheckBox.setToolTipText("Toggles grid lines visibility");

		useBestScaleCheckBox = new JCheckBox("Best Scale");
		useBestScaleCheckBox.addItemListener(new ItemListener(){
			public void itemStateChanged(ItemEvent e) {
				setUseBestScale(useBestScaleCheckBox.isSelected());
			}
		});
		useBestScaleCheckBox.setToolTipText("Toggles XY best scale enabled/disabled");

		Object[] values = Arrays.stream(AcopDisplayMode.values()).filter(d -> 
		    d != AcopDisplayMode.Status && d != AcopDisplayMode.Annotation
		).toArray(AcopDisplayMode[]::new);
		lineDotsCombo = new JComboBox<>(values);
		lineDotsCombo.addItemListener(new ItemListener(){
			public void itemStateChanged(ItemEvent e) {
				if (e.getStateChange() == ItemEvent.SELECTED && !updateInProgress) {
					Object o = lineDotsCombo.getSelectedItem();
					if (o.equals(DEFAULT_LINE_MODE)) {
						setLineDisplayMode(null);
					} else {
						setLineDisplayMode((AcopDisplayMode)o);
					}
				}
			}
		});
		GUIUtilities.setFixedSize(lineDotsCombo,comboDim);
		lineDotsCombo.setToolTipText("Sets Chart Style");
		lineDotsLabel = new JLabel("Chart Display Style:");

		this.add(extraDigitCheckBox, new GridBagConstraints(0,0,1,1,0,0,GridBagConstraints.WEST, GridBagConstraints.NONE, new Insets(2,4,0,4),0,0));
		this.add(textCheckBox, new GridBagConstraints(0,1,1,1,0,0,GridBagConstraints.WEST, GridBagConstraints.NONE, new Insets(2,4,0,4),0,0));
		this.add(cursorMarkerCheckBox, new GridBagConstraints(0,2,1,1,0,0,GridBagConstraints.WEST, GridBagConstraints.NONE, new Insets(2,4,0,4),0,0));
		this.add(showGridCheckBox, new GridBagConstraints(0,3,1,1,0,0,GridBagConstraints.WEST, GridBagConstraints.NONE, new Insets(2,4,0,4),0,0));
		this.add(useBestScaleCheckBox, new GridBagConstraints(0,4,1,1,0,0,GridBagConstraints.WEST, GridBagConstraints.NONE, new Insets(2,4,0,4),0,0));
		this.add(lineDotsLabel, new GridBagConstraints(0,5,1,1,0,0,GridBagConstraints.WEST, GridBagConstraints.NONE, new Insets(2,5,0,4),0,0));
		this.add(lineDotsCombo, new GridBagConstraints(0,6,1,1,0,0,GridBagConstraints.WEST, GridBagConstraints.NONE, new Insets(2,4,0,4),0,0));
	}

	/**
	 * Returns the curosr marker visible property value.
	 * @return the cursor marker
	 */
	public boolean isCursorMarkerVisible() {
		return cursorMarkerVisible;
	}

	/**
	 * Sets the chubby lines property.
	 *
	 * @param chubby new chubby lines value
	 */
	public void setCursorMarkerVisible(boolean cursorMarkerVisible) {
		if (this.cursorMarkerVisible == cursorMarkerVisible) return;
		boolean oldcurosroMarker = this.cursorMarkerVisible;
		this.cursorMarkerVisible = cursorMarkerVisible;
		cursorMarkerCheckBox.setSelected(this.cursorMarkerVisible);
		firePropertyChange("cursorMarkerVisible", oldcurosroMarker, this.cursorMarkerVisible);

	}

	/**
	 * Sets the extra digits property.
	 *
	 * @param extra new extra digits value
	 */
	public void setExtraDigits(boolean extra) {
		if (this.extraDigits == extra) return;
		boolean oldExtraDigits = this.extraDigits;
		this.extraDigits = extra;
		extraDigitCheckBox.setSelected(this.extraDigits);
		firePropertyChange("extraDigits", oldExtraDigits, this.extraDigits);

	}

	/**
	 * Returns the extra digits property value.
	 * @return the extra digits
	 */
	public boolean isExtraDigits() {
		return extraDigits;
	}

	/**
	 * Returns the line display mode.
	 *
	 * @return display mode
	 */
	public AcopDisplayMode getLineDisplayMode() {
		return lineStyle;
	}

	/**
	 * Sets the line display mode.
	 *
	 * @param style new display mode value
	 */
	public void setLineDisplayMode(AcopDisplayMode style) {
		if (lineStyle == style) return;
		AcopDisplayMode oldStyle = this.lineStyle;
		this.lineStyle = style;
		if (style == null && allowDefaultLineMode) {
			lineDotsCombo.setSelectedItem(DEFAULT_LINE_MODE);
		} else {
			lineDotsCombo.setSelectedItem(style);
		}
		firePropertyChange("lineDisplayMode", oldStyle, style);
	}

	/**
	 * Sets the text selected property.
	 *
	 * @param visible new text selected value
	 */
	public void setTextSelected(boolean visible) {
		if (this.textSelected == visible) return;
		boolean oldTextSelected = this.textSelected;
		this.textSelected = visible;
		textCheckBox.setSelected(this.textSelected);
		firePropertyChange("textSelected", oldTextSelected, this.textSelected);
	}

	/**
	 * Returns the text selected property value.
	 *
	 * @return the text selected
	 */
	public boolean isTextSelected() {
		return textSelected;
	}

	/**
	 * Returns the use best scale property value.
	 *
	 * @return the use best scale
	 */
	public boolean isUseBestScale() {
		return useBestScale;
	}

	/**
	 * Sets the use best scale property.
	 *
	 * @param selected new use best scale value
	 */
	public void setUseBestScale(boolean selected) {
		if (this.useBestScale == selected) return;
		boolean oldUseBestScale = this.useBestScale;
		this.useBestScale = selected;
		useBestScaleCheckBox.setSelected(this.useBestScale);
		firePropertyChange("useBestScale", oldUseBestScale, this.useBestScale);
	}

	/**
	 * Returns the show grid property value.
	 *
	 * @return show grid value
	 */
	public boolean isShowGrid() {
		return showGrid;
	}

	/**
	 * Sets the show grid property.
	 *
	 * @param selected new show grid property value
	 */
	public void setShowGrid(boolean selected) {
		if (this.showGrid == selected) return;
		boolean oldShowGrid = this.showGrid;
		this.showGrid = selected;
		showGridCheckBox.setSelected(this.showGrid);
		firePropertyChange("showGrid", oldShowGrid, this.showGrid);
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.JComponent#setEnabled(boolean)
	 */
	@Override
	public void setEnabled(boolean enabled) {
		super.setEnabled(enabled);
		textCheckBox.setEnabled(enabled);
		cursorMarkerCheckBox.setEnabled(enabled);
		lineDotsCombo.setEnabled(enabled);
		extraDigitCheckBox.setEnabled(enabled);
		showGridCheckBox.setEnabled(enabled);
		useBestScaleCheckBox.setEnabled(enabled);
	}

	/**
	 * Toggles the visibility of the cursor marker check box.
	 *
	 * @param visible
	 */
	public void setCursorMarkerVisibleView(boolean visible) {
		if (cursorMarkerCheckBox.isVisible() == visible) return;
		cursorMarkerCheckBox.setVisible(visible);
		firePropertyChange("cursorMarkerVisibleView", !visible, visible);
	}

	/**
	 * Returns true if cursor marker check box is visible.
	 *
	 * @return
	 */
	public boolean isCursorMarkerVisibleiew() {
		return cursorMarkerCheckBox.isVisible();
	}

	/**
	 * Toggles the visibility of the extra digits checkbox.
	 * @param visible
	 */
	public void setExtraDigitsView(boolean visible) {
		if (extraDigitCheckBox.isVisible() == visible) return;
		extraDigitCheckBox.setVisible(visible);
		firePropertyChange("extraDigitsView", !visible, visible);
	}

	/**
	 * Returns true if extra digits check box is visible.
	 *
	 * @return
	 */
	public boolean isExtraDigitsView() {
		return extraDigitCheckBox.isVisible();
	}

	/**
	 * Toggles the visibility of the show grid checkbox.
	 * @param visible
	 */
	public void setShowGridView(boolean visible) {
		if (showGridCheckBox.isVisible() == visible) return;
		showGridCheckBox.setVisible(visible);
		firePropertyChange("showGridView", !visible, visible);
	}

	/**
	 * Returns true if show grid check box is visible.
	 *
	 * @return
	 */
	public boolean isShowGridView() {
		return showGridCheckBox.isVisible();
	}

	/**
	 * Toggles the visibility of the text checkbox.
	 * @param visible
	 */
	public void setTextView(boolean visible) {
		if (textCheckBox.isVisible() == visible) return;
		textCheckBox.setVisible(visible);
		firePropertyChange("textView", !visible, visible);
	}

	/**
	 * Returns true if text check box is visible.
	 *
	 * @return
	 */
	public boolean isTextView() {
		return textCheckBox.isVisible();
	}

	/**
	 * Toggles the visibility of the best scale checkbox.
	 * @param visible
	 */
	public void setBestScaleView(boolean visible) {
		if (useBestScaleCheckBox.isVisible() == visible) return;
		useBestScaleCheckBox.setVisible(visible);
		firePropertyChange("bestScaleView", !visible, visible);
	}

	/**
	 * Toggles the visibility of the line display mode view.
	 *
	 * @param visible
	 */
	public void setLineDisplayModeView(boolean visible) {
		if (lineDotsCombo.isVisible() == visible) return;
		lineDotsCombo.setVisible(visible);
		lineDotsLabel.setVisible(visible);
		firePropertyChange("lineDisplayModeView",!visible, visible);
	}

	/**
	 * Returns the visibility of the line display mode combo box.
	 *
	 * @return
	 */
	public boolean isLineDisplayModeView() {
		return lineDotsCombo.isVisible();
	}

	/**
	 * Returns true if best scale check box is visible.
	 *
	 * @return
	 */
	public boolean isBestScaleView() {
		return useBestScaleCheckBox.isVisible();
	}

	/**
	 * Allows or prevents default line mode option.
	 *
	 * @param allowDefaultLineMode true if default line mode is allowed or false otherwise
	 */
	public void setAllowDefaultLineMode(boolean allowDefaultLineMode) {
		if (this.allowDefaultLineMode == allowDefaultLineMode) return;
		this.allowDefaultLineMode = allowDefaultLineMode;
		updateInProgress = true;
		lineDotsCombo.removeAllItems();
		AcopDisplayMode[] values = AcopDisplayMode.values();
		if (allowDefaultLineMode) {
			lineDotsCombo.addItem(DEFAULT_LINE_MODE);
		}
		for (AcopDisplayMode v : values) {
			lineDotsCombo.addItem(v);
		}
		updateInProgress = false;
		lineDotsCombo.setSelectedIndex(0);
		setLineDisplayMode(null);
		firePropertyChange("allowDefaultLineMode",!allowDefaultLineMode,allowDefaultLineMode);
	}

	/**
	 * Returns true if default line mode is allowed. If default line mode is
	 * selected, the {@link #getLineDisplayMode()} return null.
	 *
	 * @return true if allowed or false otherwise
	 */
	public boolean isAllowDefaultLineMode() {
		return allowDefaultLineMode;
	}

}
