/*
 * Copyright (c) 2006 Stiftung Deutsches Elektronen-Synchroton,
 * Member of the Helmholtz Association, (DESY), HAMBURG, GERMANY.
 *
 * THIS SOFTWARE IS PROVIDED UNDER THIS LICENSE ON AN "../AS IS" BASIS.
 * WITHOUT WARRANTY OF ANY KIND, EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR PARTICULAR PURPOSE AND
 * NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE. SHOULD THE SOFTWARE PROVE DEFECTIVE
 * IN ANY RESPECT, THE USER ASSUMES THE COST OF ANY NECESSARY SERVICING, REPAIR OR
 * CORRECTION. THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF THIS LICENSE.
 * NO USE OF ANY SOFTWARE IS AUTHORIZED HEREUNDER EXCEPT UNDER THIS DISCLAIMER.
 * DESY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS,
 * OR MODIFICATIONS.
 * THE FULL LICENSE SPECIFYING FOR THE SOFTWARE THE REDISTRIBUTION, MODIFICATION,
 * USAGE AND OTHER RIGHTS AND OBLIGATIONS IS INCLUDED WITH THE DISTRIBUTION OF THIS
 * PROJECT IN THE FILE LICENSE.HTML. IF THE LICENSE IS NOT INCLUDED YOU MAY FIND A COPY
 * AT HTTP://WWW.DESY.DE/LEGAL/LICENSE.HTM
 */

package de.desy.acop.displayers.chart;

import java.io.Serializable;

/**
 * <code>HistoryParameters</code> defines a connection point for the history data.
 * This class defines the remote connection which can be used to connect
 * to history or archive server and obtain data provided by those servers.
 * 
 * @author Tilen Kusterle, Cosylab
 *
 */
public class HistoryParameters implements Serializable, Cloneable {
	
	private static final long serialVersionUID = 1L;
	
	private String historyContext;
	private String historyServer;
	private String historyDevice;
	private String historyProperty;
	private int historyIndex;
	private int subIndex = 0;
	
	/**
	 * Constructs new HistoryParameters.
	 * 
	 * @param historyContext
	 * @param historyServer
	 * @param historyDevice
	 * @param historyProperty
	 * @param historyIndex
	 */
	public HistoryParameters(String historyContext, String historyServer, String historyDevice, String historyProperty, int historyIndex) {
		this.historyContext = historyContext;
		this.historyServer = historyServer;
		this.historyDevice = historyDevice;
		this.historyProperty = historyProperty;
		this.historyIndex = historyIndex;
	}
	
	/**
	 * Constructs new HistoryParameters.
	 * 
	 * @param historyContext
	 * @param historyServer
	 * @param historyDevice
	 * @param historyProperty
	 * @param historyIndex
	 */
	public HistoryParameters(String historyContext, String historyServer, String historyDevice, String historyProperty, int historyIndex, int subIndex) {
		this.historyContext = historyContext;
		this.historyServer = historyServer;
		this.historyDevice = historyDevice;
		this.historyProperty = historyProperty;
		this.historyIndex = historyIndex;
		this.subIndex = subIndex;
	}
	
	/**
	 * Constructs a new HistoryParameters from the remote name. The remote name
	 * should have th structure context/server/device/property.
	 * 
	 * @param remoteName
	 */
	public HistoryParameters(String remoteName) {
		String[] s = remoteName.split("/");
	    
		this.historyContext = (s.length>0 && s[0]!=null) ? s[0] : "";
		this.historyServer = (s.length>1 && s[1]!=null) ? s[1] : "";
		this.historyProperty = (s.length>3 && s[s.length-1]!=null) ? s[s.length-1] : "";
		
		StringBuilder sb= new StringBuilder(128);
		
		if (s.length>2 && s[2]!=null) {
			sb.append(s[2]);
		}
		for (int i = 3; i < s.length-1; i++) {
			sb.append('/');
			if (s[i]!=null) sb.append(s[i]);
		}
		
		this.historyDevice = sb.toString();
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj) {
		if (!(obj instanceof HistoryParameters)) return false;
		return this.toString().equals(obj.toString());
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return historyContext + "/" + historyServer + "/" + historyDevice + "/" + historyProperty;
	}
	
	/**
	 * Returnst the remote name.
	 * 
	 * @return the remote name
	 * @see HistoryParameters#HistoryParameters(String)
	 */
	public String getRemoteName() {
		return historyContext + "/" + historyServer + "/" + historyDevice + "/" + historyProperty;
	}

	/**
	 * Returns the context.
	 * 
	 * @return the historyContext
	 */
	public String getHistoryContext() {
		return historyContext;
	}
	/**
	 * Sets the new context.
	 * 
	 * @param historyContext the historyContext to set
	 */
	public void setHistoryContext(String historyContext) {
		this.historyContext = historyContext;
	}
	
	/**
	 * Returns the device name.
	 * 
	 * @return the historyDevice
	 */
	public String getHistoryDevice() {
		return historyDevice;
	}
	
	/**
	 * Sets new device name.
	 * 
	 * @param historyDevice the historyDevice to set
	 */
	public void setHistoryDevice(String historyDevice) {
		this.historyDevice = historyDevice;
	}
	
	/**
	 * Returns the property name
	 * 
	 * @return the historyProperty
	 */
	public String getHistoryProperty() {
		return historyProperty;
	}
	
	/**
	 * Sets new property name 
	 * @param historyProperty the historyProperty to set
	 */
	public void setHistoryProperty(String historyProperty) {
		this.historyProperty = historyProperty;
	}
	
	/** Returns the device server.
	 * 
	 * @return the historyServer
	 */
	public String getHistoryServer() {
		return historyServer;
	}
	
	/**
	 *  Sets new device server.
	 *  
	 * @param historyServer the historyServer to set
	 */
	public void setHistoryServer(String historyServer) {
		this.historyServer = historyServer;
	}

	/**
	 * Returns the index of the device.
	 * 
	 * @return the historyIndex
	 */
	public int getHistoryIndex() {
		return historyIndex;
	}

	/**
	 * Sets the index of the device.
	 * 
	 * @param historyIndex the historyIndex to set
	 */
	public void setHistoryIndex(int historyIndex) {
		this.historyIndex = historyIndex;
	}
	
	/**
	 * Returns the subindex of the data in the array if this is a spectrum.
	 * 
	 * @return the sub index
	 */
	public int getSubIndex() {
		return subIndex;
	}
	
	/**
	 * Sets the sub index of the property (the index in the array if for example a spectrum).
	 * 
	 * @param subIndex the index
	 */
	public void setSubIndex(int subIndex) {
		this.subIndex = subIndex;
	}
	
	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		return getRemoteName().hashCode();
	}
	
	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#clone()
	 */
	@Override
	public HistoryParameters clone() {
		return new HistoryParameters(historyContext,historyServer,historyDevice,historyProperty,historyIndex,subIndex);
	}
	
}
