/*
 * Copyright (c) 2006 Stiftung Deutsches Elektronen-Synchroton, Member of the Helmholtz Association, (DESY), HAMBURG,
 * GERMANY. THIS SOFTWARE IS PROVIDED UNDER THIS LICENSE ON AN "../AS IS" BASIS. WITHOUT WARRANTY OF ANY KIND, EXPRESSED
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR PARTICULAR PURPOSE AND
 * NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. SHOULD THE SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE USER
 * ASSUMES THE COST OF ANY NECESSARY SERVICING, REPAIR OR CORRECTION. THIS DISCLAIMER OF WARRANTY CONSTITUTES AN
 * ESSENTIAL PART OF THIS LICENSE. NO USE OF ANY SOFTWARE IS AUTHORIZED HEREUNDER EXCEPT UNDER THIS DISCLAIMER. DESY HAS
 * NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE FULL LICENSE SPECIFYING
 * FOR THE SOFTWARE THE REDISTRIBUTION, MODIFICATION, USAGE AND OTHER RIGHTS AND OBLIGATIONS IS INCLUDED WITH THE
 * DISTRIBUTION OF THIS PROJECT IN THE FILE LICENSE.HTML. IF THE LICENSE IS NOT INCLUDED YOU MAY FIND A COPY AT
 * HTTP://WWW.DESY.DE/LEGAL/LICENSE.HTM
 */
package de.desy.acop.displayers.selector;

import java.lang.reflect.Array;

import javax.swing.AbstractListModel;
import javax.swing.ComboBoxModel;

/**
 * <code>AcopComboBoxModel</code> is an implementation of the <code>ListModel</code> which allows adding multiple items
 * at once. This model can only work with String objects.
 * 
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 */
public class AcopComboBoxModel<T> extends AbstractListModel<T>implements ComboBoxModel<T> {

    private static final long serialVersionUID = -380035760547442817L;
    private T[] elements;
    private T selectedItem;
    private Class<T> type;

    /**
     * Constructs an empty AcopComboBoxModel object.
     */
    @SuppressWarnings("unchecked")
    public AcopComboBoxModel(Class<T> type) {
        this.type = type;
        elements = (T[])Array.newInstance(type,0);
    }

    /**
     * Constructs a AcopComboBoxModel object with the given elements as the initial available options.
     */
    public AcopComboBoxModel(Class<T> type, T[] elements) {
        this.type = type;
        setElements(elements);
    }

    /**
     * Set the value of the selected item. The selected item may be null.
     * <p>
     * 
     * @param anObject The combo box value or null for no selection.
     */
    public void setSelectedItem(Object anObject) {
        if ("".equals(anObject)) {
            anObject = null;
        }
        if ((selectedItem != null && !selectedItem.equals(anObject)) || (selectedItem == null && anObject != null)) {
            selectedItem = type.cast(anObject);
            fireContentsChanged(this,-1,-1);
        }
    }

    /*
     * (non-Javadoc)
     * @see javax.swing.ComboBoxModel#getSelectedItem()
     */
    public T getSelectedItem() {
        return selectedItem;
    }

    /*
     * (non-Javadoc)
     * @see javax.swing.ListModel#getSize()
     */
    public int getSize() {
        return elements == null ? 0 : elements.length;
    }

    /*
     * (non-Javadoc)
     * @see javax.swing.ListModel#getElementAt(int)
     */
    public T getElementAt(int index) {
        if (elements != null && index >= 0 && index < elements.length)
            return elements[index];
        else
            return null;
    }

    /**
     * Sets new elements to this model.
     * 
     * @param obj new elements
     */
    public void setElements(T[] obj) {
        if (obj == elements || (obj == null && elements == null)) {
            return;
        }
        removeAllElements();
        if (obj != null) {
            elements = obj;
            fireIntervalAdded(this,0,obj.length - 1 > 0 ? obj.length - 1 : 0);
        }
    }

    /**
     * Returns all elements in the model.
     * 
     * @return the elements
     */
    public T[] getElements() {
        return elements;
    }

    /**
     * Empties the list.
     */
    @SuppressWarnings("unchecked")
    protected void removeAllElements() {
        if (elements != null && elements.length > 0) {
            int firstIndex = 0;
            int lastIndex = elements.length - 1;
            elements = (T[])Array.newInstance(type,0);
            fireIntervalRemoved(this,firstIndex,lastIndex);
        }
    }

    /**
     * Replaces an element at the specified index.
     * 
     * @param index the index the element should be replaced at
     * @param newItem the replacement element
     */
    public void replaceElement(int index, T newItem) {
        if (elements == null || index > elements.length || index < 0) return;
        elements[index] = newItem;
        fireIntervalAdded(this,index,index);
    }

    /**
     * Checks if this model containes the given object.
     * 
     * @param o the object that existance should be checked for
     * @return true if object is in the model.
     */
    public boolean contains(Object o) {
        if (o == null) return false;
        if (elements == null) return false;
        for (T obj : elements) {
            if (o.equals(obj)) return true;
        }
        return false;
    }
}
/* __oOo__ */
