/*
 * Copyright (c) 2006 Stiftung Deutsches Elektronen-Synchroton,
 * Member of the Helmholtz Association, (DESY), HAMBURG, GERMANY.
 *
 * THIS SOFTWARE IS PROVIDED UNDER THIS LICENSE ON AN "../AS IS" BASIS.
 * WITHOUT WARRANTY OF ANY KIND, EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR PARTICULAR PURPOSE AND
 * NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE. SHOULD THE SOFTWARE PROVE DEFECTIVE
 * IN ANY RESPECT, THE USER ASSUMES THE COST OF ANY NECESSARY SERVICING, REPAIR OR
 * CORRECTION. THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF THIS LICENSE.
 * NO USE OF ANY SOFTWARE IS AUTHORIZED HEREUNDER EXCEPT UNDER THIS DISCLAIMER.
 * DESY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS,
 * OR MODIFICATIONS.
 * THE FULL LICENSE SPECIFYING FOR THE SOFTWARE THE REDISTRIBUTION, MODIFICATION,
 * USAGE AND OTHER RIGHTS AND OBLIGATIONS IS INCLUDED WITH THE DISTRIBUTION OF THIS
 * PROJECT IN THE FILE LICENSE.HTML. IF THE LICENSE IS NOT INCLUDED YOU MAY FIND A COPY
 * AT HTTP://WWW.DESY.DE/LEGAL/LICENSE.HTM
 */

package de.desy.acop.displayers.selector;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.StringSelection;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.beans.Beans;
import java.beans.Customizer;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.StringTokenizer;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.RejectedExecutionHandler;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import javax.swing.DefaultListCellRenderer;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JTextField;
import javax.swing.SwingConstants;
import javax.swing.SwingUtilities;
import javax.swing.TransferHandler;

import com.cosylab.gui.components.NumberField;
import com.cosylab.gui.components.util.ComponentPositioner;
import com.cosylab.gui.components.util.IconHelper;
import com.cosylab.gui.components.util.RunnerHelper;
import com.cosylab.gui.util.DisplayerParametersSelectorDialog;
import com.cosylab.gui.util.GUIUtilities;
import com.cosylab.util.ListenerList;

import de.desy.acop.displayers.AcopSpider;
import de.desy.acop.displayers.tools.CaseInsensitiveComparator;
import de.desy.acop.displayers.tools.ConnectionParametersReceiver;
import de.desy.acop.displayers.tools.MultipleAcopDisplayer;
import de.desy.acop.transport.AccessMode;
import de.desy.acop.transport.ConnectionParameters;

/**
 * <code>SelectorGUI</code> is a visual component which enables the construction
 * of <code>ConnectionParameters</code>. The component consists of sevral combo boxes
 * and input fields which allow user to configure the connection point. When
 * constructing a valid ConnectionParameters a SelectorGUIEvent will be
 * fired notifying the listeners that selection was completed.
 *  
 * 
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 * @see ConnectionParameters
 */
public class SelectorGUI extends JPanel implements Customizer {
		
	private class ComboRenderer extends DefaultListCellRenderer {
		
		private static final long serialVersionUID = 1L;

		@Override
		public Component getListCellRendererComponent(JList<?> list, Object value, int index, boolean isSelected, boolean cellHasFocus) {
		
			Component c = super.getListCellRendererComponent(list, value, index, isSelected, cellHasFocus);
			if (value != null && c instanceof JLabel) {
				((JLabel)c).setToolTipText(String.valueOf(value));
			}
			return c;
		}
	}
	
	private enum ExtendedData{
		BYTE ("byte"),
		SHORT ("short"),
		INT16 ("int16"),
		INT ("int"),
		INT32 ("int32"),
		LONG ("long"),
		INT64 ("int64"),
		DOUBLE ("double"),
		FLOAT ("float"),
		CHAR ("char"),
		STRING ("String");
		
		private final String displayName;
		
		private ExtendedData(String name) {
			displayName = name;
		}
		
		@Override
		public String toString() {
			return displayName;
		}
	}
	
	
	private static final long serialVersionUID = 1647591422586736408L;
	
	private static final String AUTO = "Auto";
	private static final String ACTUAL = "Actual";
	private Object disp;
	
	private JPanel commandPanel;
	private JComboBox<String> accessProtocolCombo;
	private JComboBox<String> deviceContextCombo;
	private JComboBox<String> subSystemCombo;
	private JComboBox<String> deviceServerCombo;
	private JComboBox<String> deviceNameCombo;
	private JComboBox<String> propertyNameCombo;
	private AcopComboBoxModel<String> accessProtocolModel;
	private AcopComboBoxModel<String> deviceContextModel;
	private AcopComboBoxModel<String> subSystemModel;
	private AcopComboBoxModel<String> deviceServerModel;
	private AcopComboBoxModel<String> deviceNameModel;
	private AcopComboBoxModel<String> propertyNameModel;
	private AcopComboBoxModel<String> arraySizeModel;
	private AcopComboBoxModel<ExtendedData> extendedDataModel;
	private NumberField accessRateField;
	private JLabel accessProtocolLabel;
	private JLabel deviceContextLabel;
	private JLabel subSystemLabel;
	private JLabel deviceServerLabel;
	private JLabel deviceNameLabel;
	private JLabel propertyNameLabel;
	private JLabel accessRateLabel;
	private JLabel selectionLabel;
	private JLabel descriptionLabel;
	private JCheckBox showStockProperties;
	private JComboBox<AccessMode> accessModeCombo;
	private JLabel accessModeLabel;
	
	private JLabel arraySizeLabel;
	private JComboBox<String> arraySizeCombo;

	private JLabel extendedDataLabel;
	private JComboBox<ExtendedData> extendedDataCombo;
	private JTextField extendedDataTextField;
	
	private JButton advancedSelectorButton;
	private SelectorListGUI listSelector;
	private JDialog advancedSelectorDialog;
	
	private ConnectionParameters connectionParameters;
	private volatile Selector selector; 
	
	private ListenerList selectorGUIListeners = new ListenerList(SelectorGUIListener.class);
	
	private PropertyChangeListener selectorPropertyListener;
	private QueryListener selectorQueryListener;
	private ExceptionListener exceptionListener;
	private CaseInsensitiveComparator comparator = CaseInsensitiveComparator.getDefault();
	
	private AcopSpider spiderBean;
	private boolean onlyConnectionParametersMode = false;
	private boolean settingFromSelector = false;
	private JPopupMenu popupMenu;
	private JPopupMenu selectionPopup;
	
	private Dimension comboDim = new Dimension(170,21);
	
	/**
	 * Constructs a new SelectorGUI.
	 *
	 */
	public SelectorGUI() {
		super();
		initialize();
	}
	
	private void initialize() {
		deviceContextModel= new AcopComboBoxModel<>(String.class);
		subSystemModel = new AcopComboBoxModel<>(String.class);
		deviceServerModel= new AcopComboBoxModel<>(String.class);
		deviceNameModel= new AcopComboBoxModel<>(String.class);
		propertyNameModel= new AcopComboBoxModel<>(String.class);
		accessProtocolModel= new AcopComboBoxModel<>(String.class);
		arraySizeModel = new AcopComboBoxModel<>(String.class);
		extendedDataModel = new AcopComboBoxModel<>(ExtendedData.class);
		setSelector(new Selector());
		
		this.setLayout(new GridBagLayout());
		this.add(getCommandPanel(), new GridBagConstraints(0,0,1,1,1,1,GridBagConstraints.CENTER, GridBagConstraints.BOTH, new Insets(0,0,0,0),0,0));
		this.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseReleased(MouseEvent e) {
				if (SwingUtilities.isRightMouseButton(e)) {
					getPopUpMenu().show(SelectorGUI.this, e.getX(), e.getY());
				}
			}
		});
		
	}
	
	private JPopupMenu getPopUpMenu() {
		if (popupMenu == null) {
			popupMenu = new JPopupMenu();
			JMenuItem reset = new JMenuItem("Reset selector");
			reset.addActionListener(new ActionListener(){
				public void actionPerformed(ActionEvent e) {
					getExecutorService().shutdownNow();
					executorService = null;
					setSelector(new Selector());
				}
			});
			popupMenu.add(reset);
		}
		return popupMenu;
	}
		
	private JPopupMenu getSelectionPopUpMenu() {
		if (selectionPopup == null) {
			selectionPopup = new JPopupMenu();
			JMenuItem copy = new JMenuItem("Copy");
			copy.addActionListener(new ActionListener(){
				public void actionPerformed(ActionEvent e) {
					Clipboard cb = getToolkit().getSystemClipboard();
					StringSelection strSel = new StringSelection(getSelectionLabel().getText());
					cb.setContents(strSel, strSel);
				}
			});
			selectionPopup.add(copy);
		}
		return selectionPopup;
	}
	
	private void applySettings() {
		
		try {
			String remoteName = getSelector().getRemoteName();
			int accessRate = getAccessRateField().getValue().intValue(); 
			ConnectionParameters old = connectionParameters;
			connectionParameters = new ConnectionParameters(remoteName,(AccessMode)getAccessModeCombo().getSelectedItem(), accessRate, getArraySize(), getExtendedData());
			firePropertyChange(ConnectionParametersReceiver.CONNECTION_PARAMETERS_PROPERTY,old,getConnectionParameters());
        } catch (Exception e) {
            handleException(e);
        } 
	}
	
	private int getArraySize() {
		String arraySize = arraySizeModel.getSelectedItem();
		if (arraySize == null || AUTO.equals(arraySize)) return ConnectionParameters.AUTO_PROPERTY_SIZE;
		else if (arraySize.startsWith(ACTUAL)) return ConnectionParameters.ACTUAL_PROPERTY_SIZE;
		else return Integer.parseInt(arraySize);
		
	}
	
	private JPanel getCommandPanel() {
		if (commandPanel == null) {
			commandPanel = new JPanel(new GridBagLayout());
			accessProtocolLabel = new JLabel("Access protocol");
			GUIUtilities.setFixedSize(accessProtocolLabel,new Dimension(120,21));
			commandPanel.add(accessProtocolLabel, new GridBagConstraints(0,0,1,1,1,1,GridBagConstraints.EAST, GridBagConstraints.NONE, new Insets(0,1,1,1), 0,0));
			commandPanel.add(getAccessProtocolCombo(), new GridBagConstraints(1,0,2,1,1,1,GridBagConstraints.WEST, GridBagConstraints.NONE, new Insets(0,1,1,1), 0,0));
			
			deviceContextLabel = new JLabel("Device context");
			GUIUtilities.setFixedSize(deviceContextLabel,new Dimension(120,21));
			commandPanel.add(deviceContextLabel, new GridBagConstraints(0,1,1,1,1,1,GridBagConstraints.EAST, GridBagConstraints.NONE, new Insets(0,1,1,1), 0,0));
			commandPanel.add(getDeviceContextCombo(), new GridBagConstraints(1,1,2,1,1,1,GridBagConstraints.WEST, GridBagConstraints.NONE, new Insets(0,1,1,1), 0,0));
			
			subSystemLabel = new JLabel("Device subsystem");
			GUIUtilities.setFixedSize(subSystemLabel,new Dimension(120,21));
			commandPanel.add(subSystemLabel, new GridBagConstraints(0,2,1,1,1,1,GridBagConstraints.EAST, GridBagConstraints.NONE, new Insets(0,1,1,1), 0,0));
			commandPanel.add(getSubSystemCombo(), new GridBagConstraints(1,2,2,1,1,1,GridBagConstraints.WEST, GridBagConstraints.NONE, new Insets(0,1,1,1), 0,0));
						
			deviceServerLabel = new JLabel("Device server");
			GUIUtilities.setFixedSize(deviceServerLabel,new Dimension(120,21));
			commandPanel.add(deviceServerLabel, new GridBagConstraints(0,3,1,1,1,1,GridBagConstraints.EAST, GridBagConstraints.NONE, new Insets(0,1,1,1), 0,0));
			commandPanel.add(getDeviceServerCombo(), new GridBagConstraints(1,3,2,1,1,1,GridBagConstraints.WEST, GridBagConstraints.NONE, new Insets(0,1,1,1), 0,0));
			
			deviceNameLabel = new JLabel("Device name");
			GUIUtilities.setFixedSize(deviceNameLabel,new Dimension(120,21));
			commandPanel.add(deviceNameLabel, new GridBagConstraints(0,4,1,1,1,1,GridBagConstraints.EAST, GridBagConstraints.NONE, new Insets(0,1,1,1), 0,0));
			commandPanel.add(getDeviceNameCombo(), new GridBagConstraints(1,4,2,1,1,1,GridBagConstraints.WEST, GridBagConstraints.NONE, new Insets(0,1,1,1), 0,0));
			
			propertyNameLabel = new JLabel("Property name");
			GUIUtilities.setFixedSize(propertyNameLabel,new Dimension(120,21));
			commandPanel.add(propertyNameLabel, new GridBagConstraints(0,5,1,1,1,1,GridBagConstraints.EAST, GridBagConstraints.NONE, new Insets(0,1,1,1), 0,0));
			commandPanel.add(getPropertyNameCombo(), new GridBagConstraints(1,5,2,1,1,1,GridBagConstraints.WEST, GridBagConstraints.NONE, new Insets(0,1,1,1), 0,0));
			
			commandPanel.add(getShowStockPropertiesCheckBox(), new GridBagConstraints(0,6,3,1,1,1,GridBagConstraints.CENTER, GridBagConstraints.NONE, new Insets(0,1,1,1), 0,0));
						
			accessRateLabel = new JLabel("Access rate");
			GUIUtilities.setFixedSize(accessRateLabel,new Dimension(120,21));
			commandPanel.add(accessRateLabel, new GridBagConstraints(0,7,1,1,1,1,GridBagConstraints.EAST, GridBagConstraints.NONE, new Insets(0,1,1,1), 0,0));
			commandPanel.add(getAccessRateField(), new GridBagConstraints(1,7,2,1,1,1,GridBagConstraints.WEST, GridBagConstraints.NONE, new Insets(0,1,1,1), 0,0));
			
			accessModeLabel = new JLabel("Access mode");
			GUIUtilities.setFixedSize(accessModeLabel,new Dimension(120,21));
			commandPanel.add(accessModeLabel, new GridBagConstraints(0,8,1,1,1,1,GridBagConstraints.EAST, GridBagConstraints.NONE, new Insets(0,1,1,1), 0,0));
			commandPanel.add(getAccessModeCombo(), new GridBagConstraints(1,8,2,1,1,1,GridBagConstraints.WEST, GridBagConstraints.NONE, new Insets(0,1,1,1), 0,0));
			
			
			arraySizeLabel = new JLabel("Data size");
			GUIUtilities.setFixedSize(arraySizeLabel,new Dimension(120,21));
			arraySizeLabel.setToolTipText("Size of connected array, ACTUAL takes actual size, AUTO adjust size to the widget.");
			commandPanel.add(arraySizeLabel, new GridBagConstraints(0,9,1,1,1,1,GridBagConstraints.EAST, GridBagConstraints.NONE, new Insets(0,1,1,1), 0,0));
			commandPanel.add(getArraySizeCombo(), new GridBagConstraints(1,9,2,1,1,1,GridBagConstraints.WEST, GridBagConstraints.NONE, new Insets(0,1,1,1), 0,0));
			
			extendedDataLabel = new JLabel("Extended Data");
			GUIUtilities.setFixedSize(extendedDataLabel,new Dimension(120,21));
			commandPanel.add(extendedDataLabel, new GridBagConstraints(0,10,1,1,1,1,GridBagConstraints.EAST, GridBagConstraints.NONE, new Insets(0,1,1,1),0,0));
			JPanel p = new JPanel(new GridBagLayout());
			p.add(getExtendedDataCombo(), new GridBagConstraints(0,0,1,1,0,1,GridBagConstraints.CENTER,GridBagConstraints.VERTICAL,new Insets(0,0,1,1),0,0));
			p.add(getExtendedDataTextField(), new GridBagConstraints(1,0,1,1,1,0,GridBagConstraints.CENTER,GridBagConstraints.BOTH,new Insets(0,0,0,0),0,0));
			GUIUtilities.setFixedSize(p,new Dimension(comboDim.width, comboDim.height + 1));
			commandPanel.add(p, new GridBagConstraints(1,10,2,1,1,1,GridBagConstraints.WEST, GridBagConstraints.NONE, new Insets(0,1,1,1), 0,0));
			commandPanel.add(getAdvancedSelectorButton(),new GridBagConstraints(0,11,3,1,1,0,GridBagConstraints.CENTER, GridBagConstraints.NONE, new Insets(5,1,5,1),0,0));
			commandPanel.add(getSelectionLabel(), new GridBagConstraints(0,12,3,1,1,0,GridBagConstraints.CENTER, GridBagConstraints.HORIZONTAL, new Insets(0,1,1,1),0,0));
			
			if (!Beans.isDesignTime()) {
				commandPanel.add(getDescriptionLabel(), new GridBagConstraints(0,13,3,1,1,0,GridBagConstraints.CENTER, GridBagConstraints.NONE, new Insets(0,1,1,1),0,0));
			} else {
				commandPanel.add(getSpiderBean(), new GridBagConstraints(1,13,2,1,1,0,GridBagConstraints.EAST, GridBagConstraints.NONE, new Insets(0,1,1,1),0,0));
			}
			
			setCombosEnabled(true);
		}
		return commandPanel;
	}
	
	private JComboBox<AccessMode> getAccessModeCombo() {
		if (accessModeCombo == null) {
			accessModeCombo = new JComboBox<>(AccessMode.values());
			accessModeCombo.removeItem(AccessMode.WRITE);
			accessModeCombo.removeItem(AccessMode.WRITE_CONNECT);
			accessModeCombo.setSelectedItem(AccessMode.POLL);
			GUIUtilities.setFixedSize(accessModeCombo,comboDim);
			accessModeCombo.setRenderer(new ComboRenderer());
			accessModeCombo.addActionListener(new ActionListener(){
				public void actionPerformed(ActionEvent e) {
					if (settingFromSelector) return;
					refreshSelectionLabel();
					getSelector().setAccessMode((AccessMode)accessModeCombo.getSelectedItem());
				}
			});
		}
		return accessModeCombo;
	}
	
	private JLabel getSelectionLabel() {
		if (selectionLabel == null) {
			selectionLabel = new JLabel();
			GUIUtilities.setFixedSize(selectionLabel,new Dimension(240,21));
			selectionLabel.setIcon(IconHelper.createIcon("icons/development/ApplicationDeploy16.gif"));
			selectionLabel.setHorizontalAlignment(SwingConstants.CENTER);
			selectionLabel.addMouseMotionListener(new MouseMotionAdapter() {
				
				@Override
				public void mouseDragged(MouseEvent e) {
					applySettings();
					selectionLabel.getTransferHandler().exportAsDrag(selectionLabel,e,TransferHandler.COPY);
				}
			
			});
			
			selectionLabel.addMouseListener(new MouseAdapter() {
				@Override
				public void mouseReleased(MouseEvent e) {
					if (e.getButton() == MouseEvent.BUTTON3 && e.isPopupTrigger()) {
						getSelectionPopUpMenu().show(selectionLabel, e.getX(), e.getY());
					}
				}
			});
					
			selectionLabel.setTransferHandler(new TransferHandler() {

				private static final long serialVersionUID = 1658117801619157639L;
				private DisplayerParametersSelectorDialog dialog;
				@Override
				public boolean importData(JComponent comp, Transferable t) {
					try {
						if (t.isDataFlavorSupported(DataFlavor.stringFlavor)) {
							if (getConnectionParameters() == null) return false;
							String input = t.getTransferData(DataFlavor.stringFlavor).toString();
							StringTokenizer st = new StringTokenizer(input, "\n");
							ArrayList<ConnectionParameters> al = new ArrayList<>(0);
							while (st.hasMoreTokens()) {
								al.add(new ConnectionParameters(st.nextToken(), AccessMode.POLL,1000));
							}
							ConnectionParameters[] cp = al.toArray(new ConnectionParameters[]{});
							int idx = selectByDialog(getSelectionLabel(), cp);
							if (idx < 0) return false;
							setConnectionParameters(cp[idx]);
							return true;
						}
					} catch (Exception e) {
						handleException(e);
					}
					return false;
				}

				@Override
				protected Transferable createTransferable(JComponent c) {
					return new Transferable() {
					
						public Object getTransferData(DataFlavor flavor)
								throws UnsupportedFlavorException, IOException {
							if (flavor.isFlavorTextType()) {
								return getConnectionParameters().getRemoteName();
							}
							throw new UnsupportedFlavorException(flavor);
						}
					
						public boolean isDataFlavorSupported(DataFlavor flavor) {
							return flavor.isFlavorTextType();
						}
					
						public DataFlavor[] getTransferDataFlavors() {
							return new DataFlavor[]{DataFlavor.stringFlavor};
						}
					};
				}

				@Override
				public boolean canImport(JComponent comp, DataFlavor[] transferFlavors) {
					for (int i = 0; i < transferFlavors.length; i++) {
						if (transferFlavors[i].isFlavorTextType()) {
							return true;
						}
					}
					return false;
				}
				@Override
				public int getSourceActions(JComponent c) {
					return TransferHandler.COPY;
				}
				
				protected int selectByDialog(Component c, ConnectionParameters[] dp) {
					if (dialog == null) {
						dialog = new DisplayerParametersSelectorDialog(c, "Adding Source", "Select Source:");
					}
					String[] s= new String[dp.length];
					for (int i = 0; i < s.length; i++) {
						s[i]=dp[i].getRemoteName();
					}
					return dialog.showSelectionDialog(c,s);
				}
			});
			
			refreshSelectionLabel();
		}
		return selectionLabel;
	}
	
	private JLabel getDescriptionLabel() {
		if (descriptionLabel == null) {
	        descriptionLabel = new JLabel();
	        GUIUtilities.setFixedSize(descriptionLabel,new Dimension(240,21));
        }

        return descriptionLabel;
	}
	
	private JComboBox<String> getAccessProtocolCombo() {
		if (accessProtocolCombo == null) {
			accessProtocolCombo = new JComboBox<>();
			accessProtocolCombo.setModel(accessProtocolModel);
			accessProtocolModel.setElements(new String[]{Selector.PROTOCOL_TINE});
			accessProtocolModel.setSelectedItem(null);
			GUIUtilities.setFixedSize(accessProtocolCombo,comboDim);
			accessProtocolCombo.setRenderer(new ComboRenderer());
			accessProtocolCombo.addActionListener(new ActionListener(){
				public void actionPerformed(ActionEvent e) {
					if (settingFromSelector) return;
					getExecutorService().execute(new Runnable() {
						public void run() {
							getSelector().setSelectedProtocol(accessProtocolModel.getSelectedItem());							
						}
					});
                }
			});
		}
		return accessProtocolCombo;
	}
	
	private JComboBox<String> getSubSystemCombo() {
		if (subSystemCombo == null) {
			subSystemCombo = new JComboBox<>();
//			Set<String> subSysName = Selector.SUB_SYSTEM.keySet();
//			String[] subSys = new String[subSysName.size()];
//			subSys = subSysName.toArray(subSys);
//			subSystemModel.setElements(subSys);
			subSystemCombo.setModel(subSystemModel);
			GUIUtilities.setFixedSize(subSystemCombo,comboDim);
			subSystemCombo.setRenderer(new ComboRenderer());
			subSystemCombo.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					if (settingFromSelector) return;
					getExecutorService().execute(new Runnable() {
						public void run() {
							getSelector().setSelectedSubSystem(subSystemModel.getSelectedItem());							
						}
					});
                }
			});
		}
		return subSystemCombo;
	}
	
	private JComboBox<String> getDeviceContextCombo() {
		if (deviceContextCombo == null) {
			deviceContextCombo = new JComboBox<>();
			deviceContextCombo.setModel(deviceContextModel);
			GUIUtilities.setFixedSize(deviceContextCombo,comboDim);
			deviceContextCombo.setRenderer(new ComboRenderer());
			deviceContextCombo.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					if (settingFromSelector) return;
					getExecutorService().execute(new Runnable() {
						public void run() {
							getSelector().setSelectedContext(deviceContextModel.getSelectedItem());
						}
					});
					
                }
			});
		}
		return deviceContextCombo;
	}
	
	private JComboBox<String> getDeviceServerCombo() {
		if (deviceServerCombo == null) {
			deviceServerCombo = new JComboBox<>();
			deviceServerCombo.setModel(deviceServerModel);
			GUIUtilities.setFixedSize(deviceServerCombo,comboDim);
			deviceServerCombo.setRenderer(new ComboRenderer());
			deviceServerCombo.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					if (settingFromSelector) return;
					getExecutorService().execute(new Runnable() {
						public void run() {
							getSelector().setSelectedServer(deviceServerModel.getSelectedItem());
						}
					});
                }
			});
		}
		return deviceServerCombo;
	}
	
	private JComboBox<String> getDeviceNameCombo() {
		if (deviceNameCombo == null) {
			deviceNameCombo = new JComboBox<>();
			deviceNameCombo.setModel(deviceNameModel);
			GUIUtilities.setFixedSize(deviceNameCombo,comboDim);
			deviceNameCombo.setRenderer(new ComboRenderer());
			deviceNameCombo.setEditable(true);
			deviceNameCombo.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					if (settingFromSelector) return;
					getExecutorService().execute(new Runnable(){
						public void run() {
							getSelector().setSelectedDevice(deviceNameModel.getSelectedItem());
							fireSelectorGUIEvent(SelectorUtilities.isConnectionParametersValid(getConnectionParameters()));
						}
					});
                }
			});
		}
		return deviceNameCombo;
	}
	
	private JComboBox<String> getPropertyNameCombo() {
		if (propertyNameCombo == null) {
			propertyNameCombo = new JComboBox<>();
			propertyNameCombo.setModel(propertyNameModel);
			GUIUtilities.setFixedSize(propertyNameCombo,comboDim);
			propertyNameCombo.setRenderer(new ComboRenderer());
			propertyNameCombo.setEditable(true);
			propertyNameCombo.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					if (settingFromSelector) return;
					getExecutorService().execute(new Runnable() {
						public void run() {
							String obj = ((String)propertyNameCombo.getEditor().getItem());
							if (obj != null) {
								getSelector().setSelectedProperty(obj);
								fireSelectorGUIEvent(SelectorUtilities.isConnectionParametersValid(getConnectionParameters()));
							} else if (propertyNameCombo.getSelectedIndex() < 0 || propertyNameCombo.getSelectedItem() == null) {
								getSelector().setSelectedProperty(null);
								fireSelectorGUIEvent(false);
							} else {
								getSelector().setSelectedProperty(propertyNameModel.getSelectedItem());
								fireSelectorGUIEvent(true);
							}
						}
					});
                }
			});
		}
		return propertyNameCombo;
	}
	
	private JCheckBox getShowStockPropertiesCheckBox() {
		if (showStockProperties == null) {
			showStockProperties = new JCheckBox("Show stock properties");
			showStockProperties.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					if (settingFromSelector) return;
					getExecutorService().execute(new Runnable() {
						public void run() {
							getSelector().setShowStockProperties(showStockProperties.isSelected());							
						}
					});
				}
			});
		}
		return showStockProperties;
	}
	
	private NumberField getAccessRateField() {
		if (accessRateField == null) {
			accessRateField = new NumberField();
			accessRateField.setNumberType(Integer.class);
			accessRateField.setFormat("%d");
			accessRateField.setValue(1000);
			GUIUtilities.setFixedSize(accessRateField,comboDim);
			accessRateField.setText(String.valueOf(1000));
			accessRateField.addPropertyChangeListener("value", new PropertyChangeListener() {
				public void propertyChange(PropertyChangeEvent evt) {
					if (settingFromSelector) return;
					refreshSelectionLabel();
					getSelector().setAccessRate(accessRateField.getValue().intValue());	
				}
			});
		}
		return accessRateField;
	}
	
	private JComboBox<String> getArraySizeCombo() {
		if (arraySizeCombo == null) {
			arraySizeCombo = new JComboBox<>();
			arraySizeCombo.setEditable(true);
			GUIUtilities.setFixedSize(arraySizeCombo,comboDim);
			arraySizeCombo.setRenderer(new ComboRenderer());
			arraySizeCombo.setModel(arraySizeModel);
			arraySizeModel.setElements(new String[]{"1",ACTUAL,AUTO});
			arraySizeModel.setSelectedItem(AUTO);
			arraySizeCombo.setToolTipText("Size of connected array, ACTUAL takes actual size, AUTO adjusts size to the widget.");
			arraySizeCombo.addActionListener(new ActionListener() {
				@SuppressWarnings("unchecked")
				public void actionPerformed(ActionEvent e) {
					Object o = arraySizeCombo.getEditor().getItem();
				    try {
						int size = Integer.parseInt((String)o);
						if (size <=0 ) return;
					} catch(NumberFormatException ex) {
						if (o.toString().startsWith(ACTUAL)) {
							arraySizeModel.setSelectedItem(o);
						} else {
							arraySizeModel.setSelectedItem(AUTO);
						}
						refreshSelectionLabel();
						return;
					}
					if (arraySizeModel.contains(o)) {
						refreshSelectionLabel();
						return;
					}
					String[] elements = new String[arraySizeModel.getElements().length + 1];
					for (int i = 0; i < arraySizeModel.getElements().length; i ++) {
						elements[i] = arraySizeModel.getElements()[i];
					}
					elements[elements.length-1] = (String)o;
					Arrays.sort(elements, comparator);
					arraySizeModel.setElements(elements);
					refreshSelectionLabel();
				}
			});
		}

		return arraySizeCombo;
	}
		
	private JTextField getExtendedDataTextField() {
		if(extendedDataTextField == null) {
			extendedDataTextField = new JTextField();	
			extendedDataTextField.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent arg0) {
					if (settingFromSelector) return;
					refreshSelectionLabel();
				}
			});
			extendedDataTextField.addFocusListener(new FocusListener() {
				public void focusLost(FocusEvent e) {
					if (settingFromSelector) return;
					refreshSelectionLabel();
					}
				public void focusGained(FocusEvent e) {}
			});
		}
		return extendedDataTextField;
	}

	private JComboBox<ExtendedData> getExtendedDataCombo() {
		if (extendedDataCombo == null) {
			extendedDataCombo = new JComboBox<>();
			extendedDataCombo.setRenderer(new ComboRenderer());
			extendedDataCombo.setModel(extendedDataModel);
			extendedDataModel.setElements(ExtendedData.values());
			//bugfix
			SwingUtilities.invokeLater(new Runnable() {
				public void run() {
					extendedDataModel.setSelectedItem(ExtendedData.DOUBLE);				
				}
			});
			extendedDataCombo.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					if (settingFromSelector) return;
					refreshSelectionLabel();
				}
			});
		}
		
		return extendedDataCombo;
	}
	
	private Object getExtendedData() {
		try {
    		ExtendedData ed = (ExtendedData) getExtendedDataCombo().getSelectedItem();
    		if(getExtendedDataTextField().getText().isEmpty()){
    			return null;
    		}
    		String[] stringArray = getExtendedDataTextField().getText().split(";");
    		switch (ed) {
    		case DOUBLE:{
    			double[] dbl = new double[stringArray.length];
    			for(int i=0;i<dbl.length;i++)
    				dbl[i]=Double.parseDouble(stringArray[i]);
    			return dbl;
    		}
    		case FLOAT:{
    			float[] flt = new float[stringArray.length];
    			for(int i=0;i<flt.length;i++)
    				flt[i]=Float.parseFloat(stringArray[i]);
    			return flt;
    		}
    		case INT32:
    		case INT:{
    			int[] intgr = new int[stringArray.length];
    			for(int i=0;i<intgr.length;i++)
    				intgr[i]=Integer.parseInt(stringArray[i]);
    			return intgr;			
    		}
    		case INT16:
    		case SHORT:{
    			short[] shrt = new short[stringArray.length];
    			for(int i=0;i<shrt.length;i++)
    				shrt[i]=Short.parseShort(stringArray[i]);
    			return shrt;	
    		}
    		case BYTE:{
    			byte[] byt = new byte[stringArray.length];
    			for(int i=0;i<byt.length;i++)
    				byt[i]=Byte.parseByte(stringArray[i]);
    			return byt;	
    		}
    		case INT64:
    		case LONG:{
    			long[] lng = new long[stringArray.length];
    			for(int i=0;i<lng.length;i++)
    				lng[i]=Long.parseLong(stringArray[i]);
    			return lng;	
    		}
    		case CHAR:{
    			char[] chr = new char[stringArray.length];
    			for(int i=0;i<chr.length;i++)
    				chr[i]=stringArray[i].charAt(0);
    			return chr;	
    		}
    		case STRING:{
    			return extendedDataTextField.getText();
    		}
    		default: return null;
    		}
		} catch (Exception e) {
			return null;
		}
	}
	
	private void updateArraySizeCombo(ConnectionParameters param) {
		String[] elements = arraySizeModel.getElements();
		if (param == null) {
			for (int i = 0; i < elements.length; i++) {
				if (elements[i].startsWith(ACTUAL)) {
					arraySizeModel.replaceElement(i, ACTUAL);
					return;
				}
			}
		} else {
			for (int i = 0; i < elements.length; i++) {
				if (elements[i].startsWith(ACTUAL)) {
					arraySizeModel.replaceElement(i, ACTUAL + " (" + getActualSize(param) + ")");
					return;
				}
			}
		}
	}
	
	private void refreshSelectionLabel() {
		applySettings();
		String text = connectionParameters.toString();
		getSelectionLabel().setText(text);
		getSelectionLabel().setToolTipText(text);
	}
	
	private void update(final ConnectionParameters params) {
		if (params == null || params.equals(connectionParameters)) return;
		ConnectionParameters old = connectionParameters;
		connectionParameters = params;
		getExecutorService().execute(new Runnable() {
			public void run() {
				getSelector().setSelectedProtocol(params.getAccessProtocol());
				getSelector().setSelectedContext(params.getDeviceContext());
				getSelector().setSelectedServer(params.getDeviceGroup());
				getSelector().setSelectedDevice(params.getDeviceName());
				getSelector().setSelectedProperty(params.getDeviceProperty());
                // due to device query precedence device has to be reselected
				getSelector().setSelectedDevice(params.getDeviceName());
				getSelector().setAccessMode(params.getAccessMode()); // mdavid: fixed (?)
				getSelector().setAccessRate(params.getAccessRate());
			}
		});

		int size = params.getPropertySize();
		if (size == -1) {
			getArraySizeCombo().setSelectedItem(AUTO);
		} else if (size == 0){
			String o = ACTUAL;
			for (String s :arraySizeModel.getElements()) {
				if (s.startsWith(ACTUAL)) {
					o = s;
				}
			}
			getArraySizeCombo().setSelectedItem(o);
		} else {
			getArraySizeCombo().setSelectedItem(String.valueOf(size));
		}
		
		if (SelectorUtilities.isConnectionParametersValid(connectionParameters)) {
			getExecutorService().execute(new Runnable() {
				public void run() {
					String text = SelectorUtilities.getInfo(connectionParameters);
					getDescriptionLabel().setText(text);
			        getDescriptionLabel().setToolTipText(text);
			        updateArraySizeCombo(connectionParameters);
				}
			});
		}
		
        firePropertyChange(ConnectionParametersReceiver.CONNECTION_PARAMETERS_PROPERTY,old,connectionParameters);
	}
	
	/*
	 * (non-Javadoc)
	 * @see java.beans.Customizer#setObject(java.lang.Object)
	 */
	public void setObject(Object bean) {
		if (bean instanceof MultipleAcopDisplayer<?>) {
			disp = bean;
		} else if (bean instanceof ConnectionParametersReceiver) {
			disp = bean;
			update(((ConnectionParametersReceiver)disp).getConnectionParameters());
			((ConnectionParametersReceiver)disp).addPropertyChangeListener(ConnectionParametersReceiver.CONNECTION_PARAMETERS_PROPERTY, new PropertyChangeListener(){

				public void propertyChange(PropertyChangeEvent evt) {
					if (!ConnectionParametersReceiver.CONNECTION_PARAMETERS_PROPERTY.equals(evt.getPropertyName())) return;
					update(((ConnectionParametersReceiver)disp).getConnectionParameters());
				}
				
			});   
		} else {
			throw new IllegalArgumentException("Only ConnectionParametersReceiver and" +
				" MultipleAcopDisplayer can use this Customizer.");
		}
		
		     
    }

	/**
	 * Returns the selected ConnectionParameters.
	 * 
     * @return the parameters
     */
    public ConnectionParameters getConnectionParameters() {
    	return connectionParameters;
    }

	/**
	 * Sets the ConnectionParameters and applies the values to the UI.
	 * 
     * @param parameters the parameters to set
     */
    public void setConnectionParameters(ConnectionParameters parameters) {
    	if (this.connectionParameters != null && this.connectionParameters.equals(parameters)) return;
		update(parameters);
    }
    
    protected void setSelector(Selector selector) {
    	if (this.selector != null && this.selector.equals(selector)) return;
    	if (this.selector != null) {
	    	this.selector.removePropertyChangeListener(getSelectorPropertyListener());
	    	this.selector.removeQueryListener(getSelectorQueryListener());
	    	this.selector.removeExceptionListener(getExceptionListener());
    	}
    	Selector old = this.selector;
    	this.selector = selector;
    	if (this.selector != null) {
	    	this.selector.addPropertyChangeListener(getSelectorPropertyListener());
	    	this.selector.addQueryListener(getSelectorQueryListener());
	    	this.selector.addExceptionListener(getExceptionListener());
    	}
    	getListSelector().setSelector(selector);
    	
    	getExecutorService().execute(new Runnable() {
			public void run() {
		    	SelectorGUI.this.selector.setSelectedProtocol(Selector.PROTOCOL_TINE);
		    	fillAndSelectValue();
		    	refreshSelectionLabel();
			}
    	});
//    	fillAndSelectValue();
//    	refreshSelectionLabel();
    	firePropertyChange("selector", old, this.selector);
    }
    
	protected Selector getSelector() {
        return selector;
    }
	
	private ExceptionListener getExceptionListener() {
		if (exceptionListener == null) {
			exceptionListener = new ExceptionListener() {
				public void exceptionReceived(ExceptionEvent e) {
					handleException(e.getThrowable());
				}
			};
		}
		return exceptionListener;
	}
	
	protected void handleException(Throwable e) {
		if (Beans.isDesignTime()) {
			getSpiderBean().addError(e);
		} else {
			e.printStackTrace();
		}
	}
	
	private AcopSpider getSpiderBean() {
		if (spiderBean == null) {
			spiderBean = new AcopSpider();
			spiderBean.setAutoResize(false);
			spiderBean.setSize(new Dimension(30,30));
			spiderBean.setMaximumSize(new Dimension(30,30));
			spiderBean.setMinimumSize(new Dimension(30,30));
			spiderBean.setPreferredSize(new Dimension(30,30));
			spiderBean.setSize(new Dimension(30,30));
			spiderBean.rescaleIconToFit();
		}
		return spiderBean;
	}
	
	private PropertyChangeListener getSelectorPropertyListener() {
    	if (selectorPropertyListener == null) {
    		selectorPropertyListener = new PropertyChangeListener(){
	        	public void propertyChange(PropertyChangeEvent evt) {
	        		if (getSelector() == null) return;
	        		if (!getPropertyNameCombo().isEnabled()) return;
	        		final String name = evt.getPropertyName();
	        		if ("accessRate".equals(name)) {
	        			getAccessRateField().setValue(getSelector().getAccessRate());
	        			refreshSelectionLabel();
	        			getAccessRateField().requestFocus();
	        			return;
	        		} else if ("accessMode".equals(name)) {
	        			getAccessModeCombo().setSelectedItem(getSelector().getAccessMode());
	        			refreshSelectionLabel();
	        			getAccessModeCombo().requestFocus();
	        			return;
	        		}
	        		
            		if ("selectedProtocol".equals(name)) {
            			accessProtocolModel.setSelectedItem(getSelector().getSelectedProtocol());
            		} else if ("selectedContext".equals(name)) {
            			deviceContextModel.setSelectedItem(getSelector().getSelectedContext());
            		} else if ("selectedSubSystem".equals(name)) {
            			subSystemModel.setSelectedItem(getSelector().getSelectedSubSystem());
            		} else if ("selectedServer".equals(name)) {
            			deviceServerModel.setSelectedItem(getSelector().getSelectedServer());
            		} else if ("selectedDevice".equals(name)) {
            			deviceNameModel.setSelectedItem(getSelector().getSelectedDevice());
            		} else if ("selectedProperty".equals(name)) {
            			propertyNameModel.setSelectedItem(getSelector().getSelectedProperty());
            		} else if ("availableContexts".equals(name)) {
            			deviceContextModel.setElements(getSelector().getAvailableContexts());
            		} else if ("availableSubsystems".equals(name)) {
            			subSystemModel.setElements(getSelector().getAvailableSubsystems());
            		} else if ("availableServers".equals(name)) {
            			deviceServerModel.setElements(getSelector().getAvailableServers());
            		} else if ("availableDevices".equals(name)) {
            			deviceNameModel.setElements(getSelector().getAvailableDevices());
            		} else if ("availableProperties".equals(name)) {
            			propertyNameModel.setElements(getSelector().getAvailableProperties());
            		} else if ("showStockProperties".equals(name)) {
            			getShowStockPropertiesCheckBox().setSelected(getSelector().isShowStockProperties());
            		}
            		refreshSelectionLabel();
	                
                    if (!Beans.isDesignTime()) {
		        		if ("selectedProperty".equals(evt.getPropertyName())) {
		        			if (evt.getNewValue() == null) {
		        				getDescriptionLabel().setText("");
	                            getDescriptionLabel().setToolTipText("");
	                            return;
		        			}
		        			if (evt.getNewValue() != evt.getOldValue()) {
		        				if (SelectorUtilities.isConnectionParametersValid(connectionParameters)) {
		        					String text = SelectorUtilities.getInfo(connectionParameters);
		        			        getDescriptionLabel().setText(text);
		        			        getDescriptionLabel().setToolTipText(text);
		        			        updateArraySizeCombo(connectionParameters);
		        				}
		        			}
		        		}
                    }

	        	}
	        };
    	} 
    	return selectorPropertyListener;
    }
	
	private void fillAndSelectValue() {
		settingFromSelector = true;
		getShowStockPropertiesCheckBox().setSelected(getSelector().isShowStockProperties());
		accessProtocolModel.setSelectedItem(getSelector().getSelectedProtocol());
		deviceContextModel.setElements(getSelector().getAvailableContexts());
		deviceContextModel.setSelectedItem(getSelector().getSelectedContext());
		subSystemModel.setElements(getSelector().getAvailableSubsystems());
		subSystemModel.setSelectedItem(getSelector().getSelectedSubSystem());
		deviceServerModel.setElements(getSelector().getAvailableServers());
		deviceServerModel.setSelectedItem(getSelector().getSelectedServer());
		deviceNameModel.setElements(getSelector().getAvailableDevices());
		deviceNameModel.setSelectedItem(getSelector().getSelectedDevice());
		propertyNameModel.setElements(getSelector().getAvailableProperties());
		propertyNameModel.setSelectedItem(getSelector().getSelectedProperty());
		getAccessRateField().setValue(getSelector().getAccessRate());
		getAccessModeCombo().setSelectedItem(getSelector().getAccessMode());
		settingFromSelector = false;
	}
    
    private QueryListener getSelectorQueryListener() {
    	if (selectorQueryListener == null) {
    		selectorQueryListener = new QueryListener() {
				public void queryFinished(QueryEvent e) {
					if (getSelector() == null) return;
					synchronized(selector) {
						fillAndSelectValue();
						refreshSelectionLabel();
						setCombosEnabled(true);
						if (SelectorUtilities.isConnectionParametersValid(connectionParameters)) {
							String text = SelectorUtilities.getInfo(connectionParameters);
					        getDescriptionLabel().setText(text);
					        getDescriptionLabel().setToolTipText(text);
					        updateArraySizeCombo(connectionParameters);
						} else {
							getDescriptionLabel().setText("");
					        getDescriptionLabel().setToolTipText("");
					        updateArraySizeCombo(null);
						}
					}
				}
				public void queryStarted(QueryEvent e) {
					if (getSelector() == null) return;
					synchronized (selector) {
						setCombosEnabled(false);
					}
				    
				}
    		};
    	}
    	return selectorQueryListener;
    }
	
    private boolean combosEnabled = true; 
	private void setCombosEnabled(boolean enabled) {
		
		if (enabled && getSelector().isBusy()) {
			enabled = false;
		}
		if (this.combosEnabled == enabled) return;
		this.combosEnabled = enabled;
		SwingUtilities.invokeLater(new Runnable() {
			
			@Override
			public void run() {
				getAccessProtocolCombo().setEnabled(combosEnabled);
			    getDeviceContextCombo().setEnabled(combosEnabled);
			    getSubSystemCombo().setEnabled(combosEnabled);
			    getDeviceServerCombo().setEnabled(combosEnabled);
			    getDeviceNameCombo().setEnabled(combosEnabled);
			    getPropertyNameCombo().setEnabled(combosEnabled);
			
			}
		});
		
	}
	
	/**
	 * Adds a SelectorGUIListener.
	 * 
	 * @param l listener to be added
	 */
	public void addSelectorGUIListener(SelectorGUIListener l) {
		selectorGUIListeners.add(l);
	}
	
	/**
	 * Removes SelectorGUIListener.
	 * 
	 * @param l listener to be removed
	 */
	public void removeSelectorGUIListener(SelectorGUIListener l) {
		selectorGUIListeners.remove(l);
	}
	
	protected void fireSelectorGUIEvent(boolean selectionComplete) {
		SelectorGUIListener[] listeners = (SelectorGUIListener[]) selectorGUIListeners.toArray();
		SelectorGUIEvent e = new SelectorGUIEvent(this, selectionComplete);
		for (SelectorGUIListener listener : listeners) {
			listener.selectionChanged(e);
		}
	}
	
	private int getActualSize(ConnectionParameters param) {
		return SelectorUtilities.getSequenceLength(param);
	}
	
	private JButton getAdvancedSelectorButton() {
		if (advancedSelectorButton == null) {
			advancedSelectorButton = new JButton("Advanced browsing",IconHelper.createIcon("icons/navigation/ArrowRightRight16.gif"));
			advancedSelectorButton.addActionListener(new ActionListener(){
				public void actionPerformed(ActionEvent e) {
					getAdvancedSelector().setVisible(true);					
				}
			});
		}
		return advancedSelectorButton;
	}
	
	private JDialog getAdvancedSelector() {
		if (advancedSelectorDialog == null) {
			advancedSelectorDialog = new JDialog();
			advancedSelectorDialog.setSize(800,600);
			ComponentPositioner.centerOnScreen(advancedSelectorDialog);
			advancedSelectorDialog.setTitle("Advanced browsing");
			advancedSelectorDialog.setContentPane(getListSelector());
			advancedSelectorDialog.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
			
		}
		return advancedSelectorDialog;
	}
	
	private SelectorListGUI getListSelector() {
		if (listSelector == null) {
			listSelector = new SelectorListGUI();
			listSelector.setSelector(selector);
			listSelector.setExecutorService(getExecutorService());
		}
		return listSelector;
	}
	
	public static void main(String[] args) {
		try {
			
			SelectorGUI c= new SelectorGUI();
			RunnerHelper.runComponent(c,300,370);
		} catch (Exception e) {
			e.printStackTrace();
		}
		
	}
	
	private ThreadPoolExecutor executorService;
	private ThreadPoolExecutor getExecutorService() {
		if (executorService == null) {
			executorService = new  ThreadPoolExecutor(1, 1, 0, TimeUnit.NANOSECONDS,
		              new LinkedBlockingQueue<Runnable>()){
				@Override
				protected void afterExecute(Runnable r, Throwable t) {
					super.afterExecute(r, t);
					if (t != null) {
						t.printStackTrace();
					}
				}
			};
			executorService.setRejectedExecutionHandler(new RejectedExecutionHandler(){
				public void rejectedExecution(Runnable r, ThreadPoolExecutor executor) {
					//ignore
				}
			});
//			Executors.newSingleThreadExecutor();
			
        }
        return executorService;
	}
	
	/**
	 * Flag indicating whether array size choosing should also be available or only
	 * the remote name and access paramters should be available.
	 * 
	 * @param onlyCP if true only the remote name and access parameters can be selected
	 */
	public void setOnlyConnectionParametersMode(boolean onlyCP) {
		this.onlyConnectionParametersMode = onlyCP;
		arraySizeLabel.setVisible(!onlyConnectionParametersMode);
		getArraySizeCombo().setVisible(!onlyConnectionParametersMode);
	}
	
	/**
	 * Returns true if only remote name and access parameters can be selected.
	 * @return true if only remote name and access parameters can be selected
	 */
	public boolean getOnlyConnectionParametersMode() {
		return this.onlyConnectionParametersMode;
	}

}
/* __oOo__ */
