/*
 * Copyright (c) 2006 Stiftung Deutsches Elektronen-Synchroton,
 * Member of the Helmholtz Association, (DESY), HAMBURG, GERMANY.
 *
 * THIS SOFTWARE IS PROVIDED UNDER THIS LICENSE ON AN "../AS IS" BASIS.
 * WITHOUT WARRANTY OF ANY KIND, EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR PARTICULAR PURPOSE AND
 * NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE. SHOULD THE SOFTWARE PROVE DEFECTIVE
 * IN ANY RESPECT, THE USER ASSUMES THE COST OF ANY NECESSARY SERVICING, REPAIR OR
 * CORRECTION. THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF THIS LICENSE.
 * NO USE OF ANY SOFTWARE IS AUTHORIZED HEREUNDER EXCEPT UNDER THIS DISCLAIMER.
 * DESY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS,
 * OR MODIFICATIONS.
 * THE FULL LICENSE SPECIFYING FOR THE SOFTWARE THE REDISTRIBUTION, MODIFICATION,
 * USAGE AND OTHER RIGHTS AND OBLIGATIONS IS INCLUDED WITH THE DISTRIBUTION OF THIS
 * PROJECT IN THE FILE LICENSE.HTML. IF THE LICENSE IS NOT INCLUDED YOU MAY FIND A COPY
 * AT HTTP://WWW.DESY.DE/LEGAL/LICENSE.HTM
 */

package de.desy.acop.displayers.table;

import java.util.HashMap;
import java.util.Map;

import com.cosylab.gui.components.table.cells.TableCell;
import com.cosylab.gui.displayers.DataConsumer;
import com.cosylab.gui.displayers.DataState;
import com.cosylab.util.ListenerList;

import de.desy.acop.displayers.AcopTable;

/**
 * <code>AcopTableColumn</code> is a <code>DataConsumer</code> which uses
 * TableCells for displaying values. AcopTableColumn can be used in combination
 * with <code>AcopTable</code> and can represent a single column in the
 * table.
 *   
 * @author Jaka Bobnar, Cosylab
 * @see AcopTable
 *
 */
public abstract class AcopTableColumn implements DataConsumer {
	
	private String uniqueName;
	private String shortName;
	protected AcopTable table;
	protected AcopTableParameters displayerParameters;
	protected DataState dataState = new DataState(DataState.UNDEFINED);
	protected TableCell[] cells;
	protected int rowCount = 10;
	protected HashMap<String, Object> characteristics = new HashMap<>();
	
	private ListenerList columnListeners = new ListenerList(TableColumnListener.class);
	
	/**
	 * Constructs a new AcopTableColumn
	 *
	 */
	public AcopTableColumn() {
		this(null, null, "");
	}
	
	/**
	 * Constructs a new AcopTableColumn.
	 * 
	 * @param table parent table
	 * @param uniqueName unique name of the column
	 * @param shortName short name of the column
	 */
	public AcopTableColumn(AcopTable table, String uniqueName, String shortName) {
		this.table = table;
		this.uniqueName = uniqueName;
		this.shortName = shortName;
	}
	
	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getName()
	 */
	public String getName() {
		return uniqueName;
	}
	
	/**
	 * Returns the name that is printed in the table header. This is the short name
	 * of the column.
	 * 
	 * @return the header name
	 */
	public String getHeaderName() {
		return shortName;
	}
	
	/**
	 * Returns the class of the objects that this column presents. 
	 * @return column class
	 */
	public Class<?> getColumnClass() {
		return TableCell.class;
	}
	
	/**
	 * Returns the value at specified row.
	 * @param rowIndex
	 * @return
	 */
	public abstract Object getValue(int rowIndex);
	
	/**
	 * Sets the number of rows. When update is received only the as many rows as 
	 * specified here will be updated to avoid to avoid unnecessary overload. Counting
	 * starts at the first (most top) row.
	 * 
	 * @param rowCount new number of rows
	 */
	public void setRowCount(int rowCount) {
		this.rowCount = rowCount;
	}
	
	/**
	 * Returns the number of rows.
	 * @return number of rows in the column
	 */
	public int getRowCount() {
		return rowCount;
	}
	
	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getDataConsumer(java.lang.Class)
	 */
	@SuppressWarnings({"unchecked","rawtypes"})
    public DataConsumer getDataConsumer(Class type) {
		if (type.isAssignableFrom(getClass())) {
			return this;
		}
		return null;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getDefaultDataConsumer()
	 */
	public DataConsumer getDefaultDataConsumer() {
		return this;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#updateDataState(com.cosylab.gui.displayers.DataState)
	 */
	public void updateDataState(DataState state) {
		if (dataState != null && dataState.equals(state)) return;
		dataState = state;
		if (cells != null) {
			for (int i = 0; i < cells.length && i < rowCount; i++) {
				if (cells[i] == null) getValue(i);
				setDataStateOnCell(cells[i]);
			}
		}
		fireColumnChanged();
	}
	
	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#setCharacteristics(java.util.Map)
	 */
	@SuppressWarnings({"rawtypes","unchecked"})
	public void setCharacteristics(Map characteristics) {
		
		if (cells != null) {
			for (TableCell cell : cells) {
				cell.putAllCharacteristics(characteristics);
			}
		}
		
		this.characteristics.putAll(characteristics);
	}
	
	/**
	 * Destroys this column and releases any resource associated with it.
	 *
	 */
	public void destroy() {
		table = null;
		displayerParameters = null;
	}
	
	/**
	 * Adds a TableColumnListener which receives notifications when certain column
	 * values changed.
	 * 
	 * @param l
	 */
	public void addTableColumnListener(TableColumnListener l) {
		synchronized (columnListeners) {
			columnListeners.add(l);
		}
	}
	
	/**
	 * Removes a TableColumnListener.
	 * 
	 * @param l
	 */
	public void removeTableColumnListener(TableColumnListener l) {
		synchronized (columnListeners) {
			columnListeners.add(l);
		}		
	}
	
	protected void fireColumnChanged() {
		TableColumnListener[] listeners = (TableColumnListener[]) columnListeners.toArray();
		TableColumnEvent event = new TableColumnEvent(this);
		
		for (TableColumnListener l : listeners) {
			l.columnDataChanged(event);
		}
	}

	/**
	 * Returns the current data state of this column.
	 * 
	 * @return the data state
	 */
	public DataState getDataState() {
		return dataState;
	}
	
	protected void setCharacteristicsOnCell(TableCell cell) {
		cell.putCharacteristic(TableCell.TOOLTIP, cell.getIdentifier() + " '" + cell.getValue() + "'");
	}
	
	protected void setDataStateOnCell(TableCell cell) {
		if (cell == null) return;
		DataState dataState = getDataState();
		if (dataState.type.equals(DataState.ERROR)) {
			cell.putCharacteristic(TableCell.SUSPENDED, false);
			cell.putCharacteristic(TableCell.ALARM, false);
			cell.putCharacteristic(TableCell.WARNING, false);
			cell.putCharacteristic(TableCell.TIMEOUT, false);
			cell.putCharacteristic(TableCell.ERROR, true);
		} else if (dataState.type.equals(DataState.ALARM)) {
			cell.putCharacteristic(TableCell.SUSPENDED, false);
			cell.putCharacteristic(TableCell.ALARM, true);
			cell.putCharacteristic(TableCell.WARNING, false);
			cell.putCharacteristic(TableCell.TIMEOUT, false);
			cell.putCharacteristic(TableCell.ERROR, false);
		} else if (dataState.type.equals(DataState.WARNING)) {
			cell.putCharacteristic(TableCell.SUSPENDED, false);
			cell.putCharacteristic(TableCell.ALARM, false);
			cell.putCharacteristic(TableCell.WARNING, true);
			cell.putCharacteristic(TableCell.TIMEOUT, false);
			cell.putCharacteristic(TableCell.ERROR, false);
		} else if (dataState.type.equals(DataState.TIMEOUT)) {
			cell.putCharacteristic(TableCell.SUSPENDED, false);
			cell.putCharacteristic(TableCell.ALARM, false);
			cell.putCharacteristic(TableCell.WARNING, false);
			cell.putCharacteristic(TableCell.TIMEOUT, true);
			cell.putCharacteristic(TableCell.ERROR, false);
		} else if (dataState.type.equals(DataState.TIMELAG)) {
			cell.putCharacteristic(TableCell.SUSPENDED, false);
			cell.putCharacteristic(TableCell.ALARM, false);
			cell.putCharacteristic(TableCell.WARNING, false);
			cell.putCharacteristic(TableCell.TIMEOUT, true);
			cell.putCharacteristic(TableCell.ERROR, false);
		} else if (dataState.type.equals(DataState.UNDEFINED) || dataState.type.equals(DataState.NOT_INITIALIZED)) {
			cell.putCharacteristic(TableCell.SUSPENDED, true);
			cell.putCharacteristic(TableCell.ALARM, false);
			cell.putCharacteristic(TableCell.WARNING, false);
			cell.putCharacteristic(TableCell.TIMEOUT, false);
			cell.putCharacteristic(TableCell.ERROR, false);
		} else if (dataState.type.equals(DataState.NORMAL)) {
			cell.putCharacteristic(TableCell.SUSPENDED, false);
			cell.putCharacteristic(TableCell.ALARM, false);
			cell.putCharacteristic(TableCell.WARNING, false);
			cell.putCharacteristic(TableCell.TIMEOUT, false);
			cell.putCharacteristic(TableCell.ERROR, false);
		}
	}

	/**
	 * Returns the displayer parameters associated with this column.
	 * 
	 * @return the parameters
	 */
	public AcopTableParameters getDisplayerParameters() {
		return displayerParameters;
	}

	/**
	 * Sets the displayer parameters that specify the connection point and look of
	 * the column.
	 * 
	 * @param displayerParameters the parameters
	 */
	public void setDisplayerParameters(AcopTableParameters displayerParameters) {
		this.displayerParameters = displayerParameters;
	}
}
