/*
 * Copyright (c) 2006 Stiftung Deutsches Elektronen-Synchroton,
 * Member of the Helmholtz Association, (DESY), HAMBURG, GERMANY.
 *
 * THIS SOFTWARE IS PROVIDED UNDER THIS LICENSE ON AN "../AS IS" BASIS.
 * WITHOUT WARRANTY OF ANY KIND, EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR PARTICULAR PURPOSE AND
 * NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE. SHOULD THE SOFTWARE PROVE DEFECTIVE
 * IN ANY RESPECT, THE USER ASSUMES THE COST OF ANY NECESSARY SERVICING, REPAIR OR
 * CORRECTION. THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF THIS LICENSE.
 * NO USE OF ANY SOFTWARE IS AUTHORIZED HEREUNDER EXCEPT UNDER THIS DISCLAIMER.
 * DESY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS,
 * OR MODIFICATIONS.
 * THE FULL LICENSE SPECIFYING FOR THE SOFTWARE THE REDISTRIBUTION, MODIFICATION,
 * USAGE AND OTHER RIGHTS AND OBLIGATIONS IS INCLUDED WITH THE DISTRIBUTION OF THIS
 * PROJECT IN THE FILE LICENSE.HTML. IF THE LICENSE IS NOT INCLUDED YOU MAY FIND A COPY
 * AT HTTP://WWW.DESY.DE/LEGAL/LICENSE.HTM
 */

package de.desy.acop.displayers.table;

import java.util.ArrayList;
import java.util.List;

import javax.swing.event.TableModelEvent;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.TableModel;

import de.desy.acop.displayers.AcopTable;

/**
 * <code>AcopTableModel</code> is an implementation of the <code>TableModel</code>
 * used in combinataion with <code>AcopTable</code>. AcopTableModel is a read only
 * model which can hold any number of <code>AcopTableColumn</code>s. The model
 * listenes to each of this columns and updates the table according to the changes
 * in the columns. The number of columns in the model can be changed in run time.
 * 
 * @author Jaka Bobnar, Cosylab
 * @see AcopTable
 * @see AcopTableColumn
 *
 */
public class AcopTableModel extends AbstractTableModel implements TableModel, TableColumnListener {
	
	private static final long serialVersionUID = 3387896763789247131L;

	private List<AcopTableColumn> columns = new ArrayList<>();
	
	private int rowCount = 10;
	
	/**
	 * Constructs new AcopTableModel
	 *
	 */
	public AcopTableModel() {
		super();
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.table.TableModel#getColumnClass(int)
	 */
	@Override
	public Class<?> getColumnClass(int columnIndex) {
		return columns.get(columnIndex).getColumnClass();
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.table.TableModel#getColumnCount()
	 */
	public int getColumnCount() {
		return columns.size();
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.table.TableModel#getColumnName(int)
	 */
	@Override
	public String getColumnName(int columnIndex) {
		return columns.get(columnIndex).getHeaderName();
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.table.TableModel#getRowCount()
	 */
	public int getRowCount() {
		return rowCount;
	}
	
	/**
	 * Sets the number of rows presented in the table.
	 * 
	 * @param rowCount
	 */
	public void setRowCount(int rowCount) {
		this.rowCount = rowCount;
		for (AcopTableColumn c : columns) {
			c.setRowCount(rowCount);
		}
		fireTableStructureChanged();
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.table.TableModel#getValueAt(int, int)
	 */
	public Object getValueAt(int rowIndex, int columnIndex) {
		if (columnIndex >= getColumnCount()) 
			throw new IllegalArgumentException("Column index should not exceed column count: " + columnIndex + " >= " + getColumnCount() + ".");
		
		return columns.get(columnIndex).getValue(rowIndex);
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.table.TableModel#isCellEditable(int, int)
	 */
	@Override
	public boolean isCellEditable(int rowIndex, int columnIndex) {
		return false;
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.table.TableModel#setValueAt(java.lang.Object, int, int)
	 */
	@Override
	public void setValueAt(Object aValue, int rowIndex, int columnIndex) {
		//read only
	}
	
	/**
	 * Adds a new column to the model. Column is added as the most right one.
	 * 
	 * @param column new column to be added
	 */
	public void addColumn(AcopTableColumn column) {
		columns.add(column);
		column.addTableColumnListener(this);
		column.setRowCount(rowCount);
		fireTableStructureChanged();
	}
	
	/**
	 * Removes a column from the model.
	 * 
	 * @param column the column to be removed
	 */
	public void removeColumn(AcopTableColumn column) {
		columns.remove(column);
		column.removeTableColumnListener(this);
		fireTableStructureChanged();
	}

	/*
	 * (non-Javadoc)
	 * @see de.desy.acop.displayers.tools.table.TableColumnListener#columnDataChanged(de.desy.acop.displayers.tools.table.TableColumnEvent)
	 */
	public void columnDataChanged(TableColumnEvent e) {
		AcopTableColumn column = e.getColumn();
		int columnIndex = -1;
		for (int i = 0; i < columns.size(); i++) {
			if (column.equals(columns.get(i))) {
				columnIndex = i;
			}
		}
		
		if (columnIndex != -1){
			TableModelEvent event = new TableModelEvent(this, 0, rowCount, columnIndex);
			fireTableChanged(event);
		}
		
	}	
}
