/**
 * File: ConnectionModel.java
 * Date: Jul 23, 2010
 */
package de.desy.acop.displayers.tarantula;

import java.util.Collection;
import java.util.Comparator;
import java.util.logging.Logger;

import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreeNode;

/**
 * Connection model. Includes all nodes in a hierarchical ordered structure 
 * which effects the fast binary search of a child node from its parent.
 * 
 * @author <a href="mailto:marcus.walla(at)desy.de">walla</a>
 * @version $Id: ConnectionModel.java 40186 2010-11-03 14:12:06Z walla $
 */
final class ConnectionModel extends DefaultTreeModel {
	
    // --- CLASS FIELDS --------------------------------------------------------
	
	@SuppressWarnings("unused")
	private static final Logger log = Utility.getLogger(ConnectionModel.class);
	private static final long serialVersionUID = -6802927497819817630L;
	static final String ROOT_NODE = "ROOT";
	
	// --- INSTANCE FIELDS -----------------------------------------------------
	
	private final Comparator<ANode> compareNodes = new Comparator<ANode>() {

		@Override
		public int compare(ANode node1, ANode node2) {
			if (node1 == null || node2 == null)
				return 0;
			
			final int diff = node1.getFullAddress().compareToIgnoreCase(node2.getFullAddress()); 
			return diff;
		}
	};
	private final Comparator<String> compareNodeAddress = new Comparator<String>() {

		@Override
		public int compare(String fullAddress1, String fullAddress2) {
			return fullAddress1.compareToIgnoreCase(fullAddress2);
		}
	};
	
	// --- CONSTRUCTOR ---------------------------------------------------------
	
	/**
	 * Constructor.
	 */
	ConnectionModel() {
		this(null);
	}

	/**
	 * Constructor.
	 * 
	 * @param status Initialize status of the root node.
	 */
	ConnectionModel(final EStatusRoot status) {
		super(null);
		if (status != null)
			super.setRoot(new NodeRoot(status));
		else
			super.setRoot(new NodeRoot(EStatusRoot.ROOT));
	}

	// --- INSTANCE METHODS ----------------------------------------------------

	/** This method doesn't effect the root node and hence does nothing. */
	@Override public void setRoot(final TreeNode node) {
	}
	
	@Override public NodeRoot getRoot() {
		return (NodeRoot) super.getRoot();
	}
	
	/**
	 * Inserts a new node according to the lexicographical order.
	 * @param insertNode New node
	 * @param parent Parent node, in which the new node should be inserted.
	 */
	void insertNode(final ANode insertNode, final ANode parent) {
		final int index = 
			this.searchNodeIndex(0, parent.getChildCount() - 1, insertNode, parent, this.compareNodes);
		super.insertNodeInto(insertNode, parent, index);
	}
	
	private int searchNodeIndex(final int start, final int end, final ANode node, final ANode parent, final Comparator<ANode> comparator) {
		if (start > end)
			return start;
		
		final int index = start + ((end - start) / 2);
		final int diff = comparator.compare(node, (ANode) parent.getChildAt(index));

		if (diff > 0)
			return this.searchNodeIndex(index + 1, end, node, parent, comparator);
		else if (diff < 0)
			return this.searchNodeIndex(start, index - 1, node, parent, comparator);
		
		return index;
	}

	/**
	 * Binary search (costs {@code O(log n)} for one of the parents child nodes 
	 * by comparing the delivered node. If the node doesn't exists a 
	 * {@code null} is returned otherwise the node. Moreover this method can
	 * also return a {@code null} if one of the arguments is {@code null}.
	 * 
	 * @param node Is the node to search for. This can also be {@code null} in
	 * 		which case this method returns a {@code null}.
	 * @param parent Parent node.
	 * @return Returns the node. This can also be {@code null} if no node is
	 * 		found.
	 */
	NodeConnection findNode(final ANode node, final ANode parent) {
		if (node == null || parent == null)
			return null;
		return findNode(node.getFullAddress(), parent);
	}
	
	/**
	 * Binary search (costs {@code O(log n)} for one of the parents child nodes 
	 * by comparing the full connection address. If the node doesn't exists a 
	 * {@code null} is returned otherwise the node.
	 * 
	 * @param fullAddress Full connection address. Depends on the special 
	 * 		connection framework, for example if its a TINE framework or DOOCS.
	 * 		This address is compare by ignoring the case of the characters.
	 * @param parent Parent node.
	 * @return Returns the node. This can also be {@code null} if no node is
	 * 		found.
	 */
	NodeConnection findNode(final String fullAddress, final ANode parent) {
		final int index = searchFullAddressIndex(0, parent.getChildCount(), fullAddress, parent, this.compareNodeAddress);
		if (index == -1)
			return null;
		final TreeNode node = parent.getChildAt(index);
		if (node instanceof NodeConnection)
			return (NodeConnection) parent.getChildAt(index);
		else
			return null;
	}
	
	/**
	 * Searches all child elements of the parent for the specified full address.
	 * If none can be found then a {@code -1} is returned otherwise the current
	 * index of the found node. The search mechanism is realized with a binary 
	 * search algorithm, i. e. costs {@code O(log n)}.
	 * 
	 * @param start Start search index. On the first call this will be {@code 0}.
	 * @param end End search index. On the first call this will be the length of 
	 * 		the child nodes minus 1. 
	 * @param searchFullAddress Look up address.
	 * @param parent Parent node.
	 * @param comparator Comparator instance, which compare the current full 
	 * 		addresses returned from the node.
	 * @return Index of the found node or {@code -1} if non is found.
	 */
	private int searchFullAddressIndex(
			final int start, 
			final int end, 
			final String searchFullAddress, 
			final ANode parent, 
			final Comparator<String> comparator) {
		final int index = start + ((end - start) / 2);
		
		if (start > end || index >= parent.getChildCount())
			return -1;
		
		final int diff = comparator.compare(searchFullAddress, ((ANode) parent.getChildAt(index)).getFullAddress());

		if (diff > 0)
			return this.searchFullAddressIndex(index + 1, end, searchFullAddress, parent, comparator);
		else if (diff < 0)
			return this.searchFullAddressIndex(start, index - 1, searchFullAddress, parent, comparator);
		
		return index;
	}

	/**
	 * Updates the current model of the tree with the new created one.
	 * @param model New created model.
	 */
	void update(final ConnectionModel model) {
		update(getRoot(), model.getRoot().getChildNodes());
	}
	
	private void update(final ANode parent, final Collection<NodeConnection> nodes) {
		for(final NodeConnection node: nodes) {
			final int index = searchFullAddressIndex(0, parent.getChildCount(), node.getFullAddress(), parent, this.compareNodeAddress);
			final NodeConnection currentNode;
			if (index == -1) {
				currentNode = node;
				this.insertNode(node, parent);
			}
			else {
				currentNode = (NodeConnection) parent.getChildAt(index);
				if (currentNode.getStatusLinkCode() != node.getStatusLinkCode() ||
						currentNode.getCondition() != node.getCondition()) {
					currentNode.setStatusLinkCode(node.getStatusLinkCode());
					currentNode.setCondition(node.getCondition());
					this.nodeChanged(currentNode);
				}
			}
			update(currentNode, node.getChildNodes());
		}
	}
}
/* end of file */