/**
 * File: RefreshConnections.java
 * Date: Sep 1, 2010
 */
package de.desy.acop.displayers.tarantula;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.swing.JOptionPane;
import javax.swing.JTree;
import javax.swing.SwingUtilities;
import javax.swing.tree.TreeNode;
import javax.swing.tree.TreePath;

import de.desy.tine.client.TLink;
import de.desy.tine.client.TLinkFactory;
import de.desy.tine.definitions.TErrorList;
import de.desy.tine.server.connections.TConnectionStruct;

/**
 * <p>This is the main and heavily used class for refreshing the tree model 
 * structure including the whole connection chain from the client up to
 * all servers. The depth of this structure can be set via the method 
 * {@link #setStructureDepth(int)}.</p>
 * <p>Also remember that the permanently call of the updating mechanism is done
 * outside of this class in the calling instance. This class only refreshes
 * the connection by calling the {@link #update()} method.</p>
 * 
 * @author <a href="mailto:marcus.walla(at)desy.de">walla</a>
 * @version $Id: RefreshConnections.java 63096 2014-11-14 16:36:43Z bobnar $
 */
final class RefreshConnections {
	// --- CLASS FIELDS --------------------------------------------------------
	
	private final static Logger log = 
		Utility.getLogger(RefreshConnections.class);
	public final static int MIN_TREE_DEPTH = 3;
	public final static int MAX_TREE_DEPTH = MIN_TREE_DEPTH + 7;
	public final static int INIT_LOOP_TIME = 60;
	public final static int MIN_LOOP_TIME = 30;
	public final static int MAX_LOOP_TIME = MIN_LOOP_TIME + 210;

	private enum ERefreshResult {
		FAILURE_CONNECTION,
		SUCCESS,
		SUCCESS_BUT_STOP_NODE,
		SUCCESS_BUT_DEPTH_EXCEEDED,
		SUCCESS_BUT_TIME_EXCEEDED,
		SUCCESS_BUT_CYCLE_LOOP,
		SUCCESS_BUT_IGNORE_SERVER
	}
	
	// --- INSTANCE FIELDS -----------------------------------------------------

	private final TarantulaPanel tarantulaPanel;
	private int treeDepth = RefreshConnections.MIN_TREE_DEPTH;
	private int maxLoopTime = RefreshConnections.MIN_LOOP_TIME;
	private int internalLoopTime = 0;

	// --- CONSTRUCTOR ---------------------------------------------------------
	
	/**
	 * Refreshes the connections. Remember that a permanently updating mechanism
	 * is done in the calling instance and not here. 
	 * 
	 * @param tarantulaPanel
	 * @param treeDepth
	 */
	RefreshConnections(final TarantulaPanel tarantulaPanel, final int treeDepth, final int maxLoopTime) {
		this.tarantulaPanel = tarantulaPanel;
		this.setStructureDepth(treeDepth);
		this.setMaxLoopTime(RefreshConnections.MIN_LOOP_TIME);
		this.setMaxLoopTime(maxLoopTime);
	}
	
	// --- INSTANCE METHODS ----------------------------------------------------
	
	/**
	 * Refreshes the hierarchical tree structure of the client depending on 
	 * its descending links in the chain.
	 * 
	 * @param model Connection model. Should never be {@code null}.
	 */
	void update(final ConnectionModel model) {
		update(model, null, getStructureDepth());
	}
	
	/**
	 * Refreshes the hierarchical tree structure of the client depending on 
	 * its descending links in the connection chain.
	 * 
	 * @param model Connection model. Should never be {@code null}.
	 * @param node Starting node. This can also be {@code null}. In this case 
	 * 		all nodes are refreshed starting from the root node.
	 * @param resetStructureDepth Is the current structure tree depth from the
	 * 		calling class. The structure depth value can be independent from the 
	 * 		depth of method {@link #getStructureDepth()}, which is the depth for
	 * 		the hole tree instance instead of a single brunch. For example, the 
	 * 		user can select a different structure depth via a context menu. This
	 * 		allows the user to interact with the open dialog box and enables him
	 * 		to set an independent structure depth. Internally this resets the
	 * 		structure depth via the method of {@link #setStructureDepth(int)} with
	 * 		this parameter.
	 * @param filter Refreshes only those nodes which are included on the filter 
	 * 		element.
	 */
	void update(final ConnectionModel model, final NodeConnection node, final int resetStructureDepth) {
		if (model == null)
			throw new NullPointerException("Connection model is null.");
		
		this.getTarantulaPanel().update(TarantulaPanel.ERefresh.RUNNING);
		refreshRoot(EStatusRoot.ROOT_REFRESH_CONNECTIONS);

		model.getRoot().setCondition(ENodeCondition.ORDINARY_NODE);

		try {
			if (node == null)
				updateFinished(refresh(model));
			else {
				switch(refresh(node, model)) {
				case SUCCESS_BUT_TIME_EXCEEDED: {
					ANode currentNode = node;
					while(currentNode != null && currentNode instanceof NodeConnection) {
						if (currentNode.getCondition() != ENodeCondition.TIME_EXCEEDED)
							currentNode.setCondition(ENodeCondition.TIME_EXCEEDED);
						currentNode = currentNode.getParent();
					}
					model.getRoot().setCondition(ENodeCondition.TIME_EXCEEDED);
					break;
				}
				case SUCCESS_BUT_DEPTH_EXCEEDED: {
					node.setCondition(ENodeCondition.DEPTH_EXCEEDED);
					break;
				}
				default:
					node.setCondition(ENodeCondition.ORDINARY_NODE);
				}
				model.nodeStructureChanged(node);
				updateFinished(model);
			}
		}
		catch(Throwable exception) { log.log(Level.SEVERE, "Unexpected exception is thrown", exception); }
		setStructureDepth(resetStructureDepth);
		
		if (this.getTarantulaPanel().getConnectionModel(TarantulaPanel.EFilter.ALL_CONNECTIONS).getRoot().getStatus() == EStatusRoot.ROOT_TIME_EXCEEDED)
			refreshRoot(EStatusRoot.ROOT_TIME_EXCEEDED);
		else
			refreshRoot(EStatusRoot.ROOT_REFRESH_FINISHED);
		this.getTarantulaPanel().update(TarantulaPanel.ERefresh.FINISHED);
	}

	private void updateFinished(final ConnectionModel newModel) {
		
		final TarantulaPanel tarantula = RefreshConnections.this.getTarantulaPanel();
		final JTree tree = tarantula.getTree();
		// needed to update the current model in the EDT thread
		//final ConnectionModel model = RefreshConnections.this.getTarantulaPanel().getConnectionModel();
		
		SwingUtilities.invokeLater(new Runnable() {

			// added @SuppressWarnings("unused") because in the run method 
			// I request a constant field which is set at compile time an 
			// therefore cannot be changed
			@Override public void run() {

				// ATTENTION: possibly the row height of the tree can be ignored, 
				// but is here included for saftey's sake
				tree.setRowHeight(16);

				// FIXME should be activated in contrast to the upper statements: 
				// model.update(newModel);
				
				if (newModel.getRoot().getCondition() == ENodeCondition.TIME_EXCEEDED) {
					JOptionPane.showMessageDialog(
							tarantula, 
							"<html><center>" +
							"The maximal loop time of " 
							+ RefreshConnections.this.getMaxLoopTime() + " seconds" 
							+ " is exceeded!<br>" +
							"Hence the connection refresh loop is explicitly stopped.<br><br>" +
							"Try to set a higher periode under the label '" + TarantulaPanel.STR_MAX_LOOP_TIME+ "'." +
							"</center></html>", 
							"Maximal Loop Time is Exceeded...", 
							JOptionPane.WARNING_MESSAGE);
				}
			}
		});
	}
	
	/**
	 * Returns the structure depth.
	 * @return Structure depth.
	 */
	int getStructureDepth() {
		return this.treeDepth;
	}
	
	/**
	 * Sets the tree depth. Only values between {@link #MIN_TREE_DEPTH} and
	 * {@link #MAX_TREE_DEPTH} are applicable.
	 * @param treeDepth Tree depth.
	 */
	void setStructureDepth(final int treeDepth) {
		this.treeDepth = Math.min(MAX_TREE_DEPTH, Math.max(MIN_TREE_DEPTH, treeDepth));
	}
	
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	private TarantulaPanel getTarantulaPanel() { return this.tarantulaPanel; }
	
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	
	private ConnectionModel refresh(final ConnectionModel model) {
    	
		// refresh the active links of this application
		final TLink[] lnks = TLinkFactory.getInstance().getActiveLinks();

		// collection of nodes which represent the current used connections. 
		// This list is used to remove deprecated links from the model.
		final List<NodeConnection> currentNodes = new ArrayList<>();

		// root is not a leaf and links are null or number of links is 0
		if (lnks == null || lnks.length == 0) {
			// remove all children descending from the root element
			for(final ANode node: model.getRoot().getChildNodes())
				model.removeNodeFromParent(node);
	    	
			return model;
		}
		
		// adapt nodes to root and add new ones if necessary
		for(final TLink lnk: lnks) {

			
			// ATTENTION: discussed with Phil if stock properties should be
			// included in the model. We came to the result that they should.
			// But for safety the following lines is only commented out to 
			// recall it if this wasn't such a good idea. 
			//
			// reflect if it is a stock property
			//if (TQuery.isStockProperty(lnk.getProperty()))
			//	continue;
			
			// reflects if the link element from the array is null -> continue the loop
			if (lnk == null)
				continue;
			
			// reflect if it is a special connection -> continue the loop
			if (TineConnection.isSpecialConnection(lnk))
				continue;
			
			/* linear search! But its OK because its only requested in the first 
			 * level of the tree structure! */
			final NodeConnection tmpNode = 
				findNode(model.getRoot(), TineConnection.getFullAddress(lnk)); 
			final NodeConnection node;
			
			if (tmpNode != null) {
				node = tmpNode;
				final EStatusLink currentStatus = 
					TineConnection.convertToStatus(lnk.getLinkStatus());

				currentNodes.add(node);
				
				// has the current node a failure state ...
				if (TineConnection.isNodeFailure(currentStatus)) {
					final TreePath path = new TreePath(node.getPath());
					if (getTarantulaPanel().getTree().isExpanded(path)) {
						// FIXME: bad solution - redrawing of the UI
						// component is not realized in the EventDispatchThread
						getTarantulaPanel().getTree().collapsePath(path);
					}
					node.setStatusLinkCode(lnk.getLinkStatus());
					node.removeAllChildren();
					// model structure changed
					model.nodeStructureChanged(node);
					
					continue;
				}

				if (node.getStatus() != currentStatus) {
					node.setStatusLinkCode(lnk.getLinkStatus());
					// node has been modified
					model.nodeChanged(node);
				}
			}
			else {
				
				// create a new node ...
				node = new NodeConnection(lnk);
				node.setStatusLinkCode(lnk.getLinkStatus());
				
				// insert the new created node at the children of root
				model.insertNode(node, model.getRoot());
				
				currentNodes.add(node);
				
				// has the current node a failure state ...
				if (TineConnection.isNodeFailure(node))
					continue;
			}

		}
		
		resetNodes(model, model.getRoot(), currentNodes);

		for(final NodeConnection node: currentNodes) {
			// FETCH NEXT NODE ...
			final ERefreshResult result = switchNode(node, refresh(node, model));
			if (result == ERefreshResult.SUCCESS_BUT_TIME_EXCEEDED) {
				// if time is exceeded then exit immediately
				model.getRoot().setStatus(EStatusRoot.ROOT_TIME_EXCEEDED);
				model.getRoot().setCondition(ENodeCondition.TIME_EXCEEDED);
				model.nodeChanged(model.getRoot());
				debugAddresses(node, "-| => success, time exceeded");
			}
		}
		
		expandRootPath(model.getRoot());		
		
		return model;
	}
	
	private ERefreshResult refresh(
			final NodeConnection parent, 
			final ConnectionModel model) {
		if (TineConnection.ignoreServer(parent.getServer().toUpperCase())) {
			return ERefreshResult.SUCCESS_BUT_IGNORE_SERVER;
		}
		
		// FIXME Is it right here to reset the parent status to SUCCESS?
		if (parent.getStatus() == EStatusLink.ASCENDING_WARNING)
			parent.setStatusLinkCode(TErrorList.success);
		
		// =====================================================================

		final Collection<TConnectionStruct> connections;
		try {
			connections = TineConnection.getConnections(
						parent.getContext(), 
						parent.getServer());
		}
		catch(TineException exception) {
			parent.setStatusLinkCode(exception.getErrorCode());
			return ERefreshResult.FAILURE_CONNECTION;
		}
		
		// if the parent hasn't any links then stop here and return STOP NODE!
		if (connections.size() == 0)
			return ERefreshResult.SUCCESS_BUT_STOP_NODE;
		
		// node has a status of SUCCESS but reflects a cycle connection
		if (parent.getStatus() == EStatusLink.SUCCESS && isCycleConnection(parent))
			return ERefreshResult.SUCCESS_BUT_CYCLE_LOOP;
		
		// request if tree level hasn't exceeded (tree level is a property 
		// of a node which reveals the current depth from root to this node
		// in the model).
		if (isDepthExceeded(parent, model)) {
			return ERefreshResult.SUCCESS_BUT_DEPTH_EXCEEDED;
		}
		
		for(final TConnectionStruct connection: connections) {

			if (isTimeExceeded())
				return ERefreshResult.SUCCESS_BUT_TIME_EXCEEDED;
			
			// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

			final NodeConnection node = createNode(connection, parent, model);
			
			if (TineConnection.isNodeFailure(node)) {
				setNodeFailure(node, model);
				model.nodeChanged(node);
			}
			else if (switchNode(node, refresh(node, model)) == ERefreshResult.SUCCESS_BUT_TIME_EXCEEDED) {
				return ERefreshResult.SUCCESS_BUT_TIME_EXCEEDED;
			}
		}

		return ERefreshResult.SUCCESS;
	}
	
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	private void resetNodes(final ConnectionModel model, final ANode node, final List<NodeConnection> nodes) {
		for(final NodeConnection child: node.getChildNodes()) {
			if (!nodes.contains(child)) {
				node.remove(child);
			}
		}
		model.nodeStructureChanged(node);
	}
	
	private ERefreshResult switchNode(
			final NodeConnection node, 
			final ERefreshResult result) {
		/* ATTENTION/FIXME: the following query for fetching all nodes lowers 
		 * the performance immediately! */
		switch(result) {
		case SUCCESS:
			// -> set as a normal node without any special conditions
			node.setCondition(ENodeCondition.ORDINARY_NODE);
			debugAddresses(node, "-OK => success");
			break;
			
		case SUCCESS_BUT_CYCLE_LOOP:
			// -> set the node condition to CYCLE LOOP NODE
			node.setCondition(ENodeCondition.CYCLE_LOOP_NODE);
			debugAddresses(node, "-O => succes but cycle loop (depth=" + node.getLevel() + ')');
			break;
		
		case SUCCESS_BUT_DEPTH_EXCEEDED:
			node.setCondition(ENodeCondition.DEPTH_EXCEEDED);
			debugAddresses(node, "-? => success but depth exceeded");
			break;
		
		case SUCCESS_BUT_IGNORE_SERVER:
			node.setCondition(ENodeCondition.IGNORE_SERVER);
			debugAddresses(node, "-[!] => success but ignore server");
			break;
		
		case SUCCESS_BUT_STOP_NODE:
			node.setCondition(ENodeCondition.STOP_NODE);
			debugAddresses(node, "-| => success, leaf");
			break;
		
		case SUCCESS_BUT_TIME_EXCEEDED:
			node.setCondition(ENodeCondition.TIME_EXCEEDED);
			debugAddresses(node, "-| => success, time exceeded");
			return ERefreshResult.SUCCESS_BUT_TIME_EXCEEDED;
		
		case FAILURE_CONNECTION:
			node.setCondition(ENodeCondition.FAILURE_CONNECTION);
			debugAddresses(node, "-[!] => failure:" + node.getStatusLink());
			break;
			
		default:					
			// -> set the node condition to INTERNAL FAILURE
			node.setCondition(ENodeCondition.FAILURE_INTERNAL);
			debugAddresses(node, "-[!] => internal failure");
		}

		return ERefreshResult.SUCCESS;
	}
	
	private boolean isDepthExceeded(final NodeConnection node, final ConnectionModel model) {
		if (node.getLevel() >= this.treeDepth) {
			
			// has the parent node any children?
			if (node.getChildCount() != 0) {
				// Possibly a bug in the JTree class: I first have to
				// collapse all child nodes before removing them. If I do so 
				// with the expanded children they will be removed from the 
				// model but not from the UI tree. They are included in the model
				// without any suggestion to be erased instead of removing
				// them indirectly from the tree. I think this is not a big
				// deal to be solved by Sun!
				final TreePath path = new TreePath(node.getPath());
				if (this.getTarantulaPanel().getTree().isExpanded(path)) {
					final JTree tree = this.getTarantulaPanel().getTree();
					SwingUtilities.invokeLater(new Runnable() {
						@Override public void run() {
							tree.collapsePath(path);
						}
					});
				}
				
				node.removeAllChildren();
				model.nodeStructureChanged(node);
			}
			
			try {
				if (TineConnection.getConnections(
						node.getContext(), 
						node.getServer()).size() != 0) {
					// FIXME there exists child nodes -> should trigger the 
					// calling method to include a special condition of the 
					// node
				}
			}
			catch(TineException exception) {
				// FIXME fetching all child nodes failed cause of a TINE link exception
				log.warning("unable to fetch information from the address " + node.getFullAddress() + " cause of a TINE exception: " + exception.getError());
			}
			
			return true;
		}
		
		return false;
	}
	
	private boolean isTimeExceeded() {
		// request if loop time has exceeded its maximum delay time (in 
		// seconds)
		final int currentTime = (int) (System.currentTimeMillis() / 1000L);
		final int diff = Math.max(0, currentTime - getInternalLoopTime());
		if (diff > RefreshConnections.this.getMaxLoopTime())
			return true;
		else
			return false;
	}

	/**
	 * If one of the preceding nodes contains the current node 
	 * address then a cycle connection is detected. In this case a {@code true}
	 * is returned otherwise a {@code false}.
	 *  
	 * @param node Node, can also be {@code null}!
	 * @param fullAddress - never null!
	 * @return Returns a {@code true} if the address has been mentioned in one
	 * 		of the preceding nodes.
	 */
	private boolean isCycleConnection(final ANode node) {
		if (node == null || node.isRoot())
			return false;
		return isCycleConnection(node.getParent(), node.getFullAddress());
	}
	
	private boolean isCycleConnection(final ANode node, final String fullAddress) {
		if (node == null || node.isRoot())
			return false;
		else if(fullAddress.equalsIgnoreCase(node.getFullAddress()))
			return true;
		else
			return isCycleConnection(node.getParent(), fullAddress);
	}
	
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	
	private NodeConnection findNode(final ANode parent, final String fullAddress) {
		for(final NodeConnection node: parent.getChildNodes())
			if (node.getFullAddress().equalsIgnoreCase(fullAddress))
				return node;
		return null;
	}

	private void expandRootPath(final NodeRoot root) {
		// Possibly a bug from the JTree class - its impossibly to reject the 
		// root node from being collapsed. So I have to request if it is 
		// collapsed and, when true, expand it. This is a foolish behavior of 
		// the root node or tree at all but I don't know which method to call 
		// on a JTree or model to reject this.
		// A discussion which reflects the mysterious behavior of the hole
		// Swing framework can be found under BugId 6459532 
		// (http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=6359532)
		final TreePath rootPath = new TreePath(root.getPath());
		if (this.getTarantulaPanel().getTree().isCollapsed(rootPath)) {
			// FIXME: bad solution - redrawing of the GUI
			// component is not realized in the EventDispatchThread
			this.getTarantulaPanel().getTree().expandPath(rootPath);
		}
	}
	
	private NodeConnection createNode(
			final TConnectionStruct connection, 
			final NodeConnection parent, 
			final ConnectionModel model) {
		final String fullAddress = TineConnection.getFullAddress(connection);
		final NodeConnection tmpNode = model.findNode(fullAddress, parent);
		final NodeConnection node;
		
		if (tmpNode == null) {
			node = new NodeConnection(
					connection.getContext(), 
					connection.getExpName(), 
					connection.getDevName(), 
					connection.getPrpName(), 
					TineConnection.getConnectionInfo(fullAddress));
			model.insertNode(node, parent);
		}
		else
			node = tmpNode;
		
		node.setStatusLinkCode(connection.getLinkStatus());

		return node;
	}

	private void setNodeFailure(final NodeConnection node, final ConnectionModel model) {
		if (node.getParent() instanceof NodeConnection)
			this.setAscendingWarningStatus((NodeConnection) node.getParent());

		// FIXME possibly be removed?
		this.expandNode(node);
		model.nodeStructureChanged(node);

		node.removeAllChildren();
		model.nodeStructureChanged(node);
	}
	
	private void expandNode(final ANode node) {
		final TreeNode[] nodes = node.getPath();
		for (int i = 1; i < nodes.length; i++) {
			final TreeNode[] current = new TreeNode[i];
			for (int j = 0; j < current.length; j++) current[j] = nodes[j];
//			SwingUtilities.invokeLater(new Runnable() {
//				@Override public void run() {
					RefreshConnections.this.getTarantulaPanel().getTree().expandPath(new TreePath(current));
//				}
//			});
		}
	}

	private void setAscendingWarningStatus(final NodeConnection node) {
		if (node.getStatus() != EStatusLink.ASCENDING_WARNING) {
			node.setAscendingWarning();
			this.getTarantulaPanel().getConnectionModel(TarantulaPanel.EFilter.ALL_CONNECTIONS).nodeChanged(node);
		}
		if (node.getParent() instanceof NodeConnection)
			setAscendingWarningStatus((NodeConnection) node.getParent());
	}

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	
	private void refreshRoot(final EStatusRoot refreshTree) {
		final TarantulaPanel tarantula = this.getTarantulaPanel();
		SwingUtilities.invokeLater(new Runnable() {
			@Override public void run() {
				final NodeRoot root = tarantula.getConnectionModel(TarantulaPanel.EFilter.ALL_CONNECTIONS).getRoot();
				root.refreshTree(refreshTree);
				tarantula.getConnectionModel(TarantulaPanel.EFilter.ALL_CONNECTIONS).nodeChanged(root);
				if (refreshTree == EStatusRoot.ROOT_REFRESH_CONNECTIONS)
					tarantula.getTree().setEnabled(false);
				else
					tarantula.getTree().setEnabled(true);
			}
		});
		
		// start refreshing
		if (refreshTree == EStatusRoot.ROOT_REFRESH_CONNECTIONS) {
			this.setInternalLoopTime((int) (System.currentTimeMillis() / 1000L));
		}
	}
	
	private void setInternalLoopTime(final int internalLoopTime) { this.internalLoopTime  = internalLoopTime; }
	private int getInternalLoopTime() { return this.internalLoopTime; }
	
	/**
	 * <p>During executing the refresh method of {@link #update(ConnectionModel)}
	 * or {@link #update(ConnectionModel, NodeConnection, int)} it can sometimes
	 * happen that the time exceeds the current maximum loop time. In this 
	 * situation the execution is immediately aborted and a dialog box is opened 
	 * to inform the user about the exceeded loop time. For this case the user
	 * can enlarge the loop time to higher value and retry. The time by this 
	 * method is released in seconds.</p>
	 * @param maxLoopTime Maximum loop time. Should be in the range from 
	 * 		{@link RefreshConnections#MIN_LOOP_TIME} to {@link RefreshConnections#MAX_LOOP_TIME}
	 */
	void setMaxLoopTime(final int maxLoopTime) { 
		this.maxLoopTime = Math.max(MIN_LOOP_TIME, Math.min(RefreshConnections.MAX_LOOP_TIME, maxLoopTime)); 
	}
	
	/**
	 * <p>Gets the maximum loop time. If the execution for refreshing all 
	 * connections exceeds the maximum loop time then the user can set a new
	 * one by a special UI component. See also the method {@link #setMaxLoopTime(int)} 
	 * for more information. The time is released in seconds.</p>
	 * @return Maximum loop time.
	 */
	int getMaxLoopTime() { return this.maxLoopTime; }
	
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	
	private static void debugAddresses(final ANode node, final String addresses) {
		if (node.getParent() == null) {
			log.fine(addresses);
			return;
		}
		debugAddresses(node.getParent(), node.getFullAddress() + " -> " + 
				addresses);
	}
}
/* end of file */