/*
 * Copyright (c) 2006 Stiftung Deutsches Elektronen-Synchroton,
 * Member of the Helmholtz Association, (DESY), HAMBURG, GERMANY.
 *
 * THIS SOFTWARE IS PROVIDED UNDER THIS LICENSE ON AN "../AS IS" BASIS.
 * WITHOUT WARRANTY OF ANY KIND, EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR PARTICULAR PURPOSE AND
 * NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE. SHOULD THE SOFTWARE PROVE DEFECTIVE
 * IN ANY RESPECT, THE USER ASSUMES THE COST OF ANY NECESSARY SERVICING, REPAIR OR
 * CORRECTION. THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF THIS LICENSE.
 * NO USE OF ANY SOFTWARE IS AUTHORIZED HEREUNDER EXCEPT UNDER THIS DISCLAIMER.
 * DESY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS,
 * OR MODIFICATIONS.
 * THE FULL LICENSE SPECIFYING FOR THE SOFTWARE THE REDISTRIBUTION, MODIFICATION,
 * USAGE AND OTHER RIGHTS AND OBLIGATIONS IS INCLUDED WITH THE DISTRIBUTION OF THIS
 * PROJECT IN THE FILE LICENSE.HTML. IF THE LICENSE IS NOT INCLUDED YOU MAY FIND A COPY
 * AT HTTP://WWW.DESY.DE/LEGAL/LICENSE.HTM
 */

package de.desy.acop.displayers.tools;

import java.awt.Color;
import java.beans.Beans;
import java.util.Map;

import com.cosylab.gui.displayers.CommonDisplayer;
import com.cosylab.gui.displayers.DataConsumer;
import com.cosylab.gui.displayers.DataState;
import com.cosylab.gui.displayers.Displayer;
import com.cosylab.gui.displayers.DisplayerUtilities;
import com.cosylab.gui.displayers.DoubleSeqConsumer;
import com.cosylab.util.CommonException;

import de.desy.acop.chart.AcopGraphStyleEnum;
import de.desy.acop.displayers.AcopChart;

/**
 * <code>AcopChartConsumer</code> is an implementation of <code>DoubleSeqConsumer</code>
 * which can be used in connection with <code>AcopChart</code>. AcopChartConsumer
 * receives updates from the <code>DataSource</code> and updates the chart accordingly.
 * One consumer can only handle connection to one specific channel.
 * 
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 * @see AcopChart
 */
public class AcopChartConsumer implements DoubleSeqConsumer {

	@SuppressWarnings("rawtypes")
	public static final Class[] SUPPORTED_CONSUMER_TYPES = {
		DoubleSeqConsumer.class
		};
	public static final String[] SUPPORTED_CHARACTERISTICS = {
		CommonDisplayer.C_GRAPH_MAX, CommonDisplayer.C_GRAPH_MIN
		};
	
	protected String name;
	protected AcopChart acop;

	protected double[] xData;
	protected double[] yData;
	protected int displayHandle = -1;
	protected int textHandle = -1;
	protected long lastTimestamp;
	protected Color color= Color.RED;
	protected String[] xLabels;
	protected Object disable;
	protected int arraySize;
	protected int maxNumber;
    
    private boolean isTrendChart = false;
    
    private double preferredYMax;
    private double preferredYMin;
    
    private AcopGraphParameters displayerParameters;
    	
	/**
	 * Constructs new AcopChartConsumer object with the given name.
	 * The consumer belongs to the specified chart.
	 * 
	 * @param acop parent chart
	 * @param name the name of the consumer
	 */
	public AcopChartConsumer(AcopChart acop, String name) {
		super();
		this.name=name;
		this.acop=acop;
		preferredYMax = acop.getYMax();
		preferredYMin = acop.getYMin();
		
		//color= new Color((int)(Math.random()*255),(int)(Math.random()*255),(int)(Math.random()*255));
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getDataConsumer(java.lang.Class)
	 */
	@SuppressWarnings({"rawtypes","unchecked"})
    public DataConsumer getDataConsumer(Class type) {
		if (type.isAssignableFrom(getClass())) {
			return this;
		}
		return null;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getDefaultDataConsumer()
	 */
	public DataConsumer getDefaultDataConsumer() {
		return this;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#updateDataState(com.cosylab.gui.displayers.DataState)
	 */
	public void updateDataState(DataState state) {
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#setCharacteristics(java.util.Map)
	 */
	@SuppressWarnings("rawtypes")
	public void setCharacteristics(Map characteristics) {
		
		if (acop == null) {
			return;
		}

		Object o1 = null;
		Object o2 = null;

		o1 = characteristics.get(CommonDisplayer.C_GRAPH_MAX);

		if (o1 == null) {
			o1 = characteristics.get(CommonDisplayer.C_MAXIMUM);
		}

		o2 = characteristics.get(CommonDisplayer.C_GRAPH_MIN);

		if (o2 == null) {
			o2 = characteristics.get(CommonDisplayer.C_MINIMUM);
		}

		if (o1 != null || o2 != null) {
			double max = o1 != null ? ((Number)o1).doubleValue()
				: acop.getYMax();
			double min = o2 != null ? ((Number)o2).doubleValue()
				: acop.getYMin();
			
			setPreferredYMax(max);
			setPreferredYMin(min);
			acop.updateYScale(this);
		}

		o1 = characteristics.get(CommonDisplayer.C_DISPLAY_NAME);
		
		String newName = (String) o1;
		if (newName != null && !newName.equals(acop.getName())) {
		    acop.setName(newName);
		    acop.setCaption(newName);
		}
		
		o1 = characteristics.get(Displayer.C_UNITS);
		if (o1 != null) {
			acop.setYAxisLabel((String)o1);
		}

		//avoid setting colors through characterstics to avoid different threads
		//setting the same property at the same time - consumer might end up with wrong color
		o1 = characteristics.get(Displayer.C_COLOR);

		if (o1 != null) {
			color = (Color) o1;
			acop.setForeground(color);
		}
			
		o1 = characteristics.get("xLabels");
		if (o1 != null && !isTrendChart) {
			setXLabels((String[])o1);
		}
		
		o1 = characteristics.get(CommonDisplayer.C_SEQUENCE_LENGTH);
		if ((arraySize == 0 || arraySize == -1) && o1 != null) {
			setArraySize((Integer)o1);
		}
		
		if (!Beans.isDesignTime()) {
			acop.updateXScale(this);
		}
		acop.repaint();

	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getName()
	 */
	public String getName() {
		return name;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getSupportedCharacteristics()
	 */
	public String[] getSupportedCharacteristics() {
		return DisplayerUtilities.combineCharacteristics(DisplayerUtilities.COMMON_NUMERIC_DISPLAYER_CHARACTERISTICS,
			    SUPPORTED_CHARACTERISTICS);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getSupportedConsumerTypes()
	 */
	@SuppressWarnings("unchecked")
	public Class<DataConsumer>[] getSupportedConsumerTypes() {
		return SUPPORTED_CONSUMER_TYPES;
	}
	
	private void updateChart() {
		if (acop==null) {
			return;
		}
		
		synchronized (acop) {

	        //System.out.println("disp id : " + displayHandle);
	//	    acop.clearScreen(-1,0); <- this will wipe out all display references
	        //Acop aChart = acop.getAcopChart();
	//      aChart.clearScreen(displayHandle,0); <- this removes the given reference (but Acop will reuse the handle, so we don't want to use this inside the AcopDataConsumer)
	//	    acop.draw(y,x,null,null,141,100);
	        if (displayHandle < 0)
	        { // display the initial data set ...
	          applyParameters();
		      displayHandle = acop.draw(yData,xData,disable,xLabels,arraySize,maxNumber);
	//        System.out.println("assigned display handle : " + displayHandle + "  color : " + color.toString());
	        }
	        else
	        {
	          if (!isTrendChart)
	          { // not a trend chart : just refresh the display ...
	            acop.refreshScreen(yData, displayHandle, arraySize, 0, xData);
	//          System.out.println("refresh display handle : " + displayHandle + "  color : " + color.toString());
	          }
	          else
	          { // append 1 value to the end ...  (ACOP should be in LIN_TIME or LOG_TIME graph mode)
	            double[] yValue = new double[1];
	            yValue[0] = yData[0];
	            double[] xValue = new double[1];
	            xValue[0] = lastTimestamp/1000.;
	            
//	            System.out.println("Append: ["+xValue[0]+", "+yValue[0]+"]");
	            acop.appendScreen(yData, displayHandle, xValue);            
	            
//	            double[][] data= dumpData();
//	            System.out.print("X: ");
//	            System.out.println(Arrays.toString(data[0]));
//	            System.out.print("Y: ");
//	            System.out.println(Arrays.toString(data[1]));
	          }
	        }
	//	    a.GetDataSize(d,null,-1);
	//	    acopLabel.setText(a.getStatus() + " " + a.GetDataSize(-1) + " elements " + a.getTimeStamp().toString());
		    //acop.setForeground(Color.BLUE);
	//	    a.draw(z,x,null,null,141,100);
		    //acop.draw(z);
		}
		
	}

	/**
	 * Returns all data from chart. First index is 0 for X data array and 1 for Y data array.
	 * @return all dat a from chart in X and Y array
	 */
	public double[][] dumpData() {
		
		double[][] array= new double[2][arraySize];
		
		acop.getDrawnData(array[0], array[1], arraySize, 0, displayHandle);
		
		return array;
	}
	
	/**
	 * Destroys this consumer. Removes from the chart any data that belongs
	 * to this consumer.
	 * 
	 */
	public void destroy() {
		acop.clearScreen(displayHandle);
		acop=null;
	}
	
	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DoubleSeqConsumer#updateValue(long, double[])
	 */
	public void updateValue(long timestamp, double[] value) throws CommonException {
		lastTimestamp=timestamp;
		yData=value;
		updateChart();
	}

	/**
	 * Returns the horizontal values.
	 * 
	 * @return x values 
	 */ 
	public double[] getXData() {
		return xData;
	}

	/**
	 * Sets the horizontal values.
	 * 
	 * @param xdata new horizontal values 
	 */
	public void setXData(double[] xdata) {
		this.xData = xdata;
	}

	/**
	 * Returns the vertical values.
	 * 
	 * @return y values
	 */
	public double[] getYData() {
		return yData;
	}

	/**
	 * Sets new vertical values.
	 * 
	 * @param data new y values
	 */
	public void setYData(double[] data) {
		yData = data;
	}

	/**
	 * Returns the parent chart of this consumer.
	 * 
	 * @return the acop
	 */
	public AcopChart getAcop() {
		return acop;
	}

	/**
	 * Returns the last timestamp when value was updated.
	 * 
	 * @return last timestamp
	 */
	public long getLastTimestamp() {
		return lastTimestamp;
	}

	/**
	 * Returns the color that shows data of this consumer in the chart.
	 * 
	 * @return the color
	 */
	public Color getColor() {
		return color;
	}

	/**
	 * Sets the color that shows data of this consumer in the chart. Color can 
	 * only be set before the data was first drawn.
	 * 
	 * @param color new color
	 */
	public void setColor(Color color) {
		this.color = color;
	}

	/**
	 * Returns the labels of horizontal values.
	 * 
	 * @return x labels
	 */
	public String[] getXLabels() {
		return xLabels;
	}

	/**
	 * Sets the labels of horizontal values. Labels can only be set before data
	 * was first drawn.
	 * 
	 * @param labels new labels
	 */
	public void setXLabels(String[] labels) {
		xLabels = labels;
	}

	/**
	 * Returns the size of the array that this consumer shows.
	 * 
	 * @return the array size
	 */
	public int getArraySize() {
		return arraySize;
	}

	/**
	 * Sets the size of the array that this consumer handles with. This property can only
	 * be set before data was first drawn.
	 * 
	 * @param arraySize new array size
	 */
	public void setArraySize(int arraySize) {
		this.arraySize = arraySize;
	}

//	/**
//	 * @return Returns the disable.
//	 */
//	public Object getDisable() {
//		return disable;
//	}
//
//	/**
//	 * @param disable The disable to set.
//	 */
//	public void setDisable(Object disable) {
//		this.disable = disable;
//	}
//
//	/**
//	 * @return Returns the maxNumber.
//	 */
//	public int getMaxNumber() {
//		return maxNumber;
//	}
//
//	/**
//	 * @param maxNumber The maxNumber to set.
//	 */
//	public void setMaxNumber(int maxNumber) {
//		this.maxNumber = maxNumber;
//	}

	/**
	 * Returns the preferred maximum vertical value for this graph.
	 * 
     * @return the preferred max value
     */
    public double getPreferredYMax() {
    	return preferredYMax;
    }

	/**
	 * Sets the preferred maximum vertical value for this graph.
	 * 
     * @param preferredYMax new preferred maximum value 
     */
    public void setPreferredYMax(double preferredYMax) {
    	this.preferredYMax = preferredYMax;
    }

	/**
	 * Returns the preferred minimum vertical value for this graph.
     * @return the preferred min value
     */
    public double getPreferredYMin() {
    	return preferredYMin;
    }

	/**
	 * Sets the preferred minimum vertical value for this graph.
     * @param preferredYMin new preferred minimum value
     */
    public void setPreferredYMin(double preferredYMin) {
    	this.preferredYMin = preferredYMin;
    }
    
    private void applyParameters() {
    	if (displayerParameters != null) {
	    	acop.setForeground(displayerParameters.getColor());
	    	acop.setDrawStyle(displayerParameters.getDrawStyle());
	    	acop.setFFT(displayerParameters.getFFT());
	    	acop.setDisplayMode(displayerParameters.getMode());
	    	acop.setDrawWidth(displayerParameters.getWidth());
	    	if (displayerParameters.isTrend()) {
	    		acop.setGraphStyle(AcopGraphStyleEnum.TimeLin.ordinal());
	    		double[] data = new double[displayerParameters.getTrendLength()];
	    		if (yData != null) {
		    		for (int i = 0; i < yData.length && i < data.length; i++) {
		    			data[i] = yData[i];
		    		}
	    		}
	    		yData = data;
	    		maxNumber = data.length;
	    	}
    	} else if (color != null) {
    		acop.setForeground(color);
    	}
    }

    /**
     * Returns the AcopGraphParameters associated with this consumer.
     * 
     * @return associated parameters
     */
	public AcopGraphParameters getDisplayerParameters() {
		return displayerParameters;
	}

	/**
	 * Sets the AcopGraphParameters which describes the visual appearance of
	 * the data in the chart. DisplayerParameters can only be set before
	 * data was first drawn.
	 * 
	 * @param displayerParameters the displayer parameters
	 */
	public void setDisplayerParameters(AcopGraphParameters displayerParameters) {
		this.displayerParameters = displayerParameters;
		this.color = displayerParameters.getColor() != null ? displayerParameters.getColor() : this.color;
		setTrendChart(displayerParameters.isTrend());
		if (isTrendChart()) {
			setArraySize(displayerParameters.getTrendLength());
		} else {
			setArraySize(displayerParameters.getConnectionParameters().getPropertySize());
		}
	}

	/**
	 * Returns true if this consumer shows a trend graph.
	 * 
	 * @return true if this is a trend graph data
	 */
	public boolean isTrendChart() {
		return isTrendChart;
	}

	/**
	 * Sets the trend mode for this consumer. This property can only be set before 
	 * the first data was drawn.
	 * 
	 * @param isTrendChart trend mode flag
	 */
	public void setTrendChart(boolean isTrendChart) {
		this.isTrendChart = isTrendChart;
	}

}
