/*
 * Copyright (c) 2006 Stiftung Deutsches Elektronen-Synchroton,
 * Member of the Helmholtz Association, (DESY), HAMBURG, GERMANY.
 *
 * THIS SOFTWARE IS PROVIDED UNDER THIS LICENSE ON AN "../AS IS" BASIS.
 * WITHOUT WARRANTY OF ANY KIND, EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR PARTICULAR PURPOSE AND
 * NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE. SHOULD THE SOFTWARE PROVE DEFECTIVE
 * IN ANY RESPECT, THE USER ASSUMES THE COST OF ANY NECESSARY SERVICING, REPAIR OR
 * CORRECTION. THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF THIS LICENSE.
 * NO USE OF ANY SOFTWARE IS AUTHORIZED HEREUNDER EXCEPT UNDER THIS DISCLAIMER.
 * DESY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS,
 * OR MODIFICATIONS.
 * THE FULL LICENSE SPECIFYING FOR THE SOFTWARE THE REDISTRIBUTION, MODIFICATION,
 * USAGE AND OTHER RIGHTS AND OBLIGATIONS IS INCLUDED WITH THE DISTRIBUTION OF THIS
 * PROJECT IN THE FILE LICENSE.HTML. IF THE LICENSE IS NOT INCLUDED YOU MAY FIND A COPY
 * AT HTTP://WWW.DESY.DE/LEGAL/LICENSE.HTM
 */

package de.desy.acop.displayers.tools;

import java.awt.Color;

import com.cosylab.gui.adapters.Converter;

import de.desy.acop.chart.Acop;
import de.desy.acop.chart.AcopDisplayMode;
import de.desy.acop.chart.AcopDrawStyleEnum;
import de.desy.acop.transport.ConnectionParameters;

/**
 * <code>AcopGraphParameters</code> is an extension of 
 * <code>AcopDisplayerParameters</code> which also hold data about 
 * visualization of a single graph in a chart. This parameters are intended
 * for use with <code>AcopChart</code> and only hold information that can be 
 * read by the AcopChart object.
 * 
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 *
 */
public class AcopGraphParameters extends AcopDisplayerParameters {

	private static final long serialVersionUID = -7187249526651411944L;

	/**
	 * Constructs a new AcopGraphParameters object.
	 *
	 */
	public AcopGraphParameters() {
		this((ConnectionParameters)null, null, -1, -1, -1, 1, false, 100, null);
	}
	
	/**
	 * Constructs a new AcopGraphParameters object with the given connection point.
	 * 
	 * @param param remote connection point 
	 */
	public AcopGraphParameters(ConnectionParameters param) {
		this (param, null, -1, -1, -1,1, false, 100, null);
	}
	
	/**
	 * Constructs a new AcopGraphParameters object with the specified parameters values
	 * 
	 * @param parameters remote connection point
	 * @param color color of the graph
	 * @param fft acop fft mode
	 * @param drawStyle style of the graph
	 * @param mode acop graph mode 
	 * @param width the width of the line
	 * @param trend a flag indicating if this is a trend graph
	 * @param trendLength number of trend points if this is a trend
	 * @see Acop
	 */
	public AcopGraphParameters(ConnectionParameters parameters, Color color, int fft, int drawStyle, int mode, int width, boolean trend, int trendLength) {
		this(parameters,-1,Double.NaN,Double.NaN,null,null,color, fft, drawStyle, mode, width, trend, trendLength, null);
	}
	
	/**
	 * Constructs a new AcopGraphParameters object with the specified parameters values
	 * 
	 * @param parameters remote connection point
	 * @param color color of the graph
	 * @param fft acop fft mode
	 * @param drawStyle style of the graph
	 * @param mode acop graph mode 
	 * @param width the width of the line
	 * @param trend a flag indicating if this is a trend graph
	 * @param trendLength number of trend points if this is a trend
	 * @param converter converter associated with this connection point
	 * @see Acop
	 */
	public AcopGraphParameters(ConnectionParameters parameters, Color color, int fft, int drawStyle, int mode, int width, boolean trend, int trendLength, Converter converter) {
		this(parameters,-1,Double.NaN,Double.NaN,null,null,color, fft, drawStyle, mode, width, trend, trendLength, converter);
	}

	/**
	 * Constructs a new AcopGraphParameters object with the specified parameters values
	 * 
	 * @param parameters remote connection point
	 * @param index index of the value in the array if this is a spectrum
	 * @param min minimum value
	 * @param max maximum value
	 * @param units physical units 
	 * @param format C-style format
	 * @param color color of the graph
	 * @param fft acop fft mode
	 * @param drawStyle style of the graph
	 * @param mode acop graph mode 
	 * @param width the width of the line
	 * @param trend a flag indicating if this is a trend graph
	 * @param trendLength number of trend points if this is a trend
	 * @param converter converter associated with this connection point
	 * @see Acop
	 * 
	 */
	protected AcopGraphParameters(ConnectionParameters parameters, int index, double min, double max, String units, String format, Color color, int fft, int drawStyle, int mode, int width, boolean trend, int trendLength, Converter converter) {
		super(parameters,index,min,max,units,format, converter);
		values.put("color",color);
		values.put("fft",fft);
		values.put("drawStyle",drawStyle);
		values.put("mode",mode);
		values.put("width",width);
		values.put("trend",trend);
		values.put("trendLength",trendLength);
	}
	
	/**
	 * Constructs new AcopGraphParameters using the supplied parameters.
	 * 
	 * @param adp
	 */
	public AcopGraphParameters (AcopDisplayerParameters adp) {
		this(adp, null, -1, -1, -1, 1, false, 100, null);
	}
	
	/**
	 * Constructs a new AcopGraphParameters object with the specified parameters values
	 * 
	 * @param adp additional displayer parameters
	 * @param color color of the graph
	 * @param fft acop fft mode
	 * @param drawStyle style of the graph
	 * @param mode acop graph mode 
	 * @param width the width of the line
	 * @param trend a flag indicating if this is a trend graph
	 * @param trendLength number of trend points if this is a trend
	 * @see Acop
	 */
	public AcopGraphParameters(AcopDisplayerParameters adp, Color color, int fft, int drawStyle, int mode, int width, boolean trend, int trendLength) {
		this(adp.getConnectionParameters(), adp.getArrayIndex(), adp.getMinimum(), adp.getMaximum(), adp.getUnits(), adp.getFormat(), color, fft, drawStyle, mode, width, trend, trendLength, adp.getConverter());
	}
	
	/**
	 * Constructs a new AcopGraphParameters object with the specified parameters values
	 * 
	 * @param adp additional displayer parameters
	 * @param color color of the graph
	 * @param fft acop fft mode
	 * @param drawStyle style of the graph
	 * @param mode acop graph mode 
	 * @param width the width of the line
	 * @param trend a flag indicating if this is a trend graph
	 * @param trendLength number of trend points if this is a trend
	 * @param converter converter associated with this connection point
	 * @see Acop
	 */
	public AcopGraphParameters(AcopDisplayerParameters adp, Color color, int fft, int drawStyle, int mode, int width, boolean trend, int trendLength, Converter converter) {
		this(adp.getConnectionParameters(), adp.getArrayIndex(), adp.getMinimum(), adp.getMaximum(), adp.getUnits(), adp.getFormat(), color, fft, drawStyle, mode, width, trend, trendLength, converter);
	}

	/**
	 * Sets the color of the graph.
	 * 
     * @return the color
     * @see Acop#setForeground(Color)
     */
    public Color getColor() {
    	return (Color)values.get("color");
    }

	/**
	 * Returns the acop fft mode.
	 * 
     * @return the fFT
     * @see Acop#setFFT(int)
     */
    public int getFFT() {
    	return getInt("fft");
    }

	/**
	 * Returns the acop draw style.
	 * 
     * @return the graphStyle
     * @see AcopDrawStyleEnum
     * @see Acop#setDrawStyle(int)
     */
    public int getDrawStyle() {
    	return getInt("drawStyle");
    }
    
    /**
     * Returns true if this is shown as a trend graph.
     * 
     * @return is trend
     */
    public boolean isTrend() {
    	return getBool("trend");
    }
    
    /**
     * Returns the number of points if this is a trend.
     * 
     * @return the trend length
     */
    public int getTrendLength() {
    	return getInt("trendLength");
    }
	
	/**
	 * Returns the acop display mode.
	 * 
     * @return the mode
     * @see AcopDisplayMode
     * @see Acop#setDisplayMode(int)
     */
    public int getMode() {
    	return getInt("mode");
    }

	/**
	 * Returns the width of the graph line.
	 * 
     * @return the width
     * @see Acop#setDrawWidth(int)
     */
    public int getWidth() {
    	return getInt("width");
    }
}


/* __oOo__ */
