/*
 * Copyright (c) 2006 Stiftung Deutsches Elektronen-Synchroton,
 * Member of the Helmholtz Association, (DESY), HAMBURG, GERMANY.
 *
 * THIS SOFTWARE IS PROVIDED UNDER THIS LICENSE ON AN "../AS IS" BASIS.
 * WITHOUT WARRANTY OF ANY KIND, EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR PARTICULAR PURPOSE AND
 * NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE. SHOULD THE SOFTWARE PROVE DEFECTIVE
 * IN ANY RESPECT, THE USER ASSUMES THE COST OF ANY NECESSARY SERVICING, REPAIR OR
 * CORRECTION. THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF THIS LICENSE.
 * NO USE OF ANY SOFTWARE IS AUTHORIZED HEREUNDER EXCEPT UNDER THIS DISCLAIMER.
 * DESY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS,
 * OR MODIFICATIONS.
 * THE FULL LICENSE SPECIFYING FOR THE SOFTWARE THE REDISTRIBUTION, MODIFICATION,
 * USAGE AND OTHER RIGHTS AND OBLIGATIONS IS INCLUDED WITH THE DISTRIBUTION OF THIS
 * PROJECT IN THE FILE LICENSE.HTML. IF THE LICENSE IS NOT INCLUDED YOU MAY FIND A COPY
 * AT HTTP://WWW.DESY.DE/LEGAL/LICENSE.HTM
 */

package de.desy.acop.displayers.tools;

import java.beans.PropertyChangeListener;
import java.beans.PropertyVetoException;

import com.cosylab.util.CommonException;

import de.desy.acop.transport.ConnectionParameters;

/**
 * <code>ConnectionParametersReceiver</code> describes the class that can 
 * accept <code>ConnectionParameters</code>. The purpose of the interface is to
 * identify the class as the object, whose connectionParameters property
 * can be configured through property editors. Implementor of this class
 * should follow java beans conventions.
 * 
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 *
 */
public interface ConnectionParametersReceiver {

	/** ConnectionParameters property name tag */
	public static final String CONNECTION_PARAMETERS_PROPERTY = "connectionParameters";
	
	/**
	 * Returns ConnectionParameters associated with this receiver.
	 * 
	 * @return the connection parameters
	 */
	public ConnectionParameters getConnectionParameters();
	
	/**
	 * Sets the ConnectionParameters to the receiver. If a receiver is a 
	 * a displayer it should connect to the remote point upon setting the parameters.
	 * 
	 * @param cp new parameters
	 * @throws CommonException if connection establishing failed
	 * @throws PropertyVetoException if setting failed due to improper parameters
	 */
	public void setConnectionParameters(ConnectionParameters cp) throws CommonException, PropertyVetoException;
	
	/**
	 * Adds a PropertyChangeListener to this receiver. Listener should receives
	 * notifications when the property with the name <code>propertyName</code> was 
	 * changed.
	 * 
	 * @param propertyName the name of the property which the listener listens
	 * @param listener the listener
	 */
	public void addPropertyChangeListener(String propertyName, PropertyChangeListener listener);
		
	/**
	 * Removes a PropertyChangeListener associated with the specified property.
	 * 
	 * @param propertyName the name of the property which the listener listens
	 * @param listener the listener
	 */
	public void removePropertyChangeListener(String propertyName, PropertyChangeListener listener);
	
	/**
	 * Adds a PropertyChangeListener which listenes to all properties.
	 * @param listener the listener
	 */
	public void addPropertyChangeListener(PropertyChangeListener listener);
	
	/**
	 * Removes a PropertyChangeListener.
	 * @param listener
	 */
	public void removePropertyChangeListener(PropertyChangeListener listener);
	
}
