/*
 * Copyright (c) 2006 Stiftung Deutsches Elektronen-Synchroton,
 * Member of the Helmholtz Association, (DESY), HAMBURG, GERMANY.
 *
 * THIS SOFTWARE IS PROVIDED UNDER THIS LICENSE ON AN "../AS IS" BASIS.
 * WITHOUT WARRANTY OF ANY KIND, EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR PARTICULAR PURPOSE AND
 * NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE. SHOULD THE SOFTWARE PROVE DEFECTIVE
 * IN ANY RESPECT, THE USER ASSUMES THE COST OF ANY NECESSARY SERVICING, REPAIR OR
 * CORRECTION. THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF THIS LICENSE.
 * NO USE OF ANY SOFTWARE IS AUTHORIZED HEREUNDER EXCEPT UNDER THIS DISCLAIMER.
 * DESY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS,
 * OR MODIFICATIONS.
 * THE FULL LICENSE SPECIFYING FOR THE SOFTWARE THE REDISTRIBUTION, MODIFICATION,
 * USAGE AND OTHER RIGHTS AND OBLIGATIONS IS INCLUDED WITH THE DISTRIBUTION OF THIS
 * PROJECT IN THE FILE LICENSE.HTML. IF THE LICENSE IS NOT INCLUDED YOU MAY FIND A COPY
 * AT HTTP://WWW.DESY.DE/LEGAL/LICENSE.HTM
 */

package de.desy.acop.launcher;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Toolkit;
import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.lang.reflect.Constructor;
import java.util.ArrayList;
import java.util.Optional;

import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JInternalFrame;
import javax.swing.JPanel;

import com.cosylab.gui.components.util.ComponentPositioner;


/**
 * 
 * <code>Launcher</code> is a utility class which can launch an application
 * according to its type and supplied parameters.
 * 
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 */
public class Launcher {
	
	public static final String CAMNAME = "de.desy.mst.libs.framework.mstapp.util.camInstrumentation.CAMframeControl";
	
	/**
	 * 
	 * <code>StreamGobbler</code> reads the data from the input stream
	 * and forwards it to the outputstream.
	 *
	 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
	 *
	 */
	private static class StreamGobbler extends Thread {
		private InputStream is;
		private OutputStream os;
		private Optional<Boolean> errorDetected = Optional.empty();

		StreamGobbler(InputStream is, OutputStream os) {
			this.is = is;
			this.os = os;
			setDaemon(true);
		}

		/**
		 * creates readers to handle the text created by the external program
		 */
		@Override
		public void run() {
			try (BufferedReader reader = new BufferedReader(new InputStreamReader(is))) {
			    @SuppressWarnings("resource")
                PrintWriter writer = os == null ? null : new PrintWriter(os); 
				String line;
				while ((line = reader.readLine()) != null) {
				    if (writer != null) {
				        writer.println(line);
				    }
				    if (line.contains("not recognized")) {
				        errorDetected = Optional.of(Boolean.TRUE);
				    } else {
				        errorDetected = Optional.of(Boolean.FALSE);
				    }
				}
				reader.close();
				if (!errorDetected.isPresent()) {
                    errorDetected = Optional.of(Boolean.FALSE);
                }
			} catch (IOException ioe) {
				ioe.printStackTrace();
				if (!errorDetected.isPresent()) {
                    errorDetected = Optional.of(Boolean.TRUE);
                }
			}
		}
	}
	
	/**
	 * Starts a webstart application. The provided string should be the url to the
	 * jnlp file. This method will construct a webstart start command composed of all
	 * provided program and JVM parameters and forward the call to 
	 * {@link #runWebstartApplication(String, boolean)}.
	 * 
	 * @param jnlpURL the url to the jnlp file to be started
	 * @param jvmParameters the virtual machine parameters (without -D)
	 * @param waitForCompletion true if the calling thread should be blocked unti the execution is completed
	 * @throws IOException if execution of the command failed
	 */
	public static void runWebstartApplication(String jnlpURL, String[] jvmParameters, boolean waitForCompletion) throws IOException {
		ArrayList<String> list = new ArrayList<>();
		if (jvmParameters != null) {
    		for (int i = 0; i < jvmParameters.length; i++) {
    			list.add("-J-D"+jvmParameters[i]);
    		}
		}
		list.add(jnlpURL);		
		runWebstartApplication(list.toArray(new String[list.size()]),waitForCompletion);
	}
	
	/**
	 * Method will start a webstart application. The provided String should be 
	 * a complete url to the jnlp file. The provided url can also include 
	 * several program or JVM arguments. They have to be supplied as prescribed
	 * by the Java API. This method will forward the call to 
	 * {@link #runWebstartApplication(String, OutputStream, OutputStream, boolean)}
	 * where the streams are null.
	 *  
	 * @see #runCommand(String, OutputStream, OutputStream, boolean) 
	 *  
	 * @param completeJNLP_URL the url which will be loaded using the java
	 * 			webstart command (javaws) separated by spaces
	 * @param waitForCompletion the flag indicating if the calling thread
	 * 			should be blocked until the execution is completed
	 * @throws IOException if execution of the command failed
	 */
	public static void runWebstartApplication(String[] completeJNLP_URL, boolean waitForCompletion) throws IOException {
		runWebstartApplication(completeJNLP_URL, (OutputStream)null, (OutputStream)null, waitForCompletion);
	}

	/**
	 * Method will start a webstart application. The provided String should be 
	 * a complete url to the jnlp file. The provided url can also include 
	 * several program or JVM arguments. They have to be supplied as prescribed
	 * by the Java API.
	 *  
	 * @see #runCommand(String, OutputStream, OutputStream, boolean) 
	 *  
	 * @param completeJNLP_URL the url which will be loaded using the java
	 * 			webstart command (javaws) separated by spaces
	 * @param outputStream the output stream where all the output from the
	 * 			input stream of the executed process will be forwarded
	 * @param errorOutput the output stream where all the output from the
	 * 			error stream of the executed process will be forwarded
	 * @param waitForCompletion the flag indicating if the calling thread
	 * 			should be blocked until the execution is completed
	 * @throws IOException if execution of the command failed
	 */
	public static void runWebstartApplication(String[] completeJNLP_URL, OutputStream outputStream, OutputStream errorOutput, boolean waitForCompletion) throws IOException {
	    //first try with jaws
	    String[] s = new String[completeJNLP_URL.length + 3];
        s[0] = "cmd";
        s[1] = "/c";
        s[2] = "jaws";
        System.arraycopy(completeJNLP_URL,0,s,3,completeJNLP_URL.length);
        if (!runCommand(s, outputStream, errorOutput, waitForCompletion,true)) {
    	    //if it fails try javaws
    		String jvm = System.getProperty("java.home") + File.separator + "bin" + File.separator + "javaws";
    		s = new String[completeJNLP_URL.length + 1];
    		s[0] = jvm;
    		System.arraycopy(completeJNLP_URL,0,s,1,completeJNLP_URL.length);
    		runCommand(s, outputStream, errorOutput, waitForCompletion,false);
        }
	}
	
	/**
	 * Runs the specified command. This method will invoke the {@link Runtime#exec(String)},
	 * where the provided string parameters is the command. The output generated by
	 * the execution of this command is forwarded to the provided {@link OutputStream}s.
	 * If any of these stream are null the output from that part of execution will
	 * be skipped.
	 * <p>
	 * The invoker has the possibility to block the current thread until the 
	 * process created by this method is completed. The lifecycle of the process
	 * is specified by the command itself and the process which will be executed
	 * by that command.
	 * 
	 * @param command the command to be executed  
	 * @param outputStream the output stream where all the data from the {@link Process}'s
	 * 			input stream will be forwarded (null is allowed)
	 * @param errorOutput the output stream where all the data from the {@link Process}'s
	 * 			error stream will be forwarded (null is allowed)
	 * @param waitForCompletion a flag indicating if the calling thread should be blocked
	 * 			until the process is completed
	 * @return true if the command was executed successfully (to the launcher knowledge), or false if it didn't. Command
	 *          was not executed successfully, if the error stream receives a message that command is not recognized. 
	 * @throws IOException if execution of the command failed
	 */
	public static boolean runCommand(String command[], OutputStream outputStream, OutputStream errorOutput, boolean waitForCompletion) throws IOException {
		return runCommand(command,outputStream,errorOutput,waitForCompletion,false);
	}
	
	/**
     * Runs the specified command. This method will invoke the {@link Runtime#exec(String)},
     * where the provided string parameters is the command. The output generated by
     * the execution of this command is forwarded to the provided {@link OutputStream}s.
     * If any of these stream are null the output from that part of execution will
     * be skipped.
     * <p>
     * The invoker has the possibility to block the current thread until the 
     * process created by this method is completed. The lifecycle of the process
     * is specified by the command itself and the process which will be executed
     * by that command.
     * 
     * @param command the command to be executed  
     * @param outputStream the output stream where all the data from the {@link Process}'s
     *          input stream will be forwarded (null is allowed)
     * @param errorOutput the output stream where all the data from the {@link Process}'s
     *          error stream will be forwarded (null is allowed)
     * @param waitForCompletion a flag indicating if the calling thread should be blocked
     *          until the process is completed
     * @param parseErrorStreamForError if true the error stream is parsed for any OS messages that the execution
     *          was not successful, if false the error stream messages are ignored
     * @return true if the command was executed successfully (to the launcher knowledge), or false if it didn't. Command
     *          was not executed successfully, if the error stream receives a message that command is not recognized. 
     * @throws IOException if execution of the command failed
     */
	public static boolean runCommand(String command[], OutputStream outputStream, OutputStream errorOutput, boolean waitForCompletion,
	        boolean parseErrorStreamForError) throws IOException {
        Process p = Runtime.getRuntime().exec(command);
        if (outputStream != null) {
            new StreamGobbler(p.getInputStream(), outputStream).start();
        }
        StreamGobbler errorGobler = new StreamGobbler(p.getErrorStream(), errorOutput);
        errorGobler.start();

        if (waitForCompletion) {
            try {
                int exit = p.waitFor();
                return exit == 0;
            } catch (InterruptedException e) {
                //ignore
            }
        } else {
            if (parseErrorStreamForError) {
                int exitValue = 0;
                try {
                    while(!errorGobler.errorDetected.isPresent()) {
                        Thread.sleep(500);
                    }
                    exitValue = p.exitValue();
                } catch (InterruptedException | IllegalThreadStateException e) {
                    //ignore
                }
                return !errorGobler.errorDetected.get() && exitValue == 0;
            } else {
                return true;
            }
        }
        return false;
    }
	
	
	private static String getClassName() {
		return Launcher.class.getName();
	}
	
	public static void main(String[] args) throws IOException {
//	    Process p = Runtime.getRuntime().exec(new String[]{"cmd", "/c", "jaws", "http://users.cosylab.com/~jbobnar/tine/archiveviewer.jnlp"});
	    
		Launcher.runWebstartApplication(new String[]{"http://users.cosylab.com/~jbobnar/tine/archiveviewer.jnlp"},null,null , false);
	}
	
	/**
	 * Launches an application using the input arguments. This call is forwarded
	 * to {@link #launch(Class, String[], boolean)}, where shouldExitOnClose is false.
	 * 
	 * @param <T> the object which should be constructed and returned
	 * @param applicationClass the class decribing that object
	 * @param args program arguments
	 * @return the instance of the application
	 */
	public static <T> T launch(final Class<T> applicationClass, final String[] args) {
		return launch(applicationClass, args, false);
	}
		
	/**
	 * Launches an application using the input arguments. This call is forwarded
	 * to {@link #launch(Class, String[], boolean)}, where shouldExitOnClose is given.
	 * 
	 * @param <T> the object which should be constructed and returned
	 * @param applicationClass the class decribing that object
	 * @param args program arguments
	 * @param shouldExitOnClose should the application be close on exit
	 * @return the instance of the application
	 */
	public static <T> T launch(final Class<T> applicationClass, final String[] args, boolean shouldExitOnClose) {
		return launch(applicationClass, args, shouldExitOnClose, true);
	}
	
	/**
	 * Launches an application defined by the applicationClass. The method will try
	 * to create an object of the given instance using the constructor with String[]
	 * parameter. If this fails, it will call invoke the main method on the the 
	 * give application class using the args arguments. If the main method doesn't
	 * exist an object's default constructor will be used (in this case args array is
	 * useless). 
	 * <p>
	 * If the application is not a window type but it is at least a {@link Component},
	 * a JFrame will be created and the component will be added to that frame.
	 * DefaultCloseOperation for the frame is EXIT or DISPOSE and it depends
	 * on the shouldExitOnClose parameter. 
	 * </p>
	 * <p>
	 * This method is not thread safe and will block the current thread until the
	 * application is loaded. One should take of appropriate multithreading to avoid
	 * GUI freezing.
	 * </p>
	 *  
	 * @param <T> the object which should be constructed and returned
	 * @param applicationClass the class describing that object
	 * @param args the program arguments
	 * @param shouldExitOnClose if true the frame will have EXIT_ON_CLOSE defaultCloseOperation
	 * @param shouldDisposeOnClose if true the frame will have DISPOSE_ON_CLOSE onless exit is requested
	 * @return the instance of application if it was started
	 */
	public static <T> T launch(final Class<T> applicationClass, final String[] args, boolean shouldExitOnClose, boolean shouldDisposeOnClose) {
		return launch(applicationClass,args,shouldExitOnClose,shouldDisposeOnClose,true);
	}
	/**
	 * Launches an application defined by the applicationClass. The method will try
	 * to create an object of the given instance using the constructor with String[]
	 * parameter. If this fails, it will call invoke the main method on the the 
	 * give application class using the args arguments. If the main method doesn't
	 * exist an object's default constructor will be used (in this case args array is
	 * useless). 
	 * <p>
	 * If the application is not a window type but it is at least a {@link Component},
	 * a JFrame will be created and the component will be added to that frame.
	 * DefaultCloseOperation for the frame is EXIT or DISPOSE and it depends
	 * on the shouldExitOnClose parameter. 
	 * </p>
	 * <p>
	 * This method is not thread safe and will block the current thread until the
	 * application is loaded. One should take of appropriate multithreading to avoid
	 * GUI freezing.
	 * </p>
	 *  
	 * @param <T> the object which should be constructed and returned
	 * @param applicationClass the class describing that object
	 * @param args the program arguments
	 * @param shouldExitOnClose if true the frame will have EXIT_ON_CLOSE defaultCloseOperation
	 * @param shouldDisposeOnClose if true the frame will have DISPOSE_ON_CLOSE onless exit is requested
	 * @param initializeFrameControl true if the CAM frame control should be initialized for the new window
	 * @return the instance of application if it was started
	 */
	public static <T> T launch(final Class<T> applicationClass, final String[] args, boolean shouldExitOnClose, boolean shouldDisposeOnClose, boolean initializeFrameControl) {
		if (applicationClass == null) {
			System.err.println("No class specified to launch in '"+getClassName()+".launch()' method");
			return null;
		}
		if (args == null) {
			System.err.println("No arguments specified in '"+getClassName()+".launch()' method.");
			return null;
		}
				
		Constructor<T> constructor = null;
		T obj;
		try {
			constructor = applicationClass.getConstructor(new Class[]{String[].class});
		} catch (NoSuchMethodException e2) {
			//ignore
		}
		
		if (constructor == null) {
			try {
				constructor = applicationClass.getConstructor();
				obj = constructor.newInstance();
			} catch (Exception e) {
				e.printStackTrace();
				System.err.println("Cannot launch class '"+applicationClass + "'. \nIt does not provide a Constructor with parameters 'String[]' nor default constructor.");
				return null;
			}

			/*
			 * This code below effecitvly prevents you to launch class from 
			 * main method, it makes closed loop.
			 * This happens only if you launch it from the main method that invoked
			 * this Launcher call.
			 */
			
			/*//if no String[] constructor try main method
			Method mainMethod = null;
			try {
				mainMethod = applicationClass.getMethod("main", new Class[]{String[].class});
			} catch (NoSuchMethodException e1) {
				//ignore
			}

			if (mainMethod != null) {
				try {
					mainMethod.invoke(null, new Object[]{args});
					return null;
				} catch (Exception e) {
					//ignore
				}
			}
			//if no main method try default constructor
			try {
				constructor = applicationClass.getConstructor(new Class[]{});
			} catch (NoSuchMethodException e2) {
				//ignore
			}
			//if no default constructor abort
			if (constructor == null) {
				System.out.println("Cannot launch class '"+applicationClass + "'. \nIt does not provide a Constructor with parameters 'String[]' nor default constructo or the main method.");
				return null;
			}
			
			//create new object
			try {
				obj = constructor.newInstance(new Object[]{});
			} catch (Exception e) {
				e.printStackTrace();
				return null;
			}*/
			
		} else {
			try {
				obj = constructor.newInstance(new Object[]{args});
			} catch (Exception e) {
				e.printStackTrace();
				return null;
			}
		}
		if (obj != null) {
			int closeOperation = shouldExitOnClose ? JFrame.EXIT_ON_CLOSE : shouldDisposeOnClose ? JFrame.DISPOSE_ON_CLOSE : JFrame.DO_NOTHING_ON_CLOSE;
			if (JFrame.class.isAssignableFrom(applicationClass)) {
				JFrame frame = JFrame.class.cast(obj);
				frame.setDefaultCloseOperation(closeOperation);
				ComponentPositioner.centerOnScreen(frame);
				frame.setVisible(true);
				if (initializeFrameControl)	initializeFrameControl(frame);
			} else if (JDialog.class.isAssignableFrom(applicationClass)) {
				JDialog frame = JDialog.class.cast(obj);
				frame.setDefaultCloseOperation(closeOperation);
				ComponentPositioner.centerOnScreen(frame);
				frame.setVisible(true);
			} else if (JInternalFrame.class.isAssignableFrom(applicationClass)) { 
				JInternalFrame frame = JInternalFrame.class.cast(obj);
				frame.setDefaultCloseOperation(closeOperation);
				frame.setVisible(true);
			} else if (Container.class.isAssignableFrom(applicationClass)) {
				Container comp = Container.class.cast(obj);
				JFrame frame = new JFrame();
				frame.setDefaultCloseOperation(closeOperation);
				frame.setContentPane(comp);
				Dimension d = Toolkit.getDefaultToolkit().getScreenSize();
				frame.setSize((int)(d.width*0.8), (int)(d.height*0.8));
				ComponentPositioner.centerOnScreen(frame);
				frame.setVisible(true);
				if (initializeFrameControl)	initializeFrameControl(frame);
			} else if (Component.class.isAssignableFrom(applicationClass)) {
				Component comp = Component.class.cast(obj);
				JFrame frame = new JFrame();
				frame.setDefaultCloseOperation(closeOperation);
				JPanel p = new JPanel(new BorderLayout());
				p.add(comp, BorderLayout.CENTER);
				frame.setContentPane(p);
				Dimension d = Toolkit.getDefaultToolkit().getScreenSize();
				frame.setSize((int)(d.width*0.8), (int)(d.height*0.8));
				ComponentPositioner.centerOnScreen(frame);
				frame.setVisible(true);
				if (initializeFrameControl)	initializeFrameControl(frame);
			}
			return obj;
		}
		return null;
	}
	
	/**
	 * Initializes Component Application Manager. This method uses reflection to initialize
	 * the class, and than instantiates the manager.
	 * 
	 * @param frame the frame to connect it with the manager
	 * @see de.desy.mst.libs.framework.mstapp.util.camInstrumentation.CAMframeControl
	 */
	private static final void initializeFrameControl(Frame frame) {
		try {
    		Class<?> c = Class.forName(CAMNAME);
    		Constructor<?> constructor = c.getConstructor(new Class[]{Frame.class});
    		constructor.newInstance(frame);
		} catch (Exception e) {
			System.out.println("Cannot initialize CAM frame control.");
		}
	}
}
