package de.desy.acop.print.logbook;

import java.awt.Color;
import java.awt.Component;
import java.awt.Frame;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Window;
import java.awt.image.BufferedImage;
import java.awt.print.PageFormat;
import java.awt.print.Printable;
import java.awt.print.PrinterException;
import java.awt.print.PrinterJob;
import java.io.File;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.Iterator;
import java.util.Locale;

import javax.imageio.IIOImage;
import javax.imageio.ImageIO;
import javax.imageio.ImageTypeSpecifier;
import javax.imageio.ImageWriter;
import javax.imageio.metadata.IIOMetadata;
import javax.imageio.plugins.jpeg.JPEGImageWriteParam;
import javax.imageio.stream.ImageOutputStream;
import javax.print.PrintService;
import javax.print.PrintServiceLookup;
import javax.print.attribute.Attribute;
import javax.print.attribute.HashPrintRequestAttributeSet;
import javax.print.attribute.PrintRequestAttributeSet;
import javax.print.attribute.standard.DialogTypeSelection;
import javax.swing.JOptionPane;
import javax.swing.RepaintManager;
import javax.swing.SwingUtilities;

import com.cosylab.gui.components.util.PaintHelper;


/**
 * 
 * <code>PrintUtilities</code> provides a common API to print a swing component.
 * Caller is allowed to choose to print the component directly to the default
 * printer, to logbook or to select a printer from the registered list.
 *
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 *
 */
public class PrintUtilities implements Printable {

	private PrintRequestAttributeSet printRequestAttributeSet;
	private PrintDialogForm printDialog;
	
	private final Component printableComponent;
	private final PrintCallback callback;
	private final boolean printBorder;
	private BufferedImage snapshotImage;
	
	public PrintUtilities(Component printable, PrintCallback callback, boolean printBorder) {
		this.printableComponent = printable;
		this.callback = callback;
		this.printBorder = printBorder;
	}
	
	private PrintRequestAttributeSet getPrintRequestAttributeSet() {
		if (printRequestAttributeSet == null) {
			printRequestAttributeSet = new HashPrintRequestAttributeSet();
			printRequestAttributeSet.add(DialogTypeSelection.NATIVE);	
			// Prior to Java 11, the DialogOnwer was in package sun.print.DialogOnwer,
			// since java 11 it is in javax.print
			Class<?> dialogOwnerClass = null;
			try {
			    dialogOwnerClass = Class.forName("sun.print.DialogOwner");
            } catch (ClassNotFoundException e) {
                try {
                    dialogOwnerClass = Class.forName("javax.print.attribute.standard.DialogOwner");
                } catch (ClassNotFoundException ex) {
                    //ignore
                }
            }
			if (dialogOwnerClass != null) {
			    Window window = SwingUtilities.getWindowAncestor(printableComponent);
			    if (window instanceof Frame) {
    			    Frame frame = (Frame) window;
    			    try {
                        Constructor<?> dialogConstructor = dialogOwnerClass.getConstructor(Frame.class);
                        Attribute dialogOwner = (Attribute)dialogConstructor.newInstance(frame);
                        printRequestAttributeSet.add(dialogOwner);
                    } catch (NoSuchMethodException | SecurityException | InstantiationException | 
                            IllegalAccessException | IllegalArgumentException | InvocationTargetException e) {
                        // ignore
                    }
			    }
			}
		}
		return printRequestAttributeSet;
	}
	
	
	/**
	 * Performs printing of the application. Methods takes this class as the 
	 * Printable object which redirects the call to the printAll() of the printableComponent
	 * set at constructions and prints it. It uses the local 
	 * PrintRequestAttributeSet and performs printing to the printer specified by 
	 * the parameter. If print is called with {@link PrinterType#CHOOSABLE} a selector
	 * dialog is displayed
	 * 
	 * @param printer the printer to print to
	 * @param context the context to select default logbook printer if necessary
	 */
	public void print(final PrinterType printer, final String context) {
		print(printer,context,false);
	}
	
	/**
	 * Performs printing of the application. Method can either take a snapshot
	 * of the printableComponent or it takes this class as the 
	 * Printable object which redirects the call to the printAll() of the printableComponent
	 * set at constructions and prints it. It uses the local 
	 * PrintRequestAttributeSet and performs printing to the printer specified by 
	 * the parameter. If print is called with {@link PrinterType#CHOOSABLE} a selector
	 * dialog is displayed. 
	 * 
	 * 
	 * @param printer the printer to print to
	 * @param context the context to select default logbook printer if necessary
	 * @param takeSnapshot true if the snapshot of the current application is taken and the 
	 * 			snapshot is sent to the printer or false if the component is rendered on the
	 * 			printer graphics
	 */
	public void print(final PrinterType printer, final String context, final boolean takeSnapshot) {
		this.snapshotImage = null;
		if (takeSnapshot) {
			try {
    			Point loc = printableComponent.getLocationOnScreen();
    			Rectangle size = printableComponent.getBounds();
    			this.snapshotImage = MstSnapshot.captureWindow(loc.x,loc.y,size.width,size.height);
			} catch (Exception e) {
				//ignore
			}
    		
		}
		if (printer == PrinterType.LOGBOOK) {
			try {
				getPrintDialogForm(context).showdialog(snapshotImage);
			} catch (Exception e1) {
				callback.reportException(e1,"Exception during printing to logbook.");
			} finally {
				callback.finished();
			}
		} else {
    		new Thread(new Runnable() {
    			public void run() {
    				PrinterJob printerJob = PrinterJob.getPrinterJob();
    				printerJob.setJobName(callback.getApplicationInfo().getApplication());   
    				printerJob.setPrintable(PrintUtilities.this);
    				try {
    				    if (printer == PrinterType.DEFAULT) {
    			    		printerJob.setPrintService(PrintServiceLookup.lookupDefaultPrintService());
    			    		Thread.yield();
    			    		printerJob.print(getPrintRequestAttributeSet());
    			    		JOptionPane.showMessageDialog(printableComponent, "Snapshot successfully sent to printer.", "Print", JOptionPane.INFORMATION_MESSAGE);
    				    } else if (printer == PrinterType.CHOOSABLE) {
    				    	if (printerJob.printDialog(getPrintRequestAttributeSet())) {
    					    	Thread.yield();
    					    	if (isLogBookPrinter(printerJob.getPrintService())) {
    					    		printToLogBook(printerJob.getPrintService().getName());
    					    	} else {
    					    		printerJob.print(getPrintRequestAttributeSet());
    					    	}
    					    	JOptionPane.showMessageDialog(printableComponent, "Snapshot successfully sent to printer.", "Print", JOptionPane.INFORMATION_MESSAGE);
    				    	}
    				    } else if (printer == PrinterType.LOGBOOK) {
    				    	printToLogBook(getDefaultLogbookName(context));
    				    	JOptionPane.showMessageDialog(printableComponent, "Snapshot successfully sent to printer.", "Print", JOptionPane.INFORMATION_MESSAGE);
    				    }			   
    				    
    				} catch (Exception e) {
    					callback.reportException(e,"Exception during printing.");
    				} finally {
    					printerJob.cancel();
    					callback.finished();
    				}
    			}
    		},"Print").start();
		}
		
	}

	private String getDefaultLogbookName(String context) {
		if (context == null) context = "";
		return "//adprint/" + context + "log";
	}
	
	/**
	 * Returns the print dialog form, which grabs the current content panel
	 * as the image source.
	 * 
	 * @param context the context for the logbook selection
	 * @return the print dialog form
	 */
	private PrintDialogForm getPrintDialogForm(String context) {
		if (printDialog == null) {
			printDialog = new PrintDialogForm(printableComponent,true,callback);
		}
		printDialog.setDefaultMaschine(Maschine.ensStringToMachine(context));
		return printDialog;
	}
	
	/**
	 * Checks if the provided print service points to the log book printer.
	 * This is determined if the service name starts with '//adprint/' and ends with 'log'
	 * 
	 * @param service the print service to check if it points to the log book
	 * @return true if the print service is a log book printer
	 */
	private boolean isLogBookPrinter(PrintService service) {
		String name = service.getName().toLowerCase();
		name = name.replace('\\', '/');
		return (name.startsWith("//adprint/") && name.endsWith("log"));
	}
	
	/**
	 * Prints the contentPanel to the specified log book.
	 * @param logBookName
	 */
	private void printToLogBook(String logBookName) {
		ImageWriter iw = null;
    	Iterator<ImageWriter> iter = ImageIO.getImageWritersByFormatName("JPEG");
    	if(iter.hasNext()) {
    		iw = iter.next();
    	}
    	if(iw == null) return;
    	BufferedImage image = null;
        //draw
        if (snapshotImage == null) {
            image = new BufferedImage(printableComponent.getWidth(), 
                    printableComponent.getHeight(), BufferedImage.TYPE_INT_RGB);
            Graphics2D g = image.createGraphics();
            boolean doubleBuffer = RepaintManager.currentManager(printableComponent).isDoubleBufferingEnabled();
            
            printableComponent.printAll(g);
            if (printBorder) {
                g.setColor(Color.BLACK);
                g.drawRect(0, 0, printableComponent.getWidth()-1, printableComponent.getHeight()-1);
            }
            Thread.yield();
            RepaintManager.currentManager(printableComponent).setDoubleBufferingEnabled(doubleBuffer);
            g.dispose();
        } else {
            image = snapshotImage;
        }
        
    	
    	try (ImageOutputStream imageOut =ImageIO.createImageOutputStream(new File(logBookName))){
			JPEGImageWriteParam param = new JPEGImageWriteParam(Locale.UK);
    		IIOMetadata data = iw.getDefaultImageMetadata(new ImageTypeSpecifier(image), param);
    		iw.setOutput(imageOut);
    		IIOMetadata streamData = iw.getDefaultStreamMetadata(param);
    		iw.write(streamData, new IIOImage(image, null, data), param);
		} catch (IOException e1) {
			callback.reportException(e1,"Exception during printing to logbook.");
		}
		iw.dispose();
	}
		
	/*
	 * (non-Javadoc)
	 * @see java.awt.print.Printable#print(java.awt.Graphics, java.awt.print.PageFormat, int)
	 */
	public int print(Graphics graphics, PageFormat pageFormat, int pageIndex) throws PrinterException {
		if (pageIndex > 0) {
	        return NO_SUCH_PAGE;
	    } else {	    	
	        Graphics2D g2d = (Graphics2D)graphics;
	        
	        g2d.setRenderingHints(PaintHelper.getAntialiasingHints());
	        int height = printableComponent.getHeight();
	        int width = printableComponent.getWidth();
			double w = pageFormat.getImageableWidth();
			double h = pageFormat.getImageableHeight();
			double xscale = w/width;
			double yscale = h/height;
			double scale = Math.min(xscale, yscale);
		
			g2d.translate(pageFormat.getImageableX(), pageFormat.getImageableY());
			g2d.scale(scale, scale);
	        
			boolean doubleBuffer = RepaintManager.currentManager(printableComponent).isDoubleBufferingEnabled();
			RepaintManager.currentManager(printableComponent).setDoubleBufferingEnabled(false);
			if (snapshotImage == null) {
				printableComponent.printAll(g2d);
			} else {
				g2d.drawImage(snapshotImage,0,0,null);
			}
			if (printBorder) {
				g2d.setColor(Color.BLACK);
				g2d.drawRect(0, 0, width-1, height-1);
			}
	        RepaintManager.currentManager(printableComponent).setDoubleBufferingEnabled(doubleBuffer);
	        return PAGE_EXISTS;
	    }
	}
}
