/*
 * Copyright (c) 2006 Stiftung Deutsches Elektronen-Synchroton, Member of the Helmholtz Association, (DESY), HAMBURG,
 * GERMANY. THIS SOFTWARE IS PROVIDED UNDER THIS LICENSE ON AN "../AS IS" BASIS. WITHOUT WARRANTY OF ANY KIND, EXPRESSED
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR PARTICULAR PURPOSE AND
 * NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. SHOULD THE SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE USER
 * ASSUMES THE COST OF ANY NECESSARY SERVICING, REPAIR OR CORRECTION. THIS DISCLAIMER OF WARRANTY CONSTITUTES AN
 * ESSENTIAL PART OF THIS LICENSE. NO USE OF ANY SOFTWARE IS AUTHORIZED HEREUNDER EXCEPT UNDER THIS DISCLAIMER. DESY HAS
 * NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE FULL LICENSE SPECIFYING
 * FOR THE SOFTWARE THE REDISTRIBUTION, MODIFICATION, USAGE AND OTHER RIGHTS AND OBLIGATIONS IS INCLUDED WITH THE
 * DISTRIBUTION OF THIS PROJECT IN THE FILE LICENSE.HTML. IF THE LICENSE IS NOT INCLUDED YOU MAY FIND A COPY AT
 * HTTP://WWW.DESY.DE/LEGAL/LICENSE.HTM
 */
package de.desy.acop.transport;

import java.io.Serializable;
import java.util.ArrayList;

/**
 * This objecct holds all parameters necessary to initiate ACOP transport connection.
 * 
 * @author Igor Kriznar (igor.kriznarATcosylab.com)
 */
public class ConnectionParameters implements Comparable<ConnectionParameters>, Serializable, Cloneable {

    /**
     * Predefined value for the property size connection parameter, which means that when property will be connected,
     * then actual remote size for property array will be used.
     */
    public static int ACTUAL_PROPERTY_SIZE = 0;
    /**
     * Predefined value for the property size connection parameter, which means that when property will be connected,
     * size of array will be set according to displayer's capabilities.
     */
    public static int AUTO_PROPERTY_SIZE = -1;
    private static final long serialVersionUID = -430933461297463309L;
    private String accessProtocol;
    private String deviceContext;
    private String deviceGroup;
    private String deviceName;
    private String deviceProperty;
    private AccessMode accessMode;
    private int accessRate;
    private int propertySize = AUTO_PROPERTY_SIZE;
    private Object dynamicParameters;

    /**
     * Creates new instance of ConnectionParameters.
     */
    public ConnectionParameters() {}

    /**
     * Constructs a new ConnectionParameters. This constructor assume that the given paramete string includes the remote
     * name which is obligatory as well as AccessMode, access rate and property size, which is optional (all three
     * included or excluded). The parameter can also include everything but the property size. All parameters in the
     * string should be separated by comma (eg. PROTOCOL/CONTEXT/SERVER/DEVICE/PROPERTY,AccessMode.POLL,1000,1).
     * 
     * @param parameters
     * @see #getPropertySize()
     */
    public ConnectionParameters(String parameters) {
        String[] p = parameters.split(",");
        if (p == null) {
            throw new IllegalArgumentException(
                    "Could not construct connection parameters. The provided remote name is null.");
        }
        //the following code solves problems with comma in the device name
        if (p.length > 1) {
            StringBuilder remoteName = new StringBuilder(p[0]);
            int i = 1;
            int j = p.length - 1;
            //Find out where the property is located (the last element that contains
            //the /). Anything before that is definitely part of the remote name.
            for (; j > -1; j--) {
                if (p[j].contains("/")) break;
            }
            for (; i < p.length; i++) {
                if (p[i].contains("/") || i <= j) {
                    remoteName.append(',');
                    remoteName.append(p[i]);
                } else {
                    try {
                        //if we hit the access mode, we are through with the remotename concatenation
                        AccessMode.valueOf(p[i]);
                        break;
                    } catch (IllegalArgumentException e) {
                        try {
                            //if we hit a number, we are through with the remotename concatenation
                            Integer.parseInt(p[i]);
                            break;
                        } catch (NumberFormatException ex) {
                            remoteName.append(',');
                            remoteName.append(p[i]);
                        }
                    }
                }
            }
            //construct a new array with all the stuff combined
            ArrayList<String> param = new ArrayList<>();
            param.add(remoteName.toString());
            for (; i < p.length; i++) {
                param.add(p[i]);
            }
            p = param.toArray(new String[param.size()]);
        }
        if (p.length == 4) {
            setUpRemoteName(p[0]);
            this.accessMode = AccessMode.valueOf(p[1]);
            this.accessRate = Integer.parseInt(p[2]);
            this.propertySize = Integer.parseInt(p[3]);
        } else if (p.length == 1) {
            setUpRemoteName(p[0]);
            this.accessMode = AccessMode.POLL;
            this.accessRate = 1000;
            this.propertySize = AUTO_PROPERTY_SIZE;
        } else if (p.length == 3) {
            setUpRemoteName(p[0]);
            this.accessMode = AccessMode.valueOf(p[1]);
            this.accessRate = Integer.parseInt(p[2]);
        } else {
            throw new IllegalArgumentException(
                    "Failed to construct connection parameters. The string should contain 1, 3, or 4 elements separated by comma.");
        }
    }

    /**
     * @see #ConnectionParameters(String)
     * @param parameters the parameters (remotename,accessmode,accessrate,propertySize)
     * @param dynamicParameters additional parameters that are required for making a connection
     */
    public ConnectionParameters(String parameters, Object dynamicParameters) {
        this(parameters);
        this.dynamicParameters = dynamicParameters;
    }

    /**
     * Creates new instance of ConnectionParameters with TINE access protocol.
     * 
     * @param deviceContext the context
     * @param deviceGroup the server
     * @param deviceName the device name
     * @param deviceProperty the property name
     * @param accessMode the access mode for the connection
     * @param accessRate the polling rate
     */
    public ConnectionParameters(String deviceContext, String deviceGroup, String deviceName,
            String deviceProperty, AccessMode accessMode, int accessRate) {
        this("TINE",deviceContext,deviceGroup,deviceName,deviceProperty,accessMode,accessRate);
    }

    /**
     * Creates new instance of ConnectionParameters.
     * 
     * @param accessProtocol the access protocol (TINE)
     * @param deviceContext the context
     * @param deviceGroup the server
     * @param deviceName the device name
     * @param deviceProperty the property name
     * @param accessMode the access mode for the connection
     * @param accessRate the polling rate
     */
    public ConnectionParameters(String accessProtocol, String deviceContext, String deviceGroup, String deviceName,
            String deviceProperty, AccessMode accessMode, int accessRate) {
        super();
        this.accessProtocol = accessProtocol != null ? accessProtocol : "";
        this.deviceContext = deviceContext != null ? deviceContext : "";
        this.deviceGroup = deviceGroup != null ? deviceGroup : "";
        this.deviceName = deviceName != null ? deviceName : "";
        this.deviceProperty = deviceProperty != null ? deviceProperty : "";
        this.accessMode = accessMode;
        this.accessRate = accessRate;
    }

    /**
     * @see #ConnectionParameters(String, String, String, String, String, AccessMode, int)
     * @param accessProtocol the access protocol (TINE)
     * @param deviceContext the context
     * @param deviceGroup the server
     * @param deviceName the device name
     * @param deviceProperty the property name
     * @param accessMode the access mode for the connection
     * @param accessRate the polling rate
     * @param dynamicParameters additional parameters to make a connection
     */
    public ConnectionParameters(String accessProtocol, String deviceContext, String deviceGroup, String deviceName,
            String deviceProperty, AccessMode accessMode, int accessRate, Object dynamicParameters) {
        this(accessProtocol,deviceContext,deviceGroup,deviceName,deviceProperty,accessMode,accessRate);
        this.dynamicParameters = dynamicParameters;
    }

    /**
     * Constructs a new ConnectionParameters.
     * 
     * @param accessProtocol the access protocol (TINE)
     * @param deviceContext the context
     * @param deviceGroup the server
     * @param deviceName the device name
     * @param deviceProperty the property name
     * @param accessMode the access mode for the connection
     * @param accessRate the polling rate
     * @param propertySize the size of the requested property data
     */
    public ConnectionParameters(String accessProtocol, String deviceContext, String deviceGroup, String deviceName,
            String deviceProperty, AccessMode accessMode, int accessRate, int propertySize) {
        this(accessProtocol,deviceContext,deviceGroup,deviceName,deviceProperty,accessMode,accessRate);
        this.propertySize = propertySize;
    }

    /**
     * @see #ConnectionParameters(String, String, String, String, String, AccessMode, int, int)
     * @param accessProtocol the access protocol (TINE)
     * @param deviceContext the context
     * @param deviceGroup the server
     * @param deviceName the device name
     * @param deviceProperty the property name
     * @param accessMode the access mode for the connection
     * @param accessRate the polling rate
     * @param propertySize the size of the requested property data
     * @param dynamicParameters additional parameters to make a connection
     */
    public ConnectionParameters(String accessProtocol, String deviceContext, String deviceGroup, String deviceName,
            String deviceProperty, AccessMode accessMode, int accessRate, int propertySize, Object dynamicParameters) {
        this(accessProtocol,deviceContext,deviceGroup,deviceName,deviceProperty,accessMode,accessRate,propertySize);
        this.dynamicParameters = dynamicParameters;
    }

    /**
     * Creates new instance of ConnectionParameters.
     * 
     * @param remoteName the remote name (context/server/device/property)
     * @param accessMode the access mode for the connection
     * @param accessRate the polling rate
     */
    public ConnectionParameters(String remoteName, AccessMode accessMode, int accessRate) {
        super();
        setUpRemoteName(remoteName);
        this.accessMode = accessMode;
        this.accessRate = accessRate;
    }

    /**
     * @see #ConnectionParameters(String, AccessMode, int)
     * @param remoteName the remote name (context/server/device/property)
     * @param accessMode the access mode for the connection
     * @param accessRate the polling rate
     * @param dynamicParameters additional parameters to make a connection
     */
    public ConnectionParameters(String remoteName, AccessMode accessMode, int accessRate, Object dynamicParameters) {
        this(remoteName,accessMode,accessRate);
        this.dynamicParameters = dynamicParameters;
    }

    private void setUpRemoteName(String remoteName) {
        String[] s = remoteName.split("/");
        this.accessProtocol = (s.length > 0 && s[0] != null) ? s[0] : "";
        this.deviceContext = (s.length > 1 && s[1] != null) ? s[1] : "";
        this.deviceGroup = (s.length > 2 && s[2] != null) ? s[2] : "";
        this.deviceProperty = (s.length > 4 && s[s.length - 1] != null) ? s[s.length - 1] : "";
        StringBuilder sb = new StringBuilder(128);
        if (s.length > 3 && s[3] != null) {
            sb.append(s[3]);
        }
        for (int i = 4; i < s.length - 1; i++) {
            sb.append('/');
            if (s[i] != null) sb.append(s[i]);
        }
        this.deviceName = sb.toString();
    }

    /**
     * Constructs new ConnectionParameters object.
     * 
     * @param remoteName the remote name (context/server/device/property)
     * @param accessMode the access mode for the connection
     * @param accessRate the polling rate
     * @param arraySize the size of the requested data array
     */
    public ConnectionParameters(String remoteName, AccessMode accessMode, int accessRate, int arraySize) {
        this(remoteName,accessMode,accessRate);
        this.propertySize = arraySize;
    }

    /**
     * @see #ConnectionParameters(String, AccessMode, int, int)
     * @param remoteName the remote name (context/server/device/property)
     * @param accessMode the access mode for the connection
     * @param accessRate the polling rate
     * @param arraySize the size of the requested data array
     * @param dynamicParameters additional parameters to make a connection
     */
    public ConnectionParameters(String remoteName, AccessMode accessMode, int accessRate, int arraySize,
            Object dynamicParameters) {
        this(remoteName,accessMode,accessRate,arraySize);
        this.dynamicParameters = dynamicParameters;
    }

    /**
     * Returns the accessMode.
     * 
     * @return Returns the accessMode.
     */
    public AccessMode getAccessMode() {
        return accessMode;
    }

    /**
     * Returns the accessProtocol.
     * 
     * @return Returns the accessProtocol.
     */
    public String getAccessProtocol() {
        return accessProtocol;
    }

    /**
     * Returns the accessRate.
     * 
     * @return Returns the accessRate.
     */
    public int getAccessRate() {
        return accessRate;
    }

    /**
     * Returns the deviceContext.
     * 
     * @return Returns the deviceContext.
     */
    public String getDeviceContext() {
        return deviceContext;
    }

    /**
     * Returns the deviceGroup.
     * 
     * @return Returns the deviceGroup.
     */
    public String getDeviceGroup() {
        return deviceGroup;
    }

    /**
     * Returns the deviceName.
     * 
     * @return Returns the deviceName.
     */
    public String getDeviceName() {
        return deviceName;
    }

    /**
     * Returns the deviceProperty.
     * 
     * @return Returns the deviceProperty.
     */
    public String getDeviceProperty() {
        return deviceProperty;
    }

    /*
     * (non-Javadoc)
     * @see java.lang.Object#toString()
     */
    @Override
    public String toString() {
        return getRemoteName() + "," + accessMode + "," + accessRate + "," + propertySize;
    }

    /**
     * Return URI ready name for remote TINE property.
     * 
     * @return remote TINE name delimitted with / characters
     */
    public String getRemoteName() {
        return accessProtocol + "/" + deviceContext + "/" + deviceGroup + "/" + deviceName + "/" + deviceProperty;
    }

    /**
     * Returns the dynamic parameters. These are the additional parameters, which might be required by the server to
     * establish a connection. In normal conditions this is null, but it can be something like an int array.
     * 
     * @return the dynamic parameters
     */
    public Object getDynamicParameters() {
        return dynamicParameters;
    }

    /**
     * Size of TINE property. 1 means that property will have one value in array. -1 or 0 means that when connected,
     * automatic or actual property size will be used.
     * 
     * @return size or property if it is array
     * @see #ACTUAL_PROPERTY_SIZE
     * @see #AUTO_PROPERTY_SIZE
     */
    public int getPropertySize() {
        return propertySize;
    }

    /*
     * (non-Javadoc)
     * @see java.lang.Comparable#compareTo(java.lang.Object)
     */
    public int compareTo(ConnectionParameters o) {
        if (o == null) return -1;
        return this.toString().compareTo(o.toString());
    }

    /*
     * (non-Javadoc)
     * @see java.lang.Object#hashCode()
     */
    @Override
    public int hashCode() {
        return toString().hashCode();
    }

    /*
     * (non-Javadoc)
     * @see java.lang.Object#equals(java.lang.Object)
     */
    @Override
    public boolean equals(Object o) {
        if (!(o instanceof ConnectionParameters)) return false;
        return this.toString().equals(o.toString());
        //    ConnectionParameters p = (ConnectionParameters) o;
        //    return (getRemoteName().equals(p.getRemoteName()) && accessMode.equals(p.getAccessMode()) && accessRate == p
        //        .getAccessRate());
    }

    /**
     * Returns new ConnectionParameters which resembles this object but has a different access mode.
     * 
     * @param mode access mode of the new parameters
     * @return new ConnectionParameters object
     */
    public ConnectionParameters deriveWith(AccessMode mode) {
        return new ConnectionParameters(accessProtocol,deviceContext,deviceGroup,deviceName,deviceProperty,mode,
                accessRate,propertySize,dynamicParameters);
    }

    /**
     * Returnst new ConnectionParameters which resembles this object but has a different acces mode and rate.
     * 
     * @param mode access mode of the new parameters
     * @param rate access rate of the new parameters
     * @return new COnnectionParameters object
     */
    public ConnectionParameters deriveWith(AccessMode mode, int rate) {
        return new ConnectionParameters(accessProtocol,deviceContext,deviceGroup,deviceName,deviceProperty,mode,rate,
                propertySize,dynamicParameters);
    }

    /**
     * Returnst new ConnectionParameters which resembles this object but has a different and rate.
     * 
     * @param rate access rate of the new parameters
     * @return new COnnectionParameters object
     * @deprecated use {@link #deriveWithAccessRate(int)} instead
     */
    @Deprecated
    public ConnectionParameters deriveWith(int rate) {
        return deriveWithAccessRate(rate);
    }

    /**
     * Returns new ConnectionParameters which resebles this object but has a different dynamic parameters.
     * 
     * @param newDynamicParameters the new dynamic parameters
     * @return new connection parameters
     */
    public ConnectionParameters deriveWithDynamicParameters(Object newDynamicParameters) {
        return new ConnectionParameters(accessProtocol,deviceContext,deviceGroup,deviceName,deviceProperty,accessMode,
                accessRate,propertySize,newDynamicParameters);
    }

    /**
     * Returns new ConnectionParameters which resembles this object but has a different property size.
     * 
     * @param propertySize the new property size
     * @return new ConnectionParameters object
     */
    public ConnectionParameters deriveWithPropertySize(int propertySize) {
        return new ConnectionParameters(accessProtocol,deviceContext,deviceGroup,deviceName,deviceProperty,accessMode,
                accessRate,propertySize,dynamicParameters);
    }

    /**
     * Constructs a new ConnectionParameters with the same properties as this object but a different access protocol.
     * 
     * @param accessProtocol new access protocol
     * @return new ConnectionParameters object
     */
    public ConnectionParameters deriveWithAccessProtocol(String accessProtocol) {
        return new ConnectionParameters(accessProtocol,deviceContext,deviceGroup,deviceName,deviceProperty,accessMode,
                accessRate,propertySize,dynamicParameters);
    }

    /**
     * Constructs a new ConnectionParameters with the same properties as this object but a different device context.
     * 
     * @param deviceContext new device context
     * @return new ConnectionParameters object
     */
    public ConnectionParameters deriveWithDeviceContext(String deviceContext) {
        return new ConnectionParameters(accessProtocol,deviceContext,deviceGroup,deviceName,deviceProperty,accessMode,
                accessRate,propertySize,dynamicParameters);
    }

    /**
     * Constructs a new ConnectionParameters with the same properties as this object but a different access rate.
     * 
     * @param accessRate new access rate
     * @return new ConnectionParameters object
     */
    public ConnectionParameters deriveWithAccessRate(int accessRate) {
        return new ConnectionParameters(accessProtocol,deviceContext,deviceGroup,deviceName,deviceProperty,accessMode,
                accessRate,propertySize,dynamicParameters);
    }

    /**
     * Constructs a new ConnectionParameters with the same properties as this object but a different device group.
     * 
     * @param deviceGroup new device group
     * @return new ConnectionParameters object
     */
    public ConnectionParameters deriveWithDeviceGroup(String deviceGroup) {
        return new ConnectionParameters(accessProtocol,deviceContext,deviceGroup,deviceName,deviceProperty,accessMode,
                accessRate,propertySize,dynamicParameters);
    }

    /**
     * Constructs a new ConnectionParameters with the same properties as this object but a different device name.
     * 
     * @param deviceName new device name
     * @return new ConnectionParameters object
     */
    public ConnectionParameters deriveWithDeviceName(String deviceName) {
        return new ConnectionParameters(accessProtocol,deviceContext,deviceGroup,deviceName,deviceProperty,accessMode,
                accessRate,propertySize,dynamicParameters);
    }

    /**
     * Constructs a new ConnectionParameters with the same properties as this object but a different property name.
     * 
     * @param deviceProperty new property name
     * @return new ConnectionParameters object
     */
    public ConnectionParameters deriveWithDeviceProperty(String deviceProperty) {
        return new ConnectionParameters(accessProtocol,deviceContext,deviceGroup,deviceName,deviceProperty,accessMode,
                accessRate,propertySize,dynamicParameters);
    }

    /*
     * (non-Javadoc)
     * @see java.lang.Object#clone()
     */
    @Override
    public ConnectionParameters clone() {
        return new ConnectionParameters(accessProtocol,deviceContext,deviceGroup,deviceName,deviceProperty,accessMode,
                accessRate,propertySize,dynamicParameters);
    }
}
