/*
 * Copyright (c) 2006 Stiftung Deutsches Elektronen-Synchroton, Member of the Helmholtz Association, (DESY), HAMBURG,
 * GERMANY. THIS SOFTWARE IS PROVIDED UNDER THIS LICENSE ON AN "../AS IS" BASIS. WITHOUT WARRANTY OF ANY KIND, EXPRESSED
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR PARTICULAR PURPOSE AND
 * NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. SHOULD THE SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE USER
 * ASSUMES THE COST OF ANY NECESSARY SERVICING, REPAIR OR CORRECTION. THIS DISCLAIMER OF WARRANTY CONSTITUTES AN
 * ESSENTIAL PART OF THIS LICENSE. NO USE OF ANY SOFTWARE IS AUTHORIZED HEREUNDER EXCEPT UNDER THIS DISCLAIMER. DESY HAS
 * NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE FULL LICENSE SPECIFYING
 * FOR THE SOFTWARE THE REDISTRIBUTION, MODIFICATION, USAGE AND OTHER RIGHTS AND OBLIGATIONS IS INCLUDED WITH THE
 * DISTRIBUTION OF THIS PROJECT IN THE FILE LICENSE.HTML. IF THE LICENSE IS NOT INCLUDED YOU MAY FIND A COPY AT
 * HTTP://WWW.DESY.DE/LEGAL/LICENSE.HTM
 */
package de.desy.acop.transport.adapters;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import com.cosylab.gui.displayers.CommonDisplayer;
import com.cosylab.gui.displayers.Displayer;

import de.desy.acop.transport.ConnectionFailed;
import de.desy.acop.transport.ConnectionParameters;
import de.desy.tine.definitions.TAccess;
import de.desy.tine.definitions.TArrayType;
import de.desy.tine.definitions.TFormat;
import de.desy.tine.queryUtils.TPropertyQuery;
import de.desy.tine.queryUtils.TQuery;
import de.desy.tine.types.DBLDBL;
import de.desy.tine.types.DBLTIME;
import de.desy.tine.types.FLTFLT;
import de.desy.tine.types.FLTINT;
import de.desy.tine.types.NAME16;
import de.desy.tine.types.NAME32;
import de.desy.tine.types.NAME64;
import de.desy.tine.types.NAME8;
import de.desy.tine.types.USTRING;

/**
 * <code>DataSourceUtilities</code> provides common utilities for data and connection manipulation. This class provides
 * characteristics for a specific channel and provides methods for data conversion used by the
 * <code>AcopTransportDataSource</code>.
 *
 * @author Jaka Bobnar, Cosylab
 */
public class DataSourceUtilities {

    private static Map<String,Map<String,Object>> cache = Collections.synchronizedMap(new HashMap<>());
    private static Map<String,TPropertyQuery[]> cacheInfo = Collections.synchronizedMap(new HashMap<>());

    /**
     * Returns the characteristics for the given property. It returns the cached value if it exists.
     *
     * @param context
     * @param server
     * @param device
     * @param property
     * @return
     * @throws ConnectionFailed
     */
    public static Map<String,Object> getCharacteristics(String context, String server, String device, String property)
            throws ConnectionFailed {
        String remoteName = new StringBuilder(140).append("TINE/").append(context).append('/').append(server)
                .append('/').append(device).append('/').append(property).toString();
        if (cache.containsKey(remoteName)) {
            Map<String,Object> map = cache.get(remoteName);
            if (map == null) {
                throw new ConnectionFailed("Failed to retrieve property info for " + remoteName,null);
            } else {
                return new HashMap<>(map);
            }
        }
        TPropertyQuery[] info = cacheInfo.get(remoteName);
        if (info == null) {
            try {
                info = TQuery.getPropertyInformation(context,server,device,property);
            } catch (RuntimeException e) {
                throw new ConnectionFailed("Failed to retrieve property info for " + remoteName,e);
            }
            if (info == null) {
                cache.put(remoteName,null);
                throw new ConnectionFailed("Failed to retrieve property info for " + remoteName,null);
            }
        }
        Map<String,Object> characteristics = new HashMap<>();
        characteristics.put(CommonDisplayer.C_DESCRIPTION,info[0].prpDescription);
        characteristics.put(CommonDisplayer.C_DISPLAY_NAME,remoteName);
        if (info[0].prpMaxValue != 0 || info[0].prpMinValue != 0) {
            characteristics.put(CommonDisplayer.C_MAXIMUM,info[0].prpMaxValue);
            characteristics.put(CommonDisplayer.C_MINIMUM,info[0].prpMinValue);
        }
        characteristics.put("dataFormat",TFormat.valueOf(info[0].prpFormat));
        characteristics.put(CommonDisplayer.C_FORMAT,getFormat(info[0].prpFormat));
        characteristics.put(CommonDisplayer.C_UNITS,info[0].prpUnits);
        characteristics.put(CommonDisplayer.C_SEQUENCE_LENGTH,info[0].prpSize);
        characteristics.put("redirection",info[0].prpRedirection);
        characteristics.put("tag",info[0].prpTag);
        characteristics.put("rows",(int)info[0].numRows);
        characteristics.put("rowSize",(int)info[0].rowSize);
        characteristics.put("access",TAccess.valueOf(info[0].prpAccess));
        characteristics.put(Displayer.C_EDITABLE,TAccess.valueOf(info[0].prpAccess).isWrite());
        characteristics.put("rangeUnits",info[0].rngUnits);
        characteristics.put("rangeMin",info[0].rngMinValue);
        characteristics.put("rangeMax",info[0].rngMaxValue);
        characteristics.put("graphType", info[0].prpGraphType);
        TArrayType arrayType = TArrayType.valueOf(info[0].prpArrayType);
        if (arrayType.isChannel()) {
            String[] names = TQuery.getDeviceNames(context,server,property);
            if (names != null) {
                String[] newNames = new String[names.length];
                int j = -1;
                int startPoint = -1;
                for (int i = 0; i < 2 * names.length; i++) {
                    if (i % names.length == startPoint) break;
                    if (names[i % names.length] == null) {
                        names[i % names.length] = "";
                    }
                    if (names[i % names.length].equals(device)) {
                        j = 0;
                        startPoint = i;
                    }
                    if (j >= 0) {
                        newNames[j] = names[i % names.length];
                        j++;
                    }
                }
                characteristics.put("xLabels",newNames);
            }
        }
        //DOOCS sequence and names
        if (device.length() == 1 && device.charAt(0) == '*') {
            String[] names = (String[])characteristics.get("xLabels");
            if (names == null) {
                names = TQuery.getDeviceNames(context,server,property);
                characteristics.put("xLabels",names);
                characteristics.put(CommonDisplayer.C_SEQUENCE_LENGTH,names.length);
            }
        }
        //first fallback
        //TODO this does not work OK - if type is a spectrum the returned names are empty strings
        //          if (characteristics.get("xLabels") == null) {
        //              AcopTransport atr = new AcopTransport();
        //              atr.setConnectionParameters(param);
        //              USTRING[] data = new USTRING[(Integer)characteristics.get(CommonDisplayer.C_SEQUENCE_LENGTH)];
        //              characteristics.put("xLabels", atr.getXAxisChannelNames(data));
        //          }
        //second fallback
        String[] names = (String[])characteristics.get("xLabels");
        if (names == null && info[0].prpFormat != TFormat.CF_IMAGE) {
            try {
                names = TQuery.getDeviceNames(context,server,property);
                if (names != null && names.length == 1 && names[0].equalsIgnoreCase("keyword")) {
                    names = null;
                } else if (names != null) {
                    characteristics.put("xLabels",names);
                    characteristics.put(CommonDisplayer.C_SEQUENCE_LENGTH,names.length);
                }
            } catch (Exception e) {
                //ignore
            }
            if (names == null) {
                String[] data = new String[(Integer)characteristics.get(CommonDisplayer.C_SEQUENCE_LENGTH)];
                for (int i = 0; i < data.length; i++) {
                    data[i] = '#' + String.valueOf(i);
                }
                characteristics.put("xLabels",data);
            }
        }
        characteristics.put("arrayType",arrayType);
        characteristics.put("isChannel",arrayType.isChannel());
        characteristics.put("TPropertyQuery",info[0]);
        characteristics.put("TPropertyQueries",info);
        cacheInfo.put(remoteName,info);
        cache.put(remoteName,characteristics);
        return new HashMap<>(characteristics);
    }

    /**
     * Returns the characteristics of the specified remote channel.
     *
     * @param param remote channel specification
     * @return map of all characteristics
     * @throws ConnectionFailed if characteristics query failed
     */
    public static Map<String,Object> getCharacteristics(ConnectionParameters param) throws ConnectionFailed {
        return getCharacteristics(param.getDeviceContext(),param.getDeviceGroup(),param.getDeviceName(),
                param.getDeviceProperty());
    }

    /**
     * Returns the description of the remote channel.
     *
     * @param param remote channel specification
     * @return the description
     * @throws ConnectionFailed if property information query failed
     */
    public static String getDescription(ConnectionParameters param) throws ConnectionFailed {
        return getDescription(param.getDeviceContext(),param.getDeviceGroup(),param.getDeviceName(),
                param.getDeviceProperty());
    }

    /**
     * Returns the description for the specified property. It returns the cached version if it exists.
     *
     * @param context
     * @param server
     * @param device
     * @param property
     * @return the descriptionr
     * @throws ConnectionFailed if property info could not be read
     */
    public static String getDescription(String context, String server, String device, String property)
            throws ConnectionFailed {
        return getPropertyQuery(context,server,device,property)[0].prpDescription;
    }

    /**
     * Returns the property query for the given property. Cached value is returned if it exists.
     * 
     * @param context
     * @param server
     * @param device
     * @param property
     * @return
     * @throws ConnectionFailed
     */
    public static TPropertyQuery[] getPropertyQuery(String context, String server, String device, String property)
            throws ConnectionFailed {
        String remoteName = new StringBuilder(140).append("TINE/").append(context).append('/').append(server)
                .append('/').append(device).append('/').append(property).toString();
        TPropertyQuery[] info = cacheInfo.get(remoteName);
        if (info == null) {
            info = TQuery.getPropertyInformation(context,server,device,property);
        }
        if (info == null) {
            throw new ConnectionFailed("Failed to retrieve property info for " + remoteName,null);
        }
        cacheInfo.put(remoteName,info);
        return info;
    }

    /**
     * Converts TINE timestamp to UTC in millis.
     *
     * @param stamp TINE timestamp
     * @return UTC in milis
     */
    static long toUTC(double stamp) {
        return (long)(stamp * 1E3);
    }

    /**
     * Returns the java object of the given size which corresponds to the specified TFormat.
     *
     * @param format TINE format of the data
     * @param size the size of the object
     * @return data object
     */
    public static Object getDataObject(TFormat format, int size) {
        if (format == null) return new double[size];
        switch (format.getValue()) {
            case TFormat.CF_DOUBLE:
                return new double[size];
            case TFormat.CF_INT16:
                return new short[size];
            case TFormat.CF_BYTE:
                return new byte[size];
            case TFormat.CF_INT32:
                return new int[size];
            case TFormat.CF_FLOAT:
                return new float[size];
            case TFormat.CF_INT64:
                return new long[size];
            case TFormat.CF_BIT:
                return new boolean[size];
            case TFormat.CF_SPECTRUM:
                return new double[size];
            case TFormat.CF_USTRING:
                return new USTRING[size];
            case TFormat.CF_NAME64:
                return new NAME64[size];
            case TFormat.CF_NAME32:
                return new NAME32[size];
            case TFormat.CF_NAME16:
                return new NAME16[size];
            case TFormat.CF_NAME8:
                return new NAME8[size];
            case TFormat.CF_STRING:
                return new String[size];
            case TFormat.CF_FLTFLT:
                return new FLTFLT[size];
            case TFormat.CF_FLTINT:
                return new FLTINT[size];
            case TFormat.CF_DBLDBL:
                return new DBLDBL[size];
            case TFormat.CF_DBLTIME:
                return new DBLTIME[size];
            default:
                return new double[size];
        }
    }

    /**
     * Extracts data from the <code>inData</code> which match the TINE format <code>format</code>. Method rewrites data
     * to the <code>data</code> parameter and returns this object. The <code>data</code> parameter should be the same
     * size (or larger) than the <code>inData</code>.
     *
     * @param format format of the incomming data
     * @param inData incomming data
     * @param data converted data
     * @return converted data
     */
    static double[] extractData(TFormat format, Object inData, double[] data) {
        switch (format.getValue()) {
            case TFormat.CF_DOUBLE:
                return (double[])inData;
            case TFormat.CF_SHORT: {
                short[] d = (short[])inData;
                for (int i = 0; i < d.length; i++) {
                    data[i] = d[i];
                }
                return data;
            }
            case TFormat.CF_BYTE: {
                byte[] d = (byte[])inData;
                for (int i = 0; i < d.length; i++) {
                    data[i] = d[i];
                }
                return data;
            }
            case TFormat.CF_INT32: {
                int[] d = (int[])inData;
                for (int i = 0; i < d.length; i++) {
                    data[i] = d[i];
                }
                return data;
            }
            case TFormat.CF_INT64: {
                long[] d = (long[])inData;
                for (int i = 0; i < d.length; i++) {
                    data[i] = d[i];
                }
                return data;
            }
            case TFormat.CF_FLOAT: {
                float[] d = (float[])inData;
                for (int i = 0; i < d.length; i++) {
                    data[i] = Double.parseDouble(Float.toString(d[i]));
                }
                return data;
            }
            case TFormat.CF_USTRING: {
                USTRING[] d = (USTRING[])inData;
                for (int i = 0; i < d.length; i++) {
                    data[i] = d[i].f1val;
                }
                return data;
            }
            default:
                return (double[])inData;
        }
    }

    /**
     * Extracts data from the <code>inData</code> which match the TINE dual format <code>format</code>. Method rewrites
     * data to the <code>primary data</code> and <code>secondary data</code> parameter. The parameters should be the
     * same size (or larger) than the <code>inData</code>.
     *
     * @param format format of the incomming data
     * @param inData incomming data
     * @param primaryData the primary value of the dual data
     * @param secondaryData the secondary value of the dual data
     */
    static void extractData(TFormat format, Object inData, double[] primaryData, double[] secondaryData) {
        switch (format.getValue()) {
            case TFormat.CF_DBLDBL:
                DBLDBL[] dbldbldata = (DBLDBL[])inData;
                for (int i = 0; i < dbldbldata.length; i++) {
                    primaryData[i] = dbldbldata[i].d1val;
                    secondaryData[i] = dbldbldata[i].d2val;
                }
                break;
            case TFormat.CF_DBLTIME: {
                DBLTIME[] dbltimedata = (DBLTIME[])inData;
                for (int i = 0; i < dbltimedata.length; i++) {
                    primaryData[i] = dbltimedata[i].d1val;
                    secondaryData[i] = dbltimedata[i].d2val;
                }
                break;
            }
            case TFormat.CF_FLTFLT: {
                FLTFLT[] fltfltdata = (FLTFLT[])inData;
                for (int i = 0; i < fltfltdata.length; i++) {
                    primaryData[i] = fltfltdata[i].f1val;
                    secondaryData[i] = fltfltdata[i].f2val;
                }
                break;
            }
            case TFormat.CF_FLTINT: {
                FLTINT[] fltintdata = (FLTINT[])inData;
                for (int i = 0; i < fltintdata.length; i++) {
                    primaryData[i] = fltintdata[i].fval;
                    secondaryData[i] = fltintdata[i].ival;
                }
                break;
            }
        }
    }

    /**
     * Extracts data from the <code>inData</code> which match the TINE format <code>format</code>. Method rewrites data
     * to the <code>data</code> parameter and returns this object. The <code>data</code> parameter should be the same
     * size (or larger) than the <code>inData</code>. This method handles String types of converted data.
     *
     * @param format format of the incomming data
     * @param inData incomming data
     * @param data converted data
     * @return converted data
     */
    static String[] extractData(TFormat format, Object inData, String[] data) {
        switch (format.getValue()) {
            case TFormat.CF_USTRING: {
                USTRING[] d = (USTRING[])inData;
                for (int i = 0; i < d.length; i++) {
                    data[i] = d[i].f1val + "," + d[i].f2val + "," + d[i].ival + "," + d[i].str.trim() + "," + d[i].tm;
                }
                return data;
            }
            default: {
                Object[] d = (Object[])inData;
                for (int i = 0; i < d.length; i++) {
                    data[i] = String.valueOf(d[i]);
                }
                return data;
            }
        }
    }

    /**
     * Transforms the double array to String array.
     *
     * @param data
     * @return
     */
    static String[] toStringData(double[] data) {
        String[] s = new String[data.length];
        for (int i = 0; i < data.length; i++) {
            s[i] = String.valueOf(data[i]);
        }
        return s;
    }

    /**
     * Returns the C-type String format that match the given <code>format</code> parameter. This parameter is the short
     * code of the <code>TFormat</code>.
     *
     * @param format
     * @return
     */
    private static String getFormat(short format) {
        switch (format) {
            case TFormat.CF_DOUBLE:
                return "%3.2f";
            case TFormat.CF_FLOAT:
                return "%3.2f";
            case TFormat.CF_LONG:
                return "%d";
            case TFormat.CF_SHORT:
                return "%d";
            default:
                return null;
        }
    }

    /**
     * Clears the cache.
     */
    public static void clearCache() {
        cache.clear();
    }
}
/* __oOo__ */
