/*
 * Copyright (c) 2006 Stiftung Deutsches Elektronen-Synchroton,
 * Member of the Helmholtz Association, (DESY), HAMBURG, GERMANY.
 *
 * THIS SOFTWARE IS PROVIDED UNDER THIS LICENSE ON AN "../AS IS" BASIS.
 * WITHOUT WARRANTY OF ANY KIND, EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR PARTICULAR PURPOSE AND
 * NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE. SHOULD THE SOFTWARE PROVE DEFECTIVE
 * IN ANY RESPECT, THE USER ASSUMES THE COST OF ANY NECESSARY SERVICING, REPAIR OR
 * CORRECTION. THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF THIS LICENSE.
 * NO USE OF ANY SOFTWARE IS AUTHORIZED HEREUNDER EXCEPT UNDER THIS DISCLAIMER.
 * DESY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS,
 * OR MODIFICATIONS.
 * THE FULL LICENSE SPECIFYING FOR THE SOFTWARE THE REDISTRIBUTION, MODIFICATION,
 * USAGE AND OTHER RIGHTS AND OBLIGATIONS IS INCLUDED WITH THE DISTRIBUTION OF THIS
 * PROJECT IN THE FILE LICENSE.HTML. IF THE LICENSE IS NOT INCLUDED YOU MAY FIND A COPY
 * AT HTTP://WWW.DESY.DE/LEGAL/LICENSE.HTM
 */

package de.desy.acop.transport.adapters;

/*
 * Copyright (c) 2004 by Cosylab d.o.o.
 *
 * The full license specifying the redistribution, modification, usage and other
 * rights and obligations is included with the distribution of this project in
 * the file license.html. If the license is not included you may find a copy at
 * http://www.cosylab.com/legal/abeans_license.htm or may write to Cosylab, d.o.o.
 *
 * THIS SOFTWARE IS PROVIDED AS-IS WITHOUT WARRANTY OF ANY KIND, NOT EVEN THE
 * IMPLIED WARRANTY OF MERCHANTABILITY. THE AUTHOR OF THIS SOFTWARE, ASSUMES
 * _NO_ RESPONSIBILITY FOR ANY CONSEQUENCE RESULTING FROM THE USE, MODIFICATION,
 * OR REDISTRIBUTION OF THIS SOFTWARE.
 */

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import de.desy.acop.transport.ConnectionFailed;
import de.desy.acop.transport.ConnectionParameters;


/**
 * Default implementaion of the AdapterFactory.
 * 
 * @author Jaka Bobnar, Cosylab
 *
 */
public class DefaultAdapterFactory implements AdapterFactory
{
	private Map<ConnectionParameters, AcopTransportDataSource> adapters = new HashMap<>();
	private boolean useCache = true;

	protected DefaultAdapterFactory()
	{
		super();
	}
	
	@Override
	public AcopTransportDataSource createDataSource(ConnectionParameters param, Map<String,Object> characteristics) throws ConnectionFailed {
		if (useCache) {
			AcopTransportDataSource ad = adapters.get(param);

			if (ad == null) {
				ad = createAdapter(param,characteristics);
				adapters.put(param, ad);
			}

			ad.useCount++;

			return ad;
		}

		return createAdapter(param,characteristics);
	}
	
	@Override
	public AcopTransportDataSource getDataSource(ConnectionParameters param, Map<String,Object> characteristics) throws ConnectionFailed {
		if (useCache) {
			AcopTransportDataSource ad = adapters.get(param);

			if (ad == null) {
				ad = createDataSource(param);
			}

			return ad;
		}

		return createAdapter(param,characteristics);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.adapters.dal.AdapterFactory#createDataSource(abeans.datatypes.DynamicValueProperty)
	 */
	public synchronized AcopTransportDataSource createDataSource(
			ConnectionParameters param) throws ConnectionFailed
	{
		return createDataSource(param,null);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.adapters.dal.AdapterFactory#getDataSource(abeans.datatypes.DynamicValueProperty)
	 */
	public synchronized AcopTransportDataSource getDataSource(
			ConnectionParameters param) throws ConnectionFailed
	{
		return getDataSource(param,null);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.adapters.dal.AdapterFactory#releaseDataSource(com.cosylab.adapters.dal.DynamicValuePropertyAdapter)
	 */
	public synchronized void releaseDataSource(AcopTransportDataSource ds)
	{
		ds.useCount--;

		if (ds.useCount <= 0 || !useCache) {
			adapters.remove(ds.getConnectionParameters());
			ds.disconnect();
		}
	}
	
	/*
	 * (non-Javadoc)
	 * @see de.desy.acop.transport.adapters.AdapterFactory#releaseDataSource(de.desy.acop.transport.ConnectionParameters)
	 */
	public synchronized AcopTransportDataSource releaseDataSource(ConnectionParameters param) {
		AcopTransportDataSource ds = adapters.get(param);
		
		ds.useCount--;

		if (ds.useCount <= 0 || !useCache) {
			adapters.remove(param);
			ds.disconnect();
		}
		
		return ds;
	    
    }

	/* (non-Javadoc)
	 * @see com.cosylab.adapters.dal.AdapterFactory#clear()
	 */
	public synchronized void clear()
	{
		Iterator<AcopTransportDataSource> it = adapters.values().iterator();

		while (it.hasNext()) {
			AcopTransportDataSource element = it.next();
			element.disconnect();
		}

		adapters.clear();
	}

	protected AcopTransportDataSource createAdapter(ConnectionParameters param, Map<String,Object> characteristics)
		throws ConnectionFailed
	{
		if (param == null) {
			throw new NullPointerException("Connection Parameters cannot be null.");
		}
		AcopTransportDataSource retVal;
		if (characteristics == null || characteristics.size() == 0) {
			retVal = new AcopTransportDataSource();
		} else {
			retVal = new AcopTransportDataSource(characteristics);
		}
		retVal.connect(param);
		
		return retVal;
	}

	/**
	 * Returns true if caching is enabled.
	 * 
	 * @return
	 * @see #setUseCache(boolean)
	 */
	public boolean isUseCache()
	{
		return useCache;
	}

	/**
	 * Enables/disables caching of adapters for multiple use. If adapters are
	 * cached there is only one adapter for one ConnectionParameters. If displayers
	 * are connected to the same channel, they share the adapter.
	 *
	 * @param useCache 
	 */
	public void setUseCache(boolean useCache)
	{
		this.useCache = useCache;
	}

	/*
	 * (non-Javadoc)
	 * @see de.desy.acop.transport.adapters.AdapterFactory#getCharacteristics(de.desy.acop.transport.ConnectionParameters)
	 */
	public Map<String, Object> getCharacteristics(ConnectionParameters param) throws ConnectionFailed {
		if (param == null) return new HashMap<>(0);
		AcopTransportDataSource ds = adapters.get(param);
		if (ds != null) {
			return ds.getCharacteristics();
		}
		return DataSourceUtilities.getCharacteristics(param);
	}

	
}

/* __oOo__ */
