package de.desy.acop.video;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Desktop;
import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.WindowListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.net.URI;
import java.net.URL;

import javax.swing.BorderFactory;
import javax.swing.GroupLayout;
import javax.swing.GroupLayout.Alignment;
import javax.swing.GroupLayout.ParallelGroup;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.SwingConstants;
import javax.swing.SwingUtilities;
import javax.swing.SwingWorker;
import javax.swing.UIManager;
import javax.swing.border.BevelBorder;

import com.cosylab.gui.components.customizer.AbstractCustomizerPanel;
import com.cosylab.gui.components.numberfield.AbstractNumberDocument;
import com.cosylab.gui.components.util.IconHelper;
import com.cosylab.gui.property.editors.IntegerEditor;

import de.desy.acop.transport.AccessMode;
import de.desy.acop.transport.ConnectionParameters;
import de.desy.acop.video.displayer.ColorMap;
import de.desy.acop.video.displayer.ImageCLUT;
import de.desy.acop.video.displayer.ImageDisplayer;
import de.desy.acop.video.displayer.ImageOrientation;
import de.desy.acop.video.displayer.ImageZoom;
import de.desy.acop.video.displayer.OverlayState;
import de.desy.tine.queryUtils.TPropertyQuery;
import de.desy.tine.queryUtils.TQuery;
import de.desy.tine.startup.TInitializerFactory;

/**
 * 
 * <code>AcopVideoApplication</code> is a demo application of the AcopVideo. It encapsulates the AcopVideo and offers a
 * simple GUI for manipulation of the AcopVideo. It allows setting of all its properties and setting up a connection to
 * a remote TINE point.
 * 
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 * @author <a href="mailto:david.melkumyan@desy.de">David Melkumyan</a>
 * 
 */
public class AcopVideoApplication extends JPanel {

	private static final long serialVersionUID = 1L;

	private static final String JARG_ACOPVIDEO_CONNECTION_ADDRESS = "acopvideo.connection.address";
	private static final String JARG_APP_URL = "app.url";
	private static final String JARG_LAF_DEFAULT = "laf.default";

	private static final String DEFAULT_APP_URL = "http://cosylib.cosylab.com/pub/acop/site/AcopVideoManual.html";
	private static final String DEFAULT_LAF = "false";

	/**
	 * the main control panel with all the buttons and selectors
	 */
	private JPanel controlPanel;

	/**
	 * Color map chooser
	 */
	private JComboBox<ColorMap> colorMapBox;

	/**
	 * Image zoom chooser
	 */
	private JComboBox<String> imageZoomBox;

	/**
	 * Image orientation chooser
	 */
	private JComboBox<String> imageOrientationBox;

	/**
	 * Overlay panel
	 */
	private JComboBox<OverlayState> overlayStateBox;

	/**
	 * AOIZoom switch
	 */
	private JCheckBox aoiZoomCheckBox;

	/**
	 * keepAspectRatio switch
	 */
	private JCheckBox aspectRatioCheckBox;

	/**
	 * histogramEqualisation switch
	 */
	private JCheckBox normalizationCheckBox;

	/**
	 * histogram min text field
	 */
	private IntegerEditor histogramMinTF;

	/**
	 * histogram min text field
	 */
	private IntegerEditor histogramMaxTF;

	/**
	 * address field describe the remote connection point
	 */
	private JTextField addressTF;

	/**
	 * address button opens the connection customizer
	 */
	private JButton browseButton;

	/**
	 * starts the data acquisition
	 */
	private JButton startButton;

	/**
	 * stops the data acquisition
	 */
	private JButton stopButton;

	/**
	 * saves current image as PNG
	 */
	private JButton saveAsPNGButton;

	/**
	 * exports current image to PNG
	 */
	private JButton exportToPngButton;

	/**
	 * opens the help page
	 */
	private JButton helpButton;

	/**
	 * the image displayer
	 */
	private AcopVideo acopVideo;

	/**
	 * address text field focus listener
	 */
	private WindowListener addressFocusListener;

	/**
	 * 
	 */
	private Window lastWindow;

	/**
	 * file chooser for image saving
	 */
	private boolean isLiveMode;

	private boolean changingAOIZoom;
	private boolean changingAspectRatio;
	private boolean changingNormalization;
	private boolean changingHistogramMin;
	private boolean changingHistogramMax;
	private boolean changingImageZoom;
	private boolean changingImageOrientation;
	private boolean changingOverlayState;
	private boolean changingColorMap;
	private boolean changingLiveMode;
	private boolean changingDisplayClutHeader;
	private boolean changingAddress;

	/**
	 * Constructs a new AcopVideoApplication.
	 * 
	 */
	public AcopVideoApplication() {
		super();
		initialize();
	}

	private void initialize() {
		this.setLayout(new BorderLayout());

		acopVideo = new AcopVideo();
		acopVideo.addPropertyChangeListener(new PropertyChangeListener() {
			public void propertyChange(PropertyChangeEvent e) {
//				System.err.println(" ############ [acopVideo]: " + e.getPropertyName() +
//				 " changed from '" + e.getOldValue()
//				 + "' to '" + e.getNewValue() + "'");
				acopVideoPropertyChanged(e.getPropertyName());
			}
		});

		acopVideo.getImageDisplayer().addPropertyChangeListener(new PropertyChangeListener() {
			public void propertyChange(PropertyChangeEvent e) {
//				System.err.println("############ [imageDisplayer]: " + e.getPropertyName()
//				 + " changed from '" + e.getOldValue()
//				 + "' to '" + e.getNewValue() + "'");
				imageDisplayerPropertyChanged(e.getPropertyName());
			}
		});

		this.add(acopVideo, BorderLayout.CENTER);
		this.add(getControlPanel(), BorderLayout.SOUTH);
		getControlPanel().addMouseListener(acopVideo.getPopupManager().getMouseHook());
	}

	/**
	 * ImageDisplayer property changed
	 * 
	 * @author mdavid
	 * 
	 * @param propertyChanged
	 */
	private void imageDisplayerPropertyChanged(String propertyChanged) {
		if (propertyChanged.equals(ImageDisplayer.PROPERTY_AOI_ZOOM) && !changingAOIZoom) {
			changingAOIZoom = true;
			aoiZoomCheckBox.setSelected(acopVideo.isAOIZoom());
			changingAOIZoom = false;

		} else if (propertyChanged.equals(ImageDisplayer.PROPERTY_KEEP_ASPECT_RATIO) && !changingAspectRatio) {
			changingAspectRatio = true;
			aspectRatioCheckBox.setSelected(acopVideo.isKeepAspectRatio());
			changingAspectRatio = false;

		} else if (propertyChanged.equals(ImageDisplayer.PROPERTY_HISTOGRAM_EQUALISATION) && !changingNormalization) {
			changingNormalization = true;
			normalizationCheckBox.setSelected(acopVideo.isHistogramEqualisation());
			changingNormalization = false;

		} else if (propertyChanged.equals(ImageDisplayer.PROPERTY_HISTOGRAM_MIN) && !changingHistogramMin) {
			changingHistogramMin = true;
			int value = acopVideo.getHistogramMin();
			histogramMinTF.setMaxValue(value);
			histogramMinTF.setValue(value);
			changingHistogramMin = false;

		} else if (propertyChanged.equals(ImageDisplayer.PROPERTY_HISTOGRAM_MAX) && !changingHistogramMax) {
			changingHistogramMax = true;
			int value = acopVideo.getHistogramMax();
			histogramMaxTF.setMaxValue(acopVideo.getImageDisplayer().getCLUT().getMaxValue());
			histogramMaxTF.setValue(value);
			changingHistogramMax = false;

		} else if (propertyChanged.equals(ImageDisplayer.PROPERTY_COLOR_MAP) && !changingColorMap) {
			changingColorMap = true;
			ImageCLUT clut = acopVideo.getImageDisplayer().getCLUT();
			if (((AbstractNumberDocument) histogramMaxTF.getDocument()).getMaxValue().intValue() < clut.getMaxValue()) {
				histogramMinTF.setMaxValue(clut.getMaxValue());
				histogramMaxTF.setMaxValue(clut.getMaxValue());
				
	            histogramMaxTF.setValue(acopVideo.getHistogramMax());
	            histogramMinTF.setValue(acopVideo.getHistogramMin());
			}
			colorMapBox.setEnabled(clut.getColorMap() != null && !ColorMap.NONE.equals(clut.getColorMap()));
			colorMapBox.setSelectedItem(clut.getColorMap());
			changingColorMap = false;

		} else if (propertyChanged.equals(ImageDisplayer.PROPERTY_IMAGE_ZOOM) && !changingImageZoom) {
			changingImageZoom = true;
			imageZoomBox.setSelectedItem(acopVideo.getImageZoom().getDescription());
			changingImageZoom = false;

		} else if (propertyChanged.equals(ImageDisplayer.PROPERTY_IMAGE_ORIENTATION) && !changingImageOrientation) {
			changingImageOrientation = true;
			imageOrientationBox.setSelectedItem(acopVideo.getImageOrientation().getDescription());
			changingImageOrientation = false;

		} else if (propertyChanged.equals(ImageDisplayer.PROPERTY_OVERLAY_STATE) && !changingOverlayState) {
			changingOverlayState = true;
			overlayStateBox.setSelectedItem(acopVideo.getOverlayState());
			changingOverlayState = false;
		}
	}

	/**
	 * AcopVideo property changed
	 * 
	 * @author mdavid
	 * 
	 * @param propertyChanged
	 */
	private void acopVideoPropertyChanged(String propertyChanged) {
		if (propertyChanged.equals(AcopVideo.PROPERTY_LIVE_MODE) && !changingLiveMode) {
			changingLiveMode = true;
			setLiveMode(acopVideo.isLiveMode());
			changingLiveMode = false;

		} else if (propertyChanged.equals(AcopVideo.PROPERTY_DISPLAY_CLUT_HEADER) && !changingDisplayClutHeader) {
			changingDisplayClutHeader = true;
			setDisplayClutHeader(acopVideo.isDisplayClutHeader());
			changingDisplayClutHeader = false;

		} else if (propertyChanged.equals(AcopVideo.CONNECTION_PARAMETERS_PROPERTY) && !changingAddress) {
			changingAddress = true;
			ConnectionParameters cp = acopVideo.getConnectionParameters();
			addressTF.setText(cp == null ? "" : //
					("/" + cp.getDeviceContext() //
							+ "/" + cp.getDeviceGroup() //
							+ "/" + cp.getDeviceName() //
							+ "/" + cp.getDeviceProperty()));
			changingAddress = false;
		}
	}

	/**
	 * Construct the main control panel.
	 * 
	 * @return
	 */
	private JPanel getControlPanel() {
		if (controlPanel == null) {
			controlPanel = new JPanel();
			controlPanel.setBorder(BorderFactory.createBevelBorder(BevelBorder.RAISED));

			GroupLayout layout = new GroupLayout(controlPanel);
			controlPanel.setLayout(layout);
			layout.setAutoCreateContainerGaps(true);
			layout.setAutoCreateGaps(true);

			addressFocusListener = new java.awt.event.WindowAdapter() {
				@Override
				public void windowOpened(java.awt.event.WindowEvent se) {
					addressTF.requestFocus();
				}
			};

			Component connPanel = getAddressPanel();
			Component configPanel = getConfigurationPanel();
			Component filePanel = getFilePanel();
			Component helpPanel = getHelpPanel();

			ParallelGroup hg = layout.createParallelGroup();
			layout.setHorizontalGroup(hg);

			hg //
				// .addComponent(connPanel) //
					.addGroup(layout.createSequentialGroup().addComponent(connPanel, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE, Short.MAX_VALUE)
							.addComponent(helpPanel, 0, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE))
					.addGroup(layout.createSequentialGroup() //
							.addComponent(configPanel, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE, Short.MAX_VALUE) //
							.addComponent(filePanel));

			GroupLayout.SequentialGroup vg = layout.createSequentialGroup();
			layout.setVerticalGroup(vg);
			vg //
				// .addComponent(connPanel) //
					.addGroup(layout.createParallelGroup().addComponent(connPanel, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE, Short.MAX_VALUE)
							.addComponent(helpPanel, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE, Short.MAX_VALUE))
					.addGroup(layout.createParallelGroup() //
							.addComponent(configPanel) //
							.addComponent(filePanel, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE, Short.MAX_VALUE));

		}
		return controlPanel;
	}

	/**
	 * @return Address Panel
	 */
	private Component getAddressPanel() {
		JLabel addressLabel = new JLabel("Address:");

		addressTF = new JTextField();

		int height = addressTF.getPreferredSize().height - 1;

		URL url = IconHelper.class.getClassLoader().getResource("icons/media/Play16.gif");
		if (url != null)
			startButton = new JButton(new ImageIcon(url));
		else
			startButton = new JButton("Start");
		startButton.setToolTipText("Start video");
		startButton.setPreferredSize(new Dimension(0, height));
		startButton.setEnabled(false);
		startButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				startVideo();
			}
		});

		url = IconHelper.class.getClassLoader().getResource("icons/media/Stop16.gif");
		if (url != null)
			stopButton = new JButton(new ImageIcon(url));
		else
			stopButton = new JButton("Stop");
		stopButton.setToolTipText("Stop video");
		stopButton.setPreferredSize(new Dimension(0, height));
		stopButton.setEnabled(false);
		stopButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				acopVideo.stop();
			}
		});

		browseButton = new JButton("...");
		browseButton.setToolTipText("Browse...");
		browseButton.setPreferredSize(new Dimension(height, height));
		browseButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				AbstractCustomizerPanel customizer = acopVideo.getCustomizer();
				customizer.showDialog();
				customizer.selectAspect("Connections");
			}
		});

		addressTF.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				if (startButton.isEnabled())
					startVideo();
			}
		});

		addressTF.getDocument().addDocumentListener(new javax.swing.event.DocumentListener() {

			/**
			 * Address validation swing worker
			 */
			private SwingWorker<Void, Void> validator;

			public void changedUpdate(javax.swing.event.DocumentEvent e) {
				addressChanged();
			}

			public void removeUpdate(javax.swing.event.DocumentEvent e) {
				addressChanged();
			}

			public void insertUpdate(javax.swing.event.DocumentEvent e) {
				addressChanged();
			}

			private void addressChanged() {
				if (validator != null && !validator.isDone()) {
					validator.cancel(true);
					validator = null;
				}

				validator = new SwingWorker<Void, Void>() {
					String protocol = "";
					String context = "";
					String server = "";
					String device = "";
					String property = "";

					@Override
					protected Void doInBackground() throws Exception {
						try {
							startButton.setEnabled(validate(addressTF.getText()));

						} catch (Exception e) {
							startButton.setEnabled(false);
							System.err.println(e);
						}

						if (startButton.isEnabled() && !changingAddress) {
							changingAddress = true;
							try {
								ConnectionParameters cp = acopVideo.getConnectionParameters();
								if (cp != null) {
									cp = new ConnectionParameters(protocol, context, server, device, property, //
											cp.getAccessMode(), //
											cp.getAccessRate(), //
											cp.getPropertySize(), //
											cp.getDynamicParameters());
								} else {
									cp = new ConnectionParameters(protocol, context, server, device, property, AccessMode.POLL, //
											1000);
								}
								acopVideo.setConnectionParameters(cp);

							} catch (Exception ignored) {
							}
							changingAddress = false;
						}
						return null;
					}

					/**
					 * Validates TINE address.
					 * 
					 * @param address
					 *            - Tine address to validate
					 * @return true if address is valid and accessible, otherwise - false
					 */
					private boolean validate(String address) {
						if (address == null || (address = address.trim()).isEmpty())
							return false;

						String[] split = address.split("/");
						switch (split.length) {
						case 3:
							server = split[0].trim();
							device = split[1].trim();
							property = split[2].trim();
							break;
						case 4:
							context = split[0].trim();
							server = split[1].trim();
							device = split[2].trim();
							property = split[3].trim();
							break;
						case 5:
							protocol = split[0].trim();
							context = split[1].trim();
							server = split[2].trim();
							device = split[3].trim();
							property = split[4].trim();
							break;
						default:
							return false;
						}

						if (server.isEmpty() || device.isEmpty() || property.isEmpty())
							return false;

						if (protocol.isEmpty())
							protocol = "TINE";
						if (context.isEmpty())
							context = "DEFAULT";

						TPropertyQuery[] queries = TQuery.getPropertyInformation(context, server, device, property);
						if (queries == null || queries.length == 0)
							return false;

						return true;
					}
				};
				validator.execute();
			}

		});

		String address = System.getProperty(JARG_ACOPVIDEO_CONNECTION_ADDRESS);
		if (address != null) {
			addressTF.setText(address);
		}

		JPanel panel = new JPanel(new BorderLayout());
		panel.setBorder(BorderFactory.createTitledBorder("Connection"));
		panel.setOpaque(false);

		GroupLayout layout = new GroupLayout(panel);
		panel.setLayout(layout);
		layout.setAutoCreateContainerGaps(true);
		layout.setAutoCreateGaps(true);

		GroupLayout.SequentialGroup hg = layout.createSequentialGroup();
		layout.setHorizontalGroup(hg);
		hg.addComponent(addressLabel).addComponent(addressTF). //
				addGap(0, 0, 0).addComponent(browseButton, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE). //
				addComponent(startButton). //
				addGap(1, 1, 1).addComponent(stopButton);

		GroupLayout.ParallelGroup vg = layout.createParallelGroup(Alignment.CENTER);
		layout.setVerticalGroup(vg);
		vg.addComponent(addressLabel). //
				addComponent(addressTF, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE, GroupLayout.PREFERRED_SIZE). //
				addComponent(browseButton, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE, GroupLayout.PREFERRED_SIZE). //
				addComponent(startButton, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE, GroupLayout.PREFERRED_SIZE). //
				addComponent(stopButton, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE, GroupLayout.PREFERRED_SIZE);

		return panel;
	}

	/**
	 * @return Configuration Panel
	 */
	private Component getConfigurationPanel() {

		// Color map ComboBox
		JLabel colorMapLabel = new JLabel("Color Map:");
		colorMapBox = new JComboBox<>(ColorMap.values());
		colorMapBox.setEnabled(acopVideo.getColorMap() != null && !acopVideo.getColorMap().equals(ColorMap.NONE));
		colorMapBox.setSelectedItem(acopVideo.getColorMap());
		colorMapBox.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				if (!changingColorMap) {
					changingColorMap = true;
					acopVideo.setColorMap((ColorMap) colorMapBox.getSelectedItem());
					changingColorMap = false;
				}
			}
		});

		// Image zoom ComboBox
		JLabel imageZoomLabel = new JLabel("Zoom:");
		imageZoomBox = new JComboBox<>(ImageZoom.descriptions());
		imageZoomBox.setSelectedItem(acopVideo.getImageZoom().getDescription());
		imageZoomBox.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				aspectRatioCheckBox.setEnabled(ImageZoom.AUTO.getDescription().equals(imageZoomBox.getSelectedItem()));
				if (!changingImageZoom) {
					changingImageZoom = true;
					acopVideo.setImageZoom(ImageZoom.getValue((String) imageZoomBox.getSelectedItem()));
					changingImageZoom = false;
				}
			}
		});

		// Image orientation ComboBox
		JLabel imageOrientationLabel = new JLabel("Orientation:");
		imageOrientationBox = new JComboBox<>(ImageOrientation.descriptions());
		imageOrientationBox.setSelectedItem(acopVideo.getImageOrientation().getDescription());
		imageOrientationBox.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				if (!changingImageOrientation) {
					changingImageOrientation = true;
					acopVideo.setImageOrientation(ImageOrientation.getValue((String) imageOrientationBox.getSelectedItem()));
					changingImageOrientation = false;
				}
			}
		});

		// Overlay State Information ComboBox
		JLabel overlayStateLabel = new JLabel("Overlay:");
		overlayStateBox = new JComboBox<>(OverlayState.values());
		overlayStateBox.setSelectedItem(acopVideo.getOverlayState());
		overlayStateBox.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				if (!changingOverlayState) {
					changingOverlayState = true;
					acopVideo.setOverlayState((OverlayState) overlayStateBox.getSelectedItem());
					changingOverlayState = false;
				}
			}
		});

		// AOI Zoom CheckBox
		aoiZoomCheckBox = new JCheckBox("AOI Zoom");
		aoiZoomCheckBox.setSelected(acopVideo.isAOIZoom());
		aoiZoomCheckBox.addItemListener(new ItemListener() {
			public void itemStateChanged(ItemEvent e) {
				if (!changingAOIZoom) {
					changingAOIZoom = true;
					acopVideo.setAOIZoom(aoiZoomCheckBox.isSelected());
					changingAOIZoom = false;
				}
			}
		});

		// Keep Aspect Ratio CheckBox
		aspectRatioCheckBox = new JCheckBox("Keep Aspect Ratio");
		aspectRatioCheckBox.setSelected(acopVideo.isKeepAspectRatio());
		aspectRatioCheckBox.addItemListener(new ItemListener() {
			public void itemStateChanged(ItemEvent e) {
				if (!changingAspectRatio) {
					changingAspectRatio = true;
					acopVideo.setKeepAspectRatio(aspectRatioCheckBox.isSelected());
					changingAspectRatio = false;
				}
			}
		});

		// Histogram Equalization CheckBox
		normalizationCheckBox = new JCheckBox("Normalization:");
		normalizationCheckBox.setSelected(acopVideo.isHistogramEqualisation());
		normalizationCheckBox.addItemListener(new ItemListener() {
			public void itemStateChanged(ItemEvent e) {
				// System.out.println(acopVideo.getImageDisplayer().getCLUT());
				boolean isSelected = normalizationCheckBox.isSelected();
				histogramMinTF.setEditable(isSelected);
				histogramMaxTF.setEditable(isSelected);
				if (!changingNormalization) {
					changingNormalization = true;
					acopVideo.setHistogramEqualisation(isSelected);
					changingNormalization = false;
				}
			}
		});

		JLabel minmaxLabel = new JLabel("_"); // TODO: implement set values

		histogramMinTF = new IntegerEditor();
		histogramMinTF.setEditable(false);
		histogramMinTF.setNullAllowed(true);
		histogramMinTF.setToolTipText("min value");
		histogramMinTF.setPreferredSize(new Dimension(40, normalizationCheckBox.getPreferredSize().height));
		histogramMinTF.setMinValue(0);
		histogramMinTF.setValue(acopVideo.getHistogramMin());
		histogramMinTF.setMaxValue(acopVideo.getHistogramMax() - 1);

		histogramMinTF.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				histogramMinTF.transferFocus();
			}
		});
		histogramMinTF.addFocusListener(new FocusListener() {
			@Override
			public void focusLost(FocusEvent e) {
				changedHistogramMin();
			}

			@Override
			public void focusGained(FocusEvent e) {
			}
		});

		histogramMaxTF = new IntegerEditor();
		histogramMaxTF.setEditable(false);
		histogramMaxTF.setNullAllowed(true);
		histogramMaxTF.setToolTipText("max value");
		histogramMaxTF.setPreferredSize(new Dimension(40, normalizationCheckBox.getPreferredSize().height));
		histogramMaxTF.setMinValue(acopVideo.getHistogramMin() + 1);
		histogramMaxTF.setValue(acopVideo.getHistogramMax());
		if (acopVideo.getImageDisplayer().getCLUT() != null) //
			histogramMaxTF.setMaxValue(acopVideo.getImageDisplayer().getCLUT().getMaxValue());

		// printMinMax("initialize");

		histogramMaxTF.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				histogramMaxTF.transferFocusBackward();
			}
		});
//		acopVideo.addPropertyChangeListener(new PropertyChangeListener() {
//			
//			@Override
//			public void propertyChange(PropertyChangeEvent evt) {
//				System.err.println("############### " +evt.getPropertyName() + ": " + evt.getOldValue() + " -> " + evt.getNewValue());
//				
//			}
//		});
		histogramMaxTF.addFocusListener(new FocusListener() {
			@Override
			public void focusLost(FocusEvent e) {
				changedHistogramMax();
			}

			@Override
			public void focusGained(FocusEvent e) {
			}
		});

		// Create panel
		JPanel panel = new JPanel(new BorderLayout());
		panel.setBorder(BorderFactory.createTitledBorder("Configuration"));
		panel.setOpaque(false);

		GroupLayout layout = new GroupLayout(panel);
		panel.setLayout(layout);
		layout.setAutoCreateContainerGaps(true);
		layout.setAutoCreateGaps(true);

		GroupLayout.SequentialGroup hg = layout.createSequentialGroup();
		layout.setHorizontalGroup(hg);
		layout.linkSize(SwingConstants.HORIZONTAL, colorMapBox, imageZoomBox, imageOrientationBox, overlayStateBox);

		hg.addGroup(layout.createParallelGroup(Alignment.LEADING) //
				.addComponent(colorMapLabel) //
				.addComponent(imageZoomLabel)//
				.addComponent(imageOrientationLabel)//
				.addComponent(overlayStateLabel)) /**/
				.addGroup(layout.createParallelGroup(Alignment.LEADING) //
						.addComponent(colorMapBox, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE, GroupLayout.PREFERRED_SIZE) //
						.addComponent(imageZoomBox, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE, GroupLayout.PREFERRED_SIZE) //
						.addComponent(imageOrientationBox, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE, GroupLayout.PREFERRED_SIZE) //
						.addComponent(overlayStateBox, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE, GroupLayout.PREFERRED_SIZE)) /**/
				.addGap(16, 16, 16) /**/
				.addGroup(layout.createParallelGroup(Alignment.LEADING) //
						.addComponent(aoiZoomCheckBox, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE, GroupLayout.PREFERRED_SIZE) //
						.addComponent(aspectRatioCheckBox, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE, GroupLayout.PREFERRED_SIZE) //

						.addGroup(layout.createSequentialGroup() //
								.addComponent(normalizationCheckBox, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE, GroupLayout.PREFERRED_SIZE) //
								.addComponent(histogramMinTF, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE, GroupLayout.PREFERRED_SIZE) //
								.addComponent(minmaxLabel, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE, GroupLayout.PREFERRED_SIZE) //
								.addComponent(histogramMaxTF, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE, GroupLayout.PREFERRED_SIZE) //
		) //
		);

		// addComponent(normalizationCheckBox, GroupLayout.PREFERRED_SIZE,
		// GroupLayout.DEFAULT_SIZE,
		// GroupLayout.PREFERRED_SIZE));

		GroupLayout.SequentialGroup vg = layout.createSequentialGroup();
		layout.setVerticalGroup(vg);
		vg.addGroup(layout.createParallelGroup(Alignment.CENTER) //
				.addComponent(colorMapLabel) //
				.addComponent(colorMapBox, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE, GroupLayout.PREFERRED_SIZE) //
				.addComponent(aoiZoomCheckBox, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE, GroupLayout.PREFERRED_SIZE)) /**/
				.addGroup(layout.createParallelGroup(Alignment.CENTER) //
						.addComponent(imageZoomLabel) //
						.addComponent(imageZoomBox, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE, GroupLayout.PREFERRED_SIZE) //
						.addComponent(aspectRatioCheckBox, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE, GroupLayout.PREFERRED_SIZE)) /**/
				.addGroup(layout.createParallelGroup(Alignment.CENTER) //
						.addComponent(imageOrientationLabel) //
						.addComponent(imageOrientationBox, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE, GroupLayout.PREFERRED_SIZE)) /**/
				.addGroup(layout.createParallelGroup(Alignment.CENTER) //
						.addComponent(overlayStateLabel) //
						.addComponent(overlayStateBox, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE, GroupLayout.PREFERRED_SIZE) //
						.addGroup(layout.createParallelGroup() //
								.addComponent(normalizationCheckBox, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE, GroupLayout.PREFERRED_SIZE) //
								.addComponent(histogramMinTF, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE, GroupLayout.PREFERRED_SIZE) //
								.addComponent(minmaxLabel, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE, GroupLayout.PREFERRED_SIZE) //
								.addComponent(histogramMaxTF, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE, GroupLayout.PREFERRED_SIZE) //
		));

		return panel;
	}

	private void changedHistogramMin() {
		if (!changingHistogramMin) {
			// printMinMax("before changedHistogramMin");
			changingHistogramMin = true;
			try {
				acopVideo.setHistogramMin(histogramMinTF.getValue().intValue());

			} catch (java.lang.NullPointerException e) {
				histogramMinTF.setValue(acopVideo.getHistogramMin());
			} catch (IllegalArgumentException e) {
				histogramMinTF.setValue(acopVideo.getHistogramMin());
			}
			changingHistogramMin = false;
			// printMinMax("after changedHistogramMin");
		}
	}

	private void changedHistogramMax() {
//		System.err.println(" ################ AcopVideoApp.changedHistogramMax(): changingHistogramMax = " + changingHistogramMax);
//		System.err.println(" ################ AcopVideoApp.changedHistogramMax(): histogramMaxTF.getValue().intValue() = " + histogramMaxTF.getValue().intValue());
//		System.err.println(" ################ AcopVideoApp.changedHistogramMax(): acopVideo.getHistogramMax() = " + acopVideo.getHistogramMax());
		if (!changingHistogramMax) {
			changingHistogramMax = true;
			try {
				acopVideo.setHistogramMax(histogramMaxTF.getValue().intValue());

			} catch (java.lang.NullPointerException e) {
				histogramMaxTF.setValue(acopVideo.getHistogramMax());
			} catch (IllegalArgumentException e) {
				histogramMaxTF.setValue(acopVideo.getHistogramMax());
			}
			changingHistogramMax = false;
		}

	}

	private Component getHelpPanel() {
		JPanel helpPanel = new JPanel(new GridBagLayout());
		helpPanel.setBorder(BorderFactory.createTitledBorder("Help"));

		helpButton = new JButton("Help");
		helpButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				try {
					URI uri = new URI(System.getProperty(JARG_APP_URL, DEFAULT_APP_URL));
					Desktop.getDesktop().browse(uri);
				} catch (Exception ex) {
					ex.printStackTrace();
				}
			}
		});

		helpPanel.add(helpButton, new GridBagConstraints(0, 0, 1, 1, 1, 1, GridBagConstraints.CENTER, GridBagConstraints.NONE, new Insets(2, 2, 2, 2), 0, 0));

		return helpPanel;
	}

	/**
	 * @return Configuration Panel
	 */
	private Component getFilePanel() {

		saveAsPNGButton = new JButton("Save As 1:1 PNG...");
		saveAsPNGButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				acopVideo.saveAsPNG();
			}
		});

		exportToPngButton = new JButton("Export To PNG...");
		exportToPngButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				acopVideo.exportToPNG();
			}
		});

		JPanel panel = new JPanel(new BorderLayout());
		panel.setBorder(BorderFactory.createTitledBorder("File"));
		panel.setOpaque(false);

		GroupLayout layout = new GroupLayout(panel);
		panel.setLayout(layout);
		layout.setAutoCreateContainerGaps(true);
		layout.setAutoCreateGaps(true);

		ParallelGroup hg = layout.createParallelGroup();
		layout.setHorizontalGroup(hg);

		layout.linkSize(SwingConstants.HORIZONTAL, saveAsPNGButton, exportToPngButton);

		hg.addComponent(saveAsPNGButton) //
				.addComponent(exportToPngButton);

		GroupLayout.SequentialGroup vg = layout.createSequentialGroup();
		layout.setVerticalGroup(vg);
		vg.addComponent(saveAsPNGButton) //
				.addComponent(exportToPngButton);

		return panel;
	}

	// called when added to window
	@Override
	public void addNotify() {
		super.addNotify();

		Window window = SwingUtilities.windowForComponent(this);

		if (window == null)
			throw new IllegalStateException("Unexpected");

		if (lastWindow != null)
			lastWindow.removeWindowListener(addressFocusListener);

		window.addWindowListener(addressFocusListener);
		lastWindow = window;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see javax.swing.JComponent#setEnabled(boolean)
	 */
	@Override
	public void setEnabled(boolean enabled) {
		super.setEnabled(enabled);

		acopVideo.setEnabled(enabled);

		startButton.setEnabled(!isLiveMode);
		stopButton.setEnabled(isLiveMode);
		saveAsPNGButton.setEnabled(!isLiveMode);
		exportToPngButton.setEnabled(!isLiveMode);
		addressTF.setEnabled(!isLiveMode);
		browseButton.setEnabled(!isLiveMode);
	}

	public static void main(String[] args) {

		secureJArg(ImageDisplayer.JARG_SPLASH_IMAGE_URL);
		secureJArg(ImageDisplayer.JARG_ERRORIMAGE_URL);

		secureJArg(JARG_ACOPVIDEO_CONNECTION_ADDRESS);
		secureJArg(JARG_APP_URL);
		secureJArg(JARG_LAF_DEFAULT);

		boolean b = Boolean.valueOf(System.getProperty(JARG_LAF_DEFAULT, DEFAULT_LAF));
		for (String s : args) {
			if (s.equals(JARG_LAF_DEFAULT))
				b = true;
		}
		if (!b) {
			try {
				UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
			} catch (Exception ignored) {
			}
		}

		TInitializerFactory.getInstance().getInitializer().setSrvPacketMtu(64000);
		TInitializerFactory.getInstance().getInitializer().setClnRcvBufferSize(131072);
		JFrame f = new JFrame();
		f.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		if (b) {
			f.setSize(new Dimension(720, 720));
			f.setMinimumSize(new Dimension(720, 340));
		} else {
			f.setSize(new Dimension(580, 580));
			f.setMinimumSize(new Dimension(580, 340));
		}
		f.setTitle("ACOP Video Application");
		f.setContentPane(new AcopVideoApplication());
		f.setVisible(true);
	}

	/**
	 * Puts this applications into a live mode. When in live mode all connection point control are disabled.
	 */
	private void setLiveMode(boolean liveMode) {
		isLiveMode = liveMode;
		startButton.setEnabled(!liveMode);
		stopButton.setEnabled(liveMode);
		saveAsPNGButton.setEnabled(!liveMode);
		exportToPngButton.setEnabled(!liveMode);
		addressTF.setEnabled(!liveMode);
		browseButton.setEnabled(!liveMode);
	}

	private void setDisplayClutHeader(boolean displayClutHeader) {
		// TODO Auto-generated method stub
	}

	private void startVideo() {
		try {
			acopVideo.start();
		} catch (Exception e) {
			JOptionPane.showMessageDialog(this, "Failed to start video.\n" + e.getMessage(), "Error", JOptionPane.ERROR_MESSAGE);
		}
	}

	/**
	 * Extends Java System property list with property 'jnlp.<key>' for making it available for the Java Web-Start.
	 * 
	 * @param key
	 *            - Java VM property name
	 * @return Java VM property value
	 */
	private static String secureJArg(String key) {
		String value = System.getProperty(key);
		if (value == null) {
			value = System.getProperty("jnlp." + key);
			if (value != null) {
				System.setProperty(key, value);
			}
		}
		System.out.println("secureJArg(\"" + key + "\") = " + (value == null ? null : ("\"" + value + "\"")));
		return value;
	}

	// /**
	// * Only for debuging
	// *
	// * @param msg
	// * @deprecated
	// */
	// private void printMinMax(String msg) {
	// Number minOfMin = null, min = null, maxOfMin = null;
	// if (histogramMaxTF != null) {
	// minOfMin = ((AbstractNumberDocument)
	// histogramMinTF.getDocument()).getMinValue();
	// min = histogramMinTF.getValue();
	// maxOfMin = ((AbstractNumberDocument)
	// histogramMinTF.getDocument()).getMaxValue();
	// }
	//
	// Number minOfMax = null, max = null, maxOfMax = null;
	// if (histogramMaxTF != null) {
	// minOfMax = ((AbstractNumberDocument)
	// histogramMaxTF.getDocument()).getMinValue();
	// max = histogramMaxTF.getValue();
	// maxOfMax = ((AbstractNumberDocument)
	// histogramMaxTF.getDocument()).getMaxValue();
	// }
	//
	// StringBuilder sb = new StringBuilder(msg);
	// sb.append(": ");
	// sb.append("min = ").append(min);
	// sb.append(" from [").append(minOfMin);
	// sb.append(", ").append(maxOfMin);
	// sb.append("], max = ").append(max);
	// sb.append(" from [").append(minOfMax);
	// sb.append(", ").append(maxOfMax).append("]");
	//
	// System.err.println(sb.toString());
	// }
}
