/**
 *
 */
package de.desy.acop.video;

import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.geom.GeneralPath;
import java.awt.geom.Line2D;
import java.awt.geom.Rectangle2D;

import javax.swing.JPanel;

import de.desy.acop.video.analysis.AnalysisData;
import de.desy.acop.video.analysis.BesselJ;
import de.desy.acop.video.analysis.gaussian.Gaussian2DAnalysisData;
import de.desy.acop.video.analysis.gaussian.GaussianAnalysisData;
import de.desy.acop.video.analysis.interferometer.InterferometerAnalysisData;

/**
 * </code>SideProfileDisplayer</code> is a component that displays a one dimensional profile of an image. This displayer
 * can have either horizontal or vertical orientation. It plots a </code>double[]</code> normalized to its height or
 * width.
 *
 * @author Tilen Kusterle, Cosylab
 */
public class SideProfileDisplayer extends JPanel {

    private static final long serialVersionUID = 1L;
    private boolean horizontalOrientation = true;
    private double[] profile = null;
    private double[] statisticalProfile = null;
    private double[] fittedProfile = null;
    private double max;
    private double min;
    private double canvasStart = 0;
    private double canvasSize = 0;
    private int roiStart = -1;
    private int roiLength = -1;
    private double roi1 = Double.NaN;
    private double roi2 = Double.NaN;
    private boolean displayStatistics = true;
    private Object mutex = new Object();

    /**
     * Constructs this </code>SideProfileDisplayer</code>. The construction parameter defines the orientation of this
     * displayer (horizontal or vertical).
     *
     * @param horizontalOrientation if </code>true</code> the displayer is horizontally oriented; if </code>false</code>
     *        the displayer is vertically oriented
     */
    public SideProfileDisplayer(boolean horizontalOrientation) {
        this.horizontalOrientation = horizontalOrientation;
        setBackground(Definitions.getBackgroundColor());
    }

    /**
     * Sets data to this </code>SideProfileDisplayer</code>.
     *
     * @param data the data that will be drawn on this displayer
     */
    public void setData(AnalysisData data) {
        if (data == null) return;
        synchronized (mutex) {
            this.profile = data.profile;
            this.roiStart = data.roiStart;
            this.roiLength = data.roiLength;
        }
        if (roiStart < 0 || roiLength < 0 || roiStart + roiLength > profile.length)
            throw new IllegalStateException("Illegal region of interest!");
        if (data instanceof GaussianAnalysisData) {
            setGaussian((GaussianAnalysisData)data);
        } else if (data instanceof InterferometerAnalysisData) {
            setInterferometer((InterferometerAnalysisData)data);
        } else if (data instanceof Gaussian2DAnalysisData) {
            setGaussian2D((Gaussian2DAnalysisData)data);
        }
        repaint();
    }

    private void setGaussian(GaussianAnalysisData data) {
        double[] statProfile = new double[roiLength];
        double[] fitProfile = null;
        double amplitude = data.amplitude;
        double std = data.std;
        double mean = data.mean;
        double constant = data.constant;
        if (data.drawFit) {
            double amplitudeFit = data.amplitudeFit;
            double stdFit = data.stdFit;
            double meanFit = data.meanFit;
            double constantFit = data.constantFit;
            double slopeFit = data.slopeFit;
            fitProfile = new double[roiLength];
            for (int i = 0; i < roiLength; i++) {
                statProfile[i] = amplitude * Math.exp(-(roiStart + i - mean) * (roiStart + i - mean) / (2 * std * std))
                        + constant;
                fitProfile[i] = amplitudeFit
                        * Math.exp(-(roiStart + i - meanFit) * (roiStart + i - meanFit) / (2 * stdFit * stdFit))
                        + constantFit + (roiStart + i) * slopeFit;
            }
        } else {
            for (int i = 0; i < roiLength; i++) {
                statProfile[i] = amplitude * Math.exp(-(roiStart + i - mean) * (roiStart + i - mean) / (2 * std * std))
                        + constant;
            }
        }
        min = Double.MAX_VALUE;
        max = -Double.MAX_VALUE;
        for (int i = 0; i < profile.length; i++) {
            if (profile[i] < min) min = profile[i];
            if (profile[i] > max) max = profile[i];
        }
        if (data.drawFit) {
            for (int i = 0; i < roiLength; i++) {
                if (statProfile[i] < min) min = statProfile[i];
                if (statProfile[i] > max) max = statProfile[i];
                if (fitProfile[i] < min) min = fitProfile[i];
                if (fitProfile[i] > max) max = fitProfile[i];
            }
        } else {
            for (int i = 0; i < roiLength; i++) {
                if (statProfile[i] < min) min = statProfile[i];
                if (statProfile[i] > max) max = statProfile[i];
            }
        }
        synchronized (mutex) {
            if (data.drawFit) {
                fittedProfile = fitProfile;
                statisticalProfile = statProfile;
            } else {
                statisticalProfile = statProfile;
                fittedProfile = null;
            }
        }
    }

    private void setGaussian2D(Gaussian2DAnalysisData data) {
        min = Double.MAX_VALUE;
        max = -Double.MAX_VALUE;
        for (int i = 0; i < profile.length; i++) {
            if (profile[i] < min) min = profile[i];
            if (profile[i] > max) max = profile[i];
        }

        double[] fitProfile = null;
        if (data.drawFit) {
            double amplitudeFit = max-min;
            double stdFit = data.std;
            double meanFit = data.mean;
            double constantFit = data.constant;
            int roiStart = data.roiStart;
            fitProfile = new double[roiLength];
            for (int i = 0; i < roiLength; i++) {
                fitProfile[i] = amplitudeFit
                        * Math.exp(-(roiStart + i - meanFit) * (roiStart + i - meanFit) / (2 * stdFit * stdFit))
                        + constantFit + min;
            }
        }

        if (data.drawFit) {
            for (int i = 0; i < roiLength; i++) {
                if (fitProfile[i] < min) min = fitProfile[i];
                if (fitProfile[i] > max) max = fitProfile[i];
            }
        }
        synchronized (mutex) {
            statisticalProfile = null;
            if (data.drawFit) {
                fittedProfile = fitProfile;
            } else {
                fittedProfile = null;
            }
        }
    }

    private static boolean isNotNaN(double... values) {
        for (double d : values) {
            if (Double.isNaN(d)) {
                return false;
            }
        }
        return true;
    }

    private void setInterferometer(InterferometerAnalysisData data) {
        int i = 0;
        double[] fitProfile = null;
        if (data.drawFit) {
            fitProfile = new double[roiLength];
            if ((data.isVertical && data.isInterferometerVertical)
                    || (!data.isVertical && !data.isInterferometerVertical)) {
                double bg = data.constant;
                double amp = data.amplitude;
                double asinc = data.asinc;
                double psinc = data.psinc;
                double acos = data.acos;
                double pcos = data.pcos;
                double vis = data.vis;
                if (isNotNaN(bg,amp,asinc,psinc,acos,pcos,vis)) {
                    double sinc;
                    if (data.useBesselFunction) {
                        for (; i < roiLength; i++) {
                            try {
                                sinc = BesselJ.valueAny(1,asinc * (i - psinc)) / (asinc * (i - psinc));
                                fitProfile[i] = bg + amp * sinc * sinc * (1 + vis * Math.cos(acos * (i - pcos)));
                            } catch (Exception e) {
                                //ignore
                            }
                        }
                    } else {
                        for (; i < roiLength; i++) {
                            sinc = Math.sin(asinc * (i - psinc)) / (asinc * (i - psinc));
                            fitProfile[i] = bg + amp * sinc * sinc * (1 + vis * Math.cos(acos * (i - pcos)));
                        }
                    }
                }
            }
        }
        min = Double.MAX_VALUE;
        max = -Double.MAX_VALUE;
        for (i = 0; i < profile.length; i++) {
            if (profile[i] < min) min = profile[i];
            if (profile[i] > max) max = profile[i];
        }
        if (data.drawFit) {
            for (i = 0; i < roiLength; i++) {
                if (fitProfile[i] < min) min = fitProfile[i];
                if (fitProfile[i] > max) max = fitProfile[i];
            }
        }
        synchronized (mutex) {
            statisticalProfile = null;
            if (data.drawFit) {
                fittedProfile = fitProfile;
            } else {
                fittedProfile = null;
            }
        }
    }

    /**
     * Clears this displayer's data.
     */
    public void clearData() {
        profile = null;
        statisticalProfile = null;
        fittedProfile = null;
        min = Double.MAX_VALUE;
        max = -Double.MAX_VALUE;
        canvasStart = 0;
        canvasSize = 0;
        roiStart = -1;
        roiLength = -1;
        roi1 = Double.NaN;
        roi2 = Double.NaN;
        repaint();
    }

    /**
     * Sets the position of region of interest markers on this displayer.
     *
     * @param c1 the position of the first marker
     * @param c2 the position of the second marker
     */
    public void setRoi(double c1, double c2) {
        roi1 = c1;
        roi2 = c2;
        repaint();
    }

    public void setCanvasSize(double start, double size, boolean repaintAtOnce) {
        canvasSize = size;
        canvasStart = start;
        if (repaintAtOnce) repaint();
    }

    public void setCanvasStart(double start) {
        if (canvasStart == start) return;
        canvasStart = start;
        repaint();
    }

    /**
     * Clears the region of interest markers on this displayer.
     */
    public void clearRoi() {
        roi1 = Double.NaN;
        roi2 = Double.NaN;
        repaint();
    }

    @Override
    protected void paintComponent(Graphics g) {
        super.paintComponent(g);
        drawProfile((Graphics2D)g);
        drawRoi((Graphics2D)g);
    }

    void setDisplayStatistics(boolean display) {
        this.displayStatistics = display;
        repaint();
    }

    private void drawProfile(Graphics2D g) {
        GeneralPath dataPath = new GeneralPath();
        GeneralPath statisticsPath = new GeneralPath();
        GeneralPath fitPath = new GeneralPath();
        double delta;
        synchronized (mutex) {
            if (profile == null || profile.length == 0) return;
            int i = 0;
            int length = profile.length;
            double coeff = (horizontalOrientation ? getHeight() : getWidth()) / (max - min);
            delta = canvasSize / (length - 1);
            double x, y;
            if (horizontalOrientation) {
                for (i = 0; i < length; i++) {
                    x = canvasStart + i * delta;
                    y = getHeight() - (profile[i] - min) * coeff;
                    if (i == 0)
                        dataPath.moveTo(x,y);
                    else
                        dataPath.lineTo(x,y);
                }
                if (statisticalProfile != null) {
                    for (i = 0; i < roiLength; i++) {
                        x = canvasStart + (roiStart + i) * delta;
                        y = getHeight() - (statisticalProfile[i] - min) * coeff;
                        if (i == 0)
                            statisticsPath.moveTo(x,y);
                        else
                            statisticsPath.lineTo(x,y);
                    }
                }
                if (fittedProfile != null) {
                    for (i = 0; i < roiLength; i++) {
                        x = canvasStart + (roiStart + i) * delta;
                        y = getHeight() - (fittedProfile[i] - min) * coeff;
                        if (i == 0)
                            fitPath.moveTo(x,y);
                        else
                            fitPath.lineTo(x,y);
                    }
                }
            } else {
                for (i = 0; i < length; i++) {
                    x = canvasStart + i * delta;
                    y = getWidth() - (profile[i] - min) * coeff;
                    if (i == 0)
                        dataPath.moveTo(y,x);
                    else
                        dataPath.lineTo(y,x);
                }
                if (statisticalProfile != null) {
                    for (i = 0; i < roiLength; i++) {
                        x = canvasStart + (roiStart + i) * delta;
                        y = getWidth() - (statisticalProfile[i] - min) * coeff;
                        if (i == 0)
                            statisticsPath.moveTo(y,x);
                        else
                            statisticsPath.lineTo(y,x);
                    }
                }
                if (fittedProfile != null) {
                    for (i = 0; i < roiLength; i++) {
                        x = canvasStart + (roiStart + i) * delta;
                        y = getWidth() - (fittedProfile[i] - min) * coeff;
                        if (i == 0)
                            fitPath.moveTo(y,x);
                        else
                            fitPath.lineTo(y,x);
                    }
                }
            }
        }
        double x1 = horizontalOrientation ? canvasStart : 0;
        double y1 = horizontalOrientation ? 0 : canvasStart;
        double x2 = horizontalOrientation ? canvasSize : getWidth();
        double y2 = horizontalOrientation ? getHeight() : canvasSize;
        Rectangle2D r = new Rectangle2D.Double(x1,y1,x2,y2);
        g.setPaint(Definitions.getCanvasColor());
        g.fill(r);
        g.setPaint(Definitions.getDataColor());
        g.draw(dataPath);
        if (fittedProfile != null) {
            g.setPaint(Definitions.getFitColor());
            g.draw(fitPath);
        }
        if (displayStatistics) {
            g.setPaint(Definitions.getStatisticsColor());
            g.draw(statisticsPath);
        }
        double p1 = canvasStart + roiStart * delta;
        double p2 = canvasStart + (roiStart + roiLength - 1) * delta;
        x1 = horizontalOrientation ? p1 : 0;
        y1 = horizontalOrientation ? 0 : p1;
        x2 = horizontalOrientation ? p1 : getWidth();
        y2 = horizontalOrientation ? getHeight() : p1;
        Line2D l1 = new Line2D.Double(x1,y1,x2,y2);
        x1 = horizontalOrientation ? p2 : 0;
        y1 = horizontalOrientation ? 0 : p2;
        x2 = horizontalOrientation ? p2 : getWidth();
        y2 = horizontalOrientation ? getHeight() : p2;
        Line2D l2 = new Line2D.Double(x1,y1,x2,y2);
        g.setPaint(Definitions.getRoiColor());
        g.draw(l1);
        g.draw(l2);
    }

    private void drawRoi(Graphics2D g) {
        if (!Double.isNaN(roi1) && !Double.isNaN(roi2)) {
            Line2D l1, l2;
            double x1 = horizontalOrientation ? canvasStart + roi1 : 0;
            double y1 = horizontalOrientation ? 0 : 2 * canvasStart + roi1;
            double x2 = horizontalOrientation ? canvasStart + roi1 : getWidth();
            double y2 = horizontalOrientation ? getHeight() : 2 * canvasStart + roi1;
            l1 = new Line2D.Double(x1,y1,x2,y2);
            x1 = horizontalOrientation ? canvasStart + roi2 : 0;
            y1 = horizontalOrientation ? 0 : 2 * canvasStart + roi2;
            x2 = horizontalOrientation ? canvasStart + roi2 : getWidth();
            y2 = horizontalOrientation ? getHeight() : 2 * canvasStart + roi2;
            l2 = new Line2D.Double(x1,y1,x2,y2);
            g.setPaint(Definitions.getRoiSelectingColor());
            g.draw(l1);
            g.draw(l2);
        }
    }
}
