/**
 *
 */
package de.desy.acop.video.analysis;

import java.awt.Rectangle;
import java.awt.Transparency;
import java.awt.color.ColorSpace;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.awt.image.ComponentColorModel;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferByte;
import java.awt.image.DataBufferInt;
import java.awt.image.DataBufferUShort;
import java.awt.image.DirectColorModel;
import java.awt.image.Raster;
import java.awt.image.SampleModel;
import java.awt.image.WritableRaster;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.util.Hashtable;
import java.util.concurrent.Executor;
import java.util.concurrent.Executors;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.function.Function;
import java.util.logging.Level;

import de.desy.acop.displayers.tools.DismissableBlockingQueue;
import de.desy.acop.displayers.tools.TINEThreadFactory;
import de.desy.acop.transport.ConnectionFailed;
import de.desy.acop.transport.ConnectionParameters;
import de.desy.acop.video.VideoServerConnection;
import de.desy.acop.video.analysis.gaussian.AImage;
import de.desy.acop.video.analysis.gaussian.GaussianImageAnalyzer;
import de.desy.acop.video.analysis.interferometer.InterferometerImageAnalyzer;
import de.desy.acop.video.displayer.ColorMap;
import de.desy.acop.video.displayer.LinkData;
import de.desy.acop.video.displayer.TineHandler;
import de.desy.acop.video.displayer.TineImageReceiver;
import de.desy.acop.video.displayer.timage.TImageConverter;
import de.desy.acop.video.timageio.TBufferedImage;
import de.desy.acop.video.timageio.TImageMetadata;
import de.desy.tine.types.IMAGE;
import de.desy.tine.types.IMAGE.FrameHeader;

/**
 * </code>ImageAnalysisEngine</code> provides </code>AImage</code> image analysis data. It has two modes of operation:
 * local and remote analysis. In local analysis mode, the </code>IMAGE</code> arguments of the </code>updateValue</code>
 * method are processed locally. In remote mode this engine reads </code>AImage</code> data from a video analysis
 * server.
 *
 * @author Tilen Kusterle, Cosylab
 */
@SuppressWarnings("deprecation")
public class ImageAnalysisEngine implements TineImageReceiver {

    public static final String PROPERTY_INT_NUM_OF_IMAGES = "interferometerNumOfImages";
    public static final String PROPERTY_A_IMAGE = "aImage";
    public static final String PROPERTY_BACKGROUND = "background";
    public static final String PROPERTY_PRECISE_BACKGROUND = "preciseBackground";
    public static final String PROPERTY_RESET = "reset";
    private PropertyChangeSupport pcSupport;
    private Executor executor;
    private ThreadPoolExecutor analysisExecutor;
    private boolean localAnalysis = true;
    private VideoServerConnection videoServerConnection;
    protected ColorDecoder decoder;
    private AnalysisImage aImage;
    private IMAGE image;
    private IMAGE background;
    private double[] preciseBackground;
    private int imageWidth = -1;
    private int imageHeight = -1;
    private int backgroundWidth = -1;
    private int backgroundHeight = -1;
    private int roiX = -1;
    private int roiY = -1;
    private int roiW = -1;
    private int roiH = -1;
    private int roi2X = -1;
    private int roi2Y = -1;
    private int roi2W = -1;
    private int roi2H = -1;
    private double threshold = Double.NaN;
    private boolean calculateThreshold = true;
    private boolean performFit = false;
    private boolean performSmoothing = false;
    private ColorMap colorMap = null;
    private ImageAnalyzer imageAnalyzer = new GaussianImageAnalyzer();
    private int interferometerNumOfImages = 5;
    private boolean needToUpdateBackground = false;

    /**
     * Constructs this </code>ImageAnalysisEngine</code>.
     *
     * @param decoder the </code>ColorDecoder</code> to be used with this </code>ImageAnalysisEngine</code>
     */
    public ImageAnalysisEngine(ColorDecoder decoder) {
        this.decoder = decoder;
        pcSupport = new PropertyChangeSupport(this);
    }

    @Override
    public void resetForReceiving() {
        pcSupport.firePropertyChange(PROPERTY_RESET,false,true);
    }

    @Override
    public void updateValue(IMAGE newImage, LinkData data) {
        updateValue(newImage);
    }

    @Override
    public synchronized void updateValue(IMAGE newImage) {
        if (!localAnalysis) return;
        //		logger.log(Level.INFO,"New update received.");
        //		image = TImageUtils.clone(newImage); // mdavid: commented
        image = newImage.clone();
        FrameHeader hdr = image.getFrameHeader();
        int w = hdr.aoiWidth > 0 ? hdr.aoiWidth : hdr.sourceWidth;
        int h = hdr.aoiHeight > 0 ? hdr.aoiHeight : hdr.sourceHeight;
        if (w != imageWidth || h != imageHeight) {
            roiX = 0;
            roiY = 0;
            roiW = w;
            roiH = h;
            roi2X = 0;
            roi2Y = 0;
            roi2W = w;
            roi2H = h;
        }
        imageWidth = w;
        imageHeight = h;
        if ((imageWidth != backgroundWidth || imageHeight != backgroundHeight) && background != null) {
            setBackground(null);
        } else if (needToUpdateBackground && preciseBackground != null) {
            setPreciseBackground(preciseBackground,backgroundWidth,image,true);
            needToUpdateBackground = false;
        } else {
            performAnalysis();
        }

    }

    /**
     * Sets the precise background array. This is the array that holds values per pixels for whole background image, but
     * instead of int values, pixel values are given as luminosity doubles. Negative values are permitted and will be
     * taken into account during analysis, but forced to 0, when real image is constructed.
     *
     * @param background the pixel value array
     * @param width the width of the image
     */
    public synchronized void setPreciseBackground(double[] background, int width) {
        TineHandler.logger.log(Level.INFO,"Background image set.");
        setPreciseBackground(background,width,image,true);
    }

    /**
     * Returns the double array that is used as background image.
     *
     * @see #setPreciseBackground(double[], boolean)
     * @return the double array image
     */
    public synchronized double[] getPreciseBackground() {
        return preciseBackground;
    }

    /**
     * Returns the width of the background image.
     *
     * @return the width
     */
    public synchronized int getBackgroundWidth() {
        return backgroundWidth;
    }

    /**
     * Sets the background image and initiates required actions.
     *
     * @param background the background image array
     * @param width the width of the image
     * @param reanalyze true if the image should be reanalyzed
     */
    private synchronized void setPreciseBackground(final double[] preciseBackground, final int width,
            IMAGE orgImage, boolean reanalyze) {
        this.background = null;
        this.preciseBackground = null;
        backgroundWidth = -1;
        backgroundHeight = -1;
        if (preciseBackground != null) {
            int w = width;
            int h = preciseBackground.length / width;
            if (w == imageWidth && h == imageHeight) {
                backgroundWidth = w;
                backgroundHeight = h;
                this.preciseBackground = preciseBackground;
                needToUpdateBackground = orgImage == null;
                if (!needToUpdateBackground) {
                    this.background = toImage(preciseBackground,width,orgImage);
                }
            }
        }
        if (reanalyze) {
            if (localAnalysis)
                performAnalysis();
            else {
                executeTask(() -> getVideoServerConnection().setBackground(background));
            }
        }
        pcSupport.firePropertyChange(PROPERTY_PRECISE_BACKGROUND,null,this.preciseBackground);
        if (!needToUpdateBackground) {
            pcSupport.firePropertyChange(PROPERTY_BACKGROUND,null,this.background);
        }
    }

    /**
     * Transforms the double array to a Gray image.
     *
     * @param image the image to transform
     * @param width the width of the image
     * @param orgImage original image, which specifies the bytesPerPixel, bitsPerPixel etc.
     * @return generated IMAGE
     */
    public static IMAGE toImage(double[] image, int width, IMAGE orgImage) {
        if (image == null || image.length == 0) return null;
        if (width < 1 || image.length / width < 1) return null;
        BufferedImage bi = toBufferedImage(image,width,orgImage);
        return TBufferedImage.toIMAGE(bi);
    }

    /**
     *
     * @param values the values for the pixels
     * @param width the width of the image
     * @param orgImage the original image, used to specify
     * @return
     */
    public static BufferedImage toBufferedImage(double[] values, int width, IMAGE orgImage) {
        if (orgImage == null) {
            return null;
        }
        Hashtable<?,?> props = TImageMetadata.createProperties(orgImage);
        int length = values.length;
        double bitsFactor = Math.pow(2, orgImage.getFrameHeader().effectiveBitsPerPixel)/256.;

        Function<DataBuffer, BufferedImage> toGrayImage = (buffer) -> {
            WritableRaster raster = Raster.createInterleavedRaster(buffer, width, values.length/width, width, 1,
                    new int[] { 0 }, null);
            ColorModel cm = new ComponentColorModel(ColorSpace.getInstance(ColorSpace.CS_GRAY),
                    new int[] { orgImage.getFrameHeader().effectiveBitsPerPixel }, false, true,
                    Transparency.OPAQUE, buffer.getDataType());
            return new BufferedImage(cm, raster, false, props);
        };
        Function<DataBuffer, BufferedImage> toRGBImage = (buffer) -> {
            ColorModel cm = new DirectColorModel(24, 0x00ff0000, 0x0000ff00, 0x000000ff, 0x0);
            SampleModel sm = cm.createCompatibleSampleModel(width, values.length/width);
            WritableRaster raster = WritableRaster.createWritableRaster(sm, buffer, null);
            return new BufferedImage(cm, raster, false, props);
        };
        switch (orgImage.getFrameHeader().bytesPerPixel) {
            case 1:
                byte[] byteData = new byte[length];
                for (int i = 0; i < length; i++) {
                    byteData[i] = (byte)((values[i] < 0 ? 0 : values[i]) + 0.5);
                }
                DataBuffer dataBuffer = new DataBufferByte(byteData,length);
                return toGrayImage.apply(dataBuffer);
            case 2:
                double factor = Math.pow(2, orgImage.getFrameHeader().effectiveBitsPerPixel)/256.;
                short[] shortData = new short[length];
                for (int i = 0; i < length; i++) {
                    shortData[i] = (short)(factor*(values[i] < 0 ? 0 : values[i]) + 0.5);
                }
                DataBuffer dataBuffer2 = new DataBufferUShort(shortData,length);
                return toGrayImage.apply(dataBuffer2);
            case 3:
                int[] intData3 = new int[length];
                for (int i = 0; i < length; i++) {
                    intData3[i] = ((int)((values[i] < 0 ? 0 : values[i]) * bitsFactor + 0.5)) & 0xFFFFFF;
                }
                DataBuffer dataBuffer3 = new DataBufferInt(intData3,length);
                return toRGBImage.apply(dataBuffer3);
            case 4:
                int[] intData4 = new int[length];
                for (int i = 0; i < length; i++) {
                    intData4[i] = (int)((values[i] < 0 ? 0 : values[i]) * bitsFactor + 0.5);
                }
                DataBuffer dataBuffer4 = new DataBufferInt(intData4,length);
                return toRGBImage.apply(dataBuffer4);
            default:
                return null;
            }
    }

    /**
     * Gets the background image used by this </code>ImageAnalysisEngine</code>.
     *
     * @return the background image
     */
    public synchronized IMAGE getBackground() {
        return background;
    }

    /**
     * Sets the background image to be used by this </code>ImageAnalysisEngine</code>.
     *
     * @param background the image to set
     */
    public synchronized void setBackground(IMAGE background) {
        TineHandler.logger.log(Level.INFO,"Background image set.");
        setBackground(background,true);
        this.background = background;
    }

    private synchronized void setBackground(final IMAGE background, boolean reanalyze) {
        if (background == null) {
            setPreciseBackground(null,-1);
        } else {
            int width = background.getFrameHeader().sourceWidth;
            BufferedImage image = toBufferedImage(background,ColorMap.NONE);
            double[] data = imageToDoubleValues(image,background.getFrameHeader().effectiveBitsPerPixel,getDecoder());
            setPreciseBackground(data,width,background,reanalyze);
            this.background = background;
        }
    }

    /**
     * Sets the region of interest and reanalyzes pixels.
     *
     * @param roiX the starting x coordinate
     * @param roiY the starting y coordinate
     * @param roiW the width of the region
     * @param roiH the height of the region
     */
    public synchronized void setRoi(final int roiX, final int roiY, final int roiW, final int roiH) {
        if (this.roiX == roiX && this.roiY == roiY && this.roiW == roiW && this.roiH == roiH) return;
        TineHandler.logger.log(Level.INFO,"ROI: x=" + roiX + ", y=" + roiY + ", w=" + roiW + ", h=" + roiH);
        setAnalysisParameters(threshold,roiX,roiY,roiW,roiH,roi2X,roi2Y,roi2W,roi2H,calculateThreshold,performFit,
                performSmoothing);
    }

    /**
     * Sets the region of interest and reanalyzes pixels.
     *
     * @param roiX the starting x coordinate
     * @param roiY the starting y coordinate
     * @param roiW the width of the region
     * @param roiH the height of the region
     */
    public synchronized void setRoi2(final int roi2X, final int roi2Y, final int roi2W, final int roi2H) {
        if (this.roi2X == roi2X && this.roi2Y == roi2Y && this.roi2W == roi2W && this.roi2H == roi2H) return;
        TineHandler.logger.log(Level.INFO,
                "Threshold ROI: x=" + roi2X + ", y=" + roi2Y + ", w=" + roi2W + ", h=" + roi2H);
        setAnalysisParameters(threshold,roiX,roiY,roiW,roiH,roi2X,roi2Y,roi2W,roi2H,calculateThreshold,performFit,
                performSmoothing);
    }

    /**
     * Gets the </code>IMAGE</code>.
     *
     * @return the </code>IMAGE</code>
     */
    public synchronized IMAGE getImage() {
        return image;
    }

    /**
     * Gets the </code>AnalysisImage</code>.
     *
     * @return the </code>AnalysisImage</code>
     */
    public synchronized AnalysisImage getAImage() {
        return aImage;
    }

    /**
     * Returns the rectangle describing the region of interest.
     *
     * @return the region of intereset
     */
    public synchronized Rectangle getROI() {
        return new Rectangle(roiX,roiY,roiW,roiH);
    }

    /**
     * Returns the rectangle describing the region of interest used for threshold calculation.
     *
     * @return the threshold region of interest
     */
    public synchronized Rectangle getThresholdROI() {
        return new Rectangle(roi2X,roi2Y,roi2W,roi2H);
    }

    /**
     * Gets the threshold value.
     *
     * @see #setThreshold(double)
     * @return the threshold value
     */
    public synchronized double getThreshold() {
        return threshold;
    }

    /**
     * Sets the threshold value, which discards all pixels which have returned by <code>ColorDecoder</code> lover than
     * threshold value. Threshold value has effect only between values Const2D and Amplitude2D as described by
     * <code>AImage</code>.
     *
     * @see AImage
     * @see AImage#getAmplitude2D()
     * @see AImage#getConst2D()
     * @param threshold the threshold value to set
     */
    public synchronized void setThreshold(final double threshold) {
        if (this.threshold == threshold) return;
        //		this.threshold = threshold;
        TineHandler.logger.log(Level.INFO,"Threshold: " + threshold);
        setAnalysisParameters(threshold,roiX,roiY,roiW,roiH,roi2X,roi2Y,roi2W,roi2H,calculateThreshold,performFit,
                performSmoothing);
    }

    /**
     * Sets the calculate thershold parameter. When true the threshold values is calculated at each image analysis
     * performed and the original user defined threshold is overridden with the calculated one.
     *
     * @param calculateThreshold true if the threshold should be calculated or false otherwise
     */
    public synchronized void setCalculateThreshold(final boolean calculateThreshold) {
        if (this.calculateThreshold == calculateThreshold) return;
        TineHandler.logger.log(Level.INFO,"Calculate threshold: " + calculateThreshold);
        setAnalysisParameters(roiX,roiY,roiW,roiH,roi2X,roi2Y,roi2W,roi2H,calculateThreshold,performFit,
                performSmoothing);
    }

    /**
     * Returns true if the threshold is calculated by this engine, or false if the user defined the threshold to be
     * used.
     *
     * @return true if threshold calculation is automatic
     */
    public boolean isCalculateThreshold() {
        return calculateThreshold;
    }

    /**
     * Sets the perform fit parameter. When true fitting is performed on the data. When true, the time required to
     * perform the analysis might be significantly longer
     *
     * @param performFit true if fitting should be performed or false otherwise
     */
    public synchronized void setPerformFit(final boolean performFit) {
        if (this.performFit == performFit) return;
        TineHandler.logger.log(Level.INFO,"Perform fit: " + performFit);
        setAnalysisParameters(roiX,roiY,roiW,roiH,roi2X,roi2Y,roi2W,roi2H,calculateThreshold,performFit,
                performSmoothing);
    }

    /**
     * Sets the perform smoothing parameter. When true smoothing is performed on the side view data.
     *
     * @param performSmoothing true if smoothing should be performed or false otherwise
     */
    public synchronized void setPerformSmoothing(final boolean performSmoothing) {
        if (this.performSmoothing == performSmoothing) return;
        TineHandler.logger.log(Level.INFO,"Perform smoothing: " + performSmoothing);
        setAnalysisParameters(roiX,roiY,roiW,roiH,roi2X,roi2Y,roi2W,roi2H,calculateThreshold,performFit,
                performSmoothing);
    }

    /**
     * Returns true if fitting is performed by this engine, or false if otherwise
     *
     * @return true if fitting is performed
     */
    public boolean isPerformFit() {
        return performFit;
    }

    /**
     * Gets the decoder used by this </code>ImageAnalysisEngine</code>.
     *
     * @return the </code>ColorDecoder</code> used
     */
    public synchronized ColorDecoder getDecoder() {
        return decoder;
    }

    /**
     * Sets a </code>ColorDecoder</code> to this </code>ImageAnalysisEngine</code>
     *
     * @param decoder the </code>ColorDecoder</code> to set
     */
    public synchronized void setDecoder(ColorDecoder decoder) {
        if (this.decoder == decoder) return;
        this.decoder = decoder;
        if (localAnalysis)
            performAnalysis();
        else {
            // set to server?
        }
    }

    /**
     * Sets the analysis parameters (only region of interest, threshold remains the same) and reanalyzes pixels.
     *
     * @param roiX the starting x coordinate
     * @param roiY the starting y coordinate
     * @param roiW the width of the region
     * @param roiH the height of the region
     * @param calculateThreshold true if the threshold is calculated by the server or false otherwise
     * @param performFit true if the fitting should be performed or false otherwise
     */
    public void setAnalysisParameters(final int roiX, final int roiY, final int roiW, final int roiH, final int roi2X,
            final int roi2Y, final int roi2W, final int roi2H, final boolean calculateThreshold, boolean performFit,
            boolean performSmoothing) {
        setAnalysisParameters(threshold,roiX,roiY,roiW,roiH,roi2X,roi2Y,roi2W,roi2H,calculateThreshold,performFit,
                performSmoothing);
    }

    /**
     * Sets the analysis parameters (threshold and region of interest) and reanalyzes pixels.
     *
     * @param threshold the threshold
     * @param roiX the starting x coordinate
     * @param roiY the starting y coordinate
     * @param roiW the width of the region
     * @param roiH the height of the region
     */
    public synchronized void setAnalysisParameters(final double threshold, final int roiX, final int roiY,
            final int roiW, final int roiH, final int roi2X, final int roi2Y, final int roi2W, final int roi2H,
            final boolean calculateThreshold, final boolean performFit, final boolean performSmoothing) {
        if (!localAnalysis) {
            executeTask(() -> getVideoServerConnection().setAnalysisParameters(threshold,roiX,roiY,roiW,roiH,roi2X,
                    roi2Y,roi2W,roi2H,calculateThreshold,performFit));
        } else {
            this.calculateThreshold = calculateThreshold;
            this.performFit = performFit;
            this.performSmoothing = performSmoothing;
            if (this.threshold == threshold && this.roiX == roiX && this.roiY == roiY && this.roiW == roiW
                    && this.roiH == roiH && this.roi2X == roi2X && this.roi2Y == roi2Y && this.roi2H == roi2H
                    && this.roi2W == roi2W)
                return;
            this.threshold = threshold;
            if (roiX >= imageWidth)
                this.roiX = imageWidth - 1;
            else if (roiX < 0)
                this.roiX = 0;
            else
                this.roiX = roiX;
            if (roiY >= imageHeight)
                this.roiY = imageHeight - 1;
            else if (roiY < 0)
                this.roiY = 0;
            else
                this.roiY = roiY;
            if (this.roiX + roiW > imageWidth)
                this.roiW = imageWidth - this.roiX;
            else if (roiW <= 0)
                this.roiW = 1;
            else
                this.roiW = roiW;
            if (this.roiY + roiH > imageHeight)
                this.roiH = imageHeight - this.roiY;
            else if (roiH <= 0)
                this.roiH = 1;
            else
                this.roiH = roiH;
            if (roi2X >= imageWidth)
                this.roi2X = imageWidth - 1;
            else if (roi2X < 0)
                this.roi2X = 0;
            else
                this.roi2X = roi2X;
            if (roi2Y >= imageHeight)
                this.roi2Y = imageHeight - 1;
            else if (roi2Y < 0)
                this.roi2Y = 0;
            else
                this.roi2Y = roi2Y;
            if (this.roi2X + roi2W > imageWidth)
                this.roi2W = imageWidth - this.roi2X;
            else if (roi2W <= 0)
                this.roi2W = 1;
            else
                this.roi2W = roi2W;
            if (this.roi2Y + roi2H > imageHeight)
                this.roi2H = imageHeight - this.roi2Y;
            else if (roi2H <= 0)
                this.roi2H = 1;
            else
                this.roi2H = roi2H;
            performAnalysis();
        }
    }

    /**
     * Sets the analysis mode for this </code>ImageAnalysisEngine</code>.
     *
     * @param local </code>true</code> for local analysis and </code>false</code> for remote analysis
     */
    public void setLocalAnalysisMode(boolean local) {
        if (localAnalysis == local) return;
        TineHandler.logger.log(Level.INFO,"Analysis mode: " + local);
        localAnalysis = local;
        if (localAnalysis) {
            executeTask(() -> getVideoServerConnection().disconnect());
            setRoi(0,0,imageWidth,imageHeight);
        }
    }

    /**
     * Gets the analysis mode for this </code>ImageAnalysisEngine</code>.
     *
     * @return the </code>true</code> if local analysis is selected and </code>false</code> if remote analysis is
     *         selected
     */
    public boolean isLocalAnalysisMode() {
        return localAnalysis;
    }

    /**
     * Starts connection to remote video analysis server.
     */
    public void startRemoteAnalysis() {
        TineHandler.logger.log(Level.INFO,"Start remote analysis.");
        try {
            getVideoServerConnection().connect();
        } catch (ConnectionFailed e) {
            TineHandler.logger.log(Level.SEVERE,"Cannot start remote analysis.",e);
        }
    }

    /**
     * Stops connection to remote video analysis server.
     */
    public void stopRemoteAnalysis() {
        //		TineHandler.logger.log(Level.INFO, "Stop remote analysis.");
        getVideoServerConnection().disconnect();
    }

    /**
     * Sets </code>ConnectionParameters</code> for remote video analysis server.
     *
     * @param cp the </code>ConnectionParameters</code> to set
     */
    public void setAnalysisServerConnectionParameters(ConnectionParameters cp) {
        getVideoServerConnection().setConnectionParameters(cp);
    }

    private VideoServerConnection getVideoServerConnection() {
        if (videoServerConnection == null) {
            videoServerConnection = new VideoServerConnection();
            try {
                videoServerConnection.setAnalysisType(imageAnalyzer.getAnalysisType());
            } catch (ConnectionFailed e) {
                TineHandler.logger.log(Level.SEVERE,"Cannot set analysis type.",e);
            }
            videoServerConnection.addPropertyChangeListener(evt -> {
                if (localAnalysis) return;
                String name = evt.getPropertyName();
                if (VideoServerConnection.PROPERTY_A_IMAGE.equals(name)) {
                    setAImage((AnalysisImage)evt.getNewValue());
                } else if (VideoServerConnection.PROPERTY_BACKGROUND.equals(name)) {
                    setBackground((IMAGE)evt.getNewValue(),false);
                } else if (VideoServerConnection.PROPERTY_PRECISE_BACKGROUND.equals(name)) {
                    //TODO handle this
                    //						setPreciseBackground((double[]) evt.getNewValue(), videoServerConnection.getPreciseBackgroundWidth(),false);
                }
            });
        }
        return videoServerConnection;
    }

    public synchronized void setAImage(AnalysisImage aImage) {
        this.aImage = aImage;
        pcSupport.firePropertyChange(PROPERTY_A_IMAGE,null,aImage);
    }

    /**
     * Sets the image analyzer which is used for the analysis.
     *
     * @param imageAnalyzer
     */
    public void setImageAnalyzer(ImageAnalyzer imageAnalyzer) {
        if (imageAnalyzer == null) {
            imageAnalyzer = new GaussianImageAnalyzer();
        }
        this.imageAnalyzer = imageAnalyzer;
        if (this.imageAnalyzer instanceof InterferometerImageAnalyzer) {
            ((InterferometerImageAnalyzer)this.imageAnalyzer).setNumOfImages(interferometerNumOfImages);
        }
        if (!localAnalysis) {
            try {
                getVideoServerConnection().setAnalysisType(imageAnalyzer.getAnalysisType());
            } catch (ConnectionFailed e) {
                TineHandler.logger.log(Level.SEVERE,"Remote analysis type cannot be set.",e);
            }
        }
        performAnalysis();
    }

    public void setInterferometerNumOfImages(int interferometerNumOfImages) {
        int oldValue = this.interferometerNumOfImages;
        this.interferometerNumOfImages = interferometerNumOfImages;
        if (this.imageAnalyzer instanceof InterferometerImageAnalyzer) {
            ((InterferometerImageAnalyzer)this.imageAnalyzer).setNumOfImages(interferometerNumOfImages);
        }
        pcSupport.firePropertyChange(PROPERTY_INT_NUM_OF_IMAGES,oldValue,interferometerNumOfImages);
        if (!localAnalysis) {
            getVideoServerConnection().setInterferometerNumOfImages(this.interferometerNumOfImages);
        }
    }

    public int getInterferometerNumOfImages() {
        return interferometerNumOfImages;
    }

    /**
     * Sets the color map to use for the analysis.
     *
     * @param colorMap
     */
    public void setColorMap(ColorMap colorMap) {
        this.colorMap = colorMap;
        performAnalysis();
    }

    private synchronized void performAnalysis() {
        if (image == null) return;
        //imagewidth and image height are in nochange orientation
        int iw = imageWidth;
        int ih = imageHeight;
        int rx = roiX;
        int ry = roiY;
        int rh = roiH;
        int rw = roiW;
        int r2x = roi2X;
        int r2y = roi2Y;
        int r2h = roi2H;
        int r2w = roi2W;
        double thresh = threshold;
        boolean calcThresh = calculateThreshold;
        boolean fit = performFit;
        boolean smooth = performSmoothing;
        ColorMap cmap = colorMap;
        ColorDecoder cdec = decoder;
        IMAGE im = image;
        double[] background = preciseBackground;
        executeAnalysis(() -> {
            setAImage(imageAnalyzer.performAnalysis(cmap,cdec,im,background,iw,ih,rx,ry,rw,rh,r2x,r2y,r2w,r2h,thresh,
                    calcThresh,fit,smooth));
        });
    }

    public void resetFitting() {
        executeAnalysis(() -> imageAnalyzer.clear());
    }

    protected synchronized void executeAnalysis(Runnable r) {
        if (analysisExecutor == null) {
            ThreadFactory factory = new TINEThreadFactory("TINE-ImageAnalysis",false,true);
            analysisExecutor = new ThreadPoolExecutor(1,1,0,TimeUnit.NANOSECONDS,
                    new DismissableBlockingQueue<Runnable>(2,true),factory) {

                @Override
                protected void afterExecute(Runnable r, Throwable t) {
                    if (t != null) {
                        t.printStackTrace();
                    }
                }
            };
            analysisExecutor.setRejectedExecutionHandler((runable, exception) -> {});
        }
        analysisExecutor.execute(r);
    }

    private void executeTask(Runnable r) {
        if (executor == null) {
            executor = Executors.newSingleThreadExecutor();
        }
        executor.execute(r);
    }

    /**
     * Adds a </code>PropertyChangeListener</code> for the specified propertyName.
     *
     * @param propertyName the name of the property
     * @param listener the </code>PropertyChangeListener</code> to add
     */
    public void addPropertyChangeListener(String propertyName, PropertyChangeListener listener) {
        pcSupport.addPropertyChangeListener(propertyName,listener);
    }

    /**
     * Removes a </code>PropertyChangeListener</code> for the specified propertyName.
     *
     * @param propertyName the name of the property
     * @param listener the </code>PropertyChangeListener</code> to remove
     */
    public void removePropertyChangeListener(String propertyName, PropertyChangeListener listener) {
        pcSupport.removePropertyChangeListener(propertyName,listener);
    }

    /**
     * Adds a </code>PropertyChangeListener</code>.
     *
     * @param listener the </code>PropertyChangeListener</code> to add
     */
    public void addPropertyChangeListener(PropertyChangeListener listener) {
        pcSupport.addPropertyChangeListener(listener);
    }

    /**
     * Removes a </code>PropertyChangeListener</code>.
     *
     * @param listener the </code>PropertyChangeListener</code> to remove
     */
    public void removePropertyChangeListener(PropertyChangeListener listener) {
        pcSupport.removePropertyChangeListener(listener);
    }

    /**
     * Transforms the IMAGE to BufferedImage using the appropriate algorithm for the given image.
     *
     * @param image the image to transform
     * @param colorMap the color map to use in transformation
     * @return transformed image
     */
    public static BufferedImage toBufferedImage(IMAGE image, ColorMap colorMap) {
        if (TBufferedImage.isSupported(image.getFrameHeader().imageFormat)) {
            return TBufferedImage.toBufferedImage(image,colorMap,false,false);
        } else {
            return new TImageConverter().parseTineImage(image,false);
        }
    }

    /**
     * Transform BufferedImage to colorless double values, where each value represents a pixel. If the input is
     * grayscale, then the double value is the gray value normalized 256, if the image is RGB, decoder is used
     * to transform the image to RGB.
     *
     * @param image the image to transform
     * @param decoder the decode for transforming RGB values
     * @return array of double values representing colorless pixels
     */
    public static double[] imageToDoubleValues(BufferedImage image, int bitsPerPixels,
            ColorDecoder decoder) {
        int type = image.getType();
        int width = image.getWidth();
        int height = image.getHeight();
        int length = width*height;
        double[] values = new double[length];

        switch (type) {
            case BufferedImage.TYPE_BYTE_GRAY:
                byte[] bytes = ((DataBufferByte)image.getRaster().getDataBuffer()).getData();
                for (int i = 0; i < length; i++) {
                    values[i] = bytes[i];
                }
                break;
            case BufferedImage.TYPE_USHORT_GRAY:
            case BufferedImage.TYPE_CUSTOM:
                double factor = 256./Math.pow(2, bitsPerPixels);
                short[] shorts = ((DataBufferUShort)image.getRaster().getDataBuffer()).getData();
                for (int i = 0; i < length; i++) {
                    values[i] = factor*(shorts[i] & 0xffff);
                }
                break;
            case BufferedImage.TYPE_INT_RGB:
            case BufferedImage.TYPE_INT_ARGB:
                int[] ints = ((DataBufferInt)image.getRaster().getDataBuffer()).getData();
                values = decoder.transform(ints);
                break;
            case BufferedImage.TYPE_INT_BGR:
                int[] ints2 = ((DataBufferInt)image.getRaster().getDataBuffer()).getData();
                values = decoder.transformBGR(ints2);
                break;
            case BufferedImage.TYPE_3BYTE_BGR:
                byte[] byte3 = ((DataBufferByte) image.getRaster().getDataBuffer()).getData();
                int[] vals = new int[length];
                for (int i = 0; i < length; i++) {
                    vals[i] = (byte3[i*3+2] << 16) | (byte3[i*3+1] << 8) | (byte3[i*3]);
                }
                values = decoder.transform(vals);
                break;
            case BufferedImage.TYPE_4BYTE_ABGR:
                byte[] byte4 = ((DataBufferByte) image.getRaster().getDataBuffer()).getData();
                int[] valss = new int[length];
                for (int i = 0; i < length; i++) {
                    valss[i] = (byte4[i*3+3] << 16) | (byte4[i*3+2] << 8) | (byte4[i*3+1]);
                }
                values = decoder.transform(valss);
                break;
            default:
                int[] data = image.getRGB(0,0,width,height,null,0,width);
                values = decoder.transform(data);
            }
        return values;
    }
}
