package de.desy.acop.video.analysis;

/**
 * 
 * <code>LUDecomposition</code> provides a method to solve a n-by-n system of linear
 * equations. It makes a LU decomposition of the matrix (lower triangle L and upper 
 * triangle U) and provides a solution to an arbitrary system A*x=b, where b
 * is given by the user and x is returned. The implementation supports pivotting. 
 *
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 *
 */
public class LUDecomposition {

	/** The matrix */
	private double[][] LU;
	/** Dimension */
	private int m;
	/** Pivot sign */
	private int pivsign;
	/** Pivot vector */
	private int[] pivot;
	 

	/**
	 * Creates a new LUDecomposition object.
	 * 
	 * @param A the matrix to be decomposed
	 * @throws MathException if the matrix is singular
	 */
	public LUDecomposition(double[][] A) throws MathException {
		LU = getMatrixCopy(A);
		m = A.length;
		pivot = new int[m];
		for (int i = 0; i < m; i++) {
			pivot[i] = i;
		}
		pivsign = 1;
		decompose();
	}

	/**
	 * Performs the LU decomposition of the matrix.
	 * 
	 * @throws MathException if the matrix is singular
	 */
	private void decompose() throws MathException {
		double[] LUrowi;
		double[] LUcolj = new double[m];

		for (int j = 0; j < m; j++) {
			for (int i = 0; i < m; i++) {
				LUcolj[i] = LU[i][j];
			}
			for (int i = 0; i < m; i++) {
				LUrowi = LU[i];
				int kmax = Math.min(i,j);
				double s = 0.0;
				for (int k = 0; k < kmax; k++) {
					s += LUrowi[k] * LUcolj[k];
				}
				LUrowi[j] = LUcolj[i] -= s;
			}
			int p = j;
			for (int i = j + 1; i < m; i++) {
				if (Math.abs(LUcolj[i]) > Math.abs(LUcolj[p])) {
					p = i;
				}
			}
			if (p != j) {
				for (int k = 0; k < m; k++) {
					double t = LU[p][k];
					LU[p][k] = LU[j][k];
					LU[j][k] = t;
				}
				int k = pivot[p];
				pivot[p] = pivot[j];
				pivot[j] = k;
				pivsign = -pivsign;
			}

			if (j < m & LU[j][j] != 0.0) {
				for (int i = j + 1; i < m; i++) {
					LU[i][j] /= LU[j][j];
				}
			}
		}
		if (isSingular()) {
			throw new MathException("Matrix is singular.");
		}
	}

	/**
	 * Returns true if the given matrix is singular or false if not.
	 * 
	 * @return true if matrix is singular
	 */
	public boolean isSingular() {
		for (int j = 0; j < m; j++) {
			if (LU[j][j] == 0) return true;
		}
		return false;
	}
	
	/**
	 * Solves A*x = b
	 * 
	 * @param b vector of the dimension of this LU
	 * @return x so that L*U*x = b
	 * @throws MathException if the dimensions are not equal 
	 */
	public double[] solve(double[] b) throws MathException {
		if (b.length != m) {
			throw new MathException("Vector dimension must equal LU dimension (" + m +").");
		}

		double[] x = new double[pivot.length];
        for (int i = 0; i < x.length; i++) {
       	 	x[i] = b[pivot[i]];
        }
		
		// Solve L*Y = B(piv,:)
		for (int k = 0; k < m; k++) {
			for (int i = k + 1; i < m; i++) {
				x[i] -= x[k] * LU[i][k];
			}
		}
		// Solve U*X = Y;
		for (int k = m - 1; k >= 0; k--) {
			x[k] /= LU[k][k];
			for (int i = 0; i < k; i++) {
				x[i] -= x[k] * LU[i][k];
			}
		}
		return x;
	}

	/**
	 * Returns the copy of the given matrix
	 * @param A the matrix to be copied 
	 * @return a copy of A
	 */
	private double[][] getMatrixCopy(double[][] A) {
		int m = A.length;
		double[][] C = new double[m][m];
		for (int i = 0; i < m; i++) {
			for (int j = 0; j < m; j++) {
				C[i][j] = A[i][j];
			}
		}
		return C;
	}
}
