package de.desy.acop.video.displayer;

import java.util.Arrays;

/**
 * Represents a Color LookUp Table (CLUT) for TINE images.
 * 
 * @author David Melkumyan, DESY Zeuthen
 * 
 */
public class ImageCLUT {

	private int nBits[];
	private int pixel_bits;
	private int numComponents;
	private int maxBits;
	private int maxValue;

	private ColorMap colorMap;

	public ImageCLUT(ColorMap colorMap, int[] nBits) {
		if (colorMap == null)
			throw new NullPointerException("colorMap == null!");
		if (nBits == null)
			throw new NullPointerException("nBits == null!");

		numComponents = nBits.length;
		if (numComponents != 1 && numComponents != 3)
			throw new IllegalArgumentException("numComponents == " + numComponents);
		if (numComponents == 3 && !ColorMap.NONE.equals(colorMap))
			throw new IllegalArgumentException("colorMap = " + colorMap + " for numComponents = " + numComponents);

		this.colorMap = colorMap;
		this.nBits = nBits;

		maxBits = nBits[0];
		maxValue = (1 << maxBits) - 1;
		for (int bits : nBits) {
			if (bits > maxBits)
				maxBits = bits;
			pixel_bits += bits;
		}
	}

	/**
	 * Returns the color map
	 * 
	 * @return the color map
	 */
	public ColorMap getColorMap() {
		return colorMap;
	}

	/**
	 * Returns the number of bits for the specified color component.
	 * 
	 * @param componentIdx
	 *            the index of the color component
	 * @return the number of bits for the color component at the specified
	 *         index.
	 * @throws ArrayIndexOutOfBoundsException
	 *             if <code>componentIdx</code> is greater than the number of
	 *             components or less than zero
	 * @throws NullPointerException
	 *             if the number of bits array is <code>null</code>
	 */
	public int getComponentSize(int componentIdx) {
		return nBits[componentIdx];
	}

	/**
	 * Returns an array of the number of bits per color component.
	 * 
	 * @return an array of the number of bits per color component
	 */
	public int[] getComponentSize() {
		return nBits.clone();
	}

	/**
	 * Returns the number of color components
	 * 
	 * @return the number of color components
	 */
	public int getNumComponents() {
		return numComponents;
	}

	/**
	 * Returns the number of bits per pixel.
	 * 
	 * @return the number of bits per pixel
	 */
	public int getPixelSize() {
		return pixel_bits;
	}

	/**
	 * Returns true if number of color components array is equal to {8, 8, 8},
	 * otherwise false.
	 * 
	 * @return true if number of color components array is equal to {8, 8, 8},
	 *         otherwise false
	 */
	public boolean isRGB() {
		return Arrays.equals(nBits, new int[] { 8, 8, 8 });
	}

	/**
	 * Returns true if number of color components is equal to 1, otherwise false
	 * 
	 * @return true if number of color components is equal to 1, otherwise false
	 */
	public boolean isGreyscale() {
		return (numComponents == 1);
	}

	public int getMaxBits() {
		return maxBits;
	}

	public int getMinValue() {
		return 0;
	}
	
	public int getMaxValue() {
		return maxValue;
	}

	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		sb.append("colorMap = ").append(colorMap);
		sb.append(", nBits = ").append(Arrays.toString(nBits));
		return sb.toString();
	}

}
