package de.desy.acop.video.displayer;

import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.LogRecord;
import java.util.logging.Logger;
import java.util.logging.StreamHandler;

/**
 * Images counter. Calculates dropped frames and dropped frames percentage for
 * each new subsequent incoming frame
 * 
 * @author sweisse, mdavid
 * 
 */
public class ImageCounter {

	private static final Logger logger = Logger.getLogger(ImageCounter.class.getSimpleName());

	static {
		LogFormatter formatter = new LogFormatter();
		Handler handler = new StreamHandler(System.out, formatter) {
			@Override
			public synchronized void publish(LogRecord record) {
				super.publish(record);
				flush();
			}
		};
		handler.setLevel(Level.INFO);
		logger.setLevel(Level.INFO);
		Handler[] handlers = logger.getHandlers();
		for (Handler h : handlers) {
			logger.removeHandler(h);
		}
		logger.addHandler(handler);
		logger.setUseParentHandlers(false);
	}

	/**
	 * logged if there is a "big" frame number gap (missing frames in between
	 * two received frames)
	 */
	private static final long BIG_FRAME_NUMBER_GAP = 20;

	/**
	 * last received frame number. This value is updated if a video image is
	 * scheduled for displaying. It is reset to -1 on startup of live transfer.
	 */
	private long _lastFrameNumber = -1;

	/**
	 * total dropped frames counter. This value is updated if a video image is
	 * scheduled for displaying. It is reset to zero on startup of live
	 * transfer.
	 */
	private long _droppedFrames;

	/**
	 * total subsequent received frames counter. This value is updated if a
	 * video image is scheduled for displaying. It is reset to zero on startup
	 * of live transfer.
	 */
	private long _receivedFrames;

	/**
	 * ratio (percentage) of dropped vs. total number of received frames. This
	 * value is updated if a video images is scheduled for displaying.
	 */
	private double _ratio;

	public ImageCounter() {
		super();
	}

	public long getLastFrameNumber() {
		return _lastFrameNumber;
	}

	public void setLastFrameNumber(long lastFrameNumber) {
		_lastFrameNumber = lastFrameNumber;
	}

	public void calculate(long frameNumber) {
		_receivedFrames++;

		if (frameNumber < 0) {
			// (may be) because the maximum integer number is exceeded.
			// reset of dropped frames counter is not allowed.
			logger.warning("ImageCounter: WARN: negative number of frame: " + frameNumber);

		} else if (frameNumber <= _lastFrameNumber) {
			logger.warning("ImageCounter: WARN: current frame number (" + frameNumber + ") is less/equal than/to last frame number (" + frameNumber + ")");
			reset(); // reset counter
			calculate(frameNumber); // recalculate dropped frames number

		} else { // (frameNumber > _lastFrameNumber)
			if (_lastFrameNumber == -1) {
				_lastFrameNumber = frameNumber;
				_droppedFrames = 0;

			} else {
				long diff = frameNumber - _lastFrameNumber;
				_droppedFrames += (diff - 1);
				_lastFrameNumber = frameNumber;
				if (diff > BIG_FRAME_NUMBER_GAP)
					logger.warning("ImageCounter: WARN: more than " + BIG_FRAME_NUMBER_GAP + " missing frames: " + diff);
			}
		}
	}

	public long getDroppedFrames() {
		return _droppedFrames;
	}

	public long getReceivedFrames() {
		return _receivedFrames;
	}

	public double getRatio() {
		if (_receivedFrames != 0)
			_ratio = (double) _droppedFrames / (double) _receivedFrames;

		return _ratio;
	}

	public double getDroppedPercent() {
		long allFrames = _droppedFrames + _receivedFrames;
		return (allFrames == 0) ? 0.0 : (double) _droppedFrames / (double) allFrames;
	}

	/**
	 * Resets image counter
	 */
	public void reset() {
		_lastFrameNumber = -1;
		_droppedFrames = 0;
		_receivedFrames = 0;
		_ratio = 0.0;
	}
}
