package de.desy.acop.video.displayer;

/**
 * Image flag enumeration
 * 
 * @author mdavid
 * 
 */
public enum ImageFlag {
	/**
	 * if set, appended video data is ordered little endian (on video format types
	 * where this is applicable)
	 */
	LITTLE_ENDIAN_BYTE_ORDER(0x1),

	/**
	 * If set, original image data (first digital image data) can be restored out of
	 * the appended image bits
	 */
	IMAGE_LOSSLESS(0x2),

	/**
	 * If set, image was flipped around horizontal axis on its way
	 */
	HORIZONTAL_FLIP_180(0x4),

	/**
	 * If set, image was flipped around vertical axis on its way
	 */
	VERTICAL_FLIP_180(0x8),

	/**
	 * If set, a software region of interest (ROI) has been applied to the image
	 * header and bits
	 */
	SOFTWARE_ROI_APPLIED(0x200);

	private static final String LITTLEENDIAN = "LITTLEENDIAN";
	private static final String BIGENDIAN = "BIGENDIAN";
	private static final String LOSSLESS = "LOSSLESS";
	private static final String LOSSY = "LOSSY";
	private static final String HORIZONTALFLIP180 = "HORIZONTAL_FLIP_180";
	private static final String VERTICALFLIP180 = "VERTICAL_FLIP_180";
	private static final String SOFTWAREROIAPPLIED = "SOFTWARE_ROI_APPLIED";

	private final int id;

	ImageFlag(int id) {
		this.id = id;
	}

	public int getId() {
		return id;
	}

	/**
	 * Decodes flags integer into human readable string format
	 * 
	 * @param flags - image flags
	 * @return string representation of the passed aFlags (frameHeader.imageFlags)
	 */
	public static String toString(int flags) {
		String str = (isSet(flags, ImageFlag.LITTLE_ENDIAN_BYTE_ORDER) ? LITTLEENDIAN : BIGENDIAN);
		str += (" " + (isSet(flags, ImageFlag.IMAGE_LOSSLESS) ? LOSSLESS : LOSSY));
		if (isSet(flags, ImageFlag.HORIZONTAL_FLIP_180)) {
			str += (" " + HORIZONTALFLIP180);
		}
		if (isSet(flags, ImageFlag.VERTICAL_FLIP_180)) {
			str += (" " + VERTICALFLIP180);
		}
		if (isSet(flags, ImageFlag.SOFTWARE_ROI_APPLIED)) {
			str += (" " + SOFTWAREROIAPPLIED);
		}
		return str;
	}

	/**
	 * Checks whether flag is set in the flags
	 * 
	 * @param flags - image flags
	 * @param flag  - flag to check
	 * @return true is flag is set in flags, otherwise false
	 */
	private static boolean isSet(int flags, ImageFlag flag) {
		return (flags & flag.getId()) != 0;
	}
}
