package de.desy.acop.video.displayer;

import java.awt.Image;
import java.awt.image.BufferedImage;
import java.io.BufferedWriter;
import java.io.DataInputStream;
import java.io.EOFException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;
import java.net.URL;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;

import javax.imageio.ImageIO;

import de.desy.acop.video.timageio.TImageUtil;
import de.desy.tine.types.IMAGE;

/**
 * Common TINE IMAGE encoder/decoder class.
 * 
 * @author David Melkumyan, DESY Zeuthen
 * @changed 17.05.2015
 */
public class ImageParser {

	private static final String LS = System.lineSeparator();

	private static ImageParser SINGLETON;
	static {
		SINGLETON = new ImageParser();
	}

	public static ImageParser instance() {
		return SINGLETON;
	}

	public static void refresh() {
		SINGLETON = new ImageParser();
	}

	private ImageParser() {
	}

	/**
	 * convenience overload, defaults grayscale conversion to false
	 * 
	 * @see loadImageFile( String, IMAGE, boolean )
	 * @deprecated use {@link #loadImageFile(URL, IMAGE, boolean)} instead
	 */
	@Deprecated
	public static final boolean loadImageFile(String aFileName, IMAGE aOutImage) {
		return loadImageFile(aFileName, aOutImage, false);
	}

	/**
	 * loads an image file with a give filename from disk and puts its contents to TINE IMAGE data type <br>
	 * 
	 * @param aFileName
	 *            file name (may include path)
	 * @param aOutImage
	 *            TINE IMAGE data type to put image contents to
	 * @param aChangeToGrey
	 *            if true, image data is down scaled to grayscale before returned
	 * @return true - image contents were successfully put into aOutImage<br>
	 *         false - some error (aOutImage is null, IOException, URISyntaxException, error img.getRGB)
	 * 
	 * @deprecated use {@link #loadImageFile(URL, IMAGE, boolean)} instead
	 */
	@Deprecated
	public static final boolean loadImageFile(String aFileName, IMAGE aOutImage, boolean aChangeToGrey) {
		if (aOutImage == null)
			return false;

		// (1) load file into java buffered image
		BufferedImage img = null;
		File imgFile = null;
		try {
			imgFile = new File(aFileName);
			img = ImageIO.read(imgFile);
		} catch (IOException ex) {
			return false;
		}

		return postProcess(imgFile, img, aOutImage, aChangeToGrey);
	}

	/**
	 * Loads an image file with a give filename from disk and puts its contents to TINE IMAGE data type <br>
	 * 
	 * @param url
	 *            input a <code>URL</code> to read from.
	 * @param outImage
	 *            TINE IMAGE data type to put image contents to
	 * @param toGrey
	 *            if true, image data is down scaled to grayscale before returned
	 * @return true - image contents were successfully put into aOutImage<br>
	 *         false - some error (aOutImage is null, IOException, URISyntaxException, error img.getRGB)
	 */
	public static final boolean loadImageFile(URL url, IMAGE outImage, boolean toGrey) {
		try {
			return postProcess(url.toString(), ImageIO.read(url), outImage, toGrey);

		} catch (IOException ignored) {
		}
		return false;
	}

	/*
	 * no javadoc
	 * 
	 * postprocesses the java buffered image and creates a VSv3 transport layer compatible image out of it,
	 * color-to-grayscale can be performed in this step on demand, File class is necessary for file name string only
	 * 
	 * returns false in the following cases img.getRGB error
	 */
	@Deprecated
	private static final boolean postProcess(File imgFile, BufferedImage img, IMAGE aOutImage, boolean aChangeToGrey) {
		// (1) obtain TYPE_INT_ARGB pixel array out of "img"

		int imgWidth = img.getWidth();
		int imgHeight = img.getHeight();

		int[] pixelArr = new int[imgWidth * imgHeight];

		if (img.getRGB(0, 0, imgWidth, imgHeight, pixelArr, 0, imgWidth) == null)
			return false;

		// (2) create TINE IMAGE type out of it

		IMAGE.FrameHeader frameHeader = aOutImage.getFrameHeader();
		IMAGE.SourceHeader sourceHeader = aOutImage.getSourceHeader();

		if (aChangeToGrey == true) {
			frameHeader.bytesPerPixel = 1;
			frameHeader.effectiveBitsPerPixel = 8;
			frameHeader.imageFormat = ImageFormat.IMAGE_FORMAT_GRAY.getId();
			frameHeader.sourceFormat = ImageFormat.IMAGE_FORMAT_GRAY.getId();
		} else {
			// keep RGB
			frameHeader.bytesPerPixel = 3;
			frameHeader.effectiveBitsPerPixel = 24;
			frameHeader.imageFormat = ImageFormat.IMAGE_FORMAT_RGB.getId();
			frameHeader.sourceFormat = ImageFormat.IMAGE_FORMAT_RGB.getId();
		}

		frameHeader.appendedFrameSize = imgWidth * imgHeight * frameHeader.bytesPerPixel;

		// mdavid: commented
		// frameHeader.imageFlags = (int)
		// VideoHeaderV3.CF_IMAGE_FLAG_IMAGE_LOSSLESS
		// | (int) VideoHeaderV3.CF_IMAGE_FLAG_LITTLE_ENDIAN_BYTE_ORDER;

		// mdavid: changed
		frameHeader.imageFlags = ImageFlag.IMAGE_LOSSLESS.getId() | ImageFlag.LITTLE_ENDIAN_BYTE_ORDER.getId();

		frameHeader.sourceHeight = imgHeight;
		frameHeader.sourceWidth = imgWidth;
		frameHeader.xScale = 1;
		frameHeader.yScale = 1;

		// TODO: could be too long and truncated badly
		sourceHeader.cameraPortName = "file://" + imgFile.getName();

		long msTimeEpoch = imgFile.lastModified();
		sourceHeader.timestampMicroseconds = (int) ((msTimeEpoch % 1000L) * 1000L);
		sourceHeader.timestampSeconds = (int) (msTimeEpoch / 1000L);

		sourceHeader.totalLength = IMAGE.HEADER_SIZE + frameHeader.appendedFrameSize;

		byte[] buf = new byte[frameHeader.appendedFrameSize];

		// imperformant, but fast-to-implement solution:
		// Note: maybe make faster!

		if (aChangeToGrey == true) {
			for (int i = 0; i < frameHeader.sourceWidth * frameHeader.sourceHeight; i++) {
				double r = (pixelArr[i] >> 16) & 0xff;
				double g = (pixelArr[i] >> 8) & 0xff;
				double b = pixelArr[i] & 0xff;
				buf[i] = (byte) (0.114 * b + 0.587 * g + 0.299 * r);
			}
		} else {
			int pos = 0;
			for (int i = 0; i < frameHeader.sourceWidth * frameHeader.sourceHeight; i++) {
				buf[pos++] = (byte) ((pixelArr[i] >> 16) & 0xff); // R
				buf[pos++] = (byte) ((pixelArr[i] >> 8) & 0xff); // G
				buf[pos++] = (byte) (pixelArr[i] & 0xff); // B
			}
		}

		aOutImage.setImageFrameBuffer(buf);

		return true;
	}

	/*
	 * no javadoc
	 * 
	 * postprocesses the java buffered image and creates a VSv3 transport layer compatible image out of it,
	 * color-to-grayscale can be performed in this step on demand, File class is necessary for file name string only
	 * 
	 * returns false in the following cases img.getRGB error
	 */
	@SuppressWarnings("deprecation")
	private static final boolean postProcess(String defltCameraPortName, BufferedImage img, IMAGE aOutImage, boolean aChangeToGrey) {
		// (1) obtain TYPE_INT_ARGB pixel array out of "img"

		int imgWidth = img.getWidth();
		int imgHeight = img.getHeight();

		int[] pixelArr = new int[imgWidth * imgHeight];

		if (img.getRGB(0, 0, imgWidth, imgHeight, pixelArr, 0, imgWidth) == null)
			return false;

		// (2) create TINE IMAGE type out of it

		IMAGE.FrameHeader frameHeader = aOutImage.getFrameHeader();
		IMAGE.SourceHeader sourceHeader = aOutImage.getSourceHeader();

		if (aChangeToGrey == true) {
			frameHeader.bytesPerPixel = 1;
			frameHeader.effectiveBitsPerPixel = 8;
			frameHeader.imageFormat = ImageFormat.IMAGE_FORMAT_GRAY.getId();
			frameHeader.sourceFormat = ImageFormat.IMAGE_FORMAT_GRAY.getId();
		} else {
			// keep RGB
			frameHeader.bytesPerPixel = 3;
			frameHeader.effectiveBitsPerPixel = 24;
			frameHeader.imageFormat = ImageFormat.IMAGE_FORMAT_RGB.getId();
			frameHeader.sourceFormat = ImageFormat.IMAGE_FORMAT_RGB.getId();
		}

		frameHeader.appendedFrameSize = imgWidth * imgHeight * frameHeader.bytesPerPixel;

		// mdavid: commented
		// frameHeader.imageFlags = (int)
		// VideoHeaderV3.CF_IMAGE_FLAG_IMAGE_LOSSLESS
		// | (int) VideoHeaderV3.CF_IMAGE_FLAG_LITTLE_ENDIAN_BYTE_ORDER;

		// mdavid: changed
		frameHeader.imageFlags = ImageFlag.IMAGE_LOSSLESS.getId() | ImageFlag.LITTLE_ENDIAN_BYTE_ORDER.getId();

		frameHeader.sourceHeight = imgHeight;
		frameHeader.sourceWidth = imgWidth;
		frameHeader.xScale = 1;
		frameHeader.yScale = 1;

		// TODO: could be too long and truncated badly
		sourceHeader.cameraPortName = defltCameraPortName;

		long msTimeEpoch = System.currentTimeMillis();
		sourceHeader.timestampMicroseconds = (int) ((msTimeEpoch % 1000) * 1000);
		sourceHeader.timestampSeconds = (int) (msTimeEpoch / 1000);

		sourceHeader.totalLength = IMAGE.HEADER_SIZE + frameHeader.appendedFrameSize;

		byte[] buf = new byte[frameHeader.appendedFrameSize];

		// imperformant, but fast-to-implement solution:
		// Note: maybe make faster!

		if (aChangeToGrey == true) {
			for (int i = 0; i < frameHeader.sourceWidth * frameHeader.sourceHeight; i++) {
				double r = (pixelArr[i] >> 16) & 0xff;
				double g = (pixelArr[i] >> 8) & 0xff;
				double b = pixelArr[i] & 0xff;
				buf[i] = (byte) (0.114 * b + 0.587 * g + 0.299 * r);
			}
		} else {
			int pos = 0;
			for (int i = 0; i < frameHeader.sourceWidth * frameHeader.sourceHeight; i++) {
				buf[pos++] = (byte) ((pixelArr[i] >> 16) & 0xff); // R
				buf[pos++] = (byte) ((pixelArr[i] >> 8) & 0xff); // G
				buf[pos++] = (byte) (pixelArr[i] & 0xff); // B
			}
		}

		aOutImage.setImageFrameBuffer(buf);
		return true;
	}

	/**
	 * loads an IMM (proprietary, VSV2, raw, greyscale) image file with a give filename from disk and puts its contents
	 * to TINE IMAGE data type. An IMM file can consist of multiple video frames glued together. In such case, only the
	 * first one is loaded! <br>
	 * 
	 * @param aFileName
	 *            file name (may include path)
	 * @param aOutImage
	 *            TINE IMAGE data type to put image contents to
	 * @return true - image contents were successfully put into aOutImage<br>
	 *         false - some error (aOutImage is null, IOException)
	 */
	@SuppressWarnings("deprecation")
	public static final boolean loadIMM(String aFileName, IMAGE aOutImage) {
		if (aOutImage == null)
			return false;

		int width;
		int height;
		int bytesperpixel;
		int bitsperpixel;
		double scale;
		byte[] imagebits = null;

		// read the full file into buffered array

		// read in first four bytes
		// read in second four bytes

		// Analyze
		// check whether file length is big enough
		// read in image bits

		// read in 8 bytes -> double scale

		width = -1;
		height = -1;
		bytesperpixel = -1;
		bitsperpixel = -1;
		scale = -1.0;

		try (FileInputStream fis = new FileInputStream(aFileName); DataInputStream dis = new DataInputStream(fis)) {
			long fsize = fis.getChannel().size();

			width = dis.readUnsignedShort();
			width = ((width & 0xff) << 8) + (width >> 8);
			bytesperpixel = dis.readUnsignedShort();
			bytesperpixel = ((bytesperpixel & 0xff) << 8) + (bytesperpixel >> 8);
			height = dis.readUnsignedShort();
			height = ((height & 0xff) << 8) + (height >> 8);
			bitsperpixel = dis.readUnsignedShort();
			bitsperpixel = ((bitsperpixel & 0xff) << 8) + (bitsperpixel >> 8);

			bytesperpixel /= 8;
			if (bytesperpixel == 0)
				bytesperpixel = 1;

			if ((bitsperpixel == 0) || (bitsperpixel > (8 * bytesperpixel)))
				bitsperpixel = bytesperpixel * 8;

			// now we can read in data, file size should be bigger than
			// width*height+16

			if (fsize < (width * height * bytesperpixel + 16))
				throw new EOFException();

			imagebits = new byte[width * height * bytesperpixel];

			dis.read(imagebits, 0, imagebits.length);

			byte[] buf = new byte[8];

			dis.read(buf, 0, buf.length);

			int off = 0;
			long scale_long = (((long) buf[off + 0]) & 0xFF) | ((((long) buf[off + 1]) & 0xFF) << 8) | ((((long) buf[off + 2]) & 0xFF) << 16)
					| ((((long) buf[off + 3]) & 0xFF) << 24) | ((((long) buf[off + 4]) & 0xFF) << 32) | ((((long) buf[off + 5]) & 0xFF) << 40)
					| ((((long) buf[off + 6]) & 0xFF) << 48) | ((((long) buf[off + 7]) & 0xFF) << 56);

			scale = Double.longBitsToDouble(scale_long);
		} catch (EOFException eof) {
			// System.out.println( "EOF reached " );
			return false;
		} catch (IOException ioe) {
			// System.out.println( "IO error: " + ioe );
			return false;
		}

		IMAGE.FrameHeader frameHeader = aOutImage.getFrameHeader();
		IMAGE.SourceHeader sourceHeader = aOutImage.getSourceHeader();

		frameHeader.bytesPerPixel = bytesperpixel;
		frameHeader.effectiveBitsPerPixel = bitsperpixel;
		frameHeader.imageFormat = ImageFormat.IMAGE_FORMAT_GRAY.getId();
		frameHeader.sourceFormat = frameHeader.imageFormat;

		frameHeader.appendedFrameSize = width * height * frameHeader.bytesPerPixel;

		// mdavid: changed
		frameHeader.imageFlags |= ImageFlag.IMAGE_LOSSLESS.getId();

		frameHeader.sourceHeight = height;
		frameHeader.sourceWidth = width;
		frameHeader.xScale = (float) scale;
		frameHeader.yScale = (float) scale;

		File imgFile = new File(aFileName);

		// TODO: could be too long and truncated badly
		sourceHeader.cameraPortName = "file://" + imgFile.getName();

		long msTimeEpoch = imgFile.lastModified();
		sourceHeader.timestampMicroseconds = (int) ((msTimeEpoch % 1000) * 1000);
		sourceHeader.timestampSeconds = (int) (msTimeEpoch / 1000);

		sourceHeader.totalLength = IMAGE.HEADER_SIZE + frameHeader.appendedFrameSize;

		aOutImage.setImageFrameBuffer(imagebits);

		return true;
	}

	/**
	 * prints all formats that Java ImageIO class is able to read to stdout. This helps when getting to know which file
	 * formats are supported.
	 */
	public static void printLoadableFormats() {
		String[] names = ImageIO.getReaderFormatNames();
		for (String name : names)
			System.out.println(name);
	}

	/**
	 * convenience overload, accepts filename as String instead of File class
	 * 
	 * @see write( File, Image, VideoHeaderV3)
	 */
	// public static boolean write(String aFileName, Image aImg, VideoHeaderV3
	// aHdr) { // mdavid: commented
	public static boolean write(String aFileName, Image aImg, IMAGE timage) {
		return write(new File(aFileName), aImg, timage);
	}

	/**
	 * saves the java.awt.Image passed to a newly created (overwritten) PNG file. Additional metadata describing the
	 * Image is taken from a special header and put into an additional textfile placed beside the PNG file. <br>
	 * 
	 * @param aFile
	 *            java.io.File where to store the png image
	 * @param aImg
	 *            Image data to be written to PNG
	 * @param aHdr
	 *            contains meta properties of Image that will be written to .txt file
	 * @return true - image contents and text file were successfully saved<br>
	 *         false - some error (IOException on png or txt)
	 */
	public static boolean write(File aFile, Image aImg, IMAGE timage) {
		try {
			ImageIO.write(TImageUtil.toBufferedImage(aImg), "png", aFile);

			// add .txt header afterwards
			String wholename = aFile.getAbsolutePath();
			wholename += ".txt";

			File txtF = new File(wholename);
			txtF.createNewFile();
			try (Writer output = new BufferedWriter(new FileWriter(txtF))) {
				// use buffering
				// FileWriter always assumes default encoding is OK!

				DateFormat dfmt = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.SSSZ");

				IMAGE.SourceHeader srcH = timage.getSourceHeader();
				IMAGE.FrameHeader frmH = timage.getFrameHeader();

				output.write("SourceHeader." + LS);
				output.write(" .version             " + srcH.versionTag + LS);
				output.write(" .timestamp           " + dfmt.format(new Date((srcH.timestampSeconds * 1000L) + (srcH.timestampMicroseconds / 1000L))) + LS);
				output.write(" .cameraport          \"" + srcH.cameraPortName + "\"" + LS);
				output.write(" .total_length        " + srcH.totalLength + LS);

				output.write(" FrameHeader" + LS);
				output.write("       .sourceWidth         	" + frmH.sourceWidth + LS);
				output.write("       .sourceHeight        	" + frmH.sourceHeight + LS);
				output.write("       .aoiWidth     			" + frmH.aoiWidth + LS);
				output.write("       .aoiHeight    			" + frmH.aoiHeight + LS);
				output.write("       .xStart       			" + frmH.xStart + LS);
				output.write("       .yStart       			" + frmH.yStart + LS);
				output.write("       .bytesPerPixel    		" + frmH.bytesPerPixel + LS);
				output.write("       .effectiveBitsPerPixel 	" + frmH.effectiveBitsPerPixel + LS);
				output.write("       .horizontalBinning   	" + frmH.horizontalBinning + LS);
				output.write("       .verticalBinning  		" + frmH.verticalBinning + LS);
				output.write("       .sourceFormat    		" + ImageFormat.valueOf(frmH.sourceFormat) + LS);
				output.write("       .currentFormat    		" + ImageFormat.valueOf(frmH.imageFormat) + LS);
				output.write("       .frameNumber   			" + frmH.frameNumber + LS);
				output.write("       .eventNumer    			" + frmH.eventNumber + LS);
				output.write("       .xScale       			" + frmH.xScale + LS);
				output.write("       .yScale       			" + frmH.yScale + LS);
				output.write("       .imageRotation   		" + frmH.imageRotation + LS);

				// mdavid: commented
				// output.write("       .imageFlags   			" +
				// VideoHeaderV3.flagsToString(frmH.imageFlags) + n);

				// mdavid: changed
				output.write("       .imageFlags   			" + ImageFlag.toString(frmH.imageFlags) + LS);

				output.write("       .appendedFrameSize 		" + frmH.appendedFrameSize + LS);
				output.write("Header END" + LS);
			} catch (IOException ex1) {
				return false;

			}
		} catch (IOException ex) {
			return false;
		}

		return true;
	}
}
