package de.desy.acop.video.displayer;

import java.util.Date;

import de.desy.tine.types.IMAGE;

/**
 * <code>VideoHeaderV2</code> implements marshaling of a Video System v2 (VSv2)
 * data array (usually just received via the network) to a Video System v3 TINE
 * IMAGE datatype for usage with the rest of Video System v3 Java code.<br>
 * <br>
 * <b>Note:</b>This class is only a very basic construct and might be
 * substantially changed in future. It fits its purpose currently but will not
 * win a price for meaningful class design!
 * 
 * @author <a href="mailto:stefan.weisse@desy.de">Stefan Weisse</a>
 * @author <a href="mailto:david.melkumyan@desy.de">David Melkumyan</a>
 * @version $Id: Templates.xml,v 1.10 2008/06/20 16:10:13 sweisse Exp $
 * 
 */

public final class VideoHeaderV2 {

	/** size in bytes of a VSv2 header that each VSv2 data array starts off with */
	public static final int HDRSIZE = 88;

	/** video codec typical FOURCC (four character code) of HUFFYUV 'HFYU' */
	public static final long HUFFYUV_FOURCC = 0x55594648;

	/** setting of VSv2 header: magic at startup of header */
	private static final long KENNUNG = 0x11223344L; // 2147483647L

	/** maximum 32-bit integer value for validity checking */
	private static final long MAX_INT = 0x7FFFFFFFL; // 2415919103L

	/** setting of VSv2 header: length of appended data (image bits) */
	private int len_data;

	/**
	 * setting of VSv2 header: length of an additional format header between
	 * global header and image bits
	 */
	private int len_fmthdr;

	/** setting of VSv2 header: frame number */
	private long framenumber = -1L;

	/** transformed setting of VSv2 header: time at which image was taken */
	private Date timestamp = new Date(0L);

	/** setting of VSv2 header: width of video image in pixels */
	private int width = -1;

	/** setting of VSv2 header: height of video image in pixels */
	private int height = -1;

	/**
	 * setting of VSv2 header: is compressed (1) or uncompressed (0) data
	 * appended
	 */
	private boolean compressed;

	/**
	 * setting of VSv2 header: ratio (in x and y) between pixels and millimetres
	 */
	private double scale = 1D;

	/** transformed setting of VSv2 header: bytes per pixel of appended image */
	private int bypp = 1;

	/** setting of VSv2 header: effective bits per pixels of appended image */
	private int effective_bpp = 8;

	public VideoHeaderV2() {
		super();
	}

	/**
	 * interprets an input array (aImageV2blob) as a Video System v2 data block
	 * (header plus possibly additional format header plus image bits). In case
	 * validity checking succeeds, the contained image information is wrapped
	 * into a VSv3 TINE IMAGE datatype for use with all other Java Video code
	 * parts. As a Video System v2 header does not deliver a good string to be
	 * used as VSv3 header camera port (mounting place e.g.) identification,
	 * this member in IMAGE datatype is taken from aCamPort parameter.<br>
	 * <br>
	 * 
	 * @param cameraPortName
	 *            camera port description to be put in resulting IMAGE
	 *            aDestImage
	 * @param imageV2blob
	 *            input data set (should contain VSv2 image data block)
	 * @param destTineImage
	 *            TINE IMAGE datatype output
	 * @return <ul>
	 *         <li>true - success
	 *         <li>false - error importing blob (most probably validity checking
	 *         failed)
	 *         </ul>
	 */
	public boolean packageV2BlobIntoIMAGE(String cameraPortName, byte[] imageV2blob, IMAGE destTineImage) {
		if (!parseImageV2blob(imageV2blob))
			return false;

		IMAGE.FrameHeader frameHeader = destTineImage.getFrameHeader();
		IMAGE.SourceHeader sourceHeader = destTineImage.getSourceHeader();

		frameHeader.sourceFormat = ImageFormat.IMAGE_FORMAT_GRAY.getId();
		frameHeader.sourceWidth = width;
		frameHeader.sourceHeight = height;
		frameHeader.bytesPerPixel = bypp;
		frameHeader.effectiveBitsPerPixel = effective_bpp;
		frameHeader.aoiHeight = frameHeader.aoiWidth = -1;
		frameHeader.eventNumber = -1L;
		frameHeader.frameNumber = framenumber;

		frameHeader.imageFlags = ImageFlag.IMAGE_LOSSLESS.getId() | ImageFlag.LITTLE_ENDIAN_BYTE_ORDER.getId();

		if (compressed) {
			frameHeader.imageFormat = ImageFormat.IMAGE_FORMAT_HUFFYUV.getId();
			frameHeader.appendedFrameSize = len_data;
		} else {
			frameHeader.imageFormat = ImageFormat.IMAGE_FORMAT_GRAY.getId();
			frameHeader.appendedFrameSize = width * height * bypp;
		}

		sourceHeader.totalLength = IMAGE.HEADER_SIZE + frameHeader.appendedFrameSize;

		frameHeader.ispare1 = frameHeader.ispare2 = frameHeader.ispare3 = -1;
		frameHeader.fspare1 = frameHeader.fspare2 = frameHeader.fspare3 = -1F;
		frameHeader.horizontalBinning = frameHeader.verticalBinning = 0;

		frameHeader.imageRotation = 0F;
		frameHeader.xScale = frameHeader.yScale = (float) scale;
		frameHeader.xStart = frameHeader.yStart = 0;

		sourceHeader.baseTag = IMAGE.DEFAULT_BASE_TAG;
		sourceHeader.cameraPortName = cameraPortName;
		sourceHeader.cameraPortId = IMAGE.DEFAULT_CAMERA_PORT_ID;

		long msecs = timestamp.getTime();

		sourceHeader.timestampMicroseconds = (int) ((msecs % 1000L) * 1000L);
		sourceHeader.timestampSeconds = (int) (msecs / 1000L);
		sourceHeader.versionTag = IMAGE.IMAGE_VERSION;

		System.arraycopy(imageV2blob, HDRSIZE + len_fmthdr, destTineImage.getImageFrameBuffer(), 0, len_data);

		return true;
	}

	/**
	 * interprets an input array (buffer) as a Video System v2 header. In case
	 * validity checking succeeds, the contained image information is stored
	 * into local class variables.<br>
	 * <br>
	 * 
	 * @param buf
	 *            VSv2 header as byte array
	 * 
	 * @return <ul>
	 *         <li>true - success<br>
	 *         <li>false - error, no class variable is updated (most probably
	 *         validity checking failed)
	 *         </ul>
	 */
	public boolean parseImageV2blob(byte[] buf) {
		// strong validity checking!
		long l_kennung;
		long l_width;
		long l_height;
		long l_bypp;
		long l_effective_bpp;
		long l_len_total;
		long l_len_data;
		long l_len_fmthdr;
		long l_framenumber;
		long l_cameraport;
		long l_timestamp;
		long l_compressed;
		double l_scale;
		double l_framerate;

		if (buf.length < HDRSIZE)
			return false;

		int off = 0;
		l_kennung = (((long) buf[off + 0]) & 0xFF) | ((((long) buf[off + 1]) & 0xFF) << 8)
				| ((((long) buf[off + 2]) & 0xFF) << 16) | ((((long) buf[off + 3]) & 0xFF) << 24);

		off = 8;
		l_len_total = (((long) buf[off + 0]) & 0xFF) | ((((long) buf[off + 1]) & 0xFF) << 8)
				| ((((long) buf[off + 2]) & 0xFF) << 16) | ((((long) buf[off + 3]) & 0xFF) << 24);

		off = 12;
		l_len_data = (((long) buf[off + 0]) & 0xFF) | ((((long) buf[off + 1]) & 0xFF) << 8)
				| ((((long) buf[off + 2]) & 0xFF) << 16) | ((((long) buf[off + 3]) & 0xFF) << 24);

		off = 16;
		l_len_fmthdr = (((long) buf[off + 0]) & 0xFF) | ((((long) buf[off + 1]) & 0xFF) << 8)
				| ((((long) buf[off + 2]) & 0xFF) << 16) | ((((long) buf[off + 3]) & 0xFF) << 24);

		off = 20;
		l_framenumber = (((long) buf[off + 0]) & 0xFF) | ((((long) buf[off + 1]) & 0xFF) << 8)
				| ((((long) buf[off + 2]) & 0xFF) << 16) | ((((long) buf[off + 3]) & 0xFF) << 24);

		off = 24;
		long l_timestamp_sec = (((long) buf[off + 0]) & 0xFF) | ((((long) buf[off + 1]) & 0xFF) << 8)
				| ((((long) buf[off + 2]) & 0xFF) << 16) | ((((long) buf[off + 3]) & 0xFF) << 24);
		off = 28;
		long l_timestamp_msec = (((long) buf[off + 0]) & 0xFF) | ((((long) buf[off + 1]) & 0xFF) << 8);

		l_timestamp = (l_timestamp_sec * 1000L) + l_timestamp_msec;

		off = 36;
		l_cameraport = (((long) buf[off + 0]) & 0xFF) | ((((long) buf[off + 1]) & 0xFF) << 8)
				| ((((long) buf[off + 2]) & 0xFF) << 16) | ((((long) buf[off + 3]) & 0xFF) << 24);

		off = 40;
		l_width = (((long) buf[off + 0]) & 0xFF) | ((((long) buf[off + 1]) & 0xFF) << 8)
				| ((((long) buf[off + 2]) & 0xFF) << 16) | ((((long) buf[off + 3]) & 0xFF) << 24);

		off = 44;
		l_height = (((long) buf[off + 0]) & 0xFF) | ((((long) buf[off + 1]) & 0xFF) << 8)
				| ((((long) buf[off + 2]) & 0xFF) << 16) | ((((long) buf[off + 3]) & 0xFF) << 24);

		off = 48;
		l_compressed = (((long) buf[off + 0]) & 0xFF) | ((((long) buf[off + 1]) & 0xFF) << 8)
				| ((((long) buf[off + 2]) & 0xFF) << 16) | ((((long) buf[off + 3]) & 0xFF) << 24);

		off = 56;

		long l_framerate_long = (((long) buf[off + 0]) & 0xFF) | ((((long) buf[off + 1]) & 0xFF) << 8)
				| ((((long) buf[off + 2]) & 0xFF) << 16) | ((((long) buf[off + 3]) & 0xFF) << 24)
				| ((((long) buf[off + 4]) & 0xFF) << 32) | ((((long) buf[off + 5]) & 0xFF) << 40)
				| ((((long) buf[off + 6]) & 0xFF) << 48) | ((((long) buf[off + 7]) & 0xFF) << 56);

		l_framerate = Double.longBitsToDouble(l_framerate_long);

		off = 64;
		long l_scale_long = (((long) buf[off + 0]) & 0xFF) | ((((long) buf[off + 1]) & 0xFF) << 8)
				| ((((long) buf[off + 2]) & 0xFF) << 16) | ((((long) buf[off + 3]) & 0xFF) << 24)
				| ((((long) buf[off + 4]) & 0xFF) << 32) | ((((long) buf[off + 5]) & 0xFF) << 40)
				| ((((long) buf[off + 6]) & 0xFF) << 48) | ((((long) buf[off + 7]) & 0xFF) << 56);

		l_scale = Double.longBitsToDouble(l_scale_long);

		off = 72;
		l_bypp = (((long) buf[off + 0]) & 0xFF) | ((((long) buf[off + 1]) & 0xFF) << 8)
				| ((((long) buf[off + 2]) & 0xFF) << 16) | ((((long) buf[off + 3]) & 0xFF) << 24);
		l_bypp /= 8;

		off = 76;
		l_effective_bpp = (((long) buf[off + 0]) & 0xFF) | ((((long) buf[off + 1]) & 0xFF) << 8)
				| ((((long) buf[off + 2]) & 0xFF) << 16) | ((((long) buf[off + 3]) & 0xFF) << 24);

		// safety because of buggy servers of mine :/
		if (l_effective_bpp >= l_bypp * 8)
			l_effective_bpp = l_bypp * 8;

		// check and copy into real class variables in case of success
		if ((l_kennung == KENNUNG) && ((l_width > 0) && (l_width <= MAX_INT))
				&& ((l_height > 0) && (l_height <= MAX_INT)) && ((l_bypp == 1) || (l_bypp == 2) || (l_bypp == 4))
				&& (l_effective_bpp <= l_bypp * 8) && (l_len_total <= buf.length) && (l_len_data <= buf.length)
				&& (l_len_fmthdr <= buf.length) && (l_len_total == l_len_data + l_len_fmthdr + HDRSIZE)
				&& (l_framenumber > 0) && ((l_cameraport >= 0) && (l_cameraport <= MAX_INT)) && (l_timestamp > 0)
				&& ((l_compressed == 0) || (l_compressed == 1)) && (l_scale > 0.0) && (l_framerate > 0.0)) {

			width = (int) l_width;
			height = (int) l_height;
			bypp = (int) l_bypp;
			effective_bpp = (int) l_effective_bpp;
			len_data = (int) l_len_data;
			len_fmthdr = (int) l_len_fmthdr;
			framenumber = l_framenumber;
			timestamp = new Date(l_timestamp);
			scale = l_scale;
			compressed = (l_compressed == 1);
			return true;
		}
		return false;
	}
}
