package de.desy.acop.video.timageio;

import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.Arrays;
import java.util.Iterator;

import javax.imageio.IIOException;
import javax.imageio.IIOImage;
import javax.imageio.ImageIO;
import javax.imageio.ImageReadParam;
import javax.imageio.ImageReader;
import javax.imageio.ImageWriter;
import javax.imageio.metadata.IIOMetadata;
import javax.imageio.stream.ImageInputStream;
import javax.imageio.stream.ImageOutputStream;

/**
 * A class containing static convenience methods for performing simple encoding and decoding of TINE image.
 * 
 * @author David Melkumyan, DESY Zeuthen
 * @since October 2009
 * @version $Revision: 78693 $ : $LastChangedDate: 2017-10-13 05:02:04 +0200 (Fri, 13 Oct 2017) $
 */
public final class TImageIO {

	/**
	 * PNG image format name.
	 */
	private static final String IMAGE_FORMAT_PNG = "PNG";

	/**
	 * Constructor is private to prevent instantiation.
	 */
	private TImageIO() {
		super();
	}

	/**
	 * Returns a <code>TBufferedImage</code> as the result of decoding a supplied <code>File</code><br>
	 * with an <code>ImageReader</code> chosen automatically from among those currently registered.
	 * 
	 * <p>
	 * Note that there is no <code>read</code> method that takes a filename as a <code>String</code>;<br>
	 * use this method instead after creating a <code>File</code> from the filename.
	 * 
	 * @param input
	 *            a <code>File</code> to read from.
	 * 
	 * @return a <code>TBufferedImage</code> containing the decoded contents of the input, or <code>null</code>.
	 * 
	 * @exception NullPointerException
	 *                if <code>input</code> is <code>null</code>.
	 * @exception IOException
	 *                if an error occurs during reading.
	 */
	public static TBufferedImage read(File input) throws IOException {
		if (input == null) {
			throw new NullPointerException("input == null!");
		}
		if (!input.canRead()) {
			throw new IIOException("Can't read input file!");
		}
		ImageInputStream stream = ImageIO.createImageInputStream(input);
		if (stream == null) {
			throw new IIOException("Can't create an ImageInputStream!");
		}
		TBufferedImage tbi = read(stream);
		if (tbi == null) {
			stream.close();
		}
		return tbi;
	}

	/**
	 * Returns a <code>TBufferedImage</code> as the result of decoding a supplied <code>InputStream</code><br>
	 * with an <code>ImageReader</code> chosen automatically from among those currently registered.<br>
	 * The <code>InputStream</code> is wrapped in an <code>ImageInputStream</code>.<br>
	 * If no registered <code>ImageReader</code> claims to be able to read the resulting stream, <code>null</code> is
	 * returned.
	 * 
	 * <p>
	 * The current cache settings from <code>getUseCache</code>and <code>getCacheDirectory</code><br>
	 * will be used to control caching in the <code>ImageInputStream</code> that is created.
	 * 
	 * <p>
	 * This method does not attempt to locate <code>ImageReader</code>s that can read directly from an
	 * <code>InputStream</code>;<br>
	 * that may be accomplished using <code>IIORegistry</code> and <code>ImageReaderSpi</code>.
	 * 
	 * <p>
	 * This method <em>does not</em> close the provided <code>InputStream</code> after the read operation has completed;
	 * <br>
	 * it is the responsibility of the caller to close the stream, if desired.
	 * 
	 * @param input
	 *            an <code>InputStream</code> to read from.
	 * 
	 * @return a <code>TBufferedImage</code> containing the decoded contents of the input, or <code>null</code>.
	 * 
	 * @exception NullPointerException
	 *                if <code>input</code> is <code>null</code>.
	 * @exception IOException
	 *                if an error occurs during reading.
	 */
	public static TBufferedImage read(InputStream input) throws IOException {
		if (input == null) {
			throw new NullPointerException("input == null!");
		}
		ImageInputStream stream = ImageIO.createImageInputStream(input);
		TBufferedImage tbi = read(stream);
		if (tbi == null) {
			stream.close();
		}
		return tbi;
	}

	/**
	 * Returns a <code>TBufferedImage</code> as the result of decoding a supplied <code>URL</code><br>
	 * with an <code>ImageReader</code> chosen automatically from among those currently registered.<br>
	 * An <code>InputStream</code> is obtained from the <code>URL</code>, which is wrapped in an
	 * <code>ImageInputStream</code>.<br>
	 * If no registered <code>ImageReader</code> claims to be able to read the resulting stream, <code>null</code> is
	 * returned.
	 * 
	 * <p>
	 * The current cache settings from <code>getUseCache</code>and <code>getCacheDirectory</code><br>
	 * will be used to control caching in the <code>ImageInputStream</code> that is created.
	 * 
	 * <p>
	 * This method does not attempt to locate <code>ImageReader</code>s that can read directly from a <code>URL</code>;
	 * that may be accomplished using <code>IIORegistry</code> and <code>ImageReaderSpi</code>.
	 * 
	 * @param input
	 *            a <code>URL</code> to read from.
	 * 
	 * @return a <code>TBufferedImage</code> containing the decoded contents of the input, or <code>null</code>.
	 * 
	 * @exception NullPointerException
	 *                if <code>input</code> is <code>null</code>.
	 * @exception IOException
	 *                if an error occurs during reading.
	 */
	public static TBufferedImage read(URL input) throws IOException {
		if (input == null) {
			throw new NullPointerException("input == null!");
		}
		InputStream istream = input.openStream();
		ImageInputStream stream = ImageIO.createImageInputStream(istream);
		TBufferedImage tbi = read(stream);
		if (tbi == null) {
			stream.close();
		}
		return tbi;
	}

	/**
	 * Returns a <code>TBufferedImage</code> as the result of decoding a supplied <code>ImageInputStream</code><br>
	 * with an <code>ImageReader</code> chosen automatically from among those currently registered.<br>
	 * If no registered <code>ImageReader</code> claims to be able to read the stream, <code>null</code> is returned.
	 * 
	 * <p>
	 * Unlike most other methods in this class, this method <em>does</em> close the provided
	 * <code>ImageInputStream</code><br>
	 * after the read operation has completed, unless <code>null</code> is returned,<br>
	 * in which case this method <em>does not</em> close the stream.
	 * 
	 * @param stream
	 *            an <code>ImageInputStream</code> to read from.
	 * 
	 * @return a <code>TBufferedImage</code> containing the decoded contents of the input, or <code>null</code>.
	 * 
	 * @exception NullPointerException
	 *                if <code>stream</code> is <code>null</code>.
	 * @exception IOException
	 *                if an error occurs during reading.
	 */
	public static TBufferedImage read(ImageInputStream stream) throws IOException {
		if (stream == null) {
			throw new NullPointerException("stream == null!");
		}
		Iterator<ImageReader> iter = ImageIO.getImageReaders(stream);
		if (!iter.hasNext()) {
			return null;
		}
		ImageReader reader = iter.next();
		ImageReadParam param = reader.getDefaultReadParam();
		reader.setInput(stream, true, false);
		BufferedImage image;
		IIOMetadata metadata;
		try {
			image = reader.read(0, param);
			metadata = reader.getImageMetadata(0);

		} finally {
			reader.dispose();
			stream.close();
		}
		return new TBufferedImage(image, new TImageMetadata(metadata));
	}

	public static boolean write(TBufferedImage tbi, File file) throws IOException {
		if (file == null) {
			throw new NullPointerException("file == null!");
		}
		boolean res = write(tbi, ImageIO.createImageOutputStream(file));
		if (res && !file.exists()) {
			throw new IOException("Write succeed but output file is not exists: " + file);
		}
		return res;
	}

	/**
	 * @param tbi
	 *            - TINE buffered input image
	 * @param stream
	 *            - Image output stream
	 * @return true if write succeed otherwise false
	 * @throws IOException
	 *             if an I/O error occurs during writing.
	 */
	public static boolean write(TBufferedImage tbi, ImageOutputStream stream) throws IOException {
		if (tbi == null) {
			throw new NullPointerException("tbi == null!");
		}
		if (stream == null) {
			throw new NullPointerException("stream == null!");
		}
		BufferedImage bi = tbi.getBufferedImage();
		if (tbi.getMetadata() == null) {
			write(bi, stream);
		}
		ImageWriter writer = null;
		try {
			IIOMetadata metadata = tbi.getMetadata().toPngMetadata();
			if (metadata == null) {
				write(bi, stream);
			}
			IIOImage ioImage = new IIOImage(bi, null, metadata);
			writer = getWriter();
			writer.setOutput(stream);
			writer.write(metadata, ioImage, null);

		} finally {
			if (writer != null) {
				writer.dispose();
			}
			stream.flush();
		}
		return true;
	}

	/**
	 * @param bi
	 *            - Java buffered input image
	 * @param stream
	 *            - PNG output stream
	 * @return true if write succeed otherwise false
	 * @throws IOException
	 *             if an I/O error occurs during writing.
	 */
	public static boolean write(BufferedImage bi, ImageOutputStream stream) throws IOException {
		if (bi == null) {
			throw new NullPointerException("bi == null!");
		}
		return ImageIO.write(bi, IMAGE_FORMAT_PNG, stream);
	}

	/**
	 * @param bi
	 *            - Java buffered input image
	 * @param stream
	 *            - PNG output file
	 * @return true if write succeed otherwise false
	 * @throws IOException
	 *             if an I/O error occurs during writing.
	 */
	public static boolean write(BufferedImage bi, File file) throws IOException {
		if (bi == null) {
			throw new NullPointerException("bi == null!");
		}
		return ImageIO.write(bi, IMAGE_FORMAT_PNG, file);
	}

	/**
	 * Returns registered PNG writer.
	 * 
	 * @return PNG writer
	 * @throws IOException
	 *             if an I/O error occurs during getting PNG writer.
	 */
	private static ImageWriter getWriter() throws IOException {
		ImageWriter writer = null;
		Iterator<ImageWriter> iter = ImageIO.getImageWritersByFormatName(IMAGE_FORMAT_PNG); // "png"
		if (iter == null || (writer = iter.next()) == null) {
			throw new IOException("PNG writer was not found in: " + Arrays.asList(ImageIO.getWriterFormatNames()));
		}
		return writer;
	}
}
