package de.desy.acop.video.timageio;

import java.awt.image.BufferedImage;
import java.math.BigInteger;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.Arrays;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;

import javax.imageio.metadata.IIOMetadata;
import javax.imageio.metadata.IIOMetadataNode;

import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import de.desy.acop.video.displayer.ImageFlag;
import de.desy.acop.video.displayer.ImageFormat;
import de.desy.tine.types.IMAGE;
import de.desy.tine.types.IMAGE.FrameHeader;
import de.desy.tine.types.IMAGE.SourceHeader;

/**
 * Represents metadata (non-image data) associated with TINE images.
 *
 * @author David Melkumyan, DESY Zeuthen
 * @version $Revision: 91076 $ : $LastChangedDate: 2021-01-19 20:37:15 +0100 (Tue, 19 Jan 2021) $
 */
public class TImageMetadata {

    private static final String DEFLT_SOFTWARE = "TImageIO - Java TINE Image I/O API for loading/saving TINE images (betta ver. 0.0.02)";
    private static final String DEFLT_COMMENT = "Betta version: 0.0.03";

    public static final String KEY_SOFTWARE = "Software";
    public static final String KEY_COMMENT = "Comment";

    public static final String KEY_CAMERA_PORT_ID = "VSv3.hdr.src.cameraPortId";
    public static final String KEY_VERSION_TAG = "VSv3.hdr.src.versionTag";
    public static final String KEY_TIMESTAMP_SECONDS = "VSv3.hdr.src.timestampSeconds";
    public static final String KEY_TIMESTAMP_MICROSECONDS = "VSv3.hdr.src.timestampMicroseconds";
    public static final String KEY_CAMERA_PORT_NAME = "VSv3.hdr.src.cameraPortName";
    public static final String KEY_SOURCE_WIDTH = "VSv3.hdr.frm.sourceWidth";
    public static final String KEY_SOURCE_HEIGHT = "VSv3.hdr.frm.sourceHeight";
    public static final String KEY_AOI_WIDTH = "VSv3.hdr.frm.aoiWidth";
    public static final String KEY_AOI_HEIGHT = "VSv3.hdr.frm.aoiHeight";
    public static final String KEY_XSTART = "VSv3.hdr.frm.xStart";
    public static final String KEY_YSTART = "VSv3.hdr.frm.yStart";
    public static final String KEY_BYTES_PER_PIXEL = "VSv3.hdr.frm.bytesPerPixel";
    public static final String KEY_EFFECTIVE_BITS_PER_PIXEL = "VSv3.hdr.frm.effectiveBitsPerPixel";
    public static final String KEY_HORIZONTAL_BINNING = "VSv3.hdr.frm.horizontalBinning";
    public static final String KEY_VERTICAL_BINNING = "VSv3.hdr.frm.verticalBinning";
    public static final String KEY_SOURCE_FORMAT = "VSv3.hdr.frm.sourceFormat";
    public static final String KEY_IMAGE_FORMAT = "VSv3.hdr.frm.imageFormat";
    public static final String KEY_FRAME_NUMBER = "VSv3.hdr.frm.frameNumber";
    public static final String KEY_EVENT_NUMBER = "VSv3.hdr.frm.eventNumber";
    public static final String KEY_XSCALE = "VSv3.hdr.frm.xScale";
    public static final String KEY_YSCALE = "VSv3.hdr.frm.yScale";
    public static final String KEY_IMAGE_ROTATION = "VSv3.hdr.frm.imageRotation";
    public static final String KEY_IMAGE_FLAGS = "VSv3.hdr.frm.imageFlags";
    public static final String KEY_APPENDED_FRAME_SIZE = "VSv3.hdr.frm.appendedFrameSize";
    public static final String KEY_MD5HASH = "VSV3.frameBuffer.md5hash";
    public static final String KEY_ISPARE1 = "VSv3.hdr.frm.ispare1";
    public static final String KEY_ISPARE2 = "VSv3.hdr.frm.ispare2";
    public static final String KEY_ISPARE3 = "VSv3.hdr.frm.ispare3";
    public static final String KEY_FSPARE1 = "VSv3.hdr.frm.fspare1";
    public static final String KEY_FSPARE2 = "VSv3.hdr.frm.fspare2";
    public static final String KEY_FSPARE3 = "VSv3.hdr.frm.fspare3";
    public static final String KEY_OPT_TEXT = "VSv3.opt.text";
    public static final String KEY_VTOB = "vtOb";
    public static final String KEY_IHDR = "IHDR";
    public static final String KEY_SBIT = "sBIT";
    public static final String KEY_TIME = "tIME";
    public static final String KEY_COLOR_TYPE = "colorType";
    public static final String KEY_WIDTH = "width";
    public static final String KEY_HEIGHT = "height";
    public static final String KEY_BIT_DPETH = "bitDepth";
    public static final String KEY_GRAY = "gray";
    public static final String KEY_PHYS = "pHYs";
    public static final String KEY_UNIT_SPECIFIER = "unitSpecifier";
    public static final String KEY_UNKNOWN = "unknown";
    public static final String KEY_PIXELS_PER_UNIT_X_AXIS = "pixelsPerUnitXAxis";
    public static final String KEY_PIXELS_PER_UNIT_Y_AXIS = "pixelsPerUnitYAxis";
    public static final String KEY_UNKNOWN_CHUNKS = "UnknownChunks";
    public static final String KEY_TYPE = "type";
    public static final String KEY_KEYWORD = "keyword";
    public static final String KEY_VALUE = "value";
    public static final String KEY_TEXT = "tEXt";

    public static final String nativeMetadataFormatName = "javax_imageio_png_1.0";
    private static final List<String> IHDRColorNames = Arrays.asList(
        "Grayscale", null, "RGB", "Palette","GrayAlpha", null, "RGBAlpha");

    // /**
    // * PNG Color Type:
    // * <p>
    // * <ul>
    // * <li>PNG_COLOR_GRAY = 0
    // * <li>PNG_COLOR_RGB = 2
    // * <li>PNG_COLOR_PALETTE = 3
    // * <li>PNG_COLOR_GRAY_ALPHA = 4
    // * <li>PNG_COLOR_RGB_ALPHA = 6
    // * </ul>
    // */
    // private int pngColorType = -1;

    /**
     * optional
     */
    public String software = DEFLT_SOFTWARE;

    /**
     * optional
     */
    public String comment = DEFLT_COMMENT;

    /**
     * @see de.desy.tine.types.TBufferedImage.SourceHeader#cameraPortId
     */
    public Long cameraPortId;

    /**
     * @see de.desy.tine.types.TBufferedImage.SourceHeader#versionTag
     */
    public Long versionTag;

    /**
     * @see de.desy.tine.types.TBufferedImage.SourceHeader#timestampSeconds
     */
    public Integer timestampSeconds;

    /**
     * @see de.desy.tine.types.TBufferedImage.SourceHeader#timestampMicroseconds
     */
    public Integer timestampMicroseconds;

    /**
     * @see de.desy.tine.types.TBufferedImage.SourceHeader#cameraPortName
     */
    public String cameraPortName;

    /**
     * @see de.desy.tine.types.TBufferedImage.FrameHeader#sourceWidth
     */
    public Integer sourceWidth;

    /**
     * @see de.desy.tine.types.TBufferedImage.FrameHeader#sourceHeight
     */
    public Integer sourceHeight;

    /**
     * @see de.desy.tine.types.TBufferedImage.FrameHeader#aoiWidth
     */
    public Integer aoiWidth;

    /**
     * @see de.desy.tine.types.TBufferedImage.FrameHeader#aoiHeight
     */
    public Integer aoiHeight;

    /**
     * @see de.desy.tine.types.TBufferedImage.FrameHeader#xStart
     */
    public Integer xStart;

    /**
     * @see de.desy.tine.types.TBufferedImage.FrameHeader#yStart
     */
    public Integer yStart;

    /**
     * @see de.desy.tine.types.TBufferedImage.FrameHeader#bytesPerPixel
     */
    public Integer bytesPerPixel;

    /**
     * @see de.desy.tine.types.TBufferedImage.FrameHeader#effectiveBitsPerPixel
     */
    public Integer effectiveBitsPerPixel;

    /**
     * @see de.desy.tine.types.TBufferedImage.FrameHeader#horizontalBinning
     */
    public Integer horizontalBinning;

    /**
     * @see de.desy.tine.types.TBufferedImage.FrameHeader#verticalBinning
     */
    public Integer verticalBinning;

    /**
     * @see de.desy.tine.types.TBufferedImage.FrameHeader#sourceFormat
     */
    public ImageFormat sourceFormat;

    /**
     * @see de.desy.tine.types.TBufferedImage.FrameHeader#imageFormat
     */
    public ImageFormat imageFormat;

    /**
     * @see de.desy.tine.types.TBufferedImage.FrameHeader#frameNumber
     */
    public Long frameNumber;

    /**
     * @see de.desy.tine.types.TBufferedImage.FrameHeader#eventNumber
     */
    public Long eventNumber;

    /**
     * @see de.desy.tine.types.TBufferedImage.FrameHeader#xScale
     */
    public Float xScale;

    /**
     * @see de.desy.tine.types.TBufferedImage.FrameHeader#yScale
     */
    public Float yScale;

    /**
     * @see de.desy.tine.types.TBufferedImage.FrameHeader#imageRotation
     */
    public Float imageRotation;

    /**
     * @see de.desy.tine.types.TBufferedImage.FrameHeader#imageFlags
     */
    public Integer imageFlags;

    /**
     * @see de.desy.tine.types.TBufferedImage.FrameHeader#appendedFrameSize
     */
    public Integer appendedFrameSize;

    /**
     * The md5 hash of image bits.
     * <p>
     * An md5 checksum (32 chars of text) is calculated out of the first 'appendedFrameSize' bytes found in TImage's
     * appended frame buffer.
     */
    public String md5hash;

    /**
     * Optional user text tags.
     * <p>
     * The text tags are passed by the user as key value pairs, where each key and value must consist of only Latin1
     * characters (see PNG reference documentation on this subject).<br>
     * The key must be smaller or equal to 64 characters in length.<br>
     * Each key string is prefixed with "VSv3.opt.text." and then encoded to PNG tEXt chunk.
     */
    public Map<String, String> optVars = new HashMap<>();

    public int ispare1 = -1; // optional
    public int ispare2 = -1; // optional
    public int ispare3 = -1; // optional

    public float fspare1 = -1; // optional
    public float fspare2 = -1; // optional
    public float fspare3 = -1; // optional

    /**
     * Videosystem Three Optional Binary data.
     * <p>
     * The user/programmer has the option to save a binary object as a byte array.
     */
    private byte[] vtOb; // optional

    /**
     * Constructs TImageMetadata from BufferedImage with TINE image metadata in hashtable
     *
     * @param bi
     *            - Java BufferedImage with TINE image metadata in hashtable
     */
    public TImageMetadata(BufferedImage bi) {
        read(bi, true);
    }

    /**
     * @param bi
     *            - BufferedImage
     * @param rejectNonArchival
     *            - if <code>true</code> and <code>bi</code> contains invalid metadata throws
     *            <code>IllegalArgumentException</code>
     */
    public TImageMetadata(BufferedImage bi, boolean rejectNonArchival) {
        read(bi, rejectNonArchival);
    }

    /**
     * Creates image Metadata from known Image I/O Metadata.
     *
     * @param iioMetadata
     *            Image I/O Metadata
     *
     * @exception NullPointerException
     *                If <code>iioMetadata</code> is <code>null</code>.
     * @exception NullPointerException
     *                If <code>iioMetadata</code> is unsupported.
     */
    public TImageMetadata(IIOMetadata iioMetadata) {
        if (iioMetadata == null)
            throw new NullPointerException("iioMetadata == null!");
        read(iioMetadata);
    }

    /**
     * Creates image Metadata from TINE image.
     *
     * @param timage
     *            TINE Image
     */
    public TImageMetadata(IMAGE timage) {
        read(timage);
    }


    private static Node getNode(Node rootNode, String id) {
        NodeList nodes = rootNode.getChildNodes();
        for (int i = 0; i < nodes.getLength(); i++) {
            Node node = nodes.item(i);
            if (id.equalsIgnoreCase(node.getNodeName())) {
                return node;
            }
        }
        return null;
    }

    /**
     * Initializes metadata from PNG Metadata
     *
     * @param TPNGMetadata
     *            - PNG Metadata
     *
     * @exception NullPointerException
     *                If <code>TPNGMetadata</code> is <code>null</code>.
     * @exception NumberFormatException
     *                If the numerical string cannot be parsed as a number.
     * @exception IllegalArgumentException
     *                If the specified image format type has no constant with the specified name.
     *
     */
    private void read(IIOMetadata metadata) {
        if (metadata == null)
            throw new NullPointerException("TPNGMetadata == null!");
        Node rootNode = metadata.getAsTree(nativeMetadataFormatName);
        Node textNode = getNode(rootNode, KEY_TEXT);
        if (textNode != null) {
            NodeList nodes = textNode.getChildNodes();
            int size = nodes.getLength();
            for (int index = 0; index < size; index++) {
                Node node = nodes.item(index);

                String key = node.getAttributes().getNamedItem(KEY_KEYWORD).getNodeValue();
                String value = node.getAttributes().getNamedItem(KEY_VALUE).getNodeValue();

                if (KEY_SOFTWARE.equals(key)) {
                    software = value;
                } else if (KEY_COMMENT.equals(key)) {
                    comment = value;
                } else if (KEY_CAMERA_PORT_ID.equals(key)) {
                    cameraPortId = Long.valueOf(value);
                } else if (KEY_VERSION_TAG.equals(key)) {
                    versionTag = Long.decode(value);
                } else if (KEY_TIMESTAMP_SECONDS.equals(key)) {
                    timestampSeconds = Integer.valueOf(value);
                } else if (KEY_TIMESTAMP_MICROSECONDS.equals(key)) {
                    timestampMicroseconds = Integer.valueOf(value);
                } else if (KEY_CAMERA_PORT_NAME.equals(key)) {
                    cameraPortName = value;
                } else if (KEY_SOURCE_WIDTH.equals(key)) {
                    sourceWidth = Integer.valueOf(value);
                } else if (KEY_SOURCE_HEIGHT.equals(key)) {
                    sourceHeight = Integer.valueOf(value);
                } else if (KEY_AOI_WIDTH.equals(key)) {
                    aoiWidth = Integer.valueOf(value);
                } else if (KEY_AOI_HEIGHT.equals(key)) {
                    aoiHeight = Integer.valueOf(value);
                } else if (KEY_XSTART.equals(key)) {
                    xStart = Integer.valueOf(value);
                } else if (KEY_YSTART.equals(key)) {
                    yStart = Integer.valueOf(value);
                } else if (KEY_BYTES_PER_PIXEL.equals(key)) {
                    bytesPerPixel = Integer.valueOf(value);
                } else if (KEY_EFFECTIVE_BITS_PER_PIXEL.equals(key)) {
                    effectiveBitsPerPixel = Integer.valueOf(value);
                } else if (KEY_HORIZONTAL_BINNING.equals(key)) {
                    horizontalBinning = Integer.valueOf(value);
                } else if (KEY_VERTICAL_BINNING.equals(key)) {
                    verticalBinning = Integer.valueOf(value);
                } else if (KEY_SOURCE_FORMAT.equals(key)) {
                    sourceFormat = ImageFormat.valueOf(value);
                } else if (KEY_IMAGE_FORMAT.equals(key)) {
                    imageFormat = ImageFormat.valueOf(value);
                } else if (KEY_FRAME_NUMBER.equals(key)) {
                    frameNumber = Long.valueOf(value);
                } else if (KEY_EVENT_NUMBER.equals(key)) {
                    eventNumber = Long.valueOf(value);
                } else if (KEY_XSCALE.equals(key)) {
                    xScale = Float.valueOf(value);
                } else if (KEY_YSCALE.equals(key)) {
                    yScale = Float.valueOf(value);
                } else if (KEY_IMAGE_ROTATION.equals(key)) {
                    imageRotation = Float.valueOf(value);
                } else if (KEY_IMAGE_FLAGS.equals(key)) {
                    imageFlags = Integer.decode(value);
                } else if (KEY_APPENDED_FRAME_SIZE.equals(key)) {
                    appendedFrameSize = Integer.valueOf(value);
                } else if (KEY_MD5HASH.equals(key)) {
                    md5hash = (new BigInteger(value, 16)).toString(16);
                } else if (KEY_ISPARE1.equals(key)) {
                    ispare1 = Integer.parseInt(value);
                } else if (KEY_ISPARE2.equals(key)) {
                    ispare2 = Integer.parseInt(value);
                } else if (KEY_ISPARE3.equals(key)) {
                    ispare3 = Integer.parseInt(value);
                } else if (KEY_FSPARE1.equals(key)) {
                    fspare1 = Float.parseFloat(value);
                } else if (KEY_FSPARE2.equals(key)) {
                    fspare2 = Float.parseFloat(value);
                } else if (KEY_FSPARE3.equals(key)) {
                    fspare3 = Float.parseFloat(value);
                } else if (key.startsWith(KEY_OPT_TEXT)) {
                    if (optVars.containsKey(key))
                        throw new IllegalArgumentException("Duplicate keyword: " + key);
                    optVars.put(key, value);
                }
            }
        }


        Node unknownNode = getNode(rootNode,KEY_UNKNOWN_CHUNKS);
        if (unknownNode != null) {
            NodeList nodes = unknownNode.getChildNodes();
            // initialize V3 optional binary data
            int size = nodes.getLength();
            for (int index = 0; index < size; index++) {
                Node node = nodes.item(index);
                String type = node.getAttributes().getNamedItem(KEY_TYPE).getNodeValue();
                if (KEY_VTOB.equals(type))
                    vtOb = (byte[])((IIOMetadataNode)node).getUserObject();
            }
        }

        if (!isArchival())
            setDefault(rootNode);
        validateArchivalMetadata(rootNode);
    }

    /**
     * Initializes metadata from TINE Image
     *
     * @param timage
     *            TINE Image
     * @exception NullPointerException
     *                If <code>timage</code> is <code>null</code>.
     */
    private void read(IMAGE timage) throws IllegalArgumentException {
        if (timage == null)
            throw new NullPointerException("timage == null!");

        // ----------------------- Source Header -----------------------
        SourceHeader srcHeader = timage.getSourceHeader();
        cameraPortId = srcHeader.cameraPortId;
        cameraPortName = srcHeader.cameraPortName;
        versionTag = srcHeader.versionTag;
        timestampSeconds = srcHeader.timestampSeconds;
        timestampMicroseconds = srcHeader.timestampMicroseconds;

        // ------------------------ Frame Header ------------------------
        FrameHeader frameHeader = timage.getFrameHeader();
        sourceWidth = frameHeader.sourceWidth;
        sourceHeight = frameHeader.sourceHeight;
        aoiWidth = frameHeader.aoiWidth;
        aoiHeight = frameHeader.aoiHeight;
        xStart = frameHeader.xStart;
        yStart = frameHeader.yStart;
        bytesPerPixel = frameHeader.bytesPerPixel;
        effectiveBitsPerPixel = frameHeader.effectiveBitsPerPixel;
        horizontalBinning = frameHeader.horizontalBinning;
        verticalBinning = frameHeader.verticalBinning;
        sourceFormat = ImageFormat.valueOf(frameHeader.sourceFormat);
        imageFormat = ImageFormat.valueOf(frameHeader.imageFormat);
        frameNumber = frameHeader.frameNumber;
        eventNumber = frameHeader.eventNumber;
        xScale = frameHeader.xScale;
        yScale = frameHeader.yScale;
        imageRotation = frameHeader.imageRotation;
        imageFlags = frameHeader.imageFlags;
        appendedFrameSize = frameHeader.appendedFrameSize;
        ispare1 = frameHeader.ispare1;
        ispare2 = frameHeader.ispare2;
        ispare3 = frameHeader.ispare3;
        fspare1 = frameHeader.fspare1;
        fspare2 = frameHeader.fspare2;
        fspare3 = frameHeader.fspare3;

        md5hash = md5hash(timage.getImageFrameBuffer());
    }

    private Object readProperty(BufferedImage bi, String name, Class<?> clazz, Object deflt, boolean reject) throws IllegalArgumentException {
        Object obj = bi.getProperty(name);
        if (obj == null || obj == java.awt.Image.UndefinedProperty) {
            if (reject)
                throw new IllegalArgumentException("Undefined archival metadata: " + name);
            return deflt;
        }
        try {
            return clazz.cast(obj);
        } catch (ClassCastException e) {
            throw new IllegalArgumentException("Illegal archival metadata: " + name);
        }
    }

    /**
     * Initializes metadata from Java <code>BufferedImage</code>
     *
     * @param bi
     *            - Java BufferedImage
     * @param rejectNonArchival
     *            - if <code>true</code> and <code>bi</code> contains invalid metadata throws
     *            <code>IllegalArgumentException</code>
     * @exception NullPointerException
     *                If <code>image</code> is <code>null</code>.
     */
    private void read(BufferedImage bi, boolean rejectNonArchival) throws IllegalArgumentException {

        if (bi == null)
            throw new NullPointerException("image == null!");

        frameNumber = (Long) readProperty(bi, KEY_FRAME_NUMBER, Long.class, 0L, rejectNonArchival);

        eventNumber = (Long) readProperty(bi, KEY_EVENT_NUMBER, Long.class, 0L, rejectNonArchival);

        xStart = (Integer) readProperty(bi, KEY_XSTART, Integer.class, 0, rejectNonArchival);
        if (xStart < -1)
        	throw new IllegalArgumentException("Invalid archival metadata: xStart == " + xStart);

        yStart = (Integer) readProperty(bi, KEY_YSTART, Integer.class, 0, rejectNonArchival);
        if (yStart < -1)
        	throw new IllegalArgumentException("Invalid archival metadata: yStart == " + yStart);

        xScale = (Float) readProperty(bi, KEY_XSCALE, Float.class, -1F, rejectNonArchival);
        yScale = (Float) readProperty(bi, KEY_YSCALE, Float.class, -1F, rejectNonArchival);

        imageRotation = (Float) readProperty(bi, KEY_IMAGE_ROTATION, Float.class, 0F, rejectNonArchival);

        horizontalBinning = (Integer) readProperty(bi, KEY_HORIZONTAL_BINNING, Integer.class, 0, rejectNonArchival);
        verticalBinning = (Integer) readProperty(bi, KEY_VERTICAL_BINNING, Integer.class, 0, rejectNonArchival);

        imageFlags = (Integer) readProperty(bi, KEY_IMAGE_FLAGS, Integer.class, 0x1, rejectNonArchival);

        ispare1 = (Integer) readProperty(bi, KEY_ISPARE1, Integer.class, -1, rejectNonArchival);
        ispare2 = (Integer) readProperty(bi, KEY_ISPARE2, Integer.class, -1, rejectNonArchival);
        ispare3 = (Integer) readProperty(bi, KEY_ISPARE3, Integer.class, -1, rejectNonArchival);

        fspare1 = (Float) readProperty(bi, KEY_FSPARE1, Float.class, -1F, rejectNonArchival);
        fspare2 = (Float) readProperty(bi, KEY_FSPARE2, Float.class, -1F, rejectNonArchival);
        fspare3 = (Float) readProperty(bi, KEY_FSPARE3, Float.class, -1F, rejectNonArchival);

        cameraPortName = (String) readProperty(bi, KEY_CAMERA_PORT_NAME, String.class, IMAGE.DEFAULT_CAMERA_PORT_NAME, rejectNonArchival);
        cameraPortId = (Long) readProperty(bi, KEY_CAMERA_PORT_ID, Long.class, IMAGE.DEFAULT_CAMERA_PORT_ID, rejectNonArchival);

        versionTag = (Long) readProperty(bi, KEY_VERSION_TAG, Long.class, IMAGE.IMAGE_INITIAL_VERSION, rejectNonArchival);

        timestampSeconds = (Integer) readProperty(bi, KEY_TIMESTAMP_SECONDS, 
        		Integer.class, (int) (new GregorianCalendar().getTimeInMillis() / 1000),
                rejectNonArchival);

        timestampMicroseconds = (Integer) readProperty(bi, KEY_TIMESTAMP_MICROSECONDS, Integer.class, 0, rejectNonArchival);

        Integer value = (Integer) readProperty(bi, KEY_IMAGE_FORMAT, Integer.class, null, rejectNonArchival);
        imageFormat = (value == null ? getImageFormat(bi) : ImageFormat.valueOf(value));

        value = (Integer) readProperty(bi, KEY_SOURCE_FORMAT, Integer.class, null, rejectNonArchival);
        sourceFormat = (value == null ? imageFormat : ImageFormat.valueOf(value));

        bytesPerPixel = (Integer) readProperty(bi, KEY_BYTES_PER_PIXEL, Integer.class, null, rejectNonArchival);
        if (bytesPerPixel == null)
            bytesPerPixel = getBytesPerPixel(bi);

        if (bytesPerPixel < 1 || bytesPerPixel > 4)
        	throw new IllegalArgumentException("Invalid archival metadata: bytesPerPixel == " + bytesPerPixel);

        effectiveBitsPerPixel = (Integer) readProperty(bi, KEY_EFFECTIVE_BITS_PER_PIXEL, Integer.class, null, rejectNonArchival);
        if (effectiveBitsPerPixel == null)
            effectiveBitsPerPixel = getEffBytesPerPixel(bi);

        sourceWidth = (Integer) readProperty(bi, KEY_SOURCE_WIDTH, Integer.class, bi.getWidth(), rejectNonArchival);
        if (sourceWidth < -1)
        	throw new IllegalArgumentException("Invalid archival metadata: sourceWidth == " + sourceWidth);

        sourceHeight = (Integer) readProperty(bi, KEY_SOURCE_HEIGHT, Integer.class, bi.getHeight(), rejectNonArchival);
        if (sourceHeight < -1)
        	throw new IllegalArgumentException("Invalid archival metadata: sourceHeight == " + sourceHeight);

        aoiWidth = (Integer) readProperty(bi, KEY_AOI_WIDTH, Integer.class, -1, rejectNonArchival);
        if (aoiWidth < -1)
			throw new IllegalArgumentException("Invalid archival metadata: aoiWidth == " + aoiWidth);

		if (aoiWidth > 0 && (aoiWidth + xStart > sourceWidth)) 
			throw new IllegalArgumentException("Invalid archival metadata: aoiWidth:" + aoiWidth + " + xStart:" + xStart + " > sourceWidth:" + sourceWidth);

        aoiHeight = (Integer) readProperty(bi, KEY_AOI_HEIGHT, Integer.class, -1, rejectNonArchival);
        if (aoiHeight < -1)
			throw new IllegalArgumentException("Invalid archival metadata: aoiHeight == " + aoiHeight);

		if (aoiHeight > 0 && (aoiHeight + yStart > sourceHeight)) 
			throw new IllegalArgumentException("Invalid archival metadata: aoiHeight:" + aoiHeight + " + aoiHeight:" + yStart + " > sourceHeight:" + sourceHeight);

		Integer w = ((aoiWidth > 0) ? aoiWidth : sourceWidth);
		Integer h = ((aoiHeight > 0) ? aoiHeight : sourceHeight);
		appendedFrameSize = bytesPerPixel * w * h;
		if (appendedFrameSize <= 0)
			throw new IllegalArgumentException("Invalid appendedFrameSize: " + appendedFrameSize + ", bpp = " + bytesPerPixel + ", w = " + w  + ", h = " + h);
    }

    private ImageFormat getImageFormat(BufferedImage bi) {
        int imageType = bi.getType();
        switch (imageType) {
        case BufferedImage.TYPE_BYTE_GRAY:
        case BufferedImage.TYPE_USHORT_GRAY:
            return ImageFormat.IMAGE_FORMAT_GRAY;
        default:
            return ImageFormat.IMAGE_FORMAT_RGB;
        }
    }

    private int getBytesPerPixel(BufferedImage bi) {
        int imageType = bi.getType();
        switch (imageType) {
        case BufferedImage.TYPE_BYTE_GRAY:
            return 1;
        case BufferedImage.TYPE_USHORT_GRAY:
            return 2;
        default:
            return 3;
        }
    }

    private int getEffBytesPerPixel(BufferedImage bi) {
        int[] cs = bi.getColorModel().getComponentSize();
        int effBits = 0;
        for (int componentSize : cs)
            effBits += componentSize;
        return Math.min(effBits, 24);
    }

    public void write(IMAGE ti) {
        if (ti == null)
            throw new NullPointerException("ti == null!");

        // ********************* Frame Header ********************************
        IMAGE.FrameHeader frmHdr = ti.getFrameHeader();
        frmHdr.bytesPerPixel = bytesPerPixel;
        frmHdr.effectiveBitsPerPixel = effectiveBitsPerPixel;
        frmHdr.appendedFrameSize = appendedFrameSize;
        frmHdr.imageFormat = imageFormat.getId();
        frmHdr.sourceFormat = sourceFormat.getId();
        frmHdr.sourceWidth = sourceWidth;
        frmHdr.sourceHeight = sourceHeight;
        frmHdr.frameNumber = frameNumber;
        frmHdr.eventNumber = eventNumber;
        frmHdr.aoiHeight = aoiHeight;
        frmHdr.aoiWidth = aoiWidth;
        frmHdr.xStart = xStart;
        frmHdr.yStart = yStart;
        frmHdr.xScale = xScale;
        frmHdr.yScale = yScale;
        frmHdr.ispare1 = ispare1;
        frmHdr.ispare2 = ispare2;
        frmHdr.ispare3 = ispare3;
        frmHdr.fspare1 = fspare1;
        frmHdr.fspare2 = fspare2;
        frmHdr.fspare3 = fspare3;
        frmHdr.horizontalBinning = horizontalBinning;
        frmHdr.verticalBinning = verticalBinning;
        frmHdr.imageFlags = imageFlags;

        // ********************* Source Header ********************************
        SourceHeader srcHdr = ti.getSourceHeader();
        srcHdr.cameraPortName = cameraPortName;
        srcHdr.timestampSeconds = timestampSeconds;
        srcHdr.timestampMicroseconds = timestampMicroseconds;
        srcHdr.cameraPortId = cameraPortId;
        srcHdr.versionTag = versionTag;
        srcHdr.totalLength = IMAGE.HEADER_SIZE + frmHdr.appendedFrameSize;
    }

    public static Hashtable<String, Object> createProperties(IMAGE ti) {
        Hashtable<String, Object> props = new Hashtable<>(32);

        // ----------------------- Source Header -----------------------
        SourceHeader srcHdr = ti.getSourceHeader();
        props.put(KEY_CAMERA_PORT_ID, srcHdr.cameraPortId);
        props.put(KEY_CAMERA_PORT_NAME, srcHdr.cameraPortName);
        props.put(KEY_VERSION_TAG, srcHdr.versionTag);
        props.put(KEY_TIMESTAMP_SECONDS, srcHdr.timestampSeconds);
        props.put(KEY_TIMESTAMP_MICROSECONDS, srcHdr.timestampMicroseconds);

        // ------------------------ Frame Header ------------------------
        FrameHeader frmHdr = ti.getFrameHeader();
        props.put(KEY_SOURCE_WIDTH, frmHdr.sourceWidth);
        props.put(KEY_SOURCE_HEIGHT, frmHdr.sourceHeight);
        props.put(KEY_AOI_WIDTH, frmHdr.aoiWidth);
        props.put(KEY_AOI_HEIGHT, frmHdr.aoiHeight);
        props.put(KEY_XSTART, frmHdr.xStart);
        props.put(KEY_YSTART, frmHdr.yStart);
        props.put(KEY_BYTES_PER_PIXEL, frmHdr.bytesPerPixel);
        props.put(KEY_EFFECTIVE_BITS_PER_PIXEL, frmHdr.effectiveBitsPerPixel);
        props.put(KEY_HORIZONTAL_BINNING, frmHdr.horizontalBinning);
        props.put(KEY_VERTICAL_BINNING, frmHdr.verticalBinning);
        props.put(KEY_SOURCE_FORMAT, frmHdr.sourceFormat);
        props.put(KEY_IMAGE_FORMAT, frmHdr.imageFormat);
        props.put(KEY_FRAME_NUMBER, frmHdr.frameNumber);
        props.put(KEY_EVENT_NUMBER, frmHdr.eventNumber);
        props.put(KEY_XSCALE, frmHdr.xScale);
        props.put(KEY_YSCALE, frmHdr.yScale);
        props.put(KEY_IMAGE_ROTATION, frmHdr.imageRotation);
        props.put(KEY_IMAGE_FLAGS, frmHdr.imageFlags);
        props.put(KEY_APPENDED_FRAME_SIZE, frmHdr.appendedFrameSize);
        props.put(KEY_ISPARE1, frmHdr.ispare1);
        props.put(KEY_ISPARE2, frmHdr.ispare2);
        props.put(KEY_ISPARE3, frmHdr.ispare3);
        props.put(KEY_FSPARE1, frmHdr.fspare1);
        props.put(KEY_FSPARE2, frmHdr.fspare2);
        props.put(KEY_FSPARE3, frmHdr.fspare3);

        return props;
    }

    /**
     * Converts metadata to a PNG Metadata
     *
     * @return PNG Metadata
     */
    public IIOMetadata toPngMetadata() {
        TPNGMetadata TPNGMetadata = new TPNGMetadata();
        addText(TPNGMetadata, KEY_SOFTWARE, software);
        addText(TPNGMetadata, KEY_COMMENT, comment);
        if (isArchival()) {
            addText(TPNGMetadata, KEY_CAMERA_PORT_ID, cameraPortId.toString());
            addText(TPNGMetadata, KEY_VERSION_TAG, String.format("0x%01x", versionTag));
            addText(TPNGMetadata, KEY_TIMESTAMP_SECONDS, timestampSeconds.toString());
            addText(TPNGMetadata, KEY_TIMESTAMP_MICROSECONDS, timestampMicroseconds.toString());
            addText(TPNGMetadata, KEY_CAMERA_PORT_NAME, cameraPortName);
            addText(TPNGMetadata, KEY_SOURCE_WIDTH, sourceWidth.toString());
            addText(TPNGMetadata, KEY_SOURCE_HEIGHT, sourceHeight.toString());
            addText(TPNGMetadata, KEY_AOI_WIDTH, aoiWidth.toString());
            addText(TPNGMetadata, KEY_AOI_HEIGHT, aoiHeight.toString());
            addText(TPNGMetadata, KEY_XSTART, xStart.toString());
            addText(TPNGMetadata, KEY_YSTART, yStart.toString());
            addText(TPNGMetadata, KEY_BYTES_PER_PIXEL, bytesPerPixel.toString());
            addText(TPNGMetadata, KEY_EFFECTIVE_BITS_PER_PIXEL, effectiveBitsPerPixel.toString());
            addText(TPNGMetadata, KEY_HORIZONTAL_BINNING, horizontalBinning.toString());
            addText(TPNGMetadata, KEY_VERTICAL_BINNING, verticalBinning.toString());
            addText(TPNGMetadata, KEY_SOURCE_FORMAT, sourceFormat.toString());
            addText(TPNGMetadata, KEY_IMAGE_FORMAT, imageFormat.toString());
            addText(TPNGMetadata, KEY_FRAME_NUMBER, frameNumber.toString());
            addText(TPNGMetadata, KEY_EVENT_NUMBER, eventNumber.toString());
            addText(TPNGMetadata, KEY_XSCALE, xScale.toString());
            addText(TPNGMetadata, KEY_YSCALE, yScale.toString());
            addText(TPNGMetadata, KEY_IMAGE_ROTATION, imageRotation.toString());
            addText(TPNGMetadata, KEY_IMAGE_FLAGS, String.format("0x%01x", imageFlags |= ImageFlag.IMAGE_LOSSLESS.getId()));
            addText(TPNGMetadata, KEY_APPENDED_FRAME_SIZE, appendedFrameSize.toString());
            addText(TPNGMetadata, KEY_ISPARE1, Integer.toString(ispare1));
            addText(TPNGMetadata, KEY_ISPARE2, Integer.toString(ispare2));
            addText(TPNGMetadata, KEY_ISPARE3, Integer.toString(ispare3));
            addText(TPNGMetadata, KEY_FSPARE1, Float.toString(fspare1));
            addText(TPNGMetadata, KEY_FSPARE2, Float.toString(fspare2));
            addText(TPNGMetadata, KEY_FSPARE3, Float.toString(fspare3));

            if (imageFormat == ImageFormat.IMAGE_FORMAT_GRAY && effectiveBitsPerPixel < 8 * bytesPerPixel) {
                TPNGMetadata.sBIT_present = true;
                TPNGMetadata.sBIT_grayBits = effectiveBitsPerPixel;
            }

            if (xScale != -1 || yScale != -1) {
                TPNGMetadata.pHYs_present = true;
                TPNGMetadata.pHYs_unitSpecifier = 1;
                TPNGMetadata.pHYs_pixelsPerUnitXAxis = (int) (1000.0F * xScale);
                TPNGMetadata.pHYs_pixelsPerUnitYAxis = (int) (1000.0F * yScale);
            }
        }

        // add time chunk
        addTimeChunk(TPNGMetadata, System.currentTimeMillis());

        // add md5hash string
        // TODO: only for testing/developing of loading/saving images
        // if (!exportMode) // TODO: uncomment later
        if (md5hash != null)
            addText(TPNGMetadata, KEY_MD5HASH, md5hash);

        // add optional strings
        for (String key : optVars.keySet()) {
            String value = optVars.get(key);
            if (value != null)
                addText(TPNGMetadata, key, value);
        }

        // add vtOb chunk
        if (vtOb != null)
            addUserChunk(TPNGMetadata, KEY_VTOB, vtOb);

        return TPNGMetadata;
    }

    private static void addText(TPNGMetadata TPNGMetadata, String keyword, String text) {
        TPNGMetadata.tEXt_keyword.add(keyword);
        TPNGMetadata.tEXt_text.add(text);
    }

    private static void addUserChunk(TPNGMetadata TPNGMetadata, String keyword, byte[] data) {
        TPNGMetadata.unknownChunkType.add(keyword);
        TPNGMetadata.unknownChunkData.add(data);
    }

    private static void addTimeChunk(TPNGMetadata TPNGMetadata, long millis) {
        Calendar cal = new GregorianCalendar();
        cal.setTimeInMillis(millis);
        TPNGMetadata.tIME_year = cal.get(Calendar.YEAR);
        TPNGMetadata.tIME_month = cal.get(Calendar.MONTH) + 1;
        TPNGMetadata.tIME_day = cal.get(Calendar.DAY_OF_MONTH);
        TPNGMetadata.tIME_hour = cal.get(Calendar.HOUR_OF_DAY);
        TPNGMetadata.tIME_minute = cal.get(Calendar.MINUTE);
        TPNGMetadata.tIME_second = cal.get(Calendar.SECOND);
        TPNGMetadata.tIME_present = true;
    }

    @Override
    public String toString() {

        StringBuilder sb = new StringBuilder();

        if (software != null)
            sb.append(KEY_SOFTWARE).append(" = ").append(software).append("\n");
        if (comment != null)
            sb.append(KEY_COMMENT).append(" = ").append(comment).append("\n");

        if (versionTag != null)
            sb.append(KEY_VERSION_TAG).append(" = ").append(String.format("0x%01x", versionTag)).append("\n");
        if (cameraPortId != null)
            sb.append(KEY_CAMERA_PORT_ID).append(" = ").append(cameraPortId).append("\n");
        if (cameraPortName != null)
            sb.append(KEY_CAMERA_PORT_NAME).append(" = ").append(cameraPortName).append("\n");

        if (timestampSeconds != null)
            sb.append(KEY_TIMESTAMP_SECONDS).append(" = ").append(timestampSeconds).append("\n");
        if (timestampMicroseconds != null)
            sb.append(KEY_TIMESTAMP_MICROSECONDS).append(" = ").append(timestampMicroseconds).append("\n");

        if (sourceFormat != null)
            sb.append(KEY_SOURCE_FORMAT).append(" = ").append(sourceFormat).append("\n");
        if (sourceWidth != null)
            sb.append(KEY_SOURCE_WIDTH).append(" = ").append(sourceWidth).append("\n");
        if (sourceHeight != null)
            sb.append(KEY_SOURCE_HEIGHT).append(" = ").append(sourceHeight).append("\n");

        if (aoiWidth != null)
            sb.append(KEY_AOI_WIDTH).append(" = ").append(aoiWidth).append("\n");
        if (aoiHeight != null)
            sb.append(KEY_AOI_HEIGHT).append(" = ").append(aoiHeight).append("\n");

        if (xStart != null)
            sb.append(KEY_XSTART).append(" = ").append(xStart).append("\n");
        if (yStart != null)
            sb.append(KEY_YSTART).append(" = ").append(yStart).append("\n");

        if (bytesPerPixel != null)
            sb.append(KEY_BYTES_PER_PIXEL).append(" = ").append(bytesPerPixel).append("\n");
        if (effectiveBitsPerPixel != null)
            sb.append(KEY_EFFECTIVE_BITS_PER_PIXEL).append(" = ").append(effectiveBitsPerPixel).append("\n");

        if (horizontalBinning != null)
            sb.append(KEY_HORIZONTAL_BINNING).append(" = ").append(horizontalBinning).append("\n");
        if (verticalBinning != null)
            sb.append(KEY_VERTICAL_BINNING).append(" = ").append(verticalBinning).append("\n");

        if (imageFormat != null)
            sb.append(KEY_IMAGE_FORMAT).append(" = ").append(imageFormat).append("\n");

        if (frameNumber != null)
            sb.append(KEY_FRAME_NUMBER).append(" = ").append(frameNumber).append("\n");
        if (eventNumber != null)
            sb.append(KEY_EVENT_NUMBER).append(" = ").append(eventNumber).append("\n");

        if (xScale != null)
            sb.append(KEY_XSCALE).append(" = ").append(xScale).append("\n");
        if (yScale != null)
            sb.append(KEY_YSCALE).append(" = ").append(yScale).append("\n");

        if (imageRotation != null)
            sb.append(KEY_IMAGE_ROTATION).append(" = ").append(imageRotation).append("\n");
        if (imageFlags != null)
            sb.append(KEY_IMAGE_FLAGS).append(" = ").append(String.format("0x%01x", imageFlags)).append("\n");

        if (appendedFrameSize != null)
            sb.append(KEY_APPENDED_FRAME_SIZE).append(" = ").append(appendedFrameSize).append("\n");

        sb.append(KEY_ISPARE1).append(" = ").append(ispare1).append("\n");
        sb.append(KEY_ISPARE2).append(" = ").append(ispare2).append("\n");
        sb.append(KEY_ISPARE3).append(" = ").append(ispare3).append("\n");

        sb.append(KEY_FSPARE1).append(" = ").append(ispare1).append("\n");
        sb.append(KEY_FSPARE2).append(" = ").append(ispare2).append("\n");
        sb.append(KEY_FSPARE3).append(" = ").append(ispare3).append("\n");

        if (md5hash != null)
            sb.append(KEY_MD5HASH).append(" = ").append(TImageUtil.padString(md5hash, 32, '0', true)).append("\n");

        if (vtOb != null)
            sb.append(KEY_VTOB).append(" = ").append(new String(vtOb)).append("\n");

        for (String key : optVars.keySet())
            sb.append(key).append(" = ").append(optVars.get(key)).append("\n");

        return sb.toString();
    }

    /**
     * Check whether metadata fulfills "archival" image requirements
     *
     * @exception IllegalArgumentException
     *                if <code>metadata</code> is <code>non-archival</code>.
     */
    private boolean validateArchivalMetadata(Node rootNode) throws IllegalArgumentException {

        int bits = bytesPerPixel * 8;
        int effbits = effectiveBitsPerPixel;
        if (effbits > bits)
            throw new IllegalArgumentException("effectiveBitsPerPixel == " + effbits + ", max: " + bits);

        Node hdrNode = getNode(rootNode,KEY_IHDR);
        if (hdrNode == null) {
            throw new IllegalArgumentException("IHDR metadata not present.");
        }
        Node sBitNode = getNode(rootNode,KEY_SBIT);

        NamedNodeMap attributes = hdrNode.getAttributes();
        String colorType = attributes.getNamedItem(KEY_COLOR_TYPE).getNodeValue();
        int pngColorType = IHDRColorNames.indexOf(colorType);
        int width = Integer.parseInt(attributes.getNamedItem(KEY_WIDTH).getNodeValue());
        int height = Integer.parseInt(attributes.getNamedItem(KEY_HEIGHT).getNodeValue());
        int bitDepth = Integer.parseInt(attributes.getNamedItem(KEY_BIT_DPETH).getNodeValue());
        switch (pngColorType) {
        case 0: // "Grayscale"
            if (imageFormat != ImageFormat.IMAGE_FORMAT_GRAY)
                throw new IllegalArgumentException("imageFormat == " + imageFormat + ", expected: " + ImageFormat.IMAGE_FORMAT_GRAY);

            if (effbits != bitDepth)
                throw new IllegalArgumentException("bitDepth == " + effbits + ", expected: " + bitDepth);

            if (sBitNode != null) {
                Node node = sBitNode.getChildNodes().item(0);
                Node grayBits = node.getAttributes().getNamedItem(KEY_GRAY);
                if (grayBits != null && effbits != Integer.parseInt(grayBits.getNodeValue())) //
                    throw new IllegalArgumentException("effectiveBitsPerPixel == " + effbits + ", expected: " + grayBits.getNodeValue());
            }

            break;

        case 2: // "RGB"
        case 6: // "RGBA"
        case 3: // "Palette"
            if (imageFormat != ImageFormat.IMAGE_FORMAT_RGB && (imageFormat != ImageFormat.IMAGE_FORMAT_GRAY && bytesPerPixel == 3))
                throw new IllegalArgumentException(
                        "imageFormat == " + imageFormat + ", expected: " + (bytesPerPixel == 3 ? ImageFormat.IMAGE_FORMAT_GRAY : ImageFormat.IMAGE_FORMAT_RGB));

            break;

        default:
            throw new IllegalArgumentException("pngColorType == " + pngColorType);
        }

        Node phyNode = getNode(rootNode,KEY_PHYS);
        boolean hasUnitSpecifier = false;
        if (phyNode != null) {
            String unitSpecifier = phyNode.getAttributes().getNamedItem(KEY_UNIT_SPECIFIER).getNodeValue();
            hasUnitSpecifier = !unitSpecifier.equalsIgnoreCase(KEY_UNKNOWN);
        }

        if (hasUnitSpecifier) {
            int xAxisPixels = Integer.parseInt(phyNode.getAttributes().getNamedItem(KEY_PIXELS_PER_UNIT_X_AXIS).getNodeValue());
            if (xScale != 1000.0F / xAxisPixels)
                throw new IllegalArgumentException("xScale == " + xScale + ", expected value: " + 1000.0F / xAxisPixels);

            int yAxisPixels = Integer.parseInt(phyNode.getAttributes().getNamedItem(KEY_PIXELS_PER_UNIT_Y_AXIS).getNodeValue());
            if (yScale != 1000.0F / yAxisPixels)
                throw new IllegalArgumentException("yScale == " + yScale + ", expected value: " + 1000.0F / yAxisPixels);

        } else if (xScale != -1F)
            throw new IllegalArgumentException("yScale == " + yScale + ", expected value: " + (-1F));

        else if (yScale != -1F)
            throw new IllegalArgumentException("yScale == " + yScale + ", expected value: " + (-1F));

        // ... TODO: continue validation

        return true;
    }

    /**
     * Sets default values to 'archival' metadata.
     */
    private void setDefault(Node rootNode) {
        versionTag = IMAGE.IMAGE_VERSION;
        cameraPortId = IMAGE.DEFAULT_CAMERA_PORT_ID;
        cameraPortName = IMAGE.DEFAULT_CAMERA_PORT_NAME;
        aoiWidth = aoiHeight = -1;
        xStart = yStart = 0;
        horizontalBinning = verticalBinning = 0;
        frameNumber = eventNumber = 0L;
        imageRotation = 0F;
        imageFlags = ImageFlag.LITTLE_ENDIAN_BYTE_ORDER.getId(); // TODO: check
        ispare1 = ispare2 = ispare3 = -1;
        fspare1 = fspare2 = fspare3 = -1f;

        Node hdrNode = getNode(rootNode,KEY_IHDR);
        Node phyNode = getNode(rootNode,KEY_PHYS);

        sourceWidth = hdrNode == null ? 0 : Integer.parseInt(hdrNode.getAttributes().getNamedItem(KEY_WIDTH).getNodeValue());
        sourceHeight = hdrNode == null ? 0 : Integer.parseInt(hdrNode.getAttributes().getNamedItem(KEY_HEIGHT).getNodeValue());

        Node timeNode = getNode(rootNode,KEY_TIME);
        Calendar cal = null;
        if (timeNode != null) {
            NamedNodeMap time = timeNode.getAttributes();
            int year = Integer.parseInt(time.getNamedItem("year").getNodeValue());
            int month = Integer.parseInt(time.getNamedItem("month").getNodeValue());
            int day = Integer.parseInt(time.getNamedItem("day").getNodeValue());
            int hour = Integer.parseInt(time.getNamedItem("hour").getNodeValue());
            int minute = Integer.parseInt(time.getNamedItem("minute").getNodeValue());
            int second = Integer.parseInt(time.getNamedItem("second").getNodeValue());
            cal = new GregorianCalendar(year, month - 1, day, hour, minute, second);
        } else {
            cal = new GregorianCalendar();
        }
        timestampSeconds = (int) (cal.getTimeInMillis() / 1000);
        timestampMicroseconds = 0;

        boolean hasUnitSpecifier = false;
        if (phyNode != null) {
            String unitSpecifier = phyNode.getAttributes().getNamedItem(KEY_UNIT_SPECIFIER).getNodeValue();
            hasUnitSpecifier = !unitSpecifier.equalsIgnoreCase(KEY_UNKNOWN);
        }

        if (hasUnitSpecifier) {
            xScale = 1000f/Integer.parseInt(phyNode.getAttributes().getNamedItem(KEY_PIXELS_PER_UNIT_X_AXIS).getNodeValue());
            yScale = 1000f/Integer.parseInt(phyNode.getAttributes().getNamedItem(KEY_PIXELS_PER_UNIT_Y_AXIS).getNodeValue());
        } else
            xScale = yScale = -1F;

        String colorType = hdrNode.getAttributes().getNamedItem(KEY_COLOR_TYPE).getNodeValue();
        int pngColorType = IHDRColorNames.indexOf(colorType);
        switch (pngColorType) {
        case 0: // PNG_COLOR_GRAY
        case 4: // PNG_COLOR_GRAY_ALPHA
            sourceFormat = imageFormat = ImageFormat.IMAGE_FORMAT_GRAY;
            effectiveBitsPerPixel = Integer.parseInt(hdrNode.getAttributes().getNamedItem(KEY_BIT_DPETH).getNodeValue());
            bytesPerPixel = effectiveBitsPerPixel <= 8 ? 1 : 2;
            break;

        case 2: // PNG_COLOR_RGB
        case 6: // PNG_COLOR_RGB_ALPHA
        case 3: // PNG_COLOR_PALETTE
            sourceFormat = imageFormat = ImageFormat.IMAGE_FORMAT_RGB;
            bytesPerPixel = 3;
            effectiveBitsPerPixel = 24;
            break;

        default:
            throw new IllegalArgumentException("pngColorType == " + pngColorType);
        }
        appendedFrameSize = sourceWidth * sourceHeight * bytesPerPixel;
    }

    // /**
    // * Converts TINE Image format to PNG image color type
    // *
    // * @param timageFormat
    // * - Tine IMAGE format
    // * @return PNG image color type
    // *
    // * @see FrameHeader#imageFormat
    // * @see TImageMetadata#pngColorType
    // */
    // public int toPngColorType() {
    // if (imageFormat == null)
    // throw new NullPointerException("timageFormat == null!");
    //
    // switch (imageFormat) {
    // case IMAGE_FORMAT_GRAY:
    // case IMAGE_FORMAT_HUFFYUV:
    // return 0; // PNG_COLOR_GRAY
    //
    // case IMAGE_FORMAT_RGB: // RGB24
    // return 2; // PNG_COLOR_RGB
    //
    // case IMAGE_FORMAT_RGBA:
    // return 6; // PNG_COLOR_RGB_ALPHA
    //
    // case IMAGE_FORMAT_JPEG:
    // if (bytesPerPixel == 1)
    // return 0;
    // return 2;
    //
    // default:
    // throw new IllegalArgumentException("Unsupported Tine Image format: " +
    // imageFormat);
    // }
    // }

    public static String md5hash(byte[] data) {
        MessageDigest m;
        try {
            m = MessageDigest.getInstance("MD5");

        } catch (NoSuchAlgorithmException e) {
            throw new RuntimeException("Invalid cryptographic algorithm: " + e.getMessage());
        }

        m.update(data);
        byte[] digest = m.digest();
        BigInteger bigInt = new BigInteger(1, digest);
        return bigInt.toString(16);
    }

    /**
     * Return <code>true</code> if image metadata is archival (valid), otherwise <code>false</code>
     *
     * @return <code>true</code> if image metadata is archival (valid), otherwise <code>false</code>
     */
    public boolean isArchival() {
        return versionTag != null //
                && cameraPortId != null //
                && cameraPortName != null //
                && timestampSeconds != null //
                && timestampMicroseconds != null //
                && sourceWidth != null //
                && sourceHeight != null //
                && aoiWidth != null //
                && aoiHeight != null //
                && xStart != null //
                && yStart != null //
                && bytesPerPixel != null //
                && effectiveBitsPerPixel != null //
                && horizontalBinning != null //
                && verticalBinning != null //
                && sourceFormat != null //
                && imageFormat != null //
                && frameNumber != null //
                && eventNumber != null //
                && xScale != null //
                && yScale != null //
                && imageRotation != null //
                && imageFlags != null //
                && appendedFrameSize != null;
    }

}
