/*
 * Created on Jan 16, 2007
 *
  */
package de.desy.mst.auxiliary.devl.deviceserverlib.config;

import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.bind.annotation.XmlType;

/**
 * Property specification.
 * @author jwlg
 */
@XmlRootElement
@XmlType(propOrder={"name", 
    "description", 
    "type", 
    "array", 
    "arrayLength", 
    "asynchronous", 
    "readable", 
    "writeable", 
    "exported", 
    "buffered",
    "localArchive",
    "XDataInfo",
    "YDataInfo"})
public class CfgProperty {
    
    private String name = "";
    private String description = "";
    private boolean isArray = false;
    private int arrayLength = 0;
    private boolean isAsynchronous = false;
    private boolean isWriteable = false;
    private boolean isReadable = true;
    private boolean isExported = true;
    private boolean isBuffered = false;
    
    private CfgPropertyType type = CfgPropertyType.INT;
    private CfgPropertyDataInfo xDataInfo = new CfgPropertyDataInfo();
    private CfgPropertyDataInfo yDataInfo = new CfgPropertyDataInfo();
    private CfgLocalArchive localArchive = new CfgLocalArchive();
    
    /**
     * Default Constructor.
     */
    public CfgProperty() {
    }

    /**
     * Constructor with name and type.
     * @param name The name of the property.
     * @param type The data type.
     */
    public CfgProperty(String name, CfgPropertyType type) {
        this.name = name;
        this.type = type;
    }
    
    /**
     * Returns the array length.
     * The value only makes sense when isArray is set.
     * @see #isArray
     */
    @XmlElement
    public int getArrayLength() {
        return arrayLength;
    }
    
    /**
     * Sets the array length.
     * The value only makes sense when isArray is set,
     * but can be modified independently.
     * @param arrayLength New array length.
     * @see #isArray
     */
    public void setArrayLength(int arrayLength) {
        this.arrayLength = arrayLength;
    }
    
    /**
     * Returns true if the property is an array.
     */
    @XmlElement
    public boolean isArray() {
        return isArray;
    }
    
    /**
     * Makes the property an array.
     * An array property means that a single device
     * returns an array of values. Array size must be fixed.
     * @param isArray New value.
     */
    public void setArray(boolean isArray) {
        this.isArray = isArray;
    }
    
    /**
     * Returns true if the property is asynchronous.
     */
    @XmlElement
    public boolean isAsynchronous() {
        return isAsynchronous;
    }
    
    /**
     * Makes a property asynchronous.
     * For an asynchronous property, the set in a
     * thread. 
     * @param isAsynchronous New value.
     */
    public void setAsynchronous(boolean isAsynchronous) {
        this.isAsynchronous = isAsynchronous;
    }
    
    /**
     * Returns true if a property is writeable.
     */
    @XmlElement
    public boolean isWriteable() {
        return isWriteable;
    }

    /**
     * Makes a property writeable.
     * @param isWriteable New value.
     */
    public void setWriteable(boolean isWriteable) {
        this.isWriteable = isWriteable;
    }
    
    /**
     * Returns the property's name.
     */
    @XmlElement
    public String getName() {
        return name;
    }

    /**
     * Sets the property's name.
     * @param name New name.
     */
    public void setName(String name) {
        this.name = name;
    }
    
    /**
     * Returns the description.
     */
    @XmlElement
    public String getDescription() {
        return description;
    }
    
    /**
     * Sets the description.
     * @param description New description.
     */
    public void setDescription(String description) {
        this.description = description;
    }

    /**
     * Returns the type.
     */
    @XmlElement
    public CfgPropertyType getType() {
        return type;
    }

    /**
     * Sets the type.
     * A property's data type can be any simple type
     * or a pre-defined TINE type.
     * @param type new type.
     */
    public void setType(CfgPropertyType type) {
        this.type = type;
    }

    /**
     * Returns true if a property is buffered.
     */
    @XmlElement
    public boolean isBuffered() {
        return isBuffered;
    }
    
    /**
     * Makes a proeprty buffered.
     * For a buffered proeprty, the value is read out 
     * end buffered periodically, independend of client 
     * requests.
     * @param isBuffered New value.
     */
    public void setBuffered(boolean isBuffered) {
        this.isBuffered = isBuffered;
    }

    /**
     * Returns true if the property is readable and
     * if local archiving is enabled.
     */
    public boolean hasLocalArchive() {
        return isReadable() && localArchive.isEnabled();
    }
    
    /**
     * Returns true if a property is exported.
     */
    @XmlElement
    public boolean isExported() {
        return isExported;
    }
    
    /**
     * Specifies if a property is exported.
     * An exported property canbe accessed via the
     * TINE interface.
     * @param isExported New value.
     */
    public void setExported(boolean isExported) {
        this.isExported = isExported;
    }
    
    /**
     * Returns true if a property is readable.
     */
    @XmlElement
    public boolean isReadable() {
        return isReadable;
    }

    /**
     * Specified if a property is readable.
     * @param isReadable New value.
     */
    public void setReadable(boolean isReadable) {
        this.isReadable = isReadable;
    }
    
    /**
     * Returns display information for the X axis.
     */
    @XmlElement(name="XDataInfo")
    public CfgPropertyDataInfo getXDataInfo() {
        return xDataInfo;
    }
    
    /**
     * Sets the data information for the X axis.
     * Required by JAXB. Not for general use.
     * @param dataInfo
     */
    public void setXDataInfo(CfgPropertyDataInfo dataInfo) {
        xDataInfo = dataInfo;
    }
    
    /**
     * Returns display information for the y axis.
     * The display information is rather a "meta-property"
     * than an attribute of the property itself.
     * Anyway, it is handled in the latter way here in order 
     * to make it simpler. 
     */
    @XmlElement(name="YDataInfo")
    public CfgPropertyDataInfo getYDataInfo() {
        return yDataInfo;
    }
    
    /**
     * Sets the display information for the y axis.
     * 
     * Required by JAXB. Not for general use.
     * @param dataInfo New display information.
     */
    public void setYDataInfo(CfgPropertyDataInfo dataInfo) {
        yDataInfo = dataInfo;
    }
    
    /**
     * Return the local archive information.
     * Like the display information, the local archive 
     * information is not really an attribute of a property.
     * It rather belongs to the local archive, but it's simpler
     * this way.
     */
    @XmlElement
    public CfgLocalArchive getLocalArchive() {
        return localArchive;
    }
    
    /**
     * Sets the local archive information.
     * @param localHistory New local archive information.
     */
    public void setLocalArchive(CfgLocalArchive localHistory) {
        this.localArchive = localHistory;
    }

}
