/*
 * Created on Jul 5, 2006
 *
  */
package de.desy.mst.auxiliary.devl.deviceserverlib.initialization;

import java.util.HashMap;
import java.util.Properties;

import de.desy.mst.auxiliary.devl.deviceserverlib.config.CfgInitType;


/**
 * The database driver repository.
 * The DriverRepository contains IDatabaseDrivers which are identified
 * by their type. A IDatabaseDriver is a front-end for a particular 
 * JDBC driver. Pre-defined drivers exist for CSV and XML files and ODBC sources.
 * This class should be used as a repository only. In order to
 * access different table formats transparently, users should 
 * generally favor DatabaseReader which hides DriverRepository and
 * provides a more convenient interface.
 * DriverRepository makes it possible to register custom database
 * formats as plug-ins.
 * It contains ODBC and CVS table creators by default.
 * 
 * @author jwlg
 */
public class DriverRepository extends HashMap<CfgInitType, IDatabaseDriver> {

    private static final long serialVersionUID = -1271112167235106121L;
    private static DriverRepository theInstance = new DriverRepository();
    
    /**
     * Constructor.
     *
     */
    private DriverRepository() {
        super();
        Properties props;
        
        props = new Properties();
        props.put("separator",",");
        props.put("suppressHeaders","false");
        props.put("fileExtension",".csv"); 
        props.put("commentLine","#");  
        props.put("dateFormat","yyyy-MM-dd HH:mm:ss");       

        this.addTableCreator(CfgInitType.StelsCSV, new GenericDriver(
            "jstels.jdbc.csv.CsvDriver",
            "jdbc:jstels:csv:", props));
        
        this.addTableCreator(CfgInitType.StelsDBF, new GenericDriver(
            "jstels.jdbc.dbf.DBFDriver",
            "jdbc:jstels:dbf:"));
        this.addTableCreator(CfgInitType.StelsXML, new GenericDriver(
            "jstels.jdbc.xml.XMLDriver",
            "jdbc:jstels:xml:"));
        //this.addTableCreator(CfgInitType.CSV, new CsvJdbcDriver());
        this.addTableCreator(CfgInitType.ODBC, new OdbcDriver());
        this.addTableCreator(CfgInitType.JDBC, new JdbcDriver());
    }
    
    /**
     * Returns the instance.
     */
    public static DriverRepository getInstance()  {
        return theInstance;
    }
    
    /**
     * Adds a table creator to the repository.
     * This method is used to add a custom table creator
     * to the repository. When it is known to the repository,
     * tables of its type it can be accessed via its type ID 
     * through DatabaseReader. 
     * @param creator
     */
    public void addTableCreator(CfgInitType initType, IDatabaseDriver creator) {
        put(initType, creator);
    }
    
    
    public IDatabaseDriver getDatabaseDriver(CfgInitType initType) {
        return this.get(initType);
    }
    
    
    /**
     * Needed by old-style initialization.
     * @param typeID
     */
    public IDatabaseDriver getDatabaseDriver(String typeID) {
        IDatabaseDriver result = null;
        for (CfgInitType type : CfgInitType.values()) {
            if (type.toString().equals(typeID)) {
               result = getDatabaseDriver(type);
               break;
            }
        }
        return result;
    }

 
}
