/*
 * Created on Sep 19, 2007
 *
  */
package de.desy.mst.auxiliary.devl.deviceserverlib.server;

import java.util.LinkedList;
import java.util.Set;

import org.apache.log4j.Logger;

import de.desy.mst.auxiliary.devl.deviceserverlib.alarms.Alarm;
import de.desy.mst.auxiliary.devl.deviceserverlib.config.CfgAlarm;
import de.desy.mst.auxiliary.devl.deviceserverlib.config.CfgDeviceGroup;
import de.desy.mst.auxiliary.devl.deviceserverlib.config.CfgDeviceServer;
import de.desy.mst.auxiliary.devl.deviceserverlib.history.HistoryMap;


/**
 * An abstract  base class for equipment module builders.
 * 
 * This is the base class for 
 * @author jwlg
 */
public abstract class AEquipmentModuleBuilder<T extends IBasicDevice> implements IEquipmentModuleBuilder<T> {
    
    private CfgDeviceGroup cfgDeviceGroup = null;
    private HistoryMap historyMap = null;
    private CfgDeviceServer cfgDeviceServer = null;
    private GenericDeviceServer deviceServer = null;
    private IBasicModuleFactory<T> moduleFactory = null;
    private DeviceContext deviceContext = null;
    
    private EquipmentModule<T> equipmentModule = null;
    private DeviceGroup<T> deviceGroup = null;

    private static Logger log = Logger.getLogger(AEquipmentModuleBuilder.class);
    
    /**
     * Returns the CfgDeviceGroup.
     * To be used by derived classes.
     */
    protected CfgDeviceGroup getCfgDeviceGroup() {
        return cfgDeviceGroup;
    }

    /**
     * Returns the CfgDeviceServer.
     * To be used by derived classes.
     */
    protected CfgDeviceServer getCfgDeviceServer() {
        return cfgDeviceServer;
    }

    /**
     * Returns the DeviceContext.
     * To be used by derived classes.
     */
    protected DeviceContext getDeviceContext() {
        return deviceContext;
    }
    
    /**
     * Returns the DeviceGroup.
     * To be used by derived classes.
     */
    protected DeviceGroup<T> getDeviceGroup() {
        return deviceGroup;
    }
    
    /**
     * Returns the GenericDeviceServer.
     * To be used by derived classes.
     */
    protected GenericDeviceServer getDeviceServer() {
        return deviceServer;
    }

    /**
     * Returns the HistoryMap.
     * To be used by derived classes.
     */
    protected HistoryMap getHistoryMap() {
        return historyMap;
    }

    /**
     * @throws EquipmentModuleBuilderException 
     * @override
     */
    public void createCommonData() throws EquipmentModuleBuilderException {
        if (cfgDeviceGroup == null) {
            log.error("Cannot create common data. CfgDeviceGroup is not set.");
            throw new EquipmentModuleBuilderException("createCommonData() failed");
        }
        if (moduleFactory == null) {
            log.error("Cannot create common data. Module Factory is not set.");
            throw new EquipmentModuleBuilderException("createCommonData() failed");
        }
        if (equipmentModule == null) {
            log.error("Cannot create common data. Equipment module is not initialized.");
            throw new EquipmentModuleBuilderException("createCommonData() failed");
        }
        if (deviceContext == null) {
            log.error("Cannot create common data. DeviceContext is not initialzied.");
            throw new EquipmentModuleBuilderException("createCommonData() failed.");
        }
        
        log.info("Creating common data for " + cfgDeviceGroup.getName());
        ACommonData<T> commonData = moduleFactory.createCommonData();
        commonData.setDeviceContext(deviceContext);
        equipmentModule.setCommonData(commonData);
        deviceContext.setCommonData(commonData);
    }

    /**
     * @throws EquipmentModuleBuilderException 
     * @override
     */
    public void initializeEquipmentModule() throws EquipmentModuleBuilderException {
        if (cfgDeviceGroup == null) {
            equipmentModule = null;
            log.error("Cannot create Equipment module. CfgDevicegroup is not set.");
            throw new EquipmentModuleBuilderException("initializeEquipmentModule() failed.");
        } else {
            log.info("Initializing equipment module " + cfgDeviceGroup.getName());
            equipmentModule = new EquipmentModule<T>(cfgDeviceGroup.getName());
            equipmentModule.setDeviceServerConfig(cfgDeviceServer);
            equipmentModule.setDeviceGroupConfig(cfgDeviceGroup);
        }
    }
    
    /**
     * Creates a device context.
     */
    public void createDeviceContext() {
        if (deviceServer == null) {
            log.error("Cannot create valid device context. Device server is null");
        }
        this.deviceContext = new DeviceContext();
        deviceContext.setDeviceServer(deviceServer);
    }
    
    public EquipmentModule<T> getEquipmentModule() {
        return equipmentModule;
    }

    public void registerDevices() throws EquipmentModuleBuilderException {
        if (equipmentModule == null) {
            log.error("Cannot register devices. Equipment module is not initialized.");
            throw new EquipmentModuleBuilderException();
        }
         if (deviceGroup == null) {
            log.error("Cannot register devices. Device group is not initialized.");
            throw new EquipmentModuleBuilderException();
         }
            equipmentModule.registerDevices(deviceGroup);
    }

    public void registerLocalHistories() throws EquipmentModuleBuilderException {
        if (equipmentModule == null) {
            log.error("Cannot register local histories. Equipment module is not initialized.");
            throw new EquipmentModuleBuilderException();
        }
        if (historyMap != null)  {
            equipmentModule.registerLocalHistories(historyMap);
        } else {
            log.info("No History map specified for equipment module " + equipmentModule.getExportName());
        }
    }

    public void registerProperties() throws EquipmentModuleBuilderException {
        if (deviceGroup == null) {
            log.error("Cannot register properties. Module factory is not set.");
            throw new EquipmentModuleBuilderException("registerProperties() failed.");
        }
        if (cfgDeviceGroup == null) {
            log.error("Cannot register properties. CfgDeviceGgoup is not set.");
            throw new EquipmentModuleBuilderException("registerProperties() failed.");
        }
        if (moduleFactory == null) {
            log.error("Cannot register properties. Module factory is not set.");
            throw new EquipmentModuleBuilderException("registerProperties() failed.");
        }
        if (equipmentModule == null) {
            log.error("Cannot register properties. Equipment module is not initialized.");
            throw new EquipmentModuleBuilderException("registerProperties() failed.");
        }
         if (moduleFactory == null) {
             log.error("Cannot register properties. Module factory is not set.");
             throw new EquipmentModuleBuilderException("registerProperties() failed.");
        }
        PropertySet<T> propertySet = moduleFactory.createPropertySet(deviceGroup);
        equipmentModule.registerProperties(propertySet);
            
    // TODO Auto-generated method stub
    }

    
    public void registerAlarms() throws EquipmentModuleBuilderException {
        if (moduleFactory == null) {
            log.error("Cannot register alarms. Module factory is not set.");
            throw new EquipmentModuleBuilderException("registerAlarms() failed.");
        }
        if (cfgDeviceGroup == null) {
            log.error("Cannot register alarms. CfgDeviceGgoup is not set.");
            throw new EquipmentModuleBuilderException("registerAlarms() failed.");
        }
        if (equipmentModule == null) {
            log.error("Cannot register alarms. Equipment module is not initialized.");
            throw new EquipmentModuleBuilderException("registerAlarms() failed.");
        }
        
        LinkedList<CfgAlarm> alarms = cfgDeviceGroup.getDeviceType().getAlarms();
        if (alarms.size() > 0) {
            Set<Alarm> alarmSet = moduleFactory.createAlarmSet();
            if (alarmSet != null) {
                log.info("Registering " + alarms.size() + " alarm(s) for " + cfgDeviceGroup.getName());
                equipmentModule.registerAlarms(
                    cfgDeviceGroup.getAlarmID(),
                    cfgDeviceGroup.getDescription(),
                    alarmSet);
            } else {
                log.warn("createAlarmSet returned null" );
            }
        } else {
            log.info("No alarms specified for " + cfgDeviceGroup.getName());
        }
    }
    
    public void setCfgDeviceGroup(CfgDeviceGroup cfgDeviceGroup) {
        this.cfgDeviceGroup = cfgDeviceGroup;
    }

    
    public void setCfgDeviceServer(CfgDeviceServer cfgDeviceServer) {
        this.cfgDeviceServer = cfgDeviceServer;
    }

    
    public void setDeviceServer(GenericDeviceServer deviceServer) {
        this.deviceServer = deviceServer;
    }

    
    public void setHistoryMap(HistoryMap historyMap) {
        this.historyMap = historyMap;
    }
    
    public void setModuleFactory(IBasicModuleFactory<T> moduleFactory) {
        this.moduleFactory = moduleFactory;
    }
    
    /**
     * Assigns the device group.
     * Called by createDeviceGroup().
     * @param deviceGroup The device group.
     */
    protected void setDeviceGroup(DeviceGroup<T> deviceGroup) {
    
        this.deviceGroup = deviceGroup;
        if (deviceContext != null) {
            deviceContext.setDeviceGroup(deviceGroup);
        } else {
            log.warn("Device context is not set.");
        }
    }

    /**
     * Creates the device group.
     * Must call setDeviceGroup()!
     */
    public abstract void createDeviceGroup()  throws EquipmentModuleBuilderException;
;


}
