package de.desy.mst.auxiliary.devl.deviceserverlib.server;

import java.util.LinkedList;
import java.util.TreeSet;

import org.apache.log4j.Logger;

import de.desy.mst.auxiliary.devl.deviceserverlib.initialization.controller.ControllerDefinition;
import de.desy.mst.auxiliary.devl.deviceserverlib.initialization.controller.ControllerGroup;
import de.desy.mst.auxiliary.devl.deviceserverlib.initialization.controller.DeviceRange;

/**
 * @version 1.0
 * @created 17-Sep-2007 14:26:39
 */
public class ControllerDeviceFactory<D extends IBasicDevice, T extends ABasicController<D>> implements IDeviceFactory<T> {

    @SuppressWarnings("unused")
    private final static Logger log = Logger.getLogger(ControllerDeviceFactory.class);

    ControllerGroup controllerGroupDefinition = null;
    GenericDeviceServer deviceServer = null;
    private IControllerFactoryMethod<T> factoryMethod = null;


    public ControllerDeviceFactory() {
    }

    public void setControllerGroupDefinition(
            ControllerGroup controllerGroupDefinition) {
        this.controllerGroupDefinition = controllerGroupDefinition;
    }

    public void setFactoryMethod(IControllerFactoryMethod<T> factoryMethod) {
        this.factoryMethod = factoryMethod;
    }
    
    protected void setDeviceServer(GenericDeviceServer deviceServer) {
        this.deviceServer = deviceServer;
    }

    /**
     * Populates a device group with devices.  This method creates devices, assigns a
     * device context to them and adds them to a device group.  The device server
     * normally uses a standard device factory, but it is also possible to use a
     * customized device factory. It has to implement this interface.
     * 
     * @param controllerDeviceGroup    The device group where to put the controllers.
     * @override
     */
    @SuppressWarnings("unchecked")
    public void createDevices(DeviceGroup<T> controllerDeviceGroup) {

        if (deviceServer == null) {
            log.error("Cannot create controllers. deviceServer not set.");
            return;
        }



        for (ControllerDefinition controllerDefinition :
            controllerGroupDefinition.getControllers()) 
        {

            try {
                /* Create and fill device map */
                DeviceMap<D> deviceMap = new DeviceMap<D>();
                TreeSet<String> deviceSelection = new TreeSet<String>();

                LinkedList<String> selectedDevices = controllerDefinition.getSelectedDevices();
                LinkedList<DeviceRange> selectedRanges = controllerDefinition.getSelectedRanges();


                String managedDeviceGroupName = controllerDefinition.getManagedDeviceGroup();
                DeviceGroup<D> managedDeviceGroup = (DeviceGroup<D>) deviceServer.getDeviceGroup(managedDeviceGroupName);
                if (managedDeviceGroup == null) {
                    log.error("Managed device group " + managedDeviceGroupName + " not found for controller " + controllerDefinition.getDeviceName());
                    continue;
                }
                
                /*
                 * If no device is selected and no range is specified,
                 * all devices in the group are selected. 
                 */
                if (((selectedDevices == null) || (selectedDevices.size() == 0))
                        && ((selectedRanges == null)|| (selectedRanges.size() == 0))) 
                {
                    for (D device : managedDeviceGroup.getDevices()) {
                        deviceSelection.add(device.getDeviceName());
                    }
                } 

                /*
                 * If there is a list of selected devices, put all 
                 * listed devices into the selection.
                 */
                if (selectedDevices != null) {
                    deviceSelection.addAll(selectedDevices);
                }

                /*
                 * If there are selected ranges, determine
                 * selected devices and put them in the list. 
                 */
                if (selectedRanges != null) {
                    for (DeviceRange range : selectedRanges) {
                        boolean firstDeviceFound = false;
                        String firstDeviceName = range.getFirstDevice();
                        String lastDeviceName = range.getLastDevice();
                        
                        if (firstDeviceName == null) {
                            log.warn("No first device specified. Device range is invalid.");
                            firstDeviceName = "";
                        }
                        
                        if (lastDeviceName == null) {
                            // No warning here. Null is allowed.
                            lastDeviceName = "";
                        }
                        
                        for (D device : managedDeviceGroup.getDevices()) {
                            if (!firstDeviceFound) {
                                firstDeviceFound = device.getDeviceName().matches(range.getFirstDevice());
                            }
                            if (firstDeviceFound) {
                                deviceSelection.add(device.getDeviceName());
                                if (device.getDeviceName().matches(range.getLastDevice())) {
                                    break;
                                }
                            }
                        }
                    }
                }

                /*
                 * Now put all selected devices into the device map.
                 */
                for (D device : managedDeviceGroup.getDevices()) {
                    if (deviceSelection.contains(device.getDeviceName())) {
                        deviceMap.addDevice(device);
                    }
                }

                /* Create controllers and put them into the controller group */
                if (factoryMethod != null) {

                    T controller =
                        factoryMethod.createController(controllerDefinition);

                    if (controller != null) {
                        //controller.setDeviceContext(deviceContext);
                        controller.setControllerDefinition(controllerDefinition);
                        controller.setDeviceMap(deviceMap);
                        controllerDeviceGroup.addDevice(controller);
                    } else {
                        log.error("Factory failed to create Controller " + controllerDefinition.getDeviceName());
                    }

                } else {
                    log.error("Controller factory method is missing!");
                }

            } catch (Exception e) {
                log.error("Could not create controller " + controllerDefinition.getDeviceName() + ": " +e, e);
            }

        }

    }

}