package de.desy.mst.auxiliary.devl.deviceserverlib.server;

import java.util.LinkedList;
import java.util.List;

/**
 * Performs groupwise device locks for a device map.
 * 
 * The device lock manager cares for the locking of 
 * suvbsets of devices in a device map. A locking 
 * operation returns a DeviceLockSet which knows if the
 * locking operation has been successful and which
 * devices were locked. The DeviceLockSet is also needed
 * to unlock all locked devices.
 * 
 * @version 1.0
 * @created 30-Aug-2007 11:27:38
 */
public class DeviceLockManager {

	public DeviceMap<IBasicDevice> deviceMap;
    

    public DeviceLockManager(IBasicDevice device) {
        this.deviceMap = new DeviceMap<IBasicDevice>();
        this.deviceMap.addDevice(device);
    }
    
    /**
     * Constructor.
     * @param deviceMap A device map.
     */
    @SuppressWarnings("unchecked")
    public DeviceLockManager(DeviceMap deviceMap) {
        this.deviceMap = deviceMap;
    }
    
    /**
     * Contructor.
     * @param deviceGroup
     */
    @SuppressWarnings("unchecked")
    public DeviceLockManager(DeviceGroup deviceGroup) {
        this.deviceMap = deviceGroup.getDeviceMap();
    }
    
    
    /**
     * Tries to acquire locks for a specified list of devices.
     * @param devices The list of devices to be locked.
     * @return The DeviceLockSet.
     */
    public DeviceLockSet tryAcquireLocks(List<IBasicDevice> devices) {
        
        LinkedList<IBasicDevice> lockedDevices = new LinkedList<IBasicDevice>();
        boolean lockSuccessful = true;
        
        for (IBasicDevice device : devices) {
            lockSuccessful &= device.tryAcquireLock();
            if (lockSuccessful) {
                lockedDevices.add(device);
            } else {
                break;
            }
        }
        
        if (!lockSuccessful) {
            releaseLocks(lockedDevices);
            lockedDevices = null;
        }
        return new DeviceLockSet(lockedDevices);
     }
    
    /**
     * Tries to acquire a lock for all devices.
     * @return The DeviceLockSet. 
     */
    public DeviceLockSet tryAcquireDeviceLocks() {
        return tryAcquireLocks(deviceMap.getDeviceList());
   }

    /**
     * Tries to acquire a lock for a set of devices.
     * @param deviceNames The list of devices to be locked. Non-existing devices are ignored!
     * @return The DeviceLockSet.
     */
    public DeviceLockSet tryAcquireDeviceLocks(List<String> deviceNames) {
        return tryAcquireLocks(deviceMap.getDeviceListByNames(deviceNames));
    }
    
    /**
     * Tries to acquire a lock for a set of devices.
     * @param deviceIndices Array of device indices. Non-existing
     * indices are ignored.
     * @return The DeviceLockSet.
     */
    public DeviceLockSet tryAcquireDeviceLocks(int[] deviceIndices) {
        return tryAcquireLocks(deviceMap.getDeviceListByIndex(deviceIndices));
    }
    
    /**
     * Acquires the locks for a list of devices.
     * Waits until all devices are locked. 
     * @param devices The list of devices.
     * 
     * @return The DeviceLockSet. Note that acquire locks 
     * can fail. Check DeviceLockSet.isLocked() in any case!
     */
    public DeviceLockSet acquireLocks(List<IBasicDevice> devices) {
        LinkedList<IBasicDevice> lockedDevices = new LinkedList<IBasicDevice>();
        boolean lockSuccessful = true;
        
        for (IBasicDevice device : devices) {
            try {
                device.acquireLock();
                lockedDevices.add(device);
            } catch (InterruptedException ex) {
                lockSuccessful = false;
            };
        }
        if (!lockSuccessful) {
            releaseLocks(lockedDevices);
            lockedDevices = null;
        }
        return new DeviceLockSet(lockedDevices);
    }
    
    /**
     * Acquires locks for all devices.
     * @return The DeviceLockSet. Note that acquire locks 
     * can fail. Check DeviceLockSet.isLocked() in any case!
     */
    public DeviceLockSet acquireDeviceLocks(){
        return acquireLocks(deviceMap.getDeviceList());
    }

    /**
     * Acquires locks for all devices in a list.
     * @param deviceList List of device names.
     * @return The DeviceLockSet. Note that acquire locks 
     * can fail. Check DeviceLockSet.isLocked() in any case!
     * Note that invalid names are ignored. 
     * Note that duplicate indices will lead to a deadlock situation.
     * Use with care! 
    */
    public DeviceLockSet acquireDeviceLocks(List<String> deviceNames){
        return acquireLocks(deviceMap.getDeviceListByNames(deviceNames));
    }
    
    /**
     * Acquires locks for all devices in an array of indices.
     * @param deviceList List of device names.
     * @return The DeviceLockSet. Note that acquire locks 
     * can fail. Check DeviceLockSet.isLocked() in any case!
     * Note that invalid indices are ignored.
     * Note that duplicate indices will lead to a deadlock situation.
     * Use with care! 
    */
    public DeviceLockSet acquireDeviceLocks(int[] deviceIndices) {
        return acquireLocks(deviceMap.getDeviceListByIndex(deviceIndices));
    }
    
    
    /**
     * Releases the locks of all devices in a list.
     * @param devices The devices.
     */
    private void releaseLocks(LinkedList<IBasicDevice> devices) {
        for (IBasicDevice device : devices) {
            device.releaseLock();
        }
    }
    
}