/*
 * Created on Nov 21, 2007
 *
  */
package de.desy.mst.auxiliary.devl.deviceserverlib.server;

import java.util.List;

/**
 * A set of locked devices.
 * A DeviceLockSet contains a list of locked devices
 * and cares for releasing the locks. A DeviceLockSet is
 * returned by a DeviceLockManager which operates on a set
 * of devices and returns those devices which are locked.
 * The DeviceLockSet is used to determine if the locking 
 * operation was successful, to memorize which devices are locked,
 * and to release the locks. 
 * If a DeviceLockSet is deleted, all locked devices
 * are released in the finalizer.
 * @author jwlg
 */
public class DeviceLockSet {
    
    private List<IBasicDevice> lockedDevices = null;
    
    /** 
     * Constructs a lock result.
     *
     */
    DeviceLockSet() {
    }
    
    /**
     * Constructs a lock result. 
     * 
     * @param lockedDevices List of locked devices.
     * 
     * Caution: All devices in the list must really be locked.
     */
    DeviceLockSet(List<IBasicDevice> lockedDevices) {
        this.lockedDevices = lockedDevices;
    }
    
    /**
     * Returns true if there are locked devices.
     */
    public boolean isLocked() {
        return (lockedDevices != null);
    }
    
    /**
     * Releases locks.
     * Unlocks all devices which werde locked by the operation
     * which returned this Lockresult.
     */
    public void release() {
        if (lockedDevices != null) {
            for (IBasicDevice device : lockedDevices) {
                device.releaseLock();
            }
        }
        lockedDevices = null;
    }
    
    /**
     * Returns the number of locked devices.
     */
    public int getNDevices() {
        int result = 0;
        if (lockedDevices != null) {
            result = lockedDevices.size();
        }
        return result;
    }

    @Override
    protected void finalize() throws Throwable {
        super.finalize();
        if (isLocked()) {
            release();
        }
    }
                
}
