package de.desy.mst.auxiliary.devl.deviceserverlib.server;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;

import org.apache.log4j.Logger;

import de.desy.mst.auxiliary.devl.deviceserverlib.tools.StringUtils;

/**
 * 
 * Standard implementation of a device map.
 * @version 1.0
 * @created 30-Aug-2007 10:19:42
 */
public class DeviceMap<D extends IBasicDevice> implements Iterable<D> {

	private ArrayList<D> deviceArray = null;
	private LinkedHashMap<String, Integer> nameToIndexMap = null;
	private LinkedHashMap<String,Integer> upperCaseNameToIndexMap = null;
    
    private static Logger log = Logger.getLogger(DeviceMap.class);

	public DeviceMap() {
        deviceArray = new ArrayList<D>();
        nameToIndexMap = new LinkedHashMap<String, Integer>();
        upperCaseNameToIndexMap = new LinkedHashMap<String, Integer>();
	}

	/**
	 * Adds a device.
	 * @param newDevice The new device.
	 */
	public void addDevice(D device){
        Integer deviceIndex = new Integer(deviceArray.size());
        
        if (device == null) {
            log.error("Cannot add null as a device");
        } else if (StringUtils.isEmpty(device.getDeviceName())) {
            log.error("Cannot add a device without a name.");
        } else {
            deviceArray.add(deviceIndex.intValue(), device);
            nameToIndexMap.put(device.getDeviceName(), deviceIndex);
            upperCaseNameToIndexMap.put(device.getDeviceName().toUpperCase(), deviceIndex);
        }
	}

    /**
     * Returns a device specified by a name.
     * @param deviceName The device name.
     * @return A device or null if none was found.
     */
	public D getDeviceByNameIgnoreCase(String deviceName){
        int deviceIndex = getDeviceIndexByNameIgnoreCase(deviceName);
        if (deviceIndex >= 0) {
            return getDeviceByIndex(deviceIndex);
        } else {
            return null;
        }
	}

       /**
     * Returns the name of a device specified by an index.
     * @param deviceIndex Index of the device.
     */
	public String getDeviceNameByIndex(int deviceIndex){
        String result = null;
        D device = getDeviceByIndex(deviceIndex);
        if (device != null) {
            result =  device.getDeviceName();
        } 
		return result;
	}

    /**
     * Removes all devices.
     */
	public  void removeAllDevices(){
	    deviceArray.clear();
        nameToIndexMap.clear();
        upperCaseNameToIndexMap.clear();
	}

	/**
	 * Removes a device.
     * @param device The device to be removed. 
	 */
	public  void removeDevice(D device){
        deviceArray.remove(device);
        nameToIndexMap.remove(device.getDeviceName());
        upperCaseNameToIndexMap.remove(device.getDeviceName().toUpperCase());
	}

	/**
	 * Removes a device.
	 * @param deviceName The name of the device to be removed.
	 */
	public  void removeDevice(String deviceName){
        int deviceIndex = getDeviceIndexByName(deviceName);
        removeDevice(deviceIndex);
	}

	/**
	 * Removes a device.
	 * @param deviceIndex The index of the device to be removed.
	 */
	public void removeDevice(int deviceIndex){
        D device = getDeviceByIndex(deviceIndex);
        if (device != null) {
            deviceArray.remove(deviceIndex);
            nameToIndexMap.remove(device.getDeviceName());
            upperCaseNameToIndexMap.remove(device.getDeviceName().toUpperCase());
        }
	}
    
    /**
     * Returns a device specified by a name.
     * @param deviceName The device name.
     * @return A device or null if none was found.
     */
    public D getDeviceByName(String deviceName) {
        D result = null;
        int deviceIndex = getDeviceIndexByName(deviceName);
        if (deviceIndex >= 0) {
            result = deviceArray.get(deviceIndex);
        } else {
            result = null;
        }
        return result;
    }

    /**
     * Returns a device specified by an index.
     * The index is the order of the devices in the group, starting with 0.
     * @param deviceIndex The device index
     * @return A device or null if the index is not valid.
     */
    public D getDeviceByIndex(int deviceIndex) {
        return deviceArray.get(deviceIndex);
    }

    /**
     * Returns a device index which corresponds to a given device name.
     * @param deviceName The device name. 
     * @return Device index or -1 if no device was found.
     */
    public int getDeviceIndexByName(String deviceName) {
        Integer deviceIndex = null;
        int result = -1;
        deviceIndex = nameToIndexMap.get(deviceName);
        if (deviceIndex != null) {
            result = deviceIndex.intValue();
        } else {
            result = -1;
        }
        return result;
    }

    /**
     * Returns a device index which corresponds to a TINE name.
     * A TINE device name is either the device name or the
     * device number, prepended by a hash sign "#".
     * @param tineDeviceName The device name. Case insensitive.
     * @return The device index or -1 if no device was found.
     */
    public int getDeviceIndexByTineName(String tineDeviceName) {
        int result = -1;
        if (tineDeviceName.startsWith("#")) {
            result =  Integer.parseInt(tineDeviceName.substring(1));
        } else {
            result = getDeviceIndexByName(tineDeviceName);
        } 
        return result;
    }

    /**
     * Returns a device which corresponds to a TINE name.
     * A TINE device name is either the device name or the
     * device number, prepended by a hash sign "#".
     * @param tineDeviceName The device name. Case insensitive.
     * @return A device or null if no device was found.
     */
    public D getDeviceByTineName(String tineDeviceName) {
        D result = null;
        if (tineDeviceName.startsWith("#")) {
            result =  getDeviceByIndex(Integer.parseInt(tineDeviceName.substring(1)));
        } else {
            result = getDeviceByNameIgnoreCase(tineDeviceName);
        } 
        return result;
    }


    /**
     * Returns the number of devices in the group.
     */
    public int size() {
        return deviceArray.size();
    }

    /**
     * Returns a device index which corresponds to a given device name.
     * @param deviceName The device name. Case insensitive.
     * @return Device index or -1 if no device was found.
     */
    public int getDeviceIndexByNameIgnoreCase(String deviceName) {
        String searchedDevice = deviceName.toUpperCase();
        int result = -1;
        Integer deviceIndex = upperCaseNameToIndexMap.get(searchedDevice);
        if (deviceIndex >= 0) {
            result = deviceIndex.intValue();
        }
        return result;
    }

    /**
     * Returns a list of device names.
     * The list is ordered by the sequence of devices
     * in the device map.
     */
    public List<String> getDeviceNames() {
        LinkedList<String> result = new LinkedList<String>();
        for (IBasicDevice device : deviceArray) {
            result.add(device.getDeviceName());
        }
        return result;
    }

    
    /**
     * Retrurns the list of all devices as an array of strings.
     */
    public String[] getDeviceNamesAsArray() {
        String[] result = new String[deviceArray.size()];
        for (int i = 0; i< deviceArray.size(); i++) {
            result[i] = deviceArray.get(i).getDeviceName();
        }
        return result;
    }
    
    /**
     * Returns a list of all devices.
     * The list is ordered in the same way as the device map.
     */
    public List<D> getDeviceList() {
        return new LinkedList<D>(deviceArray); 
    }
    
    /**
     * Returns a list of devices hich corresponds to a list
     * of device names.
     * @param deviceNames The device names.
     * @return A list of devices. The result contains all
     * devices which could be dereferenced. 
     * Non-existing devices are missing. The 
     * result is ordered in the sequence of the name list.
     */
    public List<D> getDeviceListByNames(List<String> deviceNames) {
        LinkedList<D> lockList = new LinkedList<D>();
        for (String deviceName : deviceNames) {
            D device = getDeviceByName(deviceName);
            if (device != null) {
                lockList.add(device);
            }
        }
        return lockList;
    }
    
    /**
     * Returns a list of devices which corresponds to an
     * array of indices.
     * @param deviceIndices Array of indices.
     * @return A list of devices. The result contains all
     * devices which could be dereferenced in the array. 
     * Non-existing devices are missing. The 
     * result is ordered in the sequence of the array.     
     */
    public final List<D> getDeviceListByIndex(int[] deviceIndices) {
        LinkedList<D> result = new LinkedList<D>();

        for (int deviceIndex : deviceIndices) {
            D device = getDeviceByIndex(deviceIndex);
            if (device != null) {
                result.add(device);
            }
        }
        
        return result;
    }
    
    /**
     * Returns an array list of all devices.
     * The list is ordered in the same way as the device map.
     */
    public ArrayList<D> getDeviceArrayList() {
        return new ArrayList<D>(deviceArray);
    }

    public Iterator<D> iterator() {
        return deviceArray.iterator();
    }


    /**
     * Returns a subset of devices.
     * The selected set is specified by a start index
     * and a size of the set. If the required size goes  
     * beyond the last device in the group, the result is 
     * filled up to the that size, starting with the first device 
     * in the group. However, the required size must be smaller than
     * or equal to the number of devices in the group, or will
     * be limited otherwise.
     * 
     * @param deviceIndex The index of the first device.
     * @param nDevices The desired size of the required subset.
     * @return Subset of devices. May be empty if invalid
     * numbers were supplied.
     */
    public LinkedList<D> getSubset(int deviceIndex, int nDevices) {
        LinkedList<D> result = new LinkedList<D>();

        // Ensure that nDevices does not exceed the number of available devices.
        if (nDevices > deviceArray.size()) {
            nDevices = deviceArray.size();
        }

        if ((deviceIndex >=0) && (deviceIndex < deviceArray.size())) {
            int deviceCount = deviceArray.size() - deviceIndex; 
            for (int i = deviceIndex; i< deviceArray.size(); i++) {
                result.add(deviceArray.get(i));
            }
            if (deviceCount < nDevices) {
                int imax = nDevices - deviceCount;
                for (int i = 0; i<imax; i++) {
                    result.add(deviceArray.get(i));
                }
            }
        }
        return result;
    }


    /**
     * Returns a subset of devices.
     * The subset is specified by the name of a device
     * which is the first device, and a number of selected 
     * devices. If the required size goes  
     * beyond the last device in the group, the result is 
     * filled up to the that size, starting with the first device 
     * in the group. However, the required size must be smaller than
     * or equal to the number of devices in the group, or will
     * be limited otherwise.
     * 
     * @param deviceName The device name. May be a TINE name.
     * @param nDevices The number of devices.
     * @return A list of devices or an empty list if a wrong device
     * name has been specified. 
     */
    public LinkedList<D> getDeviceSubset(String deviceName, int nDevices) {
        LinkedList<D> result = null;

        int firstIndex = 0;

        if (deviceName.startsWith("#")) {
            firstIndex = Integer.parseInt(deviceName.substring(1));
            result = getSubset(firstIndex, nDevices);
        } else {
            firstIndex = getDeviceIndexByName(deviceName);
            result = getSubset(firstIndex, nDevices);
        }
        return result;
    }
    

    
    
}
