/*
 * Created on Aug 3, 2006
 *
 */
package de.desy.mst.auxiliary.devl.deviceserverlib.server;

import java.io.File;
import java.io.InputStream;
import java.util.Date;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map.Entry;

import org.apache.log4j.Logger;

import de.desy.mst.auxiliary.devl.deviceserverlib.config.CfgDeviceGroup;
import de.desy.mst.auxiliary.devl.deviceserverlib.config.CfgDeviceServer;
import de.desy.mst.auxiliary.devl.deviceserverlib.config.CfgDeviceType;
import de.desy.mst.auxiliary.devl.deviceserverlib.config.CfgTypeDefinition;
import de.desy.mst.auxiliary.devl.deviceserverlib.config.ConfigReader;
import de.desy.mst.auxiliary.devl.deviceserverlib.history.HistoryMap;
import de.desy.mst.auxiliary.devl.deviceserverlib.history.HistoryMapReader;
import de.desy.mst.auxiliary.devl.deviceserverlib.history.HistoryMapWriter;
import de.desy.mst.auxiliary.devl.deviceserverlib.tools.NameFactory;
import de.desy.mst.auxiliary.devl.deviceserverlib.tools.StringUtils;
import de.desy.tine.server.equipment.TEquipmentBackgroundTask;
import de.desy.tine.server.equipment.TEquipmentModuleFactory;

/**
 * The main class of the device server.
 * 
 * This is the class which puts all components together.
 * A real device server must be derived from this class.
 * 
 * The main class can be run as a thread or by simply 
 * calling run(). It runs a TINE server with a background
 * thread for the updates.
 * 
 * @author jwlg
 */
public class GenericDeviceServer implements Runnable {

    private static Logger log = Logger.getLogger(GenericDeviceServer.class); 
    private boolean isActive = false;
    private HistoryMap historyMap = null;
    private DirectorySet directorySet = null;

    private LinkedList<IUpdateListener> updateListeners = new LinkedList<IUpdateListener>();
    private HashMap<String, EquipmentModule<?>> modules = new HashMap<String, EquipmentModule<?>>();
    private HashMap<String, IBasicModuleFactory> moduleFactories = new HashMap<String, IBasicModuleFactory>();
    private HashMap<String, String> environment = new HashMap<String, String>();
    private HashMap<String, EquipmentModule<?>> controllerModules = new HashMap<String, EquipmentModule<?>>();

    private CfgDeviceServer deviceServerConfig = null; 

    /**
     * Saves the history map.
     */
    private void saveHistoryMap() {
        
        File historyDir = new File(directorySet.getHistoryDirectory());
        File historyFile = new File(getHistoryFileName());
        
        if (!historyDir.exists()) {
            if (historyDir.mkdirs()) {
                log.info("History subdirectory created.");
            } else {
                log.error("Cannot create " + directorySet.getHistoryDirectory());
            }
        }

        HistoryMapWriter writer = new HistoryMapWriter();
        try {
            log.info("Saving history map.");
            writer.write(historyMap, historyFile);
        } catch (Exception e) {
            log.error("Cannot write history file: " + e, e);
            e.printStackTrace();
        }

    }
    
    /**
     * Reads the history map from the file system.
     * @return A history map. An empty history map if none 
     * was found.
     */
    private void loadHistoryMap() {
        File historyFile = null;
        historyFile = new File(getHistoryFileName());
        if ((historyFile != null) && (historyFile.exists())) {
            HistoryMapReader mapReader = new HistoryMapReader();
            try {
                historyMap = mapReader.read(historyFile);
            } catch (Exception e) {
                log.error("Cannot read history map. " + e, e);
                log.warn("Creating new history map");
                e.printStackTrace();
                historyMap = new HistoryMap();
            }
        } else {
            log.warn("History file " + getHistoryFileName() + " not found.");
            log.warn("Creating new history map");
            historyMap = new HistoryMap();
        }
    }

    /**
     * Returns the name of the history map file.
     */
    private String getHistoryFileName() {
        return directorySet.getHistoryDirectory() + File.separator + "HistoryMap.xml";
    }
    
    /**
     * Returns the package name of this object's class.
     * This method is used to find out the package name.
     * It also works for derived classes.
     */
    public final String getPackageName() {
        return this.getClass().getPackage().getName();
    }

    /**
     * Returns the resource directory.
     * Used to find the resource directory in the jar file.
     */
    private final String getResourceDir() {
        return StringUtils.getInstance().packageToDir(getPackageName(), true) + "/config"; 
    }

    /**
     * Returns the name of the resource file.
     * Used to finde the resouce file in the jar file.
     */
    private final String getResourceFile() {
        return getResourceDir() + "/ServerConfig.xml";
        // ALTERNATIV - relativ:  return "config/ServerConfig.xml";
    }
    
    /**
     * Returns the directory set.
     */
    public final DirectorySet getDirectorySet() {
        return this.directorySet;
    }

    /**
     * Constructor.
     */
    public GenericDeviceServer() throws Exception
    {
        Log4jConfigurator.getInstance().prepareLogging();
        readServerConfiguration();
        initializeRuntimeDirectories();
        Log4jConfigurator.getInstance().configure(getDirectorySet(), deviceServerConfig);
        //MstAppConfigurator.configure(deviceServerConfig.getProject().getMstApp());
        log.info("Starting Device Server. DeviceServerLib-" + Versions.DEVICE_SERVER_LIB);
        initializeSystemEnvironment();
        loadHistoryMap();
        initializeDataTypes();
        initializeDeviceTypes();
        initializeModules();
        saveHistoryMap();
        setupModules();
    }

    /**
     * Initializes the system environment.
     * Reads the environment variables from the configuration
     * and assigns them as system properties.
     *
     */
    private final void initializeSystemEnvironment() {
        
        log.info("initializing system environment");
        try {
            for (Entry<String,String> entry : deviceServerConfig.getConfiguration().getServer().getEnvironment().entrySet()) {
                System.setProperty(entry.getKey(), entry.getValue());
                log.info("Environment: " + entry.getKey() + "=" + entry.getValue());
            }
            log.info("System environment initialized");
        } catch (Exception e) {
            log.fatal("Cannot set System properties");
        }
        
        String historyHome = directorySet.getHistoryDirectory();
        System.setProperty("history.home", historyHome);
        log.info("Setting history home: " + historyHome);
        
    }

    /**
     * Reads the device server configuration.
     * Exists if the configuration cannot be read.
     */
    private void readServerConfiguration() throws Exception {
        
        Exception thrownException = null;
        String resourceFileName = getResourceFile(); // 
        
        // GEHT MIT LEADING "/" oder relativ zum Pfad der aktuellen Klasse
        // (abgeleitet von GenericDeviceServer)
        InputStream inputStream = getClass().getResourceAsStream(resourceFileName);
        if (inputStream == null) {
            log.warn("getClass().getResourceAsStream() did not work. Trying getClass().getClassLoader().getResourceAsStream()");
            // GEHT NICHT MIT LEADING "/", kann aber sonst gehen (??))
            inputStream = getClass().getClassLoader().getResourceAsStream(resourceFileName);
            if (inputStream == null) {
                log.warn("getClass().getClassLoader().getResourceAsStream() dit not work. Trying ClassLoader.getSystemResourceAsStream()");
                // GEHT NICHT VON WEBAPPS ODER APPLETS (benutzt -classpath)
                inputStream = ClassLoader.getSystemResourceAsStream(resourceFileName);
            }
        }
        if (inputStream == null) {
            log.fatal("Cannot read resource from Jar File: " + resourceFileName);
            System.exit(0);
        }
        
        ConfigReader reader = new ConfigReader();
        deviceServerConfig = reader.read(inputStream);
        if (inputStream != null) inputStream.close();
        if ((deviceServerConfig == null) || (thrownException != null)) {
            log.fatal("Cannot read configuration file.");
            thrownException.printStackTrace();
            System.exit(1);
        }
        
    }

    /**
     * Initializes the directory set.
     * 
     * Creates the appropriate directory set and creates 
     * missing directories.
     *
     */
    private void initializeRuntimeDirectories() {
        
        IDirectorySetFactory factory;
        
        if (deviceServerConfig.getProject().getMstApp().isEnabled()
                && deviceServerConfig.getProject().getMstApp().isDirectoryStructureAdopted()) {
            factory = MstAppDirectorySetFactory.getInstance();
        } else {
            factory = DirectorySetFactory.getInstance();
        }
        
        directorySet = factory.createDirectorySet(deviceServerConfig.getConfiguration().getServer().getApplicationName());
        log.info("Setting working directory to " + directorySet.getBaseDirectory());
        System.setProperty("user.dir", directorySet.getBaseDirectory());
        directorySet.createDirectoryTree();
        
    }

    /**
     * Flags the device server as active.
     * Ensures that only one thread can run at a time.
     * @return true if successful. False if the active
     * flag was already set.
     */
    private synchronized boolean setActive() {

        boolean result = false;
        if (!isActive) {
            isActive = true;
            result = true;
        }
        return result;
    }

    /**
     * Resets the active flag.
     * @see #setActive()
     */
    public final synchronized void setInactive() {

        isActive = false;
    }

    /**
     * Returns a device group with a certain name.
     * @param deviceGroupName Name of the device group.
     */
    public final DeviceGroup<?> getDeviceGroup(String deviceGroupName) {
        EquipmentModule selectedModule = modules.get(deviceGroupName);
        DeviceGroup<?> result = null;
        if (selectedModule != null) {
            result = selectedModule.getDeviceGroup();
        }
        return result;
    }
    
    private void updateModule(EquipmentModule<?> module) {
        try {
            module.updateDevices();
            module.updateAlarms();
            module.updatePropertyBuffers();
        } catch(Exception e) {
            log.error("Updating " + module.getModuleName() + "/" + module.getExportName() + ":" + e, e);
        }
    }

    /**
     * Updates all device lists.
     */
    private void updateModules() {
        for (EquipmentModule<?> module : modules.values()) {
            updateModule(module);
       }
        for (EquipmentModule<?> module : controllerModules.values()) {
            updateModule(module);
       }
        
    }
    
    /**
     * Initializes user-defiend data types.
     * This method creates an object of every user-defined device type.
     * This is necessary because all user defined types are 
     * derived from TTaggedStructure. Tagged structures have to
     * be registered, and the registration is done in the 
     * constructor of every instance.
     */
    private void initializeDataTypes() {
        String packageName = NameFactory.getInstance().getUserTypesPackage(deviceServerConfig);
        for (CfgTypeDefinition def : deviceServerConfig.getDefinition().getDataTypes().getTypeDefinitions()) {
            if (def.isTagged()) {
                try {
                    log.info("Registering tagged struture " + def.getClassName() + " : " + def.getTag());
                    Class aClass = Class.forName(packageName + "." + def.getClassName());
                    // A tagged structure is registered when an instance is created.
                    aClass.newInstance();
                } catch (Exception e) {
                    log.error("Cannot register tagged structure " + def.getClassName() + e, e);
                }
            }
        }
    }

    private void initializeDeviceType(CfgDeviceType deviceType) {
        
        IBasicModuleFactory moduleFactory = null;
        NameFactory nameFactory = NameFactory.getInstance();
        String moduleFactoryClassName = 
            nameFactory.getDeviceTypePackage(deviceServerConfig, deviceType)
            + "." 
            + nameFactory.getModuleFactoryClass(deviceType);
        try {
            log.debug("Loading module factory " + moduleFactoryClassName);
            Class moduleFactoryClass = Class.forName(moduleFactoryClassName);
            if (moduleFactoryClass != null) {
                moduleFactory = (IBasicModuleFactory)moduleFactoryClass.newInstance();
                moduleFactories.put(deviceType.getName(), moduleFactory);
            }
        } catch (Exception e) {
            log.error("Initializing device type", e);
        }
        
    }
    
    /*
     * Creates a module factory for each device or controller type. 
     */
    private void initializeDeviceTypes() {
        
        log.debug("Initializing device types");
        
        for (CfgDeviceType deviceType : deviceServerConfig.getDefinition().getDeviceTypes()) {
            initializeDeviceType(deviceType);
        }
        
        log.debug("Initializing Controller types");
        
        for (CfgDeviceType deviceType : deviceServerConfig.getDefinition().getControllerTypes()) {
            initializeDeviceType(deviceType);
        }

    }
    
    @SuppressWarnings("unchecked")
    private void createDeviceModule(CfgDeviceGroup cfgDeviceGroup) {
        
        CfgDeviceType deviceType = cfgDeviceGroup.getDeviceType();
        IDeviceModuleFactory deviceModuleFactory = (IDeviceModuleFactory) moduleFactories.get(deviceType.getName());
        EquipmentModuleFactory equipmentModuleFactory = new EquipmentModuleFactory();
        DeviceEquipmentModuleBuilder equipmentModuleBuilder = new DeviceEquipmentModuleBuilder();
        
        equipmentModuleBuilder.setCfgDeviceGroup(cfgDeviceGroup);
        equipmentModuleBuilder.setCfgDeviceServer(deviceServerConfig);
        equipmentModuleBuilder.setDeviceServer(this);
        equipmentModuleBuilder.setHistoryMap(historyMap);
        equipmentModuleBuilder.setDeviceModuleFactory(deviceModuleFactory);
        
        equipmentModuleFactory.setEquipmentModuleBuilder(equipmentModuleBuilder);

        String moduleName = cfgDeviceGroup.getName();
        
        try {
            
            EquipmentModule equipmentModule = null;
            equipmentModule = equipmentModuleFactory.buildEquipmentModule();
            log.info("Equipment module " 
                + equipmentModule.getExportName() 
                + " [" 
                + equipmentModule.getLocalName() 
                + "] created.");
            modules.put(equipmentModule.getExportName(), equipmentModule);

        } catch(EquipmentModuleBuilderException e) {
            log.error("Equipment module " + moduleName + " could not be created: " + e, e);
        } catch(Exception e) {
            log.error("Equipment module " + moduleName + " could not be created: "  + e, e);
        }
        
    }
    
    @SuppressWarnings("unchecked")
    private void createControllerModule(CfgDeviceGroup cfgDeviceGroup) {
        
        CfgDeviceType deviceType = cfgDeviceGroup.getDeviceType();
        IControllerModuleFactory moduleFactory = (IControllerModuleFactory) moduleFactories.get(deviceType.getName());
        EquipmentModuleFactory equipmentModuleFactory = new EquipmentModuleFactory();
        ControllerEquipmentModuleBuilder equipmentModuleBuilder = new ControllerEquipmentModuleBuilder();
        
        equipmentModuleBuilder.setCfgDeviceGroup(cfgDeviceGroup);
        equipmentModuleBuilder.setCfgDeviceServer(deviceServerConfig);
        equipmentModuleBuilder.setDeviceServer(this);
        equipmentModuleBuilder.setHistoryMap(historyMap);
        equipmentModuleBuilder.setControllerModuleFactory(moduleFactory);
        
        equipmentModuleFactory.setEquipmentModuleBuilder(equipmentModuleBuilder);
        
        String moduleName = cfgDeviceGroup.getName();
        try {
            EquipmentModule equipmentModule = null ;
            equipmentModule = equipmentModuleFactory.buildEquipmentModule();
            log.info("Equipment module " 
                + equipmentModule.getExportName() 
                + " [" 
                + equipmentModule.getLocalName() 
                + "] created.");
            controllerModules.put(equipmentModule.getExportName(), equipmentModule);
        } catch (Exception e) {
            log.error("Equipment module " + moduleName + " could not be created: " + e, e);
        }

    }

    /**
     * Creates and initializes equipment modules.
     */
    private void initializeModules() {
        
        for (CfgDeviceGroup deviceGroup : deviceServerConfig.getConfiguration().getDeviceGroups()) {
            createDeviceModule(deviceGroup);
        }

        for (CfgDeviceGroup deviceGroup : deviceServerConfig.getConfiguration().getControllerGroups()) {
            createControllerModule(deviceGroup);
        }
        
    }
    
    /**
     * Iterates through all modules and calls setupDevices().
     * 
     * @see EquipmentModule#setupDevices()
     */
    private void setupModules() {
        for (EquipmentModule<? extends IBasicDevice> module : modules.values()) {
            module.setupDevices();
        }
        for (EquipmentModule<? extends IBasicDevice> module : controllerModules.values()) {
            module.setupDevices();
        }
    }

    /**
     * Runs the device server.
     * 
     * Starts the device server as a TINE server.
     * 
     * <nl>
     * <li>Establishes and runs a TINE background task which updates all modules.</li>
     * <li>Initializes and starts the TINE main loop</li>
     * <li>Runs to an infinite loop which can be interrupted.</li>
     * </nl>
     */
    public final void run() {
        
        try {
            
            if (setActive()) {
                
                log.info("Starting device server " + this.getClass().getSimpleName());
                TEquipmentModuleFactory tineServer = TEquipmentModuleFactory.getInstance();

                TEquipmentBackgroundTask backgroundTask = new TEquipmentBackgroundTask()
                { 
                    public void call()
                    {
                        notifyUpdateStarting();
                        try {
                            updateModules();
                        } catch (Exception e) {
                            log.fatal("Unhandled Exception " + e.toString(), e);
                        }
                        notifyUpdateFinished();
                    }
                };

                long updateInterval = (long)deviceServerConfig.getConfiguration().getServer().getUpdateInterval();
                log.debug("Starting background task with interval " + updateInterval + " ms.");
                backgroundTask.setBackgroundTaskInterval(updateInterval);  // msec
                TEquipmentModuleFactory.getInstance().addEquipmentBackgroundTask(backgroundTask);

                log.debug("Starting TINE server");
                // Find the first module className; This is used to locate the 
                // FEC information from fecid.csv.
                // TODO Problem besser lsen als mit FECID.CSV
                String firstExportName = null;
                for (EquipmentModule module : modules.values()) {
                    firstExportName = module.getExportName();
                    break;
                }
                tineServer.systemInit(null, 0, firstExportName);

                log.info("Starting main loop");
                for (;;) {
                    try {
                        Thread.sleep(1000L);
                    } catch (InterruptedException e) {
                        log.info("Device Server Thread interrupted.");
                    } catch (Exception e) {
                        log.fatal("Unhandled Exception", new Throwable(e));
                    }
                }
                
            } else {
                log.error("Attempt to start running device server");
            }
            
        } catch(Exception e) {
            log.fatal("Device server cannot be started.",  e);
        }
        setInactive();
    }

    /**
     * Returns the enironment variables.
     */
    public final HashMap<String, String> getEnvironment() {
        return environment;
    }

    /**
     * Returns a value of an environment variable.
     * @param attributeName Name of the environment variable.
     * @return Value or null.
     */
    public final Object getAttribute(String attributeName) {
        return environment.get(attributeName);
    }

    /**
     * Informs all update listeners that the update begins.
     */
    private void notifyUpdateStarting() {
        Date timeStamp = new Date();
        for (IUpdateListener listener : updateListeners) {
            try {
                listener.updateStarting(this, timeStamp);
            } catch (Exception e) {
                log.warn("Error in update listener: " + e.toString());
            }
        }
    }

    /**
     * Informs all update listeners that the update is starting.
     */
    private void notifyUpdateFinished() {
        Date timeStamp = new Date();
        for (IUpdateListener listener : updateListeners) {
            try {
                listener.updateFinished(this, timeStamp);
            } catch (Exception e) {
                log.warn("Error in update listener: " + e.toString());
            }
        }
    }       

    /**
     * Adds an update listener.
     * @param updateListener New update listener.
     */
    public final void addUpdateListener(IUpdateListener updateListener) {
        updateListeners.add(updateListener);
    }

    /**
     * Removes an update listener.
     * @param updateListener Update listener to be removed.
     */
    public final void removeUpdateListener(IUpdateListener updateListener) {
        updateListeners.remove(updateListener);
    }

    /**
     * Returns the application name from the server configuration.
     */
    public final String getApplicationName() {
        return deviceServerConfig.getConfiguration().getServer().getApplicationName();
    }
    
    /**
     * Appends a log configuration to the existing one.
     * Makes it possible for users to have a separate log
     * configuration.
     * Important for MstApps.
     * The file is searched in the config subrdirectory first,
     * then in the resources/config package.
     * If it is not found, see log messages for reasons.
     * 
     * @param propertiesFileName Name of the properties file.
     */
    public void addLog4jProperties(String propertiesFileName) {
        Log4jConfigurator.getInstance().appendConfiguration(this.getDirectorySet(), propertiesFileName);
    }

}
