/*
 * Created on Aug 27, 2007
 *
 */
package de.desy.mst.auxiliary.devl.deviceserverlib.server;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.util.Properties;
import java.util.Map.Entry;
import java.util.regex.Matcher;

import org.apache.log4j.ConsoleAppender;
import org.apache.log4j.Logger;
import org.apache.log4j.PatternLayout;
import org.apache.log4j.PropertyConfigurator;

import de.desy.mst.auxiliary.devl.deviceserverlib.config.CfgDeviceServer;


/**
 * A utility class for the configuration of Log4j.
 * Used by GenericDeviceServer.
 * 
 */
public class Log4jConfigurator {

    private static Logger log = Logger.getLogger(Log4jConfigurator.class);
    private static Log4jConfigurator theInstance = new Log4jConfigurator();

    private Log4jConfigurator() {
    }

    public static Log4jConfigurator getInstance() {
        return theInstance;
    }

    /**
     * Initializes log4j without configuration.
     * This is used before the configuration is read.
     * It enables log4j to write logs to the console before
     * it is really initialized. The reason is that the device
     * server wizard writes logs befor it reads the configuration
     * of the logging system.
     */
    public void prepareLogging() {
        System.setProperty("log4j.defaultInitOverride", "yes");
        Logger rootLog = Logger.getRootLogger();
        PatternLayout layout = new PatternLayout("%d{absolute} %-5p %-24C{1} %m%n");
        ConsoleAppender appender = new ConsoleAppender(layout);
        rootLog.addAppender(appender);
    }

    public void disableLogging() {
        Logger rootLog = Logger.getRootLogger();
        rootLog.removeAllAppenders();
    }

    private Properties readProperties(String propertyFileName, String configDirectoryName) {
        Properties properties = null;

        // 1. look in file system first, then in package.
        String configFileName = configDirectoryName + File.separator + propertyFileName;
        File configFile = new File(configFileName);
        InputStream inputStream = null;

        try {
            if (configFile.exists()) {
                log.debug("Reading properties from file: " + configFileName);
                inputStream = new FileInputStream(configFile);
            } else {
                String propertiesFile= "resources/config/" + propertyFileName;
                log.debug("Reading properties from " + propertiesFile);
                inputStream = getClass().getClassLoader().getResourceAsStream(propertiesFile);
            }
            if (inputStream == null) {
                log.error("Properties file " + propertyFileName + " not loaded!");
                properties = null;
            } else {
                properties = new Properties();
                properties.load(inputStream);
            }
        } catch (Exception e) {
            log.error("Cannot load log4j configuration: " + e, e);
            properties = null;
        }

        try {
            if (inputStream != null) inputStream.close();
        } catch(Exception e) {
            log.error("Cannot close input stream: " + e, e);
        }

        return properties;

    }

    /**
     * Initializes log4j configuration.
     * Finds the configuration file of log4j and decides 
     * if an external log4j.properties is used or if the 
     * device server configuration is used.
     * @param directorySet
     * @param logConfig
     */
    public void configure(DirectorySet directorySet, CfgDeviceServer deviceServer) {

        Properties properties = null;
       
        properties = readProperties("log4j.properties", directorySet.getConfigDirectory());
        if (properties != null) {

            // create regex-safe copy - take care of backslashes in replacement strings !
            String replacementLogDir = Matcher.quoteReplacement(directorySet.getLogDirectory());
            String replacementApplicationName = deviceServer.getConfiguration().getServer().getApplicationName();
            
//          replace variables with actual values
            for(Entry<Object, Object> entry : properties.entrySet()) {
                String value = (String) entry.getValue();
                String newValue = value;
                newValue = newValue.replaceAll("@logDir@", replacementLogDir);
                newValue = newValue.replaceAll("@applicationName@", replacementApplicationName);
                entry.setValue(newValue);
            }

//          Configure log4j.
            disableLogging();
            PropertyConfigurator.configure(properties);

        }
    }

    /**
     * Appends a log configuration at runtime.
     * Does not reset existing loggers.
     * @param directorySet
     * @param logFileName
     */
    public void appendConfiguration(DirectorySet directorySet, String logFileName) {
        Properties properties = null;
        properties = readProperties(logFileName, directorySet.getConfigDirectory());
        if (properties != null) {
            PropertyConfigurator.configure(properties);
        }
    }

}
