/*
 * Created on Sep 25, 2006
 *
 */
package de.desy.mst.auxiliary.devl.deviceserverlib.server;

import org.apache.log4j.Logger;

import de.desy.tine.dataUtils.TDataType;
import de.desy.tine.definitions.TAccess;
import de.desy.tine.definitions.TErrorList;
import de.desy.tine.server.connections.TClient;
import de.desy.tine.server.properties.TPropertyHandler;

/**
 * A bridge which makes a APropertyHandler look like a TPropertyHandler.
 * 
 * APropertyHandler was originally intended to be derived from 
 * TPropertyHandler. Then the problem arose that error message strings
 * could not be returned by the call() method and the 
 * return type of that method could not be changed in TINE without
 * breaking compatibility. In TINE, a method of an equipment module
 * has to be called to return a message, and a TPropertyHandler 
 * cannot do that because it does not know its equipment module. 
 * An APropertyHandler has a call() method which returns a TineResultSet,
 * consisting of an integer return code and a message. A PropertyHandlerBridge
 * is a TPropertyHandler which contains an APropertyHandler and a reference to an
 * Equipment module. In its call() method, the PropertyHandlerBridge 
 * calles the APropertyHandler's call() method and then passes 
 * returned information to the equipment module.
 * 
 * @author jwlg
 */
public class PropertyHandlerBridge<D extends IBasicDevice> extends TPropertyHandler {

    private EquipmentModule<D> eqpModule = null;
    private ABasicPropertyHandler propertyHandler = null;

    private static Logger accessLog = Logger.getLogger("de.desy.mst.auxiliary.devl.deviceserverlib.server.PropertyAccessLog");

    /**
     * Constructor.
     * @param eqpModule The equipment module.
     * @param propertyHandler The property handler.
     */
    public PropertyHandlerBridge(EquipmentModule<D> eqpModule, ABasicPropertyHandler<D> propertyHandler) {
        this.eqpModule = eqpModule;
        this.propertyHandler = propertyHandler;
        propertyHandler.setEquipmentModule(eqpModule);
    }

    /**
     * Calls the property and transmits the result to the equipment module.
     */
    protected int call(String devName, TDataType dout, TDataType din,
            TAccess devAccess) {
        TineResultSet result;
        String errorMessage;

        TClient client = eqpModule.getCaller();
        boolean logsEnabled = ((client != null) && devAccess.isWrite());

        if (client != null) {
             if (logsEnabled) {
                String clientHost = client.IPaddress.getHostName();
                if (clientHost == null || clientHost.equals("")) {
                    clientHost = client.IPaddress.getHostAddress();
                }

                accessLog.info("Call: " 
                    + "Property=" + propertyHandler.getDeviceGroup().getName() 
                    + "/" + devName 
                    + "/" + propertyHandler.getPropertyName()
                    + " Client=" + client.userName
                    + "@" + clientHost
                    + " Access=" + devAccess.toString()
                );
            }
        } else {
            if (logsEnabled) {
                accessLog.info("Call: " 
                    + "Property=" + propertyHandler.getDeviceGroup().getName() 
                    + "/" + devName 
                    + "/" + propertyHandler.getPropertyName()
                    + " Client=unknown@unknown"
                    + " Access=" + devAccess.toString()
                );
            }
        }


        result = propertyHandler.call(devName, dout, din, devAccess);
        if (result.getReturnCode() != 0) {
            if (result.getErrorMessage() != null) {
                eqpModule.setCompletionString(result.getErrorMessage());
            } else {
                eqpModule.setCompletionString("No error message specified");
            }
        }
        if (logsEnabled) {
            errorMessage = result.getErrorMessage();
            if (errorMessage == null) {
                errorMessage = TErrorList.getErrorString(result.getReturnCode());
            }
            accessLog.info("Result: " 
                + "Property=" + propertyHandler.getDeviceGroup().getName() 
                + "/" + devName 
                + "/" + propertyHandler.getPropertyName()
                + " Code=" + result.getReturnCode() 
                + " Message=\"" + errorMessage + "\"");
        }
        return result.getReturnCode();
    }

}
