/*
 * Created on Jan 24, 2007
 *
  */
package de.desy.mst.auxiliary.devl.deviceserverlib.server;

import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.util.HashMap;

import org.apache.log4j.Logger;

import de.desy.mst.auxiliary.devl.deviceserverlib.config.CfgDeviceGroup;
import de.desy.mst.auxiliary.devl.deviceserverlib.initialization.DatabaseReader;

/**
 * The standard device factory.
 * This device factory is used when no customized
 * device factory is required. It requires a device 
 * table which contains one line for each device.
 * It reads the table via SQL and creates a device
 * for each line.
 * 
 * @author jwlg
 */
public class StandardDeviceFactory<T extends IBasicDevice> extends ADeviceFactory<T> {

    private static Logger log = Logger.getLogger(StandardDeviceFactory.class);
    
    
    /**
     * Constructs a standard device factory  
     * with a module and a factory method.
     * @param module The device group.
     * @param factoryMethod The factory method.
     */
    public StandardDeviceFactory(CfgDeviceGroup module,
            IDeviceFactoryMethod<T> factoryMethod) {
        super.setCfgDeviceGroup(module);
        super.setFactoryMethod(factoryMethod);
    }
    
    public StandardDeviceFactory() {
        super();
    }
    
    /**
     * Creates the devices.
     * This method performs an SQL query and reads the 
     * result line by line. For each line, a device is created.
     */
    public void createDevices(DeviceGroup<T> deviceGroup) {

        CfgDeviceGroup cfgDeviceGroup = getCfgDeviceGroup();
        IDeviceFactoryMethod<T> factoryMethod = getFactoryMethod();

        
        if (cfgDeviceGroup == null) {
            log.error("Cannot create devices. CfgDeviceGroup is not set.");
            return;
        }
        if (factoryMethod == null) {
            log.error("Cannot create devices. FactoryMethod is not set.");
            return;
        }
        
        T newDevice = null;
        
        HashMap<String,String> parameters = new HashMap<String, String>(cfgDeviceGroup.getInitConfig().getParameters());
        parameters.put("system.defaultDirectory", getDirectorySet().getDatabaseDirectory());
        
        DatabaseReader databaseReader = new DatabaseReader(
            cfgDeviceGroup.getInitConfig().getType(),
            parameters);
        
        try {
            ResultSet resultSet = databaseReader.getResult();
            if (resultSet == null) {
                log.error("Could not read data for module " + cfgDeviceGroup.getName());
            } else {

                log.debug("Creating devices");
                
                ResultSetMetaData metaData = resultSet.getMetaData();
                for (int i = 1; i<= metaData.getColumnCount();i++) {
                    log.info("COLUMN: " + metaData.getColumnName(i) + " / " + metaData.getColumnLabel(i));
                }

                while(resultSet.next()) {
                    try {
                        newDevice = factoryMethod.createDevice(resultSet);
                        if (newDevice != null) {
                            deviceGroup.addDevice(newDevice);
                        } else {
                            log.warn("Device factory method returned null");
                        }
                    } catch (Exception e) {
                        e.printStackTrace();
                        log.warn("Creating device: " + e.toString());
                    }
                }
                log.debug(deviceGroup.size() + " devices created");
                resultSet.close();
                databaseReader.close();
            }
        } catch (ClassNotFoundException e) {
            log.error("Could not initialize devices for " + cfgDeviceGroup.getName() + ": " + e, e); 
        } catch (SQLException e) {
            log.error("Could not initialize devices for " + cfgDeviceGroup.getName() + ": " + e, e); 
        }
    }

}
