/*
 * Created on Jun 1, 2006
 *
 */
package de.desy.mst.auxiliary.devl.deviceserverlib.tools;

import de.desy.mst.auxiliary.devl.deviceserverlib.config.CfgOperationDataType;
import de.desy.mst.auxiliary.devl.deviceserverlib.config.CfgDeviceServer;
import de.desy.mst.auxiliary.devl.deviceserverlib.config.CfgDeviceType;
import de.desy.mst.auxiliary.devl.deviceserverlib.config.CfgOperation;
import de.desy.mst.auxiliary.devl.deviceserverlib.config.CfgProperty;

/**
 * A name generator.
 * 
 * The name factory defines names of classes, packages and methods which are to
 * be generated from the device server model.
 * 
 * The name factory was designed as a singleton, not a utility class, because
 * it can be passed by reference like any other object. This feature is heavily 
 * used by velocity templates.
 * 
 * @author jwlg
 */
public class NameFactory {

    public static String DEFAULT_PACKAGE_PREFIX = "de.desy.mst";
    private String packagePrefix = DEFAULT_PACKAGE_PREFIX;

    // The only instance of this class.
    private static NameFactory theInstance = new NameFactory();

    // Default constructor - private, since this is a singleton.
    private NameFactory() {
    }

    /**
     * Returns the instance of the NameFactory. The NameFactory is a singleton,
     * which means that there is only one instance of this class and the
     * contructor is private. The instance can be accessed by this method.
     * 
     * @return The instance.
     */
    public static NameFactory getInstance() {
        return theInstance;
    }


    /**
     * Returns the name of the main class if the 
     * project is an MstApp.
     * @param deviceServer The device server project.
     */
    public String getMstAppClass(CfgDeviceServer deviceServer) {
        return deviceServer.getProject().getName() + "App";
    }

    /**
     * Returns the name of the TINE property handler class for a given property.
     * 
     * The returned name is the default name of the TINE property handler class
     * which is to be generated.
     * 
     * @param propertyCfg A property configuration.
     * @return Name of the property handler class.
     */
    public String getPropertyClass(CfgProperty propertyCfg) {
        return "Property" + getCapitalizedName(propertyCfg.getName());
    }

    /**
     * Returns the name of the TINE property handler class for a given operation.
     * 
     * The returned name is the default name of the TINE property handler class
     * which is to be generated.
     * 
     * @param operationCfg An operation configuration.
     * @return Name of the property handler class.
     */
    public String getOperationClass(CfgOperation operationCfg) {
        return "Operation" + getCapitalizedName(operationCfg.getName());
    }

    /**
     * Returns the name of the TINE property handler class for a given
     * attribute.
     * 
     * The returned name is the default name of the TINE property handler class
     * which is to be generated.
     * 
     * @param propertyCfg A property configuration.
     * @return Name of the property handler class.
     */
    public String getAttributeClass(CfgProperty propertyCfg) {
        return "Attibute" + getCapitalizedName(propertyCfg.getName());
    }

    /**
     * Returns the name of the interface class for a given device type.
     * 
     * For each device type, an interface class has to be generated. The
     * standard name of the interface class is defined by this method.
     * 
     * @param deviceType
     *            The device type
     */
    public String getInterfaceClass(CfgDeviceType deviceType) {
        return "I" + deviceType.getName() + "Device";
    }

    /**
     * Returns the name of the device factory for a given device type.
     * @param deviceType The device type.
     */
    public String getCustomDeviceFactoryClass(CfgDeviceType deviceType) {
        return deviceType.getName() + "DeviceFactory";
    }


    /**
     * Returns the name of the abstract device class for a given device type.
     * 
     * For each device type, an abstract device class has to be generated. The
     * standard name of the abstract device class is defined by this method.
     * 
     * @param deviceType
     *            The device type
     */
    public String getAbstractDeviceClass(CfgDeviceType deviceType) {
        return "A" + deviceType.getName() + "Device";
    }

    /**
     * Returns the name of the abstract device class for a given device type.
     * 
     * For each device type, an abstract device class has to be generated. The
     * standard name of the abstract device class is defined by this method.
     * 
     * @param deviceType
     *            The device type
     */
    public String getAbstractControllerClass(CfgDeviceType deviceType) {
        return "A" + deviceType.getName() + "Controller";
    }

    
    /**
     * Returns the name of the common data class for a given device type.
     * 
     * @param deviceType
     * @return The device type
     */
    public String getCommonDataClass(CfgDeviceType deviceType) {
        return deviceType.getName() + "CommonData";
    }

    /**
     * Returns the name of the client device class for a given device type.
     * 
     * @param deviceType
     * @return The device type
     */
    public String getClientDeviceClass(CfgDeviceType deviceType) {
        return deviceType.getName() + "ClientDevice";
    }



    /**
     * Returns the name of the initial record class for a given device type.
     * 
     * For each device type, an initial record class has to be generated. The
     * standard name of the initial record class is defined by this method.
     * 
     * @param deviceType
     *            The device type
     */
    public String getInitialRecordClass(CfgDeviceType deviceType) {

        return deviceType.getName() + "Init";
    }

    /**
     * Returns the name of the property set class for a given device type.
     * 
     * For each device type, a property set class has to be generated. The
     * standard name of the property set class is defined by this method.
     * 
     * @param deviceType
     *            The device type
     */
    public String getPropertySetClass(CfgDeviceType deviceType) {

        return deviceType.getName() + "Properties";
    }

    /**
     * Returns the name of the interface class for a given device type.
     * 
     * For each device type, an interface class has to be generated. The
     * standard name of the interface class is defined by this method.
     * 
     * @param deviceType
     *            The device type
     */
    public String getDeviceFactoryMethodClass(CfgDeviceType deviceType) {

        return deviceType.getName() + "FactoryMethod";
    }

    /**
     * Returns the name of the default implementation class for a given device
     * type.
     * 
     * For each device type, an example implemantation class is generated. The
     * name of the example implemantation class is defined by this method.
     * 
     * @param deviceType
     *            The device type
     */
    public String getDeviceImplementationClass(CfgDeviceType deviceType) {

        return deviceType.getName() + "Device";
    }

    /**
     * Returns the name of the generated device type class for a given device
     * type.
     * 
     * @param deviceType
     *            The device type.
     */
    public String getModuleFactoryClass(CfgDeviceType deviceType) {
        return deviceType.getName() + "ModuleFactory";
    }

    /**
     * Returns the package name for a given device type.
     * 
     * The package name of a device type is built by appending the device type
     * name to the device server's package name.
     * 
     * @param deviceServer
     *            The device server configuration.
     * @param deviceType
     *            Device class.
     */
    public String getDeviceTypePackage(CfgDeviceServer deviceServer,
            CfgDeviceType deviceType) {
        return deviceServer.getProject().getPackageName()
        + "." + deviceType.getName().toLowerCase();
    }

    /**
     * Returns the name of the package where user-defined 
     * types are located.
     * @param deviceServer The device server configuration.
     */
    public String getUserTypesPackage(CfgDeviceServer deviceServer) {
        return deviceServer.getProject().getPackageName()
        + ".types";
    }

    /**
     * Returns the class name of the package where device server 
     * configurartion files are stored.
     * 
     * @param deviceServer The device server configuration.
     * @return Java package class name of the congig package.
     */
    public String getConfigPackage(CfgDeviceServer deviceServer) {
        return deviceServer.getProject().getPackageName() + ".config";
    }

    /**
     * Returns the class name of the resource file.
     */
    public String getConfigFileName() {
        return "ServerConfig.xml";
    }

    /**
     * Returns the prefix of all packages.
     * Also known as the organization part (e.g.de.desy.de).
     */
    public String getPackagePrefix() {
        return packagePrefix;
    }

    /**
     * Saaigns the prefix of all packages.
     * Also known as the organization part (e.g.de.desy.de).
     * @param packagePrefix New prefix.
     */
    public void setPackagePrefix(String packagePrefix) {
        this.packagePrefix = packagePrefix;
    }

    /**
     * Returns the property package name for a given device type.
     * 
     * The property package is the place where all TINE property handler classes
     * are stored.
     * 
     * @param deviceServer
     *            The device server configuration.
     * @param deviceType
     *            A device type.
     */
    public String getPropertiesPackage(CfgDeviceServer deviceServer,
            CfgDeviceType deviceType) {

        return getDeviceTypePackage(deviceServer, deviceType)
        + ".properties";
    }

    /**
     * Returns the package name of the configuration library.
     */
    public String getConfigLibPackage() {
        return getDeviceServerLibPackage() + ".config";
    }

    /**
     * Returns the package name of the model library.
     */
    public String getModelPackage() {
        return getDeviceServerLibPackage() + ".model";
    }

    /**
     * Returns the package name of the tools library.
     */
    public String getToolsPackage() {
        return getDeviceServerLibPackage() + ".tools";
    }
    
    /**
     * Returns the package name of the tools library.
     */
    public String getArrayToolsPackage() {
        return getToolsPackage() + ".array";
    }


    /**
     * Returns the package name of the exceptions library.
     */
    public String getExceptionsPackage() {
        return getDeviceServerLibPackage() + ".exceptions";
    }

    /**
     * Returns the package name of the alarms library.
     */
    public String getAlarmsPackage() {
        return getDeviceServerLibPackage() + ".alarms";
    }

    /**
     * Returns the package name of the device server library.
     */
    public String getDeviceServerLibPackage() {
        return "de.desy.mst.auxiliary.devl.deviceserverlib";
    }

    /**
     * Returns the package name of the server library.
     */
    public String getServerLibPackage() {
        return getDeviceServerLibPackage() + ".server";
    }

    /**
     * Returns the package name of the initialization library.
     */
    public String getInitializationPackage() {
        return getDeviceServerLibPackage() + ".initialization";
    }

    /**
     * Returns the base package name of the TINE library.
     */
    public String getTinePackage() {
        return "de.desy.tine";
    }

    /**
     * Returns the package name of the TINE type library.
     */
    public String getTineTypesPackage() {
        return getTinePackage() + ".types";
    }


    /**
     * Returns the name of the constant which corresponds to an alarm.
     * @param alarmName Alarm name.
     * @return Alarm name in uppercase letters.
     */
    public String getAlarmConstantName(String alarmName) {
        return alarmName.toUpperCase();
    }

    /**
     * Returns the name of the class which contains the 
     * alarm constants.
     */
    public String getAlarmsClass(CfgDeviceType deviceType) {
        return deviceType.getName() + "Alarms";
    }


    /**
     * Returns a field name for a given property name.
     * 
     * The field name is defined as the property name, starting with a lower
     * case letter. The field name is particularly used in BeanUtils which
     * expects a propererty name to start with a lower case letter.
     * 
     * @param propertyName
     *            Name of the property.
     */
    public String getFieldName(String propertyName) {
        return StringUtils.getInstance().firstLetterToLowerCase(propertyName);
    }

    /**
     * Returns the capitalized name for a given property name.
     * 
     * This method is useful for finding getters and setters for given property
     * names or field names. Only the first letter is capitalized.
     * 
     * @param propertyName
     *            Name of a property.
     */
    public String getCapitalizedName(String propertyName) {
        return StringUtils.getInstance().firstLetterToUpperCase(propertyName);
    }

    /**
     * Returns the getter name for a given property.
     * 
     * The getter name starts with the prefix "get" and ends with the property
     * name which begins with an upper case letter.
     * 
     * @param propertyName
     *            Name of a property.
     */
    public String getGetterName(String propertyName) {
        return "get" + StringUtils.getInstance().firstLetterToUpperCase(propertyName);
    }

    /**
     * Returns the setter name for a given property.
     * 
     * The setter name starts with the prefix "set" and ends with the property
     * name which begins with an upper case letter.
     * 
     * @param propertyName
     *            Name of a property.
     */
    public String getSetterName(String propertyName) {
        return "set" + StringUtils.getInstance().firstLetterToUpperCase(propertyName);
    }

    /**
     * Returns the method name for a given operation.
     * 
     * The method name starts with the prefix "do" and ends with the operation
     * name which begins with an upper case letter.
     * 
     * @param operationName
     *            Name of an operation.
     */
    public String getOperationMethodName(String operationName) {
        return "do" + StringUtils.getInstance().firstLetterToUpperCase(operationName);
    }



    /**
     * Returns the getter declaration for a given property.
     * 
     * The getter declaration starts with the returned type, then follows the
     * getter name with an empty parenthesis.
     * 
     * @param propertyName
     *            Name of a property.
     * @param propertyType
     *            Type of a property - simple class name. Don't forget "[]" for
     *            arrays.
     */
    public String getGetterDeclaration(String propertyName, String propertyType) {
        return propertyType + " " + getGetterName(propertyName) + "()";
    }

    /**
     * Returns the getter declaration for a given property.
     * 
     * The getter declaration starts with the returned type, then follows the
     * getter name with an empty parenthesis.
     * 
     * @param property
     *            A property.
     */
    public String getGetterDeclaration(CfgProperty property) {
        String propertyType = property.getType().getSimpleClassName();
        if (property.isArray()) {
            propertyType = propertyType + "[]";
        }
        return getGetterDeclaration(property.getName(), propertyType);
    }

    /**
     * Returns the method declaration for a given operation.
     * 
     * The declaration starts with the returned type, then follows the
     * method name with an input parameter in parentheses.
     * 
     * @param operation
     *            An operation configuration.
     */
    public String getOperationMethodDeclaration(CfgOperation operation) {

        String inputType;
        String outputType;

        CfgOperationDataType inputData = operation.getInputData();
        CfgOperationDataType outputData = operation.getOutputData();

        if (inputData.isUserDefined()) {
            inputType = inputData.getUserType().getClassName();
        } else {
            inputType = inputData.getPropertyType().getSimpleClassName();
        }
        if (inputData.isArray()) {
            inputType = inputType + "[]";
        }

        if (outputData.isUserDefined()) {
            outputType = outputData.getUserType().getClassName();
        } else {
            outputType = outputData.getPropertyType().getSimpleClassName();
        }
        if (outputData.isArray()) {
            outputType = outputType + "[]";
        }
        return getOperationMethodDeclaration(operation.getName(), inputType, outputType);
    }

    /**
     * Returns the method declaration for a given operation.
     * 
     * @param operationName
     *            Name of the operation.
     * @param inputType
     *            Input type of the operation (simple class name).
     *            
     * @param outputType
     *            Output type of the operation (simple name).
     */
    public String getOperationMethodDeclaration(String operationName, String inputType, String outputType) {

        return outputType + " " + getOperationMethodName(operationName) 
        + "(" + inputType + " newValue)";
    }


    /**
     * Returns the setter declaration for a given property.
     * 
     * @param propertyName
     *            Name of a property.
     * @param propertyType
     *            Type of a property - simple class name.
     */
    public String getSetterDeclaration(String propertyName, String propertyType) {

        String argument = "";
        if (!propertyType.equals("void")) {
            argument = propertyType + " newValue";
        }
        return "void " + getSetterName(propertyName) 
            + "(" + argument + ")";
    }

    /**
     * Returns the setter declaration for a given property.
     * 
     * The setter declaration returns void and has a value or an object (the new
     * value to be set) as its parameter.
     * 
     * @param property
     *            A property.
     */
    public String getSetterDeclaration(CfgProperty property) {

        String propertyType = property.getType().getSimpleClassName();
        if (property.isArray()) {
            propertyType = propertyType + "[]";
        }
        return getSetterDeclaration(property.getName(), propertyType);
    }

}
