/*
 * Created on Jul 3, 2006
 *
 */
package de.desy.mst.auxiliary.devl.deviceserverlib.tools;

/**
 * Provides String utilities.
 * 
 * StringUtils is designed as a singleton, not a utility class, because
 * it can be passed by reference like any other object. This feature is heavily 
 * used by velocity templates.
 */
public class StringUtils {


    private static StringUtils theInstance = new StringUtils();

    /**
     * Returns the instance.
     */
    public static StringUtils getInstance() {
        return theInstance;
    }

    /**
     * Private constructor.
     * This is a singleton.
     */
    private StringUtils() {
    }

    /**
     * Converts the first character of a string to upper case.
     * 
     * @param inputString Input string.
     * @return Converted string.
     */
    public String firstLetterToUpperCase(String inputString) {

        String result = inputString; // Default result
        Character firstChar;

        if (inputString != null) {
            if (!inputString.equals("")) {
                firstChar = inputString.charAt(0);
                if (!Character.isUpperCase(firstChar)) {
                    result = Character.toUpperCase(firstChar) + inputString.substring(1);
                } 
            }
        }

        return result;
    }

    /**
     * Converts the first character of a string to lower case.
     * 
     * @param inputString Input string.
     * @return Converted string.
     */
    public String firstLetterToLowerCase(String inputString) {

        String result = inputString; // Default result
        Character firstChar;

        if (inputString != null) {
            if (!inputString.equals("")) {
                firstChar = inputString.charAt(0);
                if (!Character.isLowerCase(firstChar)) {
                    result = Character.toLowerCase(firstChar) + inputString.substring(1);
                }
            }
        }

        return result; 
    }

    /**
     * Replaces dots with slashes.
     * @param packageName
     */
    public String packageToDir(String packageName, boolean startSlash) {
        String result = packageName.replace('.', '/');;
        if (startSlash) {
            result = "/" + result; 
        }
        return result;
    }

    /**
     * A hash algorithm.
     * This algorithm was created for sdbm (a public-domain reimplementation 
     * of ndbm) database library.
     * @param inputString
     * @return Hash value.
     */
    public int sdbmHash(String inputString) {
        int hashValue = 0;
        for (char letter : inputString.toCharArray()) {
            hashValue = (hashValue << 6) + (hashValue << 16) - hashValue + letter;
        }
        return hashValue;
    }

    /**
     * A hash algorithm.
     * This algorithm (k=33) was first reported by Dan Bernstein 
     * many years ago in comp.lang.c
     * @param inputString
     * @return Hash value.
     */
    public int djb2Hash(String inputString) {
        int hashValue = 0;
        for (char letter : inputString.toCharArray()) {
            hashValue = ((hashValue << 5) + hashValue) + letter;
        }
        return hashValue;
    }

    /**
     * A hash algorithm.
     * This algorithm (k=33) was first reported by Dan Bernstein 
     * many years ago in comp.lang.c
     * @param inputString
     * @return Hash value.
     */
    public int xorHash(String inputString) {
        int hashValue = 0;
        for (char letter : inputString.toCharArray()) {
            hashValue = ((hashValue << 5) + hashValue) ^ letter;
        }
        return hashValue;
    }

    /**
     * A hash Algorithm.
     * Published in Josua Boch, Effective Java.
     * @param inputString
     * @return a hash value.
     */
    public int blochHash(String inputString) {
        int hashValue = 17;
        // Primitive hash calculation
        for (char letter : inputString.toCharArray()) {
            hashValue = (37*hashValue + (int)letter);
        }
        return hashValue;
    }

    /**
     * A comparer which handles null objects.
     * This is a wrapper for Comparable.compareTo(). It
     * works when one object or both are null.
     * @param thisObj First string to compare.
     * @param otherObj Second string to compare.
     * @return this.compareTo(other) if both are != null.
     * If one object is null, it is considered smaller.
     * If both are null, they are considered equal.
     * 
     * This method does not really belong to StringUtils
     * since it is rather generic.
     * 
     */
    @SuppressWarnings("unchecked")
    public int compareTo(Comparable thisObj, Comparable otherObj) {
        int result = 0;
        if (thisObj != null) {
            if (otherObj != null) {
                result = thisObj.compareTo(otherObj);
            } else { // otherObj is null => thisObj is greater.
                result = 1;
            }
        } else { // thisObj is null
            if (otherObj != null) {
                // otherObj is not null => thisObj is smaller.
                result = -1;
            } else {
                // both are null = equal
                result = 0;
            }
        }
        return result;
    }
    
    /**
     * Compares if two strings are equal.
     * @param string1
     * @param string2
     */
    public boolean equals(String string1, String string2) {
        boolean result = false;
        if (string1 == null) {
            result = (string2 == null);
        } else {
            result = string1.equals(string2);
        }
        return result;
    }

    /**
     * Converts a char array to a String.
     * Removes trailing whitespaces.
     * @param charArray
     */
    public String charArrayToString(char[] charArray) {

        int whitespaceCount = 0;
        String result = null;
        if (charArray == null) {
            result = null;
        } else if (charArray.length == 0) {
            result = "";
        } else {
            for (int i = charArray.length - 1; i>=0; i--) {
                if (Character.isWhitespace(charArray[i])) {
                    whitespaceCount++;
                } else {
                    break;
                }
            }
            result = String.valueOf(charArray, 0, 
                charArray.length - whitespaceCount);
        }
        return result;
    }

    /**
     * Copies a string to a given char array.
     * If the string is longer than the char array, only
     * as many characters as possible will be copied from
     * the start of the string. If the string is shorter,
     * the char array will be filled with a specified fill 
     * character.
     * @param src The string to be copied.
     * @param dest The destination array.
     * @param fillChar The fill character.
     */
    public void stringToCharArray(String src, char[] dest, char fillChar) {
        if ((src != null) && (dest != null) && (dest.length > 0)) { 
            char[] buffer = src.toCharArray();
            for (int i =0; i< dest.length; i++) {
                if (i<buffer.length) {
                    dest[i] = buffer[i];
                } else {
                    dest[i] = fillChar;
                }
            }
        }
    }

    /**
     * Converts a String to a fixed-length char array.
     * If the string is longer than the char array, only
     * as many characters as possible will be copied from
     * the start of the string. If the string is shorter,
     * the char array will be filled with spaces.
     * 
     * @param src The string to be copied.
     * @param dest The destination array.
     */
    public void stringToCharArray(String src, char[] dest) {
        stringToCharArray(src, dest, ' ');
    }

    /**
     * Trim trailing whitespace from the given String.
     * @param str the String to check
     * @return the trimmed String
     * @see java.lang.Character#isWhitespace
     */
    public static String trimTrailingWhitespace(String str) {
        if (!hasLength(str)) {
            return str;
        }
        StringBuffer buf = new StringBuffer(str);
        while (buf.length() > 0 && Character.isWhitespace(buf.charAt(buf.length() - 1))) {
            buf.deleteCharAt(buf.length() - 1);
        }
        return buf.toString();
    }

    /**
     * Check if a String has length.
     * <p><pre>
     * StringUtils.hasLength(null) = false
     * StringUtils.hasLength("") = false
     * StringUtils.hasLength(" ") = true
     * StringUtils.hasLength("Hello") = true
     * </pre>
     * @return <code>true</code> if the String is not null and has length
     */
    public static boolean hasLength(String str) {
        return (str != null && str.length() > 0);
    }
    
    /**
     * Returns true if a string is null 
     * or empty.
     * @param txt String to be tested.
     * @return
     */
    public static boolean isEmpty(String txt) {
        return ((txt == null) || txt.equals(""));
    }

}
