/*
 * Created on Aug 16, 2007
 *
  */
package de.desy.mst.auxiliary.devl.deviceserverlib.tools.selector;

import java.util.List;
import java.util.Set;

/**
 * Common interface for all selectors.
 * A selector is an object which determines if another object
 * is selected. If an object is selected depends on the state
 * of the object. The selector knows its class and calls its
 * methods. 
 * Any selector has a method isSelected() which determines if
 * an object is selected.
 * There are also some convenience methods which work on collections
 * of objects. They are implemented by ASelector.
 * @author jwlg
 */
public interface ISelector<T> {
    
    /**
     * Returns true if an object is selected.
     * An derived class must implement the selection
     * criteria.
     * @param item The object to be testet.
     */
    public abstract boolean isSelected(T item);

    /**
     * Selects a list of selected objects from a list of objects.
     * This method should traverse a source list and add all
     * selected objects to the target list. The target list
     * is also returned. 
     * @param source The source list.
     * @param target The target list.
     * @return The target list.
     */
    public List<T> select(List<T> source, List<T> target);

    /**
     * Selects a set of selected objects from a set of objects.
     * This method should traverse a source set and add all
     * selected objects to the target set. The target set
     * is also returned. 
     * @param source The source set.
     * @param target The target set.
     * @return The target list.
     */
    public Set<T> select(Set<T> source, Set<T> target);

    /**
     * Selects a list of selected objects from an array of objects.
     * This method should traverse a source array and add all
     * selected objects to the target list. The target list
     * is also returned. 
     * @param source The source list.
     * @param target The target list.
     * @return The target list.
     */
    public List<T> select(T[] source, List<T> target);

}
