/*
 * Created on Jun 7, 2006
 *
 */
package de.desy.mst.auxiliary.devl.deviceserverlib.tools.selector;

import de.desy.mst.auxiliary.devl.deviceserverlib.config.CfgPropertyType;


/**
 * Type selector factory.
 * 
 * The class TypeSelector provides different selector objects
 * for certain property types.
 * 
 * It is built as a singleton in order to be useable inside of 
 * Velocity templates.
 * 
 * @author jwlg
 * @see ASelector
 * 
 */
public class TypeSelector {

    private static TypeSelector theInstance = null;
    
    /**
     * A String selector.
     * This class is private. The corresponding factory method
     * should be used to get an instance. 
     * @author jwlg
     * @see getStringSelector()
     */
    private class StringSelector extends ASelector<CfgPropertyType> {
        public boolean isSelected(CfgPropertyType item) {
            return item.isString();
        }
    }
 
    /**
     * An Object selector.
      * This class is private. The corresponding factory method
     * should be used to get an instance. 
    * @author jwlg
     * @see getStringSelector()
     */
   private class ObjectSelector extends ASelector<CfgPropertyType> {
        public boolean isSelected(CfgPropertyType item) {
            return item.isObject();
        }
    }

   /**
    * A Simple Type selector.
    * This class is private. The corresponding factory method
    * should be used to get an instance. 
    * @author jwlg
    * @see getSimpleSelector()
    */
   private class SimpleSelector extends ASelector<CfgPropertyType> {
        public boolean isSelected(CfgPropertyType item) {
            return item.isSimple();
        }
    }

   /**
    * A Boolean selector.
    * This class is private. The corresponding factory method
    * should be used to get an instance. 
    * @author jwlg
    * @see getBooleanSelector()
    */
    private class BooleanSelector extends ASelector<CfgPropertyType> {
        public boolean isSelected(CfgPropertyType item) {
            return item.isBoolean();
        }
    }
    
    /**
     * A Tine Type ASelector.
     * This class is private. The corresponding factory method
     * should be used to get an instance. 
     * @author jwlg
     * @see getBooleanSelector()
     */
    private class TineTypeSelector extends ASelector<CfgPropertyType> {
        public boolean isSelected(CfgPropertyType item) {
            return item.isTineType();
        }
    }

    /**
     * A selector for everything.
     * This class is private. The corresponding factory method
     * should be used to get an instance. 
     * @author jwlg
     */
    private class AllSelector extends ASelector<CfgPropertyType> {
        public boolean isSelected(CfgPropertyType item) {
            return true;
        }
    }

    /**
     * Constructor.
     * The constructor for this class is private since
     * it is a singleton which can have only one instance.
     * @see getInstance()
     */
    private TypeSelector() {
        super();
    }

    /**
     * Returns the instance.
     * @return The only instance of TypeSelector.
     */
    public static TypeSelector getInstance() {
        if (theInstance == null) {
            theInstance = new TypeSelector();
        }
        return theInstance;
    }
    
    /**
     * Returns a String selector.
     * A String selector is an instance of ASelector<CfgPropertyType>
     * which selects Strings.
     * @return A String selector.
     */
    public ASelector<CfgPropertyType> getStringSelector() {
        return new StringSelector();
    }
 
    /**
     * Returns an Object selector.
     * A Object selector is an instance of ASelector<CfgPropertyType>
     * which selects Objects.
     * @return A Object selector.
     */
    public ASelector<CfgPropertyType> getObjectSelector() {
        return new ObjectSelector();
    }
    
    /**
     * Returns a TINE Type selector.
     * A TINE Type selector is an instance of ASelector<CfgPropertyType>
     * which selects TINE Types.
     * @return A TINE Type selector.
     */
    public ASelector<CfgPropertyType> getTineTypeSelector() {
        return new TineTypeSelector();
    }
    
    /**
     * Returns a Simple selector.
     * A Simple selector is an instance of ASelector<CfgPropertyType>
     * which selects simple types.
     * @return A Simple seletor.
     */
    public ASelector<CfgPropertyType> getSimpleSelector() {
        return new SimpleSelector();
    }
    
    /**
     * Returns a Boolean selector.
     * A Boolean selector is an instance of ASelector<CfgPropertyType>
     * which selects booleans.
     * @return A Boolean seletor.
     */
    public ASelector<CfgPropertyType> getBooleanSelector() {
        return new BooleanSelector();
    }
    
    /**
     * Returns an All selector.
     * An All selector is an instance of ASelector<CfgPropertyType>
     * which selects everything.
     * @return A All seletor.
     */
    public ASelector<CfgPropertyType> getAllSelector() {
        return new AllSelector();
    }
    
    
    /**
     * Returns a selector of types which can be used as operation types.
     */
    public ISelector<CfgPropertyType> getOperationTypeSelector() {
        OrSelectorSet<CfgPropertyType> selector = new OrSelectorSet<CfgPropertyType>();
        selector.addSelector(TypeSelector.getInstance().getSimpleSelector());
        //selector.addSelector(TypeSelector.getInstance().getStringSelector());
        selector.addSelector(TypeSelector.getInstance().getTineTypeSelector());
        return selector;
    }
}
