/*
 * Created on Jun 20, 2007
 *
 */
package de.desy.mst.auxiliary.devl.deviceserverlib.tools.tine;

import java.util.HashMap;
import java.util.TreeSet;

import de.desy.mst.auxiliary.devl.deviceserverlib.config.CfgDeviceServer;
import de.desy.mst.auxiliary.devl.deviceserverlib.config.CfgTypeDefinition;
import de.desy.mst.auxiliary.devl.deviceserverlib.tools.StringUtils;

/**
 * A facility to create tags.
 * A TagFactory is uzed to create TINE tags from class
 * names. It contains a list of class names and TINE tags
 * which are already used. When it creates a tag, it looks 
 * up that list and finds a tag which does not yet exist.
 * 
 * @author jwlg
 */
public class TagFactory {

    private static final int MAX_TAG_LENGTH = 8;
    private HashMap<String, String> typeMap = new HashMap<String, String>();
    private TreeSet<String> tagSet = new TreeSet<String>();

    /**
     * Constructor.
     * Creates a tag factory which contains no 
     * registered classes and tags.
     */
    public TagFactory() {
    }

    /**
     * Constructor.
     * Creates a tag factory and fills it with entries from data
     * type definitions of a device server configuration.
     * @param deviceServer A device server configuration.
     */
    public TagFactory(CfgDeviceServer deviceServer) {
        for (CfgTypeDefinition typeDef : deviceServer.getDefinition().getDataTypes().getTypeDefinitions()) {
            if (typeDef.isTagged()) {
                typeMap.put(typeDef.getClassName(), typeDef.getTag());
                tagSet.add(typeDef.getTag());
            }
        }
    }
    
    /**
     * Returns true if a given tag name is
     * registered for a given class name.
     * @param className A class name.
     * @param tagName A tag name.
     */
    public boolean isTagOf(String className, String tagName) {
        String foundTag = typeMap.get(className);
        return StringUtils.getInstance().equals(foundTag, tagName);
    }
    
    /**
     * Creates a new tag name for a given class name.
     * @param className
     */
    public String createNewTagName(String className) {

        String baseName = className;
        String result = null;
        int counter = 0;

        if ((className != null) && (!className.equals(""))) {
            if (baseName.length() > MAX_TAG_LENGTH) {
                result = baseName.substring(0, MAX_TAG_LENGTH);
            } else {
                result = baseName;
            }
            // increment counter while the tag exists and
            // belongs to another class.
            while (existsTag(result) && (!isTagOf(className, result))) {
                counter++;
                result = addNumber(baseName,counter);
            }
        }
        return result;
    }

    /**
     * Adds a number to a string.
     * The result fits the maximun tag length.
     */
    private String addNumber(String tagName, int number) {
        String numberString = Integer.toString(number);
        int maxStringLength = MAX_TAG_LENGTH - numberString.length();
        if (tagName.length() > maxStringLength) {
            return tagName.substring(0, maxStringLength) + numberString;
        } else {
            return tagName + numberString;
        }
    }

    /**
     * Proposes a tag name for a given class name.
     * If the class is already registered, the method returns the
     * registered tag. Otherwise, it creates a new tag. 
     * 
     * @param className A class name.
     */
    public String proposeTagName(String className) {
        String result = getTagOf(className);
        if (result == null) {
            result = createNewTagName(className);
        }
        return result;
    }

    /**
     * Returns true if a tag is registered.
     * @param tagName
     */
    public boolean existsTag(String tagName) {
        return tagSet.contains(tagName);
    }

    /**
     * Returns the tag of a registered class.
     * @param className A class name.
     * @return The corresponding tag or null if none was found.
     */
    public String getTagOf(String className) {
        return typeMap.get(className);
    }

    /**
     * Registers a class name and a corresponding tag.
     * @param className A class name.
     * @param tag A tag.
     */
    public void registerTag(String className, String tag) {
        if (typeMap.containsKey(className)) {
            String oldTag = typeMap.get(className);
            tagSet.remove(oldTag);
        } 
        typeMap.put(className, tag);
        tagSet.add(tag);
    }

}
