/*
 * Created on Mar 14, 2007
 *
 */
package de.desy.mst.auxiliary.devl.deviceserverlib.tools.tine;

import org.apache.log4j.Logger;

import de.desy.mst.auxiliary.devl.deviceserverlib.config.CfgOperationDataType;
import de.desy.mst.auxiliary.devl.deviceserverlib.config.CfgDeviceServer;
import de.desy.mst.auxiliary.devl.deviceserverlib.config.CfgLocalArchive;
import de.desy.mst.auxiliary.devl.deviceserverlib.config.CfgOperation;
import de.desy.mst.auxiliary.devl.deviceserverlib.config.CfgProperty;
import de.desy.mst.auxiliary.devl.deviceserverlib.config.CfgPropertyDataInfo;
import de.desy.mst.auxiliary.devl.deviceserverlib.config.CfgTypeDefinition;
import de.desy.tine.definitions.TAccess;
import de.desy.tine.definitions.TArrayType;
import de.desy.tine.definitions.TFormat;
import de.desy.tine.server.histories.THistorySpecification;
import de.desy.tine.server.properties.TExportProperty;
import de.desy.tine.server.properties.TPropertyDescription;
import de.desy.tine.server.properties.TPropertyEGU;
import de.desy.tine.structUtils.TTaggedStructure;


/**
 * A factory which creates instances of TINE classes.
 * 
 * TINE needs objects of special classes to register
 * properties and local history records. This factory
 * generates such objects from property or 
 * operation specifications as used by the device server
 * wizard.
 * 
 * @author jwlg
 */
public class TineClassFactory {

    private static final TineClassFactory theInstance = new TineClassFactory();
    private static Logger log = Logger.getLogger(TineClassFactory.class);

    /**
     * Returns the instance of TineClassFactory.
     */
    public static TineClassFactory getInstance() {
        return theInstance;
    }

    /**
     * The default constrictor.
     */
    private TineClassFactory() {
    }

    /**
     * Returns a history specification for a given property.
     * @param property
     * @return A history specification which is needed to 
     * register a local history record.
     */
    public THistorySpecification getTHistorySpecification(CfgProperty property) {

        THistorySpecification result = null;
        CfgLocalArchive localArchive = property.getLocalArchive();

        result = new THistorySpecification(
            localArchive.getPollingRate(),
            localArchive.getArchiveRate(),
            localArchive.getShortTermStorageDepth(),
            localArchive.getLongTermStorageDepth(),
            localArchive.getArchiveHeartbeat(),
            (float)localArchive.getPercentTolerance(),
            (float)localArchive.getAbsoluteTolerance(),
            localArchive.getRemoteServer());

        return result;

    }

    /**
     * Creates an operation description for an operation.
     * @param operation The property configuration.
     * @return A TPropertyDescription which is needed by TINE.
     */
    public TPropertyDescription getTPropertyDescription(CfgOperation operation) {

        TPropertyDescription result = new TPropertyDescription();
        CfgOperationDataType outputData = operation.getOutputData();
        short arrayType = TArrayType.AT_SINGLE;



        if (outputData.isUserDefined()) {
            arrayType |= TArrayType.AT_BLOB;
        } else {

            if (outputData.getPropertyType().isBoolean()) {
                arrayType |= TArrayType.AT_BOOLEAN;
            }
            if (outputData.getPropertyType().isObject()) {
                arrayType |= TArrayType.AT_STRUCTURED;
            } else if (outputData.isArray()) {
                arrayType |= TArrayType.AT_SPECTRUM;
            } else {
                arrayType |= TArrayType.AT_SCALAR;
            }
        }

        result.setArrayType(arrayType);
        result.setText(operation.getDescription());

        return result;
    }

    /**
     * Creates a property description for a property.
     * @param property The property configuration.
     * @return A TPropertyDescription which is needed by TINE.
     */
    public TPropertyDescription getTPropertyDescription(CfgProperty property) {

        TPropertyDescription result = new TPropertyDescription();
        TPropertyEGU yRange = new TPropertyEGU(); 

        CfgPropertyDataInfo xDataInfo = property.getXDataInfo();
        CfgPropertyDataInfo yDataInfo = property.getYDataInfo();

        yRange.setMinValue((float)yDataInfo.getMinimumValue());
        yRange.setMaxValue((float)yDataInfo.getMaximumValue());
        yRange.setUnits(yDataInfo.getUnits());
        yRange.setGraphType((short)yDataInfo.getGraphType().getKey());

        // Wahrscheinlich nur bei Array-Properties von Nutzen.
        TPropertyEGU xRange = new TPropertyEGU();
        xRange.setMinValue((float)xDataInfo.getMinimumValue());
        xRange.setMaxValue((float)xDataInfo.getMaximumValue());
        xRange.setUnits(xDataInfo.getUnits());
        xRange.setGraphType((short)xDataInfo.getGraphType().getKey());

        short arrayType = TArrayType.AT_SINGLE;
        if (property.getType().isBoolean()) {
            arrayType |= TArrayType.AT_BOOLEAN;
        }
        if (property.getType().isObject()) {
            arrayType |= TArrayType.AT_STRUCTURED;
        } else if (property.isArray()) {
            arrayType |= TArrayType.AT_SPECTRUM;
        } else {
            arrayType |= TArrayType.AT_CHANNEL;
        }
        result.setArrayType(arrayType);

        result.setText(property.getDescription());
        return result;
    }

    /**
     * Creates a TExportProperty for a property configuration.
     * 
     * The number of devices is required for scalar properties
     * which are made accessible as a multi-channel array in 
     * the TINE interface.The length of the multi-channel array
     * is the number of the devices.
     * 
     * @param cfgProperty The property configuration.
     * @param nDevices The number of devices.
     * 
     * @return A TExportProperty which is required to register a
     * property in TINE.
     */
    public TExportProperty getTExportProperty(CfgProperty cfgProperty, int nDevices) {

        int accessMode = 0;
        int arraySize = 0;
        TExportProperty result = new TExportProperty();
        short tineFormat = (short)cfgProperty.getType().getTineFormatKey();

        if (cfgProperty.isArray()) {
            arraySize = cfgProperty.getArrayLength();
        } else {
            arraySize = nDevices;
        }

        result.setName(cfgProperty.getName());

        if (cfgProperty.isReadable()) {
            accessMode |= TAccess.CA_READ;
            result.setOutputFormat(tineFormat);
            result.setOutputSize(arraySize);
        } else {
            result.setOutputFormat(tineFormat);
            result.setOutputSize(0);
        }

        if (cfgProperty.isWriteable()) {
            accessMode |= TAccess.CA_WRITE;
            result.setInputFormat(tineFormat);
            result.setInputSize(arraySize);
        } else {
            result.setInputFormat(tineFormat);
            result.setInputSize(0);
        }

        log.info("TINE Export Property generated: " 
            + result.getName() 
            + ". Input:" + TFormat.toString(result.getInputFormat()) 
            + "[" + result.getInputSize() + "]" 
            + " - Output: " + TFormat.toString(result.getOutputFormat()) 
            + "[" + result.getOutputSize() + "]");

        result.setAccessMode((short)accessMode);
        result.setDescription(getTPropertyDescription(cfgProperty));
        return result;
    }


    /**
     * Calculates the data size of a class.
     * Same as calculateDataSize(String), but with separated
     * package and class names.
     * 
     * @param typePackage The package.
     * @param className The class name.
     * @throws ClassNotFoundException
     * @throws InstantiationException
     * @throws IllegalAccessException
     */
    public int calculateDataSize(String typePackage, String className) throws ClassNotFoundException, InstantiationException, IllegalAccessException {
        return  calculateDataSize(typePackage + "." + className);
    }

    /**
     * Calculates the data size of a class.
     * The method creates an instance of the spoecified class
     * and calls TTaggedStructure.getSizeInBytes().
     * The class must be derived from TTaggedStructure!
     * @param fullClassName Full name of the class.
     * @throws ClassNotFoundException
     * @throws InstantiationException
     * @throws IllegalAccessException
     */
    public int calculateDataSize(String fullClassName) throws ClassNotFoundException, InstantiationException, IllegalAccessException {
        Class inputDataClass = Class.forName(fullClassName);
        TTaggedStructure object = (TTaggedStructure) inputDataClass.newInstance();
        return object.getSizeInBytes();
    }


    /**
     * Creates a TExportProperty for an operation.
     * @param cfgOperation The operation configuration.
     */
    public TExportProperty getTExportProperty(CfgDeviceServer deviceServer, CfgOperation cfgOperation) {

        int inputArraySize = 0;
        int outputArraySize = 0;

        CfgOperationDataType inputData = cfgOperation.getInputData();
        CfgOperationDataType outputData = cfgOperation.getOutputData();

        short inputTineFormat = 0;
        if (inputData.isUserDefined()) {
            inputTineFormat = TFormat.CF_STRUCT;
        } else {
            inputTineFormat = (short)inputData.getPropertyType().getTineFormatKey();
        }
        if (inputData.isArray()) {
            inputArraySize = inputData.getArraySize();
        } else {
            inputArraySize = 1;
        }

        short outputTineFormat = 0;
        if (outputData.isUserDefined()) {
            outputTineFormat = TFormat.CF_STRUCT;
        } else {
            outputTineFormat = (short)outputData.getPropertyType().getTineFormatKey();
        }
        if (outputData.isArray()) {
            outputArraySize = outputData.getArraySize();
        } else {
            outputArraySize = 1;
        }

        TExportProperty result = new TExportProperty();

        result.setName(cfgOperation.getName());


        if (inputData.isUserDefined()) {

            CfgTypeDefinition typeDef = deviceServer.getDefinition().getDataTypes()
            .findTypeDefinition(inputData.getUserType().getClassName());

            if ((typeDef != null) && (typeDef.isTagged())) {
                if (typeDef.isTagged()) {
                    result.setInputTag(typeDef.getTag());
                    log.info("Input type is tagged structure : " + typeDef.getClassName() + " -> " + typeDef.getTag());
               } else {
                    log.error("Input structure is untagged: " + inputData.getUserType().getClassName());
                    inputArraySize = 0;
                }
            } else {
                log.error("User-defined input type not found: " + inputData.getUserType().getClassName());
                inputArraySize = 0;
            }

        }

        if (outputData.isUserDefined()) {

            CfgTypeDefinition typeDef = deviceServer.getDefinition().getDataTypes()
            .findTypeDefinition(outputData.getUserType().getClassName());

            if (typeDef != null) {
                if (typeDef.isTagged()) {
                    result.setOutputTag(typeDef.getTag());
                    log.info("Output type is tagged structure : " + typeDef.getClassName() + " -> " + typeDef.getTag());
                } else {
                    log.error("Output structure is untagged: " + outputData.getUserType().getClassName());
                    inputArraySize = 0;
                }
            } else {
                log.error("User-defined output type not found: " + inputData.getUserType().getClassName());
                inputArraySize = 0;
            }
        }

        result.setOutputFormat(outputTineFormat);
        result.setOutputSize(outputArraySize);
        result.setInputFormat(inputTineFormat);
        result.setInputSize(inputArraySize);

        
        log.info("TINE Export Property generated: " 
            + result.getName() 
            + ". Input:" + TFormat.toString(result.getInputFormat()) 
            + "[" + result.getInputSize() + "]" 
            + " - Output: " + TFormat.toString(result.getOutputFormat()) 
            + "[" + result.getOutputSize() + "]");

        result.setAccessMode((short)(TAccess.CA_READ | TAccess.CA_WRITE));
        result.setDescription(getTPropertyDescription(cfgOperation));
        return result;

    }

}
