/*
 * File Channel.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.accesslayer;

import de.desy.tine.dataUtils.TDataType;
import de.desy.tine.definitions.TErrorList;
import de.desy.tine.definitions.TFormat;

/**
 * <code>Channel</code> represents a single connection to the underlying system. It receives notifications according to
 * the specified rate and or mode. The channel can always be asked for the value or status in which case it will return
 * the latest value or status that it received. It also provides means to set the value. In this case it is expected
 * that the set value is acceptable by the underlying system, otherwise an exception will be thrown or the callback will
 * receive an error update. When the channel is no longer used it should be closed, which will disconnect it from the
 * underlying system and free all resources. If the channel has been closed it can no longer be reopened.
 *
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 */
public interface Channel {

    /**
     * @return the callback registered with this channel
     */
    public ChannelCallback getCallback();

    /**
     * @return the full name of this channel
     */
    public String getName();

    /**
     * Returns the output data format. The return value of {@link #getValue()} will be of the array type that is
     * suggested by this output format.
     *
     * @return output data format
     */
    public TFormat getOutputFormat();

    /**
     * Returns the output data size. The return value of {@link #getValue()} will have the array size equals to this
     * output size.
     *
     * @return output data size
     */
    public int getOutputSize();

    /**
     * Returns the refresh rate in milliseconds at which this channel sends updates to the registered callback.
     *
     * @return refresh rate
     */
    public long getRate();

    /**
     * Returns the connection mode of this channel.
     *
     * @return the connection mode
     */
    public ConnectionMode getMode();

    /**
     * Returns the value as received from the underlying system.
     *
     * @return value
     */
    public Object getValue();

    /**
     * Returns the raw data object.
     *
     * @return the raw data object
     */
    public TDataType getRawValue();

    /**
     * Sets the value to the underlying channel.
     *
     * @param value the value to set
     * @throws ConnectionException if the write action failed
     */
    public void setValue(Object value) throws ConnectionException;

    /**
     * Synchronously sets the value to the underlying channel.
     *
     * @param value the value to set
     * @throws ConnectionException if the write action failed
     */
    public void setValueSync(Object value) throws ConnectionException;

    /**
     * Sets the value to the underlying channel.
     *
     * @param value the value to set
     * @param outValue the data type that will receive the readback value after write is completed
     * @param timeout the timeout in milliseconds
     * @throws ConnectionException if the write action failed
     */
    public void setValue(Object value, TDataType outValue, int timeout) throws ConnectionException;

    /**
     * Synchronously sets the value to the underlying channel.
     *
     * @param value the value to set
     * @param outValue the data type that will receive the readback value after write is completed
     * @param timeout the timeout in milliseconds
     * @throws ConnectionException if the write action failed
     */
    public void setValueSync(Object value, TDataType outValue, int timeout) throws ConnectionException;

    /**
     * Returns the tine link status as received from the underlying system. In most cases the status can be converted
     * into a readable string by {@link TErrorList#toString(int)}, but if the code is non-standard a better result can
     * be obtained by calling {@link #getStatusString()}.
     *
     * @return the link status
     */
    public int getStatus();

    /**
     * Returns the status code as a string. In most cases this is a {@link TErrorList#toString(int)}. However, if the
     * code is not standard, this is an arbitrary string provided by the server that this channel is connected to.
     *
     * @return the status as a string
     */
    public String getStatusString();

    /**
     * Close this channel and stop sending updates.
     */
    public void stop();

    /**
     * @return true if this channel has been closed or false if it is still alive
     */
    public boolean isStopped();
}
