/*
 * File ChannelCallback.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.accesslayer;

import de.desy.tine.definitions.TErrorList;

/**
 *
 * <code>ChannelCallback</code> is the callback interface that receives updates when a channel
 * receives a new value or goes into an error state.
 *
 * @see ChannelCallbackAdapter
 *
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 *
 */
public interface ChannelCallback {

    /**
     * An asynchronous write was attempted and failed with the given status.
     *
     * @param address the name of the property to which the values was written
     * @param value the value that was written
     * @param linkStatus the status (can be converted to string by {@link TErrorList#toString(int)}
     */
    public void writeFailed(String address, Object value, int linkStatus);

    /**
     * Asynchronouse write completed successfully.
     *
     * @param address the name of the property to which the values was written
     * @param value the value that was written
     */
    public void writeCompleted(String address, Object value);

    /**
     * The channel value was updated. The current value can be obtained by {@link Channel#getValue()},
     * but the value is not guaranteed to be the value that triggered this update. If another value
     * was received while this callback is being handled the {@link Channel#getValue()} will return
     * the new value.
     *
     * @param channel the channel which value was updated
     * @deprecated use {@link ChannelCallback2#updateValue(Channel, Object)} instead
     */
    @Deprecated
    public void updateValue(Channel channel);

    /**
     * The channel status was updated. The current status can be obtained by {@link Channel#getStatus()}.
     * If another update has been received while this callback is handled the {@link Channel#getStatus()}
     * will return the newest value.
     *
     * @param channel the channel which value was updated
     * @deprecated use {@link ChannelCallback2#updateErrorStatus(Channel, int)} instead
     */
    @Deprecated
    public void updateErrorStatus(Channel channel);

    /**
     * Called when the an error happened in the channels callback call.
     *
     * @param message message describing the situation
     * @param cause the cause of the error
     */
    public void channelError(String message, Throwable cause);
}
