/*
 * File ChannelReadEvent.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.accesslayer;

import java.util.EventObject;

import de.desy.tine.dataUtils.TDataType;

/**
 *
 * <code>ChannelEvent</code> describes a read event on the channel. It contains the reference to the source channel
 * as well as the read value and status.
 *
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 *
 */
public class ChannelReadEvent extends EventObject {

    private static final long serialVersionUID = -56765816451952596L;
    private final TDataType rawValue;
    private final int status;
    private final String statusString;
    private final boolean successful;

    /**
     * Construct a new read event.
     *
     * @param source the source of the event
     * @param rawValue the value
     * @param status the status code
     * @param statusString the human readable representation of the status code
     * @param successful indicates if this event describes a success (true) or an error (false)
     */
    public ChannelReadEvent(Channel source, TDataType rawValue, int status, String statusString, boolean successful) {
        super(source);
        this.rawValue = rawValue;
        this.status = status;
        this.statusString = statusString;
        this.successful = successful;
    }

    /**
     * Returns the channel, which is always the source of this event.
     *
     * @return the source channel
     */
    public Channel getChannel() {
        return (Channel)getSource();
    }

    /**
     * Returns the raw value that was read from the channel.
     *
     * @return the raw value
     */
    public TDataType getRawValue() {
        return rawValue;
    }

    /**
     * Returns the status code of the read action.
     *
     * @return the status code
     */
    public int getStatus() {
        return status;
    }

    /**
     * Returns the status code in a human readable text.
     *
     * @return the status code as a string
     */
    public String getStatusString() {
        return statusString;
    }

    /**
     * Returns true if this event describes an action completed successfully, or false if the action was completed
     * with an error
     *
     * @return true if successful or false if an error
     */
    public boolean isSuccessful() {
        return successful;
    }
}
