/*
 * File MasterChannel.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.accesslayer;

import java.util.Collections;
import java.util.Set;
import java.util.WeakHashMap;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.locks.ReentrantLock;

import de.desy.tine.client.TLink;
import de.desy.tine.client.TLinkCallback;
import de.desy.tine.client.TLinkFactory;
import de.desy.tine.dataUtils.TDataType;
import de.desy.tine.definitions.TAccess;
import de.desy.tine.definitions.TErrorList;
import de.desy.tine.definitions.TFormat;
import de.desy.tine.definitions.TMode;
import de.desy.tine.exceptions.InvalidDataReferenceException;
import de.desy.tine.exceptions.TineRuntimeErrorException;
import de.desy.tine.exceptions.UnresolvedAddressException;
import de.desy.tine.queryUtils.TPropertyQuery;
import de.desy.tine.queryUtils.TQuery;
import de.desy.tine.server.properties.TExportProperty;
import de.desy.tine.server.properties.TMetaProperties;
import de.desy.tine.server.properties.TStockProperties;

/**
 *
 * <code>MasterChannel</code> is the master channel that contains the link to the underlying system.
 * This channel takes care of all communication to the same device/property with the same data type
 * and size. It always maintains the link at the higher requested rate or on change, depending on
 * what Channels were requested. However, when a Channel is closed it does not reduce the data rate.
 *
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 *
 */
class MasterChannel implements TLinkCallback {

    private static class Callback implements TLinkCallback {

        private final ChannelCallback callback;
        private final String address;
        private final Object value;

        Callback(ChannelCallback callback, String address, Object value) {
            this.callback = callback;
            this.address = address;
            this.value = value;
        }
        @Override
        public void callback(TLink link) {
            if (link.getLinkStatus() == 0) {
                callback.writeCompleted(address,value);
            } else {
                callback.writeFailed(address,value,link.getLinkStatus());
            }
            link.close();
        }
    }

    private TLink link;
    private final String address;
    private final TFormat dataOutFormat;
    private final int dataOutSize;
    private final Object dataIn;
    private TFormat actualDataOutFormat;
    private int actualDataOutSize;

    private final Set<TINEChannel> channels = Collections.newSetFromMap(new WeakHashMap<TINEChannel,Boolean>());
//    private final List<TINEChannel> channels = new ArrayList<>();
    private AtomicInteger counter = new AtomicInteger(0);

    private int currentRate;
    private ConnectionMode currentMode;
    private volatile boolean isUpdateReceived = false;
    private volatile boolean isStatic = false;
    private volatile boolean blacklisted = false;
    private volatile boolean isError = false;

    private boolean createNewLink = false;
    private short tModeForNewLink;
    private int periodForNewLink;
    private final long keepAliveTime;

    private Object shutDownMutex = new Object();
    private Object writeMutex = new Object();
    private ScheduledFuture<?> shutDownFuture;
    private ReentrantLock lock = new ReentrantLock();

    /**
     * Constructs a new master channel. If the dataOut is null or dataOutSize is equal or less than 0, the
     * default values for the channel will be used.
     *
     * @param address the property to connect to
     * @param dataOut output data format
     * @param dataOutSize output data size
     * @param dataIn input data object
     * @param keepAliveTime the time in milliseconds how long the channel is kept alive when no longer needed
     */
    MasterChannel(String address, TFormat dataOut, int dataOutSize, Object dataIn, long keepAliveTime) {
        this.address = address;
        this.dataOutFormat = dataOut;
        this.dataOutSize = dataOutSize;
        this.dataIn = dataIn;
        this.keepAliveTime = keepAliveTime;
    }

    /**
     * @return the input request data
     */
    Object getDataIn() {
        return dataIn;
    }

    /**
     * Increments the usage counter of this channel and returns the current value.
     *
     * @see #getCounter()
     *
     * @return the current counter value
     */
    int increment() {
        synchronized(shutDownMutex) {
            if (shutDownFuture != null) {
                shutDownFuture.cancel(true);
                shutDownFuture = null;
            }
        }
        return counter.incrementAndGet();
    }

    /**
     * Decrements the usage counter of this channel and returns the current value.
     *
     * @see #getCounter()
     *
     * @return the current counter value
     */
    int decrement() {
        return counter.decrementAndGet();
    }

    /**
     * Returns the current usage counter. This is the number of times, this channel is presently used.
     * Most of the time this is equal to the number of {@link Channel}s this master holds.
     *
     * @return the counter
     */
    int getCounter() {
        return counter.get();
    }

    /**
     * Constructs a new Channel that sattisfies the given parameters.
     *
     * @param mode the mode with which that the given channel will send notifications
     * @param rate the rate at which the channel will send notifications
     * @param outputObject the object into which the output data will be written to
     * @param callback the callback to receive notifications
     * @return the channel
     * @throws ConnectionException
     */
    TINEChannel with(ConnectionMode mode, int rate, TDataType outputObject, ChannelCallback callback) throws ConnectionException {
        int localStatus = -1;
        TDataType dout = null;
        lock.lock();
        try {
            boolean trace = false;
            if (rate < 10) rate = 10; //fastest rate is 100 Hz
            if (!isStatic && !blacklisted) {
                boolean makeNewLink = false;
                if (currentMode == null) {
                    //first connection
                    makeNewLink = true;
                } else if (currentMode == ConnectionMode.SINGLE) {
                    if (currentRate > rate && (mode == ConnectionMode.POLL || mode == ConnectionMode.CHANGE)) {
                        //single get was made but at a rate slower than the new request
                        makeNewLink = true;
                    }
                    currentMode = mode;
                } else if (currentMode == ConnectionMode.POLL) {
                    if (currentRate > rate && (mode == ConnectionMode.POLL || mode == ConnectionMode.CHANGE)) {
                        //polling monitor was created but at a rate slower than the new request
                        makeNewLink = true;
                    }
                } else if (currentMode == ConnectionMode.CHANGE) {
                    if (currentRate > rate && (mode == ConnectionMode.POLL || mode == ConnectionMode.CHANGE)) {
                        //data change monitor was made but at a rate slower than the new request
                        makeNewLink = true;
                    }
                    if (mode == ConnectionMode.POLL) {
                        currentMode = ConnectionMode.POLL;
                    }
                }
                boolean reattachLink = false;
                if (makeNewLink) {
                    boolean delayCreateLink = false;
                    if (this.link == null) {
                        actualDataOutFormat = dataOutFormat;
                        actualDataOutSize = dataOutSize;
                        if (dataOutFormat == null) {
                            actualDataOutFormat = TFormat.valueOf(TFormat.CF_DEFAULT);
                        }
                        if (dataOutSize <= 0) {
                            actualDataOutSize = ChannelFactory.DEFAULT_SIZE;
                        }

                        dout = new TDataType(actualDataOutSize,actualDataOutFormat.getValue());
                        TDataType din;
                        if (dataIn == null) {
                            din = null;
                        } else if (dataIn instanceof TDataType) {
                            din = (TDataType)dataIn;
                        } else {
                            din = TDataType.toTDataType(dataIn);
                        }
                        link = new TLink(address,dout,din,TAccess.CA_READ);
                        localStatus = link.getLinkStatus();
                        if (localStatus == TErrorList.resources_exhausted
                                || localStatus == TErrorList.link_blacklisted
                                || localStatus == TErrorList.argument_list_error) {
                            isUpdateReceived = true;
                        }
                    } else {
                        delayCreateLink = !isUpdateReceived;
                        if (!delayCreateLink) {
                          if (mode == ConnectionMode.SINGLE)
                            link.close();
                          else
                            reattachLink = true;
                        }
                    }

                    trace = link.isTraceLink();
                    short tmode;
                    int rateToUse = rate;
                    if (mode == ConnectionMode.CHANGE) {
                        tmode = TMode.CM_POLL;
                        rateToUse = rate;
                    } else if (mode == ConnectionMode.SINGLE) {
                        tmode = TMode.CM_POLL | TMode.CM_WAIT;
                        rateToUse = 1000;
                    } else {
                        tmode = TMode.CM_POLL;
                        rateToUse = rate;
                    }
                    if (trace) link.traceLink("MasterChannel.with", "make new link (delay: "+delayCreateLink+", mode: "+mode+", rate: "+rateToUse+")");
                    int cc = 0;

                    boolean syncCall = false;
                    if (mode == ConnectionMode.SINGLE || mode == ConnectionMode.SYNC_SINGLE) {
                        String property = address.substring(address.lastIndexOf('/')+1);
                        if (TQuery.isStockProperty(property) || TMetaProperties.isMetaProperty(property) ||
                            TQuery.isDoocsStatic(this.address, property) || mode == ConnectionMode.SYNC_SINGLE) {
                            syncCall = true;
                            cc = link.executeAndClose(rateToUse);
                            isUpdateReceived = true;
                            if (cc == 0) {
                                if (actualDataOutFormat.getValue() == TFormat.CF_DEFAULT) {
                                    actualDataOutFormat = TFormat.valueOf(link.getOutputDataObject().getFormat());
                                }
                                if (actualDataOutSize == ChannelFactory.DEFAULT_SIZE) {
                                    actualDataOutSize = link.getOutputDataObject().getArrayLength();
                                }
                            } else if (cc == TErrorList.buffer_too_small || cc == TErrorList.illegal_data_size ||
                                       cc == TErrorList.illegal_format) {
                            	  boolean reacq = false;
                                if (dataOutFormat == null || dataOutFormat.getValue() == TFormat.CF_DEFAULT
                                        || dataOutSize <= 0 || dataOutSize == ChannelFactory.DEFAULT_SIZE) {
                                    if (TMetaProperties.isMetaProperty(property))
                                    {
	                                    actualDataOutFormat = TFormat.valueOf(TMetaProperties.getSuggestedFormat(property));
	                                    actualDataOutSize = TMetaProperties.getSuggestedDataSize(property);
	                                    reacq = true;
                                    }
                                    else if (TQuery.isStockProperty(property))
                                    {
                                    	TExportProperty xp = TStockProperties.getPropertyList().getFirstProperty(property);
                                    	if (xp != null)
                                    	{
                                    	  actualDataOutFormat = TFormat.valueOf(xp.getOutputFormat());
                                    	  actualDataOutSize = xp.getOutputSize();
  	                                    reacq = true;
                                    	}
                                    }
                                    else
                                    {
                                    	TPropertyQuery[] pq = TQuery.getPropertyInformation(link.cntName, link.expName, link.devName, property, 500);
                                    	if (pq != null && pq.length > 0)
                                    	{
                                    	  actualDataOutFormat = TFormat.valueOf(pq[0].prpFormat);
                                    	  actualDataOutSize = pq[0].prpSize;                                    		
  	                                    reacq = true;
                                    	}
                                    }
                                    if (reacq)
                                    {
	                                    link.setOutputDataObject(new TDataType(actualDataOutSize, actualDataOutFormat.getValue()));
	                                    cc = link.executeAndClose(rateToUse);
                                    }
                                }
                            }
                            if (cc > 0) {
                                cc = -cc;
                            }
                        }
                    }

                    if (!syncCall) {
                        if (delayCreateLink) {
                            createNewLink = true;
                            periodForNewLink = rateToUse;
                            tModeForNewLink = tmode;
                        } else {
                            createNewLink = false;
                            if (reattachLink) {
                              link.modifyAttributes(tmode, rateToUse);
                              cc = 0;
                            }
                            else {
                              cc = link.attach(tmode,this,rateToUse);
                            }
                            if (cc < 0) {
                                cc = -link.getLinkStatus();
                            }
                        }
                    }
                    currentMode = mode;
                    currentRate = rateToUse;
                    if (cc < 0) {
                        throw new ConnectionException("Could not create a link with mode " + mode + " at rate " + rateToUse + " [" + TErrorList.getErrorCode(-cc) +"]",-cc);
                    }
                }
            }
        } catch (UnresolvedAddressException e) {
            isUpdateReceived = true;
            localStatus = TErrorList.address_unknown;
        } catch (InvalidDataReferenceException e) {
            isUpdateReceived = true;
            localStatus = TErrorList.invalid_reference;
        } finally {
            lock.unlock();
        }

        TINEChannel channel = new TINEChannel(this,mode,rate,outputObject,callback);
        synchronized(channels) {
            channels.add(channel);
            if (isUpdateReceived) {
                if (link == null) {
                    int status = localStatus;
                    String statusString = TErrorList.getErrorString(status);
                    if (TErrorList.hasData(status)) {
                        channel.updateValue(dout,status,statusString);
                    } else {
                        channel.updateError(dout,status,statusString);
                    }
                } else {
                    int status = link.getLinkStatus();
                    String statusString = link.getLastError();
                    if (TErrorList.hasData(status)) {
                        channel.updateValue(link.getOutputDataObject(),status,statusString);
                    } else {
                        channel.updateError(link.getOutputDataObject(),status,statusString);
                    }
                }

            }
        }
        return channel;
    }

    /**
     * Asynchronously set the value.
     *
     * @param value the value to set
     * @param outValue the data type to write the output value to
     * @param callback the callback to notify upon completion.
     * @param timeout the timeout to use in milliseconds
     * @throws ConnectionException if there was an error establishing connection
     */
    void setValue(final Object value, final TDataType outValue, final int timeout, final ChannelCallback callback) throws ConnectionException {
        //do not allow two write attempts at the same time
        try {
            synchronized(writeMutex) {
                final TLink linkToUse;

                TDataType din;
                if (value instanceof TDataType) {
                    din = (TDataType) value;
                } else {
                    din = TDataType.toTDataType(value);
                    if (din == null) {
                        throw new IllegalArgumentException("The value " + value + " cannot be sent over TINE.");
                    }
                }
                // p.d. 25.9.2015: the 'din' and 'dout' objects need to be established
                //      prior to instantiating a new TLink !
    //            if (outValue != null) {
    //                if (writeReadLink == null) {
    //                    writeReadLink = new TLink(address,outValue,din,(short)(TAccess.CA_WRITE|TAccess.CA_READ));
    //                } else {
    //                    writeReadLink.close();
    //                }
    //                linkToUse = writeReadLink;
    //            } else {
    //                if (writeLink == null) {
    //                    writeLink = new TLink(address,null,din,TAccess.CA_WRITE);
    //                } else {
    //                    writeLink.close();
    //                }
    //                linkToUse = writeLink;
    //            }
                linkToUse = new TLink(address,outValue,din,(short)(TAccess.CA_WRITE|TAccess.CA_READ));
                int cc = linkToUse.getLinkStatus();
                if (cc == TErrorList.resources_exhausted) {
                    cc = -cc;
                } else {
                    cc = linkToUse.attach(TMode.CM_SINGLE,new Callback(callback,address,value),timeout);
                    if (cc < 0) {
                        cc = -linkToUse.getLinkStatus();
                    }
                }
                if (cc < 0) {
                    linkToUse.close();
                    throw new ConnectionException("Cannot write value to " + getName(),-cc);
                }
            }
        } catch (UnresolvedAddressException e) {
            throw new ConnectionException("Cannot write value to " + getName(),TErrorList.address_unknown);
        } catch (InvalidDataReferenceException e) {
            throw new ConnectionException("Cannot write value to " + getName(),TErrorList.invalid_reference);
        } catch (TineRuntimeErrorException e) {
            throw new ConnectionException("Cannot write value to " + getName(),TErrorList.link_exists);
        }
    }

    /**
     * Synchronously set the value.
     *
     * @param value the value to set
     * @param outValue the data type to write the output value to
     * @param timeout the timeout in milliseconds
     * @throws ConnectionException if there was an error establishing the connection
     */
    void setValue(Object value, TDataType outValue, int timeout) throws ConnectionException {
        //do not allow two write attempts at the same time
        try {
            synchronized(writeMutex) {
                final TLink linkToUse;

                TDataType din;
                if (value instanceof TDataType) {
                    din = (TDataType)value;
                } else {
                    din = TDataType.toTDataType(value);
                    if (din == null) {
                        throw new IllegalArgumentException("The value " + value + " cannot be sent over TINE.");
                    }
                }
            // p.d. 25.9.2015: the 'din' and 'dout' objects need to be established
            //      prior to instantiating a new TLink !
//            if (outValue != null) {
//                if (writeReadLink == null) {
//                    writeReadLink = new TLink(address,outValue,din,(short)(TAccess.CA_WRITE|TAccess.CA_READ));
//                } else {
//                    writeReadLink.close();
//                }
//                linkToUse = writeReadLink;
//            } else {
//                if (writeLink == null) {
//                     writeLink = new TLink(address,null,din,TAccess.CA_WRITE);
//                } else {
//                   writeLink.close();
//                }
//                linkToUse = writeLink;
//            }
            // p.d. just apply CA_READ|CA_WRITE no matter what and do a 1-shot execute:
                linkToUse = new TLink(address,outValue,din,(short)(TAccess.CA_WRITE|TAccess.CA_READ));
                int cc = linkToUse.getLinkStatus();
                if (cc != TErrorList.resources_exhausted) {
                    cc = linkToUse.executeAndClose(timeout);
                }
                if (cc != 0) {
                    throw new ConnectionException("Cannot write value to " + getName(),cc);
                }
            }
        } catch (UnresolvedAddressException e) {
            throw new ConnectionException("Cannot write value to " + getName(),TErrorList.address_unknown);
        } catch (InvalidDataReferenceException e) {
            throw new ConnectionException("Cannot write value to " + getName(),TErrorList.invalid_reference);
        } catch (TineRuntimeErrorException e) {
            throw new ConnectionException("Cannot write value to " + getName(),TErrorList.link_exists);
        }
    }

    /**
     * @return the full name of the channel
     */
    String getName() {
        return address;
    }

    /**
     * @return the requested data out format
     */
    TFormat getFormat() {
        return dataOutFormat;
    }

    /**
     * @return the requested data out size
     */
    int getSize() {
        return dataOutSize;
    }

    /**
     * @return the actual data out format
     */
    TFormat getActualFormat() {
        return actualDataOutFormat;
    }

    /**
     * @return the actual data out size
     */
    int getActualSize() {
        return actualDataOutSize;
    }

    /*
     * (non-Javadoc)
     * @see de.desy.tine.client.TLinkCallback#callback(de.desy.tine.client.TLink)
     */
    @Override
    public void callback(final TLink link) {
        if (lock.isHeldByCurrentThread()) {
            //the callback was fired during link creation
            ChannelFactory.getInstance().getCallbackExecutor().execute(new Runnable(){
                @Override
                public void run() {
                    processCallback(link);
                }
            });
        } else {
            processCallback(link);
        }
    }

    private void processCallback(TLink link) {
        boolean trace = TLinkFactory.getTraceLinkKey() != null && link.isTraceLink();
        int status = link.getLinkStatus();
        String statusString = link.getLastError();
        status = TErrorList.getErrorCode(status); // remove any information bits
        TDataType dout = link.getOutputDataObject();
        ConnectionException ex = null;
        if (trace) link.traceLink("MasterChannel.callback", "entering callback routine");
        buffer: {
            lock.lock();
            try {
                boolean doNotContinue = false;
                if (status == TErrorList.buffer_too_small || status == TErrorList.illegal_data_size || status == TErrorList.illegal_format
                    || status == TErrorList.illegal_property) {
                    if (dataOutFormat == null || dataOutFormat.getValue() == TFormat.CF_DEFAULT || dataOutSize <= 0
                        || dataOutSize == ChannelFactory.DEFAULT_SIZE) {
                        doNotContinue = true;
                        try {
                            if (trace) link.traceLink("MasterChannel.callback","acquire property information");
                            TPropertyQuery tpq = ChannelFactory.getInstance().getPropertyQuery(address);
                            byte fmt = link.isWildcardLink() ? (byte)TFormat.CF_USTRING : tpq.prpFormat;
                            actualDataOutFormat = TFormat.valueOf(fmt);
                            actualDataOutSize = tpq.prpSize;
                            link.close();
                            if (status == TErrorList.illegal_property) TLinkFactory.getInstance().flushBlackList();
                            link.setOutputDataObject(new TDataType(actualDataOutSize,actualDataOutFormat.getValue()));
                            createNewLink = true;
                            periodForNewLink = currentRate;
                            tModeForNewLink = currentMode == ConnectionMode.CHANGE ? TMode.CM_DATACHANGE : TMode.CM_POLL;
                        } catch (ConnectionException e) {
                            ex = e;
                        }
                    }
                }
                if (createNewLink) {
                    if (trace) link.traceLink("MasterChannel.callback","creating new link");
                    createNewLink = false;
                    link.close();
                    int cc = link.attach(tModeForNewLink,this,periodForNewLink);
                    if (cc < 0) {
                        cc = link.getLinkStatus();
                        if (trace) link.traceLink("MasterChannel.callback","could not create new link ! <" + (cc) + ">");
                        ex = new ConnectionException("Could not create a link with mode " + currentMode + " at rate " + currentRate + " ["
                            + TErrorList.getErrorCode(cc) + "]",cc);
                        lock.unlock();
                        break buffer;
                    }
                    if (doNotContinue) {
                        if (trace) link.traceLink("MasterChannel.callback","do not continue");
                        return;
                    }
                }
                if (status == TErrorList.information_static) {
                    isStatic = true;
                    link.close();
                    if (trace) link.traceLink("MasterChannel.callback","information static: close link");
                    status = 0;
                } else if (TQuery.isDoocsStatic(link.cntName,link.expName,link.devProperty)) {
                    isStatic = true;
                    link.close();
                    if (trace) link.traceLink("MasterChannel.callback","isDoocsStatic = true: close link");
                } else if (TErrorList.isBlackListOffense(status)) {
                    blacklisted = true;
                    link.close();
                } else if (TLinkFactory.getInstance().isLinkBlackListed(link)) {
                    blacklisted = true;
                    link.close();
                }
                if (actualDataOutFormat.getValue() == TFormat.CF_DEFAULT) {
                    actualDataOutFormat = TFormat.valueOf(link.getOutputDataObject().getFormat());
                    if (trace) link.traceLink("MasterChannel.callback","apply data format " + TFormat.toString(actualDataOutFormat.getValue()));
                }
                if (actualDataOutSize == ChannelFactory.DEFAULT_SIZE) {
                    actualDataOutSize = link.getOutputDataObject().getArrayLength();
                    if (trace) link.traceLink("MasterChannel.callback","apply data size " + actualDataOutSize);
                }
                isUpdateReceived = true;
            } finally {
                lock.unlock();
            }
        }

        TINEChannel[] chs = null;
        synchronized(channels) {
            if (channels.size() == 0) return;
            chs = channels.toArray(new TINEChannel[channels.size()]);
        }

        if (ex != null) {
            isError = true;
            if (trace) link.traceLink("MasterChannel.callback", "exception "+ex.getLocalizedMessage());
            for (TINEChannel c : chs) {
                c.updateFatalError(ex);
            }
        } else if (TErrorList.hasData(status)) {
            isError = false;
            if (trace) link.traceLink("MasterChannel.callback", "update data values");
            for (TINEChannel c : chs) {
                c.updateValue(dout,status,statusString);
            }
        } else {
            isError = true;
            if (trace) link.traceLink("MasterChannel.callback", "update status");
            for (TINEChannel c : chs) {
                c.updateError(dout,status,statusString);
            }
        }
    }

    /**
     * Removes the channel from this master. The channel is expected to be closed.
     *
     * @param channel the channel to remove
     */
    void remove(TINEChannel channel) {
        synchronized(channels) {
            channels.remove(channel);
        }
        if (decrement() == 0) {
            synchronized(shutDownMutex) {
                //if the link is in error state, close it immediately, ok we give it 100 ms grace period
                long delay = isError ? 100 : keepAliveTime;
                shutDownFuture = ChannelFactory.getInstance().getSchedulerService().schedule(new Runnable() {
                    @Override
                    public void run() {
                        lock.lock();
                        try {
                            if (getCounter() == 0) {
                                ChannelFactory.getInstance().releaseChannel(MasterChannel.this);
                            }
                        } finally {
                            lock.unlock();
                        }
                    }
                },delay,TimeUnit.MILLISECONDS);
            }
        }
    }

    /**
     * Destroy this master channel.
     */
    void destroy() {
        lock.lock();
        try {
            if (link != null) {
                link.close();
            }
            synchronized(channels) {
                channels.clear();
            }
        } finally {
            lock.unlock();
        }
    }

    /*
     * (non-Javadoc)
     * @see java.lang.Object#finalize()
     */
    @Override
    protected void finalize() throws Throwable {
        destroy();
        super.finalize();
    }
}
