/*
 * File TINEChannel.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.accesslayer;

import java.lang.ref.WeakReference;
import java.util.Objects;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;

import de.desy.tine.dataUtils.TDataType;
import de.desy.tine.definitions.TErrorList;
import de.desy.tine.definitions.TFormat;
import de.desy.tine.server.logger.MsgLog;

/**
 * <code>TINEChannel</code> is a channel implementation
 *
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 */
class TINEChannel implements Channel {

    static final class Update implements Runnable {

        private final int status;
        private final String statusString;
        private final TDataType rawValue;
        private final WeakReference<TINEChannel> weakChannel;
        //used in case of periodic updates, when the value always has to come from the channel
        private final boolean valueFromChannel;
        private int lastStatusSent;
        private Object lastValueSent;

        Update(TINEChannel channel) {
            this(null,-1,null,channel,true);
        }

        Update(TDataType rawValue, int status, String statusString, TINEChannel channel) {
            this(rawValue,status,statusString,channel,false);
        }

        TINEChannel getChannel() {
            return weakChannel.get();
        }

        private Update(TDataType rawValue, int status, String statusString, TINEChannel channel, boolean valueFromChannel) {
            this.status = status;
            this.statusString = statusString;
            this.rawValue = rawValue;
            this.weakChannel = new WeakReference<>(channel);
            this.valueFromChannel = valueFromChannel;
        }

        @SuppressWarnings("deprecation")
        public void run() {
            TINEChannel channel = weakChannel.get();
            if (channel == null) return;
            int st = -1;
            String ststr;
            TDataType val = null;
            synchronized (channel.mutex) {
                if (valueFromChannel) {
                    st = channel.statusSafe;
                    val = channel.doutSafe;
                    ststr = channel.statusStringSafe;
                } else {
                    st = status;
                    val = rawValue;
                    ststr = this.statusString;
                }
            }
            if (channel.getMode() == ConnectionMode.CHANGE) {
                if (lastStatusSent == st && Objects.deepEquals(lastValueSent,val.getDataObject())) {
                    //do not fire if nothing changed
                    return;
                }
            }
            ChannelCallback callback = channel.callback;
            if (TErrorList.hasData(st)) {
                if (callback instanceof ChannelListener) {
                    ((ChannelListener)callback).updateValue(new ChannelReadEvent(channel,val,st,ststr,true));
                } else if (callback instanceof ChannelCallback2) {
                    ((ChannelCallback2)callback).updateValue(channel,val);
                }
                callback.updateValue(channel);
            } else {
                if (callback instanceof ChannelListener) {
                    ((ChannelListener)callback).updateError(new ChannelReadEvent(channel,val,st,ststr,false));
                } else if (callback instanceof ChannelCallback2) {
                    ((ChannelCallback2)callback).updateErrorStatus(channel,val,st);
                }
                callback.updateErrorStatus(channel);
            }
            lastValueSent = val.getDataObject();
            lastStatusSent = st;
        }
    }

    private final ScheduledExecutorService executor;
    private final MasterChannel channel;
    private final TDataType outputObject;
    private volatile ScheduledFuture<?> future;
    private final ChannelCallback callback;
    private final long rate;
    private final ConnectionMode mode;
    private boolean isStopped = false;
    private volatile TDataType dout;
    private volatile TDataType doutSafe;
    private volatile int linkStatus = TErrorList.status_pending;
    private volatile int statusSafe = TErrorList.status_pending;
    private volatile String statusString;
    private volatile String statusStringSafe;
    private final Object mutex = new Object();

    /**
     * Constructs a new Channel, which receives updates from the master channel and sends them to the given callback at
     * the specified rate. If the connection mode is CHANGE or SINGLE, the rate is irrelevant.
     *
     * @param channel the master channel that supplies the date
     * @param mode connection mode of this channel
     * @param rate the callback notification rate in milliseconds
     * @param outputObject the object which will receive the output data upon completion
     * @param callback the callback to receive notifications
     */
    TINEChannel(MasterChannel channel, ConnectionMode mode, long rate, TDataType outputObject, ChannelCallback callback) {
        this.channel = channel;
        this.callback = callback;
        this.rate = rate;
        this.mode = mode;
        this.executor = ChannelFactory.getInstance().getSchedulerService();
        this.outputObject = outputObject;
    }

    /*
     * (non-Javadoc)
     * @see de.desy.tine.accesslayer.Channel#setValue(java.lang.Object)
     */
    @Override
    public void setValue(Object value) throws ConnectionException {
        channel.setValue(value,null,ChannelFactory.WAIT_TIME,callback);
    }

    /*
     * (non-Javadoc)
     * @see de.desy.tine.accesslayer.Channel#setValueSync(java.lang.Object)
     */
    @Override
    public void setValueSync(Object value) throws ConnectionException {
        channel.setValue(value,null,ChannelFactory.WAIT_TIME);
    }

    /*
     * (non-Javadoc)
     * @see de.desy.tine.accesslayer.Channel#setValue(java.lang.Object, de.desy.tine.dataUtils.TDataType, int)
     */
    @Override
    public void setValue(Object value, TDataType outValue, int timeout) throws ConnectionException {
        channel.setValue(value,outValue,timeout,callback);
    }

    /*
     * (non-Javadoc)
     * @see de.desy.tine.accesslayer.Channel#setValueSync(java.lang.Object, de.desy.tine.dataUtils.TDataType, int)
     */
    @Override
    public void setValueSync(Object value, TDataType outValue, int timeout) throws ConnectionException {
        channel.setValue(value,outValue,timeout);
    }

    /*
     * (non-Javadoc)
     * @see de.desy.tine.accesslayer.Channel#getName()
     */
    @Override
    public String getName() {
        return channel.getName();
    }

    /*
     * (non-Javadoc)
     * @see de.desy.tine.accesslayer.Channel#getOutputFormat()
     */
    @Override
    public TFormat getOutputFormat() {
        return channel.getActualFormat();
    }

    /*
     * (non-Javadoc)
     * @see de.desy.tine.accesslayer.Channel#getOutputSize()
     */
    @Override
    public int getOutputSize() {
        return channel.getActualSize();
    }

    /*
     * (non-Javadoc)
     * @see de.desy.tine.accesslayer.Channel#getRate()
     */
    @Override
    public long getRate() {
        return rate;
    }

    /*
     * (non-Javadoc)
     * @see de.desy.tine.accesslayer.Channel#getMode()
     */
    @Override
    public ConnectionMode getMode() {
        return mode;
    }

    /*
     * (non-Javadoc)
     * @see de.desy.tine.accesslayer.Channel#getValue()
     */
    @Override
    public Object getValue() {
        if (dout == null) return null;
        Object obj = dout.getDataObject();
        if (obj == null) {
            MsgLog.log("TINEChannel.getValue","null output object",TErrorList.invalid_data,null,0);
            return null;
        }
        if (outputObject == null) {
            if (TFormat.isString(getOutputFormat().getValue())) {
                return dout.toString();
            } else if (getOutputSize() == 1) {
                if (obj instanceof int[]) {
                    return ((int[])obj)[0];
                } else if (obj instanceof long[]) {
                    return ((long[])obj)[0];
                } else if (obj instanceof double[]) {
                    return ((double[])obj)[0];
                } else if (obj instanceof float[]) {
                    return ((float[])obj)[0];
                } else if (obj instanceof short[]) {
                    return ((short[])obj)[0];
                } else if (obj instanceof byte[]) {
                    return ((byte[])obj)[0];
                } else if (obj instanceof char[]) {
                    return ((char[])obj)[0];
                }
            }
        }
        return obj;
    }

    /*
     * (non-Javadoc)
     * @see de.desy.tine.accesslayer.Channel#getStatus()
     */
    @Override
    public int getStatus() {
        return linkStatus;
    }

    /*
     * (non-Javadoc)
     * @see de.desy.tine.accesslayer.Channel#getStatusString()
     */
    @Override
    public String getStatusString() {
        return statusString;
    }

    private static TDataType cloneDataObject(TDataType owner) {
        TDataType d = new TDataType(ChannelFactory.DEFAULT_SIZE,TFormat.CF_DEFAULT);
        if (d.dataCopy(owner) == 0) {
            return owner;
        } else {
            return d;
        }
    }

    /*
     * (non-Javadoc)
     * @see de.desy.tine.accesslayer.Channel#stop()
     */
    @Override
    public synchronized void stop() {
        if (isStopped) {
            return;
        }
        if (future != null) {
            future.cancel(true);
        }
        isStopped = true;
        channel.remove(this);
    }

    /*
     * (non-Javadoc)
     * @see de.desy.tine.accesslayer.Channel#isStopped()
     */
    @Override
    public boolean isStopped() {
        return isStopped;
    }

    /*
     * (non-Javadoc)
     * @see de.desy.tine.accesslayer.Channel#getCallback()
     */
    @Override
    public ChannelCallback getCallback() {
        return callback;
    }

    /*
     * (non-Javadoc)
     * @see de.desy.tine.accesslayer.Channel#getRawValue()
     */
    @Override
    public TDataType getRawValue() {
        return dout;
    }

    /**
     * Called by the {@link MasterChannel} when a new value is received.
     *
     * @param dout output data object
     * @param status the current status code of the link
     * @param statusString plain text description of the current status code
     */
    void updateValue(TDataType dout, int status, String statusString) {
        if (isStopped) return;
        this.dout = dout;
        this.linkStatus = status;
        this.statusString = statusString;
        if (outputObject != null) {
            outputObject.dataCopy(this.dout);
        }
        TDataType temp = cloneDataObject(dout);
        synchronized (mutex) {
            this.doutSafe = temp;
            this.statusSafe = linkStatus;
            this.statusStringSafe = statusString;
        }
        if (mode == ConnectionMode.SINGLE || mode == ConnectionMode.SYNC_SINGLE) {
            executor.execute(new Update(doutSafe,statusSafe,statusStringSafe,this));
            stop();
        } else if (future == null) {
            future = executor.scheduleAtFixedRate(new Update(this),0,rate,TimeUnit.MILLISECONDS);
        } else if (future.isDone() || future.isCancelled()) {
            future = executor.scheduleAtFixedRate(new Update(this),0,rate,TimeUnit.MILLISECONDS);
        }
    }

    /**
     * Called by the {@link MasterChannel} when a non 0 link status is received.
     *
     * @param dout output data object
     * @param status the current status code of the link
     * @param statusString plain text description of the current status code
     */
    void updateError(TDataType dout, int linkStatus, String statusString) {
        updateValue(dout,linkStatus,statusString);
    }

    /**
     * Sends an error to the callback. This can happen when the link needed to reconnect due to buffer too small or
     * invalid data size.
     *
     * @param e the exception describing the original error
     */
    void updateFatalError(final ConnectionException e) {
        executor.execute(new Runnable() {

            public void run() {
                callback.channelError("Error connecting to channel: " + getName(),e);
            }
        });
    }

    /*
     * (non-Javadoc)
     * @see java.lang.Object#finalize()
     */
    @Override
    protected void finalize() throws Throwable {
        stop();
    }
}
