/*
 * File FECAddr.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.addrUtils;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.net.InetAddress;

import de.desy.tine.definitions.TErrorList;
import de.desy.tine.definitions.TStrings;
import de.desy.tine.endianUtils.Swap;
import de.desy.tine.server.logger.MsgLog;

public final class FECAddr
{
  public static final String LOOPBACKADDR4 = "127.0.0.1";
  public static final String LOOPBACKADDR6 = "::1";
  public static final int ADDRFLAG_MCAST_LEGACY = 0x00;
  public static final int ADDRFLAG_MCAST_MODERN = 0x01;
  public static final String IPV6_IPV4PREFIX = "::ffff:";
  protected String fecName = new String(new StringBuffer(TStrings.FEC_NAME_SIZE));
  protected byte[] netAdr = new byte[16];
  protected byte[] lclAdr = new byte[8];
  protected String strAdr = new String(new StringBuffer(40));
  protected int portOffset;
  protected int inetProtocol;
  protected int tineProtocol;
  protected int addrFlags;             /**< address flags: e.g. mcast rules */
  public static int sizeInBytes = TStrings.FEC_NAME_SIZE + 16 + 8 + 40 + 4*4;
  ByteArrayOutputStream dBuffer = new ByteArrayOutputStream(sizeInBytes);
  byte[] hByteArray;
  private static boolean isIPv6Site = false;
  public static boolean isIPv6Site() { return isIPv6Site; }
  public static String getLoopBackAddr()
  {
    return isIPv6Site ? LOOPBACKADDR6 : LOOPBACKADDR4;
  }
  public FECAddr4 toFECAddrIPv4()
  {
    FECAddr4 fa4 = new FECAddr4();
    fa4.setFecName(fecName);
    String adr = (isIPv4on6()) ? strAdr.substring(IPV6_IPV4PREFIX.length()) : strAdr;
    fa4.setIpAddr(adr);
    fa4.setInetProtocol(inetProtocol);
    fa4.setTineProtocol(tineProtocol);
    System.arraycopy(netAdr, 0, fa4.ipHAddr, 0, 4);
    return fa4;
  }
  public static void setIPv6Site(boolean value) { isIPv6Site = value; }
  public boolean isIPv6() { return strAdr.contains(":"); }
  public boolean isIPv4() { return !isIPv6(); }
  public boolean isIPv4on6() 
  {
    if (!strAdr.startsWith(IPV6_IPV4PREFIX)) return false;
    String adr = strAdr.substring(IPV6_IPV4PREFIX.length());
    int i = -1, n = 0;
    while ((i=adr.indexOf(".")) > 0)
    {
      n++;
      adr = adr.substring(++i);
    }
    if (n != 3) return false;
    return true;
  }
  public boolean hasAdr() 
  {
    for (int i=0; i<16; i++) if (netAdr[i] != 0) return true;
    return false;
  }
  public int getPortOffset()
  {
    return portOffset;
  }
  public int getAddressFlags()
  {
    return addrFlags;
  }
  public void setAddressFlags(int flags)
  {
    addrFlags = flags;
  }
  public byte[] toByteArray()
  {
    try
    {
      dBuffer.reset();
      DataOutputStream ds = new DataOutputStream(dBuffer);
      byte[] bstr = new byte[64];
      byte[] tmp  = fecName.getBytes();
      for (int i=0; i<TStrings.FEC_NAME_SIZE; i++) bstr[i] = i < tmp.length ? tmp[i] : 0;
      ds.write(bstr,0,TStrings.FEC_NAME_SIZE);
      ds.write(netAdr,0,16);
      ds.write(lclAdr,0,8);
      tmp = strAdr.getBytes();
      for (int i=0; i<40; i++) bstr[i] = i < tmp.length ? tmp[i] : 0;
      ds.write(bstr,0,40);
      ds.writeInt(Swap.Int(portOffset));
      ds.writeInt(Swap.Int(inetProtocol));
      ds.writeInt(Swap.Int(tineProtocol));
      ds.writeInt(Swap.Int(addrFlags));
      ds.close();	
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("FECAddr","could not prepare byte stream",TErrorList.code_failure,e,0);
      return null;
    }	 
    hByteArray = dBuffer.toByteArray();
    return hByteArray;
  }  
  public void toStruct()
  {
    try
    {
      ByteArrayInputStream dis = new ByteArrayInputStream(hByteArray);
      DataInputStream ds = new DataInputStream(dis);
      byte[] bstr = new byte[64];
      ds.read(bstr,0,TStrings.FEC_NAME_SIZE);
      fecName = new String(bstr);
      ds.read(netAdr,0,16);
      ds.read(lclAdr,0,8);
      ds.read(bstr,0,40);
      strAdr = new String(bstr);
      portOffset = Swap.Int(ds.readInt());
      inetProtocol = Swap.Int(ds.readInt());
      tineProtocol = Swap.Int(ds.readInt());
      addrFlags = Swap.Int(ds.readInt());
      ds.close(); dis.close();
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("FECAddr","could not prepare byte stream",TErrorList.code_failure,e,0);
    }	 
  }
  public void setMulticastRules(int rules)
  {
    addrFlags |= (rules == 0) ? ADDRFLAG_MCAST_LEGACY : ADDRFLAG_MCAST_MODERN;
  }
  public int getMulticastRules()
  {
    return (addrFlags & ADDRFLAG_MCAST_MODERN);
  }
  public boolean useModernMulticastRules()
  {
    return ((addrFlags & ADDRFLAG_MCAST_MODERN) == ADDRFLAG_MCAST_MODERN);
  }
  public String getMulticastRulesAsString()
  {
    return useModernMulticastRules() ? "Modern" : "Legacy";
  }
  public String toString()
  {
    String msg = new String("FEC name : " + fecName);
    msg = msg.concat("\nFEC ip   : " + strAdr);
    msg = msg.concat("\nFEC port : " + portOffset);
    msg = msg.concat("\nFEC tine protocol : " + tineProtocol);
    msg = msg.concat("\nFEC multicast rules : " + getMulticastRulesAsString());
    return msg;  
  }
  public String getFecName()
  {
    return fecName;
  }
  public void setFecName(String fecName)
  {
    this.fecName = fecName;
  }
  public int getInetProtocol()
  {
    return inetProtocol;
  }
  public void setInetProtocol(int inetProtocol)
  {
    this.inetProtocol = inetProtocol;
  }
  public String getIpAddr()
  {
    return strAdr;
  }
  public void setIpAddr(String ipAddr)
  {
    this.strAdr = ipAddr;
  }
  public int getTineProtocol()
  {
    return tineProtocol;
  }
  public void setTineProtocol(int tineProtocol)
  {
    this.tineProtocol = tineProtocol;
  }
  public void setPortOffset(int portOffset)
  {
    this.portOffset = portOffset;
  }
  public void setMacAddr(byte[] macaddr)
  { // push into the lclAdr (for registration in the ENS)
    if (macaddr == null) return;
    for (int i=0; i<lclAdr.length && i<macaddr.length; i++)
    {
      lclAdr[i] = macaddr[i];
    }
  }
  public void setNetAddr(byte[] netaddr)
  { // push into the netAdr (for registration in the ENS)
    if (netaddr == null) return;
    for (int i=0; i<netAdr.length && i<netaddr.length; i++)
    {
      netAdr[i] = netaddr[i];
    }
  }
  public void setNetAddr(String addr)
  {
    if (addr == null || addr.length() == 0) return;
    try
    {
      byte[] baddr = InetAddress.getByName(addr).getAddress();
      setNetAddr(baddr);
    }
    catch (Exception e)
    {
      String arg0 = addr == null ? "null" : addr;
      MsgLog.log("setNetAddr(\""+arg0+"\")",e.toString(),TErrorList.code_failure,e,0);                          
    }
  }
}