/*
 * File FECAddr4.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.addrUtils;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import de.desy.tine.definitions.TErrorList;
import de.desy.tine.definitions.TStrings;
import de.desy.tine.endianUtils.Swap;
import de.desy.tine.server.logger.MsgLog;

public final class FECAddr4
{
  public static final String LOOPBACKADDR = "127.0.0.1";
  protected String fecName = new String(new StringBuffer(TStrings.FEC_NAME_SIZE));
  protected byte[] ipxNetwork = new byte[4];
  protected byte[] ipxNode = new byte[6];
  protected byte[] ipxImmediateAddress = new byte[6];
  protected String ipAddr = new String(new StringBuffer(16));
  protected byte[] ipHAddr = new byte[4];
  protected int portOffset;
  protected int inetProtocol;
  protected int tineProtocol;
  public static int sizeInBytes = TStrings.FEC_NAME_SIZE + 16 + 16 + 4 + 3*4;
  ByteArrayOutputStream dBuffer = new ByteArrayOutputStream(sizeInBytes);
  byte[] hByteArray;
  public boolean hasAdr()
  {
    for (int i=0; i<4; i++) if (ipHAddr[i] != 0) return true;
    return false;
  }
  public int getPortOffset()
  {
    return portOffset;
  }
  public byte[] toByteArray()
  {
    try
    {
      dBuffer.reset();
      DataOutputStream ds = new DataOutputStream(dBuffer);
      byte[] bstr = new byte[64];
      byte[] tmp  = fecName.getBytes();
      for (int i=0; i<TStrings.FEC_NAME_SIZE; i++) bstr[i] = i < tmp.length ? tmp[i] : 0;
      ds.write(bstr,0,TStrings.FEC_NAME_SIZE);
      ds.write(ipxNetwork,0,4);
      ds.write(ipxNode,0,6);
      ds.write(ipxImmediateAddress,0,6);
      tmp = ipAddr.getBytes();
      for (int i=0; i<16; i++) bstr[i] = i < tmp.length ? tmp[i] : 0;
      ds.write(bstr,0,16);
      ds.write(ipHAddr,0,4);
      ds.writeInt(Swap.Int(portOffset));
      ds.writeInt(Swap.Int(inetProtocol));
      ds.writeInt(Swap.Int(tineProtocol));
      ds.close();	
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("FECAddr","could not prepare byte stream",TErrorList.code_failure,e,0);
      return null;
    }	 
    hByteArray = dBuffer.toByteArray();
    return hByteArray;
  }  
  public void toStruct()
  {
    try
    {
      ByteArrayInputStream dis = new ByteArrayInputStream(hByteArray);
      DataInputStream ds = new DataInputStream(dis);
      byte[] bstr = new byte[64];
      ds.read(bstr,0,TStrings.FEC_NAME_SIZE);
      fecName = new String(bstr);
      ds.read(ipxNetwork,0,4);
      ds.read(ipxNode,0,6);
      ds.read(ipxImmediateAddress,0,6);
      ds.read(bstr,0,16);
      ipAddr = new String(bstr);
      ds.read(ipHAddr,0,4);
      portOffset = Swap.Int(ds.readInt());
      inetProtocol = Swap.Int(ds.readInt());
      tineProtocol = Swap.Int(ds.readInt());
      ds.close(); dis.close();
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("FECAddr","could not prepare byte stream",TErrorList.code_failure,e,0);
    }	 
  }
  public String toString()
  {
    String msg = new String("FEC name : " + fecName);
    msg = msg.concat("\nFEC ip   : " + ipAddr);
    msg = msg.concat("\nFEC port : " + portOffset);
    msg = msg.concat("\nFEC tine protocol : " + tineProtocol);
    return msg;  
  }
  public String getFecName()
  {
    return fecName;
  }
  public void setFecName(String fecName)
  {
    this.fecName = fecName;
  }
  public int getInetProtocol()
  {
    return inetProtocol;
  }
  public void setInetProtocol(int inetProtocol)
  {
    this.inetProtocol = inetProtocol;
  }
  public String getIpAddr()
  {
    return ipAddr;
  }
  public void setIpAddr(String ipAddr)
  {
    this.ipAddr = ipAddr;
  }
  public int getTineProtocol()
  {
    return tineProtocol;
  }
  public void setTineProtocol(int tineProtocol)
  {
    this.tineProtocol = tineProtocol;
  }
  public void setPortOffset(int portOffset)
  {
    this.portOffset = portOffset;
  }
}