/*
 * File SrvAddr.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.addrUtils;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.net.InetAddress;
import java.net.UnknownHostException;

import de.desy.tine.definitions.TErrorList;
import de.desy.tine.definitions.TStrings;
import de.desy.tine.definitions.TTransport;
import de.desy.tine.endianUtils.Swap;
import de.desy.tine.server.logger.MsgLog;
import de.desy.tine.stringUtils.Str;

/**
 * @author duval
 *
 * This class hold both the Server address information and the FEC address
 * information and corresponds to ExpDataStruct followed by FecDataStruct
 * in the C libraries.
 *  
 */
public final class SrvAddr
{
  public String fecName;  // 16
  public String subSystem;
  public String expName; // 32
  public String eqmName; // 8
  public String eqmContext; //32
  private byte[] ipx = new byte[16];
  private byte[] lclAdr = new byte[8];
  public String ipAddr; // 40
  public int portOffset;
  public byte[] netAdr = new byte[16];
  public int inetProtocol;
  public int tineProtocol;
  public int addrFlags;
  public static int expSizeInBytes =
      TStrings.FEC_NAME_SIZE +
      TStrings.SUBSYSTEM_NAME_SIZE +
      TStrings.EXPORT_NAME_SIZE +
      TStrings.EQM_NAME_SIZE +
      TStrings.CONTEXT_NAME_SIZE;
  public static int sizeInBytes = expSizeInBytes + FECAddr.sizeInBytes;
  public static int legacySizeInBytes = expSizeInBytes + FECAddr4.sizeInBytes;
  ByteArrayOutputStream dBuffer = new ByteArrayOutputStream(sizeInBytes);
  private byte[] hByteArray;
  SrvAddr()
  {
  }
  public SrvAddr(FECAddr fecAddr)
  {
    fecName = fecAddr.fecName;
    ipAddr = Str.trim(fecAddr.strAdr);
    portOffset = fecAddr.portOffset;
    addrFlags = fecAddr.addrFlags;
    System.arraycopy(fecAddr.lclAdr, 0, lclAdr, 0, 8);
    if (!fecAddr.hasAdr())
    {
      try
      {
        byte[] badr = InetAddress.getByName(ipAddr).getAddress();
        int len = badr.length; if (len > 16) len = 16;
        System.arraycopy(badr, 0, fecAddr.netAdr, 0, len);
      }
      catch (UnknownHostException e)
      {
        e.printStackTrace();
        MsgLog.log("SrvAddr",e.getMessage(),TErrorList.code_failure,e,0);
      }      
    }
    System.arraycopy(fecAddr.netAdr,0,netAdr,0,16);
  }
  public SrvAddr(FECAddr4 fecAddr)
  {
    fecName = fecAddr.fecName;
    ipAddr = Str.trim(fecAddr.ipAddr);
    portOffset = fecAddr.portOffset;
    if (!fecAddr.hasAdr())
    {
      try
      {
        byte[] badr = InetAddress.getByName(ipAddr).getAddress();
       // fecAddr.ipAddr = new String(""+(badr[0]&0xff)+"."+(badr[1]&0xff)+"."+(badr[2]&0xff)+"."+(badr[3]&0xff));
        System.arraycopy(badr, 0, fecAddr.ipHAddr, 0, 4);
      }
      catch (UnknownHostException e)
      {
        e.printStackTrace();
        MsgLog.log("SrvAddr",e.getMessage(),TErrorList.code_failure,e,0);
      }      
    }
    System.arraycopy(fecAddr.ipHAddr,0,netAdr,0,4);
  }
  public SrvAddr(String fecName,String fecIP,int fecPort,String eqmName,String expName,String ctxName)
  {
    this(fecName,fecIP,fecPort,FECAddr.ADDRFLAG_MCAST_MODERN,eqmName,expName,ctxName);
  }
  public SrvAddr(String fecName,String fecIP,int fecPort,int flags,String eqmName,String expName,String ctxName)
  {
    this.fecName = new String(fecName);
    eqmContext = new String(ctxName);
    this.expName = new String(expName);
    this.eqmName = new String(eqmName);
    ipAddr = Str.trim(new String(fecIP));
    portOffset = fecPort;
    inetProtocol = TTransport.UDP;
    tineProtocol = TTransport.thisProtocolLevel;
    addrFlags = flags;
    try
    {
      int len = ipAddr.contains(":") ? 16 : 4;
      System.arraycopy(InetAddress.getByName(ipAddr).getAddress(),0,netAdr,0,len);
    }
    catch (UnknownHostException e)
    {
      e.printStackTrace();
      MsgLog.log("SrvAddr",e.getMessage(),TErrorList.code_failure,e,0);
    }      
  }
  public int getPortOffset()
  {
    return portOffset;
  }
  public int getAddressFlags()
  {
    return addrFlags;
  }
  public void setAddressFlags(int flags)
  {
    addrFlags = flags;
  }
  private void pushStringToBytes(String s,byte[] b,int maxlen)
  {
    int i,len;
    if (s != null)
    {
      if((len=s.length()) > maxlen) len = maxlen;
      System.arraycopy(s.getBytes(),0,b,0,len);
    }
    else len = 0;
    for (i=len; i<maxlen; i++) b[i] = 0;    
  }
  public byte[] toByteArray(boolean isLegacy)
  { 
    try
    { // prepare double-address info for export to ENS
      dBuffer.reset();
      DataOutputStream ds = new DataOutputStream(dBuffer);
      // First: export name information (ExpDataStruct in c)
      byte[] bstr = new byte[64];
      pushStringToBytes(fecName, bstr, TStrings.FEC_NAME_SIZE);
      ds.write(bstr,0,TStrings.FEC_NAME_SIZE);
      pushStringToBytes(subSystem, bstr, TStrings.SUBSYSTEM_NAME_SIZE);
      ds.write(bstr,0,TStrings.SUBSYSTEM_NAME_SIZE);
      pushStringToBytes(expName, bstr, TStrings.EXPORT_NAME_SIZE);
      ds.write(bstr,0,TStrings.EXPORT_NAME_SIZE);
      pushStringToBytes(eqmName, bstr, TStrings.EQM_NAME_SIZE);
      ds.write(bstr,0,TStrings.EQM_NAME_SIZE);
      pushStringToBytes(eqmContext, bstr, TStrings.CONTEXT_NAME_SIZE);
      ds.write(bstr,0,TStrings.CONTEXT_NAME_SIZE);
      // Second: fec name information (FecDataStruct in c)
      pushStringToBytes(fecName, bstr, TStrings.FEC_NAME_SIZE);
      ds.write(bstr,0,TStrings.FEC_NAME_SIZE);
      if (isLegacy)
      {
        ds.write(ipx,0,16);
        pushStringToBytes(ipAddr, bstr, TStrings.ADDR_SIZE);
        ds.write(bstr,0,TStrings.ADDR_SIZE);
        ds.write(netAdr,0,4);
        ds.writeInt(Swap.Int(portOffset));
        ds.writeInt(Swap.Int(inetProtocol));
        ds.writeInt(Swap.Int(tineProtocol));
      }
      else
      {
        ds.write(netAdr,0,16);
        ds.write(lclAdr,0,8);
        pushStringToBytes(ipAddr,bstr,TStrings.ADDR_SIZE_IPV6);
        ds.write(bstr,0,TStrings.ADDR_SIZE_IPV6);
        ds.writeInt(Swap.Int(portOffset));
        ds.writeInt(Swap.Int(inetProtocol));
        ds.writeInt(Swap.Int(tineProtocol));
        ds.writeInt(Swap.Int(addrFlags));
      }
      ds.close();	
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("SrvAddr","could not prepare byte stream",TErrorList.code_failure,e,0);
      return null;
    }	 
    hByteArray = dBuffer.toByteArray();
    return hByteArray;
  }  
  public void toStruct(boolean isLegacy)
  {
    try
    {
      ByteArrayInputStream dis = new ByteArrayInputStream(hByteArray);
      DataInputStream ds = new DataInputStream(dis);
      // First: export information:
      byte[] bstr = new byte[64];
      ds.read(bstr,0,TStrings.FEC_NAME_SIZE);
      fecName = new String(bstr,0,TStrings.FEC_NAME_SIZE).trim();
      ds.read(bstr,0,TStrings.SUBSYSTEM_NAME_SIZE);
      subSystem = new String(bstr,0,TStrings.SUBSYSTEM_NAME_SIZE).trim();
      ds.read(bstr,0,TStrings.EXPORT_NAME_SIZE);
      expName = new String(bstr,0,TStrings.EXPORT_NAME_SIZE).trim();
      ds.read(bstr,0,TStrings.EQM_NAME_SIZE);
      eqmName = new String(bstr,0,TStrings.EQM_NAME_SHORTSIZE).trim();
      ds.read(bstr,0,TStrings.CONTEXT_NAME_SIZE);
      eqmContext = new String(bstr,0,TStrings.CONTEXT_NAME_SHORTSIZE).trim();
      // Second: FEC address information:
      ds.read(bstr,0,TStrings.FEC_NAME_SIZE);
      fecName = new String(bstr,0,TStrings.FEC_NAME_SIZE).trim();
      if (isLegacy)
      {
        ds.read(ipx,0,16);
        ds.read(bstr,0,TStrings.ADDR_SIZE);
        ipAddr = new String(bstr,0,TStrings.ADDR_SIZE).trim();
        ds.read(netAdr,0,4);
        portOffset = Swap.Int(ds.readInt());
        inetProtocol = Swap.Int(ds.readInt());
        tineProtocol = Swap.Int(ds.readInt());
      }
      else
      {
        ds.read(netAdr,0,16);
        ds.read(lclAdr,0,8);
        ds.read(bstr,0,TStrings.ADDR_SIZE_IPV6);
        ipAddr = Str.trim(new String(bstr,0,TStrings.ADDR_SIZE_IPV6));
        portOffset = Swap.Int(ds.readInt());
        inetProtocol = Swap.Int(ds.readInt());
        tineProtocol = Swap.Int(ds.readInt());
        addrFlags = Swap.Int(ds.readInt());        
      }
      ds.close(); dis.close();
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("SrvAddr","could not prepare byte stream",TErrorList.code_failure,e,0);
    }	 
  }  
  public boolean useModernMulticastRules()
  {
    return ((addrFlags & FECAddr.ADDRFLAG_MCAST_MODERN) == FECAddr.ADDRFLAG_MCAST_MODERN);
  }
  public String getMulticastRulesAsString()
  {
    return useModernMulticastRules() ? "Modern" : "Legacy";
  }
  public String toString()
  {
    String msg = new String("Exp name : " + expName);
    msg = msg.concat("\nEqm name : " + eqmName);
    msg = msg.concat("\nContext  : " + eqmContext);
    msg = msg.concat("\nFEC name : " + fecName);
    msg = msg.concat("\nFEC ip   : " + ipAddr);
    msg = msg.concat("\nFEC portOffset : " + portOffset);
    msg = msg.concat("\ntine protocol : " + tineProtocol);
    msg = msg.concat("\nFEC multicast rules : " + getMulticastRulesAsString());
    return msg;  
  }
}