/*
 * File AlarmMonitor.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.alarmUtils;

import java.util.*;
import de.desy.tine.server.alarms.TAlarmMessage;

public class AlarmMonitor
{
  private int connectionStatus = 0;
  private AlarmMonitorCallback amcb = null;
  private AlarmMonitorHandler amHdlr = null;
  public AlarmMonitor(AlarmMonitorHandler myAlarmsHandler)
  {
     amHdlr = myAlarmsHandler; 
  }
  /** 
   * Gets the alarm message array 
   * 
   * This method gets the alarm message array containing all alarms
   * within the recent time interval consisting of the last 2 hours 
   * for the AlarmMonitor passed in the TAlarmSystem.monitorAlarms() 
   * method.  Typically, an alarms application might several such monitors
   * for different alarms systems and different minimum severities.
   *  
   * @return an array of Alarm Messages 
   *
   */
  public TAlarmMessage[] getAlarms()
  {
    return amcb == null ? null : amcb.getAlarms();
  }
  private TAlarmMessage[] getPrunedAlarmList(TAlarmMessage[] lst,int minSeverity,boolean activeOnly)
  {
    if (lst == null) return null;
    Vector<TAlarmMessage> av = new Vector<TAlarmMessage>();
    for (TAlarmMessage a : lst)
    {
      if (a.getAlarmSeverity() < minSeverity) continue;
      if (activeOnly && !a.isActive()) continue;
      av.add(a);
    }
    if (av.size() == 0) return null;
    return av.toArray(new TAlarmMessage[av.size()]);    
  }
  /** 
   * Gets the alarm message array 
   * 
   * This method gets the alarm message array containing all alarms
   * within the recent time interval consisting of the last 2 hours 
   * for the AlarmMonitor passed in the TAlarmSystem.monitorAlarms() 
   * method.  Typically, an alarms application might several such monitors
   * for different alarms systems and different minimum severities.
   * 
   * Only those alarms with at least the severity given are returned in the list.
   * If 'activeOnly' is true, only active alarms are returned in the list 
   * (no termination alarm messages).
   * 
   * @param minSeverity is the minimum severity of interest.  Note that this
   * is also in conjunction with the severity used in establishing the monitor
   * @param activeOnly is a flag which if true signals the call to return only
   * those alarms which are still active.
   *  
   * @return an array of Alarm Messages 
   *
   */
  public TAlarmMessage[] getAlarms(int minSeverity,boolean activeOnly)
  {
    if (amcb == null) return null;
    return getPrunedAlarmList(amcb.getAlarms(), minSeverity, activeOnly);
  }
  /** 
   * Gets the last acquired alarm message array 
   * 
   * This method gets the last acquired alarm message array containing only 
   * the alarms which caused the handler to be called. That is, the 'new'
   * alarms since the previous call to the alarm handler.
   *  
   * @return an array of Alarm Messages 
   *
   */
  public int getNumberDisabledAlarms()
  {
    return amcb == null ? 0 : amcb.getNumberDisableAlarms();
  }
  /** 
   * Gets the last acquired alarm message array 
   * 
   * This method gets the last acquired alarm message array containing only 
   * the alarms which caused the handler to be called. That is, the 'new'
   * alarms since the previous call to the alarm handler.
   *  
   * @return an array of Alarm Messages 
   *
   */
  public TAlarmMessage[] getLastAcquiredAlarms()
  {
    return amcb == null ? null : amcb.getLastAcquiredAlarms();
  }
  /** 
   * Gets the last acquired alarm message array 
   * 
   * This method gets the last acquired alarm message array containing only 
   * the alarms which caused the handler to be called. That is, the 'new'
   * alarms since the previous call to the alarm handler.
   *  
   * Only those alarms with at least the severity given are returned in the list.
   * If 'activeOnly' is true, only active alarms are returned in the list 
   * (no termination alarm messages).
   * 
   * @param minSeverity is the minimum severity of interest.  Note that this
   * is also in conjunction with the severity used in establishing the monitor
   * @param activeOnly is a flag which if true signals the call to return only
   * those alarms which are still active.
   *  
   * @return an array of Alarm Messages 
   *
   */
  public TAlarmMessage[] getLastAcquiredAlarms(int minSeverity,boolean activeOnly)
  {
    if (amcb == null) return null;
    return getPrunedAlarmList(amcb.getLastAcquiredAlarms(), minSeverity, activeOnly);
  }
  /** 
   * Gets the last acquired alarm timestamp 
   * 
   * This method gets the last acquired alarm timestamp 
   * since the previous call to the alarm handler.
   *  
   * @return an array of Alarm Messages 
   *
   */
  public long getLastAcquiredAlarmTime()
  {
    return amcb == null ? 0 : amcb.getLastAcquired();
  }
  public String getContext()
  {
    return amcb == null ? null : amcb.getContext();
  }
  public String getServer()
  {
    return amcb == null ? null : amcb.getServer();
  }
  public String getSubsystem()
  {
    return amcb == null ? null : amcb.getSubsystem();
  }
  public AlarmMonitorHandler getAmHdlr()
  {
    return amHdlr;
  }
  public void setAmHdlr(AlarmMonitorHandler myHandler)
  {
    amHdlr = myHandler;
  }
  protected void setAmcb(AlarmMonitorCallback amcb)
  {
    this.amcb = amcb;
  }
  public int getConnectionStatus()
  {
    return connectionStatus;
  }
  public void setConnectionStatus(int status)
  {
    connectionStatus = status;
  }
  public void clearNonRecent() {
	if (amcb != null) amcb.clearNonRecent();
  }
}
