/*
 * File CasAction.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.alarmUtils;

import de.desy.tine.definitions.TFormat;
import de.desy.tine.definitions.TStrings;
import de.desy.tine.structUtils.TTaggedStructure;

public class CasAction extends TTaggedStructure {

	// CasAct
//  char email[128];
//  char ctx[CONTEXT_NAME_SIZE];
//  char srv[EXPORT_NAME_SIZE];
//  char dev[DEVICE_NAME_SIZE];
//  char prp[PROPERTY_NAME_SIZE];
//  char actionSignalString[128];
//  int actionSignalFormat;
//  int alarmCode;
	
	public void clear() {
		contextString = null;
		serverString = null;
		deviceString = null;
		propertyString = null;
		actionSignalString = null;
		emailString = null;
		actionFilterString = null;
	}

	private char[] email = new char[128]; /* email */
	private String emailString;
	private char[] ctx = new char[TStrings.CONTEXT_NAME_SIZE]; /* calling context */
	private String contextString;
	private char[] srv = new char[TStrings.EXPORT_NAME_SIZE]; /* calling equipment module */
	private String serverString;
	private char[] prp = new char[TStrings.PROPERTY_NAME_SIZE]; /* calling property */
	private String propertyString;
	private char[] dev = new char[TStrings.DEVICE_NAME_SIZE]; /* calling device name */
	private String deviceString;
	private char[] actionSignal = new char[128]; /* action signal */
	private String actionSignalString;
	private char[] actionFilter = new char[256];
	private String actionFilterString;
	private int[] actSignalFormat = new int[1]; /* action signal format */
	private int[] alarmCode = new int[1]; /* alarm code */

	private void initStructure() {
		addField(email,"email");
		addField(ctx,"context");
		addField(srv,"server");
		addField(dev,"device");
		addField(prp,"property");
		addField(actionSignal,"actionSignalString");
    addField(actionFilter, "actionFilter");
		addField(actSignalFormat, "actionSignalFormat");
		addField(alarmCode, "alarmCode");
		initDone();
	}

	public CasAction() {
		super("CasActEx");
		initStructure();
	}

	public CasAction(CasAction reference) {
		super("CasActEx");
		initStructure();
		alarmCode[0] = reference.alarmCode[0];
		actSignalFormat[0] = reference.actSignalFormat[0];
		setActionSignal(reference.getActionSignal());
		setActionFilter(reference.getActionFilter());
		setEmail(reference.getEmail());
		setContext(reference.getContext());
		setServer(reference.getServer());
		setDevice(reference.getDevice());
		setProperty(reference.getProperty());
	}

	public int getActionSignalFormat() {
		return TFormat.makeByteFormatCode(actSignalFormat[0]);
	}

	public void setActionSignalFormat(int value) {
		actSignalFormat[0] = TFormat.makeStdFormatCode(value);
	}

	public int getAlarmCode() {
		return alarmCode[0];
	}

	public void setAlarmCode(int value) {
		alarmCode[0] = value;
	}

	public String getContext() {
		if (contextString == null)
			contextString = new String(ctx).trim();
		return contextString;
	}

	public void setContext(String context) {
		pushChars(context,ctx);
		contextString = new String(context);
	}

	public String getServer() {
		if (serverString == null)
			serverString = new String(srv).trim();
		return serverString;
	}

	public void setServer(String server) {
		pushChars(server,srv);
		serverString = new String(server);
	}

	public String getDevice() {
		if (deviceString == null)
			deviceString = new String(dev).trim();
		return deviceString;
	}

	public void setDevice(String device) {
		pushChars(device,dev);
		deviceString = new String(device);
	}

	public String getProperty() {
		if (propertyString == null)
			propertyString = new String(prp).trim();
		return propertyString;
	}

	public void setProperty(String property) {
		pushChars(property,prp);
		propertyString = new String(property);
	}
	
	public String getActionSignal() {
		if (actionSignalString == null)
			actionSignalString = new String(actionSignal).trim();
		return actionSignalString;
	}

	public void setActionSignal(String actionSignal) {
		pushChars(actionSignal,this.actionSignal);
		actionSignalString = new String(actionSignal);
	}
	
  public String getActionFilter() {
    if (actionFilterString == null)
      actionFilterString = new String(actionFilter).trim();
    return actionFilterString;
  }

  public void setActionFilter(String actionFilter) {
    pushChars(actionFilter,this.actionFilter);
    actionFilterString = new String(actionFilter);
  }
  
	public String getEmail() {
		if (emailString == null)
			emailString = new String(email).trim();
		return emailString;
	}

	public void setEmail(String email) {
		pushChars(email,this.email);
		emailString = new String(email);
	}
}
