/*
 * File CasDb.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.alarmUtils;

import de.desy.tine.definitions.TStrings;
import de.desy.tine.structUtils.TTaggedStructure;

public class CasDb extends TTaggedStructure { 
//  char devContext[CONTEXT_NAME_SIZE];
//  char devServer[EXPORT_NAME_SIZE];
//  char subSystem[SUBSYSTEM_NAME_SIZE];
//  char arcExt[XTN_NAME_SIZE];

	//  int SeverityLevel;
	//  int ArchiveLevel;
	//  int Retention;
	//  int AlarmLevel;
	//  int AlarmSystem;
	//  int offline;
	public void clear() {
		contextString = null;
		serverString = null;
		extString = null;
		subsystemString = null;
	}

	private char[] ctx = new char[TStrings.CONTEXT_NAME_SIZE]; /* calling context */
	private String contextString;
	private char[] srv = new char[TStrings.EXPORT_NAME_SIZE]; /* calling server */
	private String serverString;
	private char[] subs = new char[TStrings.SUBSYSTEM_NAME_SIZE]; /* calling subsystem */
	private String subsystemString;
	private char[] ext = new char[TStrings.XTN_NAME_SIZE]; /* calling device name */
	private String extString;
	private int[] severityLevel = new int[1];
	private int[] archiveLevel = new int[1];
	private int[] retention = new int[1];
	private int[] alarmLevel = new int[1];
	private int[] alarmSystem = new int[1];
	private int[] offline = new int[1]; /*  set offline to < 0 to delete it */

	public static String makeExtensionFromServerNameAndTruncate(String srv)
	{
	  String ext = null;
	  if (srv != null)
	  {
	    char[] srvChrs = srv.toUpperCase().toCharArray();
	    char[] extChrs = new char[TStrings.XTN_NAME_SIZE];
	    int k = 0;
	    for (int i=0; i<srvChrs.length; i++)
	    {
	      if (srvChrs[i] == '.') continue;
        if (srvChrs[i] == '_') continue;
        extChrs[k] = srvChrs[i];
        if (++k == TStrings.XTN_NAME_SIZE) break;
	    }
	    ext = new String(extChrs);
	    
	  }
	  return ext;
	}
	
	public static String makeExtensionFromServerName(String srv)
	{
	  if (srv == null)
		  return null;
	  return srv.replace("_","").replace(".","");
	}
	
	private void initStructure() {
		addField(ctx,"devContext");
		addField(srv,"devServer");
		addField(subs,"subSystem");
		addField(ext,"arcExt");
		addField(severityLevel,"SeverityLevel");
		addField(archiveLevel,"ArchiveLevel");
		addField(retention,"Retention");
		addField(alarmLevel,"AlarmLevel");
		addField(alarmSystem,"AlarmSystem");
		addField(offline,"offline");
		initDone();
	}

	public CasDb() {
		super("CasDb");
		initStructure();
	}

	public CasDb(CasDb reference) {
		super("CasDb");
		initStructure();
		severityLevel[0] = reference.severityLevel[0];
		alarmLevel[0] = reference.alarmLevel[0];
		archiveLevel[0] = reference.archiveLevel[0];
		alarmSystem[0] = reference.alarmSystem[0];
		retention[0] = reference.retention[0];
		offline[0] = reference.offline[0];
		setContext(reference.getContext());
		setServer(reference.getServer());
		setSubsystem(reference.getSubsystem());
		setArcExt(reference.getArcExt());
	}

	public int getSeverityLevel() {
		return severityLevel[0];
	}

	public void setSeverityLevel(int value) {
		severityLevel[0] = value;
	}

	public int getArchiveLevel() {
		return archiveLevel[0];
	}

	public void setArchiveLevel(int archiveLevel) {
		this.archiveLevel[0] = archiveLevel;
	}

	public int getAlarmLevel() {
		return alarmLevel[0];
	}

	public void setAlarmLevel(int alarmLevel) {
		this.alarmLevel[0] = alarmLevel;
	}

	public int getAlarmSystem() {
		return alarmSystem[0];
	}

	public void setAlarmSystem(int alarmSystem) {
		this.alarmSystem[0] = alarmSystem;
	}

	public int getRetention() {
		return retention[0];
	}

	public void setRetention(int retention) {
		this.retention[0] = retention;
	}

	public String getContext() {
		if (contextString == null)
			contextString = new String(ctx).trim();
		return contextString;
	}

	public void setContext(String context) {
		pushChars(context,ctx);
		contextString = new String(context);
	}

	public String getServer() {
		if (serverString == null)
			serverString = new String(srv).trim();
		return serverString;
	}

	public void setServer(String server) {
		pushChars(server,srv);
		serverString = new String(server);
	}

	public String getSubsystem() {
		if (subsystemString == null)
			subsystemString = new String(subs).trim();
		return subsystemString;
	}

	public void setSubsystem(String subsystem) {
		pushChars(subsystem,subs);
		subsystemString = new String(subs);
	}

	public String getArcExt() {
		if (extString == null)
			extString = new String(ext).trim();
		return extString;
	}

	public void setArcExt(String arcExt) {
		pushChars(arcExt,ext);
		extString = new String(arcExt);
	}
	
	
	public int getOffline() {
		return offline[0];
	}
	
	
	public void setOffline(int offline) {
		this.offline[0] = offline;
	}
}
