/*
 * File TAwtAsyncCallback.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

/*
 * Created on Aug 16, 2007
 *
  */
package de.desy.tine.client;

import javax.swing.SwingUtilities;

/**
 * A TLink callback which is executed asynchronously on the AWT 
 * event dispatching thread.
 * <p>
 * This callback is useful for programs with an AWT or Swing GUI.
 * It is not advisable to make any GUI modifications directly
 * from TLinkCallback.callback(), since different threads are
 * involved. An indirection via the AWT event queue has to 
 * be implemented.   
 * <p>
 * This class simplifies GUI callbacks for client programmers.
 * The programmer simply replaces TLinkCallback with TAwtSyncCallback
 * and implements onCallback() instead of callback. From onCallback(),
 * GUI components can be accessed safely.
 * <p>
 * In this class, the inherited TINE callback() method calls 
 * SwingUtilities.invokeLater() and returns.
 * The AWT event dispatcher calls onCallback() later. 
 * <p>
 * <b>Warning:</b> onCallback() has the original TLink as a parameter.
 * It cannot be guaranteed that TLink is still in its original state when
 * this method is called. If this is required, better use TAwtSyncCallback.
 * However, the original link status and error message are saved and 
 * can be accessed with the methods getLinkStatus() and getErrorString().
 * 
 * @author MCS
 */
public abstract class TAwtAsyncCallback implements TLinkCallback, Runnable {

    private TLink storedLink = null;
    private int linkStatus = 0;
    private String errorString = "";
    
    /**
     * The original TINE callback method.
     */
    public final void callback(TLink link) {
        this.storedLink = link;
        this.linkStatus = link.getLinkStatus();
        this.errorString = link.getLastError();
        SwingUtilities.invokeLater(this);
    }

    /**
     * Called from the AWT event thread.
     */
    public final void run() {
        onCallback(storedLink);
    }
    
    /**
     * Returns the link status.
     * @return
     */
    public final int getLinkStatus() {
        return linkStatus;
    }
    
    /**
     * Returns the error string.
     */
    public final String getErrorString() {
        return errorString;
    }
    
    /**
     * Replaces TLinkCallback.callback().
     * This method to be implemented by derived classes. 
     * 
     * @param link The TLink which fired this callback.
     */
    public abstract void onCallback(TLink link);

}
