/*
 * File TFilterLink.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.client;

import java.util.Iterator;
import java.util.LinkedList;

import de.desy.tine.dataUtils.*;
import de.desy.tine.definitions.*;
import de.desy.tine.exceptions.TineRuntimeErrorException;

public class TFilterLink implements TLinkCallback
{
  class TFilterTarget
  {
    double dv = 0;
    String sv = "";
    comp fcmp = comp.eq;
    boolean valid;
    TFilterTarget(String value,String comparator)
    {
      sv = value;
      try
      {
        dv = Double.parseDouble(sv);
      }
      catch (Exception x) {};
      fcmp = getComparator(comparator);
    }
  }
  public String toString()
  {
    return toString(false);
  }
  public String toString(boolean verbose)
  {
    if (lnk == null) return "no assigned filter";
    if (!verbose) return filterString;
    StringBuilder sb = new StringBuilder();
    sb.append(filterString);
    sb.append("\n\tlink status: "+(lnk == null ? "not linked!" : TErrorList.toString(lnk.getLinkStatus())));
    sb.append("\n\tvalid: "+isValid());
    //return lnk.getFullDeviceNameAndProperty()+" "+getCompString()+" "+sv;
    return sb.toString();
  }
  private TLink lnk = null;
  private static enum comp { neq, eq, lt, gt, le, ge; };
  private static final String[] compToken = { "!=", "==", "<=", ">=", "=", "<", ">" };
  private static final String paramToken = "&";
  private static final String andToken = "&&";
  private static final String orToken = "||";
  private static final String regexOr = "\\|\\|";
  //private comp thisComparator = comp.eq;
  private String key = null;
  private String filterTarget = null;
  public String getFilterTarget() { return filterTarget; }
  public String getKey() { return key; }
  //private double dv = 0;
  //private String sv = "";
  LinkedList<TFilterTarget> target = new LinkedList<TFilterTarget>();
  private int altParam = -1;
  public int getAlternateParameter() { return altParam; }
  private boolean active = false;
  public boolean isActive() { return active; }
  public static String getCompString(comp value)
  {
    switch (value)
    {
      case eq: return "=";
      case neq: return "!=";
      case lt: return "<";
      case gt: return ">";
      case le: return "<=";
      case ge: return ">=";
    }
    return "?";
  }
  public void remove()
  {
    if (lnk != null) lnk.close();
    lnk = null;
    filterIsValid = true;
    active = false;
  }
  private void makeFilterLink(String addressKey,String comparator,String value)
  {
    TFilterTarget ft = new TFilterTarget(value,comparator);
    if (TLinkFactory.getFilterLink(addressKey) != null)
    { // already have a filter link
      if (target.contains(ft)) return; // and already have this target
      target.add(ft); // add another target
      return;
    }
    short dFmt = TFormat.CF_DEFAULT;
    int dSiz = 128;
    TDataType dout = new TDataType(dSiz,dFmt);
    lnk = new TLink(addressKey,dout,null,TAccess.CA_READ);
    int lid = lnk.attach(TMode.CM_TIMER, this, 1000);
    if (lid < 0)
    {
      throw new TineRuntimeErrorException(-lid);
    }
    target.add(new TFilterTarget(value,comparator));
    TLinkFactory.addFilterLink(key,this);
    active = true;
  }
  private String filterString = null;
  public String getFilterString() { return filterString; }
  private comp getComparator(String comparator)
  {
    if      (comparator.compareTo("=") == 0) return comp.eq;
    else if (comparator.compareTo("==") == 0) return comp.eq;
    else if (comparator.compareTo("!=") == 0) return comp.neq;
    else if (comparator.compareTo("<") == 0) return comp.lt;
    else if (comparator.compareTo(">") == 0) return comp.gt;
    else if (comparator.compareTo("<=") == 0) return comp.le;
    else if (comparator.compareTo(">=") == 0) return comp.ge;
    else
    {
      throw new TineRuntimeErrorException(TErrorList.invalid_parameter);      
    }    
  }
  private int getComparatorLength(String value)
  {
    if (value == null) return 0;
    if (value.startsWith("==")) return 2;
    if (value.startsWith("=")) return 1;
    if (value.startsWith("!=")) return 2;
    if (value.startsWith("<=")) return 2;
    if (value.startsWith(">=")) return 2;
    if (value.startsWith("<")) return 1;
    if (value.startsWith(">")) return 1;
    return 0;
  }
  public static String getKeyFromFilterString(String parsableFilterString)
  {
    if (parsableFilterString == null) return null;
    String[] parts = parsableFilterString.split(paramToken);
    parsableFilterString = parts[0];
    int idx = -1;
    for (int i=0; i<compToken.length && idx == -1; i++)
    {
      idx = parsableFilterString.indexOf(compToken[i]);
    }
    return idx < 0 ? null : parsableFilterString.substring(0, idx).trim();
  }
  public TFilterLink(String parsableFilterString)
  {
    if (parsableFilterString == null || parsableFilterString.length() == 0)
    {
      throw new TineRuntimeErrorException(TErrorList.argument_list_error);
    }
    int idx = -1, len = 1;
    filterString = parsableFilterString;
    if ((idx=parsableFilterString.indexOf('&')) > 0 && !parsableFilterString.substring(idx, idx+2).equals("&&"))
    { // given as /ctx/srv/dev[prp]=v&sev=3
      String ts = parsableFilterString.substring(idx+1);
      parsableFilterString = parsableFilterString.substring(0, idx);
      try
      {
        switch (ts.charAt(0))
        {
          case 't' : // 'tolerance' for history filters
          case 's' : // 'severity' for alarm filters
            int pidx = ts.indexOf('=');
            if (pidx > 0) altParam = Integer.parseInt(ts.substring(pidx+1));
            break;
          default:
            break;
        }
      }
      catch (Exception x)
      {
        x.printStackTrace();
        altParam = -1;
      }      
    }
    idx = -1;
    for (int i=0; i<compToken.length && idx == -1; i++)
    {
      idx = parsableFilterString.indexOf(compToken[i]);
      len = i < 4 ? 2 : 1;
    }
    if (idx < 0) throw new TineRuntimeErrorException(TErrorList.parsing_error);
    String key = parsableFilterString.substring(0, idx).trim();
    String cmp = parsableFilterString.substring(idx, idx+len);
    String val = parsableFilterString.substring(idx+len).trim();
    this.key = key;
    filterTarget = parsableFilterString.substring(idx);
    String ctok = (idx=val.indexOf(orToken)) >= 0 ? orToken : (idx=val.indexOf(andToken)) >= 0 ? andToken : null;
    String[] vals = new String[0];
    if (ctok != null) 
    {
      if (ctok.equals(orToken)) ctok = regexOr;
      vals = val.split(ctok);
    }
    else
    {
      vals = new String[] { val };
    }
    for (int i=0; i<vals.length; i++)
    {
      val = vals[i];
      if (val == null || val.length() == 0) continue;
      if ((len=getComparatorLength(vals[i])) > 0)
      {
        cmp = vals[i].substring(0, len);
        val = vals[i].substring(len);
      }
      makeFilterLink(key,cmp,val);
    }
  }
  public TFilterLink(String addressKey,String comparator,String value)
  {
    this(addressKey,comparator,value,(String)null);
  }
  public TFilterLink(String addressKey,String comparator,String value,String param)
  {
    if (addressKey == null || addressKey.length() == 0 ||
        comparator == null || comparator.length() == 0 ||
        value == null || value.length() == 0)
    {
      throw new TineRuntimeErrorException(TErrorList.argument_list_error);
    }
    this.key = addressKey;
    filterString = addressKey+" "+comparator+" "+value;
    filterTarget = comparator+" "+value;
    if (param != null && param.length() > 0) filterString += "&" + param;
    makeFilterLink(addressKey,comparator,value);
  }
  private boolean filterIsValid = true;
  public boolean isValid()
  {
    return filterIsValid;
  }
  @Override
  public void callback(TLink link)
  {
    if (!TErrorList.hasData(link.linkStatus))
    { // punt
      filterIsValid = true;
      return;
    }
    TDataType d = link.getOutputDataObject();
    if (d == null)
    {
      filterIsValid = true;
      link.close();
      return;      
    }
    boolean tgt_add = filterTarget.contains(andToken);
    int nvalid = 0, ntargets = 0;
    Iterator<TFilterTarget> fti = target.iterator();
    TFilterTarget ft = null;
    if (TFormat.isNumberFormat(d.dFormat))
    {
      boolean top = false;
      while (fti.hasNext())
      {
        ntargets++;
        ft = fti.next();
        switch (ft.fcmp)
        {
          case eq: top = true;
          case neq:
          {
            int[] v = new int[1];        
            d.getData(v);
            ft.valid = top ? (v[0] == (int)ft.dv) : (v[0] != (int)ft.dv);
            break;
          }
          case gt: top = true;
          case lt:
          {
            double[] v = new double[1];
            d.getData(v); 
            ft.valid = top ? (v[0] > ft.dv) : (v[0] < ft.dv);
            break;
          }                    
          case ge: top = true;
          case le:
          {
            double[] v = new double[1];
            d.getData(v); 
            ft.valid = top ? (v[0] >= ft.dv) : (v[0] <= ft.dv);
            break;
          }                    
        }
        if (ft.valid) nvalid++; 
      }
    }
    else if (TFormat.isName(d.dFormat) || TFormat.CF_TEXT == d.dFormat)
    {
      char[] cs = new char[128];
      d.getData(cs);
      String s = new String(cs).trim();
      while (fti.hasNext())
      {
        ntargets++;
        ft = fti.next();
        switch (ft.fcmp)
        {
          case eq: ft.valid = ft.sv.compareToIgnoreCase(s) == 0; break;
          case neq: ft.valid = ft.sv.compareToIgnoreCase(s) != 0; break;
          case gt: ft.valid = s.compareToIgnoreCase(ft.sv) > 0; break;
          case lt: ft.valid = s.compareToIgnoreCase(ft.sv) < 0; break;
          case ge: ft.valid = s.compareToIgnoreCase(ft.sv) >= 0; break;
          case le: ft.valid = s.compareToIgnoreCase(ft.sv) <= 0; break;          
        }
        if (ft.valid) nvalid++; 
      }
    }
    filterIsValid = tgt_add ? ntargets == nvalid : nvalid > 0;
  }
}
