/*
 * File TLinkGroup.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

/*
 * Created on Oct 2, 2006
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package de.desy.tine.client;

import java.util.Iterator;
import java.util.concurrent.ConcurrentLinkedQueue;
import de.desy.tine.dataUtils.TDataTime;
import de.desy.tine.definitions.TErrorList;
import de.desy.tine.definitions.TMode;
import de.desy.tine.server.logger.MsgLog;

public class TLinkGroup
{
  public static final int GRP_JITTER = 50;
  public static final int GRP_SYNC_NOSYNC = 0;
  public static final int GRP_SYNC_STARTSYNC = 1;
  public static final int GRP_SYNC_NOTIFICATION = 2;
  public static final int GRP_SYNC_CYCLEOFFSET = 3;
  public static final int GRP_SYNC_INTERVAL = 4;
  public static final int GRP_SYNC_INSYNC = 5;
  public static final String grpSyncLevels[] = { 
      "not synchronized", "sychronization startup", "synchronize event notification", 
      "synchronize cycle offset", "synchronize effective update interval", "is synchronized" };

  private ConcurrentLinkedQueue<TLink> lnks;
  private TLinkCallback tlcb;
  private TCallback tcb;
  private TLinkFactory tlf = null;
  private int syncLevel = GRP_SYNC_NOSYNC;
  private boolean incMono = false;
  public int getSynchronizationLevel()
  {
    return syncLevel;
  }
  public void setSynchronizationLevel(int sycnLevel)
  {
    this.syncLevel = sycnLevel;
  }
  public boolean isIncrementingMonotonically()
  {
    return incMono;
  }
  public void setIncrementingMonotonically(boolean incMono)
  {
    this.incMono = incMono;
  }
  public boolean hasHeadIncremented()
  {
    return headInc;
  }
  public void setHeadIncremented(boolean headInc)
  {
    this.headInc = headInc;
  }
  public int getUpdateInterval()
  {
    return updateInterval;
  }
  public void setUpdateInterval(int updateInterval)
  {
    this.updateInterval = updateInterval;
  }
  public TLink getGroupHead()
  {
    return lnkHead;
  }
  public void setGroupHead(TLink lnkHead)
  {
    this.lnkHead = lnkHead;
  }
  public int getSysStamp()
  {
    return sysStamp;
  }
  public void setSysStamp(int sysStamp)
  {
    this.sysStamp = sysStamp;
  }
  public int getLastStamp()
  {
    return lastStamp;
  }
  public void setLastStamp(int lastStamp)
  {
    this.lastStamp = lastStamp;
  }
  public long getLastUpdate()
  {
    return lastUpdate;
  }
  public void setLastUpdate(long lastUpdate)
  {
    this.lastUpdate = lastUpdate;
  }
  public long getTimeUpdate()
  {
    return timeUpdate;
  }
  public void setTimeUpdate(long timeUpdate)
  {
    this.timeUpdate = timeUpdate;
  }
  public int getTimeDispersion()
  {
    return bndWdthT;
  }
  public void setTimeDispersion(int bndWdthT)
  {
    this.bndWdthT = bndWdthT;
  }
  public int getCountDispersion()
  {
    return bndWdthC;
  }
  public void setCountDispersion(int bndWdthC)
  {
    this.bndWdthC = bndWdthC;
  }
  private boolean headInc = false;
  private int updateInterval = 0;
  private TLink lnkHead = null;
  private int sysStamp = 0;
  private int lastStamp = 0;
  private long timeUpdate = 0;
  private long lastUpdate = 0;
  private int bndWdthT = 0;
  private int bndWdthC = 0;
  private int grpCc = 0;
  public int getNumberInGroup()
  {
    return lnks.size();
  }
  public int getNumberPending() 
  {
    if (lnks == null || lnks.isEmpty()) return 0;
    int npending = 0;
    Iterator<TLink> i = lnks.iterator();
    do
    {
      npending += ((TLink)i.next()).notifyPending ? 1 : 0;
    } while (i.hasNext());
    
    return npending; 
  }
  public void reset()
  {
    if (lnks == null || lnks.isEmpty()) return;
    TLink lnk;
    Iterator<TLink> i = lnks.iterator();
    do
    {
      lnk = (TLink)i.next();
      lnk.notifyPending  = true;
      lnk.needsNotification = false;
      lnk.dOutput.bytesin = 0;
      if (lnk.sub.mode <= TMode.CM_SINGLE) lnk.active = false;
      if (tlf != null) tlf.checkLinkRenewalCondition(lnk);
    } while (i.hasNext());
  }
  private int shuffle = 0;
  public boolean canNotify(TLink lnk)
  {
    if (lnk == null || lnk.sub == null) return false;
    int numPending = getNumberPending();
    int numInGrp = lnks.size();
    long tv = System.currentTimeMillis();
    long tvg = 0;
    if (lnk.getGroup() == null) return false;
    if (numPending > 0 && lnk.needsNotification) 
    { // has already contributed
      return false; 
    }
    lnk.needsNotification = true;
    if (numPending == numInGrp-1)
    { // initial callback in this set
      bndWdthC = 0;
      grpCc = lnk.linkStatus;
      if (syncLevel <= GRP_SYNC_STARTSYNC)
      { // absolute initial callback
        lnkHead = lnk;
        updateInterval = lnk.sub.pollingInterval;
      }
    }
    if (grpCc == 0) grpCc |= lnk.linkStatus;
    int cycstamp = lnk.dOutput.sysDataStamp;
    if (lnkHead == lnk)
    { // peg all of this on the group master
      lastUpdate = timeUpdate;
      timeUpdate = tv;
      lastStamp = sysStamp;
      sysStamp = cycstamp;
    }
    int latency = (int)(tv - timeUpdate);
    int offset = cycstamp - sysStamp;
    lnk.dOutput.setSynchronizationOffset(offset);
    TLink cte = null;
    if (syncLevel > GRP_SYNC_CYCLEOFFSET)
    {
      if (lnk != lnkHead)
      { // not the group head 
        if ((cte=lnkHead) != null)
        {
          if (cte.notifyPending)
          { /* members updating but group head is still pending */
            if (incMono)
            { // redefine the group head
              if (cycstamp <= sysStamp)
              { // skip this extraneous update
                lnk.notifyPending = true;
                lnk.needsNotification = false;
                return false;
              }
              lnkHead = lnk;
              //goto chkGroupHead;
              lastUpdate = timeUpdate;
              timeUpdate = tv;
              cycstamp = lnk.dOutput != null ? lnk.dOutput.sysDataStamp : 0;
              sysStamp = cycstamp;
            }
            else if (latency < updateInterval - GRP_JITTER)
            { // reject this: it'c coming in too fast (scheduled ?)
              lnk.notifyPending = true;
              lnk.needsNotification = false;
              return false;
            }
            offset = latency = 0; // as yet, no group head notification !
          }
          else if (cycstamp > sysStamp) 
          { // could be advancing underneath (e.g. scheduled property)
            tvg = cte.dOutput.dTimestamp;
            sysStamp = cte.dOutput.sysDataStamp;
            timeUpdate = tvg;
          }
        }
      }
    }
    lnk.setGroupLatency(latency);
    TLink[] mems = getMembers();
    if (syncLevel == GRP_SYNC_NOTIFICATION)
    { /* find the best group order */
      int pi = lnk.getLinkPollingInterval();
      if (cycstamp > sysStamp && latency > pi/2)
      { /* re-group ! */
        MsgLog.log("TGroupLink.canNotify","group synchronization: redefine group order",0,null,1);
        lnkHead = lnk;
        updateInterval = pi;
        numPending = numInGrp;
        timeUpdate = tv;
        sysStamp = cycstamp;
        lnk.setGroupLatency(latency);
        syncLevel = GRP_SYNC_CYCLEOFFSET;
        for (TLink tlnk : mems)
        {
          if (tlnk == lnk) continue;
          tlnk.notifyPending = true;
        }
      }
    }
    if (numPending > 0) return false;
    /* now they're all in ... */
    latency = 0; offset = 0; 
    int latency_floor = 0;
    int offset_floor = 0;
    if ((cte=lnkHead) != null)
    { /* re-sync these values in case a straggler held us up */
      tvg = cte.dOutput.dTimestamp;
      sysStamp = cte.dOutput.sysDataStamp;
      timeUpdate = tvg;
    }
    int grpLatency = 0;
    int syncOffset = 0;
    for (TLink tlnk : mems)
    { /* find the real offsets */
      if (tlnk == lnkHead) continue;
      tvg = tlnk.dOutput.getDataTimeStamp();
      tlnk.dOutput.setSynchronizationOffset(tlnk.dOutput.sysDataStamp - sysStamp);
      if (incMono && tlnk.getSynchronizationOffset() < 0)
      { // grp is monotonically incrementing but some members are late : re-order !
        if (shuffle == 0)
        { // but only do this every so often !
          tlnk.notifyPending = true;
          numPending++;
        }
        else
        {
          shuffle++;
        }
      }
      grpLatency = (int)(tvg - timeUpdate);
      tlnk.setGroupLatency(grpLatency);
      syncOffset = tlnk.getSynchronizationOffset();
      if (grpLatency < latency_floor) latency_floor = grpLatency;
      if (grpLatency > latency) latency = grpLatency;
      if (syncOffset < offset_floor) offset_floor = syncOffset;
      if (syncOffset > offset) offset = syncOffset;
    }
    if (numPending > 0) 
    { // group reshuffling ?
      shuffle++; // will prevent numPending from increasing pathologically
      MsgLog.log("TGroupLink.canNotify","re-shuffle group due to "+numPending+" late arrivals",0,null,1);
      return false;
    }
    if (shuffle > mems.length * 10) 
    {
      MsgLog.log("TGroupLink.canNotify","shuffle count : "+shuffle+" -> allow group re-shuffling to continue",0,null,1);
      shuffle = 0;
    }
    
    bndWdthT = latency - latency_floor;
    bndWdthC = offset - offset_floor;
    switch (syncLevel)
    {
      case GRP_SYNC_STARTSYNC:
        syncLevel = GRP_SYNC_NOTIFICATION;
        MsgLog.log("TGroupLink.canNotify","group synchronization: find most effective group order",0,null,1);
        break;
      case GRP_SYNC_NOTIFICATION:
        syncLevel = GRP_SYNC_CYCLEOFFSET;
        break;
      case GRP_SYNC_CYCLEOFFSET:
        syncLevel = GRP_SYNC_INTERVAL;
        break;
      case GRP_SYNC_INTERVAL:
        syncLevel = GRP_SYNC_INSYNC;
        MsgLog.log("TGroupLink.canNotify","group synchronization: mark as synchronized",0,null,1);
      case GRP_SYNC_INSYNC:
        if ((cte=lnkHead) != null)
        {
          tvg = cte.dOutput.dTimestamp;
          int upd = (int)(tvg - lastUpdate) + GRP_JITTER - 1;
          upd -= (upd%GRP_JITTER);
          if (upd < updateInterval)
          {
            updateInterval = upd;
            MsgLog.log("TGroupLink.canNotify","group synchronization: fix effective interval to "+updateInterval+" msec",0,null,1);
          }
        }
        break;
    }
    incMono = (sysStamp - lastStamp == 1);
    reset();
    return true;
  }
  public TLinkGroup()
  {
    lnks = new ConcurrentLinkedQueue<TLink>();
    tlf = TLinkFactory.getInstance();
  }
  public TLink[] getMembers()
  {
    return (TLink[])lnks.toArray(new TLink[0]);
  }
  public int addMember(TLink lnk)
  {
    if (lnk.tcb != null)
    {
      if (tcb == null) tcb = lnk.tcb;
      if (lnk.tcb != tcb) return TErrorList.not_owner; 
    }
    if (lnk.tlcb != null)
    {
      if (tlcb == null) tlcb = lnk.tlcb;
      if (lnk.tlcb != tlcb) return TErrorList.not_owner; 
    }
    MsgLog.log("TLinkGroup","add "+lnk.getFullDeviceNameAndProperty()+" to group",0,null,1);
    lnks.add(lnk);
    return 0;
  }
  public int removeMember(TLink lnk)
  {
    if (lnk.tcb != null)
    {
      if (tcb == null) tcb = lnk.tcb;
      if (lnk.tcb != tcb) return TErrorList.not_owner; 
    }
    if (lnk.tlcb != null)
    {
      if (tlcb == null) tlcb = lnk.tlcb;
      if (lnk.tlcb != tlcb) return TErrorList.not_owner; 
    }
    MsgLog.log("TLinkGroup","remove "+lnk.getFullDeviceNameAndProperty()+" from group",0,null,1);
    lnks.remove(lnk);
    return 0;    
  }
  public String toString()
  {
    String s = "Group members :";
    String tag;
    try
    {
    for (TLink lnk : getMembers())
    {
      tag = lnk.getFullDeviceNameAndProperty()+" + "+lnk.getSynchronizationOffset()+" cnts";
      if (lnkHead == lnk) tag += " (*head*)";
      s += "\n"+tag;
    }
    s +=  "\n number in group : "+getNumberInGroup()+"\n"+
          " number pending : "+getNumberPending()+"\n"+
          " current group cycle stamp : "+getSysStamp()+"\n"+
          " last group cycle stamp : "+getLastStamp()+"\n"+
          " current group cycle dispersion : "+bndWdthC+" counts\n"+
          " current group time dispersion : "+bndWdthT+" msec\n"+
          " current group synchronization : "+grpSyncLevels[getSynchronizationLevel()]+"\n"+
          " effective group update interval : "+getUpdateInterval()+" msec\n"+
          " group updating monotonically : "+(incMono ? "TRUE" : "FALSE")+"\n"+
          " most recent update : "+TDataTime.toString(timeUpdate)+"\n"+
          " current group status code : "+grpCc;
    }
    catch (Exception any)
    {
      s = any.getMessage();
    }
    return s;
  }
}
