/*
 * File TWildcardSingleCallback.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

/*
 * Created on Jul 12, 2006
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package de.desy.tine.client;

import de.desy.tine.dataUtils.TDataType;
import de.desy.tine.definitions.TErrorList;
import de.desy.tine.definitions.TFormat;
import de.desy.tine.definitions.TMode;
import de.desy.tine.queryUtils.TQuery;
import de.desy.tine.server.logger.DbgLog;

public class TWildcardSingleCallback implements TLinkCallback
{
  public void callback(TLink link)
  {
    TWildcardLink wc = link.twcl;
    if (wc == null || link.devName == null) return; // get out of Dodge as quickly as possible
    String srv, dev;
    int fmt = TFormat.CF_NULL, mode, i, n, cc = 0;
    int debug = TLinkFactory.debugLevel;
    boolean isWcGroup = false;
    switch (link.getLinkStatus())
    {
      case TErrorList.illegal_property:
        /* will be returned by the group server */
      case TErrorList.data_not_local:
        /* mark this entry as CM_REGISTER */
        isWcGroup = true;
        if (debug > 1)
          DbgLog.log("TWildcardSingleCallback", link.getFullDeviceName()+" is redirecting to another source");
        mode = link.sub.mode;
        link.sub.mode = TMode.CM_REGISTER;
        /* get the device list and length and allocate the relevant arrays */
        wc.list = link.devName.compareTo("*") == 0 ? 
            TQuery.getDeviceNames(link.cntName,link.expName,null) :
            TQuery.getDeviceNames(link.cntName,link.expName,null,link.devName);
        if (wc.list == null) 
        {
          cc = TErrorList.connection_timeout;
          break;
        }
        /* now allocate the device tables */
        n = wc.list.length;
        wc.length = n;
        wc.status = new int[n];
        wc.links = new TLink[n];
        /* n is the total number, so this is a worst case allocation ... */
        switch (wc.format)
        {
          case TFormat.CF_USTRING:
          case TFormat.CF_NAME16DBLDBL:
          case TFormat.CF_NAME32DBLDBL:
          case TFormat.CF_NAME64DBLDBL:
          case TFormat.CF_NAME16FI:
            fmt = TFormat.CF_FLOAT;
            break;
          case TFormat.CF_NAME16II:
            fmt = TFormat.CF_INT32;
            break;
          default:
            if (wc.format > TFormat.CF_NAME32 && link.devName.compareTo("*") != 0) 
            {
              cc = TErrorList.illegal_format;
              break;
            }
            fmt = wc.format;
            // data reference ????
            break;
        }
        /* if the wildcard contains characters, then n returned here is likely to be smaller than above */
        wc.numPending = wc.length;
        
        TDataType din = link.getInputDataObject();
        srv = link.cntName.length() > 0 ? "/" + link.cntName + "/" : "";
        srv += link.expName;
        for (i=0; i<wc.length; i++)
        { /* start the individual links (these will be redirected by the source) */
          TDataType dout = new TDataType(1,(short)fmt);
          dev = srv + "/" + wc.list[i];
          if (debug > 1)
            DbgLog.log("TWildcardSingleCallback", "launch link to "+dev);
          wc.links[i] = new TLink(dev,link.devProperty,dout,din,link.devAccess);
          wc.links[i].callbackId = i;
          wc.links[i].twcl = wc; // pass it on
          wc.links[i].linkStatus = wc.status[i] = TErrorList.not_ready;
          wc.links[i].delayEstablishLink(true);
          wc.links[i].attach(mode,wc.gcb,link.sub.pollingInterval);
        }
        break;
      default:
        if (wc.tlcb != null) wc.tlcb.callback(link);
        else if (wc.tcb != null) wc.tcb.callback(link.linkId,link.linkStatus);
        else wc.parent.wakeUpCall();
        wc.canNotify = true;
        wc.checkActive();
        break;
    }
    if (cc != 0)
    {
      link.linkStatus = cc;
      link.linkStatusSource = TLink.STATUS_LOCAL;
      if (wc.tlcb != null) wc.tlcb.callback(link);
      else if (wc.tcb != null) wc.tcb.callback(link.linkId,cc);
      else if (!isWcGroup) wc.parent.wakeUpCall();
      link.close();
      wc.canNotify = true;
      wc.active = false;
      wc.mode = TMode.CM_CANCEL;
    }
    return;
  }
}
