/*
 * File TineClient.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

/*
 * The main contributor to this project is Institute of Materials Research,
 * Helmholtz-Zentrum Geesthacht,
 * Germany.
 *
 * This project is a contribution of the Helmholtz Association Centres and
 * Technische Universitaet Muenchen to the ESS Design Update Phase.
 *
 * The project's funding reference is FKZ05E11CG1.
 *
 * Copyright (c) 2012. Institute of Materials Research,
 * Helmholtz-Zentrum Geesthacht,
 * Germany.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

package de.desy.tine.client.ez;

import de.desy.tine.client.TCallback;
import de.desy.tine.client.TLink;
import de.desy.tine.dataUtils.TDataType;
import de.desy.tine.definitions.TAccess;
import de.desy.tine.definitions.TErrorList;
import de.desy.tine.definitions.TMode;
import de.desy.tine.queryUtils.TPropertyQuery;
import de.desy.tine.queryUtils.TQuery;

import java.lang.reflect.Array;
import java.sql.Timestamp;
import java.util.AbstractMap;
import java.util.Iterator;
import java.util.Map;
import java.util.concurrent.*;

/**
 * This instance stores a number of {@link de.desy.tine.client.TLink} and reads data from them.
 *
 * //@ThreadSafe
 *
 * @author Igor Khokhriakov <igor.khokhriakov@hzg.de>
 * @since 27.04.12
 */
public class TineClient {
    private final String url;
    private final String context;
    private final String serverName;
    private final String deviceName;
    private ConcurrentMap<String, Future<TLink>> tlinks = new ConcurrentHashMap<String, Future<TLink>>();

    public TineClient(String url) {
        this.url = url;
        String[] deviceInfo = url.split("/");
        this.context = deviceInfo[1];//skip leading '/'
        this.serverName = deviceInfo[2];
        this.deviceName = deviceInfo[3];
    }

    /**
     * Reads an attribute from this TINE server
     * <p/>
     * Implementation stores {@link de.desy.tine.client.TLink} instances wrapped by {@link java.util.concurrent.FutureTask} in private {@link this#tlinks} field.
     * This is done to prevent multiply dummy TLink instances creation.
     *
     * @param attrName
     * @return
     * @throws TineClientException
     */
    public <T> Map.Entry<T, Timestamp> readAttribute(final String attrName) throws TineClientException {
        Future<TLink> futureLink = getFutureLink(attrName);
        TLink tLink = null;
        try {
            tLink = futureLink.get();
        } catch (InterruptedException e) {
            Thread.currentThread().interrupt();
            throw new TineClientException("Can not get TLink for attribute " + attrName, e);
        } catch (Exception e) {
            throw new TineClientException("Can not get TLink for attribute " + attrName, e);
        }
        int rc = tLink.execute();
        if (rc != TErrorList.success /*TODO handle -1 rc*/ && rc != -1) {
            throw new TineClientException("Can not execute TLink: " + TErrorList.getErrorString(rc), rc);
        }
        TDataType dout = tLink.dOutput;
        long time = tLink.getLastTimeStamp();
        Timestamp timestamp = new Timestamp(time);
        return new AbstractMap.SimpleImmutableEntry<T, Timestamp>((T) getDataObject(dout), timestamp);
    }

    /**
     * Writes a value to an attribute of this TINE server
     * <p/>
     * NOT YET IMPLEMENTED!!!
     *
     * @param attrName
     * @param value
     * @param <T>
     * @throws TineClientException
     */
    public <T> void writeAttribute(String attrName, T value) throws TineClientException {
        throw new UnsupportedOperationException("Writing is not yet implemented");
    }

    private Future<TLink> getFutureLink(final String attrName) {
        Future<TLink> futureLink = tlinks.get(attrName);
        if (futureLink == null) {
            FutureTask<TLink> ft =
                    new FutureTask<TLink>(
                            new Callable<TLink>() {
                                public TLink call() throws TineClientException {
                                    TPropertyQuery meta = getTPropertyQuery(attrName);
                                    int size = getTPropertySize(meta);
                                    short dataFormat = getTPropertyFormat(meta);
                                    TDataType dout = new TDataType(size, dataFormat);
                                    // get a reference array : synchronous call ...
                                    TLink result = new TLink(url, attrName, dout, null, TAccess.CA_READ);
                                    return result;
                                }
                            });


            futureLink = tlinks.putIfAbsent(attrName, ft);
            if (futureLink == null) {
                ft.run();
                futureLink = ft;
            }
        }
        return futureLink;
    }

    /**
     * @param dout
     * @return Object or Object[]
     */
    private <T> T getDataObject(TDataType dout) {
//        int length = Array.getLength(dout.getDataObject());
        Object dataObject = dout.getDataObject();
        //TODO it appears that TINE always returns array of values
        //TODO and the only interesting value is always the first one in this array
        //if (length > 1) {
        //    return dataObject;
        //} else {
        return (T) Array.get(dataObject, 0);
        //}
    }

    /**
     * Subscribes to a change event from an attribute of this TINE server
     *
     * @param attrName
     * @param cbk
     * @param <T>
     * @throws TineClientException
     */
    public <T> void subscribeEvent(final String attrName, final EventCallback<T> cbk) throws TineClientException {
        Future<TLink> futureLink = getFutureLink(attrName);
        try {
            final TLink link = futureLink.get();
            final TDataType dout = link.dOutput;
            long time = link.getLastTimeStamp();
            //read data for the first time
            cbk.onEvent(new AbstractMap.SimpleImmutableEntry<T, Timestamp>(this.<T>getDataObject(dout), new Timestamp(time)));
            //attach event listener
            int rc = link.attach(TMode.CM_DATACHANGE, new TCallback() {
                public void callback(int LinkIndex, int LinkStatus) {
                    if (LinkStatus == TErrorList.success) {
                        long time = link.getLastTimeStamp();
                        cbk.onEvent(new AbstractMap.SimpleImmutableEntry<T, Timestamp>(TineClient.this.<T>getDataObject(dout), new Timestamp(time)));
                    } else {
                        cbk.onError(new TineClientException(TErrorList.getErrorString(LinkStatus), LinkStatus));
                    }
                }
            });
            if (rc < 0) {
                throw new TineClientException(TErrorList.getErrorString(rc), rc);
            }
        } catch (InterruptedException e) {
            Thread.currentThread().interrupt();
            throw new TineClientException("Can not get TLink for attribute " + attrName, e);
        } catch (ExecutionException e) {
            throw new TineClientException("Can not get TLink for attribute " + attrName, e);
        }
    }

    /**
     * Checks that this TINE server has an attribute
     *
     * @param attrName
     * @return true or false
     */
    public boolean hasAttribute(String attrName) {
        try {
            getTPropertyQuery(attrName);
            return true;
        } catch (TineClientException e) {
            return false;
        }
    }

    private int getTPropertySize(TPropertyQuery meta) {
        int result = meta.prpSize;
        return result;
    }

    private short getTPropertyFormat(TPropertyQuery meta) {
        short result = meta.prpFormat;
        return result;
    }

    /**
     * Returns attribute's value class
     *
     * @param attrName
     * @return
     * @throws TineClientException
     */
    public Class<?> getAttributeClass(String attrName) throws TineClientException {
        TPropertyQuery meta = getTPropertyQuery(attrName);
        int size = getTPropertySize(meta);
        short dataFormat = getTPropertyFormat(meta);
        TDataType dataType = new TDataType(size, dataFormat);
        //TODO it appears that TINE always returns array of values
        //TODO and the only interesting value is always the first one in this array
        //if (size > 1) {
//            return dataType.getDataObject().getClass();
        //      } else {
        return dataType.getDataObject().getClass().getComponentType();
        //    }
    }

    /**
     * Unsubscribes from the change event of the attribute
     *
     * @param attrName
     * @throws TineClientException
     */
    public void unsubscribeEvent(String attrName) throws TineClientException {
        Future<TLink> futureLink = tlinks.remove(attrName);
        try {
            final TLink link = futureLink.get();
            int rc = link.close();
            if (rc < 0) {
                throw new TineClientException(link.getLastError(), null);
            }
        } catch (Exception e) {
            throw new TineClientException("Failed subscribe to " + url + "/" + attrName, e);
        }
    }

    private TPropertyQuery getTPropertyQuery(String attrName) throws TineClientException {
        TPropertyQuery[] result = TQuery.getPropertyInformation(context, serverName, deviceName, attrName);
        if (result == null) {
            throw new TineClientException("Cannot read meta info for " + url + "/" + attrName, new NullPointerException());
        }
        return result[0];
    }

    public void close() {
        boolean interrupt = false;
        for (Iterator<Future<TLink>> iterator = tlinks.values().iterator(); iterator.hasNext(); ) {
            Future<TLink> link = iterator.next();
            try {
                link.get().close();
                iterator.remove();
            } catch (InterruptedException e) {
                interrupt = true;
            } catch (ExecutionException ignored) {
                //TODO notify client
            }
        }
        if (interrupt) Thread.currentThread().interrupt();
    }
}
