/*
 * File csv.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.csvUtils;
import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import de.desy.tine.definitions.TErrorList;
import de.desy.tine.server.logger.MsgLog;

public final class csv
{
  private char[] inchars = null;
  private int inpos = 0;
  private BufferedReader in;
  private String fileName;
  public String getFileName() { return fileName; }
  public csv(BufferedReader reader)
  {
    in = reader;
  }
  public boolean fileAvailable() { return in == null ? false : true; }
  public csv(String csvFile)
  {
    fileName = csvFile;
    try
    {
      in = new BufferedReader(
        new InputStreamReader(new FileInputStream(csvFile)));
    }
    catch (FileNotFoundException e) 
    { // lots of files aren't there -> don't worry about this here!
    }      
  }
  public csv(char[] csvChars)
  {
    inchars = csvChars;
  }
  public void mark()
  {
    try 
    {
      if (in != null) in.mark(1024*8);
    }
    catch (IOException e)
    { // did the best we could ...
    }    
  }
  public void rewind()
  {
    try 
    {
      if (in != null) in.reset();
      if (inchars != null) inpos = 0;
    }
    catch (IOException e)
    { // did the best we could ...
    }
  }
  public String readLine()
  {
    if (inchars != null)
    {
      StringBuffer sb = new StringBuffer(256);
      char c;
      while (inpos < inchars.length && (c=inchars[inpos++]) != '\n') 
        sb.append(c);
      return inpos < inchars.length ? sb.toString() : null;
    }
    if (in == null) return null;
    try 
    {
      return in.readLine();
    }
    catch (IOException e)
    { // End of File is normal ?...
      return null;
    }
  }
  public int readFile(csvColumn[] csvCols)
  {
    return readFile(csvCols,null);
  }
  public int readFile(csvColumn[] csvCols, RowHandler csvRowHndlr)
  {
    boolean headerRead = false, haltIteration = false;
    String s, strValue;
    int i = 0, idx = 0;
    if (in == null && inchars == null) return TErrorList.no_such_file;
    try
    {
      while((s=readLine()) != null)
      {
        if (s.length() == 0) continue;
        if (s.startsWith("#") || s.startsWith(";") || s.startsWith("%")) continue;
        if (!headerRead)
        {
          for (i=0; i<csvCols.length; i++)
          {
            if (csvCols[i] == null) continue;
            csvCols[i].setColumnPosition(findcol(s,csvCols[i].getColumnName()));
            if (csvCols[i].getColumnPosition() == -1 && csvCols[i].getColumnDefaultValue() == null)
              throw new NoSuchFieldException();
          }
          headerRead = true;
          continue;
        }
        for (i=0; i<csvCols.length; i++)
        {
          if (csvCols[i] == null) continue;
          strValue = csvCols[i].getColumnDefaultValue();
          if (csvCols[i].getColumnPosition() != -1)
            strValue = colptr(csvCols[i].getColumnPosition(),s);
          if (csvCols[i].getColumnHandler().process(strValue.trim(),idx) != 0)
          { // any individual field handler can halt the iteration
            haltIteration = true;
          }
        }
        if (haltIteration) break;
        if (csvRowHndlr != null) 
        { // halt iteration on non-zero return code
          if (csvRowHndlr.process(idx) != 0) break;
        }
        idx++;
      }
    }
    catch (NoSuchFieldException e)
    { // database corrupt ?
      MsgLog.log("csv readFile","looking for " + csvCols[i].getColumnName() + " in corrupt file !",TErrorList.no_such_column,e,0);
      return TErrorList.no_such_column;
    }
    return 0;
  }
  public void close()
  {
    try 
    {
      if (in != null) in.close();
    }
    catch (IOException e)
    {
      MsgLog.log("csv readFile","Cannot close file!",TErrorList.file_error,e,0);
      System.out.println("Cannot close file !");
      e.printStackTrace();
    }
  }  
  public int findcol(String line,String ColumnEntry)
  {
    String subs;
    String s = line.toUpperCase();
    String e = ColumnEntry.toUpperCase();
    String altEntry1=null, altEntry2=null;
    String[] altPieces = null;
    char brkc = 0, altc = 0;
    int i, n, col;

    if (e.indexOf(' ') > 0)
    {
      brkc = ' ';
      altc = '_';
    }
    else if (e.indexOf('_') > 0) 
    {
      brkc = '_';
      altc = ' ';
    }
    if (brkc != 0) /* prepare alternates */
    {
      altEntry1 = e.replace(brkc,altc);
      altPieces = e.split(String.valueOf(brkc));
      altEntry2 = new String(altPieces[0]);
      for (i = 1; i<altPieces.length; i++) altEntry2 = altEntry2.concat(altPieces[i]);
    }    
    for (i=0, col=0; i < s.length(); col++)
    {
      if ((n = s.indexOf(',',i)) < 0) // not found
      {
        n = s.length();
      }
      subs = s.substring(i,n).trim();
      if (subs.compareTo(e) == 0) return col;
      if (brkc != 0) 
      {
        if (subs.compareTo(altEntry1) == 0) return col;
        if (subs.compareTo(altEntry2) == 0) return col;
      }
      if ((i = n + 1) >= s.length()) return -1;	  
    }
    return -1;
  }  
  public int namcmp(String targetString,String searchString,int column)
  {
    String s = colptr(column,searchString);
    if (s == null) return -1;
    return s.compareToIgnoreCase(targetString);// p.d. (21.12.11) changed to ignoreCase
  }  
  public String colptr(int column,String line)
  {
    String subs;
    String s = line; //line.toUpperCase();
    int i, n, col,slen;
    slen = s.length();
    for (i=0, col=0; i < slen; col++)
    {
      while (i<slen && (s.charAt(i) == ' ' || s.charAt(i) ==  '\t')) i++;
      if (i == slen) break;
      if (s.charAt(i) == '"') 
      {
        n = s.indexOf('"',i+1); 
        if (n < 0 || n >= slen) break;
        subs = s.substring(i+1,n).trim();
        i = n;
        n = s.indexOf(',',i);
      }
      else
      {
        if ((n=s.indexOf(',',i)) == -1) n = slen; //break;
        subs = s.substring(i,n).trim();
      }
      if (col == column) return subs;
      if ((i = n + 1) >= slen) break;
    }
    return "";
  }  
  public static float getFloatFromString(String s)
  {
    float fval = 0;
    try
    {
      fval = Float.parseFloat(s);
    }
    catch (NumberFormatException e)
    {
      fval = 0;
    }
    return fval;
  }
  public static int getIntFromString(String s)
  {
    int ival = 0;
    try
    {
      if (s.startsWith("0x") || s.startsWith("0X"))
      {
        ival = Integer.parseInt(s.substring(2),16);
      }
      else
      {
        ival = Integer.parseInt(s);
      }
    }
    catch (NumberFormatException e)
    {
      ival = 0;
    }
    return ival;
  }
}