/*
 * File UnsignedUtils.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.dataUtils;

/**
 * Utility class to perform common unsigned conversion algorithms.
 * 
 * @author David Melkumyan, DESY Zeuthen
 * @since 08.2009
 */
public class UnsignedUtils
{
  /**
   * Converts the specified int into an unsigned long.<br>
   * The least significant int (32 bits) of the long will be identical to the
   * int (32 bits) provided, and the most significant int (32 bits) of the long
   * will be zero.
   * 
   * @param input
   *          - the int to convert.
   * @return An unsigned long representing the specified int.
   */
  public static long toUnsignedInt(int input)
  {
    return input & 0xFFFFFFFFL;
  }
  /**
   * Converts the specified short into an unsigned integer.<br>
   * The least significant short (16 bits) of the integer will be identical to
   * the short (16 bits) provided, and the most significant 2 bytes (16 bits) of
   * the integer will be zero.
   * 
   * @param input
   *          - the short to convert.
   * @return An unsigned int representing the specified short.
   */
  public static int toUnsignedShort(short input)
  {
    return input & 0xFFFF;
  }
  /**
   * Converts the specified byte into an unsigned short.<br>
   * The least significant byte (8 bits) of the short will be identical to the
   * byte (8 bits) provided, and the most significant byte (8 bits) of the short
   * will be zero.
   * 
   * @param input
   *          - the byte to convert.
   * @return An unsigned short representing the specified byte.
   */
  public static short toUnsignedByte(byte input)
  {
    return (short) (input & 0xFF);
  }
}
