/*
 * File TArrayType.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.definitions;

/**
 * Data set returned by a call to a property can
 * specify the type of array it is by registering this information.
 * 
 * @author duval
 */
public final class TArrayType
{
  public static final short AT_UNKNOWN = 0x00;
  public static final short AT_SCALAR = 0x01;
  public static final short AT_SINGLE = 0x02;
  public static final short AT_DOUBLE = 0x04;
  public static final short AT_BOOLEAN = 0x08;
  public static final short AT_CHANNEL = 0x10;
  public static final short AT_SPECTRUM = 0x20;
  public static final short AT_TRACE = 0x20;
  public static final short AT_WAVEFORM = 0x20;
  public static final short AT_STRUCTURED = 0x40;
  public static final short AT_COLLECTION = 0x40;
  public static final short AT_UNSTRUCTURED = 0x80;
  public static final short AT_BLOB = 0x80;
  public static final short AT_IMAGE = 0x80;
  public static final short AT_JPEG = 0x100;
  public static final short AT_MPEG = 0x200;
  public static final short AT_BMP = 0x400;
  public static final short AT_GIF = 0x800;
  public static final short AT_WAV = 0x1000;
  public static final short AT_MP3 = 0x2000;
  /** 
   * Returns the string representation of the input array type
   * @param type is the array type for which the string representation is desired
   * @return the string representation of the input array type
   */
  public static String toString(short type)
  {
    String add = "";
    String bool = "";
    if ((type & 0x08) == 0x08) bool = "boolean ";
    if ((type & 0x10) == 0x10) add = "channel ";
    if ((type & 0x20) == 0x20) add = "spectrum ";
    if ((type & 0x40) == 0x40) add = "collection ";
    if ((type & 0x48) == 0x80) add = "unstructured ";
    switch (type & 0x07)
    {
      default:
      case AT_UNKNOWN:
        return "unknown";
      case AT_SCALAR:
        return "scalar" + bool;
      case AT_SINGLE:
        return "single " + bool + add + "array";
      case AT_DOUBLE:
        return "double " + bool + add + "array";
    }
  }
  /**
   * @param typeString is an array type string 
   * @return the numerical form of the array type
   */
  public static short getType(String typeString)
  {
    short at = AT_UNKNOWN;
    if (typeString != null)
    {
      String ts = typeString.trim();
      if (ts.compareToIgnoreCase("SCALAR") == 0)
        at |= AT_SCALAR;
      else if (ts.compareToIgnoreCase("SINGLE") == 0)
        at |= AT_SINGLE;
      else if (ts.compareToIgnoreCase("DOUBLE") == 0) at |= AT_DOUBLE;
      if (ts.compareToIgnoreCase("BOOLEAN") == 0) at |= AT_BOOLEAN;
      if (ts.compareToIgnoreCase("CHANNEL") == 0)
        at |= AT_CHANNEL;
      else if (ts.compareToIgnoreCase("SPECTRUM") == 0 || ts.compareToIgnoreCase("TRACE") == 0
          || ts.compareToIgnoreCase("WAVEFORM") == 0)
        at |= AT_TRACE;
      else if (ts.compareToIgnoreCase("STRUCTURED") == 0 || ts.compareToIgnoreCase("COLLECTION") == 0)
        at |= AT_COLLECTION;
      else if (ts.compareToIgnoreCase("UNSTRUCTURED") == 0 || ts.compareToIgnoreCase("BLOB") == 0)
        at |= AT_BLOB;
      else if (ts.compareToIgnoreCase("IMAGE") == 0) at |= AT_IMAGE;
      if (ts.compareToIgnoreCase("JPEG") == 0) at |= AT_JPEG;
      if (ts.compareToIgnoreCase("MPEG") == 0) at |= AT_MPEG;
      if (ts.compareToIgnoreCase("BMP") == 0) at |= AT_BMP;
      if (ts.compareToIgnoreCase("GIF") == 0) at |= AT_GIF;
      if (ts.compareToIgnoreCase("WAV") == 0) at |= AT_WAV;
      if (ts.compareToIgnoreCase("MP3") == 0) at |= AT_MP3;
    }
    return at;
  }
  /**
   * Return new instance of TArrayType object.
   * 
   * @param val
   *          raw type value
   * @return new instance
   */
  static public TArrayType valueOf(short val)
  {
    return new TArrayType(val);
  }
  private short value;
  private TArrayType(short val)
  {
    value = val;
  }
  /** 
   * Returns the string representation of this array type
   * @return the string representation of this array type
   */
  public String toString()
  {
    return toString(value);
  }
  /**
   * Returns the value.
   * 
   * @return Returns the value.
   */
  public short getValue()
  {
    return value;
  }
  /**
   * Return <code>true</code> if this array type presents multi-channel array.
   * 
   * @return <code>true</code> if this array type presents multi-channel array.
   */
  public boolean isChannel()
  {
    return (value & AT_CHANNEL) == AT_CHANNEL;
  }
  /**
   * Return <code>true</code> if input arrayType returns a multi-channel array
   * 
   * @param arrayType is the targeted array type.
   * 
   * @return <code>true</code> if input arrayType returns a multi-channel array
   */
  public static boolean isChannel(int arrayType)
  {
    return (arrayType & AT_CHANNEL) == AT_CHANNEL;
  }
  /**
   * Return <code>true</code> if this array type presents scalar value.
   * 
   * @return <code>true</code> if this array type presents scalar value.
   */
  public boolean isScalar()
  {
    return (value & AT_SCALAR) == AT_SCALAR;
  }
  /**
   * Returns <code>true</code> if the array type presents scalar value.
   * @param arrayType is the array type to be examined
   * @return <code>true</code> if the array type presents scalar value.
   */
  public static boolean isScalar(int arrayType)
  {
    return (arrayType & AT_SCALAR) == AT_SCALAR;
  }
  /**
   * Return <code>true</code> if property returns a single dimensioned array
   * (i.e. a vector).
   * 
   * @return <code>true</code> if property returns a single dimensioned array
   *         (i.e. a vector).
   */
  public boolean isVector()
  {
    return (value & AT_CHANNEL) == AT_CHANNEL;
  }
  /**
   * Return <code>true</code> if input arrayType returns a single dimensioned array
   * (i.e. a vector).
   * 
   * @param arrayType is the targeted array type.
   * 
   * @return <code>true</code> if input arrayType returns a single dimensioned array
   *         (i.e. a vector).
   */
  public static boolean isVector(int arrayType)
  {
    return (arrayType & AT_CHANNEL) == AT_CHANNEL;
  }
  /**
   * Return <code>true</code> if property returns a trace array
   * 
   * @return <code>true</code> if property returns a trace array
   */
  public boolean isTrace()
  {
    return (value & AT_TRACE) == AT_TRACE;
  }
  /**
   * Return <code>true</code> if input arrayType returns a trace array
   * 
   * @param arrayType is the targeted array type.
   * 
   * @return <code>true</code> if input arrayType returns a trace array
   */
  public static boolean isTrace(int arrayType)
  {
    return (arrayType & AT_TRACE) == AT_TRACE;
  }
  /**
   * Return <code>true</code> if property returns a double dimensioned array
   * (a matrix).
   * 
   * @return <code>true</code> if property returns a double dimensioned array
   *         (a matrix).
   */
  public boolean isMatrix()
  {
    return (value & AT_DOUBLE) == AT_DOUBLE;
  }
  /**
   * Return <code>true</code> if input arrayType returns a double dimensioned array
   * (a matrix).
   * 
   * @param arrayType is the targeted array type.
   * 
   * @return <code>true</code> if input arrayType returns a double dimensioned array
   * (a matrix).
   */
  public static boolean isMatrix(int arrayType)
  {
    return (arrayType & AT_DOUBLE) == AT_DOUBLE;
  }
}
