/*
 * File TMode.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.definitions;
/**
 * TINE access modes and modifiers
 */
public class TMode
{
  public static final short CM_CANCEL = 0;
  public static final short CM_SINGLE = 1;
  public static final short CM_REFRESH = 2;
  public static final short CM_DATACHANGE = 2;
  public static final short CM_POLL = 3;
  public static final short CM_TIMER = 3;
  public static final short CM_REGISTER = 4;
  public static final short CM_GLOBAL = 5; 
  public static final short CM_RECEIVE = 5; 
  public static final short CM_EVENT = 6;  
  public static final short CM_NODUPLICATES = 0x10;
  public static final short CM_SEEKPATTERN = 0x20;
  public static final short CM_MISSPATTERN = 0x40;
  public static final short CM_STREAM = 0x80;
  public static final short CM_SYNCGROUP = 0x100;
  public static final short CM_NETWORK = 0x200;
  public static final short CM_BCAST = 0x200;
  public static final short CM_MCAST = 0x200;
  public static final short CM_USE_ON_ERROR = 0x400;
  public static final short CM_USEONERROR = 0x400;
  public static final short CM_RETRY = 0x800;
  public static final short CM_GROUPED = 0x1000;
  public static final short CM_CONNECT = 0x2000;
  public static final short CM_WAIT = 0x4000;

  /**
   * @param mode is the targeted transfer mode
   * @return true if the transfer mode is allowed to send a CM_CLOSE
   * signal to a server.
   */
  public static boolean canClose(int mode)
  {
    switch (getBaseMode(mode))
    {
      case CM_DATACHANGE:
      case CM_TIMER:
      case CM_EVENT:
        return true;
    }
    return false;    
  }
  /**
   * @param mode is the targeted transfer mode
   * @return true if the transfer mode can generate local timeout signals.
   */
  public static boolean canTimeOut(int mode)
  {
    switch (getBaseMode(mode))
    {
      case CM_SINGLE:
      case CM_DATACHANGE:
      case CM_TIMER:
      case CM_GLOBAL:
      case CM_EVENT:
        return true;
    }
    return false;
  }
  /**
   * The base mode for the specified mode parameter
   * 
   * @param mode is the mode parameter for which the base transfer mode 
   * (CM_SINGLE to CM_GLOBAL) is desired 
   * 
   * @return The base mode for the specified mode parameter
   */
  public static final short getBaseMode(int mode)
  {
    return (short)(mode & 0x0f);
  }
  /**
   * The string representation for the specified mode parameter
   * 
   * @param mode is the mode parameter for which the string representation 
   * is desired 
   * 
   * @return The string representation for the specified mode parameter
   */
  public static String toString(int mode)
  {
    String add = "";
    if ((mode & CM_SYNCGROUP) == CM_SYNCGROUP) add += " + synchronized"; 
    if ((mode & CM_NETWORK) == CM_NETWORK) add += " + network"; 
    if ((mode & CM_USE_ON_ERROR) == CM_USE_ON_ERROR) add += " + use on error";
    if ((mode & CM_RETRY) == CM_RETRY) add += " + retry";
    if ((mode & CM_GROUPED) == CM_GROUPED) add += " + grouped";
    if ((mode & CM_CONNECT) == CM_CONNECT) add += " + connected";
    if ((mode & CM_STREAM) == CM_STREAM) add += " + streamed";
    if ((mode & CM_WAIT) == CM_WAIT) add += " + wait";
    switch (mode & 0xff)
    {
      case CM_CANCEL: return "cancel" + add;
      case CM_SINGLE: return "single" + add; 
      case CM_DATACHANGE: return "datachange" + add;
      case CM_TIMER: return "timer" + add;
      case CM_REGISTER: return "register" + add;
      case CM_GLOBAL: return "receive" + add;  
      case CM_EVENT: return "event" + add;  
      default: return "unknown";     
    }
  }
  /**
   * @param mode is the targeted transfer mode
   * @return true if the transfer mode designates an active link
   */
  public static boolean isActive(int mode)
  {
    switch (getBaseMode(mode))
    {
      case CM_SINGLE:
      case CM_DATACHANGE:
      case CM_TIMER:
      case CM_GLOBAL:  
      case CM_EVENT:
        return true;
      default:
        break;
    }
    return false;
  }
  /**
   * @param mode is the targeted transfer mode
   * @return true if the transfer mode is using a connected tcp/ip or stream socket.
   */
  public static boolean isConnected(int mode)
  {
    return ((mode & (CM_CONNECT|CM_STREAM)) != 0);
  }
  /**
   * @param mode is the targeted transfer mode
   * @return true if the transfer mode is using a streaming tcp/ip socket.
   */
  public static boolean isStream(int mode)
  {
    return ((mode & CM_STREAM) == CM_STREAM);    
  }
  /**
   * @param mode is the targeted transfer mode
   * @return true if the transfer mode is using a connected tcp/ip socket.
   */
  public static boolean isTcpIp(int mode)
  {
    return ((mode & CM_CONNECT) == CM_CONNECT);    
  }
  /**
   * @param mode is the targeted transfer mode
   * @return true if the transfer mode is receiving network multicasts.
   */
  public static boolean isNetwork(int mode)
  {
    return ((mode & CM_NETWORK) == CM_NETWORK);    
  }
  /**
   * @param mode is the targeted transfer mode
   * @return true if the transfer mode is of a SINGLE type
   * @note this does not necessarily imply 'synchronous' transfer
   */
  public static boolean isSynchronous(int mode)
  {
    return ((mode&0x0f) == CM_SINGLE);    
  }
  /**
   * @param mode is the targeted transfer mode
   * @return true if the transfer mode is of a DATACHANGE type
   */
  public static boolean isDataChange(int mode)
  {
    return ((mode&0x0f) == CM_DATACHANGE);    
  }
  /**
   * @param mode is the targeted transfer mode
   * @return true if the transfer mode is of an EVENT type
   */
  public static boolean isEvent(int mode)
  {
    return ((mode&0x0f) == CM_EVENT);    
  }
  /**
   * @param mode is the targeted transfer mode
   * @return true if the transfer mode is of a TIMER type
   */
  public static boolean isTimer(int mode)
  {
    return ((mode&0x0f) == CM_TIMER);    
  }
  /**
   * @param mode is the targeted transfer mode
   * @return true if the transfer mode is merely CM_REGISTER
   */
  public static boolean isPlaceHolder(int mode)
  {
    return ((mode&0x0f) == CM_REGISTER);    
  }
  /**
   * @param mode is the targeted transfer mode
   * @return true if the transfer mode designates a grouped callback.
   */
  public static boolean isGrouped(int mode)
  {
    return ((mode & (CM_GROUPED|CM_SYNCGROUP)) != 0);
  }
  /**
   * @param mode is the targeted transfer mode
   * @return true if the transfer mode is starting in a 'wait' state.
   */
  public static boolean isWaiting(int mode)
  {
    return ((mode & CM_WAIT) == CM_WAIT);
  }
  /**
   * @param mode is the targeted transfer mode
   * @return true if the transfer mode designates a globals link
   * (receiving globals multicasts).
   */
  public static boolean isGlobal(int mode)
  {
    return ((mode & CM_GLOBAL) == CM_GLOBAL);    
  }
  public static boolean isCancelled(int mode)
  {
    return ((mode & 0x0f) == CM_CANCEL);        
  }
  /**
   * @param mode is the targeted transfer mode
   * @return true if the transfer mode has a long deadband before
   * registering a timeout
   */
  public static boolean hasLongDeadband(int mode)
  {
    if ((mode&0x0f) == CM_DATACHANGE) return true;
    if ((mode&0x0f) == CM_EVENT) return true;
    return false; 
  }
}