/*
 * File TStrings.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.definitions;

import de.desy.tine.server.logger.TFecLog;
import de.desy.tine.server.logger.TineLogLevel;

public final class TStrings
{
  static public final int STATUS_SHORTSIZE = 32;
  static public final int STATUS_SIZE = 192;
  static public final int FEC_NAME_SIZE = 16;
  static public final int FEC_DESC_SIZE = 64;
  static public final int FEC_LOCATION_SIZE = 48;
  static public final int FEC_VERSION_SIZE = 16;
  static public final int FEC_HDW_SIZE = 32;
  static public final int FEC_RESP_SIZE = 32;
  static public final int FEC_OS_SIZE = 16;
  static public final int EXPORT_NAME_SHORTSIZE = 16;
  static public final int EXPORT_NAME_SIZE = 32;
  static public final int XTN_NAME_SIZE = 16;  
  static public final int PROPERTY_NAME_SIZE = 64;  
  static public final int PROPERTY_NAME_SHORTSIZE = 32;  
  static public final int PROPERTY_DESC_SIZE = 64;
  static public final int PROPERTY_REDIR_SIZE = STATUS_SIZE;
  static public final int DEVICE_NAME_SIZE = 64;  
  static public final int DEVICE_NAME_SHORTSIZE = 16;
  static public final int DEVICE_NAME_MAXSIZE = 1024;
  static public final int DEVICE_REDIR_SIZE = STATUS_SIZE;
  static public final int DEVICE_DESC_SIZE = 64;  
  static public final int DEVICE_LOCATION_SIZE = 64;  
  static public final int EQM_NAME_SIZE = 8;  
  static public final int EQM_NAME_SHORTSIZE = 6;  
  static public final int TAG_NAME_SIZE = 16;  
  static public final int TAG_NAME_SHORTSIZE = 8;  
  static public final int CONTEXT_NAME_SHORTSIZE = 16;
  static public final int CONTEXT_NAME_SIZE = 32;
  static public final int SUBSYSTEM_NAME_SIZE = 16;
  static public final int SUBSYSTEM_NAME_SHORTSIZE = 5;
  static public final int SET_NAME_SIZE = 32;
  static public final int FORMAT_STR_SIZE = 64;
  static public final int ACCESS_STR_SIZE = 16;
  static public final int ADDR_SIZE = 16;
  static public final int ADDR_SIZE_IPV6 = 40;
  static public final int UNITS_SIZE = 16;
  static public final int USERNAME_SIZE = 16;
  static public final int FIELD_NAME_SIZE = 16;
  static public final int HISTORY_HOME_SIZE = 128;
  static public final int FULL_NAME_MAXSIZE = DEVICE_NAME_MAXSIZE + PROPERTY_NAME_SIZE + EXPORT_NAME_SIZE + CONTEXT_NAME_SIZE;
  static public final int ALARM_DATA_SHORTSIZE = 6;
  static public final int ALARM_DATA_SIZE = 64;
  static public final int ALARM_TAG_SIZE = 32;
  static public final int ALARM_TAG_SHORTSIZE = 16;
  static public final int ALARM_TEXT_SIZE = 64;
  static public final int ALARM_TEXT_SHORTSIZE = 40;
  static public final int ALARM_TEXT_LONGSIZE = 128;
  static public final int KEYWORD_NAME_SIZE = 64;
  static public final int KEYWORD_NAME_SHORTSIZE = 16;
  static private final String STRICTLY_ILLEGALCHARS = "[^\"*&]+";
  static private final String RUDECHARS = "[^\\\\/\n\t ]+";
  static private final String ALPHANUM = "[a-zA-Z0-9]+"; 
  static public final String MCADEV_TAG = "mcaDev=>";
  public static void validateName(String name, int maximumLength) 
  {
    if (name == null )throw new IllegalArgumentException("name is null");
    if (name.length() == 0) throw new IllegalArgumentException("name has zero length");
    if (!name.matches(STRICTLY_ILLEGALCHARS))throw new IllegalArgumentException("name \"" + name +"\" contains illegal characters");
    if (!name.chars().allMatch(c -> c < 128)) throw new IllegalArgumentException("name \"" + name +"\" contains non-ascii characters");
    if (name.length() > maximumLength)
    {
      throw new IllegalArgumentException("length of name " + name + 
              " (len: " + name.length() + ") exceeds limit " + maximumLength);
    }
  }
  public static void validateDeviceName(String deviceName)
  {
    if (deviceName != null && deviceName.length() == 0) return;
    validateName(deviceName,DEVICE_NAME_SIZE);
  }
  public static void validatePropertyName(String propertyName)
  {
    validateName(propertyName,PROPERTY_NAME_SIZE);
    if (!propertyName.matches(RUDECHARS))
      throw new IllegalArgumentException("property name \"" + propertyName +"\" contains illegal characters");
  }
  public static void validateExportName(String exportName)
  {
    validateName(exportName,EXPORT_NAME_SIZE);
    if (!exportName.matches(RUDECHARS))
      throw new IllegalArgumentException("export name \"" + exportName +"\" contains illegal characters");
  }
  public static void validateContextName(String contextName)
  {
    validateName(contextName,CONTEXT_NAME_SIZE);
    if (!contextName.matches(RUDECHARS))
      throw new IllegalArgumentException("context name \"" + contextName +"\" contains illegal characters");
  }
  public static void validateLocalName(String eqmName)
  {
    validateName(eqmName,EQM_NAME_SHORTSIZE);
  }
  public static void validateTagName(String tagName)
  {
    validateName(tagName,TAG_NAME_SIZE);
  }
  public static void validateSubSystemName(String subsystemName)
  {
    if (subsystemName == null || subsystemName.length() == 0)
    {
      TFecLog.log(TineLogLevel.WARN,"Subsystem name not supplied (TEST?)!");
      return;
    }
    validateName(subsystemName,SUBSYSTEM_NAME_SIZE);
    if (!subsystemName.matches(ALPHANUM))
      throw new IllegalArgumentException("subsystem name \"" + subsystemName +"\" contains illegal characters");
  }
  public static void validateFecName(String fecName)
  {
    validateName(fecName,FEC_NAME_SIZE);
    if (!fecName.matches(ALPHANUM))
      throw new IllegalArgumentException("fec name \"" + fecName +"\" contains illegal characters");
  }
}
