/*
 * File TTransport.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

/*
 * Created on Nov 16, 2004
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package de.desy.tine.definitions;

import java.net.InetAddress;
import de.desy.tine.addrUtils.TSrvEntry;
import de.desy.tine.io.MCastAddrMap;

/**
 * \internal
 * 
 * @author duval
 *
 * To change the template for this generated type comment go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
public class TTransport
{
  public static final short NONE = 0x00;
  public static final short IPX = 0x01;
  public static final short SPX = 0x02;
  public static final short TCP = 0x04;
  public static final short UDP = 0x08;
  public static final short NETSRV = 0x0a;
  public static final short PIPE = 0x10;
  public static final short MMF = 0x20;
  public static final short STREAM = 0x40;
  public static final short DBGPIPE = 0x80;
  public static final short TCP6 = 0x100;
  public static final short UDP6 = 0x200;
  public static final short STREAM6 = 0x400;
  public static final short DBGPIPE6 = 0x800;
  public static String toString(short transport)
  {
    switch (transport)
    {
      case IPX: return "IPX";
      case SPX: return "SPX";
      case TCP: return "TCP";
      case NETSRV: return "ADDR service";
      case UDP: return "UDP";
      case PIPE: return "LCL";
      case MMF: return "LCL";
      case STREAM: return "STREAM";
      default: return "---";
    }
  }
  public static boolean isStream(short transport)
  {
    return transport == STREAM;
  }
  public static boolean isTcp(short transport)
  {
    return (transport == TCP || transport == TCP6);
  }
  public static boolean isUdp(short transport)
  {
    if  ((transport & UDP) == UDP) return true;
    if  ((transport & UDP6) == UDP6) return true;    
    return false;
  }
  public static short getTransportFromConnectionMode(short mode)
  {
    if (TMode.isStream(mode)) return STREAM;
    if (TMode.isConnected(mode)) return TCP;
    return UDP; 
  }
  public static boolean isConnected(short transport)
  {
    switch (transport)
    {
      case TCP:
      case TCP6:
      case STREAM:
      case STREAM6:
        return true;
      default:
        break;
    }
    return false;    
  }
  public static final int minimumMcaVersionId = 4 * 65536 + 11;
  public static final int UDP_BUFFER_SIZE = 1472;
  public static final int TCP_BUFFER_SIZE = UDP_BUFFER_SIZE*4;
  public static final int UDP_MAXBUFFER_SIZE = 0x10000;
  public static final int MMF_BUFFER_SIZE = 65000;
  //public static final int DEFAULT_PROTOCOL_LEVEL = 7;
  public static final int legacyProtocolLevel = 6;
  public static final int thisProtocolLevel = 7;
  public static boolean isValidAddr(String addr)
  {
    return isValidIpV4Addr(addr) || isValidIpV6Addr(addr);
  }
  public static boolean isValidIpV4Addr(String addr)
  {
    if (addr == null || addr.length() == 0) return false;
    String[] parts = addr.split("\\.");
    if (parts != null && parts.length == 4) 
    {
      try { for (String p : parts) Integer.parseInt(p); } 
      catch (Exception e) { return false; } 
      return true;
    }
    return false;
  }
  public static boolean isValidIpV6Addr(String addr)
  {
    if (addr == null || addr.length() == 0) return false;
    String[] parts = addr.split(":");
    if (parts != null) 
    {
      try 
      { 
        for (String p : parts) 
        {
          if (p.length() == 0) continue; // probably a '::' in the string
          Integer.parseInt(p,16); 
        }
      } 
      catch (Exception e) { return false; } 
      return true;
    }
    return false;
  }
  public static String getMCastGroup(String hostAddr, String mcAddr, boolean legacyRules)
  { // apply mcast rules to get the appropriate group ...
    if (mcAddr == null) return null;
    if (hostAddr == null) return mcAddr;
    String ip = null;
    if (mcAddr.contains(":"))
    { // ipv6 rules
      String[] mp = mcAddr.split("::");
      String[] hp = hostAddr.split(":");
      ip = mp[0]+":";
      int idx = hp.length > 4 ? hp.length - 4 : 1;
      for (int i=idx; i<hp.length; i++) ip += ":"+hp[i];
    }
    else
    { // ipv4 rules;
      String[] mp = mcAddr.split("\\.");
      String[] hp = hostAddr.split("\\.");
      if (legacyRules) 
        ip = mp[0] + "." + mp[1] + "." + hp[2] + "." + hp[3];
      else 
        ip = mp[0] + "." + hp[1] + "." + hp[2] + "." + hp[3];
    }
    return ip;
  }
  public static String getMCastGroup(String hostAddr, boolean legacyRules)
  { // apply mcast rules to get the appropriate group ...
    return getMCastGroup(hostAddr,MCastAddrMap.getMCastAddr(hostAddr,legacyRules),legacyRules);
  }
  public static String getMCastGroup(String hostAddr)
  { // apply mcast rules to get the appropriate group ...
    return getMCastGroup(hostAddr,MCastAddrMap.getMCastAddr(hostAddr,false),false);
  }
  public static String getMCastGroup(TSrvEntry srv)
  {
    if (srv == null || srv.fecAddr == null || srv.fecAddr.fecHost == null) return null;
    String fa = srv.fecAddr.fecHost.getHostAddress();
    boolean legacyRules = srv.fecAddr.getAddressFlags() == 0;
    return getMCastGroup(fa,legacyRules);
  }
  public static InetAddress getMCastAddr(TSrvEntry srv)
  {
    try { return InetAddress.getByName(getMCastGroup(srv)); } catch (Exception e) {}
    return null;
  }
}
