/*
 * File DHS.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.eventUtils;
import java.io.*;
import de.desy.tine.definitions.TFormat;
import de.desy.tine.endianUtils.Swap;

public class DHS
{
  public String channelID = new String(new StringBuffer(32));
  public int eventCode;
  public short arraysize;
  public short datatype;
  public float SamplingRate;
  public int numPreTriggerValues;
  public int TimeOffset;
  public int LowCount;
  public int HighCount;
  public String egu = new String(new StringBuffer(16));
  public float LowEGU;
  public float HighEGU;  
  public int datalength;
  public String comment = new String(new StringBuffer(256));
  public static int sizeInBytes = (32 + 4 + 2*2 + 4 + 4*4 + 16 + 4*2 + 4 + 4*64);
  ByteArrayOutputStream dBuffer = new ByteArrayOutputStream(sizeInBytes);
  byte[] hByteArray;
  public String toString()
  {
    String dhs;
    
    dhs = new String(
    	"channelID    : " + channelID + "\n" +
        "eventCode    : " + eventCode + "\n" +
        "arraySize    : " + (datalength > 0 ? datalength : arraysize) + "\n" +
        "dataType     : " + TFormat.toString(datatype) + "\n" +
        "SamplingRate : " + SamplingRate + "\n" +
        "nPreTrigVals : " + numPreTriggerValues + "\n" +
        "timeOffset   : " + TimeOffset + "\n" +
        "lowCount     : " + LowCount + "\n" +
        "highCount    : " + HighCount + "\n" +
        "egu          : " + egu + "\n" + 
        "lowEGU       : " + LowEGU + "\n" +
        "highEGU      : " + HighEGU + "\n" +
        "comment      : " + comment
        );
    return dhs;
  }
  public byte[] toByteArray()
  {
    try
    {
      dBuffer.reset();
      DataOutputStream ds = new DataOutputStream(dBuffer);
      byte[] bstr = new byte[256];
      byte[] tmp  = channelID.getBytes();
      for (int i=0; i<32; i++) bstr[i] = i < tmp.length ? tmp[i] : 0;
      ds.write(bstr,0,32);
      ds.writeInt(Swap.Int(eventCode));
      ds.writeShort(Swap.Short(arraysize));
      ds.writeShort(Swap.Short(datatype)+512);
      ds.write(Swap.Float(SamplingRate),0,4);
      ds.writeInt(Swap.Int(numPreTriggerValues));
      ds.writeInt(Swap.Int(TimeOffset));
      ds.writeInt(Swap.Int(LowCount));
      ds.writeInt(Swap.Int(HighCount));
      tmp  = egu.getBytes();
      for (int i=0; i<16; i++) bstr[i] = i < tmp.length ? tmp[i] : 0;
      ds.write(bstr,0,16);
      ds.write(Swap.Float(LowEGU),0,4);
      ds.write(Swap.Float(HighEGU),0,4);
      ds.writeInt(Swap.Int(datalength));
      tmp  = comment.getBytes();
      for (int i=0; i<256; i++) bstr[i] = i < tmp.length ? tmp[i] : 0;
      ds.write(bstr,0,256);
      ds.close();
    }
    catch (IOException e)
    {
      e.printStackTrace();
      System.exit(1);
    }	 
    hByteArray = dBuffer.toByteArray();
    return hByteArray;
  }
  public void toStruct()
  {
    try
    {
      byte[] bstr = new byte[256];
      ByteArrayInputStream dis = new ByteArrayInputStream(hByteArray);
      DataInputStream ds = new DataInputStream(dis);
      ds.read(bstr,0,32);
      channelID = new String(bstr).trim();
      eventCode = Swap.Int(ds.readInt());
      arraysize = Swap.Short(ds.readShort());
      datatype = Swap.Short(ds.readShort());
      if (datatype > 512) datatype -= 512;
      ds.read(bstr,0,4);
      SamplingRate = Swap.Float(bstr);
      numPreTriggerValues = Swap.Int(ds.readInt());
      TimeOffset = Swap.Int(ds.readInt());
      LowCount = Swap.Int(ds.readInt());
      HighCount = Swap.Int(ds.readInt());
      ds.read(bstr,0,16);
      int len;
      for (len=0; len < 16 && bstr[len] != 0; len++);
      egu = new String(bstr,0,len);
      ds.read(bstr,0,4);
      LowEGU = Swap.Float(bstr);
      ds.read(bstr,0,4);
      HighEGU = Swap.Float(bstr);
      datalength = Swap.Int(ds.readInt());
      ds.read(bstr,0,256);
      for (len=0; len < 256 && bstr[len] != 0; len++);
      comment = new String(bstr,0,len);
      ds.close(); dis.close();
    }
    catch (IOException e)
    {
      e.printStackTrace();
      System.exit(1);
    }	 
  }
}
