package de.desy.tine.example.sine;

import de.desy.tine.dataUtils.*;
import de.desy.tine.definitions.*;
import de.desy.tine.server.devices.*;
import de.desy.tine.server.equipment.*;
import de.desy.tine.server.properties.*;
import de.desy.tine.types.*;

/**
 * Example equipment module. The equipment module maps the device properties to
 * de.desy.tine.client properties.
 * 
 * Notes:
 * The routine registerProperties():
 * The principal property registeration occurs via the 'exports.csv' configuration file.
 * In this case the property 'handlers' need to be attached to the registered properties.
 * An alternative would be the use the registerProperty(TExportProperty, TPropetyHandler)
 * method of TEquipmentModule.  In this case you would have to supply all property
 * information inside an instance of TExportProperty prior to making the call. Using
 * an 'export.csv' file expidites this process considerably!
 *
 * Property Handler examples:
 * Have a look at the examples for handling requests for properties "Amplitude" and
 * "Frequency".  In one case ("Frequency") the the entire request is passed to a
 * call routine which handles 'write' and 'read' requests collectively.  The other
 * case ("Amplitude") splits the 'write' and 'read' requests into two separate
 * routines.  The differences between the two is by and large cosmetic.  It is
 * generally useful to allow atomic 'write-read' requests along with individual
 * 'writes' and 'reads'. 
 * 
 * Such equipment module routines should be to a large extent generated from a
 * wizard, so that your only task might be to fill in or expand routines such as 
 * writeAmplitude to suit your needs.
 */
public class SineEquipmentModule extends TEquipmentModule
{
  boolean wrapMultiChannelArrays = true;
  TDeviceList myDeviceSet; // the devices managed by this equipment module
  SineInfo[] info; // an array of SineInfo tagged structures 
  // Wrapper which casts to sineDevice
  private SineDevice findDevice(String deviceName)
  {
    if (myDeviceSet == null) return null;
    return (SineDevice) myDeviceSet.getDevice(deviceName);
  }
  /**
   * Constructor.  Supplies a specific local name setting. 
   * 
   * @param localName the internally used 6-char local name
   * @param devices a set of device modules to be managed by this equipment module
   */
  public SineEquipmentModule(String localName, SineDevice[] devices)
  {
    super(localName);
    registerDevices(devices);
    registerProperties();
    myDeviceSet = super.getDeviceList();
    info = new SineInfo[myDeviceSet.getNumberOfDevices()];
    for (int i=0; i<myDeviceSet.getNumberOfDevices(); i++)
    {
      info[i] = new SineInfo();
      // could either fill in those things that are known are are static here
      // but this simple test will fill in the fields when the structure is asked for 
    }
  }
  /**
   * Constructor.  Attempts to determine a suitable local name. 
   * The local name used will be the first 6 characters of the class name,
   * in this case 'SineEq'.  The local name is required to be unique only within
   * a given server process. Thus, if another TEquipmentModule is to be 
   * registered with the same factory, it should have a class name whose initial
   * 6 characters differ from this one. 
   * 
   * @param devices a set of device modules to be managed by this equipment module
   */
  public SineEquipmentModule(SineDevice[] devices)
  {
    super();
    registerDevices(devices);
    registerProperties();
    myDeviceSet = super.getDeviceList();
  }
  private int readString(String devName, TDataType dout)
  {
    return dout.putData(new NAME16(devName));
  }
  /**
   * This Reads the sine values of a device. Called by PropertyHandler.
   */
  private int readSine(String devName, TDataType dout)
  {
    SineDevice device = findDevice(devName);
    if (device == null) return TErrorList.device_not_connected;
    info[device.devNumber].numberCalls[0]++;
    return dout.putData(device.getSine());
  }
  /**
   * Writes the amplitude of a device.
   * 
   * @param devName The targetted device name
   * @param din The input data object containing the set point
   * @return
   */
  public int writeAmplitude(String devName, TDataType din)
  {
    int cc = 0;
    SineDevice theDevice = null;
    if (din.getArrayLength() == 1)
    { // CLIENT WANTS TO SET AMPLITUDE
      double[] input = new double[1];
      if (din.getArrayLength() != 1) return TErrorList.dimension_error;
      theDevice = (SineDevice) myDeviceSet.getDevice(devName);
      if (theDevice == null) return TErrorList.illegal_equipment_number;
      if ((cc=din.getData(input)) != 0) return cc;
      theDevice.setAmplitude(input[0]);
    }
    return cc;
  }
  /**
   * Reads the amplitude of a device or set of devices.  
   * If 'wrapMultiChannelArrays' is true, the caller can start at
   * any element of his choosing and receive a full array as output
   * (regardless of the datalength requested and the starting point),
   * otherwise the array might be truncated if the number of requested
   * elements exceeds the number of array elements remaining.
   * 
   * @param devName The initial targetted device name in the requested series
   * @param dout The output data object to contain the returned data set.
   * @return
   */
  public int readAmplitude(String devName, TDataType dout)
  {
    int cc = 0;
    SineDevice theDevice;
    int nret = dout.getArrayLength();
    if (nret < 1) return TErrorList.dimension_error;
    int ndv = myDeviceSet.getNumberOfDevices();
    int dv = myDeviceSet.getDeviceNumber(devName);
    if (dv < 0 || dv >= ndv) return TErrorList.illegal_equipment_number;
    double[] output = new double[ndv];
    for (int i=0; i<ndv; i++)
    {
      theDevice = (SineDevice) myDeviceSet.getDevice(i);
      output[i] = theDevice.getAmplitude();
    }
    cc = dout.putData(output,nret,dv);
    return cc;
  }
  /**
   * This property handler reads or sets the frequency of one device, or reads
   * the frequencies of a subset of devices. This call method demonstrates the
   * read of a multi-channel array, where each element of the array refers to a
   * registered device.  Note the (optional) end-wrapping, allowing the caller to
   * retrieve the entire array beginning at an element of his choosing.
   * If 'wrapMultiChannelArrays' is true, the caller will can start at
   * any element of his choosing and receive a full array as output
   * (regardless of the datalength requested and the starting point),
   * otherwise the array might be truncated if the number of requested
   * elements exceeds the number of array elements remaining.
   */
  private int callFrequency(String devName, TDataType dout, TDataType din, TAccess devAccess)
  {
    int cc = 0;
    SineDevice theDevice;
    if (devAccess.isWrite())
    { // CLIENT WANTS TO SET FREQUENCY (allow single-channel write)
      double[] input = new double[1];
      if (din.getArrayLength() != 1) return TErrorList.dimension_error;
      theDevice = (SineDevice) myDeviceSet.getDevice(devName);
      if (theDevice == null) return TErrorList.illegal_equipment_number;
      if ((cc=din.getData(input)) != 0) return cc;
      theDevice.setFrequency(input[0]);
    }
    if (devAccess.isRead())
    { // CLIENT WANTS TO READ (allow multi-channel read)
      int nret = dout.getArrayLength();
      if (nret < 1) return TErrorList.dimension_error;
      int ndv = myDeviceSet.getNumberOfDevices();
      int dv = myDeviceSet.getDeviceNumber(devName);
      if (dv < 0 || dv >= ndv) return TErrorList.illegal_equipment_number;
      double[] output = new double[ndv];
      for (int i=0; i<ndv; i++)
      {
        theDevice = (SineDevice) myDeviceSet.getDevice(i);
        output[i] = theDevice.getFrequency();
      }
      cc = dout.putData(output,nret,dv);
    }
    return cc;
  }

  private int callPhase(String devName, TDataType dout, TDataType din, TAccess devAccess)
  {
    int cc = 0;
    SineDevice theDevice;
    if (devAccess.isWrite())
    { // CLIENT WANTS TO SET PHASE (allow single-channel write)
      double[] input = new double[1];
      if (din.getArrayLength() != 1) return TErrorList.dimension_error;
      theDevice = (SineDevice) myDeviceSet.getDevice(devName);
      if (theDevice == null) return TErrorList.illegal_equipment_number;
      if ((cc=din.getData(input)) != 0) return cc;
      theDevice.setPhase(input[0]);
    }
    if (devAccess.isRead())
    { // CLIENT WANTS TO READ (allow multi-channel read)
      int nret = dout.getArrayLength();
      if (nret < 1) return TErrorList.dimension_error;
      int ndv = myDeviceSet.getNumberOfDevices();
      int dv = myDeviceSet.getDeviceNumber(devName);
      if (dv < 0 || dv >= ndv) return TErrorList.illegal_equipment_number;
      double[] output = new double[ndv];
      for (int i=0; i<ndv; i++)
      {
        theDevice = (SineDevice) myDeviceSet.getDevice(i);
        output[i] = theDevice.getPhase();
      }
      cc = dout.putData(output,nret,dv);
    }
    return cc;
  }

  private int callNoise(String devName, TDataType dout, TDataType din, TAccess devAccess)
  {
    int cc = 0;
    SineDevice theDevice;
    if (devAccess.isWrite())
    { // CLIENT WANTS TO SET PHASE (allow single-channel write)
      double[] input = new double[1];
      if (din.getArrayLength() != 1) return TErrorList.dimension_error;
      theDevice = (SineDevice) myDeviceSet.getDevice(devName);
      if (theDevice == null) return TErrorList.illegal_equipment_number;
      if ((cc=din.getData(input)) != 0) return cc;
      theDevice.setNoise(input[0]);
    }
    if (devAccess.isRead())
    { // CLIENT WANTS TO READ (allow multi-channel read)
      int nret = dout.getArrayLength();
      if (nret < 1) return TErrorList.dimension_error;
      int ndv = myDeviceSet.getNumberOfDevices();
      int dv = myDeviceSet.getDeviceNumber(devName);
      if (dv < 0 || dv >= ndv) return TErrorList.illegal_equipment_number;
      double[] output = new double[ndv];
      for (int i=0; i<ndv; i++)
      {
        theDevice = (SineDevice) myDeviceSet.getDevice(i);
        output[i] = theDevice.getNoise();
      }
      cc = dout.putData(output,nret,dv);
    }
    return cc;
  }
  public int readInfo(String devName, TDataType dout)
  { // the caller wants a SineInfo structure array ...
    int cc = 0;
    SineDevice theDevice;
    int nret = dout.getArrayLength();
    if (nret < 1) return TErrorList.dimension_error;
    int ndv = myDeviceSet.getNumberOfDevices();
    int dv = myDeviceSet.getDeviceNumber(devName);  // the device called
    if (dv < 0 || dv >= ndv) return TErrorList.illegal_equipment_number;
    String dstr;
    // fill this in below on the fly for the sake of this test ...
    for (int i=0; i<ndv; i++)
    {
      theDevice = (SineDevice) myDeviceSet.getDevice(i);
      dstr = "the sine device called " + theDevice.getName();
      dstr.getChars(0,dstr.length(),info[i].description,0);
      info[i].amplitude[0] = (float)theDevice.getAmplitude();
      info[i].frequency[0] = (float)theDevice.getFrequency();
      info[i].noiseLevel[0] = (float)theDevice.getNoise();
      info[i].phase[0] = (float)theDevice.getPhase();
    }
    // now send it back (use the device called as an offset)
    cc = dout.putData(info, info.length - dv,dv);
    return cc;
  }

  private void registerProperties()
  { // all property information (except the handlers) from exports.csv file
    getExportInformationFromFile();
    // attach the property handlers ...
    attachPropertyHandler("Sine",new TPropertyHandler()
    {
      protected int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        return readSine(devName, dout);
      }
    });
    attachPropertyHandler("Amplitude",new TPropertyHandler()
    {
      protected int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        int rc = TErrorList.illegal_read_write;
        if (devAccess.isWrite())
        {
          rc = writeAmplitude(devName, din);
          if (rc != 0) return rc;
        }
        if (devAccess.isRead())
        {
          rc = readAmplitude(devName, dout);
        }
        return rc;
      }
    });
    attachPropertyHandler("Frequency",new TPropertyHandler()
    {
      protected int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        return callFrequency(devName, dout, din, devAccess);
      }
    });
    attachPropertyHandler("Phase",new TPropertyHandler()
        {
          protected int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
          {
            return callPhase(devName, dout, din, devAccess);
          }
        });
    attachPropertyHandler("Noise",new TPropertyHandler()
        {
          protected int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
          {
            return callNoise(devName, dout, din, devAccess);
          }
        });
    attachPropertyHandler("Info",new TPropertyHandler()
        {
          protected int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
          {
            int rc = TErrorList.illegal_read_write;
            if (devAccess.isRead())
            {
              rc = readInfo(devName, dout);
            }
            return rc;
          }
        });
    attachPropertyHandler("ThisDevice",new TPropertyHandler()
    {
      protected int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        return readString(devName, dout);
      }
    });
  }
}