/*
 * File TGlbDataHdr.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.headers;
import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.IOException;
import java.net.InetAddress;
import java.util.LinkedList;
import de.desy.tine.client.TLinkFactory;
import de.desy.tine.client.TLink;
import de.desy.tine.dataUtils.TDataType;
import de.desy.tine.definitions.TErrorList;
import de.desy.tine.definitions.TFormat;
import de.desy.tine.definitions.TStrings;
import de.desy.tine.definitions.TTransport;
import de.desy.tine.endianUtils.Swap;
import de.desy.tine.server.logger.DbgLog;
import de.desy.tine.server.logger.MsgLog;

public class TGlbDataHdr 
{
  short msgSizeInBytes;
  short mtu;
  int debugLevel = 0;
  byte[] globalsNameAsByteArray = new byte[TStrings.KEYWORD_NAME_SIZE];  // 64 chars
  short dArrayLength;
  short dFormat;
  short status;
  int dTimestampUSec;
  short interval;
  int dStamp;
  int sysStamp;
  int dTimestamp;
  int dTimestampMSec;
  static final int maxPayloadSizeInBytes = TTransport.UDP_BUFFER_SIZE;
  static final int glbhdrSizeInBytes = (TStrings.KEYWORD_NAME_SIZE + 4*2 + 4*4);
  static final int glbhdrSizeInBytesLegacy = (TStrings.KEYWORD_NAME_SHORTSIZE + 2*2 + 1*4);
  static ByteArrayInputStream dBuffer;
  byte[] usernameAsByteArray = new byte[TStrings.USERNAME_SIZE];
  private int  numInPacket;
  static final short hdrSizeInBytesLegacy = 2 + TStrings.USERNAME_SIZE + 2 + 2; // network header size
  static final short hdrSizeInBytes = 48; // network header size
  private int totalSizeInBytes;
  TLinkFactory tf;
 /**
   * Attach a GlobalsData object to the factory
   * Creation date: (8/21/01 6:33:36 PM)
   */
  public TGlbDataHdr(TLinkFactory tfactory) 
  {
    tf = tfactory;
  }
  private LinkedList<TLink> lnks = new LinkedList<TLink>();
  private byte[] dBytes = new byte[maxPayloadSizeInBytes];
  public TLink[] toStruct(InetAddress gaddr, byte[] data, int length)
  {
    TLink glb;
    boolean isLegacy = false;
    int bytesread = 0;
    int ghdrsize = glbhdrSizeInBytes;
    int hdrsize;
    int nlen = TStrings.KEYWORD_NAME_SIZE;
    TLinkFactory.GlobalInfo gi = tf.getGlobalsInfo(gaddr);
    TGlbHdr glbhdr = new TGlbHdr(data, 0, length);
    totalSizeInBytes = glbhdr.getTotalSizeInBytes();
    if (totalSizeInBytes == 0)
    {
      DbgLog.log("TGlobalsHdr","received incomplete datagram");
      return null;
    }
    try
    {
      lnks.clear();
      dBuffer = new ByteArrayInputStream(data);
      DataInputStream ds = new DataInputStream(dBuffer);
      int p = glbhdr.getTineProtocol();
      // overall message length:
      hdrsize = glbhdr.getHeaderSizeInBytes();
      ghdrsize = p <= TTransport.legacyProtocolLevel ? glbhdrSizeInBytesLegacy : glbhdrSizeInBytes;
      if (totalSizeInBytes < hdrsize + ghdrsize)
      {
        String msg = "recv: " + totalSizeInBytes + " bytes : hdr " + (hdrsize + ghdrsize);
        DbgLog.log("TGlobalsHdr",msg);
        return null;
      }
      bytesread = hdrsize;
      numInPacket = glbhdr.getNumberSegments();
      if (ds.skipBytes(bytesread) != bytesread)
      {
        if (TLinkFactory.debugLevel > 0) DbgLog.log("TGlbDataHdr","could not skip "+bytesread+" bytes!");
        return null;        
      }
      switch (p)
      {
        case 3: // really legacy !
          isLegacy = true;
          ghdrsize = glbhdrSizeInBytesLegacy;
          nlen = TStrings.KEYWORD_NAME_SHORTSIZE;
          break;
        case 4:
        case TTransport.thisProtocolLevel:
          break;
        default: // is not a globals packet!
          return null;
      }
      String key, ctx;
      for (int i=0; i<numInPacket && bytesread < totalSizeInBytes; i++)
      {        
        ds.read(globalsNameAsByteArray,0,nlen); 
        dArrayLength = Swap.Short(ds.readShort());
        dFormat = (short)(Swap.Short(ds.readShort()) % 256);
        if (p < TTransport.thisProtocolLevel)
        {
          short usec;
          if (isLegacy)
          {
            dTimestamp = Swap.Int(ds.readInt());
            status = 0;
            dStamp = sysStamp = dTimestampMSec = 0;
            usec = 0;
          }
          else
          {
            status = Swap.Short(ds.readShort());
            usec = Swap.Short(ds.readShort());
            dStamp = Swap.Int(ds.readInt());
            sysStamp = Swap.Int(ds.readInt());
            dTimestamp = Swap.Int(ds.readInt());
            dTimestampMSec = Swap.Int(ds.readInt());          
          }
          dTimestampUSec = dTimestampMSec * 1000 + (usec % 1000);
        }
        else
        {
          status = Swap.Short(ds.readShort());
          interval = Swap.Short(ds.readShort());
          dStamp = Swap.Int(ds.readInt());
          sysStamp = Swap.Int(ds.readInt());
          dTimestamp = Swap.Int(ds.readInt());
          dTimestampUSec = Swap.Int(ds.readInt());          
        }
        bytesread += ghdrsize;
        ctx = null;
        key = new String(globalsNameAsByteArray).trim();
        String[] parts = key.split("/");
        if (parts.length > 2)
        {
          ctx = parts[1]; key = parts[2];
        }
        if (ctx != null) gi = tf.getGlobalsInfo(ctx, gaddr);
        glb = tf.findTLink(gaddr, ctx, key);
        int dataSizeInBytes = dArrayLength * TFormat.formatSizeOf(dFormat);
        byte[] d = dBytes; //new byte[dataSizeInBytes];
        ds.read(d,0,dataSizeInBytes); //ds.read(d,0,d.length);
        bytesread += dataSizeInBytes;
        if (gi != null) 
        {
          gi.updateDataElement(key.trim(), d, dArrayLength, dFormat, dTimestamp, dTimestampUSec, sysStamp, status);
          gi.hasUpdated = true;
        }
        // push this data set into the corresponding TDataType object
        if (glb != null && glb.dOutput != null)
        {
          postOutputData(glb, d, dFormat, dArrayLength, dTimestamp, dTimestampUSec, sysStamp, dStamp, status, TLink.STATUS_REMOTE);
          lnks.add(glb);
        }
      }
      ds.close();
      dBuffer.close();
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("TGlobalsHdr.toStruct",e.toString(),TErrorList.code_failure,e,0);
      return null;
    }
    return lnks.toArray(new TLink[0]);
  }
  public static void postOutputData(TLink lnk,byte[] d,short fmt,int len,int ts,int tsUSEC,int sds,int ds,int sts,int stsrc)
  {
    lnk.dOutput.resetBuffersReady();
    if (lnk.dOutput.getFormat() == TFormat.CF_DEFAULT)
    {
      lnk.dOutput.dFormat = fmt;
      lnk.dOutput.dArrayLength = len;
      lnk.dOutput.numblks = 1;
      if (fmt != TFormat.CF_STRUCT && lnk.dOutput.getDataObject() == null)
      {
        lnk.dOutput.setDataObject(lnk.dOutput.dArrayLength,lnk.dOutput.dFormat);
      }
    }
    lnk.linkStatus = sts;
    lnk.linkStatusSource = stsrc;
    lnk.dOutput.sysDataStamp = sds;
    lnk.dOutput.usrDataStamp = ds;
    lnk.dOutput.timestamp = ts;
    lnk.dOutput.timestampUSEC = tsUSEC;
    lnk.dOutput.dTimestamp = (long)ts*1000 + tsUSEC/1000;
    if (fmt == lnk.dOutput.getFormat())
    {
      lnk.dOutput.update(d,1,0);
      lnk.dOutput.blksin = lnk.dOutput.bytesin = 0;
      lnk.dOutput.resetBuffersReady();
      lnk.dOutput.isDataObjectInSync = false;
      TLinkFactory.getInstance().fillinIncomingData(lnk);
    }
    else
    { // try to re-format (what came in to what the caller wants ...)
      TDataType dt = new TDataType(len,fmt);
      dt.update(d,1,0);
      lnk.dOutput.blksin = lnk.dOutput.bytesin = 0;
      dt.resetBuffersReady();
      Object srcData = dt.getDataObject();
      dt.getData(srcData);
      lnk.dOutput.putData(srcData);
      lnk.dOutput.hasBeenUpdated = dt.hasBeenUpdated = true;
      lnk.dOutput.dCompletionLength = len;
    }
  }
}
