/*
 * File TReqHdr.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

//: TLink.java
package de.desy.tine.headers;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Arrays;

import de.desy.tine.client.TLinkFactory;
import de.desy.tine.dataUtils.UnsignedUtils;
import de.desy.tine.definitions.*;
import de.desy.tine.endianUtils.Swap;
import de.desy.tine.server.logger.MsgLog;
import de.desy.tine.startup.TInitializerFactory;
public final class TReqHdr
{
  // header bytes:
  private int totalSizeInBytes;
  private int headerSizeInBytes;
  private int pid;
  private int tineVersion;
  private short numberSegments;
  private short tineProtocol;
  private short endianness;
  private short encoding;
  private String username = null; // 16 char
  private String usertype = null; // 8 char
  // end of header bytes
  static public final int hdrSizeInBytes = 48; // network header size
  static public final int hdrSizeInBytesL = 22; // legacy header size
  private int msgSizeInBytes;
  private int revisionId;
  ByteArrayOutputStream dBuffer;
  public int getHeaderSizeInBytes()
  {
    return tineProtocol <= TTransport.legacyProtocolLevel ? hdrSizeInBytesL : hdrSizeInBytes;
  }
  public static int getHeaderSizeInBytes(int protocol)
  {
    if (protocol > TTransport.legacyProtocolLevel) return hdrSizeInBytes;
    return hdrSizeInBytesL;
  }
  public TReqHdr(String u,String t,int protocolLevel)
  {
    username = u.toUpperCase();
    usertype = t.toUpperCase();
    tineProtocol = (short)protocolLevel;
    tineVersion = TInitializerFactory.getInstance().getInitializer().getVersionId();
    revisionId = 99 * 256 + 99; // used only in legacy headers
    headerSizeInBytes = getHeaderSizeInBytes(tineProtocol);
    dBuffer = new ByteArrayOutputStream(headerSizeInBytes);
    pid = TLinkFactory.getInstance().getPid();
    endianness = TEndianness.BO_LITTLE_ENDIAN; // for the moment
    encoding = TEncoding.CE_ASCII;
    numberSegments = 1;
  }
  public TReqHdr(String u,String t,int segments,int protocolLevel,int versionId)
  {
    username = u.toUpperCase();
    usertype = t.toUpperCase();
    tineProtocol = (short)protocolLevel;
    tineVersion = versionId;
    revisionId = 99 * 256 + 99; // used only in legacy headers
    headerSizeInBytes = getHeaderSizeInBytes(tineProtocol);
    dBuffer = new ByteArrayOutputStream(headerSizeInBytes);
    pid = TLinkFactory.getInstance().getPid();
    endianness = TEndianness.BO_LITTLE_ENDIAN; // for the moment
    encoding = TEncoding.CE_ASCII;
    numberSegments = (short)segments;
  }
  public int getNumberSegments() { return numberSegments; }
  public void setNumberSegments(int value)
  {
    numberSegments = (short)(value > 1 ? value : 1);
  }
  public int getPid() { return pid; }
  public int getEndianness() { return endianness; }
  public int getEncoding() { return encoding; }
  public int getVersionId() { return tineVersion; }
  public String getUserName() { return username; }
  public void setUserName(String name) 
  { 
    if (name == null) return;
    if (username != name && username.compareToIgnoreCase(name) != 0)
    {
      username = name.toUpperCase();
    }
  }
  public String getUserType() { return usertype; }
  public void setUserType(String type) 
  { 
    if (type == null) type = "JAVA";
    if (usertype != type && usertype.compareToIgnoreCase(type) != 0)
    {
      usertype = type.toUpperCase();
    }
  }  
  public int getTotalSizeInBytes() { return totalSizeInBytes; }
  public int getMsgSizeInBytes() { return msgSizeInBytes;}
  public void setMsgSizeInBytes(int sizeInBytes) 
  { 
    msgSizeInBytes = sizeInBytes;
  }
  private byte[] bstr = new byte[16];
  public byte[] toByteArray()
  { // prepare outgoing bytestream (e.g. client sending request to server)
    totalSizeInBytes = (int)(msgSizeInBytes + getHeaderSizeInBytes());
    try
    {
      dBuffer.reset();
      DataOutputStream ds = new DataOutputStream(dBuffer);
      if (tineProtocol > TTransport.legacyProtocolLevel)
      {        
        ds.writeInt(Swap.Int(totalSizeInBytes));
        ds.writeInt(Swap.Int(headerSizeInBytes));
        ds.writeInt(Swap.Int(pid));
        ds.writeInt(Swap.Int(tineVersion));
        ds.writeShort(Swap.Short(numberSegments));
        ds.writeShort(Swap.Short(tineProtocol));
        ds.writeShort(Swap.Short(endianness));
        ds.writeShort(Swap.Short(encoding));   
        Arrays.fill(bstr, (byte)0);
        byte[] tmp  = username.getBytes();
        for (int i=0; i<16 && i<tmp.length; i++) bstr[i] = tmp[i];
        ds.write(bstr,0,16);
        Arrays.fill(bstr, (byte)0);
        tmp  = usertype.getBytes();
        for (int i=0; i<8 && i<tmp.length; i++) bstr[i] = tmp[i];
        ds.write(bstr,0,8);
      }
      else
      { // legacy server
        ds.writeShort(Swap.Short(totalSizeInBytes));
        Arrays.fill(bstr, (byte)0);
        byte[] tmp  = username.getBytes();
        for (int i=0; i<16 && i<tmp.length; i++) bstr[i] = tmp[i];
        ds.write(bstr,0,16);
        ds.writeShort(Swap.Short(tineProtocol));
        ds.writeShort(Swap.Short(revisionId));        
      }
      ds.close();
      return dBuffer.toByteArray();
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("TReqHdr.toByteArray",e.toString(),TErrorList.code_failure,e,0);
    }
    return null;  
  }
  public TReqHdr(byte[] data,int off, int len)  
  { // prepare from incoming bytestream (e.g. server getting request from client)
    try
    {
      ByteArrayInputStream instream = new ByteArrayInputStream(data,off,len);
      DataInputStream ds = new DataInputStream(instream);
      ds.skipBytes(18);
      tineProtocol = Swap.Short(ds.readShort());
      ds.reset();      
      if (tineProtocol > TTransport.legacyProtocolLevel)
      {
        totalSizeInBytes = Swap.Int(ds.readInt());
        headerSizeInBytes = Swap.Int(ds.readInt());
        pid = Swap.Int(ds.readInt());
        tineVersion = Swap.Int(ds.readInt());      
        numberSegments = Swap.Short(ds.readShort());
        tineProtocol = Swap.Short(ds.readShort());
        endianness = Swap.Short(ds.readShort());
        encoding = Swap.Short(ds.readShort());      
        Arrays.fill(bstr, (byte)0);
        ds.read(bstr,0,16);
        username = (new String(bstr)).trim();
        Arrays.fill(bstr, (byte)0);
        ds.read(bstr,0,8);
        usertype = (new String(bstr)).trim();
      }
      else
      {
        totalSizeInBytes = UnsignedUtils.toUnsignedShort(Swap.Short(ds.readShort()));
        Arrays.fill(bstr, (byte)0);
        ds.read(bstr,0,16);
        username = (new String(bstr)).trim();
        tineProtocol = Swap.Short(ds.readShort());
        revisionId = Swap.Short(ds.readShort());
        tineVersion = 4 * 65536 + revisionId;
        pid = 0;
        numberSegments = 0; // unknown
        endianness = TEndianness.BO_LITTLE_ENDIAN;
        encoding = TEncoding.CE_ASCII;
        usertype = "LEGACY";
        headerSizeInBytes = hdrSizeInBytesL;
      }
      ds.close();
      instream.close();
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("TReqHdr",e.toString(),TErrorList.code_failure,e,0);
    }
  }      
  public String toString()
  {
    //TODO: expand upon this ...
    return "User : " + username + "\nTotal size : " + totalSizeInBytes + "\nTotal number : " + numberSegments;
  }
  public short getProtocol()
  {
    return tineProtocol;
  }
  public short getTineProtocol()
  {
    return tineProtocol;
  }
  public void setTineProtocol(int protocolLevel)
  {
    if (protocolLevel != tineProtocol)
    {
      headerSizeInBytes = getHeaderSizeInBytes(protocolLevel);      
    }
    tineProtocol = (short)protocolLevel;
  }
}