/*
 * File AvailCor.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.histUtils;

import de.desy.tine.structUtils.TTaggedStructure;

public class AvailCor extends TTaggedStructure
{
  public void clear()
  {
  	startString = null;
  	stopString = null;
  	systemString = null;
  	stateString = null;
  	usrString = null;
  	cmtString = null;
  }
  private double[] start = new double[1];
  private double[] stop = new double[1];
  private int[] counts = new int[1];
  private int[] to = new int[1];
  private char[] startStr = new char[64];
  private String startString;
  private char[] stopStr = new char[64];
  private String stopString;
  private char[] system = new char[64];
  private String systemString;
  private char[] state = new char[64];
  private String stateString;
  private char[] usr = new char[16];
  private String usrString;
  private char[] cmt = new char[256];
  private String cmtString;
  private void initStructure()
  {
    addField(start,"startUtc");
    addField(stop,"stopUtc");
    addField(counts,"counts");
    addField(to,"hasProblems");
    addField(system,"system");
    addField(state,"state");    
    addField(startStr,"startStr");
    addField(stopStr,"stopStr");
    addField(usr,"user");
    addField(cmt,"comment");
    initDone();    
  }
  public AvailCor()
  {
    super();
    initStructure();
  }
  public AvailCor(AvailCor reference)
  {
    super();
    initStructure();
    start[0] = reference.start[0];
    stop[0] = reference.stop[0];
    counts[0] = reference.counts[0];
    to[0] = reference.to[0];
    System.arraycopy(reference.startStr, 0, startStr, 0, 64);    
    System.arraycopy(reference.stopStr, 0, stopStr, 0, 64);    
    System.arraycopy(reference.system, 0, system, 0, 64);    
    System.arraycopy(reference.state, 0, state, 0, 64);    
    System.arraycopy(reference.usr, 0, usr, 0, 16);    
    System.arraycopy(reference.cmt, 0, cmt, 0, 256);    
  }
  public boolean getIsAvailable()
  {
    return to[0] == 0;
  }
  public void setIsAvailable(boolean value)
  {
    if (value)
    {
      to[0] = 0;
      setState("Available");
    }
    else
    {
      to[0] = 1;
      setState("Problems");      
    }
  }
  public void markForRemoval()
  {
    to[0] = -1;
  }
  public String getStartString()
  {
    if (startString == null) startString = new String(startStr).trim();
    return startString;
  }
  public String getStopString()
  {
    if (stopString == null) stopString = new String(stopStr).trim();
    return stopString;
  }
  public String getState()
  {
    if (stateString == null) stateString = new String(state).trim();
    return stateString;
  }
  public void setState(String state) 
  {
    pushChars(state,this.state);
    this.stateString = new String(state);
  }  
  public String getSystem()
  {
    if (systemString == null) systemString = new String(system).trim();
    return systemString;
  }
  public void setSystem(String system) 
  {
    pushChars(system,this.system);
    this.systemString = new String(system);
  }
  public String getUser()
  {
    if (usrString == null) usrString = new String(usr).trim();
    return usrString;
  }
  public void setUser(String user) 
  {
    pushChars(user,this.usr);
    this.usrString = new String(user);
  }
  public String getComment()
  {
    if (cmtString == null) cmtString = new String(cmt).trim();
    return cmtString;
  }
  public void setComment(String comment) 
  {
    pushChars(comment,this.cmt);
    this.cmtString = new String(comment);
  }
  public double getStartTime()
  {
  	return start[0];
  }
  public void setStartTime(double time)
  {
  	start[0] = time;
  	//TODO: also set the string variant according to dd.mm.yy hh:mm:ss TZV
  	//TODO: also set the counts (stop - start)
  }
  public void setStartTime(long time)
  {
  	start[0] = ((double)time)/1000.0;
  }
  public void setStopTime(double time)
  {
  	stop[0] = time;
  	//TODO: also set the string variant according to dd.mm.yy hh:mm:ss TZV
  	//TODO: also set the counts (stop - start)
  }
  public void setStopTime(long time)
  {
  	stop[0] = ((double)time)/1000.0;
  }
  public double getStopTime()
  {
  	return stop[0];
  }
  public double getCounts()
  {
  	return counts[0];
  }
}
