/*
 * File DAQRec.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.histUtils;

import de.desy.tine.definitions.*;
import de.desy.tine.structUtils.TTaggedStructure;

public class DAQRec extends TTaggedStructure
{
  public void clear()
  {
    contextString = null;
    serverString = null;
    deviceString = null;
    propertyString = null;
    fn_keyString = null;
    tagString = null;
  }
  private char[] fn_key = new char[80]; /* keyword structure file name */
  private String fn_keyString; 
  private char[] tag = new char[80]; /* structure tag or input file name */
  private String tagString; 
  private char[] ctx = new char[TStrings.CONTEXT_NAME_SIZE]; /* calling context */
  private String contextString;
  private char[] srv = new char[TStrings.EXPORT_NAME_SIZE]; /* calling equipment module */
  private String serverString;
  private char[] prp = new char[TStrings.PROPERTY_NAME_SIZE]; /* calling property */
  private String propertyString;
  private char[] dev = new char[TStrings.DEVICE_NAME_SIZE]; /* calling device name */
  private String deviceString;
  private int[] len = new int[1]; /* length of call */
  private int[] fmt = new int[1]; /* call format */
  private int[] hbt = new int[1]; /* archiving heartbeat (not to be exceeded) */
  private int[] tmr = new int[1]; /* polling interval */
  private int[] infmt = new int[1]; /* input data format (if any) */
  private int[] inlen = new int[1]; /* number of input data points */
  private int[] opt = new int[1]; /* option word */
  private int[] active = new int[1];/* is table entry active */
  private int[] rec = new int[1]; /* entry in record table */  
  private void initStructure()
  {
    addField(fn_key,"fn_key");
    addField(tag,"tag");
    addField(ctx,"context");
    addField(srv,"server");
    addField(prp,"property");
    addField(dev,"device");
    addField(len,"len");
    addField(fmt,"fmt");
    addField(hbt,"hbt");
    addField(tmr,"tmr");
    addField(infmt,"infmt");
    addField(inlen,"inlen");
    addField(opt,"opt");
    addField(active,"active");
    addField(rec,"rec");
    initDone();   
    // some 'standard' default values:
    tmr[0] = 1000;
    hbt[0] = 900;
    infmt[0] = TFormat.CF_NULL;
  }
  public DAQRec()
  {
    super();
    initStructure();
  }
  public DAQRec(DAQRec reference)
  {
    super();
    initStructure();
    len[0] = reference.len[0];
    fmt[0] = TFormat.makeStdFormatCode(reference.fmt[0]);
    inlen[0] = reference.inlen[0];
    infmt[0] = TFormat.makeStdFormatCode(reference.infmt[0]);
    hbt[0] = reference.hbt[0];
    tmr[0] = reference.tmr[0];
    opt[0] = reference.opt[0];
    active[0] = reference.active[0];
    rec[0] = reference.rec[0];
    setKeyFileName(reference.getKeyFileName());
    setInputFileName(reference.getInputFileName());
    setContext(reference.getContext());
    setServer(reference.getServer());
    setDevice(reference.getDevice());
    setProperty(reference.getProperty());
  }
  public int getLength() { return len[0]; }
  public void setLength(int value) { len[0] = value; }
  public int getFormat() { return TFormat.makeByteFormatCode(fmt[0]); }
  public void setFormat(int value) { fmt[0] = TFormat.makeStdFormatCode(value); }
  public int getInputLength() { return inlen[0]; }
  public void setInputLength(int value) { inlen[0] = value % 256; }
  public int getInputFormat() { return TFormat.makeByteFormatCode(infmt[0]); }
  public void setInputFormat(int value) { infmt[0] = TFormat.makeStdFormatCode(value); }
  public int getHeartbeat() { return hbt[0]; }
  public void setHeartbeat(int value) { hbt[0] = value; }
  public int getTimer() { return tmr[0]; }
  public void setTimer(int value) { tmr[0] = value; }
  public int getFilterOption() { return opt[0]; }
  public void setFilterOption(int value) { opt[0] = value; }
  public int getActiveFlag() { return active[0]; }
  public void setActiveFlag(int value) { active[0] = value; }
  public int getRecordNumber() { return rec[0]; }
  public void setRecordNumber(int value) { rec[0] = value; }
  public String getKeyFileName()
  {
    if (fn_keyString == null) fn_keyString = new String(fn_key).trim();
    return fn_keyString;
  }
  public void setKeyFileName(String fileName)
  {
    pushChars(fileName, fn_key);
    fn_keyString = new String(fileName);
  }
  public String getInputFileName()
  {
    if (tagString == null) tagString = new String(tag).trim();
    return tagString;
  }
  public String getTag()
  {
    return getInputFileName();
  }
  public void setTag(String tag)
  {
    pushChars(tag, this.tag);
    tagString = new String(tag);
  }
  public void setInputFileName(String fileName)
  {
    pushChars(fileName, tag);
    tagString = new String(fileName);
  }
  public String getContext()
  {
    if (contextString == null) contextString = new String(ctx).trim();
    return contextString;
  }
  public void setContext(String context)
  {
    pushChars(context, ctx);
    contextString = new String(context);
  }
  public String getServer()
  {
    if (serverString == null) serverString = new String(srv).trim();
    return serverString;
  }
  public void setServer(String server)
  {
    pushChars(server, srv);
    serverString = new String(server);
  }
  public String getDevice()
  {
    if (deviceString == null) deviceString = new String(dev).trim();
    return deviceString;
  }
  public void setDevice(String device)
  {
    pushChars(device, dev);
    deviceString = new String(device);
  }
  public String getProperty()
  {
    if (propertyString == null) propertyString = new String(prp).trim();
    return propertyString;
  }
  public void setProperty(String property)
  {
    pushChars(property, prp);
    propertyString = new String(property);
  }
}
