/*
 * File HstCfg.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.histUtils;

import de.desy.tine.structUtils.TTaggedStructure;

public class HstCfg extends TTaggedStructure
{
  public void clear()
  {
    contextString = null;
    serverString = null;
    deviceString = null;
    propertyString = null;
    unitsString = null;
    descriptionString = null;
  }
  private char[] context = new char[32];
  private String contextString;
  private char[] server = new char[32];
  private String serverString;
  private char[] device = new char[64];
  private String deviceString;
  private char[] property = new char[64];
  private String propertyString;
  private char[] units = new char[16];
  private String unitsString;
  private char[] description = new char[64];
  private String descriptionString;
  private int index[] = new int[1];
  private int subIndex[] = new int[1];
  private float scale[] = new float[1];
  private float offset[] = new float[1];
  private float max[] = new float[1];
  private float min[] = new float[1];
  private int plot[] = new int[1];

  private void initStructure()
  {
    addField(context, "context");
    addField(server, "server");
    addField(device, "device");
    addField(property, "property");
    addField(units, "units");
    addField(description, "description");
    addField(index, "index");
    addField(subIndex, "subIndex");
    addField(scale, "scale");
    addField(offset, "offset");
    addField(max, "max");
    addField(min, "min");
    addField(plot, "plot");
    initDone();
  }
  public HstCfg()
  {
    super();
    initStructure();
  }
  public HstCfg(HstCfg reference)
  {
    super();
    initStructure();
    System.arraycopy(reference.context, 0, context, 0, 32);
    System.arraycopy(reference.server, 0, server, 0, 32);
    System.arraycopy(reference.device, 0, device, 0, 64);
    System.arraycopy(reference.property, 0, property, 0, 64);
    System.arraycopy(reference.units, 0, units, 0, 16);
    System.arraycopy(reference.description, 0, description, 0, 64);
    index[0] = reference.index[0];
    subIndex[0] = reference.subIndex[0];
    scale[0] = reference.scale[0];
    offset[0] = reference.offset[0];
    max[0] = reference.max[0];
    min[0] = reference.min[0];
    plot[0] = reference.plot[0];
  }
  public String getContext()
  {
    if (contextString == null)
      contextString = new String(context).trim();
    return contextString;
  }
  public String getDescription()
  {
    if (descriptionString == null)
      descriptionString = new String(description).trim();
    return descriptionString;
  }
  public String getDevice()
  {
    if (deviceString == null)
      deviceString = new String(device).trim();
    return deviceString;
  }
  public int getIndex()
  {
    return index[0];
  }
  public float getMax()
  {
    return max[0];
  }
  public float getMin()
  {
    return min[0];
  }
  public float getOffset()
  {
    return offset[0];
  }
  public int getPlot()
  {
    return plot[0];
  }
  public int getGraphMode()
  {
    return (plot[0] & 0xffff);
  }
  public int getDrawMode()
  {
    return (plot[0] >> 16) & 0xffff;
  }
  public String getProperty()
  {
    if (propertyString == null)
      propertyString = new String(property).trim();
    return propertyString;
  }
  public float getScale()
  {
    return scale[0];
  }
  public String getServer()
  {
    if (serverString == null)
      serverString = new String(server).trim();
    return serverString;
  }
  public int getSubIndex()
  {
    return subIndex[0];
  }
  public String getUnits()
  {
    if (unitsString == null)
      unitsString = new String(units).trim();
    return unitsString;
  }
  public void setContext(String context)
  {
    pushChars(context, this.context);
    this.contextString = new String(context);
  }
  public void setServer(String server)
  {
    pushChars(server, this.server);
    this.serverString = new String(server);
  }
  public void setDevice(String device)
  {
    pushChars(device, this.device);
    this.deviceString = new String(device);
  }
  public void setProperty(String property)
  {
    pushChars(property, this.property);
    this.propertyString = new String(property);
  }
  public void setDescription(String description)
  {
    pushChars(description, this.description);
    this.descriptionString = new String(description);
  }
  public void setIndex(int index)
  {
    this.index[0] = index;
  }
  public void setSubIndex(int subIndex)
  {
    this.subIndex[0] = subIndex;
  }
  public void setMax(float max)
  {
    this.max[0] = max;
  }
  public void setMin(float min)
  {
    this.min[0] = min;
  }
  public void setOffset(float offset)
  {
    this.offset[0] = offset;
  }
  public void setPlot(int plot)
  {
    this.plot[0] = plot;
  }
  public void setGraphMode(int mode)
  {
    plot[0] = (plot[0] & 0xffff0000) + (mode & 0xffff);
  }
  public void setDrawMode(int mode)
  {
    plot[0] = (mode << 16) + (plot[0] & 0xffff);
  }
  public void setScale(float scale)
  {
    this.scale[0] = scale;
  }
  public void setUnits(String units)
  {
    pushChars(units, this.units);
    this.unitsString = new String(units);
  }
}
