/*
 * File TArchiver.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.histUtils;

import java.io.IOException;
import java.util.Arrays;

import de.desy.tine.client.TLink;
import de.desy.tine.dataUtils.TDataType;
import de.desy.tine.definitions.TAccess;
import de.desy.tine.definitions.TErrorList;
import de.desy.tine.server.logger.DbgLog;
import de.desy.tine.stringUtils.KeyAliasPair;
import de.desy.tine.stringUtils.NameToString;
import de.desy.tine.stringUtils.NamesToKeyAlias;
import de.desy.tine.types.*;

public final class TArchiver
{
  private static final int number_stock_filters = 10;
  private static final int nice_query_size = 5000;
  private static final int filter_table_size = 32;
  private static DAQRec[] recs = new DAQRec[nice_query_size];
  private static DAQPrm[] prms = new DAQPrm[nice_query_size/10];
  private static DAQFltr[] fltrs = new DAQFltr[filter_table_size];
  private static boolean isInitialized = false;
  private static void initStructs()
  {
    if (isInitialized) return;
    for (int i=0; i<nice_query_size; i++)
    {
      recs[i] = new DAQRec();
    }
    for (int i=0; i<nice_query_size/10; i++)
    {
      prms[i] = new DAQPrm();
    }
    for (int i=0; i<filter_table_size; i++)
    {
      fltrs[i] = new DAQFltr();
    }
    isInitialized = true;
  }
  private static String lastTransactionStatus = null;
  public static String getLastTransactionStatus()
  {
    if (lastTransactionStatus == null) return "none";
    return lastTransactionStatus;
  }
  public static NAME64DBLDBL[] getKeywordPriorRecords(String context, String keyword) throws IOException
  {
    if (context == null) return null;
    NAME64DBLDBL[] priors = new NAME64DBLDBL[100];
    TDataType dtp = new TDataType(priors);
    String dname = "/" + context + "/HISTORY/" + keyword;
    TLink tl = new TLink(dname,"KEYWORD.PRIORS",dtp,null,TAccess.CA_READ);
    lastTransactionStatus = tl.getFullDeviceNameAndProperty() + " : " + "Success";
    if (tl.executeAndClose(1000) != 0) 
    {
      lastTransactionStatus = tl.getFullDeviceNameAndProperty() + " : " + tl.getLastError();
      throw new IOException(lastTransactionStatus);
    }
    int len = dtp.getCompletionLength();
    if (len == 100) DbgLog.log("getKeywordPriorRecords", "acquired exactly "+len+" records -> increase query size?");
    return Arrays.copyOf(priors, len);
  }
  public static String[] getKeywordPriorRecords(String context) throws IOException
  {
    if (context == null) return null;
    NAME64[] priors = new NAME64[100];
    TDataType dtp = new TDataType(priors);
    String dname = "/" + context + "/HISTORY/keyword";
    TLink tl = new TLink(dname,"KEYWORD.PRIORS",dtp,null,TAccess.CA_READ);
    lastTransactionStatus = tl.getFullDeviceNameAndProperty() + " : " + "Success";
    if (tl.executeAndClose(1000) != 0) 
    {
      lastTransactionStatus = tl.getFullDeviceNameAndProperty() + " : " + tl.getLastError();
      throw new IOException(lastTransactionStatus);
    }
    int len = dtp.getCompletionLength();
    if (len == 100) DbgLog.log("getKeywordPriorRecords", "acquired exactly "+len+" records -> increase query size?");
    return NameToString.nameArrayToString(Arrays.copyOf(priors, len));
  }
  public static int setKeywordPriorRecords(String context, String keyword, NAME64DBLDBL[] priors) 
  {
    if (context == null) return TErrorList.illegal_context;
    TDataType dtp = new TDataType(priors);
    String dname = "/" + context + "/HISTORY/" + keyword;
    TLink tl = new TLink(dname,"KEYWORD.PRIORS",null,dtp,TAccess.CA_WRITE);
    return tl.executeAndClose(1000); 
  }
  /** 
   * keyword-alias pairs: keyword is the stored keyword, alias is the current keyword 
   */
  public static KeyAliasPair[] getKeywordPriorDevices(String context, String keyword, String priorKeyword) throws IOException
  {
    if (context == null || keyword == null) return null;
    NAME64[] priors = new NAME64[200];
    NAME64[] pkwrd = new NAME64[] { new NAME64(keyword) };
    TDataType dtpd = new TDataType(priors);
    if (priorKeyword == null) priorKeyword = keyword;
    String dname = "/" + context + "/HISTORY/" + priorKeyword;
    TLink tl = new TLink(dname,"KEYWORD.PRIOR.DEVICES",dtpd,new TDataType(pkwrd),TAccess.CA_READ);
    lastTransactionStatus = tl.getFullDeviceNameAndProperty() + " : " + "Success";
    if (tl.executeAndClose(1000) != 0) 
    {
      lastTransactionStatus = tl.getFullDeviceNameAndProperty() + " : " + tl.getLastError();
      throw new IOException(lastTransactionStatus);
    }
    int len = dtpd.getCompletionLength();
    if (len == 200) DbgLog.log("getKeywordPriorDevices", "acquired exactly "+len+" records -> increase query size?");
    return NamesToKeyAlias.nameArrayToKeyAlias(Arrays.copyOf(priors, len));
  }
  public static int setKeywordPriorDevices(String context, String priorKeyword, KeyAliasPair[] lst)
  {
    if (context == null) return TErrorList.illegal_context;
    if (priorKeyword == null) return TErrorList.invalid_keyword;
    if (lst == null) return TErrorList.invalid_parameter;
    NAME64[] priors = new NAME64[lst.length*2];
    for (int i=0; i<lst.length; i++)
    {
      priors[2*i] = new NAME64(lst[i].getKeyword());
      priors[2*i+1] = new NAME64(lst[i].getAlias());
    }
    String dname = "/" + context + "/HISTORY/" + priorKeyword;
    TLink tl = new TLink(dname,"KEYWORD.PRIOR.DEVICES",null,new TDataType(priors),TAccess.CA_WRITE);
    return tl.executeAndClose(1000); 
  }
  public static DAQRec[] getDAQRecords(String context) throws IOException
  {
    if (context == null) return null;
    if (!isInitialized) initStructs();
    for (DAQRec r : recs) r.clear();
    TDataType dtrecs = new TDataType(recs);
    String dname = "/" + context + "/HISTORY";
    TLink tl = new TLink(dname,"RECORD.INFO",dtrecs,null,TAccess.CA_READ);
    lastTransactionStatus = tl.getFullDeviceNameAndProperty() + " : " + "Success";
    if (tl.executeAndClose(1000) != 0) 
    {
      lastTransactionStatus = tl.getFullDeviceNameAndProperty() + " : " + tl.getLastError();
      throw new IOException(lastTransactionStatus);
    }
    int len = dtrecs.getCompletionLength();
    if (len == recs.length) DbgLog.log("getDAQRecords", "acquired exactly "+len+" records -> increase query size?");
    DAQRec[] results = new DAQRec[len];
    for (int i=0; i<len; i++) results[i] = new DAQRec(recs[i]);
    return results;
  }
  public static DAQRec getRecordFromKeyword(String context, String keyword) throws IOException
  {
    if (context == null) return null;
    if (!isInitialized) initStructs();
    DAQRec[] thisRec = new DAQRec[] {new DAQRec()};
    thisRec[0].clear();
    TDataType dt = new TDataType(thisRec);
    dt.dArrayLength = 1;
    String dname = "/" + context + "/HISTORY/" + keyword;
    TLink tl = new TLink(dname,"RECORD.INFO",dt,null,TAccess.CA_READ);
    lastTransactionStatus = tl.getFullDeviceNameAndProperty() + " : " + "Success";
    if (tl.executeAndClose(1000) != 0) 
    {
      lastTransactionStatus = tl.getFullDeviceNameAndProperty() + " : " + tl.getLastError();
      throw new IOException(lastTransactionStatus);
    }
    return thisRec[0];    
  }
  public static int setDAQRecords(String context,DAQRec[] recs) throws IOException
  {
    if (context == null || recs == null) return TErrorList.argument_list_error;
    if (!isInitialized) initStructs();
    TDataType dtrecs = new TDataType(recs);
    String dname = "/" + context + "/HISTORY";
    TLink tl = new TLink(dname,"RECORD.INFO",null,dtrecs,TAccess.CA_WRITE);
    int cc = tl.executeAndClose(2000);
    lastTransactionStatus = tl.getFullDeviceNameAndProperty() + " : " + tl.getLastError();      
    if (cc == TErrorList.connection_timeout || cc == TErrorList.link_timeout)
    {
      throw new IOException(lastTransactionStatus);
    }
    return cc;
  }
  public static DAQPrm[] getDAQKeyInfo(String context,String keyFile) throws IOException
  {
    if (context == null || keyFile == null) return null;
    if (!isInitialized) initStructs();
    for (DAQPrm p : prms) p.clear();
    TDataType dtprms = new TDataType(prms);
    String dname = "/" + context + "/HISTORY/"+keyFile;
    TLink tl = new TLink(dname,"KEYWORD.INFO",dtprms,null,TAccess.CA_READ);
    lastTransactionStatus = tl.getFullDeviceNameAndProperty() + " : " + "Success";
    if (tl.executeAndClose(1000) != 0) 
    {
      lastTransactionStatus = tl.getFullDeviceNameAndProperty() + " : " + tl.getLastError();
      throw new IOException(lastTransactionStatus);
    }
    int len = dtprms.getCompletionLength();
    DAQPrm[] results = new DAQPrm[len];
    for (int i=0; i<len; i++) results[i] = new DAQPrm(prms[i]);
    return results;
  }
  public static int setDAQKeyInfo(String context,String keyFile,DAQPrm[] prms) throws IOException
  {
    if (context == null || keyFile == null || prms == null) 
      return TErrorList.argument_list_error;
    if (context.length() == 0 || keyFile.length() == 0 || prms.length == 0)
      return TErrorList.argument_list_error;
    if (!isInitialized) initStructs();
    TDataType dtprms = new TDataType(prms);
    if (!keyFile.endsWith(".csv")) keyFile += ".csv";
    String dname = "/" + context + "/HISTORY/"+keyFile;
    TLink tl = new TLink(dname,"KEYWORD.INFO",null,dtprms,TAccess.CA_WRITE);
    int cc = tl.executeAndClose(2000);
    lastTransactionStatus = tl.getFullDeviceNameAndProperty() + " : " + tl.getLastError();      
    if (cc == TErrorList.connection_timeout || cc == TErrorList.link_timeout)
      throw new IOException(lastTransactionStatus);
    return cc;
  }
  public static TDataType getDAQRecInput(String context,DAQRec rec) throws IOException
  {
    if (context == null || rec == null) return null;
    if (!isInitialized) initStructs();
    TDataType dtinpt = new TDataType(rec.getInputLength(),(short)rec.getInputFormat());
    String dname = "/" + context + "/HISTORY/"+rec.getInputFileName();
    TLink tl = new TLink(dname,"KEYWORD.INPT",dtinpt,null,TAccess.CA_READ);
    lastTransactionStatus = "Success";
    if (tl.executeAndClose(1000) != 0) 
    {
      lastTransactionStatus = tl.getFullDeviceNameAndProperty() + " : " + tl.getLastError();
      throw new IOException(lastTransactionStatus);
    }
    return dtinpt;
  }
  public static int setDAQRecInput(String context,DAQRec rec,TDataType din) throws IOException
  {
    if (context == null || rec == null || din == null) 
      return TErrorList.argument_list_error;
    if (!isInitialized) initStructs();
    String dname = "/" + context + "/HISTORY/"+rec.getInputFileName();
    TLink tl = new TLink(dname,"KEYWORD.INPT",null,din,TAccess.CA_WRITE);
    int cc = tl.executeAndClose(2000);
    lastTransactionStatus = tl.getFullDeviceNameAndProperty() + " : " + tl.getLastError();      
    if (cc == TErrorList.connection_timeout || cc == TErrorList.link_timeout)
      throw new IOException(lastTransactionStatus);
    return cc;    
  }
  public static DAQFltr[] getDAQFilterTable(String context) throws IOException
  {
    if (context == null) return null;
    if (!isInitialized) initStructs();
    for (DAQFltr f : fltrs) f.clear();
    TDataType dtfltrs = new TDataType(fltrs);
    String dname = "/" + context + "/HISTORY";
    TLink tl = new TLink(dname,"FILTERTABLE",dtfltrs,null,TAccess.CA_READ);
    lastTransactionStatus = "Success";
    if (tl.executeAndClose(1000) != 0)
    {
      lastTransactionStatus = tl.getFullDeviceNameAndProperty() + " : " + tl.getLastError();
      throw new IOException(lastTransactionStatus);
    }
    int len = dtfltrs.getCompletionLength();
    DAQFltr[] results = new DAQFltr[len];
    for (int i=0; i<len; i++) results[i] = new DAQFltr(fltrs[i]);
    return results;
  }
  public static int setDAQFilterTable(String context,DAQFltr[] fltrs) throws IOException
  {
    if (context == null || fltrs == null) return TErrorList.argument_list_error;
    if (!isInitialized) initStructs();
    TDataType dtfltrs = new TDataType(fltrs);
    String dname = "/" + context + "/HISTORY";
    TLink tl = new TLink(dname,"FILTERTABLE",null,dtfltrs,TAccess.CA_WRITE);
    int cc = tl.executeAndClose(2000);
    lastTransactionStatus = tl.getFullDeviceNameAndProperty() + " : " + tl.getLastError();      
    if (cc == TErrorList.connection_timeout || cc == TErrorList.link_timeout)
      throw new IOException(lastTransactionStatus);
    return cc;
  }
  public int getFilterOptionFromString(DAQFltr[] fltrs,String filterString)
  {
    int v = 0;
    if (fltrs == null || filterString == null) return 0;
    for (int i = 0; i<fltrs.length; i++)
    {
      if (filterString.contains(fltrs[i].getTag()))
      {
        v += (1<<i);
      }
    }
    return v;    
  }
  public static String getFilterOptionAsString(DAQFltr[] fltrs,int filterOptions)
  {
    String filterString = "";
    if (fltrs == null || filterOptions == 0) return filterString;
    int b = 0;
    for (int i = 0; i<fltrs.length; i++)
    {
      b = 1 << i;
      if ((filterOptions & b) == b)
      {
         if (filterString.length() > 0) filterString +=  "|";
         filterString += fltrs[i].getTag();
      }
    }
    return filterString;
  }
  public static int reloadDatabase(String context) throws IOException
  {
    return reloadDatabase(context, null);
  }
  public static int reloadDatabase(String context, String server) throws IOException
  {
    lastTransactionStatus = null;
    if (context == null) return TErrorList.argument_list_error;
    if (!isInitialized) initStructs();
    boolean reloadAll = server == null;
    if (reloadAll) server = "HISTORY";
    String dname = "/" + context + "/" + server;
    TLink tl = new TLink(dname,"DB.RELOAD",null,null,TAccess.CA_WRITE);
    int cc = tl.executeAndClose(3000);
    if (cc == TErrorList.connection_timeout || cc == TErrorList.link_timeout)
      throw new IOException(tl.getLastError());
    if (cc != 0) 
    {
      lastTransactionStatus = tl.getFullDeviceNameAndProperty() + " : " + tl.getLastError();
      return cc;
    }
    if (!reloadAll) return cc; // then we've reloaded the one we wanted already ...
    server = "ARCHIVER";
    dname = "/" + context + "/" + server;
    tl = new TLink(dname,"DB.RELOAD",null,null,TAccess.CA_WRITE);
    cc = tl.executeAndClose(3000);
    lastTransactionStatus = tl.getFullDeviceNameAndProperty() + " : " + tl.getLastError();      
    if (cc == TErrorList.connection_timeout || cc == TErrorList.link_timeout)
      throw new IOException(lastTransactionStatus);
    return cc;
  }
  public static int updateServer(String context) throws IOException
  {
    if (context == null) return TErrorList.argument_list_error;
    if (!isInitialized) initStructs();
    String dname = "/" + context + "/HISTORY";
    TLink tl = new TLink(dname,"RESTART",null,null,TAccess.CA_WRITE);
    int cc = tl.executeAndClose(3000);
    lastTransactionStatus = tl.getFullDeviceNameAndProperty() + " : " + tl.getLastError();      
    if (cc == TErrorList.connection_timeout || cc == TErrorList.link_timeout)
      throw new IOException(lastTransactionStatus);
     return cc;
  }
  public static int getNumberStockFilters()
  {
    return number_stock_filters;
  }
}
