/*
 * File TBucket.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

/*
 * Created on Feb 6, 2006
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package de.desy.tine.io;

import java.io.*;
import java.net.*;

import de.desy.tine.client.TLinkFactory;
import de.desy.tine.definitions.TErrorList;
import de.desy.tine.server.logger.DbgLog;
import de.desy.tine.server.logger.MsgLog;

public class TBucket
{
  public final static int ALLOWED_IDLE_TIME = 300000;
  Socket sck; // TCP connection socket
  public Socket getSocket() { return sck; }
  public void setSocket(Socket socket) { sck = socket; }
  int ptr; // bucket pointer
  public int getBucketPointer() { return ptr; }
  public void setBucketPointer(int pointer) { ptr = pointer; }
  private int siz; // bucket target size
  private int capacity = TPacket.MAX_DATAGRAM_SIZE;
  public int getBucketSize() { return siz; }
  public void setBucketSize(int bucketSize) { siz = bucketSize; }
  private byte[] buf = null; // bucket buffer
  public byte[] getBucketBuffer() { return buf; }
  int port;
  int localPort;
  public int getBucketPort() { return port; }
  public int getLocalPort() { return sck != null ? sck.getLocalPort() : 0; }
  public void setBucketPort(int portAddr) { port = portAddr; }
  InetAddress iaddr;
  public InetAddress getBucketEndpoint() { return iaddr; }
  public void setBucketEndpoint(InetAddress inetAddr) { iaddr = inetAddr; }
  public boolean active = false;
  public boolean isDeactivating = false;
  int number;
  private long lastActive = System.currentTimeMillis(); // creation time
  public long getTimeLastActive() { return lastActive; }
  public void touch() { lastActive = System.currentTimeMillis(); }
  private Thread bckThrd;
  OutputStream os;
  InputStream is;
  public Thread getBucketThread() { return bckThrd; }
  public void setBucketThread(Thread thread) { bckThrd = thread; } 
  public InputStream getInputStream() { return is; }
  public void setInputStream(InputStream input) { is = input; }
  public OutputStream getOutputStream() { return os; }
  public void setOutputStream(OutputStream output) { os = output; }
  public void reconnect()
  {
    if (sck == null) return;
    try
    {
      sck.close();
      os.close();
      is.close();
      initBucket(iaddr, port, bckThrd, capacity);
      //sck.connect(new InetSocketAddress(iaddr,port));;
    }
    catch (Exception e)
    {
      e.printStackTrace();
      MsgLog.log("TBucket.reconnect",e.toString(),TErrorList.code_failure,e,0);      
    }
  }
  public void setCapacity(int newBufferSize)
  {
    byte[] b=null;
    try
    {
      b = new byte[newBufferSize];
    }
    catch (OutOfMemoryError e)
    {
      MsgLog.log("TBucket.setCapacity", "not enough memory to allocate "+newBufferSize+" bytes",TErrorList.out_of_local_memory,e,0);
      return;
    }
    buf = b;
    capacity = newBufferSize;
  }
  public int getCapacity() { return capacity; }
  protected void initBucket(Socket socket,Thread thread,int capacity)
  {
    sck = socket;
    try
    {
      int bufsiz = TLinkFactory.getSckRcvBufferSize();
      if (bufsiz > 0x1000) 
      {
        sck.setReceiveBufferSize(bufsiz);
        MsgLog.log("TBucket.initBucket","set server socket receive buffer size to "+bufsiz,0,null,1);
      }
      os = socket.getOutputStream();  
      is = socket.getInputStream(); 
      iaddr = socket.getInetAddress();
      port = socket.getPort();
      bckThrd = thread;
      this.capacity = capacity;
      buf = new byte[capacity];
    }
    catch (Exception e)
    {
      e.printStackTrace();
      MsgLog.log("TBucket.initBucket",e.toString(),TErrorList.code_failure,e,0);
    }    
  }
  public TBucket() 
  { 
  }
  protected int initBucket(InetAddress inetAddr,int portAddr,Thread thread,int capacity)
  {
    try
    {
      int bufsiz = TLinkFactory.getSckRcvBufferSize();
      sck = new Socket(inetAddr,portAddr);
      if (thread != null) thread.setName("tcp port " + sck.getLocalPort());
      if (bufsiz > 0x1000) 
      {
        sck.setReceiveBufferSize(bufsiz);
        MsgLog.log("TBucket.initBucket","set client socket receive buffer size to "+bufsiz,0,null,1);
      }
      bufsiz = TLinkFactory.getSckSndBufferSize();
      if (bufsiz > 0x1000) 
      {
        sck.setSendBufferSize(bufsiz);
        MsgLog.log("TBucket.initBucket","set client socket send buffer size to "+bufsiz,0,null,1);
      }      
      os = sck.getOutputStream();  
      is = sck.getInputStream(); 
      iaddr = inetAddr;
      port = sck.getPort();
      bckThrd = thread;
      this.capacity = capacity;
      buf = new byte[capacity];
    }
    catch (IOException e)
    { // host not there ?
      if (TLinkFactory.debugLevel > 1) DbgLog.log("TBucket.initBucket", "cannot open connected socket");
      return TErrorList.tcp_connect_error;
    }
    catch (Exception e)
    {
      e.printStackTrace();
      MsgLog.log("TBucket.initBucket",e.toString(),TErrorList.code_failure,e,0);
      return TErrorList.tcp_connect_error;
    }
    return 0;
  }
  public void activate()
  {
    if (bckThrd != null) bckThrd.start(); 
  }
}
