/*
 * File TDataOutputStream.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

/*
 * Created on Feb 23, 2005
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package de.desy.tine.io;

import java.io.DataOutput;
import java.io.DataOutputStream;
import java.io.FilterOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import de.desy.tine.endianUtils.SwapUtil;


/**
 * \internal
 * 
 * A data output stream which performs byte swapping.
 * TDataOutputStream behaves like java.io.DataOutputStream,
 * but internally swaps bytes of numeric data types 
 *
 * @author MCS
 * @see DataOutput
 * @see DataOutputStream
 * @see FilterOutputStream
 * 
 * TODO A direct implementation without DataOutputStream would
 * improve the performance. 
 */
public class TDataOutputStream extends FilterOutputStream implements DataOutput {
  
  private DataOutputStream stream = null;
  
 /**
   * @param out
   */
  public TDataOutputStream(OutputStream out) {
    super(out);
    stream = new DataOutputStream(out);
  }

  
  /* (non-Javadoc)
   * @see java.io.FilterOutputStream#close()
   */
  public void close() throws IOException {
    stream.close();
  }
  
  /* (non-Javadoc)
   * @see java.io.FilterOutputStream#flush()
   */
  public void flush() throws IOException {
    stream.flush();
  }
  
  /**
   * @return
   */
  public int size() {
    return stream.size();
  }
  
  /* (non-Javadoc)
   * @see java.lang.Object#toString()
   */
  public String toString() {
    return stream.toString();
  }
  
  /* (non-Javadoc)
   * @see java.io.FilterOutputStream#write(byte[])
   */
  public void write(byte[] b) throws IOException {
    stream.write(b);
  }
  
  /* (non-Javadoc)
   * @see java.io.FilterOutputStream#write(byte[], int, int)
   */
  public void write(byte[] b, int off, int len) throws IOException {
    stream.write(b, off, len);
  }
  
  /* (non-Javadoc)
   * @see java.io.FilterOutputStream#write(int)
   */
  public void write(int b) throws IOException {
    stream.write(b);
  }
  
  /**
   * @param v
   * @throws java.io.IOException
   */
  public void writeBoolean(boolean v) throws IOException {
    stream.writeBoolean(v);
  }
  
  /**
   * @param v
   * @throws java.io.IOException
   */
  public void writeByte(int v) throws IOException {
    stream.writeByte(v);
  }
  
  /**
   * @param s
   * @throws java.io.IOException
   */
  public void writeBytes(String s) throws IOException {
    stream.writeBytes(s);
  }
  
  /**
   * @param v
   * @throws java.io.IOException
   */
  public void writeChar(int v) throws IOException {
    stream.writeChar(SwapUtil.swapShort((short)v));
  }
  
  /**
   * @param s
   * @throws java.io.IOException
   */
  public void writeChars(String s) throws IOException {
    stream.writeChars(s);
  }
  
  /**
   * @param v
   * @throws java.io.IOException
   */
  public void writeDouble(double v) throws IOException {
    // avoid NaN !
    //stream.writeDouble(SwapUtil.swapDouble(v));
    stream.writeLong(SwapUtil.swapLong(Double.doubleToLongBits(v)));
  }
  
  /**
   * @param v
   * @throws java.io.IOException
   */
  public void writeFloat(float v) throws IOException {
    // avoid NaN !
    //stream.writeFloat(SwapUtil.swapFloat(v));
    stream.writeInt(SwapUtil.swapInt(Float.floatToIntBits(v)));
  }
  
  /**
   * @param v
   * @throws java.io.IOException
   */
  public void writeInt(int v) throws IOException {
    stream.writeInt(SwapUtil.swapInt(v));
  }
  
  /**
   * @param v
   * @throws java.io.IOException
   */
  public void writeLong(long v) throws IOException {
    stream.writeLong(SwapUtil.swapLong(v));
  }
  
  /**
   * @param v
   * @throws java.io.IOException
   */
  public void writeShort(int v) throws IOException {
    stream.writeShort(SwapUtil.swapShort((short)v));
  }
  
  /**
   * @param str
   * @throws java.io.IOException
   * TODO Unclear if swapping is necessary here!
   */
  public void writeUTF(String str) throws IOException {
    stream.writeUTF(str);
  }
}
