/*
 * File TPacket.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

/*
 * Created on Nov 10, 2004
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package de.desy.tine.io;
//import java.io.*;
import java.net.*;

import de.desy.tine.definitions.TErrorList;
import de.desy.tine.server.logger.MsgLog;
import de.desy.tine.server.logger.TFecLog;
import de.desy.tine.server.logger.TineLogLevel;
import de.desy.tine.startup.TInitializer;
import de.desy.tine.startup.TInitializerFactory;

/**
 * \internal
 * TODO: This now appears to work -> just use a single MulticastSocket called 's'
 * 
 * @author duval
 *
 */
public class TPacket
{
  private static final int defaultRcvBufSiz = 0x10000;
  private static final int defaultSndBufSiz = 0x2000;
  private static final int defaultTtl = 16;
  public static final int MAX_DATAGRAM_SIZE = 65535; //64000;
  private static final int MCAST_REATTACH_INTERVAL = 300000;
  private TInitializer initializer=TInitializerFactory.getInstance().getInitializer();
  private MulticastSocket sck;
  private int port = 0;
  public int getPort() { return port; }
  public MulticastSocket getSocket() { return sck; }
  public boolean canAttachToServiceGroup = false;
  private static long lastjoin = 0;
  private String os = null;
  public boolean isServiceGroupStale()
  {
    return isServiceGroupStale(System.currentTimeMillis());
  }
  public boolean isServiceGroupStale(long now)
  {
    if (now < lastjoin + MCAST_REATTACH_INTERVAL) return false;
    return true;
  }
  public void reAttachToServiceGroup()
  { // some routers will drop the member if there's no activity for a long time
    if (!canAttachToServiceGroup) return;
    try
    {
      InetAddress grp = InetAddress.getByName(initializer.getNetCastAddress());
      sck.joinGroup(grp);
    }
    catch (Exception e)
    {
      MsgLog.log("reAttachToServiceGroup", "could not reattach to group "+e.getMessage(),TErrorList.io_error,e,1);      
    }
    lastjoin = System.currentTimeMillis();
  }
  public void shutdown()
  {
    if (sck == null) return;
    try
    {
      sck.close();
    }
    catch (Exception e)
    {
      e.printStackTrace();
      System.err.println("could not close socket on port "+sck.getLocalPort());
    }
  }
  InetAddress hostAddress;
  private boolean isMCastListener = false;
  public boolean isMulticastListener() { return isMCastListener; }
  public DatagramPacket dpOut;
  byte[] bufIn = new byte[MAX_DATAGRAM_SIZE];
  public DatagramPacket dpIn = new DatagramPacket(bufIn, bufIn.length);
  public TPacket()
  {
    this(0,defaultRcvBufSiz,defaultSndBufSiz,defaultTtl);
  }
  public TPacket(int rcvBufferSize, int timeToLive)
  {
    this(0,rcvBufferSize,defaultSndBufSiz,timeToLive);
  }
  public TPacket(int port)
  {
    this(port,defaultRcvBufSiz,defaultSndBufSiz,defaultTtl);
  }
  public TPacket(int port,int rcvBufferSize,int timeToLive)
  {
    this(port,rcvBufferSize,defaultSndBufSiz,timeToLive);
  }  
  public TPacket(int port,int rcvBufferSize,int sndBufferSize,int timeToLive)
  {
    int clnport = initializer.getClnPort();
    boolean useNextPort = (port == 0) ? true : false;
    boolean socketBound = false;
    InetAddress grp;
    String ip = TInitializerFactory.getInstance().getInitializer().getMyIpAddr();
    if (os == null) os = System.getProperty("os.name");
    InetAddress iaddr = null;
    try
    {
      iaddr = InetAddress.getByName(ip);
      if (os != null && os.contains("Linux")) iaddr = null;
      if (os != null && os.contains("Windows")) iaddr = null;
    }
    catch (Exception e)
    {
      System.err.println("Can't open socket");
      e.printStackTrace();
      MsgLog.log("TPacket",e.toString(),TErrorList.udp_socket_error,e,0);
      TFecLog.log(TineLogLevel.ERROR,"Cannot initialize socket: "+e.toString());
      return;
    }
    for (int i=0; !socketBound && i<1000; i++)
    {
      try
      {
        if (port == 0) 
        { // general client side port
          InetSocketAddress addr = new InetSocketAddress(iaddr,clnport+i);
          sck = new MulticastSocket(null);
          sck.setReuseAddress(false);
          sck.bind(addr);
          sck.setSoTimeout(0);
          sck.setReceiveBufferSize(rcvBufferSize);
          sck.setSendBufferSize(sndBufferSize);
          sck.setTimeToLive(timeToLive);
          socketBound = true;
          this.port = clnport + i;
          if (rcvBufferSize > 0x1000)
            MsgLog.log("TPacket", "set client socket ("+this.port+") receive buffer size to "+rcvBufferSize,0,null,1);
        } 
        else if (port == initializer.getGCastPort()) 
        { // globals listening port
          grp = InetAddress.getByName(initializer.getGCastAddress());
          sck = new MulticastSocket(port);
          sck.joinGroup(grp); // this is the canonical globals multicast group
          sck.setSoTimeout(0);
          sck.setTimeToLive(timeToLive);
          sck.setSendBufferSize(sndBufferSize);
          socketBound = true;
          isMCastListener = true;
          this.port = port;
        } 
        else if (port == initializer.getMCastPort()) 
        { // general multi-cast listening port
          grp = InetAddress.getByName(initializer.getMCastAddress());
          sck = new MulticastSocket(port);
          sck.joinGroup(grp); // this is the canonical server multicast group
          sck.setSoTimeout(0);
          sck.setReceiveBufferSize(rcvBufferSize);
          sck.setSendBufferSize(sndBufferSize);
          sck.setTimeToLive(timeToLive);
          socketBound = true;
          isMCastListener = true;
          this.port = port;
          if (rcvBufferSize > 0x1000)
            MsgLog.log("TPacket", "set multicast client socket ("+this.port+") receive buffer size to "+rcvBufferSize,0,null,1);
        } 
        else if (port == initializer.getNetCastPort()) 
        { // server port for responding to services request
          grp = InetAddress.getByName(initializer.getNetCastAddress());
          sck = new MulticastSocket(port);
          sck.joinGroup(grp);
          sck.setSoTimeout(0);
          sck.setReceiveBufferSize(rcvBufferSize);
          sck.setSendBufferSize(sndBufferSize);
          socketBound = true;
          isMCastListener = true;
          this.port = port;
          lastjoin = System.currentTimeMillis();
          canAttachToServiceGroup = true;
        } 
        else 
        { // server's listening port
          InetSocketAddress addr = new InetSocketAddress(iaddr,port);
          sck = new MulticastSocket(null);
          sck.setReuseAddress(false);
          sck.bind(addr);          
          sck.setSoTimeout(0);
          sck.setReceiveBufferSize(rcvBufferSize);
          sck.setSendBufferSize(sndBufferSize);
          sck.setTimeToLive(timeToLive);
          socketBound = true;
          this.port = port;
          if (rcvBufferSize > 0x1000)
            MsgLog.log("TPacket", "set server socket ("+this.port+") receive buffer size to "+rcvBufferSize,0,null,1);
        }
      }
      catch (Exception e)
      {
        if (!useNextPort)
        {
          System.err.println("Can't open socket");
          e.printStackTrace();
          MsgLog.log("TPacket",e.toString(),TErrorList.udp_socket_error,e,0);
          TFecLog.log(TineLogLevel.ERROR,"Cannot initialize socket: "+e.toString());
          break;
        }
      }
    }
  }
}
