/*
 * File PropertyQueryLegacy.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.queryUtils;
/**
 * @version 	1.0
 * @author
 */
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

import de.desy.tine.definitions.TErrorList;
import de.desy.tine.endianUtils.Swap;
import de.desy.tine.server.logger.MsgLog;

public class PropertyQueryLegacy
{
  //  Extended Property Query structure (legacy -2)
  public  String    prpName;       // 32 character Property Name
  public  String    prpDescription;// 64 character Property Description
  public  String    prpRedirection;// 32 character Property redirection string
  public  String    prpTag;        // 8 character data tag (outbound)
  public  String    prpTagIn;      // 8 character data tag (inbound)
  public  String    prpUnits;      // 16 character engineering units
  public  float     prpMinValue;  // minimum value
  public  float     prpMaxValue;  // maximum value
  public  int       prpSize;      // Property dimension (outbound)
  public  int       prpSizeIn;    // Property dimension (inbound)
  public  int       prpNumOverloads;// number of overloads for Property Name
  public  short     prpHistoryDepthShort;// Short term history depth (ring buffer size)
  public  short     prpHistoryDepthLong; // Long term history depth in months
  public  byte      prpFormat;    // Datatype (outbound)  
  public  byte      prpFormatIn;  // Datatype (inbound)
  public  byte      prpAccess;    // Access mode(s) for Proptery
  public  byte      prpGraphType; // Graph stype for Property
  
  public static int sizeInBytes = 188;
  ByteArrayOutputStream dBuffer = new ByteArrayOutputStream(sizeInBytes);
  byte[] hByteArray;
  public PropertyQueryLegacy()
  {
    prpName = new String(new StringBuffer(32));
    prpDescription = new String(new StringBuffer(64));
    prpRedirection = new String(new StringBuffer(32));
    prpTag = new String(new StringBuffer(8));
    prpTagIn = new String(new StringBuffer(8));
    prpUnits = new String(new StringBuffer(16));
  }  
  public byte[] toByteArray()
  {
    try
    {
      dBuffer.reset();
      DataOutputStream ds = new DataOutputStream(dBuffer);
      // prpName
      byte[] bname = prpName.getBytes();
      byte[] b = new byte[sizeInBytes];
      for (int i=0; i<32 && i<bname.length; i++) b[i] = bname[i];
      ds.write(b,0,32);
      // prpDescription
      bname = prpDescription.getBytes();
      for (int i=0; i<64 && i<bname.length; i++) b[i] = bname[i];
      ds.write(b,0,64);
      // prpRedirection
      bname = prpRedirection.getBytes();
      for (int i=0; i<32 && i<bname.length; i++) b[i] = bname[i];
      ds.write(b,0,32);
      // prpTag
      bname = prpTag.getBytes();
      for (int i=0; i<8 && i<bname.length; i++) b[i] = bname[i];
      ds.write(b,0,8);
      // prpTagIn
      bname = prpTagIn.getBytes();
      for (int i=0; i<8 && i<bname.length; i++) b[i] = bname[i];
      ds.write(b,0,8);
      // prpUnits
      bname = prpUnits.getBytes();
      for (int i=0; i<16 && i<bname.length; i++) b[i] = bname[i];
      ds.write(b,0,16);
      // prpMinValue;
      ds.write(Swap.Float(prpMinValue),0,4);
      // prpMaxValue;
      ds.write(Swap.Float(prpMaxValue),0,4);
      // prpSize;
      ds.writeInt(Swap.Int(prpSize));
      // prpSizeIn;
      ds.writeInt(Swap.Int(prpSizeIn));
      // prpNumOverloads;
      ds.writeInt(Swap.Int(prpNumOverloads));
      // prpHistoryDepthShort;
      ds.writeInt(Swap.Short(prpHistoryDepthShort));
      // prpHistoryDepthLong;
      ds.writeInt(Swap.Short(prpHistoryDepthLong));
      // prpFormat;   
      ds.writeByte(prpFormat);
      // prpFormatIn; 
      ds.writeByte(prpFormatIn);
      // prpAccess;   
      ds.writeByte(prpAccess);
      // prpGraphType;
      ds.writeByte(prpGraphType);
      ds.close();
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("PropertyQueryLegacy",e.toString(),TErrorList.code_failure,e,0);
      return null;
    }
    hByteArray = dBuffer.toByteArray();
    return hByteArray;
  }  
  public void toStruct()
  {
    int len;
    try
    {
      ByteArrayInputStream dis = new ByteArrayInputStream(hByteArray);
      DataInputStream ds = new DataInputStream(dis);
      // prpName
      byte[] b = new byte[64];
      ds.read(b,0,32);
      len = 0; for (int i=0; i<32 && b[i] != 0; i++, len++);
      prpName = new String(b,0,len);
      // prpDescription      
      ds.read(b,0,64);
      len = 0; for (int i=0; i<64 && b[i] != 0; i++, len++);
      prpDescription = new String(b,0,len);
      // prpRedirection
      ds.read(b,0,32);
      len = 0; for (int i=0; i<32 && b[i] != 0; i++, len++);
      prpRedirection = new String(b,0,len);
      // prpTag
      ds.read(b,0,8);
      len = 0; for (int i=0; i<8 && b[i] != 0; i++, len++);
      prpTag = new String(b,0,len);
      // prpTagIn
      ds.read(b,0,8);
      len = 0; for (int i=0; i<8 && b[i] != 0; i++, len++);
      prpTagIn = new String(b,0,len);
      // prpUnits
      ds.read(b,0,16);
      len = 0; for (int i=0; i<16 && b[i] != 0; i++, len++);
      prpUnits = new String(b,0,len);
      // prpMinValue
      ds.read(b,0,4);
      prpMinValue = Swap.Float(b);
      // prpMaxValue
      ds.read(b,0,4);
      prpMaxValue = Swap.Float(b);
      // prpSize
      prpSize = Swap.Int(ds.readInt());
      // prpSizeIn
      prpSizeIn = Swap.Int(ds.readInt());
      // prpNumOverloads
      prpNumOverloads = Swap.Int(ds.readInt());
      // prpHistoryDepthShort
      prpHistoryDepthShort = Swap.Short(ds.readShort());
      // prpHistoryDepthLong
      prpHistoryDepthLong = Swap.Short(ds.readShort());
      // prpFormat
      prpFormat = ds.readByte();
      // prpFormatIn
      prpFormatIn = ds.readByte();
      // prpAccess
      prpAccess = ds.readByte();
      // prpGraphType
      prpGraphType = ds.readByte();
      ds.close(); dis.close();
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("PropertyQueryLegacy",e.toString(),TErrorList.code_failure,e,0);
    }
  }  
}