/*
 * File XPropertyQuery.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.queryUtils;

/**
 * @version 1.0
 * @author
 * CHANGES
 * 2005-02-23 JW toByteArray() removed, toStruct() removed, hByteArray removed.
 */
import java.io.ByteArrayOutputStream;
import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import de.desy.tine.server.properties.TExportProperty;
import de.desy.tine.stringUtils.StringToBytes;
import de.desy.tine.structUtils.TStructBase;

public class XPropertyQuery extends TStructBase
{
  // Extended Property Query structure (legacy -1)
  public String prpName; // 32 character Property Name
  public String prpDescription;// 64 character Property Description
  public String prpRedirection;// 32 character Property redirection string
  public String prpTag; // 8 character data tag (outbound)
  public String prpTagIn; // 8 character data tag (inbound)
  public String prpUnits; // 16 character engineering units
  public float prpMinValue; // minimum value
  public float prpMaxValue; // maximum value
  public int prpSize; // Property dimension (outbound)
  public int prpSizeIn; // Property dimension (inbound)
  public int prpNumOverloads;// number of overloads for Property Name
  public short prpHistoryDepthShort;// Short term history depth (ring buffer
  // size)
  public short prpHistoryDepthLong; // Long term history depth in months
  public byte prpFormat; // Datatype (outbound)
  public byte prpFormatIn; // Datatype (inbound)
  public byte prpAccess; // Access mode(s) for Proptery
  public byte prpGraphType; // Graph stype for Property
  public String rngUnits; // 16 character engineering units for range
  public float rngMinValue; // range minimum
  public float rngMaxValue; // range maximum
  public short numRows; // number of rows in double array
  public short rowSize; // row size in double array
  public short prpArrayType; // property array type
  public byte[] reserved = new byte[6];
  public static int sizeInBytes = 224;
  ByteArrayOutputStream dBuffer = new ByteArrayOutputStream(sizeInBytes);
  public XPropertyQuery()
  {
    prpName = new String(new StringBuffer(32));
    prpDescription = new String(new StringBuffer(64));
    prpRedirection = new String(new StringBuffer(32));
    prpTag = new String(new StringBuffer(8));
    prpTagIn = new String(new StringBuffer(8));
    prpUnits = new String(new StringBuffer(16));
    rngUnits = new String(new StringBuffer(16));
  }
  /**
   * Creates an XPropertyQuery from a TExportProperty
   * 
   * @param prp
   */
  public XPropertyQuery(TExportProperty prp, int numOverloads)
  {
    numRows = (short) prp.getDescription().getArrayNumRows();
    this.prpAccess = (byte) prp.getAccessMode();
    this.prpArrayType = prp.getDescription().getArrayType();
    this.prpDescription = prp.getDescription().getText();
    this.prpFormat = (byte) prp.getOutputFormat();
    this.prpFormatIn = (byte) prp.getInputFormat();
    this.prpGraphType = (byte) prp.getDescription().getYRange().getGraphType();
    this.prpHistoryDepthLong = 0; // fill this in later
    this.prpHistoryDepthShort = 0;
    this.prpMaxValue = prp.getDescription().getYRange().getMaxValue();
    this.prpMinValue = prp.getDescription().getYRange().getMinValue();
    this.prpName = prp.getName();
    this.prpNumOverloads = numOverloads;
    this.prpRedirection = prp.getDescription().getRedirection();
    if (this.prpRedirection == null) this.prpRedirection = "";
    this.prpSize = prp.getOutputSize();
    this.prpSizeIn = prp.getInputSize();
    this.prpTag = prp.getOutputTag();
    this.prpTagIn = prp.getInputTag();
    this.prpUnits = prp.getDescription().getYRange().getUnits();
    this.rngMaxValue = prp.getDescription().getXRange().getMaxValue();
    this.rngMinValue = prp.getDescription().getXRange().getMinValue();
    this.rngUnits = prp.getDescription().getXRange().getUnits();
    this.rowSize = (short) prp.getDescription().getArrayRowLength();
  }
  public void writeData(DataOutput dout) throws IOException
  {
    // NOTE No byte swapping here - use TOutputDataStream!
    dout.write(StringToBytes.stringToFixedBytes(prpName, 32));
    dout.write(StringToBytes.stringToFixedBytes(prpDescription, 64));
    dout.write(StringToBytes.stringToFixedBytes(prpRedirection, 32));
    dout.write(StringToBytes.stringToFixedBytes(prpTag, 8));
    dout.write(StringToBytes.stringToFixedBytes(prpTagIn, 8));
    dout.write(StringToBytes.stringToFixedBytes(prpUnits, 16));
    dout.writeFloat(prpMinValue);
    dout.writeFloat(prpMaxValue);
    dout.writeInt(prpSize);
    dout.writeInt(prpSizeIn);
    dout.writeInt(prpNumOverloads);
    dout.writeShort(prpHistoryDepthShort);
    dout.writeShort(prpHistoryDepthLong);
    dout.writeByte(prpFormat);
    dout.writeByte(prpFormatIn);
    dout.writeByte(prpAccess);
    dout.writeByte(prpGraphType);
    dout.write(StringToBytes.stringToFixedBytes(rngUnits, 16));
    dout.writeFloat(rngMinValue);
    dout.writeFloat(rngMaxValue);
    dout.writeShort(numRows);
    dout.writeShort(rowSize);
    dout.writeShort(prpArrayType);
    dout.write(reserved);
  }
  /*
   * (non-Javadoc)
   * 
   * @see de.desy.tine.client.TStructIfc#readData(java.io.DataInput)
   */
  public void readData(DataInput din) throws IOException
  {
    // prpName
    prpName = StringToBytes.readFixedString(din, 32);
    prpDescription = StringToBytes.readFixedString(din, 64);
    prpRedirection = StringToBytes.readFixedString(din, 32);
    prpTag = StringToBytes.readFixedString(din, 8);
    prpTagIn = StringToBytes.readFixedString(din, 8);
    prpUnits = StringToBytes.readFixedString(din, 16);
    prpMinValue = din.readFloat();
    prpMaxValue = din.readFloat();
    prpSize = din.readInt();
    prpSizeIn = din.readInt();
    prpNumOverloads = din.readInt();
    prpHistoryDepthShort = din.readShort();
    prpHistoryDepthLong = din.readShort();
    prpFormat = din.readByte();
    prpFormatIn = din.readByte();
    prpAccess = din.readByte();
    prpGraphType = din.readByte();
    rngUnits = StringToBytes.readFixedString(din, 16);
    rngMinValue = din.readFloat();
    rngMaxValue = din.readFloat();
    numRows = din.readShort();
    rowSize = din.readShort();
    prpArrayType = din.readShort();
    din.readFully(reserved, 0, 6);
  }
}