/*
 * File TAlarmDefinition.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.server.alarms;

import de.desy.tine.client.TFilterLink;
import de.desy.tine.definitions.*;
import de.desy.tine.server.logger.TFecLog;
import de.desy.tine.stringUtils.Str;
import de.desy.tine.structUtils.TTaggedStructure;

public class TAlarmDefinition extends TTaggedStructure
{
  public static final int ALM_MASK_NONE = 0x00000000;
  public static final int ALM_MASK_DATA_TO_TAG = 0x00000001;
  public boolean isDataToTag()
  {
    return ((alarmMask[0] & ALM_MASK_DATA_TO_TAG) == ALM_MASK_DATA_TO_TAG);
  }
  private char[] alarmTag = new char[TStrings.ALARM_TAG_SIZE];
  private String alarmTagString;
  private int[] alarmCode = new int[1];
  private int[] alarmMask = new int[1];
  private short[] alarmSystem = new short[1];
  private short[] alarmSeverity = new short[1];
  private byte[] alarmDataFormat = new byte[1];
  private byte[] alarmDataArraySize = new byte[1];
  private byte[] alarmOscillationWindow = new byte[1];
  private byte[] alarmOscillationWindowPinned = new byte[1];
  private char[] alarmText = new char[TStrings.ALARM_TEXT_SIZE];
  private String alarmTextString;
  private char[] deviceText = new char[TStrings.ALARM_TEXT_SIZE];
  private String deviceTextString;
  private char[] dataText = new char[TStrings.ALARM_TEXT_SIZE];
  private String dataTextString;
  private char[] url = new char[TStrings.ALARM_TEXT_LONGSIZE];
  private String urlString;
  private TFilterLink filter = null;
  public TFilterLink getFilter() { return filter; }
  public void setFilter(TFilterLink filter) 
  {
    String fstr = filter == null ? "none" : filter.toString();
    TFecLog.log("TAlarmDefinition","set filter to "+fstr);
    this.filter = filter; 
  }
  public String getFilterAsString()
  {
    return filter != null ? filter.toString() : "";
  }
  public boolean isPastFilter()
  {
    if (filter != null && !filter.isActive())
    { // marked for removal;
      filter = null;
    }
    return filter == null ? true : filter.isValid();
  }
  public int getAlarmCode() { return alarmCode[0]; }
  public int getAlmCode() { return alarmCode[0]; } // for compatibility
  public int getAlarmMask() { return alarmMask[0]; }
  public int getAlarmSystem() { return alarmSystem[0]; }
  public int getAlarmSeverity() { return alarmSeverity[0]; }
  public int getAlarmDataFormat() { return alarmDataFormat[0]; }
  public int getAlarmDataArraySize() { return alarmDataArraySize[0]; }
  public int getAlarmOscillationWindow() { return alarmOscillationWindow[0]; }
  public boolean isOscillationWindowPinned() { return (alarmOscillationWindowPinned[0] != 0); }
  private String trimmedString(char[] charArray)
  {
    int len=0;
    if (charArray == null || charArray.length == 0) return "";
    while (len < charArray.length && charArray[len] != 0 ) len++;
    return new String(charArray,0,len);
  }
  public String getAlarmTag()
  {
    if (alarmTagString == null) 
    {
    	if (alarmTag[0] == 0) System.arraycopy(alarmText, 0, alarmTag, 0, TStrings.ALARM_TAG_SIZE);
    	alarmTagString = trimmedString(alarmTag);
    }
    return alarmTagString;
  }
  public String getAlarmText()
  {
    if (alarmTextString == null) alarmTextString = trimmedString(alarmText);//new String(alarmText).trim();
    return alarmTextString;
  }
  public String getAlmText() { return getAlarmText(); } // for compatibility
  public String getDeviceText()
  {
    if (deviceTextString == null) deviceTextString = trimmedString(deviceText);//new String(deviceText).trim();
    return deviceTextString;
  }
  public String getDataText()
  {
    if (alarmDataArraySize[0] == 0) dataTextString = new String("no data associated with alarm");
    if (dataTextString == null) dataTextString = trimmedString(dataText);//new String(dataText).trim();
    return dataTextString;
  }
  public String getUrl()
  {
    if (urlString == null) urlString = trimmedString(url);//new String(url).trim();
    return urlString;
  }
  private static boolean initialized = false;
  private void initStructDescription()
  {
    if (initialized) 
    {
      //TODO:  this seems stupid to repeat this over and over ....
      //return;
    }
    alarmOscillationWindow[0] =  TAlarm.ALM_OSCILLATION_DEFAULT;
    alarmOscillationWindowPinned[0] = 0;
    addField(alarmTag,"almTag");
    addField(alarmCode,"almCode");
    addField(alarmMask,"almMask");
    addField(alarmSystem,"almSystem");
    addField(alarmSeverity,"almSeverity");
    addField(alarmDataFormat,"almDataFormat");
    addField(alarmDataArraySize,"almDataArraySize");
    addField(alarmOscillationWindow,"almOscWindow");
    addField(alarmOscillationWindowPinned,"almOscPinned");
    addField(alarmText,"alarmText");
    addField(deviceText,"deviceText");
    addField(dataText,"dataText");
    addField(url,"url");
    initDone();
    initialized = true;
  }
  public TAlarmDefinition()
  {
    super("ADSr4");
    initStructDescription();
  }
  private void pushString(String s,char[] c)
  {
    if (s != null) 
    {
      s = Str.toAscii(s);
      int len = s.length();
      if (len > c.length) len = c.length; 
      s.getChars(0, len, c, 0);
      for (int i=len; i<c.length; i++) c[i]=0;
    }    
  }
  public TAlarmDefinition(String tag,String txt,String devTxt,int code,int severity,int system,
      int mask,short dataFormat,int dataSize,String dataTxt,String urlTxt)
  {
    this(tag,txt,devTxt,code,severity,system,mask,dataFormat,dataSize,dataTxt,urlTxt,null);
  }
  public TAlarmDefinition(String tag,String txt,String devTxt,int code,int severity,int system,
      int mask,short dataFormat,int dataSize,String dataTxt,String urlTxt,String filterString)
  {
    super("ADSr4");
    initStructDescription();
    if (tag == null || tag.length() == 0) tag = txt;
    pushString(tag,alarmTag);
    pushString(txt,alarmText);
    pushString(devTxt,deviceText);
    setAlarmCode(code);
    setAlarmSeverity(severity);
    setAlarmSystem(system);
    setAlarmMask(mask);
    pushString(dataTxt,dataText);
    setAlarmDataFormat(dataFormat);
    setAlarmDataArraySize(dataSize);
    pushString(urlTxt,url);
    if (filterString != null && filterString.length() > 0) setFilter(new TFilterLink(filterString));
  }
  public TAlarmDefinition(String tag,String txt,String devTxt,int code,int severity,int system)
  {
    this(tag,txt,devTxt,code,severity,system,0,TFormat.CF_NULL,0,null,null,null);
  }
  public TAlarmDefinition(TAlarmDefinition reference,int code, String tag)
  {
    this(reference.getAlarmTag(),reference.getAlarmText(),reference.getDeviceText(),
        reference.getAlarmCode(),reference.getAlarmSeverity(),reference.getAlarmSystem(),
        reference.getAlarmMask(),(short)reference.getAlarmDataFormat(),reference.getAlarmDataArraySize(),
        reference.getDataText(),reference.getUrl(),reference.getFilterAsString());
    if (code >= 512) setAlarmCode(code);
    if (tag != null && tag.length() > 0) setAlarmTag(tag);
  }
  public void setAlarmCode(int almCode)
  {
    alarmCode[0] = almCode;
  }
  public void setAlarmDataArraySize(int almDataArraySize)
  {
    alarmDataArraySize[0] = (byte)almDataArraySize;
  }
  public void setAlarmDataFormat(int almDataFormat)
  {
    int bf = almDataFormat%512;
    if (bf > 127) bf =- 256;
    alarmDataFormat[0] = (byte)(bf);
  }
  public void setAlarmOscillationWindow(int almOscillationWindow)
  {
    int bf = almOscillationWindow;
    if (bf > 127) bf = 127;
    alarmOscillationWindow[0] = (byte)(bf);
  }
  public void setOscillationWindowPinned(boolean value)
  {
    alarmOscillationWindowPinned[0] = (byte)(value ? 0xff : 0);
  }
  public void setAlarmMask(int almMask)
  {
    alarmMask[0] = almMask;
  }
  public void setAlarmSeverity(int almSeverity)
  {
    alarmSeverity[0] = (short)almSeverity;
  }
  public void setAlarmSystem(int almSystem)
  {
    alarmSystem[0] = (short)almSystem;
  }
  public void setAlarmTag(String tag)
  {
    pushString(tag,alarmTag);
  }
  public void setAlarmText(String alarmTxt)
  {
    pushString(alarmTxt,alarmText);
  }
  public void setDataText(String dataTxt)
  {
    pushString(dataTxt,dataText);
  }
  public void setDeviceText(String deviceTxt)
  {
    pushString(deviceTxt,deviceText);
  }
  public void setUrl(String urlTxt)
  {
    pushString(urlTxt,url);
  }
  public void setFields(TAlarmDefinition reference)
  {
    alarmSeverity[0] = (short)reference.getAlarmSeverity();
    alarmSystem[0] = (short)reference.getAlarmSystem();
    pushString(reference.getAlarmTag(),alarmTag);
    pushString(reference.getAlarmText(),alarmText);
    pushString(reference.getDataText(),dataText);
    pushString(reference.getDeviceText(),deviceText);
    pushString(reference.getUrl(),url);
  }
  
}
