/*
 * File TAlarmStockAlarm.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.server.alarms;

import java.util.HashMap;

import de.desy.tine.definitions.TErrorList;
import de.desy.tine.definitions.TFormat;

public class TAlarmStockAlarm
{
  private static HashMap<String, TAlarmDefinition> stkAlarmDefinitions = new HashMap<String, TAlarmDefinition>();
  public static HashMap<String, TAlarmDefinition> getStockAlarmDefinitions() 
  { 
    if (!stockAlarmDefinitionsRegistered) registerStockAlarmDefinitions();
    return stkAlarmDefinitions; 
  }
  private static boolean stockAlarmDefinitionsRegistered = false;
  /** 
   * Gets the alarm definition structure array for known 'stock' alarm
   * definitions. 
   * 
   * @return an array list of alarm definitions for the input given.
   *
   */
  public static TAlarmDefinition[] getAlarmDefinitions()
  {
    if (!stockAlarmDefinitionsRegistered) registerStockAlarmDefinitions();
    return stkAlarmDefinitions.values().toArray(new TAlarmDefinition[0]);
  }
  public static TAlarmDefinition getAlarmDefinition(int code)
  {
    if (!stockAlarmDefinitionsRegistered) registerStockAlarmDefinitions();
    return stkAlarmDefinitions.get(Integer.toString(code));
  }
  public static void registerStockAlarmDefinitions()
  {
    if (stockAlarmDefinitionsRegistered) return;
    // register some stock alarm definitions ...
    stkAlarmDefinitions.put(Integer.toString(TErrorList.invalid_data), 
      new TAlarmDefinition("invalid", "value not at set point","device",TErrorList.invalid_data,TAlarm.ALM_SEVERITY_WARN,0, 
          -1,TFormat.CF_INT32,1,"last measured value",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.value_too_high), 
      new TAlarmDefinition("too high", "value exceeds threshold","device",TErrorList.value_too_high,TAlarm.ALM_SEVERITY_ERR,0,
          -1,TFormat.CF_FLOAT,1,"last measured value",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.warn_too_high), 
      new TAlarmDefinition("near limit", "value is near threshold","device",TErrorList.warn_too_high,TAlarm.ALM_SEVERITY_WARN,0,
          -1,TFormat.CF_FLOAT,1,"last measured value",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.value_too_low), 
      new TAlarmDefinition("too low", "value exceeds threshold","device",TErrorList.value_too_low,TAlarm.ALM_SEVERITY_ERR,0,
          -1,TFormat.CF_FLOAT,1,"last measured value",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.warn_too_low), 
      new TAlarmDefinition("near limit", "value is near threshold","device",TErrorList.warn_too_low,TAlarm.ALM_SEVERITY_WARN,0,
          -1,TFormat.CF_FLOAT,1,"last measured value",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.max_alarms_exceeded), 
        new TAlarmDefinition("number of alarms too large", "number of alarms too large for individual reporting","device",TErrorList.max_alarms_exceeded,TAlarm.ALM_SEVERITY_WARN,TAlarm.ALM_SYSTEM_HARDWARE,
            -1,TFormat.CF_INT16,32,"total number alarms + representative alarm codes",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.io_error), 
        new TAlarmDefinition("I/O error", "hardware input/ouput error","device",TErrorList.io_error,TAlarm.ALM_SEVERITY_WARN,TAlarm.ALM_SYSTEM_HARDWARE,
            -1,TFormat.CF_NULL,0,"",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.file_error), 
        new TAlarmDefinition("File error", "file input/ouput error","device",TErrorList.file_error,TAlarm.ALM_SEVERITY_WARN,TAlarm.ALM_SYSTEM_SYSTEM,
            -1,TFormat.CF_TEXT,64,"file/path name which caused the alarm",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.hardware_error), 
        new TAlarmDefinition("Hardware error", "general hardware error","device",TErrorList.hardware_error,TAlarm.ALM_SEVERITY_WARN,TAlarm.ALM_SYSTEM_HARDWARE,
            -1,TFormat.CF_TEXT,64,"hardware device description",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.net_read_error), 
        new TAlarmDefinition("Network read error", "network read error","device",TErrorList.net_read_error,TAlarm.ALM_SEVERITY_WARN,TAlarm.ALM_SYSTEM_SYSTEM,
            -1,TFormat.CF_NULL,0,"",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.net_write_error), 
        new TAlarmDefinition("Network write error", "network write error","device",TErrorList.net_write_error,TAlarm.ALM_SEVERITY_WARN,TAlarm.ALM_SYSTEM_SYSTEM,
            -1,TFormat.CF_NULL,0,"",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.link_error), 
        new TAlarmDefinition("Link error", "link to dependent server is not open","device",TErrorList.link_error,TAlarm.ALM_SEVERITY_WARN,TAlarm.ALM_SYSTEM_SYSTEM,
            -1,TFormat.CF_TEXT,64,"link which has the problem",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.driver_not_installed), 
        new TAlarmDefinition("Driver not installed", "hardware driver not installed or not initialized","device",TErrorList.driver_not_installed,TAlarm.ALM_SEVERITY_WARN,TAlarm.ALM_SYSTEM_HARDWARE,
            -1,TFormat.CF_TEXT,64,"name of hardware driver",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.port_not_available), 
        new TAlarmDefinition("Port not available", "hardware port not avaliable","device",TErrorList.port_not_available,TAlarm.ALM_SEVERITY_WARN,TAlarm.ALM_SYSTEM_HARDWARE,
            -1,TFormat.CF_TEXT,64,"Port Id",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.not_running), 
        new TAlarmDefinition("Not Running", "hardware device is not running","device",TErrorList.not_running,TAlarm.ALM_SEVERITY_WARN,TAlarm.ALM_SYSTEM_HARDWARE,
            -1,TFormat.CF_NULL,0,"",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.sedac_error), 
        new TAlarmDefinition("Sedac error", "sedac module has an error","device",TErrorList.sedac_error,TAlarm.ALM_SEVERITY_WARN,TAlarm.ALM_SYSTEM_HARDWARE,
            -1,TFormat.CF_INT16,3,"line/crate/subaddress of defective module",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.gpib_error), 
        new TAlarmDefinition("GPIB error", "GPIB bus error","device",TErrorList.gpib_error,TAlarm.ALM_SEVERITY_WARN,TAlarm.ALM_SYSTEM_HARDWARE,
            -1,TFormat.CF_TEXT,64,"hardware module description",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.rs232_error), 
        new TAlarmDefinition("RS232 error", "RS232 bus communication error","device",TErrorList.rs232_error,TAlarm.ALM_SEVERITY_WARN,TAlarm.ALM_SYSTEM_HARDWARE,
            -1,TFormat.CF_TEXT,64,"hardware module description",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.device_error), 
        new TAlarmDefinition("Device error", "general device error/device defective","device",TErrorList.device_error,TAlarm.ALM_SEVERITY_WARN,TAlarm.ALM_SYSTEM_HARDWARE,
            -1,TFormat.CF_TEXT,64,"module description",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.canbus_error), 
        new TAlarmDefinition("CAN bus error", "CAN module has an error","device",TErrorList.canbus_error,TAlarm.ALM_SEVERITY_WARN,TAlarm.ALM_SYSTEM_HARDWARE,
            -1,TFormat.CF_INT32,4,"address of CAN module",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.profibus_error), 
        new TAlarmDefinition("PROFIBUS error", "PROFIBUS module has an error","device",TErrorList.profibus_error,TAlarm.ALM_SEVERITY_WARN,TAlarm.ALM_SYSTEM_HARDWARE,
            -1,TFormat.CF_INT32,4,"address of PROFIBUS module",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.warn_disk_space), 
        new TAlarmDefinition("disk space is low", "disk space is getting low","device",TErrorList.warn_disk_space,TAlarm.ALM_SEVERITY_WARN,TAlarm.ALM_SYSTEM_SYSTEM,
            0,TFormat.CF_TEXT,64,"disk + available space (blocks)",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.low_disk_space), 
        new TAlarmDefinition("disk space is very low", "disk space is extremely low","device",TErrorList.low_disk_space,TAlarm.ALM_SEVERITY_WARN,TAlarm.ALM_SYSTEM_SYSTEM,
            0,TFormat.CF_TEXT,64,"disk + available space (blocks)",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.low_main_memory), 
        new TAlarmDefinition("low main memory", "low available main memory detected","device",TErrorList.low_main_memory,TAlarm.ALM_SEVERITY_WARN,TAlarm.ALM_SYSTEM_SYSTEM,
            0,TFormat.CF_TEXT,64,"available vs. threshold (blocks)",""));
    stockAlarmDefinitionsRegistered = true;
  }
}
