/*
 * File TCASAlarmMessage.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.server.alarms;

import de.desy.tine.client.*;
import de.desy.tine.dataUtils.*;
import de.desy.tine.structUtils.*;
import de.desy.tine.definitions.*;

public class TCASAlarmMessage extends TTaggedStructure implements Comparable<Object>
{
  private static TLinkFactory gLinkFactory = null;
  private static TLinkFactory getLinkFactory()
  {
    if (gLinkFactory == null) gLinkFactory = TLinkFactory.getInstance();
    return gLinkFactory;
  }
  private char[] server = new char[TStrings.EXPORT_NAME_SIZE];
  private String serverString;
  private char[] device = new char[TStrings.DEVICE_NAME_SIZE];
  private String deviceString;
  private char[] alarmText = new char[TStrings.ALARM_TEXT_SIZE];
  private String alarmTextString;
  private int[] alarmCode = new int[1];
  private int[] timestamp = new int[1];
  private int[] timestampUSec = new int[1];
  private int[] starttime = new int[1];
  private int[] starttimeUSec = new int[1];
  private int[] alarmMask = new int[1];
  private int[] alarmSystem = new int[1];
  private byte[] alarmData = new byte[TStrings.ALARM_DATA_SIZE];
  private byte[] alarmDataFormat = new byte[1];
  private byte[] alarmDataArraySize = new byte[1];
  private TDataType tAlarmData;
  private byte[] severity = new byte[1];
  private byte[] descriptor = new byte[1];
  public int getAlarmCode() { return alarmCode[0]; }
  public int getAlarmSeverity() { return severity[0]; }
  public int getAlarmDescriptor() { return descriptor[0]; }
  public String getAlarmDescriptorAsString() 
  {
    return TAlarmDescriptor.toLongString(descriptor[0]); 
  }
  public boolean isTransient()
  {
    if ((descriptor[0] & TAlarmDescriptor.TRANSIENT) != TAlarmDescriptor.TRANSIENT)
    {
      return true;
    }
    return false;    
  }
  public boolean isActive()
  {
    if ((descriptor[0] & TAlarmDescriptor.TERMINATE) != TAlarmDescriptor.TERMINATE)
    { // definitely active !
      return true;
    }
    if ((descriptor[0] & TAlarmDescriptor.NEW) == TAlarmDescriptor.NEW)
    { // this will let the transient alarms appear in the active list
      return true;
    }
    return false;
  }
  public int getAlarmSystem() { return alarmSystem[0]; }
  public long getTimeStamp() 
  { 
    return((long)timestamp[0]*1000) + (timestampUSec[0]/1000);
  }
  public long getStartTime()
  { 
    return ((long)starttime[0]*1000) + (starttimeUSec[0]/1000); 
  }
  public int getAlarmMask() { return alarmMask[0]; }
  public boolean hasData() { return alarmDataArraySize[0] > 0 ? true : false; }
  public TDataType getAlarmData()
  {
    if (tAlarmData == null)
    {
      int dsiz = alarmDataArraySize[0];
      short dfmt = (short)(alarmDataFormat[0]);
      tAlarmData = new TDataType(dsiz,dfmt);
      if (dsiz > 0 && dfmt != TFormat.CF_NULL)
      { // the data are in TINE (i.e. little-endian) byte order!
        tAlarmData.pushBytes(alarmData);
        getLinkFactory().fillinIncomingData(tAlarmData);
      }
    }
    return tAlarmData;
  }
  public String getServer()
  {
    if (serverString == null) serverString = new String(server).trim();
    return serverString;
  }
  public String getDevice()
  {
    if (deviceString == null) deviceString = new String(device).trim();
    return deviceString;
  }
  public String getAlarmText()
  {
    if (alarmTextString == null) alarmTextString = new String(alarmText).trim();
    return alarmTextString;
  }
  private void initStructDescription()
  {
    addField(server,"server");
    addField(device,"device");
    addField(alarmText,"almText");
    addField(alarmCode,"almCode");
    addField(timestamp,"timestamp");
    addField(timestampUSec,"timestampUSec");
    addField(starttime,"starttime");
    addField(starttimeUSec,"starttimeUSec");
    addField(alarmMask,"almMask");
    addField(alarmSystem,"almSystem");
    addField(alarmData,"almData");
    addField(alarmDataFormat,"almDataFormat");
    addField(alarmDataArraySize,"almDataArraySize");
    addField(severity,"severity");
    addField(descriptor,"descriptor");
    initDone();    
  }
  public TCASAlarmMessage()
  {
    super("XAMSr4");
    initStructDescription();
  }
  private void pushByteArray(byte[] barray)
  {
    for (int i=0; i<TStrings.ALARM_DATA_SIZE; i++) alarmData[i] = 0;
    if (barray == null) return; 
    int len = Math.min(barray.length, TStrings.ALARM_DATA_SIZE);
    // TODO: don't we have to worry about swapping ?
    for (int i=0; i<len; i++) alarmData[i] = barray[i];    
  }
  private void pushString(String s,char[] c)
  {
    if (s != null) 
    {
      int len = s.length();
      if (len > c.length) len = c.length; 
      s.getChars(0, len, c, 0);
    }    
  }
  public TCASAlarmMessage(TAlarmMessage ams)
  {
    super("XAMSr4");
    initStructDescription();
    pushString(ams.getServer(),server);
    pushString(ams.getDevice(),device);
    pushString(ams.getAlarmTag(),alarmText);
    alarmCode[0] = ams.getAlarmCode();
    timestamp[0] = (int)(ams.getTimeStamp()/1000);
    timestampUSec[0] = (int)(ams.getTimeStamp()%1000)*1000;
    starttime[0] = (int)(ams.getStartTime()/1000);
    starttimeUSec[0] = (int)(ams.getStartTime()%1000)*1000;
    pushByteArray(ams.getAlarmData().getDataBuffer());
    descriptor[0] = (byte)ams.getAlarmDescriptor();
  }
  public static final int sizeInBytes = 
    TStrings.EXPORT_NAME_SIZE +
    TStrings.DEVICE_NAME_SIZE +
    TStrings.ALARM_TAG_SIZE +
    6 * 4 + TStrings.ALARM_DATA_SIZE + 8;
  public int compareTo(Object a)
  {
    if (a == null) return -999;
    if (a instanceof TCASAlarmMessage)
    {
      int dt = timestamp[0] - ((TCASAlarmMessage)a).timestamp[0];
      if (dt != 0) return dt;
      
      return timestampUSec[0] - ((TCASAlarmMessage)a).timestampUSec[0];
    }
    return -998;
  }
}
