/*
 * File TClient.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

/*
 * Created on Nov 9, 2004
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package de.desy.tine.server.connections;

import java.io.DataOutput;
import java.io.IOException;
import java.io.OutputStream;
import java.net.InetAddress;
import java.net.Socket;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;

import de.desy.tine.definitions.*;
import de.desy.tine.server.equipment.TEquipmentModuleFactory;
import de.desy.tine.server.logger.MsgLog;
import de.desy.tine.stringUtils.*;
import de.desy.tine.structUtils.*;

/**
 * \internal
 * 
 * @author duval
 */
public class TClient //extends TStructBase
{
  static public boolean isSubNet(int mask,byte[] tgt,byte[] src)
  {
    int i, k, bit, nb = mask/8, nr = mask%8;
    for (i=0; i<nb; i++)
    { /* these bytes must match */
      if (tgt[i] != src[i]) return false;    
    }
    for (k=0; k<nr; k++)
    { /* these bits must match */
      bit = (1 << (7-k));
      if ((tgt[i] & bit) != (src[i] & bit)) return false;
    }
    return true;
  }
  public boolean isMemberControlNets(LinkedList<TNetAcl> nets)
  {
    if (nets == null || nets.size() == 0) return true; // everyone is !
    // check vs subnet, vs specified address ...
    byte[] tgt = IPaddress.getAddress();
    TNetAcl netAcl;
    Iterator<TNetAcl> it = nets.iterator();
    while (it.hasNext())
    {
      netAcl = it.next();
      if (isSubNet(netAcl.getMask(),netAcl.getRawAddress(),tgt)) return true;
    }
    return false;    
  }
  public boolean isMemberUsersList(LinkedList<String> usrs)
  {
    if (usrs == null || usrs.size() ==  0) return true; // everyone is !
    if (usrs.contains(userName.toUpperCase())) return true;
    return false;
  }
  public static final int ALLOWED_DANGLE_TIME = 20000;
  private long timeAppendedToList = 0;
  public long getTimeAppendedToList() { return timeAppendedToList; }
  public void setTimeAppendedToList(long t) { timeAppendedToList = t; }
  // below used in queries ...
  public String userName = "(unknown)"; // 16 char
  public String userAddr; // 40 char
  public String userType = ""; // 8 char
  byte[] IPXaddress = new byte[12]; //[4]; // 4 bytes
  byte[] node = new byte[6]; // 6 bytes IPX Immediate Address
  short IPfamily = 2;
  public InetAddress IPaddress; // 4 bytes struct sockaddr_in or 16 byte sockaddr_in6
  byte[] IPsine_zero = new byte[8];
  public short ncontracts; // number of registrations
  public int pid = 0;
  public short tineProtocol = 0; // target client protocol level
  public int versionId;
  public short inetProtocol; // transport protocol
  public int port; // communication port
  // above used in queries ...
  public String toString()
  {
    String s;
    s = new String(userName)+" @ "+IPaddress.getHostAddress()+":"+port;
    s = s.concat(" "+ userType+" (pid: "+pid+")");
    s = s.concat(" "+ TTransport.toString(inetProtocol));
    s = s.concat(" ("+ncontracts+" contracts)");
    return s;
  }
//  NEED the equals to work on the object reference !!!!
//  n.b. LocateClientInList() does the address comparisons ...
//  public boolean equals(TClient reference)
//  {
//    if (reference == null) return false;
//    if (!reference.IPaddress.equals(IPaddress)) return false;
//    if (reference.IPport != IPport) return false;
//    return true;
//  }
  private ArrayList<TContractTable> conLst = new ArrayList<TContractTable>();
  private LinkedList<TContractTable> conLstAdd = new LinkedList<TContractTable>();
  public void addContract(TContractTable con) 
  { 
    synchronized (conLstAdd) 
    {
      conLstAdd.add(con); 
    }
  }
  public void rmvContract(TContractTable con) 
  { 
    conLst.remove(con); 
  }
  public void clearAddedContracts()
  {
    synchronized (conLstAdd)
    {
      conLstAdd.clear();
    }
  }
  public void appendAddedContracts()
  {
    TContractTable tct = null;
    synchronized (conLstAdd)
    {
      Iterator<TContractTable> li = conLstAdd.iterator();
      while (li.hasNext())
      {
        tct = (TContractTable)li.next();
        conLst.add(tct);
      }
      conLstAdd.clear();
    }
  }
  public boolean hasAddedContracts()
  {
    return conLstAdd.size() > 0 ? true : false;
  }
  public TContractTable[] getContractList()
  {
    TContractTable[] cta = null;
    try
    {
      cta = conLst.toArray(new TContractTable[0]);
    }
    catch (Exception e) 
    {
      MsgLog.log("getContractList", e.toString(),TErrorList.code_failure,e,1);      
    };
    return cta;
  }
  public OutputStream output; // return communication stream if connected
  public Socket sck;
  public TEquipmentModuleFactory.TEqmFactoryBucketThread tbt;
  static private TStructDescription strcDscL; // Export Data structure (legacy)
  static private TStructDescription strcDsc;  // Export Data structure (modern)
  static private final int ARRAY_SIZE_HINT = 20;
  public static final int nice_query_size = 50;
  static
  {
    initClientStruct();
  }
  public static TStructDescription getLegacyStructDescription()
  {
    return strcDscL;
  }
  private static void initClientStruct()
  {
    if (TStructRegistry.contains("CLNQS")) return;
    strcDscL = new TStructDescription("CLNQS"); 
    strcDscL.beginDefinition();
    strcDscL.addField("username",TFormat.CF_TEXT, 16);
    strcDscL.addField("ipxAddr",TFormat.CF_BYTE, 12);
    strcDscL.addField("ipxNode",TFormat.CF_BYTE, 6);
    strcDscL.addField("ipFamily",TFormat.CF_INT16, 1);
    strcDscL.addField("ipPort",TFormat.CF_INT16, 1);
    strcDscL.addField("ipAddr",TFormat.CF_BYTE, 4);
    strcDscL.addField("ipZero",TFormat.CF_BYTE, 8);   
    strcDscL.addField("numContracts", TFormat.CF_INT16, 1);
    strcDscL.addField("transport", TFormat.CF_INT16, 1);
    strcDscL.setArraySize(ARRAY_SIZE_HINT);
    strcDscL.endDefinition();
    TStructRegistry.add(strcDscL);
    
    if (TStructRegistry.contains("ClnInfoQS")) return;
    strcDsc = new TStructDescription("ClnInfoQS"); 
    strcDsc.beginDefinition();
    strcDsc.addField("username",TFormat.CF_TEXT, 16);
    strcDsc.addField("address",TFormat.CF_TEXT, 40);
    strcDsc.addField("appType",TFormat.CF_TEXT, 8);
    strcDsc.addField("pid", TFormat.CF_INT32, 1);
    strcDsc.addField("port",TFormat.CF_UINT16, 1);
    strcDsc.addField("numContracts", TFormat.CF_INT16, 1);
    strcDsc.addField("tineProtocol", TFormat.CF_INT16, 1);
    strcDsc.addField("inetProtocol", TFormat.CF_INT16, 1);
    strcDsc.setArraySize(ARRAY_SIZE_HINT);
    strcDsc.endDefinition();
    TStructRegistry.add(strcDsc);
    
  }
  /**
   * used to send the CLNQS structure to callers to stock property CLIENTS
   */
  public void writeData(DataOutput dout, boolean isLegacy) throws IOException
  { 
    if (isLegacy)
    {
      dout.write(StringToBytes.stringToFixedBytes(userName, 16));
      dout.write(IPXaddress);
      dout.write(node);
      dout.writeShort(IPfamily);
      dout.writeShort(port);
      dout.write(IPaddress.getAddress(), 0, 4);
      dout.write(IPsine_zero);
      dout.writeShort(ncontracts);
      dout.writeShort(tineProtocol);
    }
    else
    {
      dout.write(StringToBytes.stringToFixedBytes(userName, 16));
      userAddr = IPaddress.getHostAddress();
      dout.write(StringToBytes.stringToFixedBytes(userAddr, 40));
      dout.write(StringToBytes.stringToFixedBytes(userType, 8));
      dout.writeInt(pid);
      dout.writeShort(port);
      dout.writeShort(ncontracts);
      dout.writeShort(tineProtocol);
      dout.writeShort(inetProtocol);
    }
  }
}
