/*
 * File TDeviceList.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

/*
 * Created on Aug 24, 2005
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package de.desy.tine.server.devices;

import java.util.ArrayList;
//import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.Set;

import de.desy.tine.definitions.TErrorList;
import de.desy.tine.server.equipment.TEquipmentModule;
import de.desy.tine.server.logger.MsgLog;
import de.desy.tine.server.logger.TFecLog;

public class TDeviceList
{
  private final static int TDeviceListCapacity = 256;
  private LinkedHashMap<String, TDevice> myDeviceData = new LinkedHashMap<String, TDevice>(TDeviceListCapacity);
  private LinkedList<String> myDeviceNames = new LinkedList<String>();
  private int lastRegisteredDevice = -1;
  private boolean propertyOriented = false;
  public void setPropertyOriented(boolean value) { propertyOriented = value; }
  public boolean isPropertyOriented() { return propertyOriented; }
  public int getLastRegisteredDevice() { return lastRegisteredDevice; }
  public int getNumberOfDevices() { return myDeviceNames.size(); }
  public void refreshAlarms()
  {
    TDevice[] dlst = getDeviceList();
    if (dlst == null) return;
    
    for (int i=0; i<dlst.length; i++)
    {
      dlst[i].processAlarms();
    }
    return;        
  }
  public void clearAlarms()
  {
    TDevice[] dlst = getDeviceList();
    if (dlst == null) return;
    for (int i=0; i<dlst.length; i++)
    {
      dlst[i].clearAlarms();
    }
    return;
  }
  public void removeAllAlarms()
  {
    TDevice[] dlst = getDeviceList();
    if (dlst == null) return;
    for (int i=0; i<dlst.length; i++)
    {
      dlst[i].removeAlarms();
    }
  }
  public void touchActiveAlarms()
  {
    int thisTimeStamp = (int)(System.currentTimeMillis()/1000);
    touchActiveAlarms(thisTimeStamp);
  }
  public void touchActiveAlarms(int timestamp)
  {
    TDevice[] dlst = getDeviceList();
    if (dlst == null) return;
    for (int i=0; i<dlst.length; i++)
    {
      dlst[i].touchAlarms(timestamp);
    }    
  }
  public Set<?> getDeviceSet()
  {
    return myDeviceData.entrySet();
  }
  public String[] getDeviceNameList()
  {
    return getDeviceNameList(0);
  }
  public String[] getDeviceNameList(int startIndex)
  {
	//return (String[])myDeviceNames.toArray(new String[0]);
//	  String[] namesArray = (String[])myDeviceNames.toArray(new String[0]);
//    ArrayList<String> lst = new ArrayList<String>();
//    int nameIdx = 0;
//    for (String name : namesArray) 
//    {
//      if (nameIdx < startIndex) continue;
//      lst.add( name);
//      ++nameIdx;
//    }
//    return lst.toArray(new String[0]);

    TDevice[] devs = (TDevice[])myDeviceData.values().toArray(new TDevice[0]);
    if (devs == null) return (String[])myDeviceNames.toArray(new String[0]);
    ArrayList<String> lst = new ArrayList<String>();
    for (TDevice d : devs) 
    {
      if (d.devNumber < startIndex) continue;
      lst.add(d.devName);
    }
    return lst.toArray(new String[0]);
  
  
  }
  public TDevice[] getDeviceList()
  {
    return (TDevice[])myDeviceData.values().toArray(new TDevice[0]);
  }
  public LinkedList<String> getDeviceLinkedList() { return myDeviceNames; }
  public TDevice getDevice(String devName)
  {
    if (devName == null) return null;
    if (devName.startsWith("#"))
    {
      try 
      {
        return getDevice(Integer.parseInt((devName.substring(1))));
      }
      catch (Exception e)
      {
        MsgLog.log("getDevice",e.toString(),TErrorList.code_failure,e,0);                          
        return null;
      }
    }
    String devNameUC = devName.toUpperCase();
    if (!myDeviceData.containsKey(devNameUC)) return null;
    return (TDevice)myDeviceData.get(devNameUC);
  }
  public TDevice getDevice(int devNumber)
  {
    if (devNumber < 0 || devNumber >= myDeviceNames.size()) return null;
    String devName = (String)myDeviceNames.get(devNumber);
    TDevice d = (TDevice)myDeviceData.get(devName.toUpperCase());
    if (d == null)
    { // try a registered number ...
      d = (TDevice)myDeviceData.get("#"+devNumber);
    }
    return d;
  }
  public void renameDevice(String oldName,String newName)
  {
    TDevice tdv = getDevice(oldName);
    if (tdv == null) return;
    tdv.setName(newName);
    int dv = tdv.devNumber;
    myDeviceData.remove(oldName.toUpperCase());
    myDeviceData.put(newName.toUpperCase(),tdv);
    myDeviceNames.set(dv,newName);
  }
  public void addDevice(String devName,int devNumber,TEquipmentModule eqm)
  {
    if (devNumber < 0 || devName == null) return;
    String ucName = devName.toUpperCase();
    if (devNumber < myDeviceNames.size())
    {
      myDeviceNames.set(devNumber,devName);
      if (!myDeviceData.containsKey(ucName))
      {
        TDevice d = getDevice(devNumber);
        myDeviceData.remove(d.devName.toUpperCase());
        myDeviceData.put(ucName,new TDevice(devName,devNumber,eqm));
      }
      return;
    }
    else
    {
      if (devNumber > lastRegisteredDevice + 1)
      {
        TFecLog.log("device "+devName+" assigned to device "+devNumber+
            " : assign names for "+(lastRegisteredDevice+1)+ " to "+(devNumber-1));
        for (int i=lastRegisteredDevice+1; i<devNumber; i++) 
        {
          myDeviceNames.add("#"+i);
        }
      }
      lastRegisteredDevice = devNumber;
      myDeviceData.put(ucName,new TDevice(devName,devNumber,eqm));
      myDeviceNames.add(devName);
    }
  }
  public void addDevice(String devName,TEquipmentModule eqm)
  {
    addDevice(devName,lastRegisteredDevice + 1,eqm);
  }
  public void addDevice(TDevice device)
  {
    if (device == null) return;
    myDeviceData.put(device.devName.toUpperCase(),device);
    myDeviceNames.add(device.devName);
  }
  
  public void removeDevice(String devName,int devNumber,TEquipmentModule eqm)
  {
    if (devNumber < 0) return;
    TDevice device = myDeviceData.get(devName.toUpperCase());
    if (device == null) return;
    if (eqm != device.getEqm()) return;
    if (device.devNumber != devNumber) return;
    if (devNumber == lastRegisteredDevice) lastRegisteredDevice--;
    myDeviceNames.remove(device.devName);
    myDeviceData.remove(device.devName.toUpperCase());  
  }
  public void removeDevice(String devName,TEquipmentModule eqm)
  {
    TDevice device = myDeviceData.get(devName.toUpperCase());
    if (device == null) return;
    if (eqm != device.getEqm()) return;
    myDeviceNames.remove(device.devName);
    myDeviceData.remove(device.devName.toUpperCase());
  }
  public void removeDevice(TDevice device)
  {
    if (device == null) return;
    myDeviceNames.remove(device.devName);
    myDeviceData.remove(device.devName.toUpperCase());
  }
  
  public int getDeviceNumber(String devName)
  {
    if (devName.startsWith("#")) 
    {
      try 
      {
        return Integer.parseInt((devName.substring(1)));
      }
      catch (NumberFormatException e)
      {
        MsgLog.log("getDeviceNumber",e.toString(),TErrorList.code_failure,e,0);                          
        return -1; 
      }
    }
    TDevice td;
    if ((td=(TDevice)myDeviceData.get(devName.toUpperCase())) == null) return -1;
    return td.devNumber;
  }
  public int getNumberOfAlarms()
  {
    TDevice[] dlst = getDeviceList();
    if (dlst == null) return 0;
    int n = 0;
    for (int i=0; i<dlst.length; i++) n += dlst[i].getNumberOfAlarms();
    return n;
  }
  public int getNumberActiveAlarms()
  {
    TDevice[] dlst = getDeviceList();
    if (dlst == null) return 0;
    int n = 0;
    for (int i=0; i<dlst.length; i++) n += dlst[i].getNumberActiveAlarms();
    return n;    
  }
  public int getMostRecentAlarmTimestamp(int[] nalarms,int index)
  {
    TDevice[] dlst = getDeviceList();
    if (dlst == null) return 0;
    int mrt = 0, ts;
    int n = 0;
    for (int i=0; i<dlst.length; i++)
    {
      if ((ts=dlst[i].getMostRecentTimestamp()) > mrt) 
      {
        mrt = ts;
        n = 0;
      }
      if (ts == mrt) n += dlst[i].getNumberOfAlarmsAtMostRecentTimestamp();
    }
    if (nalarms != null && nalarms.length > index) nalarms[index] = n;
    return mrt;    
  }
  public int getHighestAlarmSeverity(int[] nalarms,int index)
  {
    TDevice[] dlst = getDeviceList();
    if (dlst == null) return 0;
    int hsv = 0, sv;
    int n = 0;
    for (int i=0; i<dlst.length; i++)
    {
      if ((sv=dlst[i].getHighestAlarmSeverity()) > hsv) 
      {
        hsv = sv;
        n = 0;
      }
      if (sv == hsv) n += dlst[i].getNumberOfAlarmsAtHighestSeverity();
    }
    if (nalarms != null && nalarms.length > index) nalarms[index] = n;
    return hsv;    
  }
}
