/*
 * File TEquipmentManifest.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.server.equipment;

import java.io.*;
import java.text.NumberFormat;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Date;
import java.util.LinkedList;

import de.desy.tine.csvUtils.*;
import de.desy.tine.definitions.*;
import de.desy.tine.server.logger.MsgLog;
import de.desy.tine.server.properties.TStockProperties;
import de.desy.tine.startup.TInitializerFactory;
import de.desy.tine.structUtils.TTaggedStructure;

public class TEquipmentManifest extends TTaggedStructure implements RowHandler
{
  private static final String mfHdr = "EXPORT, CONTEXT, EQM, FEC, PORT, VERSION, FEC_HOME, LAST_STARTED, COMMAND_LINE, WORKING_DIRECTORY";
  private static LinkedList<TEquipmentManifest> mfLst = null;
  private static String crlf = System.getProperty("line.separator");
  private char[] exp = new char[TStrings.EXPORT_NAME_SIZE];
  public String export;
  private char[] ctx = new char[TStrings.CONTEXT_NAME_SIZE];
  public String context;
  private char[] eqmName = new char[TStrings.EQM_NAME_SIZE];
  public String eqm;
  private char[] fecName = new char[TStrings.FEC_NAME_SIZE];
  public String fec;
  private char[] ver = new char[16];
  public String version;
  private char[] fecHome = new char[32];
  public String fec_home;
  private char[] started = new char[32];
  public String last_started;
  private int[] portOffset = new int[1];
  public int port;
  private char[] cmdline = new char[132];
  public String cmdln;
  private char[] wd = new char[132];
  public String cwd;
  private void pushString(String s,char[] c)
  {
    int len = s.length();
    if (len > c.length) len = c.length;   
    Arrays.fill(c, (char)0);
    s.getChars(0, len, c, 0);
  }
  private void initStructure()
  {
    addField(exp,"export");
    addField(ctx,"context");
    addField(eqmName,"eqmName");
    addField(fecName,"fecName");
    addField(ver,"version");
    addField(fecHome,"fecHome");
    addField(started,"lastStarted");
    addField(portOffset,"port");
    addField(cmdline,"cmdLine");
    addField(wd,"curWorkDir");
    initDone();    
  }
  TEquipmentManifest()
  {
    super("FECmf");
    initStructure();
  }
  TEquipmentManifest(TEquipmentManifest reference)
  {
    super("FECmf");
    initStructure();
    pushString(reference.export,exp);
    export = new String(exp).trim();
    pushString(reference.context,ctx);
    context = new String(ctx).trim();
    pushString(reference.eqm,eqmName);
    eqm = new String(eqmName).trim();
    pushString(reference.fec,fecName);
    fec = new String(fecName).trim();
    port = portOffset[0] = reference.port;
    pushString(reference.version,ver);
    version = new String(ver).trim();
    pushString(reference.fec_home,fecHome);
    fec_home = new String(fecHome).trim();
    pushString(reference.last_started,started);
    last_started = new String(started).trim();
    pushString(reference.cmdln,cmdline);
    cmdln = new String(cmdline).trim();
    pushString(reference.cwd,wd);
    cwd = new String(wd).trim();
  }
  private String getTimeAndPid()
  {
    String ds = "";
    try
    {
      Calendar c = Calendar.getInstance();
      NumberFormat f = NumberFormat.getInstance();
      f.setMinimumIntegerDigits(2);
      ds += f.format(c.get(Calendar.HOUR_OF_DAY))+":"+
                  f.format(c.get(Calendar.MINUTE))+":"+
                  f.format(c.get(Calendar.SECOND))+" "+
                  c.get(Calendar.DAY_OF_MONTH)+"."+
                  (c.get(Calendar.MONTH)+1)+"."+
                  c.get(Calendar.YEAR);
    }
    catch (Exception ignore) {}
    ds += " ("+TInitializerFactory.getInstance().getInitializer().getPid()+")";
    return ds;
  }
  private void replace(TEquipmentModule em)
  {
    TEquipmentModuleFactory tf = TEquipmentModuleFactory.getInstance();
    pushString(em.getExportName(),exp);
    export = new String(exp).trim();
    pushString(em.getContext(),ctx);
    context = new String(ctx).trim();
    pushString(em.getLocalName(),eqmName);
    eqm = new String(eqmName).trim();
    pushString(tf.getFecName(),fecName);
    fec = new String(fec).trim();
    portOffset[0] = tf.getPort();
    port = portOffset[0];
    pushString(TStockProperties.getStockSrvVersion(),ver);
    version = new String(ver).trim();
    pushString(tf.getInitializer().getFecHome(),fecHome);
    fec_home = new String(fecHome).trim();
    pushString(getTimeAndPid(),started);
    last_started = new String(started).trim();
    String s = TInitializerFactory.getInstance().getInitializer().getSystemStartup();
    pushString(s,cmdline);
    cmdln = new String(cmdline).trim();
    s = TInitializerFactory.getInstance().getInitializer().getSystemCwd();
    pushString(s,wd);
    cwd = new String(wd).trim();
  }
  private static String getEntryAsString(TEquipmentModule em)
  {
    TEquipmentModuleFactory tf = TEquipmentModuleFactory.getInstance();
    String mfEntry = em.getExportName().trim();
    mfEntry += ", " + em.getContext().trim();
    mfEntry += ", " + em.getLocalName().trim();
    mfEntry += ", " + tf.getFecName().trim();
    mfEntry += ", " + tf.getPort();
    mfEntry += ", " + TStockProperties.getStockSrvVersion().trim();
    mfEntry += ", " + tf.getInitializer().getFecHome().trim();
    mfEntry += ", " + new Date().toString().trim();
    mfEntry += ", " + TInitializerFactory.getInstance().getInitializer().getSystemStartup();
    mfEntry += ", " + TInitializerFactory.getInstance().getInitializer().getSystemCwd();
    mfEntry += crlf;
    return mfEntry;    
  }
  private static String getEntryAsString(TEquipmentManifest tem)
  {
    String mfEntry = tem.export.trim();
    mfEntry += ", " + tem.context.trim();
    mfEntry += ", " + tem.eqm.trim();
    mfEntry += ", " + tem.fec.trim();
    mfEntry += ", " + tem.port;
    mfEntry += ", " + tem.version.trim();
    mfEntry += ", " + tem.fec_home.trim();
    mfEntry += ", " + tem.last_started.trim();
    mfEntry += ", " + tem.cmdln.trim();
    mfEntry += ", " + tem.cwd.trim();
    mfEntry += crlf;
    return mfEntry;    
  }
  private static String mfFilePath = null;
  public static String getFilePath() 
  { 
    if (mfFilePath == null)
    {
      mfFilePath = TInitializerFactory.getInstance().getInitializer().getTineCache();
      mfFilePath += File.separator + "tine" + File.separator + "cache";
      if (File.separatorChar == '\\') crlf = "\r\n";
      File csvw = new File(mfFilePath);
      if (!csvw.exists()) csvw.mkdirs();
    }
    return new String(mfFilePath); 
  }
  public static synchronized int update(TEquipmentModule em)
  { // update the manifest for this equipment module
    if (mfLst != null) mfLst.clear();
    
    csvColumn[] mfCols = new csvColumn[10];
    TEquipmentManifest mfRows = new TEquipmentManifest();
    mfCols[0] = new csvColumn("EXPORT","",new StringFieldHandler(mfRows,"export"));
    mfCols[1] = new csvColumn("CONTEXT","",new StringFieldHandler(mfRows,"context"));
    mfCols[2] = new csvColumn("EQM","",new StringFieldHandler(mfRows,"eqm"));
    mfCols[3] = new csvColumn("FEC","",new StringFieldHandler(mfRows,"fec"));
    mfCols[4] = new csvColumn("PORT","",new IntFieldHandler(mfRows,"port"));    
    mfCols[5] = new csvColumn("VERSION","",new StringFieldHandler(mfRows,"version"));
    mfCols[6] = new csvColumn("FEC_HOME","",new StringFieldHandler(mfRows,"fec_home"));
    mfCols[7] = new csvColumn("LAST_STARTED","",new StringFieldHandler(mfRows,"last_started"));
    mfCols[8] = new csvColumn("COMMAND_LINE","",new StringFieldHandler(mfRows,"cmdln"));
    mfCols[9] = new csvColumn("WORKING_DIRECTORY","",new StringFieldHandler(mfRows,"cwd"));

    String mffp = getFilePath();
    String mfFileName = mffp + File.separator + "fecmf.csv";
    // open it
    csv mfFile = new csv(mfFileName);
    // read it
    int rc = mfFile.readFile(mfCols,mfRows);
    // close it
    mfFile.close();
    if (rc != 0)
    { // TODO: remove this legacy in the next version ...
      mfFileName = mffp + File.separator + "fec.mf";
      mfFile = new csv(mfFileName);
      rc = mfFile.readFile(mfCols,mfRows);
      mfFile.close();
    }
    if (em == null) return 0; // only a read manifest required ...
    BufferedReader mfin;
    BufferedWriter mfout;
    if (rc != 0)
    { // create a new manifest
      File mf = new File(mffp);
      mf.mkdirs();
      try
      {
        mfout = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(mfFileName)));
        mfout.write(mfHdr.trim() + crlf);
        mfout.write(getEntryAsString(em));
        mfout.close();
        return 0; // we're finished
      }
      catch (Exception e)
      {
        MsgLog.log("TEquipmentManifest","cannot update FEC manifest !",TErrorList.file_error,e,0);
        return TErrorList.file_error;
      }
    }
    try
    {
      String mfTempName = mffp + File.separator + "fecmf.tmp";
      mfin = new BufferedReader(new InputStreamReader(new FileInputStream(mfFileName)));
      mfout = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(mfTempName)));
      String hdr = mfin.readLine();
      mfout.write(hdr + crlf);
      boolean found = false;
      int nfecs = mfLst.size();
      TEquipmentManifest tem;
      for (int i=0; i<nfecs; i++)    
      {
        tem = mfLst.get(i);
        if (tem.fec.length() == 0) continue;
        if (tem.eqm.length() == 0) continue;
        if (tem.export.compareToIgnoreCase(em.getExportName()) == 0 &&
            tem.context.compareToIgnoreCase(em.getContext()) == 0)
        { // this one is re-starting
          found = true;
          tem.replace(em);
        }
        mfout.write(getEntryAsString(tem));
      }
      if (!found)
      { // there's a manifest, but this one's not yet in it !
        mfout.write(getEntryAsString(em));        
      }
      mfin.close();
      mfout.close();
      File mf = new File(mfFileName);
      mf.delete();
      mf = new File(mfTempName);      
      mf.renameTo(new File(mfFileName));
      mf.setWritable(true);
    }
    catch (Exception e)
    {
      MsgLog.log("TEquipmentManifest", e.toString(),TErrorList.code_failure,e,1);      
    }
    return 0;
  }
  public int process(int index)
  { // if reading a file, add this one to a link list
    if (mfLst == null) mfLst = new LinkedList<TEquipmentManifest>();
    mfLst.add(new TEquipmentManifest(this));
    return 0;
  }
  public static LinkedList<TEquipmentManifest> getFecManifestList()
  {
    if (mfLst == null) update(null);
    return mfLst;
  }
}
