/*
 * File TStateChange.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.server.equipment;

import de.desy.tine.client.TLink;
import de.desy.tine.client.TLinkCallback;
import de.desy.tine.dataUtils.TDataType;
import de.desy.tine.definitions.TAccess;
import de.desy.tine.definitions.TErrorList;
import de.desy.tine.definitions.TMode;
import de.desy.tine.server.logger.MsgLog;
import de.desy.tine.server.logger.TFecLog;
import de.desy.tine.types.NAME64;

public class TStateChange implements TLinkCallback
{
  private int last_cc = TErrorList.not_initialized;
  private int to_counter = 0;
  private String prevState = "";
  private String thisState = "";
  private NAME64[] dStateChange = new NAME64[1]; 
  private TDataType dtStateChange = new TDataType(dStateChange);
  private TLink lnk;
  public int getTriggerStatus() { return last_cc; }
  public String getCurrentState() { return thisState; }
  public String getPreviousState() { return prevState; }
  public TStateChange(String context,String key)
  {
    if (context == null) 
    {
      MsgLog.log("TStateChange","State information unavailable for null context ",TErrorList.argument_list_error,null,0);
      return;
    }
    String dev = "/"+context+"/GLOBALS/keyword";
    try
    {
      last_cc = TErrorList.address_unknown;
      lnk = new TLink(dev,key,dtStateChange,null,TAccess.CA_READ);
      lnk.attach(TMode.CM_RECEIVE, this, 1000);
      last_cc = lnk.linkStatus;
    }
    catch (Exception e)
    {
      MsgLog.log("TStateChange","State information unavailable for context "+context,last_cc,e,0);
    }
  }
  public void callback(TLink link)
  {
    int cc = link.getLinkStatus();
    if (cc != last_cc)
    {
      last_cc = cc;
      MsgLog.log("TStateTriggerCallback","state trigger : " + link.getLastError(),cc,null,1);
    }
    if (cc != 0) 
    {
      if (to_counter++ < 5) return;
      thisState = "unavailable";
    }
    else
    {
      to_counter = 0;
    }
    thisState = dStateChange[0].getName();
    if (thisState.compareToIgnoreCase(prevState) == 0) return;
    TFecLog.log("detected state change from "+prevState+" to "+thisState);
    String ctx = link.getContext();
    String ectx = null;
    TStateChangeTrigger sct = null;
    TEquipmentModule[] eTbl = TEquipmentModuleFactory.getInstance().getEquipmentModuleTable();
    for (int i=0; i<eTbl.length; i++)
    {
      if ((ectx=eTbl[i].getContext()) == null) continue;
      if (ctx.compareToIgnoreCase(ectx) == 0)
      {
        if ((sct=eTbl[i].getStateChangeTrigger()) != null)
          sct.update(prevState,thisState);
      }
    }
    prevState = thisState;
  }    
}
