/*
 * File TSyncCallback.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.server.equipment;

import de.desy.tine.client.TLink;
import de.desy.tine.client.TLinkCallback;
import de.desy.tine.client.TLinkFactory;
import de.desy.tine.dataUtils.TDataTime;
import de.desy.tine.server.logger.DbgLog;
import de.desy.tine.server.logger.TFecLog;
import de.desy.tine.server.logger.TineLogLevel;

public class TSyncCallback implements TLinkCallback
{
  private static final int TIMESYNC_DEPTH = 6;
  private static final int TIMESYNC_NOSYNC = 0;
  private static final int TIMESYNC_TRACKED = 1;
  private static final int TIMESYNC_MONOTONIC = 2;
  private static final double gSyncTolerance = 0.1; 
  private static int last_cc = -1, last_ptr = 0, ncorrections = 0;
  private static double last_timestamp = 0;
  private static double last_logged = 0;
  private static double last_dataTimeOffset = 0, largest_correction = 0, last_correction = 0;
  private static double[] last_dataTimeRbDeltas = new double[TIMESYNC_DEPTH];
  private static double[] last_dataTimeDtDeltas = new double[TIMESYNC_DEPTH];
  private static double[] last_dataTimeStamps = new double[TIMESYNC_DEPTH];
  private static double[] last_dataTimeRbTimes = new double[TIMESYNC_DEPTH];
  private void checkTimestamp(double gDataTimeStamp,double gSyncTimeStamp)
  {
    boolean applyCorrection = false;
    int i,k,p,n=0,validCorrection = TIMESYNC_NOSYNC;;
    double correction = 0, delta_t = 0, delta_rb, delta_ts, last_dt = 0, last_rb;
    last_rb = last_dataTimeRbTimes[last_ptr];
    last_dt = last_dataTimeStamps[last_ptr];

    if (gSyncTimeStamp == last_dt) 
    { /* extra broadcast/multicast of the same */
      if (TLinkFactory.debugLevel > 1) DbgLog.log("checkTimestamp","discard extra synchronization timestamp");
      return; 
    }
    
    last_ptr = (last_ptr + 1) % TIMESYNC_DEPTH;
    last_dataTimeRbTimes[last_ptr] = gDataTimeStamp;            /* my clock's current timestamp */
    last_dataTimeRbDeltas[last_ptr] = gDataTimeStamp - last_rb; /* time difference based on my clock */
    last_dataTimeStamps[last_ptr] = gSyncTimeStamp;             /* remote clock's current timestamp */
    last_dataTimeDtDeltas[last_ptr] = gSyncTimeStamp - last_dt; /* time difference based on remote clock */

    for (i=0; i<TIMESYNC_DEPTH; i++)
    {
      k = (last_ptr + i) % TIMESYNC_DEPTH;
      p = (k - 1 + TIMESYNC_DEPTH) % TIMESYNC_DEPTH;
      if (last_dataTimeStamps[k] < 1000) return;
      if (last_dataTimeStamps[p] < 1000) return;
      delta_rb = last_dataTimeRbDeltas[k] - last_dataTimeRbDeltas[p];
      delta_ts = last_dataTimeDtDeltas[k] - last_dataTimeDtDeltas[p];
      if (k != last_ptr + 1)
      {
        if (delta_rb > 10.0 || delta_rb < -10.0) return; /* just getting started ? */
        if (delta_ts > 10.0 || delta_ts < -10.0) return;
      }
      delta_t = last_dataTimeRbDeltas[k] - last_dataTimeDtDeltas[k];
      if (validCorrection != TIMESYNC_MONOTONIC && Math.abs(delta_t) < 0.01) 
      { /* both local and remote clocks are tracking smoothly */
        validCorrection = TIMESYNC_TRACKED;
      }
      else if (validCorrection != TIMESYNC_TRACKED && 
               Math.abs(delta_rb) < 0.01 && 
               Math.abs(delta_ts) < 0.01)
      { /* both clocks are increasing monotonically */
        validCorrection = TIMESYNC_MONOTONIC;
      }
      else
      { /* these deltas should be really close if the data is coming in and being read at the correct rate */
        return;
      }
      correction += last_dataTimeStamps[k] - last_dataTimeRbTimes[k];
      n++;
      if (validCorrection == TIMESYNC_MONOTONIC && i == TIMESYNC_DEPTH-2) break;
    }
    if (n == 0) return;
    correction /= n;
    if (correction > gSyncTimeStamp)
    {
      TFecLog.log(TineLogLevel.WARN,"invalid time correction " + correction + 
                  " greater than time server clock " + gSyncTimeStamp);
      return;
    }

    if (correction > gSyncTolerance || correction < -gSyncTolerance) applyCorrection = true;
    else if (last_correction < gSyncTolerance && correction > gSyncTolerance/2) applyCorrection = true;
    else if (last_correction > gSyncTolerance && correction < -gSyncTolerance/2) applyCorrection = true;
    if (!applyCorrection) return;
    synchronized (TDataTime.syncMutex) 
    {
      double tsOffset = (double)(TDataTime.getDataTimeStampOffset()) / 1000.0;
      tsOffset += correction;
      TDataTime.setDataTimeStampOffset((long)(tsOffset*1000));
      if (TLinkFactory.isRunningAsServer())
      {
        TEquipmentModuleFactory.setDataTimeStampOffset((long)(tsOffset*1000));
      }
      last_correction = correction;
      if (TLinkFactory.debugLevel > 0) 
        DbgLog.log("checkTimestamp","global synchronization (" + 
            (validCorrection == TIMESYNC_TRACKED ? "tracking" : "monotonic") +
            ") : " + correction + " sec (current offset " + tsOffset + " sec)");
      ncorrections++;
      if (Math.abs(correction) > largest_correction) largest_correction = Math.abs(correction);
      if (gDataTimeStamp < last_logged + 3600) return;
      last_logged = gDataTimeStamp;
      if (Math.abs(tsOffset-last_dataTimeOffset) < 0.1) return;
      TFecLog.log("global synchronization offset : " + tsOffset + "sec");
      TFecLog.log("largest correction : " + largest_correction + " sec  (nr. corrections " + ncorrections + ")");
      last_logged = gDataTimeStamp; ncorrections = 0; largest_correction = 0;
      last_dataTimeOffset = tsOffset;      
    }
  }
  public void callback(TLink link)
  {
    int cc = link.getLinkStatus();
    if (cc != last_cc)
    {
      last_cc = cc;
      TFecLog.log((cc==0 ? TineLogLevel.INFO : TineLogLevel.ERROR),
                  "global synchronization : " + link.getLastError());
    }
    if (cc != 0) return;
    double this_timestamp = TDataTime.getSyncTimeStamp();
    if (last_timestamp == this_timestamp)
    { // multiple callbacks (due to extra broadcasting of the same thing)
      if (TLinkFactory.debugLevel > 3) 
        DbgLog.log("checkTimestamp","discard extra synchronization timestamp");
      return;
    }
    last_timestamp = this_timestamp;
    checkTimestamp(TDataTime.getDataTimeStamp(), this_timestamp);
  }    
}
