/*
 * File THistoryRecord.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

/*
 * Created on Aug 31, 2005
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package de.desy.tine.server.histories;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.io.*;
import java.time.LocalDateTime;
import java.time.LocalTime;
import java.time.ZoneOffset;

import de.desy.tine.client.TFilterLink;
import de.desy.tine.client.TLinkFactory;
import de.desy.tine.console.TCommandItem;
import de.desy.tine.dataUtils.*;
import de.desy.tine.definitions.*;
import de.desy.tine.endianUtils.Swap;
import de.desy.tine.exceptions.TineRuntimeErrorException;
import de.desy.tine.histUtils.THistory;
import de.desy.tine.server.equipment.TEquipmentModule;
import de.desy.tine.server.equipment.TEquipmentModuleFactory;
import de.desy.tine.server.logger.DbgLog;
import de.desy.tine.server.logger.MsgLog;
import de.desy.tine.server.logger.TFecLog;
import de.desy.tine.server.logger.TineLogLevel;
import de.desy.tine.server.properties.TExportProperty;
import de.desy.tine.server.properties.TPropertyDescription;
import de.desy.tine.server.properties.TPropertyList;
import de.desy.tine.startup.TInitializerFactory;
import de.desy.tine.structUtils.TStructDescription;
import de.desy.tine.structUtils.TStructDescription.Field;
import de.desy.tine.structUtils.TStructRegistry;
import de.desy.tine.structUtils.TTaggedStructure;
import de.desy.tine.types.*;

public class THistoryRecord
{
	public class THistoryRequest
	{
		int index;
		int fieldindex;
		int sample;
		public THistoryRequest(int index,int fieldindex,int sample)
		{
			this.index = index;
			this.fieldindex = fieldindex;
			this.sample = sample;
		}
	}
	public class THistoryRange
	{
	  long starttime; /**< time range start */
	  long stoptime; /**< time range stop */
	  float lwrLimit;  /**< lower value limit (if filter given) */
	  float uprLimit;  /**< upper value limit (if filter given) */
	  long startstamp; /**< system/user stamp range start */
	  long stopstamp; /**< system/user stamp range stop */
	  boolean userstamp; /**< true => filter on user-stamp instead of system stamp */
	  public THistoryRange(long starttime,long stoptime,long startstamp,long stopstamp,float lwr,float upr,boolean userStamp)
	  {
	  	this.starttime = starttime;
	  	this.stoptime = stoptime;
	  	this.startstamp = startstamp;
	  	this.stopstamp = stopstamp;
	  	this.lwrLimit = lwr;
	  	this.uprLimit = upr;
	  	this.userstamp = userStamp;
	  }
	  public THistoryRange(long starttime,long stoptime,int startstamp,int stopstamp,float lwr,float upr,boolean userStamp)
    {
      this.starttime = starttime;
      this.stoptime = stoptime;
      this.startstamp = startstamp;
      this.stopstamp = stopstamp;
      this.lwrLimit = lwr;
      this.uprLimit = upr;
      this.userstamp = userStamp;
    }
    public void setStartTime(long start) { starttime = start; }
	}
  private static final long msecondsInMonth = (long)3600 * 24 * 31 * 1000;
  private static final int AS_STATE_SIZE = 16;
  public static final int HRS_FREE_RECORD_INDEX_OFFSET = 100;
  private TEquipmentModule eqm;
  private String prp;   
  private String dev; 
  private int siz;
  private byte fmt;
  private String tag = null;
  private int  aType;
  private THistorySpecification hspec;
  public THistorySpecification getHistorySpecification() { return hspec; }
  private int  rngPtr;
  private int  rcdIdx;       // local index identifier
  private int  rngBufferSize;
  private float range = 0;
  private long lastArchiveTime;
  private long lastAccessTime;
  private int lastArchiveStatus = 0;
  public int getLastArchiveStatus() { return lastArchiveStatus; }
  public long getLastAccessTime() { return lastAccessTime; }
  public void setLastAccessTime(long time) { lastAccessTime = time; }
  private String ArchiveFileName; // current working file name
  private int lastArchivedDayOfTheMonth = -1;
  private boolean useMonthlyHistoryFiles = false;
  private boolean useMinimalStorage = false; // changed to false 15.1.09
  private boolean useStandardFileSet = false;
  private boolean storeSummary = false;
  public boolean isStoreSummary() { return storeSummary; }
  private TFilterLink filter = null;
  public TFilterLink getFilter() { return filter; }
  public void setFilter(TFilterLink filter) 
  { 
    String fstr = filter == null ? "none" : filter.toString();
    TFecLog.log("THistoryRecord","set filter to "+fstr);
    this.filter = filter; 
  }
  public boolean isPastFilter()
  {
    if (filter != null && !filter.isActive())
    { // marked for removal;
      filter = null;
    }
    return filter == null ? true : filter.isValid();
  }
  public void setUseStandardFileSet(boolean value)
  {
    useStandardFileSet = true;
  }
  public boolean getUseStandardFileSet() { return useStandardFileSet; }
  private Calendar calendar;
  private long[] timestamp; // timestamp array
  private int[] sysstamp; // system stamp array
  private int[] usrstamp; // user stamp array
  private long lastReadTimestamp = 0;
  private long lastReadSystemStamp = 0;
  private long lastReadUserStamp = 0;
  private boolean isScheduled = false;
  private boolean keepPointsOfInterest = false;
  public void setKeepPointsOfInterest(boolean value)
  {
    keepPointsOfInterest = value;
  }
  public boolean getKeepPointsOfInterest() { return keepPointsOfInterest; }
  private boolean isInitialized = false;
  public boolean isInitialized() { return isInitialized; }
  private static String arcDbPath = TInitializerFactory.getInstance().getInitializer().getHistoryHome();
  private static String arcDbPathSaved = arcDbPath + File.separator + "SAVED";
  private boolean hasSavedFiles = false;
  public boolean hasSavedFiles() { return hasSavedFiles; }
  public String dump()
  {
    StringBuffer sb = new StringBuffer(512);
    sb.append(prp+" "+dev+" : "+siz+" "+TFormat.toString(fmt)+" values");
    sb.append("("+hspec.getDepthShort()+" deep, "+hspec.getDepthLong()+" months)\n");
    sb.append("\tinitialized : "+ (rngPtr != -1) + "\n");
    sb.append("\tscheduled : "+isScheduled+"\n");
    sb.append("\tkeeps points of interest : "+keepPointsOfInterest+"\n");
    sb.append("\tpolling interval : "+hspec.getPollngRate()+" msec\n");
    sb.append("\tarchive intervals : minimum "+hspec.getArchiveRate()+" msec, maximum "+hspec.getHeartbeat()+" sec\n");
    sb.append("\tarchive tolerance : absolute "+hspec.getAbsoluteTolerance()+", relative "+hspec.getPercentTolerance()+" percent\n");
    sb.append("\trecord length : "+getRecordLength()+" bytes\n");
    sb.append("\tcurrent file name : "+ArchiveFileName+"\n");
    sb.append("\tcurrent data timestamp : "+TDataTime.toString(lastAccessTime)+"\n");
    sb.append("\tlast archived : "+TDataTime.toString(lastArchiveTime)+"\n");
    if (filter != null)
    {
      sb.append("\filter : "+filter.getFilterString()+" : valid = "+filter.isValid()+"\n");
    }
    return sb.toString();
  }
  public static void setHistoryStaticFilesRepository(String path)
  {
    if (path == null) return;
    if (path.length() == 0) path = ".";
    if (path.endsWith("/") || path.endsWith("\\")) path = path.substring(0, path.length()-1);
    arcDbPathSaved = path;
  }
  public static String getHistoryStaticFilesRepository() { return arcDbPathSaved; }
  private boolean checkForStaticFiles()
  {
    boolean ff = false;
    String fn = null;
    String ext = "." + Integer.toHexString(rcdIdx);
    try
    {
      File adbp = new File(arcDbPathSaved);
      for (File f : adbp.listFiles())
      {
        if (!f.isFile()) continue;
        fn = f.getName();
        if (!fn.endsWith(ext)) continue;
        if (fn.startsWith("ts")) ff = true;
        if (fn.startsWith("ta")) ff = true;
        if (fn.startsWith("ar")) ff = true;
        if (!ff) continue;
        break;
      }
    } catch (Exception ignore) {}      
    return ff;
  }
  public void setScheduled(boolean value) { isScheduled = value; }
  public boolean isScheduled() { return isScheduled; }
  private int getRecordPrefixLength() { return useMinimalStorage ? 4 : 16; }
  Object rngBuffer;
  public static boolean isAllowableFormat(String fmt)
  {
    return isAllowableFormat(TFormat.getFormatCode(fmt));
  }
  public static boolean isAllowableFormat(short fmt)
  {
    switch (fmt)
    {
    	case TFormat.CF_HISTORY:
    	case TFormat.CF_HITTS:
    	case TFormat.CF_ASPECTRUM:
    	case TFormat.CF_SPECTRUM:
    	case TFormat.CF_IMAGE:
    	case TFormat.CF_AIMAGE:
    	case TFormat.CF_STRING:
    	case TFormat.CF_DEFAULT:
    	case TFormat.CF_NULL:
    	case TFormat.CF_XML:
    	case TFormat.CF_DBLINT:
    	case TFormat.CF_SHTINT:
    	case TFormat.CF_MDA:
    		return false;
    }
    return true;
  }
  private static boolean checkForStandardFilesDone = false;
  private static String[] stdfiles = null;
  private boolean checkForStandardFiles()
  {
    if (!checkForStandardFilesDone)
    {
      ArrayList<String> lst = new ArrayList<String>(); 
      String fn = null;
      try
      {
        File adbp = new File(arcDbPath);
        for (File f : adbp.listFiles())
        {
          if (!f.isFile()) continue;
          fn = f.getName();
          if (!fn.startsWith("ts")) continue;
          lst.add(fn);
        }
        stdfiles = lst.toArray(new String[0]);
      } catch (Exception ignore) {}      
      checkForStandardFilesDone = true;
    }
    if (stdfiles == null || stdfiles.length == 0) return false;
    for (String f : stdfiles)
    {
      if (f.endsWith("."+Integer.toHexString(rcdIdx))) return true;
    }
    return false;
  }
  private THistoryRecord(THistoryRecord reference)
  {
    eqm = reference.eqm;
    prp = reference.prp;
    dev = reference.dev;
    siz = reference.siz;
    fmt = reference.fmt;
    aType = reference.aType;
    rcdIdx = reference.rcdIdx;
    hspec = reference.hspec;
    rngBufferSize = reference.rngBufferSize;
    rngBuffer = reference.rngBuffer;
    rngPtr = reference.rngPtr;
    timestamp = reference.timestamp;
    sysstamp = reference.sysstamp;
    usrstamp = reference.usrstamp;
    useMinimalStorage = reference.useMinimalStorage;
    useMonthlyHistoryFiles = reference.useMonthlyHistoryFiles;
    isInitialized = reference.isInitialized;
  }
  private Object makeRingBuffer(byte fmt, String tag, int size)
  {
  	Object rb = null;
    switch (fmt)
    {
      case TFormat.CF_BYTE:
        rb = new byte[size];
        break;
      case TFormat.CF_INT16:
        rb = new short[size];
        break;
      case TFormat.CF_BOOLEAN:
      case TFormat.CF_INT32:
        rb = new int[size];
        break;
      case TFormat.CF_INT64:
        rb = new long[size];
        break;        
      case TFormat.CF_FLOAT:
        rb = new float[size];
        break;
      case TFormat.CF_DOUBLE:
        rb = new double[size];
        break;
      case TFormat.CF_NAME16:
        rb = new NAME16[size];
        for (int i=0; i<size; i++) ((NAME16[])rb)[i] = new NAME16();
        break;
      case TFormat.CF_NAME32:
        rb = new NAME32[size];
        for (int i=0; i<size; i++) ((NAME32[])rb)[i] = new NAME32();
        break;
      case TFormat.CF_NAME64:
        rb = new NAME64[size];
        for (int i=0; i<size; i++) ((NAME64[])rb)[i] = new NAME64();
        break;
      case TFormat.CF_FLTINT:
        rb = new FLTINT[size];
        for (int i=0; i<size; i++) ((FLTINT[])rb)[i] = new FLTINT();
        break;
      case TFormat.CF_FLTFLT:
        rb = new FLTFLT[size];
        for (int i=0; i<size; i++) ((FLTFLT[])rb)[i] = new FLTFLT();
        break;
      case TFormat.CF_INTINT:
        rb = new INTINT[size];
        for (int i=0; i<size; i++) ((INTINT[])rb)[i] = new INTINT();
        break;
      case TFormat.CF_DBLDBL:
        rb = new DBLDBL[size];
        for (int i=0; i<size; i++) ((DBLDBL[])rb)[i] = new DBLDBL();
        break;
      case TFormat.CF_DBLTIME:
        rb = new DBLTIME[size];
        for (int i=0; i<size; i++) ((DBLTIME[])rb)[i] = new DBLTIME();
        break;
      case TFormat.CF_NAME16I:
        rb = new NAME16I[size];
        for (int i=0; i<size; i++) ((NAME16I[])rb)[i] = new NAME16I();
        break;
      case TFormat.CF_NAME32I:
        rb = new NAME32I[size];
        for (int i=0; i<size; i++) ((NAME32I[])rb)[i] = new NAME32I();
        break;
      case TFormat.CF_NAME64I:
        rb = new NAME64I[size];
        for (int i=0; i<size; i++) ((NAME64I[])rb)[i] = new NAME64I();
        break;
      case TFormat.CF_NAME64DBL:
        rb = new NAME64DBL[size];
        for (int i=0; i<size; i++) ((NAME64DBL[])rb)[i] = new NAME64DBL();
        break;
      case TFormat.CF_NAME64TIME:
        rb = new NAME64TIME[size];
        for (int i=0; i<size; i++) ((NAME64TIME[])rb)[i] = new NAME64TIME();
        break;
      case TFormat.CF_NAME16FI:
        rb = new NAME16FI[size];
        for (int i=0; i<size; i++) ((NAME16FI[])rb)[i] = new NAME16FI();
        break;
      case TFormat.CF_NAME16II:
        rb = new NAME16II[size];
        for (int i=0; i<size; i++) ((NAME16II[])rb)[i] = new NAME16II();
        break;
      case TFormat.CF_NAME16DBLDBL:
        rb = new NAME16DBLDBL[size];
        for (int i=0; i<size; i++) ((NAME16DBLDBL[])rb)[i] = new NAME16DBLDBL();
        break;
      case TFormat.CF_NAME32DBLDBL:
        rb = new NAME32DBLDBL[size];
        for (int i=0; i<size; i++) ((NAME32DBLDBL[])rb)[i] = new NAME32DBLDBL();
        break;
      case TFormat.CF_NAME64DBLDBL:
        rb = new NAME64DBLDBL[size];
        for (int i=0; i<size; i++) ((NAME64DBLDBL[])rb)[i] = new NAME64DBLDBL();
        break;
      case TFormat.CF_INTINTINT:
        rb = new INTINTINT[size];
        for (int i=0; i<size; i++) ((INTINTINT[])rb)[i] = new INTINTINT();
        break;
      case TFormat.CF_INTFLTINT:
        rb = new INTFLTINT[size];
        for (int i=0; i<size; i++) ((INTFLTINT[])rb)[i] = new INTFLTINT();
        break;
      case TFormat.CF_FLTINTINT:
        rb = new FLTINTINT[size];
        for (int i=0; i<size; i++) ((FLTINTINT[])rb)[i] = new FLTINTINT();
        break;
      case TFormat.CF_FLTFLTINT:
        rb = new FLTFLTINT[size];
        for (int i=0; i<size; i++) ((FLTFLTINT[])rb)[i] = new FLTFLTINT();
        break;
      case TFormat.CF_DBLDBLDBL:
        rb = new DBLDBLDBL[size];
        for (int i=0; i<size; i++) ((DBLDBLDBL[])rb)[i] = new DBLDBLDBL();
        break;
      case TFormat.CF_INTINTINTINT:
        rb = new INTINTINTINT[size];
        for (int i=0; i<size; i++) ((INTINTINTINT[])rb)[i] = new INTINTINTINT();
        break;
      case TFormat.CF_FLTINTFLTINT:
        rb = new FLTINTFLTINT[size];
        for (int i=0; i<size; i++) ((FLTINTFLTINT[])rb)[i] = new FLTINTFLTINT();
        break;
      case TFormat.CF_INTINTFLTFLT:
        rb = new INTINTFLTFLT[size];
        for (int i=0; i<size; i++) ((INTINTFLTFLT[])rb)[i] = new INTINTFLTFLT();
        break;
      case TFormat.CF_INTFLTFLTFLT:
        rb = new INTFLTFLTFLT[size];
        for (int i=0; i<size; i++) ((INTFLTFLTFLT[])rb)[i] = new INTFLTFLTFLT();
        break;
      case TFormat.CF_NAME64DBLDBLDBL:
        rb = new NAME64DBLDBLDBL[size];
        for (int i=0; i<size; i++) ((NAME64DBLDBLDBL[])rb)[i] = new NAME64DBLDBLDBL();
        break;
      case TFormat.CF_USTRING:
        rb = new USTRING[size];
        for (int i=0; i<size; i++) ((USTRING[])rb)[i] = new USTRING();
        break;
      case TFormat.CF_UNAME:
        rb = new UNAME[size];
        for (int i=0; i<size; i++) ((UNAME[])rb)[i] = new UNAME();
        break;
      case TFormat.CF_STRUCT:
      	if (TStructRegistry.get(tag) == null) return null;
      	rb = new TTaggedStructure[size];
      	for (int i=0; i<size; i++)
      	{
      		((TTaggedStructure[])rb)[i] = new TTaggedStructure(tag,true);
      	}
      	break;
      default:
        TFecLog.log(TineLogLevel.ERROR,"history record of format type " + TFormat.toString(fmt) + " not allowed!");
        return null;
    }
  	return rb;
  }
  public THistoryRecord(TEquipmentModule equipmentModule,String devName,String devProperty,int size,
      int format,int arrayType,int recordIndex,THistorySpecification hstSpecification)
  {
  	this(equipmentModule,devName,devProperty,size,format,null,arrayType,recordIndex,hstSpecification);
  }
  public THistoryRecord(TEquipmentModule equipmentModule,String devName,String devProperty,int size,
      int format,String tag,int arrayType,int recordIndex,THistorySpecification hstSpecification)
  {
    rngPtr = -1; // ring buffer as yet empty !
    eqm = equipmentModule;
    prp = devProperty;
    dev = devName;
    siz = size;
    fmt = (byte)format;
    aType = arrayType;
    rcdIdx = recordIndex;
    hspec = hstSpecification;
    if (eqm == null)
    {
      TFecLog.log(TineLogLevel.ERROR,"THistoryRecord: equipment module null pointer!");
      return;
    }
    int rblen = siz;
    if (fmt == TFormat.CF_STRUCT)
    {
    	int ssiz = TStructRegistry.getSizeInBytes(tag);
    	if (ssiz > 0) rblen /= ssiz; // the ring buffer is an array of TTaggedStructures ...
    }
    if (hspec == null) hspec = THistorySpecification.getDefaultInstance();
    int pintvl = hspec.getPollngRate();
    if (pintvl < eqm.getHistoryCycleInterval())
    {
      eqm.setHistoryCycleInterval(pintvl);
    }
    if (!eqm.hasProperty(prp))
    {
      TFecLog.log(TineLogLevel.WARN,"THistoryRecord: equipment module does not have property "+prp+"!");
      return;      
    }
    if (siz == 1) keepPointsOfInterest = true;
    TPropertyList plst = eqm.getPropertyList();
    TPropertyDescription dsc = plst.getFirstProperty(prp).getDescription();
    if (dsc != null)
    {
      if (TArrayType.isChannel(dsc.getArrayType())) keepPointsOfInterest = true;
      range = dsc.getYRange().getMaxValue() - dsc.getYRange().getMinValue(); 
    }
    if (hspec.getPollngRate() <= 0)
    {
      TFecLog.log(TineLogLevel.ERROR,"THistoryRecord: polling interval "+hspec.getPollngRate()+" for property "+prp+" is not allowed");
      return;            
    }
    if (fmt == TFormat.CF_STRUCT && (tag == null || tag.length() == 0))
    {
    	tag = plst.getFirstProperty(prp).getOutputTag();
    }
    this.tag = tag;
    TFecLog.log("keep history of " + devName + " " + devProperty);
    if (hspec.getDepthLong() == 0 && hspec.getArchiveRate() > hspec.getPollngRate())
    { // signal to keep a data summary in long-term storage
      storeSummary = true;
      keepPointsOfInterest = false;
    }
    useStandardFileSet = checkForStandardFiles();
    rngBufferSize = rblen * hspec.getDepthShort();
    rngBufferSize += rblen; // add space to store last archived record
    rngBuffer = makeRingBuffer(fmt, tag, rngBufferSize);
    if (rngBuffer == null) return;
    rngBufferSize -= rblen; // remove extra storage space from official size 
    timestamp = new long[rngBufferSize];
    sysstamp = new int[rngBufferSize];
    usrstamp = new int[rngBufferSize];
    if (checkForStaticFiles()) hasSavedFiles = true;
    if (TArrayType.isChannel(dsc.getArrayType())) chkMcaIntegrity();
    isInitialized = true;
  }
  void chkMcaIntegrity()
  {
    if (!TEquipmentModuleFactory.getAutoRemoveUnsynchronizedHistories()) return;
    boolean insync = true;
    String[] devs = eqm.getDevicesFromManifest(prp);
    TPropertyList tpl = eqm.getPropertyList();
    TExportProperty p = tpl != null ? tpl.getFirstProperty(prp) : null;
    ArrayList<String> dal = p != null ? p.getDeviceList() : null;
    String[] pdevs = dal != null ? dal.toArray(new String[0]) : new String[] {};
    if (pdevs.length == 0) pdevs = eqm.getDeviceNames().toArray(new String[0]);
    String err = "";
    if (devs.length != pdevs.length) 
    {
      err = "device manifest contains "+ (devs.length > pdevs.length ? "more" : "fewer")+
            "devices than current configuration";
      insync = false;
    }
    for (int i=0; insync && i<devs.length; i++)
    {
      if (!devs[i].equalsIgnoreCase(pdevs[i])) 
      {
        err = devs[i]+" from manifest does not equal "+pdevs[i]+" from current configration";
        insync = false;
      }
    }
    if (!insync)
    {
      resetLTF();
      TFecLog.log(TineLogLevel.WARN,"reset local history for "+ eqm+"/"+dev+"["+prp+"] : " + err);
    }
  }
  void resetLTF()
  {
    String ext = "." + Integer.toHexString(rcdIdx);
    String fn;
    File folder = new File(arcDbPath); 
    for (File f : folder.listFiles()) 
    {
      fn = f.getName();
      if (fn.endsWith(ext)) 
      {
        if (!useStandardFileSet) 
        {
          f.delete(); 
        }
        else
        {
          try
          {
            RandomAccessFile raf = new RandomAccessFile(fn,"rw");
            setFilePosition(raf, getRecordLength());
          } catch (Exception x) {}
        }
      }
    }    
  }
  public int writeRecordToPIF(TDataType dout)
  {
    THistoryRecord thr = new THistoryRecord(this);
    char[] afn = this.ArchiveFileName.toCharArray();
    "pi".getChars(0, 2, afn, 0);
    thr.ArchiveFileName = new String(afn);
    thr.useStandardFileSet = false;
    thr.writeRecordToLTS(dout);
    return 0;
  }
  public int writeRecordToLTS(TDataType dout)
  {
    return writeRecordToLTS(dout,null);
  }
  public int writeRecordToLTS(TDataType dout, String path)
  { // write to Long-Term-Storage (i.e. disk)
    Object values = dout.getDataObject();
    long ts = dout.getDataTimeStamp();
    short fmt = getFmt();
    String fileName;
    if (ts < 1000)
    { // either not set or invalid
      DbgLog.log("writeRecordToLTS","invalid timestamp "+ts+" refused");
      return 0;
    }
    try 
    {
      if (path == null) path = arcDbPath;
      File d = new File(path);
      if (!d.exists()) d.mkdirs();
      DSUMMARY ds = null;
      if (storeSummary)
      {
        double si = hspec.getAbsoluteTolerance()/1000.0;
        int cc = 0;
        try { ds = getDataSummaryFromSTS(si); } 
        catch (TineRuntimeErrorException e) { cc = e.getErrorCode(); }
        if (cc != 0) return cc;
        double[] dvals = new double[6];
        dvals[0] = ds.getAverage();
        dvals[1] = ds.getDeviation();
        dvals[2] = ds.getMax();
        dvals[3] = ds.getMin();
        values = dvals;
        fmt = TFormat.CF_DOUBLE;
        siz = 6; // DSUMMARY = effectively 6 doubles
      }
      int nsiz = TFormat.formatSizeOf(fmt);
      ByteArrayOutputStream baos = new ByteArrayOutputStream(getRecordPrefixLength()+siz*nsiz);
      DataOutputStream os = new DataOutputStream(baos);
      //n.b. writeDouble(), etc. writes BIG-Endian to disk
      //     the compound types write LITTLE-Endian to disk
      if (!useMinimalStorage)
      {
        os.writeDouble((double)ts/1000.0);
        os.writeInt(dout.getSystemDataStamp());
        os.writeInt(dout.getUserDataStamp());
      }
      else
      {
        os.writeInt((int)(ts/1000));
      }
      switch (fmt)
      {
        case TFormat.CF_BYTE:
          for (int i=0; i<siz; i++) os.writeByte(((byte[])values)[i]);
          break;
        case TFormat.CF_INT16:
          for (int i=0; i<siz; i++) os.writeShort(((short[])values)[i]);
          break;
        case TFormat.CF_INT32:
          for (int i=0; i<siz; i++) os.writeInt(((int[])values)[i]);
          break;
        case TFormat.CF_INT64:
          for (int i=0; i<siz; i++) os.writeLong(((long[])values)[i]);
          break;          
        case TFormat.CF_FLOAT:
          for (int i=0; i<siz; i++) os.writeFloat(((float[])values)[i]);
          break;
        case TFormat.CF_DOUBLE:
          for (int i=0; i<siz; i++) os.writeDouble(((double[])values)[i]);
          break;
        case TFormat.CF_NAME16:
        case TFormat.CF_NAME32:
        case TFormat.CF_NAME64:
          for (int i=0; i<siz; i++) os.write(((NAME[])values)[i].toByteArray(),0,nsiz);
          break;
        case TFormat.CF_STRUCT:
          for (int i=0; i<siz; i++) os.writeByte(((byte[])values)[i]);
          break;        	
        default:
          if (values instanceof TCompoundDataObject[])
          {
            for (int i=0; i<siz; i++) 
              os.write(((TCompoundDataObject[])values)[i].toByteArray(),0,nsiz);
          }
      }
      fileName = path +  File.separator + ArchiveFileName;
      RandomAccessFile raf = new RandomAccessFile(fileName,"rw");
      int p = getFilePosition(raf);
      raf.seek(p); 
      raf.write(baos.toByteArray());
      setFilePosition(raf, p + getRecordLength());
      raf.close();
      lastArchiveTime = ts;
      // write the 'last archived data into the extended space of the ring buffer :
      writeRecordToRingBuffer(values, hspec.getDepthShort());
      lastArchiveStatus = 0;
    }
    catch (Exception e)
    {
      TFecLog.log(TineLogLevel.ERROR,"HST: cannot write record " + getDev() + " " + getPrp());
      TEquipmentModuleFactory.getInstance().setFileErrorAlarm(getArchiveFileName());
      lastArchiveStatus = TErrorList.file_error;
    }
    return 0;
  }
  public int saveShortTermData(THistoryRange rng)
  {
    int depth = hspec.getDepthShort();
    int cnt,cc = 0;
    int cp,n,number;
    int prate = hspec.getPollngRate();
    boolean useSysStamp = false;
    long dts;
    int sds, uds;
    Object ov = null;
    TDataType dout;
    lastReadTimestamp = 0;
    lastReadSystemStamp = 0;
    if (prate == 0) return 0;
    if (depth == 0) return TErrorList.un_allocated;
    if (rng.startstamp > 0 && rng.stopstamp >= rng.startstamp) useSysStamp = true;
    if (rng.starttime < rng.stoptime - depth*prate) rng.starttime = rng.stoptime - depth*prate;
    cp = n = rngPtr; // current position
    n = (n + 1 + depth) % depth; // move raster past the current position to start at the beginning
    lastReadTimestamp = rng.starttime; lastReadSystemStamp = rng.startstamp;
    dout = new TDataType(siz,fmt);
    for (number=0,cnt=0; number <= depth && cnt < depth; cnt++)
    {
      dts = timestamp[n];
      sds = sysstamp[n];
      uds = usrstamp[n];
      n = (n + 1 + depth) % depth;
      if (useSysStamp)
      {
        if (sds < lastReadSystemStamp || sds > rng.stopstamp) continue;
      }
      else 
      {
        if (dts < lastReadTimestamp || dts > rng.stoptime) continue;
      }
      lastReadTimestamp = dts;
      lastReadSystemStamp = sds;
      switch (fmt)
      {
        case TFormat.CF_DOUBLE:
          ov = getDoubleArrayAtPosition(n);
          break;
        case TFormat.CF_FLOAT:
          ov = getFloatArrayAtPosition(n);
          break;
        case TFormat.CF_INT32:
          ov = getLongArrayAtPosition(n);
          break;
        case TFormat.CF_INT16:
          ov = getShortArrayAtPosition(n);
          break;
        default:
          ov = getNAMEArrayAtPosition(n);
          if (ov == null) ov = getCompoundArrayAtPosition(n);
          if (ov == null) ov = getDoubleArrayAtPosition(n);
          if (ov == null) ov = getStructArrayAtPosition(n);
          break;
      }
      if (ov == null) 
      {
        cc = TErrorList.illegal_format;
        break;
      }
      dout.dTimestamp = dts;
      dout.sysDataStamp = sds;
      dout.usrDataStamp = uds;
      dout.putData(ov);
      writeRecordToLTS(dout,arcDbPathSaved);      
    }        
    hasSavedFiles = true;    
    rngPtr = cp; // restore the current pointer
    return cc;
  }
  public long getLastReadTimestamp() { return lastReadTimestamp; }
  public long getLastReadSystemStamp() { return lastReadSystemStamp; }
  public long getLastReadUserStamp() { return lastReadUserStamp; }
  public int getNumDaysInMonth(int year,int month)
  {
    boolean isLeapYear = false;
    if ((year % 4) == 0) isLeapYear = true;
    switch (month)
    {
      case 0: case 2: case 4: case 6: case 7: case 9: case 11: return 31;
      case 1: return isLeapYear ? 29 : 28;
      default: return 30;
    }
  }
  public int getNumDaysInTimeSpan(long t1,long t2)
  {
    Calendar calendar = Calendar.getInstance();    
    int n = 0, nx, y, d, ms, me, i;

    calendar.setTimeInMillis(t1);
    int y1 = calendar.get(Calendar.YEAR);
    int m1 = calendar.get(Calendar.MONTH);
    int d1 = calendar.get(Calendar.DAY_OF_MONTH);
    
    calendar.setTimeInMillis(t2);
    int y2 = calendar.get(Calendar.YEAR);
    int m2 = calendar.get(Calendar.MONTH);
    int d2 = calendar.get(Calendar.DAY_OF_MONTH);

    d = getNumDaysInMonth(y2,m2);
    nx = d1 + (d-d2-1);
    for (y=y1; y<y2+1; y++)
    {
      ms = y == y1 ? m1 : 0;
      me = y == y2 ? m2 : 11;
      for (i=ms; i<me+1; i++) 
      {
        n += getNumDaysInMonth(y,i);
      }
    }
    return n - nx;
  }
  public int getDataFromLTS(THistoryRequest req,THistoryRange rng,
      Object dataTarget,int arraysize,int arrayOffset)
  {
    String fn;
    int ps;
    int recsread=0,recsfound=0,rnumber=0,asize,daystotal,daysinmonth;
    boolean finalPass=false, needAnotherPass=false;
    long t = TDataTime.now(),itime = 0,stsTime;
    THistorySpecification spc = getHspec();
    
    if (req.index < 0 || req.index > siz) return -TErrorList.out_of_range;
    if (arraysize < 1) return -TErrorList.buffer_too_small;
    int depthShort = spc.getDepthShort();
    if (depthShort <= 0) return -TErrorList.server_redirection;
    int depthLong = spc.getDepthLong();
    if (storeSummary) depthLong = 1;
    stsTime = depthShort*spc.getPollngRate();
    if (depthLong <= 0)
    {
      if (hasSavedFiles) 
      { // nominally irrelevant as we'll be looking specifically in the saved area
        depthLong = 0;
      }
      else
      {
        if (rng.stoptime < t - stsTime) return -TErrorList.server_redirection;
        if (rng.starttime < t - stsTime) return -TErrorList.at_limit;
        return 0;
      }
    }
    if (rng.starttime >= t - stsTime) return 0;
    if (rng.stoptime < t  - depthLong * msecondsInMonth )
    {
      return -TErrorList.server_redirection;
    }
    short targetFmt = TDataType.getObjectFormat(dataTarget);
    if (targetFmt == TFormat.CF_NULL) return -TErrorList.illegal_format;
    
    //LocalDateTime lt = LocalDateTime.ofEpochSecond(rng.stoptime/1000, 0, ZoneOffset.UTC); 
    Calendar calendar = Calendar.getInstance();    
    calendar.setTimeInMillis(rng.stoptime); // end time
    int ye = calendar.get(Calendar.YEAR);
    int de = calendar.get(Calendar.DAY_OF_MONTH);
    int me = calendar.get(Calendar.MONTH);
    //me = lt.getMonthValue() - 1;
    calendar.setTimeInMillis(rng.starttime); // start time
    //lt = LocalDateTime.ofEpochSecond(rng.starttime/1000, 0, ZoneOffset.UTC); 
    int ys = calendar.get(Calendar.YEAR); 
    int ds = calendar.get(Calendar.DAY_OF_MONTH);
    int ms = calendar.get(Calendar.MONTH);
    //ms = lt.getMonthValue() - 1;
    int yi, di, mi; // interim time
    boolean isSnapshot = (rng.starttime == rng.stoptime);
    if (rng.startstamp > 0 && rng.startstamp == rng.stopstamp) isSnapshot = true;
    if (me == ms && de == ds)
    {
      asize = arraysize;
      finalPass = true;
    }
    else if (arraysize == 1)   
    { // just a number query
      asize = arraysize;
    }
    else                       
    { // spans multiple files
      asize = 0;
    }
    daystotal = getNumDaysInTimeSpan(rng.starttime,rng.stoptime);  
    while (recsread < arraysize)
    {
      if (asize < arraysize || arraysize == 1 || needAnotherPass)
      {
        yi = ys; mi = ms; di = ds;
        if (yi == ye && mi == me && 
            (di == de || useMonthlyHistoryFiles))
        {
          asize = arraysize;
          finalPass = true;
        }
        else  
        { // move along
          di = (di+1)%(getNumDaysInMonth(yi,mi)+1);
          if (di == 0)
          {
            di++;
            mi = (mi+1)%12;
            if (mi == 0) yi++;
          }
          // start at midnight 
          calendar.set(yi,mi,di,0,0,0);
          itime = calendar.getTimeInMillis();
          daysinmonth = 1;
          if ((daysinmonth =(int)((itime-rng.starttime)/(3600*24*1000))) < 1) daysinmonth = 1;
          if (arraysize > 1) 
          {
            if (TFormat.isSimpleFormat(targetFmt) || isSnapshot) 
            { // almost certainly a snapshot if simple format
              asize = arraysize;
              if (asize >= arraysize) finalPass = true; // just in case
            }
            else if (req.sample > 0) 
            { // specific sampling raster given
              asize = arraysize;
            }
            else 
            { // estimate the num points in the data set
              int add = (arraysize * daysinmonth)/daystotal;
              if (add < 1) add = 1;
              asize += add;
              if (asize >= arraysize) finalPass = true; // just in case
            }
          }
        }
        needAnotherPass = false;
      }
      if (asize > arraysize) asize = arraysize;
      fn = makeFileNameFromTime(rng.starttime);
      ps = loadLTSData(fn,req,rng,dataTarget,asize,recsread);
      if (ps < 0) break;
      if (req.sample > 0) asize = 0;
      if (asize == 1) recsfound += (int)ps; else recsread += (int)ps;
      needAnotherPass = (ps == 0 && isSnapshot);
      if (finalPass && !needAnotherPass) break;
      rng.starttime = itime;
      calendar.setTimeInMillis(rng.starttime);
      ys = calendar.get(Calendar.YEAR);
      ds = calendar.get(Calendar.DAY_OF_MONTH);
      ms = calendar.get(Calendar.MONTH);
    }
    rnumber = recsfound > 0 ? recsfound : recsread;
    if (arraysize == 1) recsread = recsfound;
    if (!isSnapshot && arraysize < 2) assignNumberOfPoints(dataTarget,rnumber);

    return rnumber;
  }
  private String makeFileNameFromTime(long time)
  {
    Calendar calendar = Calendar.getInstance();
    calendar.setTimeInMillis(time);
    int dd = calendar.get(Calendar.DAY_OF_MONTH);
    int yy, mm;
    // !!! unnecessary ??? -> lastArchivedDayOfTheMonth = dd;
    yy = calendar.get(Calendar.YEAR) % 100;
    mm = calendar.get(Calendar.MONTH) + 1; // January = 0
    if (useMonthlyHistoryFiles) dd = 1; 
    return getFilePrefix() +  twoDigitString(yy) + twoDigitString(mm) + 
                      twoDigitString(dd) + "." + Integer.toHexString(rcdIdx);
  }
  private int getFilePosition(RandomAccessFile raf)
  {
    int fpos = -1;
    try
    {
      if (useStandardFileSet)
      { // read the current position from the header
        raf.seek(0);
        fpos = raf.readInt();
        if (fpos < 0)
        { // file converted via mkhstfiles ?
          fpos = Swap.Int(fpos);
        }
        fpos += AS_STATE_SIZE;
      }
      else
      { // old style : SEEK_END
        fpos = (int)raf.length(); 
      }
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("THistoryRecord.getFilePosition","record ("+rcdIdx+"): "+e.toString(),TErrorList.code_failure,e,0);        
    } 
    return fpos;
  }
  private int setFilePosition(RandomAccessFile raf, int fpos)
  {
    int cc = 0;
    try
    {
      if (useStandardFileSet)
      {
        raf.seek(0);
        raf.writeInt(fpos - AS_STATE_SIZE);
        raf.writeInt((int)(System.currentTimeMillis()/1000));
      }
    }
    catch (IOException e)
    {
      cc = -1;
    }   
    return cc;
  }
  private long getFileLength(RandomAccessFile raf)
  {
    try
    {
      int fpos = 0;
      if (useStandardFileSet)
      {
        raf.seek(0);
        fpos = raf.readInt();
        if (fpos < 0)
        { // file converted via mkhstfiles ?
          fpos = Swap.Int(fpos);
        }
        return fpos;
      }
      else
      { // old-style: we're always at the end
        return raf.length();
      }
    }
    catch (IOException e) 
    { 
      return 0;
    }
  }
  private int getNumberOfRecords(RandomAccessFile raf)
  {
    int nrecs = 0;
    try
    {
      int fpos = 0, rsiz = getRecordLength();
      if (useStandardFileSet)
      {
        boolean writeback = false;
        raf.seek(0);
        fpos = raf.readInt();
        if (fpos < 0)
        { // file converted via mkhstfiles ?
          fpos = Swap.Int(fpos);
          if (fpos > 0) writeback = true;
        }
        int ts = raf.readInt(); // don't need the file time
        if (writeback) ts = Swap.Int(ts);
        int sz = raf.readInt();
        if (writeback) sz = Swap.Int(sz);
        if (rsiz != sz)
        { // consistency check failed 
          if (TLinkFactory.debugLevel > 0)
            DbgLog.log("getNumberOfRecords","index ("+rcdIdx+"): stored record size is invalid : " + rsiz + " vs " + sz);
          nrecs = -1;
          if ((fpos%rsiz) != 0) fpos = 0; // start over !
          ts = (int)(System.currentTimeMillis()/1000);
          writeback = true; // when will this work ??
        }
        if (writeback)
        {
          raf.seek(0);
          raf.writeInt(fpos);
          raf.writeInt(ts);
          raf.writeInt(rsiz);
        }
      }
      else
      { // old-style: we're always at the end
        fpos = (int)raf.length();
      }
      nrecs = fpos/rsiz;
    }
    catch (IOException e)
    {
      nrecs = -1;
    }
    return nrecs;
  }
  private double readTimeStamp(RandomAccessFile raf) throws IOException
  {
    double dt;
    @SuppressWarnings("unused")
    int sstamp, ustamp; //TODO add these in later ...
    if (!useMinimalStorage)
    {
      dt = raf.readDouble(); 
      sstamp = raf.readInt(); 
      ustamp = raf.readInt();
    }
    else
    {
      dt = (double)raf.readInt();
    }
    return dt;
  }
  private int locateRecordFromLTS(String fn,double dt,boolean isPoiFile)
  {
    int lts_offset = useStandardFileSet ? AS_STATE_SIZE : 0;
    double dt0,dt1=0,dt2=0;
    int nrecs,p0,p1 = lts_offset,p2;
    double f;
    int rlen = getRecordLength();
    if (isPoiFile)
    { // points-of-interest file is always stored dynamic and not standard
      lts_offset = 0;
    }
    try
    {
      String rafm = useStandardFileSet ? "rw" : "r";
      RandomAccessFile raf = new RandomAccessFile(fn,rafm);
      if ((nrecs = getNumberOfRecords(raf)) <= 0) nrecs = 0;
      p0 = p1 = lts_offset;
      p2 = (nrecs-1)*rlen + lts_offset;
      if (p2 < 0) p2 = 0;
      raf.seek(p0); 
      dt0 = readTimeStamp(raf);
      if (dt < dt0) p2 = 0;  // stay out of the while loop
      raf.seek(p2);
      dt2 = readTimeStamp(raf);
      while (p2 > p0 + rlen)
      {
        if (dt > (dt1=dt2)) { p1 = p2; break; }
        while (dt1 > dt0)   
        { // seek <--
          f = ((dt-dt0)/(dt1-dt0))*((double)(p2-p0)/rlen);
          if ((p1=p0+rlen*(int)f) >= p2) break;
          raf.seek(p1);
          dt1 = readTimeStamp(raf);
          if (dt1 < dt) break;
          p2 = p1; dt2 = dt1;
        }
        dt0 = dt1; p0 = p1;
        while (dt2 > dt1) 
        {  // seek -->
          f = ((dt2-dt)/(dt2-dt1))*((double)(p2-p0)/rlen);
          if ((p1=p2-rlen*(int)f) <= p0) break;
          raf.seek(p1);
          dt1 = readTimeStamp(raf);
          if (dt1 > dt) break;
          p0 = p1; dt0 = dt1;
        }
        p2 = p1; dt2 = dt1;
      }
      raf.close();
    }
    catch (Exception e)
    {
      DbgLog.log("locateRecordFromLTS",fn+"(rec len "+rlen+", dt "+dt2+" - "+dt1+", pos "+p1+") : "+e.toString());
      return -TErrorList.file_error;
    }
    return p1;
  }
  
  private int getFormattedHistoryDataPoint(Object dataTarget,int number,Object dataSource,int index,double tstmp,int sstmp,int ustmp)
  {
  	return getFormattedHistoryDataPoint(dataTarget,number,dataSource,index,-1,tstmp,sstmp,ustmp);
  }
  private int getFormattedHistoryDataPoint(Object dataTarget,int number,Object dataSource,int index,int fieldIndex, double tstmp,int sstmp,int ustmp)
  {
  	if (dataSource instanceof double[])
  	{
  		double[] v = (double[])dataSource;
      return getFormattedHistoryDataPoint(dataTarget,number,v[index],tstmp,sstmp,ustmp);
  	}
  	if (dataSource instanceof NAME[])
  	{
  		NAME[] v = (NAME[])dataSource;
      return getFormattedHistoryDataPoint(dataTarget,number,v[index].getName(),tstmp,sstmp,ustmp);
  	}
  	if (dataSource instanceof TCompoundDataObject[])
  	{
  		TCompoundDataObject[] v = (TCompoundDataObject[])dataSource;
  		if (fieldIndex < 0)
  		  return getFormattedHistoryDataPoint(dataTarget,number,v[index],tstmp,sstmp,ustmp);
  		else
  		  return getFormattedHistoryDataPoint(dataTarget,number,v[index],fieldIndex,tstmp,sstmp,ustmp);
  	}
  	if (dataSource instanceof TTaggedStructure[])
  	{
  		TTaggedStructure[] v = (TTaggedStructure[])dataSource;
  		if (fieldIndex < 0)
    		return getFormattedHistoryDataPoint(dataTarget,number,v[index],tstmp,sstmp,ustmp);
  		else
    		return getFormattedHistoryDataPoint(dataTarget,number,v[index],fieldIndex,tstmp,sstmp,ustmp);
  	}
    return TErrorList.illegal_format;
  }
  @SuppressWarnings("unused")
  private int getFormattedHistoryDataPoint(Object dataTarget,int number,double[] v,NAME[] n,TCompoundDataObject[] cdo,int index,double tstmp,int sstmp,int ustmp)
  {
    if (v != null) return getFormattedHistoryDataPoint(dataTarget,number,v[index],tstmp,sstmp,ustmp);
    if (n != null) return getFormattedHistoryDataPoint(dataTarget,number,n[index].getName(),tstmp,sstmp,ustmp);
    if (cdo != null) return getFormattedHistoryDataPoint(dataTarget,number,cdo[index],tstmp,sstmp,ustmp);
    return TErrorList.illegal_format;
  }
  private int getFormattedHistoryDataPoint(Object dataTarget,int number,double v,double tstmp,int sstmp,int ustmp)
  {
    short fmt;
    if (dataTarget instanceof HISTORY[])
    {
      fmt = TFormat.CF_HISTORY;
    }
    else
    {
      fmt = TDataType.getObjectFormat(dataTarget);
    }
    switch (fmt)
    {
      case TFormat.CF_HISTORY:
        switch (((HISTORY[])dataTarget)[0].getCarriedFormat())
        {
          case TFormat.CF_FLOAT:
            ((HISTORY[])dataTarget)[number] = new HISTORY(tstmp,(float)v,sstmp,ustmp);
            break;
          case TFormat.CF_DOUBLE:
            ((HISTORY[])dataTarget)[number] = new HISTORY(tstmp,v,sstmp,ustmp);
            break;
          case TFormat.CF_INT32:
            ((HISTORY[])dataTarget)[number] = new HISTORY(tstmp,(int)v,sstmp,ustmp);
            break;
          case TFormat.CF_INT64:
            ((HISTORY[])dataTarget)[number] = new HISTORY(tstmp,(long)v,sstmp,ustmp);
            break;
          default: 
            return -TErrorList.illegal_format;
        }        
        break;
      case TFormat.CF_FLOAT:
        ((float[])dataTarget)[number] = (float)v;
        break;
      case TFormat.CF_FLTINT:
        if (((FLTINT[])dataTarget)[number] == null) ((FLTINT[])dataTarget)[number] = new FLTINT();
        ((FLTINT[])dataTarget)[number].ival = (int)tstmp;
        ((FLTINT[])dataTarget)[number].fval = (float)v;
        break;
      case TFormat.CF_INTFLTINT:            
        if (((INTFLTINT[])dataTarget)[number] == null) ((INTFLTINT[])dataTarget)[number] = new INTFLTINT();
        ((INTFLTINT[])dataTarget)[number].i1val = (int)tstmp;
        ((INTFLTINT[])dataTarget)[number].fval = (float)v;
        break;
      case TFormat.CF_INT32:
        ((int[])dataTarget)[number] = (int)v;
        break;
      case TFormat.CF_INT64:
        ((long[])dataTarget)[number] = (long)v;
        break;            
      case TFormat.CF_INTINT:            
        if (((INTINT[])dataTarget)[number] == null) ((INTINT[])dataTarget)[number] = new INTINT();
        ((INTINT[])dataTarget)[number].i2val = (int)tstmp;
        ((INTINT[])dataTarget)[number].i1val = (int)v;
        break;
      case TFormat.CF_DOUBLE:
        ((double[])dataTarget)[number] = (double)v;
        break;
      case TFormat.CF_DBLTIME:
        if (((DBLTIME[])dataTarget)[number] == null) ((DBLTIME[])dataTarget)[number] = new DBLTIME();
        ((DBLTIME[])dataTarget)[number].d2val = tstmp;
        ((DBLTIME[])dataTarget)[number].d1val = v;
        break;
      case TFormat.CF_DBLDBL:            
        if (((DBLDBL[])dataTarget)[number] == null) ((DBLDBL[])dataTarget)[number] = new DBLDBL();
        ((DBLDBL[])dataTarget)[number].d2val = tstmp;
        ((DBLDBL[])dataTarget)[number].d1val = v;
        break;
      case TFormat.CF_DBLDBLDBL:
        if (((DBLDBLDBL[])dataTarget)[number] == null) ((DBLDBLDBL[])dataTarget)[number] = new DBLDBLDBL();
        ((DBLDBLDBL[])dataTarget)[number].d3val = sstmp;
        ((DBLDBLDBL[])dataTarget)[number].d2val = tstmp;
        ((DBLDBLDBL[])dataTarget)[number].d1val = v;
        break;
      case TFormat.CF_FLTINTINT:
        if (((FLTINTINT[])dataTarget)[number] == null) ((FLTINTINT[])dataTarget)[number] = new FLTINTINT();
        ((FLTINTINT[])dataTarget)[number].i2val = sstmp;
        ((FLTINTINT[])dataTarget)[number].i1val = (int)tstmp;
        ((FLTINTINT[])dataTarget)[number].f1val = (float)v;
        break;
      case TFormat.CF_NAME64DBLDBL:
      case TFormat.CF_NAME64DBLDBLDBL:
        break;
      default:
        return -TErrorList.illegal_format;
    }
    return 0;
  }
  private int getFormattedHistoryDataPoint(Object dataTarget,int number,String n,double tstmp,int sstmp,int ustmp)
  {
    switch (TDataType.getObjectFormat(dataTarget))
    {
      case TFormat.CF_HISTORY:
        switch (((HISTORY[])dataTarget)[0].getCarriedFormat())
        {
          case TFormat.CF_NAME16:
            NAME16 n16 = new NAME16(n);
            ((HISTORY[])dataTarget)[number] = new HISTORY(tstmp,n16,sstmp,ustmp);
            break;
          case TFormat.CF_NAME32:
            NAME32 n32 = new NAME32(n);
            ((HISTORY[])dataTarget)[number] = new HISTORY(tstmp,n32,sstmp,ustmp);
            break;
          case TFormat.CF_NAME64:
            NAME64 n64 = new NAME64(n);
            ((HISTORY[])dataTarget)[number] = new HISTORY(tstmp,n64,sstmp,ustmp);
            break;
          default:
            return -TErrorList.illegal_format;
        }
        break;
    case TFormat.CF_NAME16I:
        if (((NAME16I[])dataTarget)[number] == null) ((NAME16I[])dataTarget)[number] = new NAME16I();
        ((NAME16I[])dataTarget)[number].ival = (int)tstmp;
        ((NAME16I[])dataTarget)[number].name = n;
        break;
      case TFormat.CF_NAME32I:
        if (((NAME32I[])dataTarget)[number] == null) ((NAME32I[])dataTarget)[number] = new NAME32I();
        ((NAME32I[])dataTarget)[number].ival = (int)tstmp;
        ((NAME32I[])dataTarget)[number].name = n;
        break;
      case TFormat.CF_NAME64I:
        if (((NAME64I[])dataTarget)[number] == null) ((NAME64I[])dataTarget)[number] = new NAME64I();
        ((NAME64I[])dataTarget)[number].ival = (int)tstmp;
        ((NAME64I[])dataTarget)[number].name = n;
        break;
      case TFormat.CF_NAME64DBL:
        if (((NAME64DBL[])dataTarget)[number] == null) ((NAME64DBL[])dataTarget)[number] = new NAME64DBL();
        ((NAME64DBL[])dataTarget)[number].dval = tstmp;
        ((NAME64DBL[])dataTarget)[number].name = n;
        break;
      case TFormat.CF_NAME64TIME:
        if (((NAME64TIME[])dataTarget)[number] == null) ((NAME64TIME[])dataTarget)[number] = new NAME64TIME();
        ((NAME64TIME[])dataTarget)[number].dval = tstmp;
        ((NAME64TIME[])dataTarget)[number].name = n;
        break;
      case TFormat.CF_NAME64DBLDBL:
        if (((NAME64DBLDBL[])dataTarget)[number] == null) ((NAME64DBLDBL[])dataTarget)[number] = new NAME64DBLDBL();
        ((NAME64DBLDBL[])dataTarget)[number].d2val = sstmp;
        ((NAME64DBLDBL[])dataTarget)[number].d1val = tstmp;
        ((NAME64DBLDBL[])dataTarget)[number].name = n;
        break;
      case TFormat.CF_NAME64DBLDBLDBL:
        break;
      default:
        return -TErrorList.illegal_format;
    }
    return 0;
  }
  private int getFormattedHistoryDataPoint(Object dataTarget,int number,TTaggedStructure tts,double tstmp,int sstmp,int ustmp)
  {
    if (tts == null) return -TErrorList.illegal_format;
    short fmt = TDataType.getObjectFormat(dataTarget);
    switch (fmt)
    {
      case TFormat.CF_HISTORY:
        short dfmt = ((HISTORY[])dataTarget)[0].getCarriedFormat();
        if (dfmt != TFormat.CF_STRUCT) return -TErrorList.illegal_format;
        String tag = ((HISTORY[])dataTarget)[0].getTag();
        if (tts.getName().compareTo(tag) != 0) return -TErrorList.invalid_structure_tag;        
        ((HISTORY[])dataTarget)[number] = new HISTORY(tstmp,tts,sstmp,ustmp);
        break;
      default:
      	return -TErrorList.not_applicable;
    }  	
    return 0;
  }
  /**
   * with fieldIndex this call is only applicable for dataTarget receiving a number or a name 
   * 
   * @param dataTarget
   * @param number
   * @param tts
   * @param fieldIndex
   * @param tstmp
   * @param sstmp
   * @param ustmp
   * @return
   */
  private int getFormattedHistoryDataPoint(Object dataTarget,int number,TTaggedStructure tts,int fieldIndex,double tstmp,int sstmp,int ustmp)
  {
    if (tts == null) return -TErrorList.illegal_format;
    short[] ffmt = new short[1];
    int[] foffset = new int[1];
    int cc = TFormat.getFieldInFormat(TFormat.CF_STRUCT, null, tts.getName(), fieldIndex, foffset, ffmt);
    if (cc != 0) return -cc;
    double dv = 0;
    NAME64 n = null;
    TStructDescription tsd = new TStructDescription(tts.getName());
    int fsiz = TFormat.formatSizeOf(ffmt[0]);
    byte[] sbytes = new byte[fsiz];
    byte[] nbytes = new byte[64];
    Field fld = tsd.getField(fieldIndex);
    int fldlen = 0;
    try
    {
    	System.arraycopy(tts.getByteArray(), foffset[0], sbytes, 0, fsiz);
    	Swap.Bytes(sbytes, fsiz);
	    ByteArrayInputStream bais = new ByteArrayInputStream(sbytes);
	    DataInputStream dis = new DataInputStream(bais);
	    switch (ffmt[0])
	    {
	    	case TFormat.CF_INT16: dv = dis.readShort(); break;
	    	case TFormat.CF_INT32: dv = dis.readInt(); break;
	    	case TFormat.CF_INT64: dv = (double)dis.readLong(); break;
	    	case TFormat.CF_FLOAT: dv = dis.readFloat(); break;
	    	case TFormat.CF_DOUBLE:dv = dis.readDouble(); break; 
	    	case TFormat.CF_CHAR:
	    		fldlen = Math.min(fld.getArraySize(), 64);
	    	case TFormat.CF_NAME8:
	    	case TFormat.CF_NAME16:
	    	case TFormat.CF_NAME32:
	    	case TFormat.CF_NAME48:
	    	case TFormat.CF_NAME64:
	    		if (fldlen == 0) fldlen = TFormat.formatSizeOf(ffmt[0]);
	    		dis.read(nbytes, 0, fldlen);
	    		n = new NAME64(new String(nbytes));
	    		break;
	      default:
	      	//TODO: what to do about structs that have complex fields?
	    		break;
	    }
    } 
    catch (Exception x) 
    {
    	return TErrorList.code_failure;
    }
  	if (n == null)
  	{
      return getFormattedHistoryDataPoint(dataTarget,number,dv,tstmp,sstmp,ustmp);
  	}
  	else
  	{
      return getFormattedHistoryDataPoint(dataTarget,number,n.getName(),tstmp,sstmp,ustmp);
  	}
  }
  /**
   * with fieldIndex this call is only applicable for dataTarget receiving a number or a name 
   * 
   * @param dataTarget
   * @param number
   * @param cdo
   * @param fieldIndex
   * @param tstmp
   * @param sstmp
   * @param ustmp
   * @return
   */
  private int getFormattedHistoryDataPoint(Object dataTarget,int number,TCompoundDataObject cdo,int fieldIndex,double tstmp,int sstmp,int ustmp)
  {
    if (cdo == null) return -TErrorList.illegal_format;
    short[] ffmt = new short[1];
    int[] foffset = new int[1];
    int cc = TFormat.getFieldInFormat(fmt, null, null, fieldIndex, foffset, ffmt);
    if (cc != 0) return -cc;
    double dv = 0;
    NAME64 n = null;
    int fsiz = TFormat.formatSizeOf(ffmt[0]);
    byte[] sbytes = new byte[fsiz];
    byte[] nbytes = new byte[64];
    try
    {
    	System.arraycopy(cdo.getByteArray(), foffset[0], sbytes, 0, fsiz);
    	Swap.Bytes(sbytes, fsiz);
	    ByteArrayInputStream bais = new ByteArrayInputStream(sbytes);
	    DataInputStream dis = new DataInputStream(bais);
	    switch (ffmt[0])
	    {
	    	case TFormat.CF_INT16: dv = dis.readShort(); break;
	    	case TFormat.CF_INT32: dv = dis.readInt(); break;
	    	case TFormat.CF_INT64: dv = (double)dis.readLong(); break;
	    	case TFormat.CF_FLOAT: dv = dis.readFloat(); break;
	    	case TFormat.CF_DOUBLE:dv = dis.readDouble(); break; 
	    	case TFormat.CF_NAME8:
	    	case TFormat.CF_NAME16:
	    	case TFormat.CF_NAME32:
	    	case TFormat.CF_NAME48:
	    	case TFormat.CF_NAME64:
	    		int fldlen = TFormat.formatSizeOf(ffmt[0]);
	    		dis.read(nbytes, 0, fldlen);
	    		n = new NAME64(new String(nbytes));
	    		break;
	      default:
	    		break;
	    }
    } 
    catch (Exception x) 
    {
    	return TErrorList.code_failure;
    }
  	if (n == null)
  	{
      return getFormattedHistoryDataPoint(dataTarget,number,dv,tstmp,sstmp,ustmp);
  	}
  	else
  	{
      return getFormattedHistoryDataPoint(dataTarget,number,n.getName(),tstmp,sstmp,ustmp);
  	}    
  }
  private int getFormattedHistoryDataPoint(Object dataTarget,int number,TCompoundDataObject cdo,double tstmp,int sstmp,int ustmp)
  {
    float f=0; double d1=0; String n = "";
    short fmt = TDataType.getObjectFormat(dataTarget);
    if (cdo == null) return -TErrorList.illegal_format;
    switch (fmt)
    {
      case TFormat.CF_HISTORY:
        short sfmt = cdo.getFormat();
        short dfmt = ((HISTORY[])dataTarget)[0].getCarriedFormat();
        if (sfmt != dfmt) return -TErrorList.illegal_format;
        ((HISTORY[])dataTarget)[number] = new HISTORY(tstmp,cdo.clone(),sstmp,ustmp);
        break;
      case TFormat.CF_DBLTIME:
        d1 = cdo.getDoubleValue();
        if (((DBLTIME[])dataTarget)[number] == null) ((DBLTIME[])dataTarget)[number] = new DBLTIME();
        ((DBLTIME[])dataTarget)[number].setValues(d1, tstmp);
        break;
      case TFormat.CF_DBLDBL:
        d1 = cdo.getDoubleValue();
        if (((DBLDBL[])dataTarget)[number] == null) ((DBLDBL[])dataTarget)[number] = new DBLDBL();
        ((DBLDBL[])dataTarget)[number].setValues(d1, tstmp);
        break;
      case TFormat.CF_FLTINT:
        f = (float)cdo.getDoubleValue();
        if (((FLTINT[])dataTarget)[number] == null) ((FLTINT[])dataTarget)[number] = new FLTINT();
        ((FLTINT[])dataTarget)[number].setValues(f, (int)tstmp);
        break;
      case TFormat.CF_INTFLTINT:            
        f = (float)cdo.getDoubleValue();
        int sts = (int)cdo.getIntegerValue();
        sts &= 0x0fff;
        if (((INTFLTINT[])dataTarget)[number] == null) ((INTFLTINT[])dataTarget)[number] = new INTFLTINT();
        ((INTFLTINT[])dataTarget)[number].i1val = (int)tstmp;
        ((INTFLTINT[])dataTarget)[number].fval = f;
        ((INTFLTINT[])dataTarget)[number].i2val = sts;        
        break;
      case TFormat.CF_FLTINTINT:
        f = (float)cdo.getDoubleValue();
        if (((FLTINTINT[])dataTarget)[number] == null) ((FLTINTINT[])dataTarget)[number] = new FLTINTINT();
        ((FLTINTINT[])dataTarget)[number].setValues(f, (int)tstmp, sstmp);
        break;
      case TFormat.CF_DBLDBLDBL:
        d1 = cdo.getDoubleValue();
        if (((DBLDBLDBL[])dataTarget)[number] == null) ((DBLDBLDBL[])dataTarget)[number] = new DBLDBLDBL();
        ((DBLDBLDBL[])dataTarget)[number].setValues(d1, tstmp, sstmp);
        break;
      case TFormat.CF_NAME64TIME:
      case TFormat.CF_NAME64DBL:
      case TFormat.CF_NAME64DBLDBL:
        if (cdo instanceof NAME)
        {
          NAME d = ((NAME)cdo);
          n =  d.name; 
        }
        else if (cdo instanceof NAMEINT)
        {
          NAMEINT d = ((NAMEINT)cdo);
          n =  d.name; 
        }
        else if (cdo instanceof NAME64DBL)
        {
          NAME64DBL d = ((NAME64DBL)cdo);
          n = d.name; 
        }
        else if (cdo instanceof USTRING)
        {
          USTRING d = ((USTRING)cdo);
          n = d.str; 
        }
        if (fmt == TFormat.CF_NAME64DBLDBL)
        {
          if (((NAME64DBLDBL[])dataTarget)[number] == null) ((NAME64DBLDBL[])dataTarget)[number] = new NAME64DBLDBL();
          ((NAME64DBLDBL[])dataTarget)[number].setValues(n, tstmp, sstmp);
        }
        else if (fmt == TFormat.CF_NAME64TIME)
        {
          if (((NAME64TIME[])dataTarget)[number] == null) ((NAME64TIME[])dataTarget)[number] = new NAME64TIME();
          ((NAME64TIME[])dataTarget)[number].setValues(n, tstmp);
        }
        else
        {
          if (((NAME64DBL[])dataTarget)[number] == null) ((NAME64DBL[])dataTarget)[number] = new NAME64DBL();
          ((NAME64DBL[])dataTarget)[number].setValues(n, tstmp);        
        }
        break;  
      default:
        return -TErrorList.not_implemented;
    }
    return 0;
  }
  private int getFormattedHistoryDataSet(Object dataTarget,Object dataSource,double tstmp,int sstmp,int ustmp)
  {
  	if (dataSource instanceof double[])
  	{
      return getFormattedHistoryDataSet(dataTarget,(double[])dataSource,tstmp,sstmp,ustmp);
  	}
  	if (dataSource instanceof NAME[])
  	{
      return getFormattedHistoryDataSet(dataTarget,(NAME[])dataSource,tstmp,sstmp,ustmp);
  	}
  	if (dataSource instanceof TCompoundDataObject[])
  	{
  		return getFormattedHistoryDataSet(dataTarget,(TCompoundDataObject[])dataSource,tstmp,sstmp,ustmp);
  	}
  	if (dataSource instanceof TTaggedStructure[])
  	{
  		return getFormattedHistoryDataSet(dataTarget,(TTaggedStructure[])dataSource,tstmp,sstmp,ustmp);
  	}
    return TErrorList.illegal_format;
  }
  @SuppressWarnings("unused")
  private int getFormattedHistoryDataSet(Object dataTarget,double[] v,NAME[] n,TCompoundDataObject[] cdo,double tstmp,int sstmp,int ustmp)
  {
    if (v != null) return getFormattedHistoryDataSet(dataTarget,v,tstmp,sstmp,ustmp);
    if (n != null) return getFormattedHistoryDataSet(dataTarget,n,tstmp,sstmp,ustmp);
    return getFormattedHistoryDataSet(dataTarget,cdo,tstmp,sstmp,ustmp);
  }
  private int getFormattedHistoryDataSet(Object dataTarget,double[] v,double tstmp,int sstmp,int ustmp)
  {
    int number = TDataType.getObjectSize(dataTarget);
    if (number > v.length) number = v.length;
    lastReadTimestamp = (long)(tstmp * 1000);
    lastReadSystemStamp = sstmp;
    lastReadUserStamp = ustmp;
    switch (TDataType.getObjectFormat(dataTarget))
    {
      case TFormat.CF_FLOAT:
        for (int i=0; i<number; i++) ((float[])dataTarget)[i] = (float)v[i];
        return number;          
      case TFormat.CF_INT32:
        for (int i=0; i<number; i++) ((int[])dataTarget)[i] = (int)v[i];
        return number;          
      case TFormat.CF_INT64:
        for (int i=0; i<number; i++) ((long[])dataTarget)[i] = (long)v[i];
        return number;          
      case TFormat.CF_DOUBLE:
        for (int i=0; i<number; i++) ((double[])dataTarget)[i] = (double)v[i];
        return number;      
    }
    return 0;
  }
  private int getFormattedHistoryDataSet(Object dataTarget,NAME[] n,double tstmp,int sstmp,int ustmp)
  {
    int number = TDataType.getObjectSize(dataTarget);
    if (number > n.length) number = n.length;
    if (TDataType.getObjectFormat(dataTarget) != TDataType.getObjectFormat(n))
      return -TErrorList.illegal_format;
    lastReadTimestamp = (long)(tstmp * 1000);
    lastReadSystemStamp = sstmp;
    lastReadUserStamp = ustmp;
    for (int i=0; i<number; i++) ((NAME[])dataTarget)[i] = n[i];
    return number;
  }
  private double[] getPrimaryNumberArray(TCompoundDataObject[] cdo)
  {
  	double[] v = new double[cdo.length];
  	for (int i=0; i<v.length; i++) v[i] = cdo[i].getDoubleValue();
  	return v;
  }
  private NAME[] getPrimaryNameArray(TCompoundDataObject[] cdo)
  {
  	NAME64[] v = new NAME64[cdo.length];
  	for (int i=0; i<v.length; i++) v[i] = new NAME64(((NAME)cdo[i]).getName());
  	return v;
  }
  private int getFormattedHistoryDataSet(Object dataTarget,TTaggedStructure[] tts,double tstmp,int sstmp,int ustmp)
  {
    int number = TDataType.getObjectSize(dataTarget);
    if (number > tts.length) number = tts.length;
    if (!(dataTarget instanceof TTaggedStructure[])) return -TErrorList.illegal_format;
    lastReadTimestamp = (long)(tstmp * 1000);
    lastReadSystemStamp = sstmp;
    lastReadUserStamp = ustmp;
    TTaggedStructure[] ttsTarget = (TTaggedStructure[])dataTarget;
    if (ttsTarget[0].getName().compareTo(tts[0].getName()) != 0)
    	return -TErrorList.invalid_structure_tag;
    for (int i=0; i<number; i++) ttsTarget[i] = tts[i];
    return number;
  }
  private int getFormattedHistoryDataSet(Object dataTarget,TCompoundDataObject[] cdo,double tstmp,int sstmp,int ustmp)
  {
    int number = TDataType.getObjectSize(dataTarget);
    if (number > cdo.length) number = cdo.length;
    short fcdo = TDataType.getObjectFormat(cdo);
    short ft = TDataType.getObjectFormat(dataTarget);
    if (ft == fcdo)
    { // same type requested (easy one!)
      for (int i=0; i<number; i++) ((TCompoundDataObject[])dataTarget)[i] = cdo[i];
      return number;
    }
    else
    {
    	if (TFormat.isNumberFormat(ft)) 
    	{ 
    		double[] dataSource = getPrimaryNumberArray(cdo);
    		return getFormattedHistoryDataSet(dataTarget,dataSource,tstmp,sstmp,ustmp);
    	}
      if (TFormat.isName(ft) && TFormat.isNameInt(fcdo))
      {
    		NAME[] dataSource = getPrimaryNameArray(cdo);
    		return getFormattedHistoryDataSet(dataTarget,dataSource,tstmp,sstmp,ustmp);             	
      }
    }
    return -TErrorList.illegal_format;
  }
  private static boolean debug = false;
  public static void setDebug(boolean value) { debug = value; }
  public static void setDebug(String value) 
  { 
  	if (value == null) return;
  	debug = TCommandItem.getBoolean(value);
  	getDebugString();
  }
  public static void setDebug(int value) { debug = (value != 0); }
  public static boolean getDebug() { return debug; }
  public static String getDebugString() 
  {
  	String s = debug ? "true" : "false";
  	DbgLog.log("THistoryRecord", "debug currently set to "+s);
  	return s; 
  }
  private int loadLTSData(String fileName,THistoryRequest req,THistoryRange rng,
      Object dataTarget,int arraysize,int arrayoffset)
  {
    int depth = hspec.getDepthLong();
    if (storeSummary) depth = 1;
    if (arraysize < 1) return 0;
    if (depth <= 0) return 0;
    long tnow = TDataTime.now();
    if ((tnow - rng.starttime) < hspec.getDepthShort()) return 0;
    if (req.index < 0 || req.index >= getArraySize()) return 0;
    double dstarttime = (double)(rng.starttime/1000.0);
    double dendtime = (double)(rng.stoptime/1000.0);
    double dtnow = (double)(tnow)/1000.0;
    double datatime = 0.0;
    double dlaststamp = -1;
    long datasys = rng.startstamp;
    int n, nrecs,raster = 1;
    int timebin = 1,pstart,pstop,number = 0;
    int sstamp = 0, ustamp = 0;
    long usstamp = 0;
    boolean isSnapshot = false, useSysStamp = false;
    boolean hasPointsOfInterest = false, isPointsOfInterestFile = false;
    String poiName = null;
    long poiStart,poiStop;
    int dindex = req.index;
    int rlen = getRecordLength(); // include time/data stamps 'prefix' length
    int hlen = getRecordPrefixLength(); 
    if (fileName.startsWith("pi") == false)
    { // isn't already a points-of-interest file 
      poiName = fileName.replace(fileName.substring(0, 2), "pi");
      File poif = new File(arcDbPath +  File.separator + poiName);
      if (!poif.exists())
      { // not there : try the 'static' area
        poif = new File(arcDbPathSaved +  File.separator + poiName);
      }
      if (poif.exists())
      { // there IS a companion points-of-interest file
        if (poif.length() < rlen * 100)
        { // only turn this on if the points-of-interest file doesn't appear to contain a lot of garbage
          hasPointsOfInterest = true;
        }
        else
        {
          MsgLog.log("loadLTSData","ignoring points-of-interest file (too large)",TErrorList.not_allowed,null,1);
        }
      }
    }
    else
    {
      isPointsOfInterestFile = true;
    }
    String fn = arcDbPath +  File.separator + fileName;
    if (!(new File(fn).exists()))
    { // not there : try the 'static' area
      fn = arcDbPathSaved +  File.separator + fileName;
      if (!(new File(fn).exists())) return 0;
    }
    if (dendtime < dstarttime) dendtime = dtnow;
    if (debug) DbgLog.log("loadLTSData", "find start/stop locations for "+fn);
    if ((pstart=locateRecordFromLTS(fn,dstarttime,isPointsOfInterestFile)) < 0) return 0;
    if ((pstop=locateRecordFromLTS(fn,dendtime,isPointsOfInterestFile)) < 0) return 0;
    if (debug) DbgLog.log("loadLTSData", "found start @ "+pstart+", stop @ "+pstop);
    isSnapshot = (dendtime-dstarttime) < 0.01;
    if (rng.startstamp > 0 && rng.startstamp == rng.stopstamp) isSnapshot = true;
    if (rng.startstamp > 0 && rng.stopstamp >= rng.startstamp && !useMinimalStorage) useSysStamp = true;
    if (rlen <= 0) return 0; // ???
    nrecs = (int)((pstop-pstart)/rlen);
    if (arraysize == 1 && !isSnapshot) 
    {
      if (isPointsOfInterestFile) return 0;
      lastReadTimestamp = rng.stoptime;
      return nrecs > 0 ? nrecs : 0;
    }
    arraysize -= arrayoffset;  // subtract off the offset
    if (arraysize <= 0) return 0; // ???
    if (arraysize < nrecs) raster += nrecs/(arraysize+1);
    if (raster > 1 && isPointsOfInterestFile) return 0;
    if (nrecs == 0 && isPointsOfInterestFile) return 0;
    if (req.sample > 0) raster = req.sample;
    short targetFmt = TDataType.getObjectFormat(dataTarget);
    if (targetFmt == TFormat.CF_NULL) return 0;
    if (isSnapshot)
    {
    	req.index = 0;
    	raster = 1;
      if (!useSysStamp) datatime = dstarttime - timebin;
    }
    else if (nrecs == 0)
    {
      return 0;
    }
    if (raster == 1) hasPointsOfInterest = false; // no need to check
    try 
    {
      long nxtpos = 0;
      boolean advance = false;
      int cc;
      double dts = 0.0;
      short sourceFmt = getFmt();
      String sourceTag = getTag();
      double[] dv = null;
      NAME[] nv = null;
      TCompoundDataObject[] cdo = null;
      TTaggedStructure[] tts = null;
      Object ov = null;
      int nsiz = TFormat.formatSizeOf(sourceFmt), dsiz;
      byte[] b;
      RandomAccessFile raf = new RandomAccessFile(fn,"r");
      long flen = getFileLength(raf); //raf.length();
      raf.seek(pstart);
      
      while (raf.getFilePointer() < flen) 
      { // read through file
        if (!useMinimalStorage)
        {
          dts = raf.readDouble();
          sstamp = raf.readInt();
          ustamp = raf.readInt();
        }
        else
        {
          dts = (double)raf.readInt();          
        }
        usstamp = UnsignedUtils.toUnsignedInt(sstamp);
        if (debug) DbgLog.log("loadLTSData","examine data @ time "+dts);
        if (hasPointsOfInterest && dlaststamp > 0 && dts > dlaststamp-2)
        {
          poiStart = (long)(dlaststamp*1000.0) + 500;
          poiStop = (long)(dts*1000.0) - 500;
          int poisize = arraysize < 2 ? arraysize : 2;
          THistoryRequest poireq = new THistoryRequest(req.index,req.fieldindex,1);
        	THistoryRange poirng = new THistoryRange(poiStart,poiStop,rng.startstamp,rng.stopstamp,rng.lwrLimit,rng.uprLimit,rng.userstamp);          
          n = loadLTSData(poiName,poireq,poirng,dataTarget,poisize,arrayoffset+number);
          if (n >= arraysize - number) break;
          if (n > 0) number += n;
        }
        nxtpos = raf.getFilePointer() + rlen*raster - hlen;
        if (isSnapshot)
        {
          if (dts > dendtime) dendtime = dts;
          if (usstamp > rng.stopstamp) rng.stopstamp = usstamp;
          if (useSysStamp && usstamp > 0 && usstamp < rng.startstamp)
          { // targeting a system/user stamp in a snapshot and this isn't it!
            raf.seek(nxtpos);
            continue;            
          }
        }
        if (dts > dendtime) 
        {
          if (debug) DbgLog.log("loadLTSData", "timestamp > "+dendtime+"; return "+number+" records");
          return number;
        }
        if (useSysStamp && usstamp > rng.stopstamp) return number;
        if ((useSysStamp && usstamp < datasys) ||
            dts < datatime + (double)timebin/1000.0)
        { // 'too old', skip over this data set
          raf.seek(nxtpos);
          continue;
        }
        datatime = dts;
        datasys = sstamp;
        dsiz = siz;  // stored data size for this record
        if (storeSummary)
        {
          sourceFmt = TFormat.CF_DOUBLE;
          dsiz = 6;
          if (req.index > 3) return 0;
        }
        // read the data
        boolean needToSwapBytes = false;
        byte[] bb = new byte[8];
        if (!isSnapshot && !storeSummary)
        { // for non-snapshot : only have to fill in the 'index' in the array
          int off = req.index * TFormat.formatSizeOf(fmt);
          if (targetFmt != TFormat.CF_HISTORY && !TFormat.isSimpleFormat(fmt))
          { // take care of any input field index here 
          	int[] foff = new int[1];
          	short[] ffmt = new short[1];
          	if ((cc=TFormat.getFieldInFormat(fmt,null,sourceTag,req.fieldindex,foff,ffmt)) < 0)
          		return -cc;
            off += foff[0];
            sourceFmt = ffmt[0];
            sourceTag = null;
            needToSwapBytes = true; // it won't be stored in java big-endian format !
          }
          raf.seek(raf.getFilePointer()+off);
          dsiz = 1;
          dindex = 0;
          advance = true;
        }
        switch (sourceFmt)
        {
          case TFormat.CF_DOUBLE:
            if (dv == null) dv = new double[dsiz];
            for (int i=0; i<dsiz; i++) 
            {
            	if (needToSwapBytes)
            	{
            		raf.read(bb, 0, 8);
            		dv[i] = Swap.Double(bb);
            	}
            	else
            	{
            		dv[i] = raf.readDouble();
            	}
            }
            ov = dv;
            break;
          case TFormat.CF_FLOAT:
            if (dv == null) dv = new double[dsiz];
            for (int i=0; i<dsiz; i++) 
            {
            	if (needToSwapBytes)
            	{
            		raf.read(bb, 0, 4);
            		dv[i] = Swap.Float(bb);
            	}
            	else
            	{
            		dv[i] = raf.readFloat();
            	}
            }
            ov = dv;
            break;
          case TFormat.CF_INT64:
            if (dv == null) dv = new double[dsiz];
            for (int i=0; i<dsiz; i++) 
            {
            	if (needToSwapBytes)
            	{
            		raf.read(bb, 0, 8);
            		dv[i] = Swap.Long(bb);
            	}
            	else
            	{
            		dv[i] = raf.readLong();
            	}
            }
            ov = dv;
            break;            
          case TFormat.CF_INT32:
            if (dv == null) dv = new double[dsiz];
            for (int i=0; i<dsiz; i++) 
            {
            	if (needToSwapBytes)
            	{
            		raf.read(bb, 0, 4);
            		dv[i] = Swap.Int(bb);
            	}
            	else
            	{
            		dv[i] = raf.readInt();
            	}
            }
            ov = dv;
            break;
          case TFormat.CF_INT16:
            if (dv == null) dv = new double[dsiz];
            short sval;
            for (int i=0; i<dsiz; i++) 
            {
            	sval = raf.readShort();
            	dv[i] = needToSwapBytes ? Swap.Short(sval) : sval;
            }
            ov = dv;
            break;
          case TFormat.CF_BYTE:
            if (dv == null) dv = new double[dsiz];
            for (int i=0; i<dsiz; i++) dv[i] = raf.readByte();
            ov = dv;
            break;
          case TFormat.CF_NAME16:
          case TFormat.CF_NAME32:
          case TFormat.CF_NAME64:
            if (nv == null) nv = new NAME[dsiz];
            b = new byte[nsiz];
            for (int i=0; i<dsiz; i++) 
            {
              nv[i] = new NAME(nsiz);
              raf.read(b,0,nsiz);
              nv[i].setName(b);
            }
            ov = nv;
            break;
          case TFormat.CF_STRUCT:
            if (tts == null) tts = new TTaggedStructure[dsiz];
            nsiz = TStructRegistry.getSizeInBytes(sourceTag);
            b = new byte[nsiz];
            for (int i=0; i<dsiz; i++) 
            {
              tts[i] = new TTaggedStructure(sourceTag);
              raf.read(b,0,nsiz);
              tts[i].toStruct(b);
            }
            ov = tts;
            break;
          default:
            if (cdo == null) 
            {
              cdo = TFormat.makeCompoundDataObjectArray(sourceFmt, sourceTag, dsiz);
            }
            if (cdo != null)
            { 
              for (int i=0; i<dsiz; i++) 
              {
                raf.read(cdo[i].getByteArray(), 0, nsiz);
                cdo[i].toStruct();
              }          
            }
            ov = cdo;
            break;  
        }
        if (isSnapshot)
        { // we're finished 
          number = getFormattedHistoryDataSet(dataTarget,ov,dts,sstamp,ustamp);
          break; 
        }
        cc = getFormattedHistoryDataPoint(dataTarget,number+arrayoffset,ov,dindex,dts,sstamp,ustamp);
        if (cc < 0)
        {
          number = cc;
          break;
        }
        if (req.sample < 0 && dv != null && !isInRange((float)dv[dindex],rng.lwrLimit,rng.uprLimit)) continue;
        number++;
        if (number >= arraysize) break;
        if (raster > 1) 
        { // move along through file
          raf.seek(raf.getFilePointer()+rlen*(raster-1));
        }
        dlaststamp = dts;
        if (advance) raf.seek(nxtpos); // a summary won't need to do this ...
      }
      raf.close();
      if (datatime > 0.0) lastReadTimestamp = (long)(datatime*1000);
      if (datasys > 0) lastReadSystemStamp = datasys;
      if (debug) DbgLog.log("loadLTSData", "found "+number+" records in range");
      return number;
    }
    catch (Exception e)
    {
      e.printStackTrace();
      TFecLog.log(TineLogLevel.ERROR,"Error reading history record from "+fileName+" : " + e.toString());
    }    
    if (debug) DbgLog.log("loadLTSData", "nothing found");
    return 0; 
  }
  public int getDataFromSTS(THistoryRequest req,THistoryRange rng,
      Object targetData,int arraysize,int arrayOffset)
  {
  	int depth = hspec.getDepthShort();
    int raster=1,cnt,cc,m;
    int n,number,rnumber=0;
    int prate = hspec.getPollngRate();
    boolean isSnapshot = false, useSysStamp = false;
    long thistimestamp;
    int sds, uds;
    double dts;
    Object ov = null;
    
    lastReadTimestamp = 0;
    lastReadSystemStamp = 0;
    lastReadUserStamp = 0;
    if (prate == 0) return 0;
    if (depth == 0) return -TErrorList.un_allocated;
    if (req.index >= siz) return -TErrorList.out_of_range;
    if (arraysize < 1) return -TErrorList.buffer_too_small;    
    if (rng.startstamp > 0 && rng.stopstamp >= rng.startstamp) useSysStamp = true;
    if (rng.starttime < rng.stoptime - depth*prate) rng.starttime = rng.stoptime - depth*prate;
    raster = (int)(rng.stoptime-rng.starttime)/(prate*arraysize);
    m = (int)(rng.stoptime-rng.starttime)%(prate*arraysize);
    if (m != 0) raster++;
    if (raster < 1) raster = 1;
    if (arraysize == 1) raster = 1;
    if (req.sample > 0) raster = req.sample;
    if (rng.starttime == rng.stoptime) isSnapshot = true;
    if (useSysStamp && rng.startstamp == rng.stopstamp) isSnapshot = true;
    if (isSnapshot) raster = 1;
    n = rngPtr;
    /* move raster past the current position to start at the beginning */
    n = (n + raster + depth) % depth;
    if (arraysize == rng.stoptime-rng.starttime) 
    { // caller wants the array filled
      if (arraysize <= depth)
      {
        n = (n + (depth-arraysize) + depth) % depth;
        rng.starttime = timestamp[n];
        rng.stoptime = TDataTime.now();
      }
    }
    int fieldIndex = -1;
    short targetFmt = TDataType.getObjectFormat(targetData);
    if (targetFmt != TFormat.CF_HISTORY && !TFormat.isSimpleFormat(fmt))
    {
    	fieldIndex = req.fieldindex;
    }
    int this_n = n;
    lastReadTimestamp = rng.starttime; lastReadSystemStamp = rng.startstamp;
    for (number=arrayOffset,cnt=0; number < arraysize && number <= depth && cnt < depth; cnt += raster)
    {
      thistimestamp = timestamp[n];
      sds = sysstamp[n];
      uds = usrstamp[n];
      this_n = n;
      n = (n + raster + depth) % depth;
      if (isSnapshot)
      {
        if (thistimestamp > rng.stoptime) rng.stoptime = thistimestamp;
        if (sds > rng.stopstamp) rng.stopstamp = sds;
      }      
      if (useSysStamp)
      {
        if (sds < lastReadSystemStamp || sds > rng.stopstamp) continue;
      }
      else 
      {
        if (thistimestamp < lastReadTimestamp || thistimestamp > rng.stoptime) continue;
      }
      lastReadTimestamp = thistimestamp;
      lastReadSystemStamp = sds;
      lastReadUserStamp = uds;
      rnumber++;
      if (!isSnapshot && arraysize < 2) continue;
      ov = getNAMEArrayAtPosition(this_n);
      if (ov == null) ov = getCompoundArrayAtPosition(this_n);
      if (ov == null) ov = getDoubleArrayAtPosition(this_n);
      if (ov == null) ov = getStructArrayAtPosition(this_n);
      dts = ((double)thistimestamp)/1000.0;
      if (isSnapshot)
      {
        number = getFormattedHistoryDataSet(targetData,ov,dts,sds,uds);
        break; // we're finished 
      }
      cc = getFormattedHistoryDataPoint(targetData,number,ov,req.index,fieldIndex,dts,sds,uds);
      if (cc < 0)
      {
        number = cc;
        break;
      }
      if (req.sample < 0 && ov instanceof double[])
      {
        double[] dv = (double[])ov;
    		if (!isInRange((float)dv[req.index],rng.lwrLimit,rng.uprLimit)) continue;
      }
      number++;    
    }    
    if (!isSnapshot && arraysize < 2) assignNumberOfPoints(targetData,rnumber);
    return number;
  }
  public DSUMMARY getDataSummaryFromSTS(double sampleInterval)
  {
    double si = 1.0; // default sample interval in seconds
    DSUMMARY ds = new DSUMMARY();
    if (sampleInterval > 0) si = sampleInterval;
    
    double pps = 1000.0 / (double)hspec.getPollngRate();
    double depthInSeconds = (double)hspec.getDepthShortInSeconds();
    if (si > depthInSeconds) si = depthInSeconds;
    int npts = (int)(pps * si);
    if (npts < 1) throw new TineRuntimeErrorException(TErrorList.invalid_interval);
    int depth = hspec.getDepthShort();
    if (npts > depth) npts = depth;
    if (npts == 1 && depth > 2) npts = 2; // get it some lee way
    
    int k=0, cp = (rngPtr + 1) % depth;
    int n = (cp + depth - npts) % depth; // beginning of sampling interval
    double tnow = TDataTime.nowInSeconds();
    double ts, v, ra, ave=0, max=0, min=0, del=0;
    double[] dv;
    for (int np=-1; n != cp; n=np)
    {
      np = (n + 1 + depth) % depth;
      ts = (double)timestamp[n]/1000.0;
      if (ts < tnow - si) continue; 
      dv = getDoubleArrayAtPosition(n);
      if (dv == null)  throw new TineRuntimeErrorException(TErrorList.not_implemented);
      v = dv[0];
      ave += v;
      if (k == 0) 
      {
        max = min = v;
      }
      else
      {
        if (v > max) max = v;
        else if (v < min) min = v;
        ra = ave/(double)(k+1);
        if (v > ra) del += v - ra; else del += ra - v;
      }
      k++;
    }
    if (k == 0) throw new TineRuntimeErrorException(TErrorList.no_data_in_range);
    ave /= (double)k;
    del /= (double)k;
    ds.ave[0] = ave;
    ds.max[0] = max;
    ds.min[0] = min;
    ds.dev[0] = del;
    ds.cnt[0] = k;
    ds.ts[0] = (int)(si + (tnow - si)/2.0);
    ds.tMax[0] = ds.ts[0];
    ds.tMin[0] = ds.ts[0];    
    return ds;
  }
  private void assignNumberOfPoints(Object data,int numberOfPoints)
  {
    if (data instanceof float[])
    {
      ((float[])data)[0] = (float)numberOfPoints;
    }
    else if (data instanceof FLTINT[])
    {
      if (((FLTINT[])data)[0] == null) ((FLTINT[])data)[0] = new FLTINT();
      ((FLTINT[])data)[0].ival = numberOfPoints;
      ((FLTINT[])data)[0].fval = (float)numberOfPoints;
    }
    else if (data instanceof INTFLTINT[])
    {
      if (((INTFLTINT[])data)[0] == null) ((INTFLTINT[])data)[0] = new INTFLTINT();
      ((INTFLTINT[])data)[0].i1val = numberOfPoints;
      ((INTFLTINT[])data)[0].fval  = (float)numberOfPoints;
      ((INTFLTINT[])data)[0].i2val = numberOfPoints;
    }
    else if (data instanceof FLTINTINT[])
    {
      if (((FLTINTINT[])data)[0] == null) ((FLTINTINT[])data)[0] = new FLTINTINT();
      ((FLTINTINT[])data)[0].i1val = numberOfPoints;
      ((FLTINTINT[])data)[0].i2val = numberOfPoints;
      ((FLTINTINT[])data)[0].f1val  = (float)numberOfPoints;
    }
    else if (data instanceof int[])
    {
      ((int[])data)[0] = numberOfPoints;
    }
    else if (data instanceof INTINT[])
    {
      if (((INTINT[])data)[0] == null) ((INTINT[])data)[0] = new INTINT();
      ((INTINT[])data)[0].i2val = numberOfPoints;
      ((INTINT[])data)[0].i1val = numberOfPoints;
    }
    else if (data instanceof double[])
    {
      ((double[])data)[0] = (double)numberOfPoints;
    }
    else if (data instanceof DBLDBL[])
    {
      if (((DBLDBL[])data)[0] == null) ((DBLDBL[])data)[0] = new DBLDBL();
      ((DBLDBL[])data)[0].d2val = (double)numberOfPoints;
      ((DBLDBL[])data)[0].d1val = (double)numberOfPoints;
    }
    else if (data instanceof NAME16I[])
    {
      if (((NAME16I[])data)[0] == null) ((NAME16I[])data)[0] = new NAME16I();
      ((NAME16I[])data)[0].ival = numberOfPoints;
      ((NAME16I[])data)[0].name = "";
    }
    else if (data instanceof NAME32I[])
    {
      if (((NAME32I[])data)[0] == null) ((NAME32I[])data)[0] = new NAME32I();
      ((NAME32I[])data)[0].ival = numberOfPoints;
      ((NAME32I[])data)[0].name = "";
    }
    else if (data instanceof NAME64I[])
    {
      if (((NAME64I[])data)[0] == null) ((NAME64I[])data)[0] = new NAME64I();
      ((NAME64I[])data)[0].ival = numberOfPoints;
      ((NAME64I[])data)[0].name = "";
    }    
  }
  private boolean isInRange(float value,float lower, float upper)
  {
    if (value < lower) return false;
    if (value > upper) return false;
    return true;
  }
  public float getFloatValueAtPosition(int ptr,int index)
  {
    if (index < 0 || index >= siz) return -1;
    if (ptr < 0 || ptr > hspec.getDepthShort()) return -1;
    if (rngBuffer instanceof float[]) return ((float[])rngBuffer)[ptr*siz+index];
    if (rngBuffer instanceof byte[]) return (float)((byte[])rngBuffer)[ptr*siz+index];
    if (rngBuffer instanceof short[]) return (float)((short[])rngBuffer)[ptr*siz+index];
    if (rngBuffer instanceof int[]) return (float)((int[])rngBuffer)[ptr*siz+index];
    if (rngBuffer instanceof double[]) return (float)((double[])rngBuffer)[ptr*siz+index];
    return -1;    
  }  
  public double getDoubleValueAtPosition(int ptr,int index)
  {
    if (index < 0 || index >= siz) return -1;
    if (ptr < 0 || ptr > hspec.getDepthShort()) return -1;
    if (rngBuffer instanceof double[]) return ((double[])rngBuffer)[ptr*siz+index];
    if (rngBuffer instanceof byte[]) return (double)((byte[])rngBuffer)[ptr*siz+index];
    if (rngBuffer instanceof short[]) return (double)((short[])rngBuffer)[ptr*siz+index];
    if (rngBuffer instanceof int[]) return (double)((int[])rngBuffer)[ptr*siz+index];
    if (rngBuffer instanceof float[]) return (double)((float[])rngBuffer)[ptr*siz+index];
    return -1;    
  }
  public int getLongValueAtPosition(int ptr,int index)
  {
    if (index < 0 || index >= siz) return -1;
    if (ptr < 0 || ptr > hspec.getDepthShort()) return -1;
    if (rngBuffer instanceof int[]) return ((int[])rngBuffer)[ptr*siz+index];
    if (rngBuffer instanceof byte[]) return (int)((byte[])rngBuffer)[ptr*siz+index];
    if (rngBuffer instanceof short[]) return (int)((short[])rngBuffer)[ptr*siz+index];
    if (rngBuffer instanceof double[]) return (int)((double[])rngBuffer)[ptr*siz+index];
    if (rngBuffer instanceof float[]) return (int)((float[])rngBuffer)[ptr*siz+index];
    return -1;    
  }
  public short getShortValueAtPosition(int ptr,int index)  
  {
    if (index < 0 || index >= siz) return -1;
    if (ptr < 0 || ptr > hspec.getDepthShort()) return -1;
    if (rngBuffer instanceof byte[]) return (short)((byte[])rngBuffer)[ptr*siz+index];
    if (rngBuffer instanceof short[]) return ((short[])rngBuffer)[ptr*siz+index];
    if (rngBuffer instanceof int[]) return (short)((int[])rngBuffer)[ptr*siz+index];
    if (rngBuffer instanceof double[]) return (short)((double[])rngBuffer)[ptr*siz+index];
    if (rngBuffer instanceof float[]) return (short)((float[])rngBuffer)[ptr*siz+index];
    return -1;    
  }
  public NAME16 getNAME16ValueAtPosition(int ptr,int index)
  {
    if (index < 0 || index >= siz) return null;
    if (ptr < 0 || ptr > hspec.getDepthShort()) return null;
    if (rngBuffer instanceof NAME16[]) return ((NAME16[])rngBuffer)[ptr*siz+index];
    if (rngBuffer instanceof NAME32[]) return new NAME16(((NAME32[])rngBuffer)[ptr*siz+index].toString());
    if (rngBuffer instanceof NAME64[]) return new NAME16(((NAME64[])rngBuffer)[ptr*siz+index].toString());
    return null;    
  }
  public NAME32 getNAME32ValueAtPosition(int ptr,int index)
  {
    if (index < 0 || index >= siz) return null;
    if (ptr < 0 || ptr > hspec.getDepthShort()) return null;
    if (rngBuffer instanceof NAME32[]) return ((NAME32[])rngBuffer)[ptr*siz+index];
    if (rngBuffer instanceof NAME16[]) return new NAME32(((NAME16[])rngBuffer)[ptr*siz+index].toString());
    if (rngBuffer instanceof NAME64[]) return new NAME32(((NAME64[])rngBuffer)[ptr*siz+index].toString());
    return null;    
  }
  public NAME64 getNAME64ValueAtPosition(int ptr,int index)
  {
    if (index < 0 || index >= siz) return null;
    if (ptr < 0 || ptr > hspec.getDepthShort()) return null;
    if (rngBuffer instanceof NAME64[]) return ((NAME64[])rngBuffer)[ptr*siz+index];
    if (rngBuffer instanceof NAME16[]) return new NAME64(((NAME16[])rngBuffer)[ptr*siz+index].toString());
    if (rngBuffer instanceof NAME32[]) return new NAME64(((NAME32[])rngBuffer)[ptr*siz+index].toString());
    return null;    
  }
  public TTaggedStructure[] getStructArrayAtPosition(int ptr)
  {
  	int ssiz = TStructRegistry.getSizeInBytes(tag);
  	int len = ssiz > 0 ? siz/ssiz : siz;
  	TTaggedStructure[] values = new TTaggedStructure[len];
  	if (rngBuffer instanceof TTaggedStructure[])
  	{
      for (int i=0; i<len; i++) values[i] = ((TTaggedStructure[])rngBuffer)[ptr*len+i]; 
      return values;  		
  	}
  	return null;
  }
  public TCompoundDataObject[] getCompoundArrayAtPosition(int ptr)
  {
    TCompoundDataObject[] values = new TCompoundDataObject[siz];
    if (rngBuffer instanceof TCompoundDataObject[])
    {
      for (int i=0; i<siz; i++) values[i] = ((TCompoundDataObject[])rngBuffer)[ptr*siz+i]; 
      return values;
    }
    return null;
  }
  public NAME[] getNAMEArrayAtPosition(int ptr)
  {
    NAME[] values = new NAME[siz];
    if (rngBuffer instanceof NAME64[]) 
    {
      for (int i=0; i<siz; i++) values[i] = ((NAME64[])rngBuffer)[ptr*siz+i]; 
      return values;
    }
    if (rngBuffer instanceof NAME16[]) 
    {
      for (int i=0; i<siz; i++) values[i] = ((NAME16[])rngBuffer)[ptr*siz+i]; 
      return values;
    }
    if (rngBuffer instanceof NAME32[])
    {
      for (int i=0; i<siz; i++) values[i] = ((NAME32[])rngBuffer)[ptr*siz+i]; 
      return values;      
    }
    return null;
  }
  public float[] getFloatArrayAtPosition(int ptr)
  {
    float[] values = new float[siz];
    
    if (rngBuffer instanceof float[])
    {
      System.arraycopy((float[])rngBuffer,ptr*siz,values,0,siz);
    }
    else if (rngBuffer instanceof byte[])
    {
      for (int i=0; i<siz; i++) values[i] = (float)((byte[])rngBuffer)[ptr*siz+i];
    }
    else if (rngBuffer instanceof short[])
    {
      for (int i=0; i<siz; i++) values[i] = (float)((short[])rngBuffer)[ptr*siz+i];
    }
    else if (rngBuffer instanceof int[])
    {
      for (int i=0; i<siz; i++) values[i] = (float)((int[])rngBuffer)[ptr*siz+i];
    }
    else if (rngBuffer instanceof double[])
    {
      for (int i=0; i<siz; i++) values[i] = (float)((double[])rngBuffer)[ptr*siz+i];
    }
    else
    {
      return null;
    }
    return values;
  }
  public double[] getDoubleArrayAtPosition(int ptr)
  {
    double[] values = new double[siz];
    
    if (rngBuffer instanceof double[])
    {
      System.arraycopy((double[])rngBuffer,ptr*siz,values,0,siz);
    }
    else if (rngBuffer instanceof byte[])
    {
      for (int i=0; i<siz; i++) values[i] = (double)((byte[])rngBuffer)[ptr*siz+i];
    }
    else if (rngBuffer instanceof short[])
    {
      for (int i=0; i<siz; i++) values[i] = (double)((short[])rngBuffer)[ptr*siz+i];
    }
    else if (rngBuffer instanceof int[])
    {
      for (int i=0; i<siz; i++) values[i] = (double)((int[])rngBuffer)[ptr*siz+i];
    }
    else if (rngBuffer instanceof float[])
    {
      for (int i=0; i<siz; i++) values[i] = (double)((float[])rngBuffer)[ptr*siz+i];
    }
    else if (rngBuffer instanceof TCompoundDataObject[])
    {
      for (int i=0; i<siz; i++) values[i] = ((TCompoundDataObject[])rngBuffer)[ptr*siz+i].getDoubleValue();
    }
    else
    {
      return null;
    }
    return values;
  }
  public int[] getLongArrayAtPosition(int ptr)
  {
    int[] values = new int[siz];
    
    if (rngBuffer instanceof int[])
    {
      System.arraycopy((int[])rngBuffer,ptr*siz,values,0,siz);
    }
    else if (rngBuffer instanceof short[])
    {
      for (int i=0; i<siz; i++) values[i] = (int)((short[])rngBuffer)[ptr*siz+i];
    }
    else if (rngBuffer instanceof byte[])
    {
      for (int i=0; i<siz; i++) values[i] = (int)((byte[])rngBuffer)[ptr*siz+i];
    }
    else if (rngBuffer instanceof double[])
    {
      for (int i=0; i<siz; i++) values[i] = (int)((double[])rngBuffer)[ptr*siz+i];
    }
    else if (rngBuffer instanceof float[])
    {
      for (int i=0; i<siz; i++) values[i] = (int)((float[])rngBuffer)[ptr*siz+i];
    }
    else
    {
      return null;
    }
    return values;
  }
  public short[] getShortArrayAtPosition(int ptr)
  {
    short[] values = new short[siz];
    
    if (rngBuffer instanceof short[])
    {
      System.arraycopy((short[])rngBuffer,ptr*siz,values,0,siz);
    }
    else if (rngBuffer instanceof byte[])
    {
      for (int i=0; i<siz; i++) values[i] = (short)((byte[])rngBuffer)[ptr*siz+i];
    }
    else if (rngBuffer instanceof int[])
    {
      for (int i=0; i<siz; i++) values[i] = (short)((int[])rngBuffer)[ptr*siz+i];
    }
    else if (rngBuffer instanceof double[])
    {
      for (int i=0; i<siz; i++) values[i] = (short)((double[])rngBuffer)[ptr*siz+i];
    }
    else if (rngBuffer instanceof float[])
    {
      for (int i=0; i<siz; i++) values[i] = (short)((float[])rngBuffer)[ptr*siz+i];
    }
    else
    {
      return null;
    }
    return values;
  }
  public NAME16[] getNAME16ArrayAtPosition(int ptr)
  {
    NAME16[] values = new NAME16[siz];
    
    if (rngBuffer instanceof NAME16[])
    {
      for (int i=0; i<siz; i++) values[i] = new NAME16(((NAME16[])rngBuffer)[i].toString());
    }
    if (rngBuffer instanceof NAME32[])
    {
      for (int i=0; i<siz; i++) values[i] = new NAME16(((NAME32[])rngBuffer)[i].toString());
    }
    if (rngBuffer instanceof NAME64[])
    {
      for (int i=0; i<siz; i++) values[i] = new NAME16(((NAME64[])rngBuffer)[i].toString());
    }
   return values;
  }
  public NAME32[] getNAME32ArrayAtPosition(int ptr)
  {
    NAME32[] values = new NAME32[siz];
    
    if (rngBuffer instanceof NAME16[])
    {
      for (int i=0; i<siz; i++) values[i] = new NAME32(((NAME16[])rngBuffer)[ptr*siz+i].toString());
    }
    if (rngBuffer instanceof NAME32[])
    {
      for (int i=0; i<siz; i++) values[i] = new NAME32(((NAME32[])rngBuffer)[ptr*siz+i].toString());
    }
    if (rngBuffer instanceof NAME64[])
    {
      for (int i=0; i<siz; i++) values[i] = new NAME32(((NAME64[])rngBuffer)[ptr*siz+i].toString());
    }
    return values;
  }
  public NAME64[] getNAME64ArrayAtPosition(int ptr)
  {
    NAME64[] values = new NAME64[siz];
    
    if (rngBuffer instanceof NAME16[])
    {
      for (int i=0; i<siz; i++) values[i] = new NAME64(((NAME16[])rngBuffer)[ptr*siz+i].toString());
    }
    if (rngBuffer instanceof NAME32[])
    {
      for (int i=0; i<siz; i++) values[i] = new NAME64(((NAME32[])rngBuffer)[ptr*siz+i].toString());
    }
    if (rngBuffer instanceof NAME64[])
    {
      for (int i=0; i<siz; i++) values[i] = new NAME64(((NAME64[])rngBuffer)[ptr*siz+i].toString());
    }
    return values;
  }
  private int writeRecordToRingBuffer(Object values,int position)
  {
    if (values instanceof byte[]) 
    {
      if (rngBuffer instanceof byte[]) 
      { 
	      int len = ((byte[])values).length;
	      if (len > siz) len = siz;
	      System.arraycopy(values,0,(byte[])rngBuffer,position*siz,len);
	      return 0;
      }
      if (rngBuffer instanceof TTaggedStructure[]) 
      {
      	int ssiz = TStructRegistry.getSizeInBytes(tag);
      	int len = ssiz > 0 ? siz/ssiz : siz;
	      for (int i=0; i<len; i++)
	      {
	      	((TTaggedStructure[])rngBuffer)[position*len + i].toStruct((byte[])values,i*ssiz,ssiz);
	      }
	      return 0;
      }
      return TErrorList.illegal_format;
    }
    else if (values instanceof short[]) 
    {
      if (!(rngBuffer instanceof short[])) return TErrorList.illegal_format;
      int len = ((short[])values).length;
      if (len > siz) len = siz;
      System.arraycopy(values,0,(short[])rngBuffer,position*siz,len);            
    }
    else if (values instanceof int[]) 
    {
      if (!(rngBuffer instanceof int[])) return TErrorList.illegal_format;
      int len = ((int[])values).length;
      if (len > siz) len = siz;
      System.arraycopy(values,0,(int[])rngBuffer,position*siz,len);            
    }
    else if (values instanceof float[]) 
    {
      if (!(rngBuffer instanceof float[])) return TErrorList.illegal_format;
      int len = ((float[])values).length;
      if (len > siz) len = siz;
      System.arraycopy(values,0,(float[])rngBuffer,position*siz,len);      
    }
    else if (values instanceof double[]) 
    {
      if (!(rngBuffer instanceof double[])) return TErrorList.illegal_format;
      int len = ((double[])values).length;
      if (len > siz) len = siz;
      System.arraycopy(values,0,(double[])rngBuffer,position*siz,len);            
    }
    else if (values instanceof TCompoundDataObject[]) 
    {
      if (!(rngBuffer instanceof TCompoundDataObject[])) return TErrorList.illegal_format;
      if (((TCompoundDataObject[])values)[0].getFormat() != 
          ((TCompoundDataObject[])rngBuffer)[0].getFormat())
        return TErrorList.illegal_format;
      int len = ((TCompoundDataObject[])values).length;
      if (len > siz) len = siz;
      System.arraycopy(values,0,rngBuffer,position*siz,len);            
    }
    else if (values instanceof TTaggedStructure[]) 
    {
      if (!(rngBuffer instanceof TTaggedStructure[])) return TErrorList.illegal_format;
      if (((TTaggedStructure[])values)[0].getName().compareTo(((TTaggedStructure[])rngBuffer)[0].getName()) != 0)
        return TErrorList.illegal_format;
      int len = ((TCompoundDataObject[])values).length;
      int ssiz = TStructRegistry.getSizeInBytes(tag);
      int dlen = ssiz > 0 ? siz / ssiz : siz;
      if (len > dlen) len = dlen;
      System.arraycopy(values,0,rngBuffer,position*dlen,len);            
    }
    else
    {
      return TErrorList.illegal_format;
    }
    return 0;
  }
  public int writeRecordToSTS(TDataType d)
  { // write to Short-Term-Storage (i.e. ram)
    if (d == null || hspec == null) return TErrorList.code_failure;
    if (rngBuffer == null || timestamp == null) return TErrorList.un_allocated;
    rngPtr = (rngPtr + 1) % hspec.getDepthShort();
    int cc = writeRecordToRingBuffer(d.getDataObject(),rngPtr);
    timestamp[rngPtr] = d.getDataTimeStamp();
    sysstamp[rngPtr] = d.getSystemDataStamp();
    usrstamp[rngPtr] = d.getUserDataStamp();
    return cc;
  }
  public boolean needsToFillSTS = false;
  public int fillSTS()
  {
    try
    {
      needsToFillSTS = false;
      if (hspec.getDepthLong() > 1) return 0; // not necessary
      if (!TFormat.isNumberFormat(fmt)) return TErrorList.not_implemented;
      DBLDBL[] dda = new DBLDBL[rngBufferSize];
      String aprp = eqm.getExportName()+"."+prp;
      double stop = TDataTime.getDataTimeStamp(); 
      double start = stop - hspec.getDepthShortInSeconds() - 1800.0;
      int npts = THistory.getArchivedData(eqm.getContext(), "HISTORY", aprp, dev, 0, 0, start, stop, dda,500);
      if (npts > 0)
      {
        TDataType d = new TDataType(1,fmt);
        double[] dd = new double[1];
        for (int i=0; i<npts; i++)
        {
          dd[0] = dda[i].d1val;
          d.putData(dd);
          d.setDataTimeStamp(dda[i].d2val);
          writeRecordToSTS(d);
        }
      }
      return npts < 0 ? -npts : 0;
    }
    catch (Exception e)
    {
      TFecLog.log(TineLogLevel.ERROR,"THistoryRecord","could not fill short-term storage for "+prp+" : "+e.getMessage());
      return TErrorList.code_failure;
    }
  }
  private String twoDigitString(int n)
  {
    return n < 10 ? "0" + n : "" + n;
  }
  public void setArchiveFileName(long cycleTime)
  {
    if (calendar == null) calendar = Calendar.getInstance();
    calendar.setTimeInMillis(cycleTime);
    int dd = calendar.get(Calendar.DAY_OF_MONTH);
    if (dd == lastArchivedDayOfTheMonth) return;
    int yy, mm;
    lastArchivedDayOfTheMonth = dd;
    yy = calendar.get(Calendar.YEAR) % 100;
    mm = calendar.get(Calendar.MONTH) + 1; // January = 0
    if (useMonthlyHistoryFiles) dd = 1; 
    ArchiveFileName = getFilePrefix() +  twoDigitString(yy) + twoDigitString(mm) + 
                      twoDigitString(dd) + "." + Integer.toHexString(rcdIdx);

    if (useStandardFileSet)
    {
      int mm0 = ((mm+12) - ((hspec.getDepthLong()+1)%12))%12;
      yy -= hspec.getDepthLong()/12;
      if (mm0 > mm) yy--;
      
      String rmvFileName = getFilePrefix() +  twoDigitString(yy) + twoDigitString(mm0) + 
                           twoDigitString(dd) + "." + Integer.toHexString(rcdIdx);
      rotateHistoryFile(rmvFileName,ArchiveFileName);
    }
    else
    {
      removeOutdatedLTSFiles(cycleTime);
    }
  }
  private boolean unlinkHistFile(int yy,int mm,int dd)
  {
    String fileName = null, poiName = null, pfix = getFilePrefix();
    fileName = pfix+twoDigitString(yy)+twoDigitString(mm)+twoDigitString(dd)+"."+Integer.toHexString(rcdIdx);
    poiName = new String(fileName); poiName.replace(pfix, "pi");
    File f = null;
    try
    {
      f = new File(arcDbPath+File.separator+fileName);
      if (!f.exists()) return false;
      f.delete();
      f = new File(arcDbPath+File.separator+poiName);
      if (f.exists()) f.delete();
      return true;
    }
    catch (Exception e)
    {
      String msg = "unable to delete file " + f != null ? f.getName() : "(null file)";
      MsgLog.log("removeOutdatedLTSFiles", msg,TErrorList.file_error,e,0);
      TFecLog.log(TineLogLevel.ERROR,msg);
    }
    return false;
  }
  public void removeOutdatedLTSFiles(long cycleTime)
  {
    int yy,mm,mm0,dd;
    calendar.setTimeInMillis(cycleTime);
    yy = calendar.get(Calendar.YEAR) % 100;
    mm0 = calendar.get(Calendar.MONTH) + 1; // January = 0
    mm = ((mm0+12) - ((hspec.getDepthLong()+1)%12))%12;
    yy -= hspec.getDepthLong()/12;
    if (mm > mm0) yy--;
    dd = calendar.get(Calendar.DAY_OF_MONTH);
    
    do
    {
      for (int i=dd-1; i>0; i--)
      {
        if (!unlinkHistFile(yy,mm,i) && i<27) break;
      }
      mm--;
      dd = 32;
    } while (mm > 0);    
  }
  private String getFilePrefix()
  {
    if (useStandardFileSet) return "ts";
    return useMinimalStorage ? "ar" : "ta"; /* "ha" */
  }
  private void rotateHistoryFile(String src, String dst)
  {
    File fdst = new File(arcDbPath +  File.separator + dst);
    if (fdst.exists()) return;
    File fsrc = new File(arcDbPath +  File.separator + src);
    fsrc.renameTo(fdst);
    try
    {
      RandomAccessFile raf = new RandomAccessFile(fdst,"rw");
      setFilePosition(raf,AS_STATE_SIZE);
    }
    catch (Exception e)
    {
      MsgLog.log("rotateHistoryFile","could not set file position in "+fdst.getName(),TErrorList.file_error,e,0);
    }
    TFecLog.log("rotate history file " + src + " to " + dst);
  }
  public String getArchiveFileName()
  {
    return ArchiveFileName;
  }
  public int getArrayType()
  {
    return aType;
  }
  public int getRingBufferPointer()
  {
    return rngPtr;
  }
  public String getDev()
  {
    return dev;
  }
  public TEquipmentModule getEqm()
  {
    return eqm;
  }
  public byte getFmt()
  {
    return fmt;
  }
  public String getTag()
  {
    return tag;
  }
  public String getRemoteHistoryServer() 
  { 
    return hspec.getRemoteServerSpecification();
  }
  public THistorySpecification getHspec()
  {
    return hspec;
  }
  public long getLastArchiveTime()
  {
    return lastArchiveTime;
  }
  public String getPrp()
  {
    return prp;
  }
  public int getRcdIdx()
  {
    return rcdIdx;
  }
  public int getRngBufferSize()
  {
    return rngBufferSize;
  }
  public int getArraySize()
  {
    return siz;
  }
  public int getRecordLength()
  {
    int rlen = storeSummary ? DSUMMARY.sizeInBytes  :  siz * TFormat.formatSizeOf(fmt);
    return rlen + getRecordPrefixLength();
  }
  public long getTimestamp()
  {
    return rngPtr < 0 ? 0 : timestamp[rngPtr];
  }
  public int getSystemStamp()
  {
    return rngPtr < 0 ? 0 : sysstamp[rngPtr];
  }
  public int getUserStamp()
  {
    return rngPtr < 0 ? 0 : usrstamp[rngPtr];
  }
  public boolean isUseMonthlyHistoryFiles()
  {
    return useMonthlyHistoryFiles;
  }
  public void setUseMonthlyHistoryFiles(boolean useMonthlyHistoryFiles)
  {
    this.useMonthlyHistoryFiles = useMonthlyHistoryFiles;
  }
  public boolean isUseMinimalStorage()
  {
    return useMinimalStorage;
  }
  public void setUseMinimalStorage(boolean value)
  {
    useMinimalStorage = value;
  }
  /**
   * @param useMsecHistoryTimeStamps
   * @deprecated
   * use setUseMinimalStorage() instead
   */
  public void setUseMsecHistoryTimeStamps(boolean useMsecHistoryTimeStamps)
  {
    useMinimalStorage = !useMsecHistoryTimeStamps;
  }
  public boolean isWithinTolerance()
  {
    double t, d0, d1;
    String s0, s1;
    byte[] ba0, ba1;
    float aTol = hspec.getAbsoluteTolerance();
    float pTol = hspec.getPercentTolerance();
    int p;
    int ptr0 = rngPtr;
    int ptr1 = hspec.getDepthShort(); //(ptr0 + rngBufferSize - 1) % rngBufferSize; 
    switch (fmt)
    {
      case TFormat.CF_DOUBLE:
      case TFormat.CF_INT64:          
      case TFormat.CF_FLOAT:
      case TFormat.CF_INT16:
      case TFormat.CF_INT32:
        for (p=0; p<siz; p++)
        {
          d0 = getDoubleValueAtPosition(ptr0,p);
          d1 = getDoubleValueAtPosition(ptr1,p);
          t = aTol + pTol * Math.abs(d1);
          if (Math.abs(d0 - d1) > t) return false;
        }
        break;
      case TFormat.CF_NAME16:
      case TFormat.CF_NAME32:
      case TFormat.CF_NAME64:
        for (p=0; p<siz; p++)
        {
          s0 = getNAME64ValueAtPosition(ptr0,p).toString();
          s1 = getNAME64ValueAtPosition(ptr1,p).toString();
          if (s0.compareTo(s1) == 0) continue;
          return false;
        }
        break;
      case TFormat.CF_STRUCT:
        try
        {
	      	int ssiz = TStructRegistry.getSizeInBytes(tag);
	      	int len = ssiz > 0 ? siz/ssiz : siz;
	      	TTaggedStructure[] tts0 = getStructArrayAtPosition(ptr0);
	      	TTaggedStructure[] tts1 = getStructArrayAtPosition(ptr1);
	        for (p=0; p<len; p++)
	        {
	          ba0 = tts0[p].getByteArray();
	          ba1 = tts1[p].getByteArray();
	          if (!Arrays.equals(ba0, ba1)) return false;
	        }      	
        }
        catch (Exception ignore)
        { // whatever it is, break and return true
        }
        break;
      default:
        try
        {
          TCompoundDataObject[] tco0 = getCompoundArrayAtPosition(ptr0);
          TCompoundDataObject[] tco1 = getCompoundArrayAtPosition(ptr1);
          for (p=0; p<siz; p++)
          {
            ba0 = tco0[p].getByteArray();
            ba1 = tco1[p].getByteArray();
            if (!Arrays.equals(ba0, ba1)) return false;
          }
        }
        catch (Exception ignore)
        { // possibly an invalid cast exception
          // whatever it is, break and return true
        }
        break;
    }
    return true;
  }
  private static double poiRangeFactor = 0.10;
  private static double poiToleranceFactor = 10.0;
  public boolean isPointOfInterest()
  {
    if (!keepPointsOfInterest) return false;
    if (rngPtr == 1 && timestamp[rngBufferSize-1] == 0)
    { // just starting
      return false;
    }
    double t, d0, d1;
    float aTol = hspec.getAbsoluteTolerance();
    float pTol = hspec.getPercentTolerance();
    boolean useRange = pTol > 0 ? false : true;
    if (useRange && range <= 0) 
    {
      useRange = false;
      if (aTol == 0) return false;
    }
    int p;
    int ptr0 = rngPtr;
    int ptr1 = (ptr0 + rngBufferSize - 1) % rngBufferSize; 
    switch (fmt)
    {
      case TFormat.CF_DBLDBL:
      case TFormat.CF_DBLTIME:  
      case TFormat.CF_FLTINT:
      case TFormat.CF_FLTFLT:
      case TFormat.CF_DOUBLE:
      case TFormat.CF_INT64:          
      case TFormat.CF_FLOAT:
      case TFormat.CF_INT16:
      case TFormat.CF_INT32:
        for (p=0; p<siz; p++)
        {
          d0 = getDoubleValueAtPosition(ptr0,p);
          d1 = getDoubleValueAtPosition(ptr1,p);
          t = useRange ? poiRangeFactor * range : poiToleranceFactor *(aTol + pTol * Math.abs(d1));
          if (Math.abs(d0 - d1) > t) return true;
        }
        break;
      default:
        break;
    }
    return false;    
  }
  public static double getPoiRangeFactor()
  {
    return poiRangeFactor;
  }
  public static void setPointsOfInterestRangeFactor(double poiRangeFactor)
  {
    THistoryRecord.poiRangeFactor = poiRangeFactor;
  }
  public static double getPoiToleranceFactor()
  {
    return poiToleranceFactor;
  }
  public static void setPointsOfInterestToleranceFactor(double poiToleranceFactor)
  {
    THistoryRecord.poiToleranceFactor = poiToleranceFactor;
  }
}
