/*
 * File THistoryRecordStruct.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.server.histories;

import de.desy.tine.definitions.TStrings;
import de.desy.tine.structUtils.TTaggedStructure;

public class THistoryRecordStruct extends TTaggedStructure
{
  private String devString = null;
  private String prpString = null;
  private char[] prp = new char[TStrings.PROPERTY_NAME_SIZE];
  private char[] dev = new char[TStrings.DEVICE_NAME_SIZE];
  private int[] siz = new int[1];
  private int[] fmt = new int[1];
  private int[] pRate = new int[1]; // polling rate
  private int[] aRate = new int[1]; // archive rate
  private int[] depthShort = new int[1];
  private int[] depthLong = new int[1];
  private int[] heartbeat = new int[1];
  private float[] pTolerance = new float[1];
  private float[] aTolerance = new float[1];
  
  private void pushString(String s,char[] c)
  {
    if (s != null) 
    {
      int len = s.length();
      if (len > c.length) len = c.length; 
      s.getChars(0, len, c, 0);
    }    
  }
  private void initStructure()
  {
    addField(prp,"property");
    addField(dev,"device");
    addField(siz,"dataSize");
    addField(fmt,"dataFormat");
    addField(pRate,"pollingRate");
    addField(aRate,"archiveRate");
    addField(depthShort,"depthShort");
    addField(depthLong,"depthLong");
    addField(heartbeat,"heartbeat");
    addField(pTolerance,"pTolerance");
    addField(aTolerance,"aTolerance");
    initDone();      
  }
  public THistoryRecordStruct()
  {
    super("HRSr4");
    initStructure();
  }
  public THistoryRecordStruct(String devName,String devProperty,int size,
      int format,THistorySpecification hsp)
  {
    super("HRSr4");
    initStructure();
    pushString(devName,dev);
    pushString(devProperty,prp);
    siz[0] = size;
    fmt[0] = format;
    pRate[0] = hsp.getPollngRate();
    aRate[0] = hsp.getArchiveRate();
    depthShort[0] = hsp.getDepthShort();
    depthLong[0] = hsp.getDepthLong();
    heartbeat[0] = hsp.getHeartbeat();
    pTolerance[0] = hsp.getPercentTolerance();
    aTolerance[0] = hsp.getAbsoluteTolerance();
  }
  public int getArchiveRate()
  {
    return aRate[0];
  }
  public float getAbsoluteTolerance()
  {
    return aTolerance[0];
  }
  public int getDepthLong()
  {
    return depthLong[0];
  }
  public int getDepthMonths()
  {
    int ldm = depthLong[0];
    ldm &= 0xffff;
    return ldm;    
  }
  public int getDepthDays()
  {
    int ldd = depthLong[0];
    ldd = ldd >> 16;
    return ldd & 0x1f;        
  }
  public int getDepthShort()
  {
    return depthShort[0];
  }
  public String getDevice()
  {
    if (devString == null) devString = new String(dev);
    return devString;
  }
  public String getProperty()
  {
    if (prpString == null) prpString = new String(prp);
    return prpString;
  }
  public int getFormat()
  {
    return fmt[0];
  }
  public int getHeartbeat()
  {
    return heartbeat[0];
  }
  public int getPollingRate()
  {
    return pRate[0];
  }
  public float getPercentTolerance()
  {
    return pTolerance[0];
  }
  public int getSize()
  {
    return siz[0];
  }
}
