/*
 * File MsgLog.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.server.logger;

import java.util.ArrayList;
import java.util.logging.FileHandler;
import java.util.logging.Level;
import java.util.logging.LogRecord;

import de.desy.tine.client.TLinkFactory;
import de.desy.tine.dataUtils.TDataTime;
import de.desy.tine.definitions.TErrorList;

public class MsgLog
{
  public class MsgEntry
  {
    private long time;
    private int code;
    private String msg;
    private Throwable e;
    MsgEntry(long t, String m) { time = t; msg = m; code = 0;}
    MsgEntry(long t, String m, int c, Throwable x) { time = t; msg = m; code = c; e = x; }
    public int getCode() { return code; }
    public String getMsg() { return msg; }
    public long getTime() { return time; }
    public Throwable getExecption() { return e; }
    public String toString() { return msg; } 
  }
  private static final int MAX_NUM_MESSAGES = 512;
  private static MsgEntry[] msgs = new MsgEntry[MAX_NUM_MESSAGES];
  private static int msgptr = 0;
  private static int dbgLevel = 0;
  public static void setDebugLevel(int level) { dbgLevel = level; }
  public static int getDebugLevel() { return dbgLevel; }
  private MsgLog()
  {
    String dbglvlStr;
    if ((dbglvlStr = System.getProperty("debug.level")) != null)
    {
      try { dbgLevel = Integer.parseInt(dbglvlStr);} catch (Exception ignore) {}
    }
    msgs[0] = new MsgEntry(System.currentTimeMillis(),"Message Log started");
  }
  private static MsgLog instance = new MsgLog();
  public static void log(String message,int code)
  {
    if (message == null) return;
    msgptr = (msgptr+1) % MAX_NUM_MESSAGES;
    msgs[msgptr] = instance.new MsgEntry(System.currentTimeMillis(),message,code,null);
  }
  public static void log(String source,String message,int code,Throwable exception,int debugLevel)
  {
    if (exception != null && exception instanceof OutOfMemoryError)
    { // there's probably no hope of recovery
      if (TLinkFactory.isRunningAsServer())
      { // if I'm a server then exit !
        System.exit(1);
      }
    }
    if (message == null) return;
    if (source == null) source = "(null)"; 
    if (instance == null) instance = new MsgLog(); // a servlet class loader issue ? 
    msgptr = (msgptr+1) % MAX_NUM_MESSAGES;
    msgs[msgptr] = instance.new MsgEntry(System.currentTimeMillis(),message,code,exception);
    if (dbgLevel >= debugLevel) DbgLog.log(source,message);
    if (TLinkFactory.isRunningAsServer() && code > 0 && !TErrorList.isSystemInternal(code) && code != TErrorList.operation_busy)
    {
      TFecLog.log(TineLogLevel.ERROR,source+": "+message);
    }
    //srvlog(source+": "+message);
  }
  static FileHandler logfileHandler = null;
  public static int srvlog(String text) 
  {
    int cc = 0;
    try
    {
      if (logfileHandler == null) 
      {
        String fn = "msg.log";
        logfileHandler = new FileHandler(fn, 50000, 3, true);
        logfileHandler.setFormatter(new TFecLogFormatter());
      }      
      LogRecord lr = new LogRecord(Level.INFO,text);
      logfileHandler.publish(lr);
    }
    catch (Exception e)
    {
      e.printStackTrace();
      cc = TErrorList.file_error;
    }
    return cc;
  }
  
  public static MsgEntry[] getMessages(long starttime)
  {
    int i=msgptr,n;
    ArrayList<MsgEntry> m = new ArrayList<MsgEntry>();
    for (n=0; n<MAX_NUM_MESSAGES; n++)
    { // first find the start (go backwards in time)
      i = (msgptr - n + MAX_NUM_MESSAGES) % MAX_NUM_MESSAGES;
      if (msgs[i] == null || msgs[i].time < starttime) break;
      m.add(msgs[i]);
    }
    return m.size() > 0 ? m.toArray(new MsgEntry[0]) : null;
  }
  public static MsgEntry[] getMessages()
  {
    return getMessages(MAX_NUM_MESSAGES);
  }
  public static MsgEntry[] getMessages(int depth)
  {
    return getMessages(depth,0);
  }
  public static MsgEntry[] getMessages(int depth,long start)
  {
    int i=msgptr,n;
    ArrayList<MsgEntry> m = new ArrayList<MsgEntry>();
    for (n=0; n<MAX_NUM_MESSAGES && n<depth; n++)
    { // first find the start (go backwards in time)
      i = (msgptr - n + MAX_NUM_MESSAGES) % MAX_NUM_MESSAGES;
      if (msgs[i] == null) break;
      if (msgs[i].getTime() < start) break;
      m.add(msgs[i]);
    }
    return m.size() > 0 ? m.toArray(new MsgEntry[0]) : null;
  }
  public static MsgEntry getLastMessage()
  {
    return msgs[msgptr];
  }
  public static void dumpMessages()
  {
    MsgEntry[] msgs = getMessages();
    String msg;
    for (MsgLog.MsgEntry m: msgs)
    {
      msg = m.getMsg();
      if (m.getCode() != 0) msg += " ("+TErrorList.getErrorString(m.getCode())+")";
      TLinkFactory.dbgPrint(TDataTime.toString(m.getTime())+" "+msg);
    }
  }
}
