/*
 * File TFecLog.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

/*
 * Created on Aug 31, 2005
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package de.desy.tine.server.logger;

import java.io.*;
import java.util.logging.*;

import de.desy.tine.client.TLinkFactory;
import de.desy.tine.definitions.TErrorList;
import de.desy.tine.server.equipment.TEquipmentModuleFactory;
import de.desy.tine.startup.TInitializer;
import de.desy.tine.startup.TInitializerFactory;
import de.desy.tine.types.NAME64I;

public class TFecLog
{
  public static final int LOG_SCAN_NONE = 0;
  public static final int LOG_SCAN_FEC = 1;
  public static final int LOG_SCAN_FULL = 2;
  private static int logFileAllowScan = LOG_SCAN_FEC;  
  private static TineLogLevel errorLogLevel = TineLogLevel.INFO;
  public static boolean isTineLogFile(String filename)
  {
    if (filename == null) return false;
    if (filename.startsWith("fec.log")) return true;
    if (filename.startsWith("commands.log")) return true;
    if (filename.startsWith("error.log")) return true;
    return false;
  }
  /**
   * @brief Sets the allowed range of log file retrieval from a FEC process
   * (via stock property SRVLOGFILE).
   *
   * A running FEC process will attempt to read and return the contents of
   * the file name input to stock property SRVLOGFILE.  The FEC process must
   * have read permission in order to succeed. The setting of
   * the LogFileAllowScan will further restrict the allowed directories from
   * which files may be read.  LogFileAllowScan = LOG_SCAN_FEC will only return
   * the contents of files within the FEC_HOME or FEC_LOG directories and their
   * subdirectories.  LogFileAllowScan = LOG_SCAN_FULL will itself allow the
   * entire file system.  LogFileAllowScan = LOG_SCAN_NONE will not allow any 
   * files to be retrieved other than fec.log.
   *
   * @param value is the desired scan range, one of: LOG_SCAN_NONE,
   * LOG_SCAN_FEC, LOG_SCAN_FULL (default = LOG_SCAN_FEC).  
   */
  public static void setLogFileAllowScan(int value)
  {
    if (value < LOG_SCAN_NONE) value = LOG_SCAN_NONE;
    if (value > LOG_SCAN_FULL) value = LOG_SCAN_FULL;
    logFileAllowScan = value;
  }
  /**
   * @brief Gets the allowed range of log file retrieval from a FEC process
   * (via stock property SRVLOGFILE).
   *
   * A running FEC process will attempt to read and return the contents of
   * the file name input to stock property SRVLOGFILE.  The FEC process must
   * have read permission in order to succeed. The setting of
   * the LogFileAllowScan will further restrict the allowed directories from
   * which files may be read.  LogFileAllowScan = LOG_SCAN_FEC will only return
   * the contents of files within the FEC_HOME or FEC_LOG directories and their
   * subdirectories.  LogFileAllowScan = LOG_SCAN_FULL will itself allow the
   * entire file system.  LogFileAllowScan = LOG_SCAN_NONE will not allow any 
   * files to be retrieved other than fec.log.
   *
   * @return the current scan range setting, one of: LOG_SCAN_NONE,
   * LOG_SCAN_FEC, LOG_SCAN_FULL (default = LOG_SCAN_FEC).  
   */
  public static int getLogFileAllowScan() 
  { 
    return logFileAllowScan; 
  }
  public static boolean isAllowedLogFileAccess(String filename)
  {
    if (logFileAllowScan == LOG_SCAN_FULL) return true;
    if (logFileAllowScan == LOG_SCAN_NONE)
    {
      if (filename.compareTo("fec.log.0") == 0) return true;
      return false;
    }
    if (!isAbsolutePath(filename)) return true;
    String plh = initializer.getLogHome();
    if (plh == null) return false;
    if (plh.length() > 0 && filename.startsWith(plh)) return true;
    String pfh = initializer.getFecHome();
    if (pfh.compareToIgnoreCase(plh) != 0)
    { /* FEC_LOG != FEC_HOME */
      if (pfh.length() > 0 && filename.startsWith(pfh)) return true;
    }
    return false;
  }
  private static TInitializer initializer = TInitializerFactory.getInstance().getInitializer();
  private static FileHandler feclogHandler = null;
  private static FileHandler errorlogHandler = null;
  private static int logDepth = 1000;
  private static int debugLevel = 0;
  public static int getDebugLevel() { return debugLevel; }
  public static void setDebugLevel(int debuglevel) { debugLevel = debuglevel; }
  public static int getLogDepth() { return logDepth; }
  public static void setLogDepth(int depthInLines) { if (depthInLines > 0) logDepth = depthInLines; }
  private static String tag = null;
  private static final String fecLogFileName = "fec.log.0";
  private static final String errorLogFileName = "error.log.0";
  public static void close()
  {
  	if (feclogHandler != null) 
  	{
  		feclogHandler.close();
  		feclogHandler =  null;
  	}
  	if (errorlogHandler != null) 
  	{
  		errorlogHandler.close();
  		errorlogHandler =  null;
  	}
  }
  /**
   * @brief logs the entry into the server's 'fec.log' file
   * 
   * The fec.log is rotated with the current file instance named fec.log.0.
   *
   * @param text is the log entry
   * 
   * @return 0 upon success or a TINE error code
   */
  public static int log(String text)
  {
    return log(TineLogLevel.INFO,text);
  }
  public static int log(TineLogLevel loglevel, String text)
  {
    if (!TLinkFactory.isRunningAsServer()) return TErrorList.not_running;
    if (tag == null) tag = TEquipmentModuleFactory.getInstance().getFecName();
    return log(loglevel,tag,text);
  }
  private static boolean hasLoggingError = false;
  public static boolean hasLoggingError() { return hasLoggingError; }
  /**
   * @brief logs the entry into the server's 'fec.log' file
   * 
   * The fec.log is rotated with the current file instance named fec.log.0.
   * 
   * @param tag is a tag identify preceding the logged entry
   * @param text is the log entry
   * 
   * @return 0 upon success or a TINE error code
   */
  public static int log(TineLogLevel loglevel, String tag, String text) {
    String txt = "[" + tag + "][" + loglevel.toString() + "] : " + text;
    // the feclog part
    if (feclogHandler == null) 
    {
      feclogHandler = initializer.getFecLogResource();
      if (feclogHandler == null)
      { // still null !!!
        DbgLog.log("TFecLog","cannot write to log file : no resource !");
        TEquipmentModuleFactory.getInstance().setFileErrorAlarm(fecLogFileName);
        hasLoggingError = true;
        return -1;
      }
      feclogHandler.setFormatter(new TFecLogFormatter());
    }
    try
    {
      
      LogRecord lr = new LogRecord(Level.INFO,txt);
      feclogHandler.publish(lr);
      hasLoggingError = false;
    }
    catch (OutOfMemoryError e)
    {
      e.printStackTrace();      
    }
    catch (Exception e)
    {
      DbgLog.log("TFecLog","cannot write to log file : " + e.toString());
      TEquipmentModuleFactory.getInstance().setFileErrorAlarm(fecLogFileName);
      hasLoggingError = true;
    }
    // the errorlog part
    if( loglevel.ordinal() > errorLogLevel.ordinal()) 
    {
      if (errorlogHandler == null) 
      {
        errorlogHandler = initializer.getErrorLogResource();
        if (errorlogHandler == null)
        { // still null !!!
          DbgLog.log("TFecLog","cannot write to error log file : no resource !");
          TEquipmentModuleFactory.getInstance().setFileErrorAlarm(errorLogFileName);
          hasLoggingError = true;
          return -1;
        }
        errorlogHandler.setFormatter(new TFecLogFormatter());
      }
      try
      {

        LogRecord lr = new LogRecord(Level.INFO,txt);
        errorlogHandler.publish(lr);
        hasLoggingError = false;
      }
      catch (OutOfMemoryError e)
      {
        e.printStackTrace();      
      }
      catch (Exception e)
      {
        DbgLog.log("TFecLog","cannot write to error log file : " + e.toString());
        TEquipmentModuleFactory.getInstance().setFileErrorAlarm(errorLogFileName);
        hasLoggingError = true;
      }
    }
    // finish
    if (TLinkFactory.debugLevel > 0) DbgLog.log("TFecLog",txt);
    return 0;
  }
  public static int log(String tag, String text)
  {
    return log( TineLogLevel.INFO, tag, text);
  }
  public static String[] getFiles()
  {
    return getFiles(initializer.getLogHome());
  }
  public static String[] getFiles(String filePath)
  {
    File path = new File(filePath);
    File[] rawList = path.listFiles();
    String[] files = new String[rawList.length];
    for (int i=0; i<rawList.length; i++)
    {
      files[i] = rawList[i].getName();      
    }
    return files;
  }
  public static NAME64I[] getFilesAsName64I()
  {
    return getFilesAsName64I(initializer.getLogHome());
  }
  public static NAME64I[] getFilesAsName64I(String filePath)
  {
    File path = new File(filePath);
    File[] rawList = path.listFiles();
    NAME64I[] files = new NAME64I[rawList.length];
    for (int i=0; i<rawList.length; i++)
    {
      files[i] = new NAME64I(rawList[i].getName(),(int)rawList[i].length());      
    }
    return files;
  }
  public static String getLines(int maxNumberOfLines)
  {
    return getLines(fecLogFileName,maxNumberOfLines);
  }
  public static boolean fileExists(String fileName)
  {
    String fn = null;
    if (fileName == null || fileName.length() == 0) return false;
    if (fileName.indexOf('/') == -1 && fileName.indexOf('\\') == -1 )
    { // just the file name given => look in the default log directory
      fn = initializer.getLogHome() + "/" + fileName;
    }
    else
    { // full path given
      fn = fileName;
    }
    return new File(fn).exists();
  }
  public static int writeTextFile(String fileName,char[] fileText)
  {
    String fn = null;
    try
    {
      if (fileName.indexOf('/') == -1 && fileName.indexOf('\\') == -1 )
      { // just the file name given => look in the default log directory
        fn = initializer.getLogHome() + "/" + fileName;
      }
      else
      { // full path given
        fn = fileName;
      }
      File f = new File(fn);
      if (!f.exists()) f.createNewFile();
      FileWriter fw = new FileWriter(f);
      fw.write(new String(fileText));
      fw.close();
//      RandomAccessFile raf = new RandomAccessFile(fn,"rw");
//      raf.write(new String(fileText).getBytes());
//      raf.close();
    }
    catch (Exception e)
    {
      DbgLog.log("TFecLog",e.toString());
      return TErrorList.file_error;
    }
    return 0;
  }
  public static boolean isAbsolutePath(String filename)
  {
    if (filename == null || filename.length() == 0) return false;
    if (filename.startsWith("/")) return true;
    if (filename.startsWith("\\\\")) return true;
    if (filename.startsWith(":\\", 1)) return true;
    if (filename.startsWith(":/",1)) return true; 
    return false;
  }
  public static int getFileInfo(String fileName,long[] len,long[] tim)
  {
    try
    {
      String fn = null;
      if (fileName.indexOf('/') == -1 && fileName.indexOf('\\') == -1 )
      { // just the file name given => look in the default log directory
        fn = initializer.getLogHome() + "/" + fileName;
      }
      else
      { // full path given
        fn = fileName;
      }
      File f = new File(fn);
      if (!f.exists()) return TErrorList.no_such_file;
      if (len != null && len.length > 0) len[0] = f.length();
      if (tim != null && tim.length > 0) tim[0] = f.lastModified();
    }
    catch (Exception e)
    {
      DbgLog.log("TFecLog.getFileInfo",e.toString());
      return TErrorList.file_error;
    }
    return 0;
  }
  public static String getLines(String fileName,int maxNumberOfLines)
  {
    String fn = null;
    StringBuffer sb = null;
    String text=null;
    try
    {
      if (fileName.indexOf('/') == -1 && fileName.indexOf('\\') == -1 )
      { // just the file name given => look in the default log directory
        fn = initializer.getLogHome() + "/" + fileName;
      }
      else
      { // full path given
        fn = fileName;
      }
      RandomAccessFile raf = new RandomAccessFile(fn,"r");
      if (maxNumberOfLines < 1) maxNumberOfLines = 1;
      int flen = (int)raf.length();
      int fpos = flen - maxNumberOfLines*80;
      if (fpos < 0) fpos = 0;
      raf.seek(fpos);
      sb = new StringBuffer(flen-fpos);
      while ((text = raf.readLine()) != null)
      {
        sb.append(text);
        if (sb.length() < flen) sb.append("\n");
      }
      raf.close();
    } 
    catch (Exception e)
    {
      DbgLog.log("TFecLog.getLines",e.toString());
    }
    if (sb == null) sb = new StringBuffer("file unavailable");
    text = sb.toString();
    if (debugLevel > 0) DbgLog.log("TFecLog",text);    
    return text;
  }
}
