/*
 * File TTaggedStructure.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package de.desy.tine.structUtils;

import de.desy.tine.dataUtils.UnsignedUtils;
import de.desy.tine.definitions.TErrorList;
import de.desy.tine.definitions.TFormat;
import de.desy.tine.endianUtils.Swap;
import de.desy.tine.exceptions.TineRuntimeErrorException;
import de.desy.tine.server.logger.MsgLog;
import de.desy.tine.stringUtils.Str;
import de.desy.tine.types.*;
import java.util.*;
import java.io.*;
import java.lang.reflect.Array;

/**
 * Tagged Structure The transfer of 'user-defined' data structures is
 *        allowed in TINE. In order to avoid nasty problems concerning alignment
 *        and byte swapping it is strongly suggested to extend the
 *        TTaggedStructure class when creating a data structure.
 */
public class TTaggedStructure implements Cloneable
{
  private String structName;
  protected boolean registered = false;
  private boolean hasExtendedSpace = false;
  private int nextFieldIndex = 0;
  private int sizeInBytes;
  private int extendedOffset;
  private int extendedSizeInBytes = 0;
  //private ArrayList<structField> fieldArrayList = new ArrayList<structField>();
  private LinkedHashMap<String, structField> fieldArrayList = new LinkedHashMap<String,structField>();
  static private int indent = 0; // used by toString()
  public int getNextFieldIndex() { return nextFieldIndex; }
  public boolean hasExtendedSpace() { return hasExtendedSpace; }
  public int getExtendedSizeInBytes() { return extendedSizeInBytes; }
  public int getExtendedOffset() { return extendedOffset; }
  public void setExtendedOffset(int value)
  {
    if (value > sizeInBytes) extendedOffset = value;
  }
  public structField[] getFields()
  {
    return fieldArrayList.values().toArray(new structField[0]);
  }
  public String[] getFieldNames()
  {
    return fieldArrayList.keySet().toArray(new String[0]);
  }
  public Object getFieldValue(String name)
  {
    if (name == null || name.length() == 0) throw new TineRuntimeErrorException(TErrorList.invalid_name);
    structField sf = fieldArrayList.get(name);
    if (sf == null) throw new TineRuntimeErrorException(TErrorList.invalid_field, name + " is not a registered field of structure "+structName);
    return sf.getValue();
  }
  @Override
  public Object clone()
  {
    Object o = null;
    try
    {
      Class<?> clz = this.getClass();
      o = clz.getConstructor().newInstance();
    }
    catch (Exception e)
    {
      MsgLog.log("TTaggedStructure", "cannot instantiate " + e.getMessage(), TErrorList.code_failure, e, 0);
      try
      {
        o = super.clone();
      }
      catch (Exception x)
      {
        MsgLog.log("TTaggedStructure", "cannot clone " + e.getMessage(), TErrorList.code_failure, x, 0);
      }
    }
    return o;
  }
  /**
   * Creates a tagged structure with the given name.
   * @param name
   *          is the name to be used as the structure tag. Note that only the
   *          first 16 characters of the given name will be used as the tag
   *          name!
   *          
   * @note The addField() method must be called for each element of the
   *       structure data element.
   *       
   * @note Properties using tagged structures should register themselves
   *       correctly, either by passing the structure name when creating the
   *       TExportProperty() or if using a configuration file such as 
   *       exports.csv or fec.xml, then by supplying the structure name 
   *       there (as in FORMAT -> struct.MYTAG, where "MYTAG" is the 
   *       tagged structure name).
   */
  public TTaggedStructure(String name)
  {
    structName = new String(name);
    sizeInBytes = 0;
    if (!this.getClass().getSimpleName().equals("TTaggedStructure")) return;
    if (TStructRegistry.contains(structName))
    {
      TStructDescription tsd = TStructRegistry.get(structName);
      //if (tsd.isComplex()) return;
      TTaggedStructure tts = tsd.getTaggedStructure();
      if (tts != null) addFieldsFrom(tts); else addFieldsFrom(tsd);
      TStructRegistry.addToSystem(structName);
      registered = true;
    }
  }
  
  public TTaggedStructure(String name, boolean addFields)
  {
    structName = new String(name);
    sizeInBytes = 0;
    if (TStructRegistry.contains(structName))
    {
      registered = true;
      TStructRegistry.addToSystem(structName);
      if (addFields)
      {
      	TStructDescription tsd = TStructRegistry.get(name, null); 
	      TTaggedStructure ref = tsd.getTaggedStructure();
	      if (ref != null) addFieldsFrom(ref); 
	      else addFieldsFrom(tsd);
      }
    }
  }
  public void addFieldsFrom(TStructDescription tsd)
  {
    if (tsd != null)
    {
      try
      {
        Object[] elementData = tsd.getFields().toArray();
        TStructDescription.Field fld;
        for (Object e : elementData) 
        {
          fld = (TStructDescription.Field)e;
          switch (fld.format)
          {
            case TFormat.CF_BYTE:
              addField(new byte[fld.arraySize],fld.getName()); 
              break;
            case TFormat.CF_TEXT:
              addField(new char[fld.arraySize],fld.getName()); 
              break;
            case TFormat.CF_DOUBLE:
              addField(new double[fld.arraySize],fld.getName()); 
              break;
            case TFormat.CF_FLOAT:
              addField(new float[fld.arraySize],fld.getName()); 
              break;
            case TFormat.CF_IMAGE:
              addField(new IMAGE[fld.arraySize],fld.getName()); 
              break;
            case TFormat.CF_UINT16:
              addField(new short[fld.arraySize],fld.getName(),true); 
              break;
            case TFormat.CF_INT16:
              addField(new short[fld.arraySize],fld.getName()); 
              break;
            case TFormat.CF_BOOLEAN:
              addField(new boolean[fld.arraySize],fld.getName()); 
              break;
            case TFormat.CF_UINT32:
              addField(new int[fld.arraySize],fld.getName(),true); 
              break;              
            case TFormat.CF_INT32:
              addField(new int[fld.arraySize],fld.getName()); 
              break;
            case TFormat.CF_INT64:
              addField(new long[fld.arraySize],fld.getName()); 
              break;
            case TFormat.CF_ASPECTRUM:
              addField(new ASPECTRUM[fld.arraySize],fld.getName()); 
              break;
            case TFormat.CF_STRING:
              addField(new String[fld.arraySize],fld.getName()); 
              break;
            case TFormat.CF_STRUCT:
              addField(new TTaggedStructure[fld.arraySize],fld.getName()); 
              break;
            case TFormat.CF_AIMAGE:
              addField(new IMAGE[fld.arraySize],fld.getName()); 
              break;
            default:
              addField(new TCompoundDataObject[fld.arraySize],fld.getFormat(),fld.getArraySize(),"",fld.getName()); 
              break;
              
          }
        }
        initDone();
      }
      catch (Exception x)
      {
        MsgLog.log("TTaggedStruct.addFieldsFrom","could not register structure fields",TErrorList.code_failure,x,1);
      }
    }
  }
  public void addFieldsFrom(TTaggedStructure tts)
  {
    if (tts != null) 
    {
      Object[] elementData = tts.fieldArrayList.values().toArray();
      String fieldName;
      for (Object e : elementData) 
      {
        fieldName = ((structField)e).getName();
        if (e instanceof byteArrayField) 
        {
          byte[] field  = new byte[((byteArrayField)e).data.length];
          addField(field,fieldName);
        }
        else if (e instanceof charArrayField)
        {
          char[] field  = new char[((charArrayField)e).data.length];
          addField(field,fieldName);
        }
        else if (e instanceof doubleArrayField)
        {
          double[] field  = new double[((doubleArrayField)e).data.length];
          addField(field,fieldName);
        }
        else if (e instanceof floatArrayField) 
        {
          float[] field  = new float[((floatArrayField)e).data.length];
          addField(field,fieldName);
        }
        else if (e instanceof imageArrayField)
        {
          IMAGE[] field  = new IMAGE[((imageArrayField)e).data.length];
          addField(field,fieldName);
        }
        else if (e instanceof intArrayField)
        {
          int[] field  = new int[((intArrayField)e).data.length];
          addField(field,fieldName);
        }
        else if (e instanceof boolArrayField)
        {
          boolean[] field  = new boolean[((boolArrayField)e).data.length];
          addField(field,fieldName);
        }
        else if (e instanceof longArrayField)
        {
          long[] field  = new long[((longArrayField)e).data.length];
          addField(field,fieldName);
        }
        else if (e instanceof shortArrayField)
        {
          short[] field  = new short[((shortArrayField)e).data.length];
          addField(field,fieldName);
        }
        else if (e instanceof spectrumArrayField)
        {
          ASPECTRUM[] field  = new ASPECTRUM[((spectrumArrayField)e).data.length];
          addField(field,fieldName);
        }
        else if (e instanceof StringArrayField) 
        {
          String[] field  = new String[((StringArrayField)e).data.length];
          addField(field,fieldName);
        }
        else if (e instanceof StringBufferField)
        {
          StringBuffer field  = new StringBuffer();
          addField(field,fieldName);
        }
        else if (e instanceof TaggedStructureArrayField)
        {
          try
          {
            TaggedStructureArrayField tsaf =(TaggedStructureArrayField)e;
            TTaggedStructure[] tta = (TTaggedStructure[])tsaf.getValue();
            String tag = tta[0].structName;
            TTaggedStructure[] field  = new TTaggedStructure[tsaf.data.length];
            Class<?> clz = tsaf.data.getClass().getComponentType();
            for (int i=0; i< field.length; i++)
            {
              if (clz.getSimpleName().equals("TTaggedStructure"))
                field[i] = new TTaggedStructure(tag);
              else
                field[i] = (TTaggedStructure)clz.getConstructor().newInstance();
            }
            addField(field,fieldName);
          }
          catch (Exception ex)
          {
            ex.printStackTrace();
            MsgLog.log("TTaggedStructure.addFieldsFrom", "cannot add field "+fieldName,TErrorList.code_failure,ex,0);
          }
        }
        else if (e instanceof TaggedStructureField)
        {
          try
          {
            TaggedStructureField tsf = (TaggedStructureField)e;
            TTaggedStructure ts = (TTaggedStructure)tsf.getValue();
            Class<?> clz = tsf.data.getClass();
            TTaggedStructure field;
            if (clz.getSimpleName().equals("TTaggedStructure"))
              field =  new TTaggedStructure(ts.structName);
            else
              field = tsf.data.getClass().getConstructor().newInstance();
            addField(field,fieldName);
          }
          catch (Exception ex)
          {
            ex.printStackTrace();
            MsgLog.log("TTaggedStructure.addFieldsFrom", "cannot add field "+fieldName,TErrorList.code_failure,ex,0);
          }
        }
        else if (e instanceof TCompoundDataArrayField)
        {
          try
          {
            TCompoundDataObject[] field  = new TCompoundDataObject[((TCompoundDataArrayField)e).data.length];
            for (int i=0; i< field.length; i++)
              field[i] = ((TCompoundDataObject[])((TCompoundDataArrayField)e).data)[i].getClass().getConstructor().newInstance();
            addField(field,fieldName);
          }
          catch (Exception ex)
          {
            ex.printStackTrace();
            MsgLog.log("TTaggedStructure.addFieldsFrom", "cannot add field "+fieldName,TErrorList.code_failure,ex,0);
          }
        }
        else if (e instanceof TCompoundDataField)
        {
          TCompoundDataObject field;
          try
          {
            field = ((TCompoundDataField)e).data.getClass().getConstructor().newInstance();
            addField(field,fieldName);
          }
          catch (Exception ex)
          {
            ex.printStackTrace();
            MsgLog.log("TTaggedStructure.addFieldsFrom", "cannot add field "+fieldName,TErrorList.code_failure,ex,0);
          }
        }
        else if (e instanceof TStringField)
        {
          TString field  = new TString();
          addField(field,fieldName);
        }
      }
      initDone();
    }    
  }
  /**
   * Creates a tagged structure using the class name as the structure
   *        tag. Note that only the first 16 characters of the class name will
   *        be used as the tag name!
   * @note The addField() method must be called for each element of the
   *       structure data element.
   *        
   * \b Example:
   * @include eg_tineTaggedStructure1.java
   * 
   * @note Properties using tagged structures should register themselves
   *       correctly, either by passing the structure name when creating the
   *       TExportProperty() or if using a configuration file such as 
   *       exports.csv or fec.xml, then by supplying the structure name 
   *       there (as in FORMAT -> struct.MYTAG, where "MYTAG" is the 
   *       tagged structure name).
   */
  public TTaggedStructure()
  {
    structName = new String(this.getClass().getSimpleName());
    if (!structName.equals("TTaggedStructure") && TStructRegistry.contains(structName))
    {
      registered = true;
      sizeInBytes = TStructRegistry.getSizeInBytes(structName);
      TStructRegistry.addToSystem(structName);
    }
    else
      sizeInBytes = 0;
  }
  /**
   * Returns the structure's assigned tag name.
   * @return The structure's assigned tag name as a string.
   */
  public String getName()
  {
    return structName;
  }
  /**
   * Returns the structure's data size in bytes.
   * @return The structure's data size in bytes.
   */
  public int getSizeInBytes()
  {
    return sizeInBytes;
  }
  private void resetExtendedSizeInBytes()
  {
    extendedSizeInBytes = 0;
  }
  /**
   * scan thru the extended space and fix the reference and the total size
   */
  public void fixExtendedBytes()
  {
    resetExtendedSizeInBytes();
    for (structField sf : fieldArrayList.values())
    {
      sf.fixExtendedBytes();
    }    
  }
  public byte[] toByteArray()
  {
    ByteArrayOutputStream baos = new ByteArrayOutputStream(getSizeInBytes());
    baos.reset();
    DataOutputStream ds = new DataOutputStream(baos);
    try
    {
      resetExtendedSizeInBytes();
      for (structField sf : fieldArrayList.values())
      {
        sf.fixExtendedBytes();
        sf.writeToOutputStream(ds);
      }
      ds.close();
    }
    catch (IOException e)
    {
      MsgLog.log("TTaggedStructure", "---Internal error: cannot convert TaggedStructure: " + getName()
          + " to byte array ", TErrorList.code_failure, e, 0);
      e.printStackTrace();
      return null;
    }
    return baos.toByteArray();
  }
  public byte[] toExtendedByteArray()
  {
    ByteArrayOutputStream baos = new ByteArrayOutputStream(getExtendedSizeInBytes());
    baos.reset();
    DataOutputStream ds = new DataOutputStream(baos);
    try
    {
      for (structField sf : fieldArrayList.values())
      {
        sf.writeExtendedBytes(ds);
      }
      ds.close();
    }
    catch (IOException e)
    {
      MsgLog.log("TTaggedStructure", "---Internal error: cannot write extended bytes: " + getName()
          + " to byte array ", TErrorList.code_failure, e, 0);
      e.printStackTrace();
      return null;
    }
    return baos.toByteArray();
  }
  public byte[] toTotalByteArray()
  {
    ByteArrayOutputStream baos = new ByteArrayOutputStream(getSizeInBytes());
    baos.reset();
    DataOutputStream ds = new DataOutputStream(baos);
    try
    {
      resetExtendedSizeInBytes();
      for (structField sf : fieldArrayList.values())
      {
        sf.fixExtendedBytes();
        sf.writeToOutputStream(ds);
      }
      for (structField sf : fieldArrayList.values())
      {
        sf.writeExtendedBytes(ds);
      }     
      ds.close();
    }
    catch (IOException e)
    {
      MsgLog.log("TTaggedStructure", "---Internal error: cannot convert TaggedStructure: " + getName()
          + " to byte array ", TErrorList.code_failure, e, 0);
      e.printStackTrace();
      return null;
    }
    return baos.toByteArray();
  }
  public byte[] getByteArray()
  {
    return toByteArray();
  }
  public void toStruct(byte[] data)
  {
    toStruct(data,0,data.length);
  }
  public void toStruct(byte[] data,int offset,int length)
  {
    ByteArrayInputStream dis = new ByteArrayInputStream(data,offset,length);
    DataInputStream ds = new DataInputStream(dis);
    try
    {
      for (structField sf : fieldArrayList.values())
      {
        sf.readFromInputStream(ds);
        sf.readExtendedBytes(data);
      }
      ds.close();
      dis.close();
    }
    catch (IOException e)
    {
      MsgLog.log("TTaggedStructure",
          "---Internal error: cannot convert input data stream to TaggedStructure: " + getName(),
          TErrorList.code_failure, e, 0);
      e.printStackTrace();
    }
  }
  private String arraySeparator = " ";
  public void setArraySeparator(String value)
  {
    if (value != null && value.length() > 0) arraySeparator = value;
  }
  private String fieldSeparator = ", ";
  public void setFieldSeparator(String value)
  {
    if (value != null && value.length() > 0) fieldSeparator = value;
  }
  synchronized public String toString(boolean useKeyValueFields)
  {
    if (!useKeyValueFields) return toString();
    StringBuffer sb = new StringBuffer("");
    
    sb.append("[");
    for (structField sf : fieldArrayList.values())
    {
      sf.setFieldSeparator(fieldSeparator);
      sb.append(sf.getName() + ": ");
      sb.append(sf.toString());
      if (sf.getArrayIndex() < fieldArrayList.size()-1) sb.append(fieldSeparator);
    }
    sb.append("]");   
    return sb.toString();
  }
  public FieldValue[] toFieldValue()
  {
    FieldValue[] fva = new FieldValue[fieldArrayList.size()];
    Object value;
    for (structField sf : fieldArrayList.values())
    {
      value = sf.getValue();
      if (value instanceof char[])
        value = Str.trim(new String((char[])value));
      fva[sf.arrayIndex] = new FieldValue(sf.getName(),value);
    }
    return fva;
  }

  protected void pushChars(String s,char[] c)
  {
    Arrays.fill(c,(char)0);
    if (s == null) s = "";
    int len = s.length();
    if (len > c.length) len = c.length;     
    s.getChars(0, len, c, 0);
  }
  /*
   * Returns a string representation of this tagged structure
   */
  @Override
  public String toString()
  {
    return toString("\n");
  }
  synchronized public String toString(String eolTok)
  {
    StringBuffer sb = new StringBuffer("");
    String indentString = new String("");
    boolean multiLine = eolTok.contains("\n");
    if (multiLine)
    {
      indent += 3;
      for (int i = 0; i < indent; i++) indentString += " ";
    }
    else
    {
      indentString = " ";
    }
    String pad =  multiLine ? eolTok + indentString : " ";
    sb.append(pad + "Struct: " + getName());
    sb.append(pad + "{\n");
    for (structField sf : fieldArrayList.values())
    {
      sb.append(indentString + " " + sf.getName() + " ->");
      sb.append(sf.toString());
      sb.append(sf.getArrayIndex() < fieldArrayList.size()-1 ? eolTok : "\n");
    }
    sb.append(indentString + "}");
    indent -= 3;
    return sb.toString();
  }
  /**
   * Terminates the tagged structure definition and signals that the
   *        tagged structure is ready for use. This method plays an integral
   *        role in defining the tagged structure. It should be called following
   *        all calls to addField().
   */
  public void initDone()
  {
    if (!registered)
    {
      TStructRegistry.fill(structName, TFormat.CF_NULL, 1);
      registered = true;
    }
    extendedOffset += sizeInBytes;
    TStructDescription tsd = TStructRegistry.get(structName);
    if (tsd != null) 
    {
      tsd.fieldsComplete = true;
      tsd.setTaggedStructure(this);
      TStructRegistry.assignServerKey(structName,null);
    }
    nextFieldIndex = 0;
  }
  private void chkNewField(String fieldName)
  {
    if (!registered) return;
    if (fieldArrayList.size() > TStructRegistry.get(structName).getNFields())
    {
      fieldArrayList.remove(fieldName);
    }    
  }
  /**
   * @param field
   *          is the structure field to add to the tagged structure
   * @param fieldName
   *          is the name of the structure field
   */
  public void addField(byte[] field, String fieldName)
  {
    if (fieldName == null) fieldName = "field-"+nextFieldIndex;
    Str.toAscii(fieldName);
    fieldArrayList.put(fieldName,new byteArrayField(field, fieldName));
    if (!registered)
    {
      TStructRegistry.fill(structName, fieldName, TFormat.CF_BYTE, field.length);
      sizeInBytes += field.length;
      nextFieldIndex++;
    }
    chkNewField(fieldName);
  }
  /**
   * @param field
   *          is the structure field to add to the tagged structure
   * @param fieldName
   *          is the name of the structure field
   */
  public void addField(char[] field, String fieldName)
  {
    if (fieldName == null) fieldName = "field-"+nextFieldIndex;
    fieldName = Str.toAscii(fieldName);
    Str.toAscii(field);
    fieldArrayList.put(fieldName,new charArrayField(field, fieldName));
    if (!registered)
    {
      TStructRegistry.fill(structName, fieldName, TFormat.CF_TEXT, field.length);
      sizeInBytes += field.length;
      nextFieldIndex++;
    }
    chkNewField(fieldName);
  }
  /**
   * @param field
   *          is the structure field to add to the tagged structure
   * @param fieldName
   *          is the name of the structure field
   */
  public void addField(short[] field, String fieldName)
  {
    if (fieldName == null) fieldName = "field-"+nextFieldIndex;
    fieldName = Str.toAscii(fieldName);
    fieldArrayList.put(fieldName,new shortArrayField(field, fieldName));
    if (!registered)
    {
      TStructRegistry.fill(structName, fieldName, TFormat.CF_INT16, field.length);
      sizeInBytes += field.length * TFormat.formatSizeOf(TFormat.CF_INT16);
      nextFieldIndex++;
    }
    chkNewField(fieldName);
  }
  public void addField(short[] field, String fieldName,boolean unsigned)
  {
    if (fieldName == null) fieldName = "field-"+nextFieldIndex;
    fieldName = Str.toAscii(fieldName);
    fieldArrayList.put(fieldName,new shortArrayField(field, fieldName,unsigned));
    if (!registered)
    {
      TStructRegistry.fill(structName, fieldName, TFormat.CF_INT16, field.length);
      sizeInBytes += field.length * TFormat.formatSizeOf(TFormat.CF_INT16);
      nextFieldIndex++;
    }
    chkNewField(fieldName);
  }
  /**
   * @param field
   *          is the structure field to add to the tagged structure
   * @param fieldName
   *          is the name of the structure field
   */
  public void addField(int[] field, String fieldName)
  {
    addField(field,fieldName,false);
  }
  public void addField(int[] field, String fieldName, boolean unsigned)
  {
    if (fieldName == null) fieldName = "field-"+nextFieldIndex;
    fieldName = Str.toAscii(fieldName);
    fieldArrayList.put(fieldName,new intArrayField(field, fieldName,unsigned));
    if (!registered)
    {
      TStructRegistry.fill(structName, fieldName, TFormat.CF_INT32, field.length);
      sizeInBytes += field.length * TFormat.formatSizeOf(TFormat.CF_INT32);
      nextFieldIndex++;
    }
    chkNewField(fieldName);
  }
  public void addField(boolean[] field, String fieldName)
  {
    if (fieldName == null) fieldName = "field-"+nextFieldIndex;
    fieldName = Str.toAscii(fieldName);
    fieldArrayList.put(fieldName,new boolArrayField(field, fieldName));
    if (!registered)
    {
      TStructRegistry.fill(structName, fieldName, TFormat.CF_INT32, field.length);
      sizeInBytes += field.length * TFormat.formatSizeOf(TFormat.CF_INT32);
      nextFieldIndex++;
    }
    chkNewField(fieldName);
  }
  /**
   * @param field
   *          is the structure field to add to the tagged structure
   * @param fieldName
   *          is the name of the structure field
   */
  public void addField(float[] field, String fieldName)
  {
    if (fieldName == null) fieldName = "field-"+nextFieldIndex;
    fieldName = Str.toAscii(fieldName);
    fieldArrayList.put(fieldName,new floatArrayField(field, fieldName));
    if (!registered)
    {
      TStructRegistry.fill(structName, fieldName, TFormat.CF_FLOAT, field.length);
      sizeInBytes += field.length * TFormat.formatSizeOf(TFormat.CF_FLOAT);
      nextFieldIndex++;
    }
    chkNewField(fieldName);
  }
  /**
   * @param field
   *          is the structure field to add to the tagged structure
   * @param fieldName
   *          is the name of the structure field
   */
  public void addField(double[] field, String fieldName)
  {
    if (fieldName == null) fieldName = "field-"+nextFieldIndex;
    fieldName = Str.toAscii(fieldName);
    fieldArrayList.put(fieldName,new doubleArrayField(field, fieldName));
    if (!registered)
    {
      TStructRegistry.fill(structName, fieldName, TFormat.CF_DOUBLE, field.length);
      sizeInBytes += field.length * TFormat.formatSizeOf(TFormat.CF_DOUBLE);
      nextFieldIndex++;
    }
    chkNewField(fieldName);
  }
  /**
   * @param field
   *          is the structure field to add to the tagged structure
   * @param fieldName
   *          is the name of the structure field
   */
  public void addField(long[] field, String fieldName)
  {
    if (fieldName == null) fieldName = "field-"+nextFieldIndex;
    fieldName = Str.toAscii(fieldName);
    fieldArrayList.put(fieldName,new longArrayField(field, fieldName));
    if (!registered)
    {
      TStructRegistry.fill(structName, fieldName, TFormat.CF_INT64, field.length);
      sizeInBytes += field.length * TFormat.formatSizeOf(TFormat.CF_INT64);
      nextFieldIndex++;
    }
    chkNewField(fieldName);
  }
  /**
   * @param field
   *          is the structure field to add to the tagged structure
   * @param fieldName
   *          is the name of the structure field
   */
  public void addField(TString field, String fieldName)
  {
    if (fieldName == null) fieldName = "field-"+nextFieldIndex;
    fieldName = Str.toAscii(fieldName);
    fieldArrayList.put(fieldName,new TStringField(field, fieldName));
    if (!registered)
    {
      TStructRegistry.fill(structName, fieldName, TFormat.CF_TEXT, field.getCapacity());
      sizeInBytes += field.getCapacity();
      nextFieldIndex++;
    }
    chkNewField(fieldName);
  }
  /**
   * @param field
   *          is the structure field to add to the tagged structure
   * @param fieldName
   *          is the name of the structure field
   */
  public void addField(StringBuffer field, String fieldName)
  {
    if (fieldName == null) fieldName = "field-"+nextFieldIndex;
    fieldName = Str.toAscii(fieldName);
    fieldArrayList.put(fieldName,new StringBufferField(field, fieldName));
    if (!registered)
    {
      TStructRegistry.fill(structName, fieldName, TFormat.CF_TEXT, field.capacity());
      sizeInBytes += field.capacity();
      nextFieldIndex++;
    }
    chkNewField(fieldName);
  }
  public void addField(String[] field, String fieldName)
  {
    if (fieldName == null) fieldName = "field-"+nextFieldIndex;
    fieldName = Str.toAscii(fieldName);
    fieldArrayList.put(fieldName,new StringArrayField(field, fieldName));
    if (!registered)
    {
      TStructRegistry.fill(structName, fieldName, TFormat.CF_STRING, field.length);
      sizeInBytes += 8 * field.length;
      // has extended bytes
      nextFieldIndex++;
    }
    chkNewField(fieldName);
  }
  public void addField(IMAGE[] field, String fieldName)
  {
    if (fieldName == null) fieldName = "field-"+nextFieldIndex;
    fieldName = Str.toAscii(fieldName);
    fieldArrayList.put(fieldName,new imageArrayField(field, fieldName));
    if (!registered)
    {
      TStructRegistry.fill(structName, fieldName, TFormat.CF_AIMAGE, field.length);
      sizeInBytes += (IMAGE.HEADER_SIZE+8) * field.length;
      // has extended bytes
      nextFieldIndex++;
    }
    chkNewField(fieldName);
  }
  public void addField(ASPECTRUM[] field, String fieldName)
  {
    if (fieldName == null) fieldName = "field-"+nextFieldIndex;
    fieldName = Str.toAscii(fieldName);
    fieldArrayList.put(fieldName,new spectrumArrayField(field, fieldName));
    if (!registered)
    {
      TStructRegistry.fill(structName, fieldName, TFormat.CF_ASPECTRUM, field.length);
      sizeInBytes += (ASPECTRUM.hdrSizeInBytes+8) * field.length;
      // has extended bytes
      nextFieldIndex++;
    }
    chkNewField(fieldName);
  }
  /**
   * @param field
   *          is the structure field to add to the tagged structure
   * @param fieldName
   *          is the name of the structure field
   */
  public void addField(TCompoundDataObject field, String fieldName)
  {
    if (fieldName == null) fieldName = "field-"+nextFieldIndex;
    fieldName = Str.toAscii(fieldName);
    fieldArrayList.put(fieldName,new TCompoundDataField(field, fieldName));
    if (!registered)
    {
      int num = 1;
      if (TFormat.hasHeader(field.getFormat()))
      {
        int hsiz = TFormat.getFormatHeaderSize(field.getFormat());
        num = (field.getSizeInBytes() - hsiz) / TFormat.formatSizeOf(field.getFormat());
      }
      TStructRegistry.fill(structName, fieldName, field.getFormat(), num);
      sizeInBytes += field.getSizeInBytes();
      nextFieldIndex++;
    }
    chkNewField(fieldName);
  }
  /**
   * @param field
   *          is the structure field to add to the tagged structure
   * @param fieldName
   *          is the name of the structure field
   */
  public void addField(TCompoundDataObject[] field, String fieldName)
  {
    if (fieldName == null) fieldName = "field-"+nextFieldIndex;
    fieldName = Str.toAscii(fieldName);
    fieldArrayList.put(fieldName,new TCompoundDataArrayField(field, fieldName));
    if (!registered)
    {
      Object tdo = field[0];
      if (field[0] == null) try
      {
        tdo = field.getClass().getComponentType().getConstructor().newInstance();
      }
      catch (Exception e)
      {
        e.printStackTrace();
        MsgLog.log("addField", "cannot add field : "+e.getMessage(),TErrorList.code_failure,e,0);
        return;
      }
      TStructRegistry.fill(structName, fieldName, ((TCompoundDataObject) tdo).getFormat(), field.length);
      sizeInBytes += ((TCompoundDataObject) tdo).getSizeInBytes() * field.length;
      nextFieldIndex++;
    }
    chkNewField(fieldName);
  }
  public void addField(TCompoundDataObject[] field, short format, int size, String tag, String fieldName)
  {
    if (fieldName == null) fieldName = "field-"+nextFieldIndex;
    fieldName = Str.toAscii(fieldName);
    fieldArrayList.put(fieldName,new TCompoundDataArrayField(field, format, size, tag, fieldName));
    if (!registered)
    {
      Object tdo = field[0];
      if (field[0] == null) try
      {
        tdo = TFormat.makeCompoundDataObjectArray(format, tag, size)[0];
      }
      catch (Exception e)
      {
        e.printStackTrace();
        MsgLog.log("addField", "cannot add field : "+e.getMessage(),TErrorList.code_failure,e,0);
        return;
      }
      TStructRegistry.fill(structName, fieldName, ((TCompoundDataObject) tdo).getFormat(), field.length);
      sizeInBytes += ((TCompoundDataObject) tdo).getSizeInBytes() * field.length;
      nextFieldIndex++;
    }
    chkNewField(fieldName);
  }
  /**
   * @param field
   *          is the structure field to add to the tagged structure
   * @param fieldName
   *          is the name of the structure field
   */
  public void addField(TTaggedStructure field, String fieldName)
  {
    if (fieldName == null) fieldName = "field-"+nextFieldIndex;
    fieldName = Str.toAscii(fieldName);
    String cname = field != null ? field.getName() : null;
    String ttname = fieldName.contains(">") ? fieldName.substring(1, fieldName.indexOf(">")) : null;
    String stag = ttname != null ? ttname : cname;
    int size = TStructRegistry.getSizeInBytes(stag);
    if (size < 1)
    {
      MsgLog.log("TTaggedStructure", "---Error: attempt to register a field: " + fieldName
          + " of unknown data type: " + field.getName(), TErrorList.code_failure, null, 0);
      return;
    }
    if (field != null)
    {
      if (field.fieldArrayList.size() == 0)
        field.addFieldsFrom(TStructRegistry.get(stag));
    }
    else
    {
      field = new TTaggedStructure(stag,true);
    }
    field.sizeInBytes = size;
    fieldArrayList.put(fieldName,new TaggedStructureField(field, fieldName));
    if (!registered)
    {
      String fieldNameTag = "<" + field.getName() + ">" + fieldName;
      TStructRegistry.fill(structName, fieldNameTag, TFormat.CF_STRUCT, 1);
      sizeInBytes += size;
      nextFieldIndex++;
    }
    chkNewField(fieldName);
  }
  /**
   * @param field
   *          is the structure field to add to the tagged structure
   * @param fieldName
   *          is the name of the structure field
   */
  public void addField(TTaggedStructure[] field, String fieldName)
  {
    if (fieldName == null) fieldName = "field-"+nextFieldIndex;
    fieldName = Str.toAscii(fieldName);
    String cname = field[0] != null ? field[0].getName() : null;
    String ttname = fieldName.contains(">") ? fieldName.substring(1, fieldName.indexOf(">")) : null;
    String stag = ttname != null ? ttname : cname;
    int size = TStructRegistry.getSizeInBytes(stag);
    if (size < 1)
    {
      MsgLog.log("TTaggedStructure", "attempt to register a field: " + fieldName
          + " being an array of unknown data type: " + field[0].getName(), TErrorList.code_failure, null, 0);
      return;
    }
    for (int i=0; i<field.length; i++)
    {
      if (field[i] != null)
      {
        if (field[i].fieldArrayList.size() == 0)
          field[i].addFieldsFrom(TStructRegistry.get(stag));
      }
      else
      {
        field[i] = new TTaggedStructure(stag,true);
      }
      field[i].sizeInBytes = size;
    }
    fieldArrayList.put(fieldName,new TaggedStructureArrayField(field, fieldName));
    if (!registered)
    {
      String fieldNameTag = fieldName.contains("<") ? fieldName : "<" + field[0].getName() + ">" + fieldName;
      TStructRegistry.fill(structName, fieldNameTag, TFormat.CF_STRUCT, field.length);
      sizeInBytes += size * field.length;
      nextFieldIndex++;
    }
    chkNewField(fieldName);
  }
  /*
   * General Byte Alignment Rules -------Taken from:
   * http://www.eventhelix.com/RealtimeMantra/ByteAlignmentAndOrdering.htm The
   * following byte padding rules will generally work with most 32 bit
   * processor. You should consult your compiler and microprocessor manuals to
   * see if you can relax any of these rules. - Single byte numbers can be
   * aligned at any address - Two byte numbers should be aligned to a two byte
   * boundary - Four byte numbers should be aligned to a four byte boundary -
   * Structures between 1 and 4 bytes of data should be padded so that the total
   * structure is 4 bytes. - Structures between 5 and 8 bytes of data should be
   * padded so that the total structure is 8 bytes. - Structures between 9 and
   * 16 bytes of data should be padded so that the total structure is 16 bytes.
   * - Structures greater than 16 bytes should be padded to 16 byte boundary.
   * Karol's comment: it seems not to work for Linux GCC... For now I decide to
   * limit the padding to 4 bytes
   */
  // private void checkPadding()
  // {
  // return;
  // // if( offset < 1 ) return;
  // //
  // // int n = offset % 4;
  // // if( n == 0 ) return;
  // // addField( new byte[4-n], "reserved");
  // }
  /*
   * private void checkPadding_not_working() { if( offset < 1 ) return; if(
   * offset < 4 ) { addField( new byte[ 4 - offset ], "reserved"); return; } if(
   * offset < 8 ) { addField( new byte[ 8 - offset ], "reserved"); return; } if(
   * offset < 16 ) { addField( new byte[ 16 - offset ], "reserved"); return; }
   * int n = offset % 16; if( n == 0 ) return; addField( new byte[16-n],
   * "reserved"); }
   */
  /*
   * old version: private int padding = 0; private void checkAligment(int size)
   * { //check the size for the correct alignment. Add some placeholders if
   * needed int d = size % 4; int fieldPadding = 4 - d; if( d > 0 ) { // there
   * are extra padding bytes if (padding > 0 && d > padding) { // too many bytes
   * to append to current padding set so clear it padding = fieldPadding;
   * addField( new byte[padding], "reserved"); } if (padding == 0) padding = 4;
   * padding -= d; } else { // new field is aligned; are there padding bytes? if
   * (padding > 0 && padding < 4) { padding = fieldPadding; addField( new
   * byte[padding], "reserved"); } } }
   */
  // ----------------------------------------------
  public abstract class structField
  {
    private String name;
    private Object value;
    protected int arrayIndex;
    private String arraySeparator = ", ";
    private String fieldSeparator = ", ";
    public void setArraySeparator(String value)
    {
      if (value != null && value.length() > 0) arraySeparator = value;
    }
    public void setFieldSeparator(String value)
    {
      if (value != null && value.length() > 0) fieldSeparator = value;
    }
    public String getName()
    {
      return name;
    }
    public void setName(String name)
    {
      this.name = name;
    }
    public Object getValue()
    {
      return value;
    }
    public void setValue(Object value)
    {
      this.value = value;
    }
    public String getValueAsString()
    {
      if (value == null) return "(null)";
      if (value instanceof char[])
        return new String((char[])value);
      int len = Array.getLength(value);
      StringBuilder sb = new StringBuilder();
      for (int i=0; i<len; i++)
      {
        if (value instanceof short[])
          sb.append(Short.toString(((short[])value)[i]));
        else if (value instanceof int[])
          sb.append(Integer.toString(((int[])value)[i]));
        else if (value instanceof long[])
          sb.append(Long.toString(((long[])value)[i]));
        else if (value instanceof float[])
          sb.append(Float.toString(((float[])value)[i]));
        else if (value instanceof double[])
          sb.append(Double.toString(((double[])value)[i]));
        else if (value instanceof boolean[])
          sb.append(((boolean[])value)[i]);
        else if (value instanceof byte[])
          sb.append(((byte[])value)[i]);
        else if (value instanceof String[])
          sb.append(((String[])value)[i]);
        else if (value instanceof TCompoundDataObject[])
        {
          ((TCompoundDataObject[])value)[i].setFieldSeparator(fieldSeparator);
          sb.append(((TCompoundDataObject[])value)[i]);
        }
        else
          return value.toString();
        if (i < len-1) sb.append(arraySeparator);
      }
      return sb.toString();
    }
    public int getArrayIndex()
    {
      return arrayIndex;
    }
    abstract void writeToOutputStream(DataOutputStream ds) throws IOException;
    abstract void readFromInputStream(DataInputStream ds) throws IOException;
    abstract void fixExtendedBytes();
    abstract void readExtendedBytes(byte[] bytes) throws IOException;
    abstract void writeExtendedBytes(DataOutputStream ds) throws IOException;
    @Override
    abstract public String toString();
  }
  private class byteArrayField extends structField
  {
    byte[] data;
    byteArrayField(byte[] data, String name)
    {
      setName(name);
      setValue(data);
      this.data = data;
      arrayIndex = nextFieldIndex;
    }
    @Override
    void readFromInputStream(DataInputStream ds) throws IOException
    {
      ds.read(data, 0, data.length);
    }
    @Override
    void writeToOutputStream(DataOutputStream ds) throws IOException
    {
      ds.write(data, 0, data.length);
    }
    @Override
    public String toString()
    {
      StringBuffer sb = new StringBuffer("");
      int ibyte;
      for (int i = 0; i < data.length; i++)
      {
        if ((ibyte = data[i]) < 0) ibyte += 256;
        sb.append((i == 0 ? " " : arraySeparator) + ibyte);
      }
      return sb.toString();
    }
    @Override
    void fixExtendedBytes()
    {
    }
    @Override
    void readExtendedBytes(byte[] bytes) throws IOException
    {
    }
    @Override
    void writeExtendedBytes(DataOutputStream ds) throws IOException
    {
    }
  }
  private class charArrayField extends structField
  {
    char[] data;
    charArrayField(char[] data, String name)
    {
      setName(name);
      setValue(data);
      this.data = data;
      arrayIndex = nextFieldIndex;
    }
    @Override
    void readFromInputStream(DataInputStream ds) throws IOException
    {
      for (int i = 0; i < data.length; i++)
      {
        data[i] = (char) (ds.readByte());
      }
    }
    @Override
    void writeToOutputStream(DataOutputStream ds) throws IOException
    {
      for (int i = 0; i < data.length; i++)
      {
        ds.writeByte((byte) (data[i]));
      }
    }
    @Override
    public String toString()
    {
      StringBuffer sb = new StringBuffer("");
      byte[] bdata = new byte[data.length];
      for (int i=0; i<data.length; i++) bdata[i] = (byte)data[i];
      sb.append(Str.trim(new String(bdata)));
      return sb.toString().trim();
    }
    @Override
    void fixExtendedBytes()
    {
    }
    @Override
    void readExtendedBytes(byte[] bytes) throws IOException
    {
    }
    @Override
    void writeExtendedBytes(DataOutputStream ds) throws IOException
    {
    }
  }
  private class shortArrayField extends structField
  {
    boolean isUnsigned = false;
    short[] data;
    shortArrayField(short[] data, String name)
    {
      setName(name);
      setValue(data);
      this.data = data;
      arrayIndex = nextFieldIndex;
    }
    shortArrayField(short[] data, String name,boolean unsigned)
    {
      setName(name);
      setValue(data);
      this.data = data;
      isUnsigned = unsigned;
      arrayIndex = nextFieldIndex;
    }
    @Override
    void readFromInputStream(DataInputStream ds) throws IOException
    {
      for (int i = 0; i < data.length; i++)
      {
        data[i] = Swap.Short(ds.readShort());
      }
    }
    @Override
    void writeToOutputStream(DataOutputStream ds) throws IOException
    {
      for (int i = 0; i < data.length; i++)
      {
        ds.writeShort(Swap.Short(data[i]));
      }
    }
    @Override
    public String toString()
    {
      StringBuffer sb = new StringBuffer("");
      int ival;
      for (int i = 0; i < data.length; i++)
      {
        ival = isUnsigned ? UnsignedUtils.toUnsignedShort(data[i]) : data[i];
        sb.append((i == 0 ? " " : arraySeparator) + ival);
      }
      return sb.toString();
    }
    @Override
    void fixExtendedBytes()
    {
    }
    @Override
    void readExtendedBytes(byte[] bytes) throws IOException
    {
    }
    @Override
    void writeExtendedBytes(DataOutputStream ds) throws IOException
    {
    }
  }
  private class boolArrayField extends structField
  {
    boolean[] data;
    @SuppressWarnings("unused")
    boolArrayField(int[] data, String name)
    {
      setName(name);
      setValue(data);
      this.data = new boolean[data.length];
      arrayIndex = nextFieldIndex;
      for (int i=0; i<data.length; i++) this.data[i] = data[i] != 0;
    }
    boolArrayField(boolean[] data, String name)
    {
      setName(name);
      setValue(data);
      this.data = data;
      arrayIndex = nextFieldIndex;
    }
    @Override
    void writeToOutputStream(DataOutputStream ds) throws IOException
    {
      for (int i = 0; i < data.length; i++)
      {
        ds.writeInt(Swap.Int((int)(data[i] ? 1 : 0)));
      }      
    }

    @Override
    void readFromInputStream(DataInputStream ds) throws IOException
    {
      for (int i = 0; i < data.length; i++)
      {
        data[i] = ds.readInt() != 0;
      }      
    }

    @Override
    void fixExtendedBytes()
    {
    }

    @Override
    void readExtendedBytes(byte[] bytes) throws IOException
    {
    }

    @Override
    void writeExtendedBytes(DataOutputStream ds) throws IOException
    {
    }

    @Override
    public String toString()
    {
      StringBuffer sb = new StringBuffer("");
      for (int i = 0; i < data.length; i++)
      {
        sb.append((i == 0 ? " " : arraySeparator) + data[i]);
      }
      return sb.toString();
    }
    
  }
  private class intArrayField extends structField
  {
    int[] data;
    boolean isUnsigned = false;
    @SuppressWarnings("unused")
    intArrayField(int[] data, String name)
    {
      setName(name);
      setValue(data);
      this.data = data;
      arrayIndex = nextFieldIndex;
    }
    intArrayField(int[] data, String name,boolean unsigned)
    {
      setName(name);
      setValue(data);
      this.data = data;
      isUnsigned = unsigned;
      arrayIndex = nextFieldIndex;
    }
    @Override
    void readFromInputStream(DataInputStream ds) throws IOException
    {
      for (int i = 0; i < data.length; i++)
      {
        data[i] = Swap.Int(ds.readInt());
      }
    }
    @Override
    void writeToOutputStream(DataOutputStream ds) throws IOException
    {
      for (int i = 0; i < data.length; i++)
      {
        ds.writeInt(Swap.Int(data[i]));
      }
    }
    @Override
    public String toString()
    {
      StringBuffer sb = new StringBuffer("");
      long lval;
      for (int i = 0; i < data.length; i++)
      {
        lval = isUnsigned ? UnsignedUtils.toUnsignedInt(data[i]) : data[i];
        sb.append((i == 0 ? " " : arraySeparator) + lval);
      }
      return sb.toString();
    }
    @Override
    void fixExtendedBytes()
    {
    }
    @Override
    void readExtendedBytes(byte[] bytes) throws IOException
    {
    }
    @Override
    void writeExtendedBytes(DataOutputStream ds) throws IOException
    {
    }
  }
  private class floatArrayField extends structField
  {
    float[] data;
    floatArrayField(float[] data, String name)
    {
      setName(name);
      setValue(data);
      this.data = data;
      arrayIndex = nextFieldIndex;
    }
    @Override
    void readFromInputStream(DataInputStream ds) throws IOException
    {
      byte[] buffer = new byte[4];
      for (int i = 0; i < data.length; i++)
      {
        ds.read(buffer);
        data[i] = Swap.Float(buffer);
      }
    }
    @Override
    void writeToOutputStream(DataOutputStream ds) throws IOException
    {
      for (int i = 0; i < data.length; i++)
      {
        ds.write(Swap.Float(data[i]));
      }
    }
    @Override
    public String toString()
    {
      StringBuffer sb = new StringBuffer("");
      for (int i = 0; i < data.length; i++)
      {
        sb.append((i == 0 ? " " : arraySeparator) + data[i]);
      }
      return sb.toString();
    }
    @Override
    void fixExtendedBytes()
    {
    }
    @Override
    void readExtendedBytes(byte[] bytes) throws IOException
    {
    }
    @Override
    void writeExtendedBytes(DataOutputStream ds) throws IOException
    {
    }
  }
  private class doubleArrayField extends structField
  {
    double[] data;
    doubleArrayField(double[] data, String name)
    {
      setName(name);
      setValue(data);
      this.data = data;
      arrayIndex = nextFieldIndex;
    }
    @Override
    void readFromInputStream(DataInputStream ds) throws IOException
    {
      byte[] buffer = new byte[8];
      for (int i = 0; i < data.length; i++)
      {
        ds.read(buffer);
        data[i] = Swap.Double(buffer);
      }
    }
    @Override
    void writeToOutputStream(DataOutputStream ds) throws IOException
    {
      for (int i = 0; i < data.length; i++)
      {
        ds.write(Swap.Double(data[i]));
      }
    }
    @Override
    public String toString()
    {
      StringBuffer sb = new StringBuffer("");
      for (int i = 0; i < data.length; i++)
      {
        sb.append((i == 0 ? " " : arraySeparator) + data[i]);
      }
      return sb.toString();
    }
    @Override
    void fixExtendedBytes()
    {
    }
    @Override
    void readExtendedBytes(byte[] bytes) throws IOException
    {
    }
    @Override
    void writeExtendedBytes(DataOutputStream ds) throws IOException
    {
    }
  }
  private class longArrayField extends structField
  {
    long[] data;
    longArrayField(long[] data, String name)
    {
      setName(name);
      setValue(data);
      this.data = data;
      arrayIndex = nextFieldIndex;
    }
    @Override
    void readFromInputStream(DataInputStream ds) throws IOException
    {
      byte[] buffer = new byte[8];
      for (int i = 0; i < data.length; i++)
      {
        ds.read(buffer);
        data[i] = Swap.Long(buffer);
      }
    }
    @Override
    void writeToOutputStream(DataOutputStream ds) throws IOException
    {
      for (int i = 0; i < data.length; i++)
      {
        ds.writeLong(Swap.Long(data[i]));
      }
    }
    @Override
    public String toString()
    {
      StringBuffer sb = new StringBuffer("");
      for (int i = 0; i < data.length; i++)
      {
        sb.append((i == 0 ? " " : arraySeparator) + data[i]);
      }
      return sb.toString();
    }
    @Override
    void fixExtendedBytes()
    {
    }
    @Override
    void readExtendedBytes(byte[] bytes) throws IOException
    {
    }
    @Override
    void writeExtendedBytes(DataOutputStream ds) throws IOException
    {
    }
  }
  private class TStringField extends structField
  {
    TString data;
    TStringField(TString data, String name)
    {
      setName(name);
      setValue(data);
      this.data = data;
      arrayIndex = nextFieldIndex;
    }
    @Override
    void readFromInputStream(DataInputStream ds) throws IOException
    {
      byte[] b = new byte[data.getCapacity()];
      ds.read(b);
      data.set(b);
    }
    @Override
    void writeToOutputStream(DataOutputStream ds) throws IOException
    {
      ds.write(data.getBytes());
    }
    @Override
    public String toString()
    {
      return data.toString();
    }
    @Override
    void fixExtendedBytes()
    {
    }
    @Override
    void readExtendedBytes(byte[] bytes) throws IOException
    {
    }
    @Override
    void writeExtendedBytes(DataOutputStream ds) throws IOException
    {
    }
  }
  private class StringBufferField extends structField
  {
    StringBuffer data;
    StringBufferField(StringBuffer data, String name)
    {
      setName(name);
      setValue(data);
      this.data = data;
      arrayIndex = nextFieldIndex;
    }
    @Override
    void readFromInputStream(DataInputStream ds) throws IOException
    {
      for (int i = 0; i < data.capacity(); i++)
        data.setCharAt(i, (char) ds.readByte());
    }
    @Override
    void writeToOutputStream(DataOutputStream ds) throws IOException
    {
      for (int i = 0; i < data.capacity(); i++)
        ds.writeByte((byte) (data.charAt(i)));
    }
    @Override
    public String toString()
    {
      return data.toString();
    }
    @Override
    void fixExtendedBytes()
    {
    }
    @Override
    void readExtendedBytes(byte[] bytes) throws IOException
    {
    }
    @Override
    void writeExtendedBytes(DataOutputStream ds) throws IOException
    {
    }
  }
  private class StringArrayField extends structField
  {
    String[] data;
    int[] xoff;
    int[] xsiz;
    StringArrayField(String[] data, String name)
    {
      setName(name);
      setValue(data);
      this.data = data;
      arrayIndex = nextFieldIndex;
      if (data == null) return;
      xoff = new int[data.length];
      xsiz = new int[data.length];
      hasExtendedSpace = true;
    }
    @Override
    void readFromInputStream(DataInputStream ds) throws IOException
    {
      if (data == null) return;
      for (int i = 0; i < data.length; i++)
      {
        xoff[i] = Swap.Int(ds.readInt());
        xsiz[i] = Swap.Int(ds.readInt());
      }
      int xo = getExtendedOffset();
      if (xo == 0 || xoff[0] < xo) setExtendedOffset(xoff[0]);
    }
    void readExtendedBytes(byte[] bytes)
    {
      if (data == null) return;
      for (int i = 0; i<data.length; i++)
      {
        data[i] = new String(bytes,xoff[i],xsiz[i]-1);
      }
    }
    @Override
    void writeToOutputStream(DataOutputStream ds) throws IOException
    {
      if (data == null) return;
      for (int i = 0; i < data.length; i++)
      {
        ds.writeInt(Swap.Int(xoff[i]));
        ds.writeInt(Swap.Int(xsiz[i]));
      }
    }
    void fixExtendedBytes()
    {
      int slen;
      if (data == null) return;
      for (int i=0; i<data.length; i++)
      {
        xoff[i] = extendedOffset + extendedSizeInBytes;
        slen = data[i] == null ? 0 : data[i].length() + 1;
        xsiz[i] = slen;
        extendedSizeInBytes += xsiz[i];
      }
    }
    void writeExtendedBytes(DataOutputStream ds) throws IOException
    { // must be called in proper order ...
      if (data == null) return;
      for (int i = 0; i < data.length; i++)
      {
        if (data[i] != null) ds.write(data[i].getBytes());
        ds.write(0);
      }
    }
    @Override
    public String toString()
    {
//      String indentString = "\n    "; 
//      for (int i=0; i < getName().length(); i++)
//        indentString += " ";
      StringBuffer sb = new StringBuffer("");
      if (data == null) return "null";
      for (int i = 0; i < data.length; i++)
      {
        //if (i > 0) sb.append(indentString + " ->");        
        sb.append((i == 0 ? " " : arraySeparator)+data[i]);
      }
      return sb.toString();
    }
  }
  private class TCompoundDataField extends structField
  {
    TCompoundDataObject data;
    TCompoundDataField(TCompoundDataObject data, String name)
    {
      setName(name);
      setValue(data);
      this.data = data;
      arrayIndex = nextFieldIndex;
   }
    @Override
    void readFromInputStream(DataInputStream ds) throws IOException
    {
      if (data instanceof IMAGE)
      {
        ((IMAGE) data).read(ds);
      }
      else
      {
        ds.read(data.getByteArray());
        data.toStruct();
      }
    }
    @Override
    void writeToOutputStream(DataOutputStream ds) throws IOException
    {
      ds.write(data.toByteArray());
    }
    @Override
    public String toString()
    {
      return data.toString();
    }
    @Override
    void fixExtendedBytes()
    {
    }
    @Override
    void readExtendedBytes(byte[] bytes) throws IOException
    {
    }
    @Override
    void writeExtendedBytes(DataOutputStream ds) throws IOException
    {
    }
  }
  private class TCompoundDataArrayField extends structField
  {
    TCompoundDataObject[] data;
    TCompoundDataArrayField(TCompoundDataObject[] data, String name)
    {
      setName(name);
      this.data = data;
      arrayIndex = nextFieldIndex;
      if (data == null) return;
      Class<?> t = data.getClass().getComponentType();
      if (t == null) return;
      if (t == TCompoundDataObject.class && data[0] == null) return;
      if (data[0] == null)
      try
      {
        data[0] = (TCompoundDataObject) t.getConstructor().newInstance();
      } catch (Exception e)
      {
        e.printStackTrace();
      }
      for (int i=1; i<data.length; i++)
      { // make sure something is instantiated
        if (data[i] == null) data[i] = data[0].clone();
      }
      setValue(data);
    }
    TCompoundDataArrayField(TCompoundDataObject[] data, short format, int size, String tag, String name)
    {
      setName(name);
      if (data[0] != null)
      {
        this.data = data;
        for (int i=1; i<data.length; i++)
        { // make sure something is instantiated
          if (data[i] == null)
          { // wasn't there -> make one
            try
            {
              data[i] = data[0].newInstance();
            }
            catch (Exception e)
            {
              e.printStackTrace();
              MsgLog.log("makeNewCompoundDataObject", e.toString(), TErrorList.code_failure, e, 0);
            }        
          }
        }
        setValue(data);
        return;
      }
      this.data = TFormat.makeCompoundDataObjectArray(format, tag, size);
      setValue(data);
    }
    @Override
    void readFromInputStream(DataInputStream ds) throws IOException
    {
      for (TCompoundDataObject obj : data)
      {
        if (obj instanceof IMAGE)
        {
          ((IMAGE) obj).read(ds);
        }
        else
        {
          ds.read(obj.getByteArray());
          obj.toStruct();
        }
      }
    }
    @Override
    void writeToOutputStream(DataOutputStream ds) throws IOException
    {
      for (int i = 0; i < data.length; i++)
        ds.write(data[i].toByteArray());
    }
    @Override
    public String toString()
    {
      StringBuffer sb = new StringBuffer("");
      for (int i = 0; i < data.length; i++)
        sb.append((i == 0 ? " " : arraySeparator) + data[i]);
      return sb.toString();
    }
    @Override
    void fixExtendedBytes()
    {
    }
    @Override
    void readExtendedBytes(byte[] bytes) throws IOException
    {
    }
    @Override
    void writeExtendedBytes(DataOutputStream ds) throws IOException
    {
    }
  }
  public class TaggedStructureField extends structField
  {
    TTaggedStructure data;
    TaggedStructureField(TTaggedStructure data, String name)
    {
      setName(name);
      setValue(data);
      this.data = data;
      arrayIndex = nextFieldIndex;
    }
    @Override
    void readFromInputStream(DataInputStream ds) throws IOException
    {
      byte[] buffer = new byte[data.getSizeInBytes()];
      ds.read(buffer);
      data.toStruct(buffer);
    }
    @Override
    void writeToOutputStream(DataOutputStream ds) throws IOException
    {
      ds.write(data.toByteArray());
    }
    @Override
    public String toString()
    {
      return data.toString();
    }
    @Override
    void fixExtendedBytes()
    {
    }
    @Override
    void readExtendedBytes(byte[] bytes) throws IOException
    {
    }
    @Override
    void writeExtendedBytes(DataOutputStream ds) throws IOException
    {
    }
  }
  private class TaggedStructureArrayField extends structField
  {
    TTaggedStructure[] data;
    TaggedStructureArrayField(TTaggedStructure[] data, String name)
    {
      setName(name);
      this.data = data;
      arrayIndex = nextFieldIndex;
      String stag = name.startsWith("<") ? name.substring(1, name.indexOf('>')) : name;
      for (int i=0; i<data.length; i++)
      {
        if (data[i] == null) data[i] = new TTaggedStructure(stag, true);
      }
      setValue(data);
    }
    @Override
    void readFromInputStream(DataInputStream ds) throws IOException
    {
      byte[] buffer = new byte[data[0].getSizeInBytes()];
      for (int i = 0; i < data.length; i++)
      {
        ds.read(buffer);
        data[i].toStruct(buffer);
      }
    }
    @Override
    void writeToOutputStream(DataOutputStream ds) throws IOException
    {
      byte[] b;
      for (int i = 0; i < data.length; i++)
      {
        b = data[i] != null ? data[i].toByteArray() : new byte[sizeInBytes];
        ds.write(b);
      }
    }
    @Override
    public String toString()
    {
      StringBuffer sb = new StringBuffer("");
      for (int i = 0; i < data.length; i++)
        sb.append((i == 0 ? " " : arraySeparator) + data[i]);
      return sb.toString();
    }
    @Override
    void fixExtendedBytes()
    {
    }
    @Override
    void readExtendedBytes(byte[] bytes) throws IOException
    {
    }
    @Override
    void writeExtendedBytes(DataOutputStream ds) throws IOException
    {
    }
  }
  private class imageArrayField extends structField
  {
    int[] xoff;
    int[] xsiz;
    IMAGE[] data;
    imageArrayField(IMAGE[] data, String name)
    {
      setName(name);
      this.data = data;
      arrayIndex = nextFieldIndex;
      if (data == null) return;
      for (int i=0; i<data.length; i++)
      {
        if (data[i] == null) data[i] = new IMAGE(0x10000);
      }
      xoff = new int[data.length];
      xsiz = new int[data.length];
      hasExtendedSpace = true;
      setValue(data);
    }
    void readFromInputStream(DataInputStream ds) throws IOException
    {
      if (data == null) return;
      for (int i = 0; i < data.length; i++)
      {
        data[i].readHeaders(ds);
        xoff[i] = Swap.Int(ds.readInt());
        xsiz[i] = Swap.Int(ds.readInt());
      }
      int xo = getExtendedOffset();
      if (xo == 0 || xoff[0] < xo) setExtendedOffset(xoff[0]);
    }
    void readExtendedBytes(byte[] bytes)
    {
      if (data == null) return;
      for (int i = 0; i<data.length; i++)
      {
        if (xsiz[i] + xoff[i] > bytes.length) 
        {
          MsgLog.log("TTaggedStructure.readExtendedBytes", "extended byte information is corrupt",TErrorList.code_failure,null,1);
          return;
        }
        data[i].ensureFrameBufferCapacity(xsiz[i]);
        data[i].fillFrameBuffer(bytes, xoff[i], 0, xsiz[i]);
      }
    }
    @Override
    void writeToOutputStream(DataOutputStream ds) throws IOException
    {
      if (data == null) return;
      for (int i = 0; i < data.length; i++)
      {
        data[i].writeHeaders(ds);
        ds.writeInt(Swap.Int(xoff[i]));
        ds.writeInt(Swap.Int(xsiz[i]));
      }
    }
    void fixExtendedBytes()
    {
      int slen;
      if (data == null) return;
      for (int i=0; i<data.length; i++)
      {
        xoff[i] = extendedOffset + extendedSizeInBytes;
        slen = data[i] == null ? 0 : data[i].getFrameHeader().appendedFrameSize;
        xsiz[i] = slen;
        extendedSizeInBytes += xsiz[i];
      }
    }
    void writeExtendedBytes(DataOutputStream ds) throws IOException
    { // must be called in proper order ...
      if (data == null) return;
      for (int i = 0; i < data.length; i++)
      {
        ds.write(data[i].getImageFrameBuffer(),0,xsiz[i]);
      }
    }
    @Override
    public String toString()
    {
      StringBuffer sb = new StringBuffer("");
      if (data == null) return "null";
      for (int i = 0; i < data.length; i++)
      {
        sb.append((i == 0 ? " " : arraySeparator) + data[i].toString());
      }
      return sb.toString();
    }
    
  }
  private class spectrumArrayField extends structField
  {
    int[] xoff;
    int[] xsiz;
    ASPECTRUM[] data;
    spectrumArrayField(ASPECTRUM[] data, String name)
    {
      setName(name);
      this.data = data;
      arrayIndex = nextFieldIndex;
      if (data == null) return;
      for (int i=0; i<data.length; i++)
      {
        if (data[i] == null) data[i] = new ASPECTRUM(TFormat.SPECTRUM_STD_SIZE);
      }
      xoff = new int[data.length];
      xsiz = new int[data.length];
      hasExtendedSpace = true;
      setValue(data);
    }
    void readFromInputStream(DataInputStream ds) throws IOException
    {
      if (data == null) return;
      for (int i = 0; i < data.length; i++)
      {
        ds.read(data[i].getByteArray());
        data[i].toStruct();
        xoff[i] = Swap.Int(ds.readInt());
        xsiz[i] = Swap.Int(ds.readInt());
      }
      int xo = getExtendedOffset();
      if (xo == 0 || xoff[0] < xo) setExtendedOffset(xoff[0]);
    }
    void readExtendedBytes(byte[] bytes)
    {
      byte[] ba;
      if (data == null) return;
      for (int i = 0; i<data.length; i++)
      {
        if (xsiz[i] + xoff[i] > bytes.length) 
        {
          MsgLog.log("TTaggedStructure.readExtendedBytes", "extended byte information is corrupt",TErrorList.code_failure,null,1);
          return;
        }
        ba = new byte[xsiz[i]];
        System.arraycopy(bytes, xoff[i], ba, 0, xsiz[i]);
        data[i].setSpectrumByteArray(ba);
        data[i].toSpectrumStruct();
      }
    }
    @Override
    void writeToOutputStream(DataOutputStream ds) throws IOException
    {
      if (data == null) return;
      for (int i = 0; i < data.length; i++)
      {
        ds.write(data[i].toByteArray());
        ds.writeInt(Swap.Int(xoff[i]));
        ds.writeInt(Swap.Int(xsiz[i]));
      }
    }
    void fixExtendedBytes()
    {
      int slen;
      if (data == null) return;
      for (int i=0; i<data.length; i++)
      {
        xoff[i] = extendedOffset + extendedSizeInBytes;
        slen = data[i] == null ? TFormat.SPECTRUM_STD_SIZE : data[i].getLength() * 4;
        xsiz[i] = slen;
        extendedSizeInBytes += xsiz[i];
      }
    }
    void writeExtendedBytes(DataOutputStream ds) throws IOException
    { // must be called in proper order ...
      if (data == null) return;
      for (int i = 0; i < data.length; i++)
      {
        ds.write(data[i].toSpectrumByteArray(),0,xsiz[i]);
      }
    }
    @Override
    public String toString()
    {
      StringBuffer sb = new StringBuffer("");
      if (data == null) return "null";
      for (int i = 0; i < data.length; i++)
      {
        sb.append((i == 0 ? " " : arraySeparator) + data[i].toString());
      }
      return sb.toString();
    }
  }
  public class FieldValue
  {
    private String field;
    private Object value;
    public FieldValue(String field, Object value)
    {
      this.field = field;
      this.value = value;
    }
    public void setField(String field)
    {
      this.field = field;
    }
    public void setValue(Object value)
    {
      this.value = value;
    }
    public String getField() { return field; }
    public Object getValue() { return value; }
  }
}
