/*
 * File TTrace.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.traceUtils;

import java.io.IOException;
import javax.naming.NamingException;
import de.desy.tine.client.TLink;
import de.desy.tine.client.TLinkCallback;
import de.desy.tine.dataUtils.TDataType;
import de.desy.tine.definitions.*;
import de.desy.tine.queryUtils.TQuery;
import de.desy.tine.server.logger.DbgLog;

public class TTrace implements TLinkCallback
{
  private String[] tracePrps = null;
  private boolean hasXAxis = false;
  private boolean hasReference = false;
  private boolean hasSequence = false;
  private boolean hasAverage = false;
  private boolean isScheduled = false;
  private String schTag = ".SCH";
  private String ctx;
  private String srv;
  private String prp;
  private String dev = "#0";
  private int length;
  private int nseq;
  private TTraceHeader[] hdr = new TTraceHeader[1];
  private float[] xaxis = null;
  private float[] ref = null;
  private float[] trace = null;
  private float[][] seq = null;
  private TTraceHandler hndlr = null;
  private TLink lnk = null;
  private TLink reflnk = null;
  private TLink seqlnk = null;
  private TLink hdrlnk = null;
  private int traceStatus = 0;
  public int getTraceStatus() { return traceStatus; }
  private String traceStatusString = null;
  public String getTraceStatusString()
  {
    return  traceStatusString == null ? "not active" : traceStatusString;  
  }
  private long timestamp = 0;
  private long[] refTimestamp = new long[1];
  public void setReferenceTrace(float[] reference)
  {
    ref = reference;
    refTimestamp[0] = System.currentTimeMillis();
  }
  public void setReferenceTrace(float[] reference,long referenceTimestamp)
  {
    ref = reference;
    refTimestamp[0] = referenceTimestamp;
  }
  public long getReferenceTimestamp()
  {
    if (reflnk != null && reflnk.linkStatus == 0) 
    {
      long ts = reflnk.getLastTimeStamp();
      if (ts > 0) refTimestamp[0] = ts;
    }
    return refTimestamp[0];
  }
  public void setReferenceTimestamp(long referenceTimestamp)
  {
    refTimestamp[0] = referenceTimestamp;
  }
  private String getUndecoratedTraceProperty(String prop)
  {
    if (prop != null && prop.endsWith(".AVE")) 
      return prop.substring(0, prop.lastIndexOf(".AVE"));
    return prop;
  }
  private synchronized float[] getPropertyData(String property,String message,long[] timeStamp) throws IOException
  {
    float[] y = new float[length];
    TDataType yt = new TDataType(y);
    String devName = "/"+ctx+"/"+srv+"/"+dev;
    TLink tl = new TLink(devName,property,yt,null,TAccess.CA_READ);
    int cc = tl.execute(500, true);
    if (timeStamp != null) timeStamp[0] = tl.getLastTimeStamp();
    tl.close();
    traceStatus = cc;
    traceStatusString = tl.getLastError();
    if (cc == TErrorList.link_timeout || cc == TErrorList.connection_timeout)
    {
      throw new IOException("timeout retrieving " + message);
    }
    if (property.compareToIgnoreCase(prp) == 0) timestamp = yt.getDataTimeStamp();
    return y;    
  }
  private int getNumberInSequence()
  {
    if (!hasSequence) return 0;
    if (nseq > 0) return nseq;
    int[] n = new int[1];
    TDataType nt = new TDataType(n);
    String devName = "/"+ctx+"/"+srv+"/"+dev;
    TLink tl = new TLink(devName,getUndecoratedTraceProperty(prp)+".NSEQ",nt,null,TAccess.CA_READ);
    int cc = tl.execute(500, true);
    traceStatus = cc;
    traceStatusString = tl.getLastError();
    tl.close();
    if (cc == TErrorList.link_timeout || cc == TErrorList.connection_timeout)
    {
      return 0;
    }
    nseq = n[0];
    return nseq;
  }
  private void mapTraceToSequence(float[][] tseq,float[] t)
  {
    if (nseq == 0) return;
    if (tseq == null) tseq = new float[nseq][];
    for (int i=0; i<nseq; i++)
    {
      tseq[i] = new float[length]; 
      System.arraycopy(t, i*length, tseq[i], 0, length);
    }
    return;    
  }
  public float[][] getTraceSequence() throws IOException
  {
    if (seq != null) return seq;
    if (getNumberInSequence() == 0) return null;
    String devName = "/"+ctx+"/"+srv+"/"+dev;
    float[] y = new float[nseq*length];
    TDataType yt = new TDataType(y);
    TLink tl = new TLink(devName,getUndecoratedTraceProperty(prp)+".SEQ",yt,null,TAccess.CA_READ);
    int cc = tl.execute(500, true);
    traceStatus = cc;
    traceStatusString = tl.getLastError();
    tl.close();
    timestamp = yt.getDataTimeStamp();
    if (cc == TErrorList.link_timeout || cc == TErrorList.connection_timeout)
    {
      throw new IOException("timeout retrieving sequence");
    }
    float[][] tseq = null;
    mapTraceToSequence(tseq,y);
    return tseq;
  }
  public void clearReference() { ref = null; }
  public float[] getReference() throws IOException
  {
    if (!hasReference) return ref;
    if (ref != null) return ref;
    float[] r = getPropertyData(getUndecoratedTraceProperty(prp)+".REF","trace reference",refTimestamp);
    ref = r;
    return r;
  }
  public float[] getAverage() throws IOException
  {
    if (!hasAverage) return null;
    return getPropertyData(getUndecoratedTraceProperty(prp)+".AVE","trace average",null);
  }
  public float[] getTrace() throws IOException
  {
    if (trace != null && lnk != null) return trace;
    return getPropertyData(prp,"trace",null);
  }
  public long getTraceTimestamp()
  {
    if (trace != null && lnk != null) return lnk.getLastTimeStamp();
    return timestamp;
  }
  public void callback(TLink link)
  {
    if (link.linkStatus != 0)
      DbgLog.log("TTrace callback","link error : " + link.getLastError());
    if (link == lnk)
    { // let only the trace callback fire the update handler etc.
      traceStatus = link.getLinkStatus();
      traceStatusString = link.getLastError();
      if (hndlr != null) hndlr.update(this);
    }
    if (link == seqlnk)
    {
      float[] y = (float[])seqlnk.getOutputDataObject().getDataObject();
      mapTraceToSequence(seq,y);
    }
  }
  public void stopMonitor()
  {
    if (lnk != null) lnk.close();
    lnk = null;
    if (reflnk != null) reflnk.close();
    reflnk = null;
    if (seqlnk != null) seqlnk.close();
    seqlnk = null;
    if (hdrlnk != null) hdrlnk.close();
    hdrlnk = null;
  }
  public int monitorTrace(TTraceHandler myTraceHandler,boolean scheduled,int pollingRate)
  {
    if (trace == null) trace = new float[length];
    TDataType tracet = new TDataType(trace);
    String devName = "/"+ctx+"/"+srv+"/"+dev;
    int lid = -1;
    if (hasReference)
    {
      if (ref == null) ref = new float[length];
      TDataType traceref = new TDataType(ref);
      try
      { // the CM_WAIT below doesn't seem to always work so do this once up front!
        float[] r = getPropertyData(getUndecoratedTraceProperty(prp)+".REF","trace reference",refTimestamp);
        System.arraycopy(r, 0, ref, 0, length);
        refTimestamp[0] = reflnk.dOutput.dTimestamp;
      }
      catch (Exception e)  {};
      reflnk = new TLink(devName,getUndecoratedTraceProperty(prp)+".REF",traceref,null,TAccess.CA_READ);
      lid = reflnk.attach(TMode.CM_DATACHANGE|TMode.CM_WAIT, this,pollingRate);
      if (lid < 0) 
      {
        reflnk = null;
        ref = null;
      }
    }
    if (hasSequence)
    {
      int nseq = getNumberInSequence();
      if (nseq > 0)
      {
        if (seq == null) seq = new float[nseq][];
        float[] y = new float[nseq*length];
        TDataType yt = new TDataType(y);
        seqlnk = new TLink(devName,getUndecoratedTraceProperty(prp)+".SEQ",yt,null,TAccess.CA_READ);
        lid = seqlnk.attach(TMode.CM_TIMER, this,pollingRate);
        if (lid < 0) 
        {
          seqlnk = null;
          seq = null;
        }
      }
    }
    if (hdr[0] == null) hdr[0] = new TTraceHeader();
    TDataType hdrt = new TDataType(hdr);
    hdrlnk = new TLink(devName,getUndecoratedTraceProperty(prp)+".INFO",hdrt,null,TAccess.CA_READ);
    lid = hdrlnk.attach(TMode.CM_DATACHANGE, this,pollingRate);
    if (lid < 0) 
    {
      hdrlnk = null;
    }
      
    String property = prp;
    if (scheduled && isScheduled) property += schTag;
    lnk = new TLink(devName,property,tracet,null,TAccess.CA_READ);
    lid = lnk.attach(TMode.CM_TIMER, this,pollingRate);
    if (lid < 0) 
    {
      lnk = null;
      return -lid;
    }
    hndlr = myTraceHandler;
    return 0;
  }
  public int getPreTriggerPosition()
  {
    return hdr[0].getPreTrigger();
  }
  public float[] getXAxis()
  {
    if (xaxis != null) return xaxis;
    float[] x = new float[length];
    if (!hasXAxis)
    { // make the x-axis here
      float off = hdr[0].getXOffset();
      float inc = hdr[0].getXScaleFactor();
      for (int i=0; i<length; i++)
      {
        x[i] = off + i * inc;
      }
    }
    else
    { // get the x-axis from the server
      try
      {
        x = getPropertyData(prp+".XAXIS","trace x axis",null);
      }
      catch (IOException e)
      {
        return null;
      }
    }
    xaxis = x;
    return xaxis;
  }
  public String getXAxisUnits()
  {
    return hdr[0].getXUnits();
  }
  public String getYAxisUnits()
  {
    return hdr[0].getYUnits();
  }
  public float getTracePlotMaximum()
  {
    return hdr[0].getYPlotMaxmimum();
  }
  public float getTracePlotMinimum()
  {
    return hdr[0].getYPlotMinimum();
  }
  private boolean isTaggedWith(String property,String tag,int index)
  {
    if (property != null && property.endsWith(tag))
    { // so far, so good
      int i = property.indexOf(tag);
      if (i == index) return true;
    }
    return false;
  }
  public String[] getProperties(String devName, String devProperty)
  {
    if (tracePrps == null)
    {
      String prps = devProperty + ".*";
      try
      {
        tracePrps = TQuery.getPropertiesWithPattern(ctx, srv, dev, prps, 500);
      }
      catch (IOException e)
      {
        return null;
      }
      int pidx = devProperty.length();
      for (int i=0; i<tracePrps.length; i++)
      {
        if (isTaggedWith(tracePrps[i],".XAXIS",pidx)) hasXAxis = true;
        if (isTaggedWith(tracePrps[i],".SEQ",pidx)) hasSequence = true;
        if (isTaggedWith(tracePrps[i],".REF",pidx)) hasReference = true;
        if (isTaggedWith(tracePrps[i],".AVE",pidx)) hasAverage = true;
        if (isTaggedWith(tracePrps[i],".SCH",pidx)) 
        {
          isScheduled = true;
        }
        else if (isTaggedWith(tracePrps[i],".SCHED",pidx)) 
        {
          schTag = ".SCHED";
          isScheduled = true;
        }
      }
    }
    return tracePrps;
  }
  private void getInfoHeader() throws IOException
  {
    hdr[0] = new TTraceHeader();
    TDataType hdrt = new TDataType(hdr);
    String devName = "/"+ctx+"/"+srv+"/"+dev;
    TLink tl = new TLink(devName,getUndecoratedTraceProperty(prp)+".INFO",hdrt,null,TAccess.CA_READ);
    int cc = tl.execute(500, true);
    traceStatus = cc;
    traceStatusString = tl.getLastError();
    tl.close();
    if (cc == TErrorList.link_timeout || cc == TErrorList.connection_timeout)
    {
      throw new IOException("could not acquire trace header information");
    }
    length = hdr[0].getTraceLength();
  }
  public TTrace(String context, String server, String device, String property) throws NamingException, IOException
  {
    if (context == null) throw new NamingException("context not given");
    if (server == null) throw new NamingException("server name not given");
    if (property == null || property.length() == 0) property = "Trace";
    if (device == null) device = "#0";
    ctx = context;
    srv = server;
    prp = property;
    dev = device;
    
    getInfoHeader();
    getProperties(device, property);
  }
  public TTrace(String context, String server, String device) throws NamingException, IOException
  {
    this(context,server,device,null);
  }
  public TTrace(String context, String server) throws NamingException, IOException
  {
    this(context,server,null,null);
  }
}
