/*
 * File DBLDBLDBL.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.types;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

import de.desy.tine.definitions.TErrorList;
import de.desy.tine.definitions.TFormat;
import de.desy.tine.endianUtils.Swap;
import de.desy.tine.server.logger.MsgLog;

public class DBLDBLDBL implements TCompoundDataObject
{
  public double d1val;
  public double d2val;
  public double d3val;
  public double getDoubleValue() { return (double)d1val; }
  public int getIntegerValue() { return (int)d2val; }
  public static final int sizeInBytes = 24;
  public static final short dFormat = TFormat.CF_DBLDBLDBL;
  public void setValues(double newD1Value, double newD2Value, double newD3Value)
  {
    d1val = newD1Value; d2val = newD2Value; d3val = newD3Value;
  }
  public void setByteArray(byte[] byteArray) { hByteArray = byteArray; }
  public int getSizeInBytes() { return sizeInBytes; }
  public short getFormat() { return dFormat; }    
  public String getTag() { return null; }
  ByteArrayOutputStream dBuffer = new ByteArrayOutputStream(sizeInBytes);
  public byte[] hByteArray;
  public byte[] getByteArray() 
  { 
    return hByteArray == null ? toByteArray() : hByteArray; 
  }
  public DBLDBLDBL() {d1val = 0; d2val = 0; d3val = 0; }  
  public DBLDBLDBL(double d) { d1val = d; d2val = d; d3val = d;}  
  public DBLDBLDBL(double d1, double d2) { d1val = d1; d2val = d2; d3val = 0;}  
  public DBLDBLDBL(double d1, double d2, double d3) { d1val = d1; d2val = d2; d3val = d3;}  
  public DBLDBLDBL(DBLDBLDBL instance) 
  { 
    d1val = instance.d1val; d2val = instance.d2val; d3val = instance.d3val; 
  }  
  @Override
  public void copy(TCompoundDataObject from)
  {
    if (from instanceof DBLDBLDBL)
    {
      DBLDBLDBL instance = (DBLDBLDBL)from;
      d1val = instance.d1val; d2val = instance.d2val; d3val = instance.d3val; 
    }
  }
  private String separator = ", ";
  public void setFieldSeparator(String value)
  {
    if (value != null) separator = value;
  }
  @Override
  public String toString()
  {
      return "["+ d1val + separator + d2val + separator + d3val + "]";
  }
  public String getCsvString()
  {
    return d1val+" "+d2val+" "+d3val;
  }  
  public byte[] toByteArray()
  {
    try
    {
      dBuffer.reset();
      DataOutputStream ds = new DataOutputStream(dBuffer);
      byte[] b = Swap.Double(d1val);
      ds.write(b,0,8);
      b = Swap.Double(d2val);
      ds.write(b,0,8);
      b = Swap.Double(d3val);
      ds.write(b,0,8);
      ds.close();
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("DBLDBLDBL",e.getMessage(),TErrorList.code_failure,e,0);        
    }
    hByteArray = dBuffer.toByteArray();
    return hByteArray;
  }    
  public void toStruct()
  {
    try
    {
      ByteArrayInputStream dis = new ByteArrayInputStream(hByteArray);
      DataInputStream ds = new DataInputStream(dis);
      byte[] b = new byte[8];
      ds.read(b,0,8);
      d1val = Swap.Double(b);
      ds.read(b,0,8);
      d2val = Swap.Double(b);
      ds.read(b,0,8);
      d3val = Swap.Double(b);
      ds.close(); dis.close();
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("DBLDBLDBL",e.getMessage(),TErrorList.code_failure,e,0);        
    }  
  }
  @Override
  public int hashCode()
  {
    final int PRIME = 31;
    int result = 1;
    long temp;
    temp = Double.doubleToLongBits(d1val);
    result = PRIME * result + (int) (temp ^ (temp >>> 32));
    temp = Double.doubleToLongBits(d2val);
    result = PRIME * result + (int) (temp ^ (temp >>> 32));
    temp = Double.doubleToLongBits(d3val);
    result = PRIME * result + (int) (temp ^ (temp >>> 32));
    return result;
  }
  public DBLDBLDBL newInstance()
  {
    return new DBLDBLDBL();
  }
  public DBLDBLDBL clone()
  {
    return new DBLDBLDBL(this);
  }
  @Override
  public boolean equals(Object obj)
  {
    if (this == obj) return true;
    if (obj == null) return false;
    if (getClass() != obj.getClass()) return false;
    final DBLDBLDBL other = (DBLDBLDBL) obj;
    if (Double.doubleToLongBits(d1val) != Double.doubleToLongBits(other.d1val)) return false;
    if (Double.doubleToLongBits(d2val) != Double.doubleToLongBits(other.d2val)) return false;
    if (Double.doubleToLongBits(d3val) != Double.doubleToLongBits(other.d3val)) return false;
    return true;
  }
}
