/*
 * File FLTINT.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.types;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

import de.desy.tine.definitions.TErrorList;
import de.desy.tine.definitions.TFormat;
import de.desy.tine.endianUtils.Swap;
import de.desy.tine.server.logger.MsgLog;
public final class FLTINT implements TCompoundDataObject
{
  public float fval;
  public int ival;
  public double getDoubleValue() { return (double)fval; }
  public int getIntegerValue() { return (int)ival; }
  public static final int sizeInBytes = 8;
  public static final short dFormat = TFormat.CF_FLTINT;
  public int getSizeInBytes() { return sizeInBytes; }
  public void setValues(float newFValue, int newIValue)
  {
    ival = newIValue; fval = newFValue;
  }
  public short getFormat() { return dFormat; }    
  public String getTag() { return null; }
  ByteArrayOutputStream dBuffer = new ByteArrayOutputStream(sizeInBytes);
  public byte[] hByteArray;
  public void setByteArray(byte[] byteArray) { hByteArray = byteArray; }
  public byte[] getByteArray() 
  { 
    return hByteArray == null ? toByteArray() : hByteArray; 
  }
  public FLTINT() {fval = 0; ival = 0; }  
  public FLTINT(float f) { fval = f; ival = 0; }  
  public FLTINT(int i) { ival = i; }  
  public FLTINT(float f,int i) { fval = f; ival = i; }  
  public FLTINT(FLTINT instance) { fval = instance.fval; ival = instance.ival; }
  
  protected String separator = ", ";
  public void setFieldSeparator(String value)
  {
    if (value != null) separator = value;
  }
   public String toString()
  {
      return "["+ fval + separator + ival + "]";
  }
   public String getCsvString()
   {
     return fval+" "+ival;
   }  
   public FLTINT newInstance()
   {
     return new FLTINT();
   }
   public FLTINT clone()
   {
     return new FLTINT(this);
   }
  
  public boolean equals(FLTINT instance)
  {
    if (instance.ival != ival) return false;
    if (instance.fval != fval) return false;
    return true;
  }
  public byte[] toByteArray()
  {
    try
    {
      dBuffer.reset();
      DataOutputStream ds = new DataOutputStream(dBuffer);
      byte[] b = Swap.Float(fval);
      ds.write(b,0,4);
      ds.writeInt(Swap.Int(ival));
      ds.close();
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("FLTINT",e.getMessage(),TErrorList.code_failure,e,0);        
    }
    hByteArray = dBuffer.toByteArray();
    return hByteArray;
  }    
  public void toStruct()
  {
    try
    {
      ByteArrayInputStream dis = new ByteArrayInputStream(hByteArray);
      DataInputStream ds = new DataInputStream(dis);
      byte[] b = new byte[4];
      ds.read(b,0,4);
      fval = Swap.Float(b);
      ival = Swap.Int(ds.readInt());
      ds.close(); dis.close();
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("FLTINT",e.getMessage(),TErrorList.code_failure,e,0);        
    }	 
  }
  public int hashCode()
  {
    final int PRIME = 31;
    int result = 1;
    result = PRIME * result + Float.floatToIntBits(fval);
    result = PRIME * result + ival;
    return result;
  }
  public boolean equals(Object obj)
  {
    if (this == obj) return true;
    if (obj == null) return false;
    if (getClass() != obj.getClass()) return false;
    final FLTINT other = (FLTINT) obj;
    return equals(other);
  }
  @Override
  public void copy(TCompoundDataObject from)
  {
    if (from instanceof FLTINT)
    {
      FLTINT instance = (FLTINT)from;
      fval = instance.fval; ival = instance.ival;
    }
  }
}
