/*
 * File GSPECTRUM.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.types;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Arrays;
import de.desy.tine.definitions.TErrorList;
import de.desy.tine.definitions.TFormat;
import de.desy.tine.endianUtils.Swap;
import de.desy.tine.server.logger.MsgLog;
public final class GSPECTRUM implements TCompoundDataObject
{
  public static final int STD_LENGTH = 1048576;
  public static final int GSP_STSZ = 24;
  public static final int GSP_TISZ = 8;
  public String comment;
  public int id;
  public int tm;
  public float s_start;
  public float s_inc;
  public int status;
  public int groups;
  public int group_size;
  public float group_inc;
  public int stat_mask;
  public float[] statistics = new float[GSP_STSZ];
  public int[] timing_param = new int[GSP_TISZ];
  
  public double getDoubleValue() { return (double)d_spect_array[0]; }
  public int getIntegerValue() { return (int)status; }
  public float[] d_spect_array;
  public static int sizeInBytes;
  public static int hdrSizeInBytes = 244;
  public int length;
  public static final short dFormat = TFormat.CF_GSPECTRUM;
  public int getSizeInBytes() { return sizeInBytes; }
  public short getFormat() { return dFormat; }    
  public String getTag() { return null; }
  ByteArrayOutputStream dBuffer = new ByteArrayOutputStream(sizeInBytes);
  public byte[] hByteArray;
  public void setByteArray(byte[] byteArray) { hByteArray = byteArray; }
  public byte[] getByteArray() 
  { 
    return hByteArray == null ? toByteArray() : hByteArray; 
  }
  protected String separator = ", ";
  public void setFieldSeparator(String value)
  {
    if (value != null) separator = value;
  }
  public String toString()
  {
    StringBuffer sb = new StringBuffer(length*4 + hdrSizeInBytes);
    sb.append("["+ comment + separator + id + separator + tm + separator + s_start + separator + s_inc + separator + status);
    sb.append(separator+ groups + separator + group_size + separator + group_inc + separator + stat_mask + "]");
    sb.append("\n[");
    for (int i=0; i<GSP_STSZ; i++) sb.append(" "+statistics[i]);
    sb.append("]");
    sb.append("\n[");
    for (int i=0; i<GSP_TISZ; i++) sb.append(" "+timing_param[i]);
    sb.append("]\n");
    for (int i=0; i<length; i++) sb.append(" " + d_spect_array[i]); 
    return sb.toString();
  }
  public String getCsvString()
  {
    return "\""+comment + ", " + id + ", " + tm + ", " + s_start + ", " + s_inc + ", " + status+"\"";
  }  
  public void setValues(int newId, int newTM, float newStart, float newInc, int newStatus,
                        int newGroups, int newGroupSize, float newGroupInc, int newStatMask,
                        float[] newStatistics, int[] newTimingParameters)
  {
    id = newId; tm = newTM; s_start = newStart; s_inc = newInc; status = newStatus;
    groups = newGroups; group_size = newGroupSize; group_inc = newGroupInc; stat_mask = newStatMask;
    System.arraycopy(newStatistics, 0, statistics, 0, GSP_STSZ);
    System.arraycopy(newTimingParameters, 0, timing_param, 0, GSP_TISZ);
    
  }
  public GSPECTRUM() 
  {
    s_start = 0; s_inc = 0; status = 0; tm  = 0; comment = ""; 
    d_spect_array = new float[STD_LENGTH];
    sizeInBytes = STD_LENGTH * 4 + hdrSizeInBytes;
    length = STD_LENGTH;
    hByteArray = new byte[sizeInBytes];
  }  
  public GSPECTRUM(int spectrumLength) 
  {
    s_start = 0; s_inc = 0; status = 0; tm  = 0; comment = "";
    d_spect_array = new float[spectrumLength];
    sizeInBytes = spectrumLength * 4 + hdrSizeInBytes;
    length = spectrumLength;
    hByteArray = new byte[sizeInBytes];
  }  
  public GSPECTRUM(GSPECTRUM instance) 
  {
    if (instance == null) return;
    s_start = instance.s_start; s_inc = instance.s_inc; status = instance.status; tm  = instance.tm;
    id = instance.id; groups = instance.groups; group_size = instance.group_size;
    group_inc = instance.group_inc; stat_mask = instance.stat_mask;
    System.arraycopy(instance.statistics, 0, statistics, 0, GSP_STSZ);
    System.arraycopy(instance.timing_param, 0, timing_param, 0, GSP_TISZ);
    comment = new String(instance.comment); 
    length = instance.length;
    hByteArray = new byte[sizeInBytes];
    d_spect_array = new float[length];
    if (instance.d_spect_array != null)
      System.arraycopy(instance.d_spect_array, 0, d_spect_array, 0, length);
  }  
  @Override
  public void copy(TCompoundDataObject from)
  {
    if (from instanceof GSPECTRUM)
    {
      GSPECTRUM instance = (GSPECTRUM)from;
      s_start = instance.s_start; s_inc = instance.s_inc; status = instance.status; tm  = instance.tm; 
      id = instance.id; groups = instance.groups; group_size = instance.group_size;
      group_inc = instance.group_inc; stat_mask = instance.stat_mask;
      System.arraycopy(instance.statistics, 0, statistics, 0, GSP_STSZ);
      System.arraycopy(instance.timing_param, 0, timing_param, 0, GSP_TISZ);
      comment = new String(instance.comment); 
      length = Math.min(length, instance.length);
      if (instance.d_spect_array != null)
        System.arraycopy(instance.d_spect_array, 0, d_spect_array, 0, length);
    }
  }
  public boolean equals(GSPECTRUM instance)
  {
    if (instance.id != id) return false;
    if (instance.s_start != s_start) return false;
    if (instance.s_inc != s_inc) return false;
    if (instance.status != status) return false;
    if (instance.groups != groups) return false;
    if (instance.group_size != group_size) return false;
    if (instance.group_inc != group_inc) return false;
    if (instance.stat_mask != stat_mask) return false;
    if (!Arrays.equals(instance.statistics, statistics)) return false;
    if (!Arrays.equals(instance.timing_param, timing_param)) return false;
    if (instance.length != length) return false;
    if (instance.comment == null || comment == null) return false;
    if (instance.comment.compareTo(comment) != 0) return false;
    if (!Arrays.equals(instance.d_spect_array, d_spect_array)) return false;
    return true;
  }
  public byte[] toByteArray()
  {
	try
	{
		dBuffer.reset();
		DataOutputStream ds = new DataOutputStream(dBuffer);
		byte[] b = new byte[80];
		ds.write(comment.getBytes());
		ds.write(b,0,80-comment.length());
    ds.writeInt(Swap.Int(id));
		ds.writeInt(Swap.Int(tm));
		b = Swap.Float(s_start); ds.write(b,0,4);
		b = Swap.Float(s_inc); ds.write(b,0,4);
		ds.writeInt(Swap.Int(status));

		ds.writeInt(Swap.Int(groups));
    ds.writeInt(Swap.Int(group_size));
    b = Swap.Float(group_inc); ds.write(b,0,4);
    ds.writeInt(Swap.Int(stat_mask));
    for (int i=0; i<GSP_STSZ; i++)
    {
      b = Swap.Float(statistics[i]); ds.write(b,0,4);
    }
    for (int i=0; i<GSP_TISZ; i++)
    {
      b = Swap.IntToBytes(timing_param[i]); ds.write(b,0,4);
    }
  	for (int i=0; i<d_spect_array.length; i++)
		{
		  b = Swap.Float(d_spect_array[i]); ds.write(b,0,4);
		}
		ds.close();
	}
	catch (IOException e)
	{
	  e.printStackTrace();
      MsgLog.log("SPECTRUM",e.getMessage(),TErrorList.code_failure,e,0);        
	}
	hByteArray = dBuffer.toByteArray();
	return hByteArray;
  }  
  public void toStruct()
  {
    try
    {
    	ByteArrayInputStream dis = new ByteArrayInputStream(hByteArray);
    	DataInputStream ds = new DataInputStream(dis);
    	byte[] b = new byte[hdrSizeInBytes];
    	ds.read(b,0,80);
    	comment = (new String(b)).trim();
    	id = Swap.Int(ds.readInt());
      tm = Swap.Int(ds.readInt());
    	ds.read(b,0,4); s_start = Swap.Float(b);
    	ds.read(b,0,4); s_inc = Swap.Float(b);
    	status = Swap.Int(ds.readInt());
    	
    	groups = Swap.Int(ds.readInt());
    	group_size = Swap.Int(ds.readInt());
      ds.read(b,0,4); group_inc = Swap.Float(b);
      stat_mask = Swap.Int(ds.readInt());
      for (int i=0; i<GSP_STSZ; i++)
      {
        if (ds.read(b,0,4) < 0) break;
        statistics[i] = Swap.Float(b);
      }
      for (int i=0; i<GSP_TISZ; i++)
      {
        if (ds.read(b,0,4) < 0) break;
        timing_param[i] = Swap.Int(b);
      }
    	for (int i=0; i<d_spect_array.length && i<length; i++)
    	{
    	  if (ds.read(b,0,4) < 0) break;
    	  d_spect_array[i] = Swap.Float(b);
    	}
    	ds.close(); dis.close();
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("SPECTRUM.toStruct",e.getMessage(),TErrorList.code_failure,e,0);        
    }	 
  }
  public int hashCode()
  {
    final int PRIME = 31;
    int result = 1;
    result = PRIME * result + ((comment == null) ? 0 : comment.hashCode());
    result = PRIME * result + Arrays.hashCode(d_spect_array);
    result = PRIME * result + length;
    result = PRIME * result + Float.floatToIntBits(s_inc);
    result = PRIME * result + Float.floatToIntBits(s_start);
    result = PRIME * result + status;
    result = PRIME * result + tm;
    result = PRIME * result + id;
    result = PRIME * result + groups;
    result = PRIME * result + group_size;
    result = PRIME * result + Float.floatToIntBits(group_inc);
    result = PRIME * result + stat_mask;
    result = PRIME * result + Arrays.hashCode(statistics);
    result = PRIME * result + Arrays.hashCode(timing_param);
    return result;
  }
  public GSPECTRUM newInstance()
  {
    return new GSPECTRUM();
  }
  public GSPECTRUM clone()
  {
    return new GSPECTRUM(this);
  }              
  public boolean equals(Object obj)
  {
    if (this == obj) return true;
    if (obj == null) return false;
    if (getClass() != obj.getClass()) return false;
    final GSPECTRUM other = (GSPECTRUM) obj;
    return equals(other);
  }
  public void setSpectrum(GSPECTRUM newSpectrum)
  {
    if (newSpectrum == null) return;
    setValues(newSpectrum.getId(), newSpectrum.getTime(), newSpectrum.getStart(), newSpectrum.getIncrement(), 
        newSpectrum.getStatus(),newSpectrum.getGroups(),newSpectrum.getGroupSize(),newSpectrum.getGroupIncrement(),
        newSpectrum.getStatMask(), newSpectrum.getStatistics(), newSpectrum.getTimingParameters());
    setComment(newSpectrum.getComment());
    setValues(newSpectrum.getValues());
  }
  public float[] getValues()
  {
    return d_spect_array;
  }
  public void setValues(float[] values)
  {
    if (values == null) return;
    int len = values.length;
    if (len > d_spect_array.length) len = d_spect_array.length;
    System.arraycopy(values, 0, d_spect_array, 0, len);
  }
  public void setComment(String newComment)
  {
    comment = newComment;
  }
  public String getComment()
  {
    return comment;
  }
  public int getLength()
  {
    return length;
  }
  public float getIncrement()
  {
    return s_inc;
  }
  public float getStart()
  {
    return s_start;
  }
  public int getTime()
  {
    return tm;
  }
  public int getStatus()
  {
    return status;
  }
  public int getId()
  {
    return id;
  }
  public int getGroups()
  {
    return groups;
  }
  public int getGroupSize()
  {
    return group_size;
  }
  public float getGroupIncrement()
  {
    return group_inc;
  }
  public int getStatMask()
  {
    return stat_mask;
  }
  public float[] getStatistics()
  {
    return statistics;
  }
  public int[] getTimingParameters()
  {
    return timing_param;
  }
}