/*
 * File IMAGE.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.types;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

import de.desy.tine.definitions.TErrorList;
import de.desy.tine.definitions.TFormat;
import de.desy.tine.endianUtils.Swap;
import de.desy.tine.server.logger.MsgLog;
import de.desy.tine.dataUtils.UnsignedUtils;

/**
 * TINE type CF_IMAGE java class implementation. Documentation mainly done by
 * Stefan Weisse. Initial class design and implementation by Philip Duval, based on
 * negotiations and agreement (between DESY Zeuthen-based controls group (now
 * ERS), MST (now MCS 1) and MVP (now MCS 4)) on a defined control system type
 * for video image data transport.  Refactoring and optimizations by David Melkymyan.
 * 
 * @author Philip Duval (pd)
 * @author Stefan Weisse (sw)
 * @author David Melkumyan (mdavid)
 * @version 1.1 Aug 10, 2009
 * @version last documented Jan 28, 2009
 */
public class IMAGE implements TCompoundDataObject,Cloneable
//public class IMAGE implements Cloneable
{
  /**
   * Total size in bytes of frame/source headers
   */
  public static final int HEADER_SIZE = TFormat.getFormatHeaderSize(TFormat.CF_IMAGE); // mdavid:
                                                                                       // added
  /**
   * Default capacity of an image data buffer
   */
  public static final int DEFAULT_DATA_SIZE = 0x600000; // mdavid: added

  /**
   * Magic id for VSv3 'VSV3'
   */
  public static final int DEFAULT_BASE_TAG = 0x33565356; // mdavid: added

  /**
   * No camera port is set
   */
  public static final long DEFAULT_CAMERA_PORT_ID = 0x0; // mdavid: added

  /**
   * Initial IMAGE version
   */
  public static final long IMAGE_INITIAL_VERSION = 0x1; // mdavid: added

  /**
   * Current IMAGE version
   */
  public static final long IMAGE_VERSION = 0x1; // mdavid: added

  /**
   * Maximum IMAGE version
   */
  public static final long IMAGE_MAX_VERSION = 0x1; // mdavid: added

  /**
   * Default camera port name
   */
  public static final String DEFAULT_CAMERA_PORT_NAME = "unknown"; // mdavid: added 

  public static final int IF_GRAY = 0,
        IF_MIN = 0,
        IF_RGB = 1,
        IF_RGBA = 2,
        IF_RGB15 = 3,
        IF_RGB16 = 4,
        IF_JPEG = 5,
        IF_TIFF = 6,
        IF_YUV411 = 7,
        IF_YUV422 = 8,
        IF_YUV444 = 9,
        IF_HUFFYUV = 10,
        IF_BITMAP = 11,  /**< Windows BitMaP bits follow headers */
        IF_BAYER = 12,    /**< Bayer-packed image bits */
        IF_PNG = 13,
        IF_BASE64 = 14;
  // some important variables for sending/receiving arrays or including in structs ...
  private int xoff = 0; // extended space offset
  private int xsiz = 0; // extended space size in bytes
  public int getExtendedOffset() { return xoff; }
  public void setExtendedOffset(int value) { xoff = value; }
  public int getExtendedSize() { return xsiz; }
  public void setExtendedSize(int value) { xsiz = value; }

  private SourceHeader sourceHeader = new SourceHeader();
  private FrameHeader frameHeader = new FrameHeader();
  private byte[] imageData;
  /**
   * Constructs an empty image frame buffer with an initial capacity of
   * {@link #DEFAULT_DATA_SIZE}.
   * 
   * @author mdavid
   */
  public IMAGE()
  {
    this(DEFAULT_DATA_SIZE);
  }
  /**
   * Constructs an empty image frame buffer with the specified initial capacity.
   * 
   * @author mdavid
   * @param initialCapacity
   *          - the initial capacity of the data buffer
   * @exception IllegalArgumentException
   *              if the specified initial capacity is negative
   */
  public IMAGE(int initialCapacity)
  {
    if (initialCapacity < 0) throw new IllegalArgumentException("Illegal Capacity: " + initialCapacity);
    this.imageData = new byte[initialCapacity];
  }
  /**
   * Returns a reference to the appended image bits buffer. One should note that
   * this buffer might only be filled to a certain extend, because buffer can be
   * bigger than contained data is.<br>
   * To obtain buffer that contains only actual/delivered data (based on
   * {@link FrameHeader#appendedFrameSize}) use {@link #trimFrameBuffer()}
   * instead.
   * 
   * @return reference to attached frame buffer (byte array)
   */
  public byte[] getImageFrameBuffer()
  {
    return imageData;
  }
  /**
   * Sets the internal image bits buffer reference to the passed byte buffer
   * argument.<br>
   * <i>Note (mdavid): do not use this method for constructing empty buffer, use
   * constructor {@link #IMAGE(int)} instead.</i>
   * 
   * @deprecated temporary annotated as deprecated (see <i>Note</i>)
   * @param buffer
   *          - reference to image bits buffer
   */
  public void setImageFrameBuffer(byte[] buffer)
  {
    imageData = buffer;
  }
  /**
   * Trims the capacity of the image data buffer that contains actual/delivered
   * data. An application can use this operation to minimize the storage of an
   * image data.
   * 
   * @author mdavid
   * @return trimmed image data buffer
   * @see FrameHeader#appendedFrameSize
   * @throws IllegalStateException
   *           if <code>appendedFrameSize</code> is undefined
   * @throws IndexOutOfBoundsException
   *           if <code>appendedFrameSize</code> is greater than image data size
   */
  public byte[] trimFrameBuffer()
  {
    if (frameHeader.appendedFrameSize == -1) throw new IllegalStateException("appendedFrameSize == -1!");
    // mdavid: I know, may be, this is not a good place for such validation...
    // but.. better was not found
    if (imageData.length < frameHeader.appendedFrameSize)
      throw new IndexOutOfBoundsException("data length (" + imageData.length + ") < frame size"
          + frameHeader.appendedFrameSize + "!");
    if (imageData.length == frameHeader.appendedFrameSize) return imageData;
    byte[] data = new byte[frameHeader.appendedFrameSize];
    System.arraycopy(imageData, 0, data, 0, frameHeader.appendedFrameSize);
    return (this.imageData = data);
  }
  /**
   * needed for arrays of IMAGE (or within structs)
   * @return
   */
  public void ensureFrameBufferCapacity(int length)
  {
    if (imageData.length < length)
    {
      length += 4096 - (length%4096);
      imageData = new byte[length];
    }
    return;
  }
  /**
   * mdavid note: currently unused, can be removed later
   * 
   * @param dis
   *          - data input stream
   * @throws IOException
   * @deprecated mdavid: use {@link #readHeaders(byte[])} instead
   */
  public void setHeaderBytes(DataInputStream dis) throws IOException
  {
    // dis.read(hHeaderByteArray); // mdavid: commented
  }
  /**
   * mdavid: unused, can be removed later
   * 
   * @deprecated mdavid: use {@link #read(DataInputStream)} instead
   * @deprecated
   * @param dis
   * @throws IOException
   */
  public void setImageBytes(DataInputStream dis) throws IOException
  {
    dis.read(imageData);
  }
  /**
   * mdavid: unused, can be removed (?)
   * 
   * @deprecated mdavid: use {@link #read(DataInputStream)} instead
   * @param dis
   * @param off
   * @param len
   * @throws IOException
   */
  public void setImageBytes(DataInputStream dis, int off, int len) throws IOException
  {
    dis.read(imageData, off, len);
  }
  /**
   * Reads headers structure from byte array.
   * 
   * @author mdavid
   * @param src
   *          - source byte array
   * @throws IOException
   */
  public void readHeaders(byte[] src) throws IOException
  {
    if (src.length < HEADER_SIZE) throw new IllegalArgumentException("src.length < " + HEADER_SIZE + "!");
    byte[] buffer = null;
    if (src.length > HEADER_SIZE)
    {
      buffer = new byte[HEADER_SIZE];
      System.arraycopy(src, 0, buffer, 0, HEADER_SIZE);
    }
    else
      buffer = src; // src.length == HEADER_SIZE
    ByteArrayInputStream bais = new ByteArrayInputStream(buffer);
    DataInputStream dis = new DataInputStream(bais);
    try
    {
      readHeaders(dis);
    }
    finally
    {
      try
      {
        dis.close();
      }
      catch (IOException ignored)
      {
      }
    }
  }
  public void readImageArrayHeaders(byte[] src,int offset) throws IOException
  {
    if (src.length < HEADER_SIZE+8+offset) throw new IllegalArgumentException("src.length < " + HEADER_SIZE + "!");
    byte[] buffer = null;
    buffer = new byte[HEADER_SIZE+8];
    System.arraycopy(src, offset, buffer, 0, HEADER_SIZE+8);
    ByteArrayInputStream bais = new ByteArrayInputStream(buffer);
    DataInputStream dis = new DataInputStream(bais);
    try
    {
      readHeaders(dis);
      xoff = Swap.Int(dis.readInt());
      xsiz = Swap.Int(dis.readInt());
    }
    finally
    {
      try
      {
        dis.close();
      }
      catch (IOException ignored)
      {
      }
    }
  }
  /**
   * @author mdavid
   * 21.12.09 duval: made this public for handling adjustable arrays inside structs
   */
  public void readHeaders(DataInputStream input) throws IOException
  {
    sourceHeader.read(input);
    frameHeader.read(input);
  }
  /**
   * @author mdavid
   * 21.12.09 duval: made this public for handling adjustable arrays inside structs
   */
  public void writeHeaders(DataOutputStream output) throws IOException
  {
    sourceHeader.write(output);
    frameHeader.write(output);
  }
  /**
   * Copies an array from the specified source array, beginning at the specified
   * position, to the specified position of the image data array.
   * <p>
   * Note: unused - can be removed
   * 
   * @author mdavid
   * @param src
   *          the source array.
   * @param srcPos
   *          starting position in the source array.
   * @param destPos
   *          starting position in the destination data.
   * @param length
   *          the number of array elements to be copied.
   * @exception IllegalArgumentException
   *              if <code>imageData</code> is <code>null</code> and
   *              <code>destPos != 0</code>.
   * @exception IndexOutOfBoundsException
   *              if copying would cause access of data outside array bounds.
   * @exception NullPointerException
   *              if either <code>src</code> or <code>imageData</code> is
   *              <code>null</code>.
   */
  public void fillFrameBuffer(byte[] src, int srcPos, int destPos, int length)
  { // TODO: mdavid - remove this method later
    if (imageData == null) throw new NullPointerException("imageData == null!");
    if (imageData.length < destPos + length) throw new IndexOutOfBoundsException();
    System.arraycopy(src, srcPos, imageData, destPos, length);
  }
  /**
   * Copies a byte into the designated position of the image buffer.
   * Used in IMAGE array transfers
   */
  public void fillFrameBuffer(byte src, int destPos)
  { // TODO: duval - don't remove this one 
    //if (imageData == null) throw new NullPointerException("imageData == null!");
    //if (imageData.length < destPos) throw new IndexOutOfBoundsException();
    // let the exceptions fall where they may ...
    imageData[destPos] = src;
  }
  /**
   * Returns gross length of the appended image bits buffer in bytes. Value
   * might be significantly bigger than appended frame size, because buffer must
   * be bigger or equal to appended frame size.
   * 
   * @return size of the attached buffer in bytes
   */
  public int getImageFrameSize()
  {
    return imageData.length;
  }
  /**
   * Returns the header size by questioning TFormat.
   * 
   * @return size of the header in bytes, if serialized
   * @see TFormat
   * @see TFormat#getFormatHeaderSize(short)
   */
  public int getHeaderSize()
  {
    return HEADER_SIZE; // mdavid: changed
  }
  // public byte[] toByteArray() { // mdavid: optimized
  // if (imageData == null)
  // return null; // TODO: mdavid: is it OK ? Proposal: return only headers byte
  // array data instead of null (?)
  //
  // ByteArrayOutputStream dBuffer = new
  // ByteArrayOutputStream(getSizeInBytes());
  // DataOutputStream ds = new DataOutputStream(dBuffer);
  // try {
  // // writeImageSourceHeader(ds); // mdavid: commented
  // // writeImageFrameHeader(ds); // mdavid: commented
  // sourceHeader.write(ds); // mdavid: added
  // frameHeader.write(ds); // mdavid: added
  //
  // ds.write(imageData, 0, imageData.length);
  // return dBuffer.toByteArray();
  //
  // } catch (IOException e) {
  // e.printStackTrace();
  // MsgLog.log("IMAGE.toByteArray", e.getMessage(), TErrorList.code_failure, e,
  // 0);
  //
  // } finally {
  // try {
  // ds.close();
  // } catch (IOException ignored) {
  // }
  // }
  // return null;
  // }
  /**
   * Convert IMAGE object into a byte array.
   * 
   * @author mdavid
   * @return the byte array representing this object.
   */
  public byte[] toByteArray()
  {
    byte[] bytes = new byte[getSizeInBytes()];
    System.arraycopy(getHeaderBytes(), 0, bytes, 0, HEADER_SIZE);
    System.arraycopy(imageData, 0, bytes, HEADER_SIZE, imageData.length);
    return bytes;
  }
  /**
   * for arrays, inclusion in structs, things are serialized 
   * differently : need separate routines for adjustable header and
   * adjustable frame parts 
   * 
   * @author duval
   * @return the byte array representing the header + 8 bytes frame location info
   */
  public byte[] toAdjustableHeaderByteArray()
  {
    byte[] bytes = new byte[HEADER_SIZE+8];
    System.arraycopy(getHeaderBytes(), 0, bytes, 0, HEADER_SIZE);
    // offset and sizeInBytes in the extended space ...
    System.arraycopy(Swap.IntToBytes(xoff),0,bytes,HEADER_SIZE,4);
    xsiz = imageData.length;
    System.arraycopy(Swap.IntToBytes(xsiz),0,bytes,HEADER_SIZE+4,4);
    // xoff += xsiz; < bad idea !
    return bytes;
  }
  public byte[] toFrameByteArray()
  {
    return imageData;
  }
  // mdavid: commented
  // public void toStruct() {
  // try {
  // ByteArrayInputStream bais = new ByteArrayInputStream(hHeaderByteArray);
  // DataInputStream dis = new DataInputStream(bais);
  // readImageSourceHeader(dis);
  // readImageFrameHeader(dis);
  // } catch (IOException e) {
  // e.printStackTrace();
  // MsgLog.log("IMAGE", e.getMessage(), TErrorList.code_failure, e, 0);
  // }
  // }
  /**
   * Initializes header structure
   * 
   * @not yet deprecated mdavid: use {@link #readHeaders(byte[])} instead
   */
  public void toStruct()
  { // mdavid: optimized
    // ByteArrayInputStream bais = new ByteArrayInputStream(hHeaderByteArray);
    // DataInputStream dis = new DataInputStream(bais);
    // try {
    // readImageSourceHeader(dis);
    // readImageFrameHeader(dis);
    //
    // } catch (IOException e) {
    // e.printStackTrace();
    // MsgLog.log("IMAGE.toStruct", e.getMessage(), TErrorList.code_failure, e,
    // 0);
    //
    // } finally {
    // try {
    // dis.close();
    // } catch (IOException ignored) {
    // }
    // }
  }
  /**
   * Returns headers byte array.<br>
   * Note: unused (not yet true: TTaggedStruct doesn't yet treat IMAGE explicitly)!
   * 
   * @not yet deprecated mdavid: use {@link #toByteArray()} instead
   */
  public byte[] getByteArray()
  {
    return getHeaderBytes();
    // return hHeaderByteArray; // mdavid: commented
  }
  // mdavid: commented
  // public void setByteArray(byte[] byteArray)
  // {
  // if (byteArray == null) return;
  // if (hHeaderByteArray == null)
  // hHeaderByteArray = new byte[TFormat.getFormatHeaderSize(TFormat.CF_IMAGE)];
  // try
  // {
  // System.arraycopy(byteArray,0,hHeaderByteArray,0,hHeaderByteArray.length);
  // byte[] fb = new byte[byteArray.length-hHeaderByteArray.length];
  // System.arraycopy(byteArray,hHeaderByteArray.length,fb,0,byteArray.length-hHeaderByteArray.length);
  // setImageFrameBuffer(fb);
  // }
  // catch (Exception e)
  // {
  // MsgLog.log("IMAGE.setByteArray",
  // e.getMessage(),TErrorList.code_failure,e,0);
  // }
  // }
  /**
   * Initializes object from byte array
   * 
   * @not yet deprecated mdavid: use {@link #read(byte[])} instead
   */
  public void setByteArray(byte[] byteArray)
  { // mdavid: optimized
    if (byteArray == null) return;
    try
    {
      // initializes headers
      readHeaders(byteArray);
      // calculate data size
      int dataSize = byteArray.length - HEADER_SIZE;
      // create new data buffer if needed
      if (imageData == null || imageData.length != dataSize) imageData = new byte[dataSize];
      // copy frame bytes
      System.arraycopy(byteArray, HEADER_SIZE, imageData, 0, dataSize);
    }
    catch (Exception e)
    {
      MsgLog.log("IMAGE.setByteArray", e.getMessage(), TErrorList.code_failure, e, 0);
    }
  }
  /**
   * Reads the <code>IMAGE</code> from the supplied byte array.
   * 
   * @author mdavid
   * @param src
   *          - input byte array
   * @exception IOException
   *              if I/O error occurs
   */
  public void read(byte[] src) throws IOException
  {
    if (src == null) throw new IllegalArgumentException("src == null");
    readHeaders(src);
    System.arraycopy(src, HEADER_SIZE, imageData, 0, src.length - HEADER_SIZE);
  }
  /**
   * Reads the <code>IMAGE</code> from the supplied <code>DataInputStream</code>
   * .
   * 
   * @author mdavid
   * @param input
   *          - an <code>DataInputStream</code> to read from.
   * @exception NullPointerException
   *              if <code>input</code> is <code>null</code>.
   * @exception IOException
   *              if an error occurs during reading.
   */
  public void read(DataInputStream input) throws IOException
  {
    if (input == null) throw new IllegalArgumentException("input == null");
    readHeaders(input);
    input.read(imageData);
  }
  /**
   * Appends a complete image containing a <b>trimmed</b> image data with
   * source/frame headers to the output.
   * 
   * @author mdavid
   * @param output
   *          - an <code>DataOutputStream</code> to write to.
   * @throws IOException
   *           if an error occurs during writing.
   * @see #trimFrameBuffer()
   */
  public void write(DataOutputStream output) throws IOException
  {
    if (output == null) throw new NullPointerException("output == null!");
    writeHeaders(output);
    output.write(imageData, 0, frameHeader.appendedFrameSize);
  }
  /**
   * Convert image headers (SourceHeader/FrameHeader) into a byte array.
   * 
   * @author mdavid
   * @return the byte array representing the headers.
   */
  public byte[] getHeaderBytes()
  {
    ByteArrayOutputStream baos = new ByteArrayOutputStream(HEADER_SIZE);
    DataOutputStream ds = new DataOutputStream(baos);
    try
    {
      writeHeaders(ds);
      return baos.toByteArray();
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("IMAGE.getHeaderBytes", e.getMessage(), TErrorList.code_failure, e, 0);
    }
    finally
    {
      try
      {
        ds.close();
      }
      catch (IOException ignored)
      {
      }
    }
    return null;
  }
  public short getFormat()
  {
    return TFormat.CF_IMAGE;
  }
  public String getTag() { return null; }
  // mdavid: commented and changed
  // public int getSizeInBytes() {
  // return sizeInBytes;
  // }
  /*
   * mdavid: changed
   */
  public int getSizeInBytes()
  {
    return (imageData == null) ? HEADER_SIZE : (HEADER_SIZE + imageData.length);
  }
  /**
   * @deprecated mdavid
   */
  public double getDoubleValue()
  {
    return 0D; // mdavid: changed
  }
  /**
   * @deprecated mdavid
   */
  public int getIntegerValue()
  {
    return 0; // mdavid: changed
  }
  /**
   * mdavid: annotated as 'deprecated'
   * <p>
   * reason(s):
   * <ul>
   * <li>duplicated (see
   * <code>"de.desy.tine.dataUtils.TDataType.UnsignedInt"<code>)
     * <li>not-optimized
   * </ul>
   * <ul>
   * <li>Note: doesn't follow the Java naming convention
   * </ul>
   * 
   * @deprecated use {@link UnsignedUtils#toUnsignedInt(int)} instead (~6 times
   *             faster!)
   * @since 08.2009
   */
  static long UnsignedInt(int ival)
  {
    long lval = (long) ival;
    if (lval < 0) lval += 4294967296L;
    return lval;
    // return ival < 0 ? (ival + 4294967296L) : ival; // mdavid: optimized (~2
    // times faster)
    // return ival < 0 ? (ival += 1L << 32) : ival; // mdavid: optimized (~3
    // times faster)
    // return ival & 0xFFFFFFFFL; // mdavid: optimized (~6 times faster)
  }
  /**
   * Returns reference to subclass SourceHeader.
   * 
   * @return reference to subclass SourceHeader
   */
  public SourceHeader getSourceHeader()
  {
    return sourceHeader;
  }
  /**
   * Sets the image source header.
   * 
   * @author mdavid
   * @param sourceHeader
   *          - source header
   */
  public void setSourceHeader(SourceHeader sourceHeader)
  {
    this.sourceHeader = sourceHeader;
  }
  /**
   * Returns reference to subclass FrameHeader.
   * 
   * @return reference to subclass FrameHeader
   */
  public FrameHeader getFrameHeader()
  {
    return frameHeader;
  }
  /**
   * Sets the image frame header.
   * 
   * @author mdavid
   * @param frameHeader
   *          - frame header
   */
  public void setFrameHeader(FrameHeader frameHeader)
  {
    this.frameHeader = frameHeader;
  }
  protected String separator = ", ";
  public void setFieldSeparator(String value)
  {
    if (value != null) separator = value;
  }
  /*
   * mdavid: added
   */
  @Override
  public String toString()
  {
    StringBuilder sb = new StringBuilder();
    sb.append("sourceHeader: [").append(sourceHeader.toString()).append("]\n");
    sb.append("frameHeader: [").append(frameHeader.toString()).append("]");
    return sb.toString();
  }
  public String getCsvString()
  {
    return "[IMAGE data]";
  }  
  public IMAGE newInstance()
  {
    return new IMAGE();
  }
  /*
   * mdavid: added
   */
  @Override
  public IMAGE clone() //throws CloneNotSupportedException
  {
    IMAGE timage = new IMAGE(0);
    timage.sourceHeader = sourceHeader.clone();
    timage.frameHeader = frameHeader.clone();
    timage.imageData = imageData.clone();
    return timage;
  }
  @Override
  public void copy(TCompoundDataObject from)
  {
    if (from instanceof IMAGE)
    {
    }
  }
  /**
   * Subclass of first part of header (subclass used for logical grouping of
   * class member variables which map almost 1:1 to header structures
   * transferred in binary form.
   */
  public class SourceHeader implements Cloneable
  {
    /**
     * Magic ID in front of binary data block. Value is fixed to four character
     * code 'VSV3' (little endian: 33565356 hex). VSV3 stands for VideoSystem
     * Version 3.
     * <p>
     * In C/C++ world, member is an unsigned int32.
     */
    public int baseTag = DEFAULT_BASE_TAG;
    /**
     * Unique ID that identifies one single location of image creation, usually
     * named Image Source in VSV3. There must not be two image sources sharing
     * the same camera port Id. If cameraPortId is set to 0 it means that no
     * camera port id is attached to the image buffer so Image Source (origin of
     * image) is undefined.
     * <p>
     * In C/C++ world, member is an unsigned int32.
     */
    public long cameraPortId = DEFAULT_CAMERA_PORT_ID; // uint32->long sw! default->0L
                                     // (no_camera_port_id) mdavid!
    /**
     * Defines the version of the IMAGE header. At the moment only version 1 is
     * defined so this tag is fixed to '1'. Applications are asked to check this
     * version number whether it meets their minimum requirements. It is
     * foreseen that future header versions will be backward compatible. New
     * applications should be able to work with old headers and new headers. Old
     * applications can only work with old headers. Checking is performed based
     * on integer comparison.
     * <p>
     * <code>if (versionTag == 0x1) { good_header_version } else
     *  { bad header }</code>
     * <p>
     * In C/C++ world, this header member is an unsigned int32.
     */
    public long versionTag = IMAGE_INITIAL_VERSION; // uint32->long sw! default->0x1L mdavid!
    /**
     * Contains the whole length of the data block as it would be transferred
     * bytewise on network level. This value contains physical amount of bytes
     * necessary for sourceHeader, for frameHeader and the net size of appended
     * frame bits. If no frame bits are appended, the size of sourceHeader plus
     * frameHeader must be contained within.
     * <p>
     * For practical reason, current C/C++ implementations of the VideoSystem V3
     * are limited to 6 Megabytes of frame bits size plus header size
     * (sourceHeader + frameHeader). But the general design rule is that an
     * application should never expect to receive image frames that are less
     * than or equal to 6 megabytes in size.
     * <p>
     * In C/C++ world, this header member is an unsigned int32.
     */
    public long totalLength; // uint32->long sw!
    /**
     * Contains the time the image was originally taken at image source, UTC
     * seconds part. The design rule is that this timestamp should be acquired
     * as early as possible after image acquire from hardware sensor. Some
     * digital cameras are even capable of tagging image data with a proper
     * timestamp. This timestamp, as it is most precise one can get, should be
     * transformed and copied here. The aim of the game is not to take a second
     * timestamp in source code once the image is received in software, except
     * if no other possibility exist.
     * <p>
     * In C/C++ world, this header member is an unsigned int32.
     */
    public int timestampSeconds;
    /**
     * Contains the time the image was originally taken at image source,
     * microseconds offset to UTC time. The same design rules as for
     * timestampSeconds member applies. If the granularity of system clock or
     * camera clock (if image bits are time-tagged at hardware) is not good
     * enough to deliver microseconds, close-by values like milliseconds or
     * tenths of a second are also fine.
     * <p>
     * In C/C++ world, this header member is an signed int32.
     */
    public int timestampMicroseconds;
    /** < image timestamp (micro seconds) */
    /**
     * A human-readable string giving imaginative representation of image
     * source. Image source here does not only mean Image Source in VSv3 terms
     * (position of image acquire location), but might also contain a string
     * describing file name where the image was taken from (if loaded).
     * <p>
     * In C/C++ world, this header member is a 64 char array (64 bytes). It may
     * not contain a null byte at the end, but 64 truly useful characters!
     */
    public String cameraPortName  = DEFAULT_CAMERA_PORT_NAME; // [64];
    /**
     * Sets the totalLength member. This function never fails.
     * 
     * @param lengthInBytes
     *          - passed totalLength value
     * @deprecated mdavid: {@link #totalLength} declared as public
     * @see #totalLength
     */
    public void setTotalLength(long lengthInBytes)
    {
      totalLength = lengthInBytes;
    }
    /**
     * Sets the timestamp member variables (UTC seconds and microseconds). This
     * function never fails.
     * 
     * @param time
     *          Contains UTC time*1000 plus fractional millisecond offset*1000.
     * @see #timestampSeconds
     * @see #timestampMicroseconds
     */
    public void setTimeStamp(long time)
    {
      timestampSeconds = (int) (time / 1000);
      timestampMicroseconds = (int) ((time % 1000) * 1000);
    }
    // mdavid: commented
    // public void setStaticValues(int baseId, int specId, int version, String
    // cameraPort) {
    // this.baseTag = baseId;
    // this.cameraPortId = specId;
    // versionTag = version;
    // cameraPortName = cameraPort;
    // }
    /**
     * @deprecated mdavid: unused, can be replaced by constructor
     */
    public void setStaticValues(int baseId, long specId, long version, String cameraPort)
    {
      this.baseTag = baseId;
      this.cameraPortId = specId;
      versionTag = version;
      cameraPortName = cameraPort;
    }
    // mdavid: added
    @Override
    public String toString()
    {
      StringBuilder sb = new StringBuilder();
      sb.append("baseTag = ").append(baseTag);
      sb.append(", cameraPortId = ").append(cameraPortId);
      sb.append(", cameraPortName = ").append(cameraPortName);
      sb.append(", timestampMicroseconds = ").append(timestampMicroseconds);
      sb.append(", timestampSeconds = ").append(timestampSeconds);
      sb.append(", totalLength = ").append(totalLength);
      sb.append(", versionTag = ").append(versionTag);
      return sb.toString();
    }
    // mdavid: added
    @Override
    public SourceHeader clone() //throws CloneNotSupportedException
    {
      SourceHeader header = new SourceHeader();
      header.baseTag = baseTag;
      header.cameraPortId = cameraPortId;
      header.cameraPortName = cameraPortName;
      header.timestampMicroseconds = timestampMicroseconds;
      header.timestampSeconds = timestampSeconds;
      header.totalLength = totalLength;
      header.versionTag = versionTag;
      return header;
    }
    /**
     * Reads the <code>SourceHeader</code> from the supplied
     * <code>DataInputStream</code>.
     * 
     * @author mdavid
     * @param input
     *          - an <code>DataInputStream</code> to read from.
     * @exception IOException
     *              if an error occurs during reading.
     */
    protected void read(DataInputStream dis) throws IOException
    {
      baseTag = Swap.Int(dis.readInt());
      cameraPortId = UnsignedUtils.toUnsignedInt(Swap.Int(dis.readInt())); // mdavid:
                                                                            // changed
      versionTag = UnsignedUtils.toUnsignedInt(Swap.Int(dis.readInt())); // mdavid:
                                                                          // changed
      totalLength = UnsignedUtils.toUnsignedInt(Swap.Int(dis.readInt())); // mdavid:
                                                                           // changed
      timestampSeconds = Swap.Int(dis.readInt());
      timestampMicroseconds = Swap.Int(dis.readInt());
      byte[] b = new byte[64];
      dis.read(b, 0, 64);
      cameraPortName = (new String(b)).trim();
    }
    /**
     * @author mdavid
     * @param ds
     *          - output stream
     * @throws IOException
     */
    protected void write(DataOutputStream ds) throws IOException
    {
      ds.writeInt(Swap.Int(baseTag));
      ds.writeInt(Swap.Int(cameraPortId));
      ds.writeInt(Swap.Int(versionTag));
      ds.writeInt(Swap.Int(totalLength));
      ds.writeInt(Swap.Int(timestampSeconds));
      ds.writeInt(Swap.Int(timestampMicroseconds));
      ds.write(cameraPortName.getBytes());
      ds.write(new byte[64], 0, 64 - cameraPortName.length());
    }
  }
  /**
   * Subclass of second (and last) part of header (subclass used for logical
   * grouping of class member variables which map almost 1:1 to header
   * structures physically transferred in binary form.
   * <p>
   * This class public member variables are mainly a clone of the
   * TTF2_IMAGE_STRUCT defined and agreed in 2008 between ERS, MCS 1 and MCS 4.
   */
  public class FrameHeader
  {
    /**
     * Width in pixels of the source image (if aoi is specified) or the raw
     * appended image size. The variable is unnecessary if an aoi is set and
     * very necessary otherwise. See aoiWidth, aoiHeight, xStart, yStart and
     * sourceWidth.
     * <p>
     * In C/C++ world, this member variable is a signed int32. If it is unset
     * (unlikely, but possible), value is '-1'.
     */
    public int sourceWidth;
    /**
     * Width in pixels of the source image (if aoi is specified) or the raw
     * appended image size. The variable is unnecessary if an aoi is set and
     * very necessary otherwise. See aoiWidth, aoiHeight, xStart, yStart and
     * sourceHeight.
     * <p>
     * In C/C++ world, this member variable is a signed int32. If it is unset
     * (unlikely, but possible), value is '-1'.
     */
    public int sourceHeight;
    /**
     * Width in pixels of the area of interest. A special mechanism is
     * implemented in a CF_IMAGE. If at the image source it can be decided that
     * only a part of the sensor is read out or contains significant information
     * (invaluable surroundings (mounting plates etc.) does not need to get
     * transferred to remote peers), not the whole image may be transferred but
     * only a rectangular region out of it. Even this rectangular piece is break
     * down into two pieces: horizontal truncation (sourceWidth, aoiWidth and
     * xStart) and vertical truncation (sourceHeight, aoiHeight and yStart).
     * <p>
     * If aoiWidth is specified (bigger than zero), it defines the number of
     * horizontal pixels that are attached to the header as frame. The
     * sourceWidth member in that case should be bigger than aoiWidth and
     * defines the source width in pixels (at source level) before an aoi was
     * cut out. The whole mechanism is referred to sometimes as 'aoi cutout'.
     * <p>
     * If aoiWidth is set to a value of 0 or less, it is treated as NOT set. If
     * xStart is bigger than zero the aoi is considered to be set in horizontal
     * dimension. If aoiWidth is zero or less in that case, the header is
     * considered to be NOT valid!
     * <p>
     * In C/C++ world, this member variable is a signed int32. If it is unset
     * (very possible), value is less than or equal to zero. A recommendation
     * for unset value is '-1', though.
     */
    public int aoiWidth = -1; // default: -1 mdavid!
    /**
     * Height in pixels of the area of interest. A special mechanism is
     * implemented in a CF_IMAGE. If at the image source it can be decided that
     * only a part of the sensor is read out or contains significant information
     * (invaluable surroundings (mounting plates etc.) does not need to get
     * transferred to remote peers), not the whole image may be transferred but
     * only a rectangular region out of it. Even this rectangular piece is break
     * down into two pieces: horizontal truncation (sourceWidth, aoiWidth and
     * xStart) and vertical truncation (sourceHeight, aoiHeight and yStart).
     * <p>
     * If aoiHeight is specified (bigger than zero), it defines the number of
     * vertical pixels that are attached to the header as frame. The
     * sourceHeight member in that case should be bigger than aoiHeight and
     * defines the source width in pixels (at source level) before an aoi was
     * cut out. The whole mechanism is referred to sometimes as 'aoi cutout'.
     * <p>
     * If aoiHeight is set to a value of 0 or less, it is treated as NOT set. If
     * yStart is bigger than zero the aoi is considered to be set in horizontal
     * dimension. If aoiHeight is zero or less in that case, the header is
     * considered to be NOT valid!
     * <p>
     * In C/C++ world, this member variable is a signed int32. If it is unset
     * (very possible), value is less than or equal to zero. A recommendation
     * for unset value is '-1', though.
     */
    public int aoiHeight = -1; // default: -1 mdavid!
    /**
     * Pixel index location (counting starts from zero) of left index from
     * top-left position if an aoi is set. If no aoi is set (here: aoiWidth <=
     * 0); xStart is defined to be exactly zero. If it is not, the header is
     * invalid. xStart plus aoiWidth must be less than or equal to sourceWidth.
     * In C/C++ world, this member variable is a signed int32. If it is unset
     * (very possible), value must be zero. A value of less than zero is
     * invalid!
     * 
     * @see #aoiWidth
     * @see #sourceWidth
     * @see #aoiHeight
     * @see #yStart
     * @see #sourceHeight
     */
    public int xStart = 0; // default: 0 mdavid!
    /**
     * Pixel index location (counting starts from zero) of top index from
     * top-left position if an aoi is set. If no aoi is set (here: aoiHeight <=
     * 0); xStart is defined to be exactly zero. If it is not, the header is
     * invalid. yStart plus aoiHeight must be less than or equal to
     * sourceHeight. In C/C++ world, this member variable is a signed int32. If
     * it is unset (very possible), value must be zero. A value of less than
     * zero is invalid!
     * 
     * @see #aoiHeight
     * @see #sourceHeight
     * @see #aoiWidth
     * @see #xStart
     * @see #sourceWidth
     */
    public int yStart = 0; // default: 0 mdavid!
    /**
     * Specifies amount of bytes per pixel. Currently defined values are 1
     * (exactly one byte per pixel), 2 (exactly two bytes per pixel), 3 (three
     * bytes per pixel) and 4 (four bytes per pixel). Fractional bytes per pixel
     * (as with packed 12 bit grayscale values) are not supported at the moment.
     * It is foreseen to use the '0' to specify that.
     * <p>
     * Generally speaking, header version 1 supports raw grayscale images of 1,2
     * and 3 bytes per pixel and raw RGB colour images of 3 or 4 bytes per pixel
     * (24 bits or 24 bits padded to 32 bits).
     * <p>
     * In C/C++ world, this member variable is a signed int32. If it is unset
     * (unlikely), value must be '-1'. Values of less than one or bigger than
     * four are treated as illegal for header version 1.
     */
    public int bytesPerPixel;
    /**
     * Specifies amount of effective bits per pixel of appended image bits. For
     * performance reasons, physical resolution of 12 bits per pixel (in example
     * Prosilica GE / GC-type monochrome cameras) is wrapped into 16 bit
     * integers for improving handling by computers. For such scheme effective
     * bits per pixel value was invented to distinguish valid from invalid pixel
     * values and to have more input for image processing algorithms. The value
     * should be set appropriate. If it is not set, it is believed to be set to
     * the maximum value the current data format is capable of. If the value is
     * bigger than theoretically possible for a given bytesPerPixel member, the
     * header is considered to be invalid.
     * <p>
     * In C/C++ world, this member variable is a signed int32. If it is unset
     * (unlikely), value must be '-1'. A value of zero is treated as 'unset' for
     * header version 1.
     */
    public int effectiveBitsPerPixel;
    /**
     * Pure informal value describes whether integer binning in horizontal
     * direction was performed on the image. It is not said whether this binning
     * was performed by hardware or software.
     * <ul>
     * <li>0 - binning status unknown
     * <li>1 - no binning (known status!)
     * <li>2 - 1 pixel was created out of 2 pixels (2:1)
     * <li>3 - 1 pixel was created out of 3 pixels (3:1)
     * <li>4 - 4:1
     * <li>...
     * </ul>
     * <p>
     * In C/C++ world, this member variable is a signed int32. If it is unset
     * (very likely), value must be '0'. If value is less than zero header is
     * considered invalid.
     */
    public int horizontalBinning = 0; // default: 0 mdavid!
    /**
     * Pure informal value describes whether integer binning in vertical
     * direction was performed on the image. It is not said whether this binning
     * was performed by hardware or software.
     * <ul>
     * <li>0 - binning status unknown
     * <li>1 - no binning (known status!)
     * <li>2 - 1 pixel was created out of 2 pixels (2:1)
     * <li>3 - 1 pixel was created out of 3 pixels (3:1)
     * <li>4 - 4:1
     * <li>...
     * </ul>
     * In C/C++ world, this member variable is a signed int32. If it is unset
     * (very likely), value must be '0'. If value is less than zero, header is
     * considered invalid.
     */
    public int verticalBinning = 0; // default: 0 mdavid!
    /**
     * Rather informal value declaring the image format taken by camera hardware
     * at image source. Once set at image acquire level, this member should not
     * be changed. One practical case is that colour information was originally
     * delivered by sensor as bayer pattern but decoded by image acquire
     * software to 24 bit RGB data, which was subsequently be appended to this
     * image header. The image source format should be bayer in that case.
     * Certain formats are implemented at the moment, see {@link #imageFormat}.
     * <p>
     * In C/C++ world, this member variable is a signed int32. If it is unset
     * (unlikely), value must be '-1'. If value is less than -1, header is
     * considered invalid.
     */
    public int sourceFormat;
    /**
     * Very important member declaring the binary format in which image bits,
     * appended to this header taken by camera, are encoded. In general, its
     * value can be very different from sourceFormat value. It should always
     * reflect encoding type of image bits. Various formats can be set:
     * <ul>
     * <li>0 - Grayscale
     * <li>1 - RGB (24bpp)
     * <li>2 - RGB (32bpp)
     * <li>3 - RGB (15bpp)*
     * <li>4 - RGB (16bpp)*
     * <li>5 - JPEG
     * <li>6 - TIFF*
     * <li>7 - YUV411*
     * <li>8 - YUV422*
     * <li>9 - YUV444*
     * <li>10 - HUFFYUV
     * <li>11 - BMP (Windows BMP)*
     * <li>11 - BAYER*
     * </ul>
     * <p>
     * - not specified in more detail up to now, should be used as informal
     * values at {@link #sourceFormat} only
     * <p>
     * In C/C++ world, this member variable is a signed int32. If it is unset
     * (very unlikely), value must be '-1'. If value is less than -1, header is
     * considered invalid.
     */
    public int imageFormat;
    /**
     * Specifies unique frame identification number. This number is not really
     * unique. Starting at one, for each frame (as early as possible after
     * acquiring, maybe even done inside camera hardware) the image is tagged
     * with this number. The next image (no losses in between!) should be tagged
     * with number+1 and so on. By performing this method, it can be checked
     * later whether a sequence contains really subsequent frames as they were
     * acquired at source level and no video frame is lost in between.
     * <p>
     * In C/C++ world, this member variable is an unsigned int32. If it is
     * unset, value must be '0'.
     */
    public long frameNumber; // uint32->long sw!
    /**
     * Specifies event number when frame was taken. In some modern timing
     * systems, together with the main repetition rate (triggered accelerator
     * like PITZ) an information is modulated on the trigger signal and properly
     * decoded to computer registers by timing hardware. This information can be
     * helpful in joining or synchronization cases. One use-case is DAQ
     * correlation
     * ("where did the lightning hit first before a lot of systems went crazy").
     * <p>
     * The header members eventNumber, frameNumber, timeStamp and cameraPortId
     * build up a primary key which should identify an image throughout a whole
     * facility for the lifetime of the facility.
     * <p>
     * In C/C++ world, this member variable is an unsigned int32. If it is
     * unset, value must be '0'. Any positive value is considered a valid event
     * number and usually event number are incremented like frame numbers, but
     * for each event instead of each subsequent frame.
     */
    public long eventNumber; // uint32->long sw!
    /**
     * Ratio between pixels and millimetres for x-pixel dimension. If xScale is
     * 3.0, one pixel is 3 mm "wide". If unset or unspecified, value must be set
     * to ~ -1.0. Any negative value except ~ -1.0 is treated as faulty. A value
     * of ~ 0.0 is considered to be illegal.
     * <p>
     * In C/C++ world, this member is a standard 32 bit floating point number.
     */
    public float xScale = -1F; // default: -1 mdavid!
    /**
     * Ratio between pixels and millimetres for x-pixel dimension. If yScale is
     * 2.0, one pixel is 2 mm "tall". If unset or unspecified, value must be set
     * to ~ -1.0. Any negative value except ~ -1.0 is treated as faulty. A value
     * of ~ 0.0 is considered to be illegal.
     * <p>
     * In C/C++ world, this member is a standard 32 bit floating point number.
     */
    public float yScale = -1F; // default: -1 mdavid!
    /**
     * Specifies the rotation balance (compare: account balance, saldo in
     * German) that was applied since image acquire at source level. For example
     * lossless orientation changes can be performed in CoreProvider TINE Server
     * video component to undo perspective changes introduced because of awkward
     * or improper mounting of camera hardware at readout position. This member
     * is always initialized to 0.0 (no rotation performed). Positive values
     * specify clockwise rotation, negative values specify counterclockwise
     * rotation. Regarding value of this field, a general agreement is to keep
     * it in range of -360.0&nbsp;<&nbsp;imageRotation&nbsp;<&nbsp;360.0 .
     * <p>
     * In C/C++ world, this member is a standard 32 bit floating point number.
     */
    public float imageRotation = 0F; // default: 0F mdavid!
    /**
     * First out of three spare floats eserved for future extensions. Current
     * value of this float should be '-1.0' (uninitialized). In
     * <p>
     * C/C++ world, this member is a standard 32 bit floating point number.
     */
    public float fspare1 = -1F; // default: -1F mdavid!
    /**
     * Second out of three spare floats reserved for future extensions. Current
     * value of this float should be '-1.0' (uninitialized). In
     * <p>
     * C/C++ world, this member is a standard 32 bit floating point number.
     */
    public float fspare2 = -1F; // default: -1F mdavid!
    /**
     * Third out of three spare floats reserved for future extensions. Current
     * value of this float should be '-1.0' (uninitialized). In
     * <p>
     * C/C++ world, this member is a standard 32 bit floating point number.
     */
    public float fspare3 = -1; // default: -1F mdavid!
    /**
     * Contains flags that describe boolean 'states' of an image. Each state is
     * encoded into one bit. Up to now, four bits are already used (LSB 1, MSB
     * 32):
     * <ul>
     * <li>0: if set, image bits are appended in little endian byte order, big
     * endian otherwise
     * <li>1: if set, image acquired from source is possible to restore out of
     * data contained within losslessly. If unset, a lossy change was applied on
     * the way...
     * <li>2: if set, image was flipped along horizontal axis on its way. If bit
     * 2 and 3 are set, horizontal flipping takes precedence over vertical
     * flipping. Any rotation degree, if specified, is performed before any of
     * the flipping.
     * <li>3: if set, image was flipped along vertical axis on its way. If bit 2
     * and 3 are set, horizontal flipping takes precedence over vertical
     * flipping. Any rotation degree, if specified, is performed before any of
     * the flipping.
     * <li>4 to 32: reserved, should be unset
     * </ul>
     * <p>
     * In C/C++ world, this member is a standard 32 bit signed integer.
     */
    public int imageFlags = 0x1; // default: 0x1 (little_endian_byte_order)
                                 // mdavid!
    /**
     * First out of three spare integers reserved for future extensions. Current
     * value should be '-1' (uninitialized).
     * <p>
     * In C/C++ world, this member is a signed 32 bit integer.
     */
    public int ispare1 = -1; // default: -1 mdavid!
    /**
     * Second out of three spare integers reserved for future extensions.
     * Current value should be '-1' (uninitialized).
     * <p>
     * In C/C++ world, this member is a signed 32 bit integer.
     */
    public int ispare2 = -1; // default: -1 mdavid!
    /**
     * Third out of three spare integers reserved for future extensions. Current
     * value should be '-1' (uninitialized).
     * <p>
     * In C/C++ world, this member is a signed 32 bit integer.
     */
    public int ispare3 = -1; // default: -1 mdavid!
    /**
     * Contains number of bytes of 'image bits' data that follows the CF_IMAGE
     * header. A value of -1 means that the member is not yet initialized. A
     * value of 0 means that no bytes are appended at all. A positive integer
     * value is equal to the number of bytes that are appended. A value of less
     * than -1 is treated as invalid!
     * <p>
     * In C/C++ world, this member is a signed 32 bit integer.
     */
    public int appendedFrameSize;
    // mdavid: added
    @Override
    public String toString()
    {
      StringBuilder sb = new StringBuilder();
      sb.append("aoiHeight = ").append(aoiHeight);
      sb.append(", aoiWidth = ").append(aoiWidth);
      sb.append(", appendedFrameSize = ").append(appendedFrameSize);
      sb.append(", bytesPerPixel = ").append(bytesPerPixel);
      sb.append(", effectiveBitsPerPixel = ").append(effectiveBitsPerPixel);
      sb.append(", eventNumber = ").append(eventNumber);
      sb.append(", frameNumber = ").append(frameNumber);
      sb.append(", fspare1 = ").append(fspare1);
      sb.append(", fspare2 = ").append(fspare2);
      sb.append(", fspare3 = ").append(fspare3);
      sb.append(", horizontalBinning = ").append(horizontalBinning);
      sb.append(", imageFlags = ").append(imageFlags);
      sb.append(", imageFormat = ").append(imageFormat);
      sb.append(", imageRotation = ").append(imageRotation);
      sb.append(", ispare1 = ").append(ispare1);
      sb.append(", ispare2 = ").append(ispare2);
      sb.append(", ispare3 = ").append(ispare3);
      sb.append(", sourceFormat = ").append(sourceFormat);
      sb.append(", sourceHeight = ").append(sourceHeight);
      sb.append(", sourceWidth = ").append(sourceWidth);
      sb.append(", verticalBinning = ").append(verticalBinning);
      sb.append(", xScale = ").append(xScale);
      sb.append(", xStart = ").append(xStart);
      sb.append(", yScale = ").append(yScale);
      sb.append(", yStart = ").append(yStart);
      return sb.toString();
    }
    // mdavid: added
    @Override
    public FrameHeader clone() //throws CloneNotSupportedException
    {
      FrameHeader header = new FrameHeader();
      header.aoiHeight = aoiHeight;
      header.aoiWidth = aoiWidth;
      header.appendedFrameSize = appendedFrameSize;
      header.bytesPerPixel = bytesPerPixel;
      header.effectiveBitsPerPixel = effectiveBitsPerPixel;
      header.eventNumber = eventNumber;
      header.frameNumber = frameNumber;
      header.fspare1 = fspare1;
      header.fspare2 = fspare2;
      header.fspare3 = fspare3;
      header.horizontalBinning = horizontalBinning;
      header.imageFlags = imageFlags;
      header.imageFormat = imageFormat;
      header.imageRotation = imageRotation;
      header.ispare1 = ispare1;
      header.ispare2 = ispare2;
      header.ispare3 = ispare3;
      header.sourceFormat = sourceFormat;
      header.sourceHeight = sourceHeight;
      header.sourceWidth = sourceWidth;
      header.verticalBinning = verticalBinning;
      header.xScale = xScale;
      header.xStart = xStart;
      header.yScale = yScale;
      header.yStart = yStart;
      return header;
    }
    /**
     * @author mdavid
     * @param dis
     *          - input stream
     * @throws IOException
     */
    protected void read(DataInputStream dis) throws IOException
    {
      sourceWidth = Swap.Int(dis.readInt());
      sourceHeight = Swap.Int(dis.readInt());
      aoiWidth = Swap.Int(dis.readInt());
      aoiHeight = Swap.Int(dis.readInt());
      xStart = Swap.Int(dis.readInt());
      yStart = Swap.Int(dis.readInt());
      bytesPerPixel = Swap.Int(dis.readInt());
      effectiveBitsPerPixel = Swap.Int(dis.readInt());
      horizontalBinning = Swap.Int(dis.readInt());
      verticalBinning = Swap.Int(dis.readInt());
      sourceFormat = Swap.Int(dis.readInt());
      imageFormat = Swap.Int(dis.readInt());
      frameNumber = UnsignedUtils.toUnsignedInt(Swap.Int(dis.readInt())); // mdavid:
                                                                           // changed
      eventNumber = UnsignedUtils.toUnsignedInt(Swap.Int(dis.readInt())); // mdavid:
                                                                           // changed
      byte[] b = new byte[4];
      dis.read(b, 0, 4);
      xScale = Swap.Float(b);
      dis.read(b, 0, 4);
      yScale = Swap.Float(b);
      dis.read(b, 0, 4);
      imageRotation = Swap.Float(b);
      dis.read(b, 0, 4);
      fspare1 = Swap.Float(b);
      dis.read(b, 0, 4);
      fspare2 = Swap.Float(b);
      dis.read(b, 0, 4);
      fspare3 = Swap.Float(b);
      imageFlags = Swap.Int(dis.readInt());
      ispare1 = Swap.Int(dis.readInt());
      ispare2 = Swap.Int(dis.readInt());
      ispare3 = Swap.Int(dis.readInt());
      appendedFrameSize = Swap.Int(dis.readInt());
    }
    /**
     * @author mdavid
     * @param ds
     *          - output stream
     * @throws IOException
     */
    protected void write(DataOutputStream ds) throws IOException
    {
      ds.writeInt(Swap.Int(sourceWidth));
      ds.writeInt(Swap.Int(sourceHeight));
      ds.writeInt(Swap.Int(aoiWidth));
      ds.writeInt(Swap.Int(aoiHeight));
      ds.writeInt(Swap.Int(xStart));
      ds.writeInt(Swap.Int(yStart));
      ds.writeInt(Swap.Int(bytesPerPixel));
      ds.writeInt(Swap.Int(effectiveBitsPerPixel));
      ds.writeInt(Swap.Int(horizontalBinning));
      ds.writeInt(Swap.Int(verticalBinning));
      ds.writeInt(Swap.Int(sourceFormat));
      ds.writeInt(Swap.Int(imageFormat));
      ds.writeInt(Swap.Int(frameNumber));
      ds.writeInt(Swap.Int(eventNumber));
      ds.write(Swap.Float(xScale));
      ds.write(Swap.Float(yScale));
      ds.write(Swap.Float(imageRotation));
      ds.write(Swap.Float(fspare1));
      ds.write(Swap.Float(fspare2));
      ds.write(Swap.Float(fspare3));
      ds.writeInt(Swap.Int(imageFlags));
      ds.writeInt(Swap.Int(ispare1));
      ds.writeInt(Swap.Int(ispare2));
      ds.writeInt(Swap.Int(ispare3));
      ds.writeInt(Swap.Int(appendedFrameSize));
    }
  }
}
// ****** mdavid: replaced by "SourceHeader.read(DataInputStream)" method
// private void readImageSourceHeader(DataInputStream dis) throws IOException {
// sourceHeader.baseTag = Swap.Long(dis.readInt());
//
// // sourceHeader.cameraPortId = UnsignedInt(Swap.Long(dis.readInt())); //
// mdavid: commented
// // sourceHeader.versionTag = UnsignedInt(Swap.Long(dis.readInt())); //
// mdavid: commented
// // sourceHeader.totalLength = UnsignedInt(Swap.Long(dis.readInt())); //
// mdavid: commented
//
// sourceHeader.cameraPortId =
// UnsignedUtils.toUnsignedInt(Swap.Long(dis.readInt())); // mdavid: changed
// sourceHeader.versionTag =
// UnsignedUtils.toUnsignedInt(Swap.Long(dis.readInt())); // mdavid: changed
// sourceHeader.totalLength =
// UnsignedUtils.toUnsignedInt(Swap.Long(dis.readInt())); // mdavid: changed
//
// sourceHeader.timestampSeconds = Swap.Long(dis.readInt());
// sourceHeader.timestampMicroseconds = Swap.Long(dis.readInt());
//
// byte[] b = new byte[64];
// dis.read(b, 0, 64);
// sourceHeader.cameraPortName = (new String(b)).trim();
// }
//
// ****** mdavid: replaced by "SourceHeader.write(DataOutputStream)" method
// private void writeImageSourceHeader(DataOutputStream ds) throws IOException {
// ds.writeInt(Swap.Long(sourceHeader.baseTag));
// ds.writeInt(Swap.Long(sourceHeader.cameraPortId));
// ds.writeInt(Swap.Long(sourceHeader.versionTag));
// ds.writeInt(Swap.Long(sourceHeader.totalLength));
// ds.writeInt(Swap.Long(sourceHeader.timestampSeconds));
// ds.writeInt(Swap.Long(sourceHeader.timestampMicroseconds));
// // byte[] b = new byte[64]; // mdavid: commented
// ds.write(sourceHeader.cameraPortName.getBytes());
// ds.write(new byte[64]/* b */, 0, 64 - sourceHeader.cameraPortName.length());
// }
//
// ****** mdavid: replaced by "FrameHeader.read(DataInputStream)" method
// private void readImageFrameHeader(DataInputStream dis) throws IOException {
// frameHeader.sourceWidth = Swap.Long(dis.readInt());
// frameHeader.sourceHeight = Swap.Long(dis.readInt());
// frameHeader.aoiWidth = Swap.Long(dis.readInt());
// frameHeader.aoiHeight = Swap.Long(dis.readInt());
// frameHeader.xStart = Swap.Long(dis.readInt());
// frameHeader.yStart = Swap.Long(dis.readInt());
// frameHeader.bytesPerPixel = Swap.Long(dis.readInt());
// frameHeader.effectiveBitsPerPixel = Swap.Long(dis.readInt());
// frameHeader.horizontalBinning = Swap.Long(dis.readInt());
// frameHeader.verticalBinning = Swap.Long(dis.readInt());
// frameHeader.sourceFormat = Swap.Long(dis.readInt());
// frameHeader.imageFormat = Swap.Long(dis.readInt());
//
// // frameHeader.frameNumber = UnsignedInt(Swap.Long(dis.readInt())); //
// mdavid: commented
// // frameHeader.eventNumber = UnsignedInt(Swap.Long(dis.readInt())); //
// mdavid: commented
//
// frameHeader.frameNumber =
// UnsignedUtils.toUnsignedInt(Swap.Long(dis.readInt())); // mdavid: changed
// frameHeader.eventNumber =
// UnsignedUtils.toUnsignedInt(Swap.Long(dis.readInt())); // mdavid: changed
//
// byte[] b = new byte[4];
// dis.read(b, 0, 4);
// frameHeader.xScale = Swap.Float(b);
// dis.read(b, 0, 4);
// frameHeader.yScale = Swap.Float(b);
// dis.read(b, 0, 4);
// frameHeader.imageRotation = Swap.Float(b);
// dis.read(b, 0, 4);
// frameHeader.fspare1 = Swap.Float(b);
// dis.read(b, 0, 4);
// frameHeader.fspare2 = Swap.Float(b);
// dis.read(b, 0, 4);
// frameHeader.fspare3 = Swap.Float(b);
// frameHeader.imageFlags = Swap.Long(dis.readInt());
// frameHeader.ispare1 = Swap.Long(dis.readInt());
// frameHeader.ispare2 = Swap.Long(dis.readInt());
// frameHeader.ispare3 = Swap.Long(dis.readInt());
// frameHeader.appendedFrameSize = Swap.Long(dis.readInt());
// }
//
// ******* mdavid: replaced by "FrameHeader.write(DataOutputStream)" method
// private void writeImageFrameHeader(DataOutputStream ds) throws IOException {
// ds.writeInt(Swap.Long(frameHeader.sourceWidth));
// ds.writeInt(Swap.Long(frameHeader.sourceHeight));
// ds.writeInt(Swap.Long(frameHeader.aoiWidth));
// ds.writeInt(Swap.Long(frameHeader.aoiHeight));
// ds.writeInt(Swap.Long(frameHeader.xStart));
// ds.writeInt(Swap.Long(frameHeader.yStart));
// ds.writeInt(Swap.Long(frameHeader.bytesPerPixel));
// ds.writeInt(Swap.Long(frameHeader.effectiveBitsPerPixel));
// ds.writeInt(Swap.Long(frameHeader.horizontalBinning));
// ds.writeInt(Swap.Long(frameHeader.verticalBinning));
// ds.writeInt(Swap.Long(frameHeader.sourceFormat));
// ds.writeInt(Swap.Long(frameHeader.imageFormat));
// ds.writeInt(Swap.Long(frameHeader.frameNumber));
// ds.writeInt(Swap.Long(frameHeader.eventNumber));
// ds.write(Swap.Float(frameHeader.xScale));
// ds.write(Swap.Float(frameHeader.yScale));
// ds.write(Swap.Float(frameHeader.imageRotation));
// ds.write(Swap.Float(frameHeader.fspare1));
// ds.write(Swap.Float(frameHeader.fspare2));
// ds.write(Swap.Float(frameHeader.fspare3));
// ds.writeInt(Swap.Long(frameHeader.imageFlags));
// ds.writeInt(Swap.Long(frameHeader.ispare1));
// ds.writeInt(Swap.Long(frameHeader.ispare2));
// ds.writeInt(Swap.Long(frameHeader.ispare3));
// ds.writeInt(Swap.Long(frameHeader.appendedFrameSize));
// }
//
// implementation of TCompoundDataObject interface further on
//
// mdavid: commnented
// public byte[] toByteArray() {
// try {
// int thisSizeInBytes = TFormat.getFormatHeaderSize(TFormat.CF_IMAGE);
// if (imageFrame != null) {
// int frameSizeInBytes = imageFrame.length;
// // TODO: if java video server, may need to worry about limiting
// // the byte array length to the net size of the frame
// // if (frameHeader.appendedFrameSize < imageFrame.length)
// // {
// // frameSizeInBytes = frameHeader.appendedFrameSize;
// // }
// thisSizeInBytes += frameSizeInBytes;
//
// // dBuffer = new ByteArrayOutputStream(thisSizeInBytes); // mdavid: commented
// ByteArrayOutputStream dBuffer = new ByteArrayOutputStream(thisSizeInBytes);
// // mdavid: changed
// DataOutputStream ds = new DataOutputStream(dBuffer);
// writeImageSourceHeader(ds);
// writeImageFrameHeader(ds);
// ds.write(imageFrame, 0, frameSizeInBytes);
// ds.close();
// return dBuffer.toByteArray();
// }
//
// } catch (IOException e) {
// e.printStackTrace();
// MsgLog.log("IMAGE", e.getMessage(), TErrorList.code_failure, e, 0);
// }
// return null;
// }