/*
 * File MDA.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.types;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.lang.reflect.Array;
import java.nio.charset.Charset;
//import java.nio.charset.StandardCharsets;  // java 1.7
import java.util.Arrays;
import de.desy.tine.dataUtils.TDataType;
import de.desy.tine.definitions.TErrorList;
import de.desy.tine.definitions.TFormat;
import de.desy.tine.endianUtils.Swap;
import de.desy.tine.exceptions.TineRuntimeErrorException;
import de.desy.tine.server.logger.MsgLog;

public class MDA implements TCompoundDataObject
{
  public final static int COMMENT_SIZE = 80;
  public final static int MAX_MDA_DIMS = 6;
  public final static int HDR_SIZE = 376;
  public final static int mda_fmt_offset = 84;
  private byte[] hByteArray;
  private ByteArrayOutputStream dBuffer;
  private int sizeInBytes;
  //private byte[] hdrbytes = new byte[HDR_SIZE];
  private Charset tineCharset = Charset.forName("US-ASCII");
  public class MDX
  {
    public final static int MDX_SIZE = 48;
    private int len;        /**< the number of elements in this dimension */
    private int reserved = 0; /**< reserved field */
    private byte[] lbl = new byte[16];/**< the axis label */
    private String lblStr = null;
    private byte[] egu = new byte[16]; /**< the axis units */
    private String eguStr = null;
    private float min;      /**< the axis value of the initial element */
    private float max;      /**< the axis value of the last element */
    private byte[] hByteArray = new byte[MDX_SIZE];
    private ByteArrayOutputStream dBuffer = null;
    public MDX(int index,int length,String label,String units,float minValue,float maxValue)
    {
      len = length;
      if (label == null) label = "";
      if (units == null) units = "";
      int slen = label.length();
      if (slen > 16) slen = 16;
      byte[] b = label.getBytes(tineCharset);
      Arrays.fill(lbl, (byte)0);
      System.arraycopy(b, 0, lbl, 0, slen);
      lblStr = new String(lbl);
      slen = units.length();
      if (slen > 16) slen = 16;
      b = units.getBytes(tineCharset);
      Arrays.fill(egu, (byte)0);
      System.arraycopy(b, 0, egu, 0, slen);
      eguStr = new String(egu);
      min = minValue;
      max = maxValue;
      if (index >= 0 && index < MAX_MDA_DIMS) axis[index] = this;
    }
    public MDX(int length,String label,String units,float minValue,float maxValue)
    {
      this(-1,length,label,units,minValue,maxValue);
    }
    public MDX(MDX instance)
    {
      this(instance.len,instance.lblStr,instance.eguStr,instance.min,instance.max);      
    }
    public MDX()
    {
      len = 0;
      max = min = 0;
    }
    public int getLength() { return len; }
    public float getMinValue() { return min; }
    public float getMaxValue() { return max; }
    public String getLabel() { return lblStr; }
    public String getUnits() { return eguStr; }
    public byte[] toByteArray()
    {
      try
      {
        if (dBuffer == null)
          dBuffer = new ByteArrayOutputStream(MDX_SIZE);
        else
          dBuffer.reset();
        DataOutputStream ds = new DataOutputStream(dBuffer);
        ds.writeInt(Swap.Int(len));
        ds.writeInt(Swap.Int(reserved));
        ds.write(lbl);
        ds.write(egu);
        ds.write(Swap.Float(min));
        ds.write(Swap.Float(max));
        ds.close();
      }
      catch (Exception e)
      {
        e.printStackTrace();
        MsgLog.log("MDX.toByteArray", "could not get MDA axis bytes",TErrorList.code_failure,e,0);
      }
      hByteArray = dBuffer.toByteArray();
      return hByteArray;
    }
    public void toStruct()
    {
      try
      {
        ByteArrayInputStream dis = new ByteArrayInputStream(hByteArray);
        DataInputStream ds = new DataInputStream(dis);
        len = Swap.Int(ds.readInt());
        reserved = Swap.Int(ds.readInt());
        Arrays.fill(lbl, (byte)0);
        ds.read(lbl, 0, 16);      
        Arrays.fill(egu, (byte)0);
        ds.read(egu, 0, 16);      
        byte[] b = new byte[4];
        ds.read(b,0,4);
        min = Swap.Float(b);
        ds.read(b,0,4);
        max = Swap.Float(b);
        ds.close(); dis.close();
        eguStr = new String(egu);
        lblStr = new String(lbl);
      }
      catch (IOException e)
      {
        e.printStackTrace();
        MsgLog.log("MDX",e.getMessage(),TErrorList.code_failure,e,0);        
      }  
    }
  }
  private byte[] cmt = new byte[COMMENT_SIZE];
  private String cmtStr = null;
  public String getComment() 
  {
    if (cmtStr == null) cmtStr = new String(cmt);
    return cmtStr;
  }
  private int ndim = 0;
  public int getNumberDimension() { return ndim; }
  private int fmt = TFormat.CF_NULL;
  private MDX[] axis = new MDX[MAX_MDA_DIMS];
  public MDX getAxis(int i) 
  {
    if (i<0 || i>= MAX_MDA_DIMS) return null;
    return axis[i];
  }
  private Object data;
  public Object getDataObject() { return data; }
  private int totalLength = 0;
  public void addAxisInformation(int index,MDX axis)
  {
    if (index < 0 || index > ndim || index >= MAX_MDA_DIMS)
      throw new TineRuntimeErrorException(TErrorList.invalid_index);
    this.axis[index] = axis;
  }
  public void setCarriedFormat(int format)
  {
    if (fmt != TFormat.CF_NULL) throw new TineRuntimeErrorException(TErrorList.already_assigned);
    fmt = format;
  }
  public int getCarriedFormat() { return fmt; }
  public MDA(int maximumLength)
  { // set a few things temporarily:
    totalLength = maximumLength;
  }
  public MDA(String comment,Object data)
  {
    if (comment == null) comment = "";
    int slen = comment.length();
    if (slen > COMMENT_SIZE) slen = COMMENT_SIZE;
    byte[] b = comment.getBytes(tineCharset);
    Arrays.fill(cmt, (byte)0);
    System.arraycopy(b, 0, cmt, 0, slen);
    cmtStr = new String(cmt);
    sizeInBytes = HDR_SIZE;
    if (data != null)
    {
      if (!data.getClass().isArray()) throw new TineRuntimeErrorException(TErrorList.dimension_error);
      Object o = data;
      int alen, tlen = 1;
      int ndims = 0;
      ndims = 1 + data.getClass().getName().lastIndexOf('[');
      if (ndims > 6) throw new TineRuntimeErrorException(TErrorList.dimension_error);
      for (int i=0; i<ndims; i++)
      {
        if (i == ndims-1) fmt = TDataType.getObjectFormat(o);
        alen = Array.getLength(o);
        tlen *= alen;
        if (alen > 0) o = Array.get(o, 0);
        axis[i] = new MDX(alen,"axis-"+i,"",0,0);
      }
      ndim = ndims;
      this.data = data;
      sizeInBytes += tlen*TFormat.formatSizeOf((short)fmt);
      totalLength = tlen;
    }
  }
  public MDA(MDA instance)
  {
    this(instance.cmtStr,instance.data);
  }
  public MDA newInstance()
  {
    return new MDA(1000);
  }
  @Override
  public void copy(TCompoundDataObject from)
  {
    if (from instanceof MDA)
    {
      MDA instance = (MDA)from;
      cmtStr = instance.cmtStr; data = instance.data;
    }
  }  
  public boolean dimsEqual(MDA reference)
  {
    if (reference != null)
    {
      if (reference.ndim != ndim) return false;
      for (int i=0; i<ndim; i++)
      {
        if (reference.axis[i].len != axis[i].len) return false;
      }
      return true;
    }
    return false;
  }
  @Override
  public byte[] toByteArray()
  {
    try
    {
      if (dBuffer == null) 
        dBuffer = new ByteArrayOutputStream(sizeInBytes);
      else
        dBuffer.reset();
      DataOutputStream ds = new DataOutputStream(dBuffer);
      ds.write(cmt);
      ds.writeInt(Swap.Int(ndim));
      ds.writeInt(Swap.Int(TFormat.makeStdFormatCode(fmt)));
      for (int i=0; i<MAX_MDA_DIMS; i++)
      {
        if (axis[i] != null && axis[i].len > 0)
          ds.write(axis[i].toByteArray());
        else
          ds.write(new MDX().hByteArray);
      }
      switch (fmt)
      {
        case TFormat.CF_FLOAT:
          for (int i=0; i<totalLength; i++) ds.write(Swap.Float(getFloatElement(i)));
          break;
        case TFormat.CF_DOUBLE:
          for (int i=0; i<totalLength; i++) ds.write(Swap.Double(getDoubleElement(i)));
          break;
        case TFormat.CF_INT32:
          for (int i=0; i<totalLength; i++) ds.writeInt(Swap.Int(getIntElement(i)));
          break;
        case TFormat.CF_INT16:
          for (int i=0; i<totalLength; i++) ds.writeShort(Swap.Short(getShortElement(i)));
          break;
      }
      int nwritten = ds.size();
      if (nwritten < sizeInBytes)
      {
        ds.write(new byte[sizeInBytes-nwritten]);
      }
      ds.close();
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("MDA",e.getMessage(),TErrorList.code_failure,e,0);        
    }
    hByteArray = dBuffer.toByteArray();
    return hByteArray;
  }
  float getFloatElement(int index)
  {
    if (fmt != TFormat.CF_FLOAT) throw new TineRuntimeErrorException(TErrorList.illegal_format);
    int i1,i2,i3,i4,i5,i6;
    if (index < 0 || index > totalLength) throw new TineRuntimeErrorException(TErrorList.invalid_index);
    switch (ndim)
    {
      case 1: return ((float[])data)[index];
      case 2:
        i1 = index/(axis[1].len);
        i2 = index%(axis[1].len);
        return ((float[][])data)[i1][i2];
      case 3:
        if (data == null) data = new float[axis[0].len][axis[1].len][axis[2].len];
        i1 = index/(axis[1].len * axis[2].len);
        i2 = (index%(axis[1].len * axis[2].len))/axis[2].len;
        i3 = (index%(axis[1].len * axis[2].len))%(axis[2].len);
        return ((float[][][])data)[i1][i2][i3];
      case 4:
        if (data == null) data = new float[axis[0].len][axis[1].len][axis[2].len][axis[3].len];
        i1 = index/(axis[1].len * axis[2].len * axis[3].len);
        i2 = (index%(axis[1].len * axis[2].len * axis[3].len))/(axis[2].len * axis[3].len);
        i3 = (index%(axis[1].len * axis[2].len * axis[3].len))%(axis[2].len * axis[3].len)/axis[3].len;
        i4 = (index%(axis[1].len * axis[2].len * axis[3].len))%(axis[2].len * axis[3].len)%(axis[3].len);
        return ((float[][][][])data)[i1][i2][i3][i4];
      case 5:
        if (data == null) data = new float[axis[0].len][axis[1].len][axis[2].len][axis[3].len][axis[4].len];
        i1 = index/(axis[1].len * axis[2].len * axis[3].len * axis[4].len);
        i2 = (index%(axis[1].len * axis[2].len * axis[3].len * axis[4].len))/(axis[2].len * axis[3].len * axis[4].len);
        i3 = (index%(axis[1].len * axis[2].len * axis[3].len * axis[4].len))%(axis[2].len * axis[3].len * axis[4].len)/(axis[3].len * axis[4].len);
        i4 = (index%(axis[1].len * axis[2].len * axis[3].len * axis[4].len))%(axis[2].len * axis[3].len * axis[4].len)%(axis[3].len * axis[4].len)/(axis[4].len);
        i5 = (index%(axis[1].len * axis[2].len * axis[3].len * axis[4].len))%(axis[2].len * axis[3].len * axis[4].len)%(axis[3].len * axis[4].len)%(axis[4].len);
        return ((float[][][][][])data)[i1][i2][i3][i4][i5];
      case 6:
        if (data == null) data = new float[axis[0].len][axis[1].len][axis[2].len][axis[3].len][axis[4].len][axis[5].len];
        i1 = index/(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len);
        i2 = (index%(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len))/(axis[2].len * axis[3].len * axis[4].len * axis[5].len);
        i3 = (index%(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len))%(axis[2].len * axis[3].len * axis[4].len * axis[5].len)/(axis[3].len * axis[4].len * axis[5].len);
        i4 = (index%(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len))%(axis[2].len * axis[3].len * axis[4].len * axis[5].len)%(axis[3].len * axis[4].len * axis[5].len)/(axis[4].len * axis[5].len);
        i5 = (index%(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len))%(axis[2].len * axis[3].len * axis[4].len * axis[5].len)%(axis[3].len * axis[4].len * axis[5].len)%(axis[4].len * axis[5].len)/(axis[5].len);
        i6 = (index%(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len))%(axis[2].len * axis[3].len * axis[4].len * axis[5].len)%(axis[3].len * axis[4].len * axis[5].len)%(axis[4].len * axis[5].len)%(axis[5].len);
        return ((float[][][][][][])data)[i1][i2][i3][i4][i5][i6];
    }
    throw new TineRuntimeErrorException(TErrorList.dimension_error);
  }
  double getDoubleElement(int index)
  {
    if (fmt != TFormat.CF_DOUBLE) throw new TineRuntimeErrorException(TErrorList.illegal_format);
    int i1,i2,i3,i4,i5,i6;
    if (index < 0 || index > totalLength) throw new TineRuntimeErrorException(TErrorList.invalid_index);
    switch (ndim)
    {
      case 1: return ((double[])data)[index];
      case 2:
        i1 = index/(axis[1].len);
        i2 = index%(axis[1].len);
        return ((double[][])data)[i1][i2];
      case 3:
        if (data == null) data = new float[axis[0].len][axis[1].len][axis[2].len];
        i1 = index/(axis[1].len * axis[2].len);
        i2 = (index%(axis[1].len * axis[2].len))/axis[2].len;
        i3 = (index%(axis[1].len * axis[2].len))%(axis[2].len);
        return ((double[][][])data)[i1][i2][i3];
      case 4:
        if (data == null) data = new float[axis[0].len][axis[1].len][axis[2].len][axis[3].len];
        i1 = index/(axis[1].len * axis[2].len * axis[3].len);
        i2 = (index%(axis[1].len * axis[2].len * axis[3].len))/(axis[2].len * axis[3].len);
        i3 = (index%(axis[1].len * axis[2].len * axis[3].len))%(axis[2].len * axis[3].len)/axis[3].len;
        i4 = (index%(axis[1].len * axis[2].len * axis[3].len))%(axis[2].len * axis[3].len)%(axis[3].len);
        return ((double[][][][])data)[i1][i2][i3][i4];
      case 5:
        if (data == null) data = new float[axis[0].len][axis[1].len][axis[2].len][axis[3].len][axis[4].len];
        i1 = index/(axis[1].len * axis[2].len * axis[3].len * axis[4].len);
        i2 = (index%(axis[1].len * axis[2].len * axis[3].len * axis[4].len))/(axis[2].len * axis[3].len * axis[4].len);
        i3 = (index%(axis[1].len * axis[2].len * axis[3].len * axis[4].len))%(axis[2].len * axis[3].len * axis[4].len)/(axis[3].len * axis[4].len);
        i4 = (index%(axis[1].len * axis[2].len * axis[3].len * axis[4].len))%(axis[2].len * axis[3].len * axis[4].len)%(axis[3].len * axis[4].len)/(axis[4].len);
        i5 = (index%(axis[1].len * axis[2].len * axis[3].len * axis[4].len))%(axis[2].len * axis[3].len * axis[4].len)%(axis[3].len * axis[4].len)%(axis[4].len);
        return ((double[][][][][])data)[i1][i2][i3][i4][i5];
      case 6:
        if (data == null) data = new float[axis[0].len][axis[1].len][axis[2].len][axis[3].len][axis[4].len][axis[5].len];
        i1 = index/(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len);
        i2 = (index%(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len))/(axis[2].len * axis[3].len * axis[4].len * axis[5].len);
        i3 = (index%(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len))%(axis[2].len * axis[3].len * axis[4].len * axis[5].len)/(axis[3].len * axis[4].len * axis[5].len);
        i4 = (index%(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len))%(axis[2].len * axis[3].len * axis[4].len * axis[5].len)%(axis[3].len * axis[4].len * axis[5].len)/(axis[4].len * axis[5].len);
        i5 = (index%(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len))%(axis[2].len * axis[3].len * axis[4].len * axis[5].len)%(axis[3].len * axis[4].len * axis[5].len)%(axis[4].len * axis[5].len)/(axis[5].len);
        i6 = (index%(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len))%(axis[2].len * axis[3].len * axis[4].len * axis[5].len)%(axis[3].len * axis[4].len * axis[5].len)%(axis[4].len * axis[5].len)%(axis[5].len);
        return ((double[][][][][][])data)[i1][i2][i3][i4][i5][i6];
    }
    throw new TineRuntimeErrorException(TErrorList.dimension_error);
  }
  int getIntElement(int index)
  {
    if (fmt != TFormat.CF_INT32) throw new TineRuntimeErrorException(TErrorList.illegal_format);
    int i1,i2,i3,i4,i5,i6;
    if (index < 0 || index > totalLength) throw new TineRuntimeErrorException(TErrorList.invalid_index);
    switch (ndim)
    {
      case 1: return ((int[])data)[index];
      case 2:
        i1 = index/(axis[1].len);
        i2 = index%(axis[1].len);
        return ((int[][])data)[i1][i2];
      case 3:
        if (data == null) data = new float[axis[0].len][axis[1].len][axis[2].len];
        i1 = index/(axis[1].len * axis[2].len);
        i2 = (index%(axis[1].len * axis[2].len))/axis[2].len;
        i3 = (index%(axis[1].len * axis[2].len))%(axis[2].len);
        return ((int[][][])data)[i1][i2][i3];
      case 4:
        if (data == null) data = new float[axis[0].len][axis[1].len][axis[2].len][axis[3].len];
        i1 = index/(axis[1].len * axis[2].len * axis[3].len);
        i2 = (index%(axis[1].len * axis[2].len * axis[3].len))/(axis[2].len * axis[3].len);
        i3 = (index%(axis[1].len * axis[2].len * axis[3].len))%(axis[2].len * axis[3].len)/axis[3].len;
        i4 = (index%(axis[1].len * axis[2].len * axis[3].len))%(axis[2].len * axis[3].len)%(axis[3].len);
        return ((int[][][][])data)[i1][i2][i3][i4];
      case 5:
        if (data == null) data = new float[axis[0].len][axis[1].len][axis[2].len][axis[3].len][axis[4].len];
        i1 = index/(axis[1].len * axis[2].len * axis[3].len * axis[4].len);
        i2 = (index%(axis[1].len * axis[2].len * axis[3].len * axis[4].len))/(axis[2].len * axis[3].len * axis[4].len);
        i3 = (index%(axis[1].len * axis[2].len * axis[3].len * axis[4].len))%(axis[2].len * axis[3].len * axis[4].len)/(axis[3].len * axis[4].len);
        i4 = (index%(axis[1].len * axis[2].len * axis[3].len * axis[4].len))%(axis[2].len * axis[3].len * axis[4].len)%(axis[3].len * axis[4].len)/(axis[4].len);
        i5 = (index%(axis[1].len * axis[2].len * axis[3].len * axis[4].len))%(axis[2].len * axis[3].len * axis[4].len)%(axis[3].len * axis[4].len)%(axis[4].len);
        return ((int[][][][][])data)[i1][i2][i3][i4][i5];
      case 6:
        if (data == null) data = new float[axis[0].len][axis[1].len][axis[2].len][axis[3].len][axis[4].len][axis[5].len];
        i1 = index/(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len);
        i2 = (index%(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len))/(axis[2].len * axis[3].len * axis[4].len * axis[5].len);
        i3 = (index%(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len))%(axis[2].len * axis[3].len * axis[4].len * axis[5].len)/(axis[3].len * axis[4].len * axis[5].len);
        i4 = (index%(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len))%(axis[2].len * axis[3].len * axis[4].len * axis[5].len)%(axis[3].len * axis[4].len * axis[5].len)/(axis[4].len * axis[5].len);
        i5 = (index%(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len))%(axis[2].len * axis[3].len * axis[4].len * axis[5].len)%(axis[3].len * axis[4].len * axis[5].len)%(axis[4].len * axis[5].len)/(axis[5].len);
        i6 = (index%(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len))%(axis[2].len * axis[3].len * axis[4].len * axis[5].len)%(axis[3].len * axis[4].len * axis[5].len)%(axis[4].len * axis[5].len)%(axis[5].len);
        return ((int[][][][][][])data)[i1][i2][i3][i4][i5][i6];
    }
    throw new TineRuntimeErrorException(TErrorList.dimension_error);
  }
  short getShortElement(int index)
  {
    if (fmt != TFormat.CF_INT16) throw new TineRuntimeErrorException(TErrorList.illegal_format);
    int i1,i2,i3,i4,i5,i6;
    if (index < 0 || index > totalLength) throw new TineRuntimeErrorException(TErrorList.invalid_index);
    switch (ndim)
    {
      case 1: return ((short[])data)[index];
      case 2:
        i1 = index/(axis[1].len);
        i2 = index%(axis[1].len);
        return ((short[][])data)[i1][i2];
      case 3:
        if (data == null) data = new float[axis[0].len][axis[1].len][axis[2].len];
        i1 = index/(axis[1].len * axis[2].len);
        i2 = (index%(axis[1].len * axis[2].len))/axis[2].len;
        i3 = (index%(axis[1].len * axis[2].len))%(axis[2].len);
        return ((short[][][])data)[i1][i2][i3];
      case 4:
        if (data == null) data = new float[axis[0].len][axis[1].len][axis[2].len][axis[3].len];
        i1 = index/(axis[1].len * axis[2].len * axis[3].len);
        i2 = (index%(axis[1].len * axis[2].len * axis[3].len))/(axis[2].len * axis[3].len);
        i3 = (index%(axis[1].len * axis[2].len * axis[3].len))%(axis[2].len * axis[3].len)/axis[3].len;
        i4 = (index%(axis[1].len * axis[2].len * axis[3].len))%(axis[2].len * axis[3].len)%(axis[3].len);
        return ((short[][][][])data)[i1][i2][i3][i4];
      case 5:
        if (data == null) data = new float[axis[0].len][axis[1].len][axis[2].len][axis[3].len][axis[4].len];
        i1 = index/(axis[1].len * axis[2].len * axis[3].len * axis[4].len);
        i2 = (index%(axis[1].len * axis[2].len * axis[3].len * axis[4].len))/(axis[2].len * axis[3].len * axis[4].len);
        i3 = (index%(axis[1].len * axis[2].len * axis[3].len * axis[4].len))%(axis[2].len * axis[3].len * axis[4].len)/(axis[3].len * axis[4].len);
        i4 = (index%(axis[1].len * axis[2].len * axis[3].len * axis[4].len))%(axis[2].len * axis[3].len * axis[4].len)%(axis[3].len * axis[4].len)/(axis[4].len);
        i5 = (index%(axis[1].len * axis[2].len * axis[3].len * axis[4].len))%(axis[2].len * axis[3].len * axis[4].len)%(axis[3].len * axis[4].len)%(axis[4].len);
        return ((short[][][][][])data)[i1][i2][i3][i4][i5];
      case 6:
        if (data == null) data = new float[axis[0].len][axis[1].len][axis[2].len][axis[3].len][axis[4].len][axis[5].len];
        i1 = index/(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len);
        i2 = (index%(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len))/(axis[2].len * axis[3].len * axis[4].len * axis[5].len);
        i3 = (index%(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len))%(axis[2].len * axis[3].len * axis[4].len * axis[5].len)/(axis[3].len * axis[4].len * axis[5].len);
        i4 = (index%(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len))%(axis[2].len * axis[3].len * axis[4].len * axis[5].len)%(axis[3].len * axis[4].len * axis[5].len)/(axis[4].len * axis[5].len);
        i5 = (index%(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len))%(axis[2].len * axis[3].len * axis[4].len * axis[5].len)%(axis[3].len * axis[4].len * axis[5].len)%(axis[4].len * axis[5].len)/(axis[5].len);
        i6 = (index%(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len))%(axis[2].len * axis[3].len * axis[4].len * axis[5].len)%(axis[3].len * axis[4].len * axis[5].len)%(axis[4].len * axis[5].len)%(axis[5].len);
        return ((short[][][][][][])data)[i1][i2][i3][i4][i5][i6];
    }
    throw new TineRuntimeErrorException(TErrorList.dimension_error);
  }
  @Override
  public void toStruct()
  {
    try
    {
      ByteArrayInputStream dis = new ByteArrayInputStream(hByteArray);
      DataInputStream ds = new DataInputStream(dis);
      Arrays.fill(cmt, (byte)0);
      ds.read(cmt, 0, COMMENT_SIZE);
      cmtStr = new String(cmt);
      ndim = Swap.Int(ds.readInt());
      fmt = TFormat.makeByteFormatCode(Swap.Int(ds.readInt()));
      int tlen = 1;
      for (int i=0; i<MAX_MDA_DIMS; i++)
      {
        axis[i] = new MDX();
        ds.read(axis[i].hByteArray, 0, MDX.MDX_SIZE);
        axis[i].toStruct();
        if (axis[i].len > 1) tlen *= axis[i].len;
      }
      // that was the header, now the data:
      tlen *= TFormat.formatSizeOf((short)fmt);
      byte[] ba = new byte[tlen];
      ds.read(ba,0,tlen);
      switch (fmt)
      {
        case TFormat.CF_FLOAT:
          assignFloatValues(Swap.FloatArray(ba));
          break;
        case TFormat.CF_DOUBLE:
          assignDoubleValues(Swap.DoubleArray(ba));
          break;
        case TFormat.CF_INT32:
          assignIntValues(Swap.IntArray(ba));
          break;
        case TFormat.CF_INT16:
          assignShortValues(Swap.ShortArray(ba));
          break;
      }      
      ds.close(); dis.close();
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("MDA",e.getMessage(),TErrorList.code_failure,e,0);        
    }  
  }
  void assignFloatValues(float[] fa)
  {
    if (fmt != TFormat.CF_FLOAT) return;
    int i1,i2,i3,i4,i5,i6;
    for (int i=0; i<fa.length; i++)
    {
      switch (ndim)
      {
        case 1:
          if (data == null) data = new float[axis[0].len];
          ((float [])data)[i] = fa[i];
          break;
        case 2:
          if (data == null) data = new float[axis[0].len][axis[1].len];
          i1 = i/(axis[1].len);
          i2 = i%(axis[1].len);
          ((float[][])data)[i1][i2] = fa[i];
          break;
        case 3:
          if (data == null) data = new float[axis[0].len][axis[1].len][axis[2].len];
          i1 = i/(axis[1].len * axis[2].len);
          i2 = (i%(axis[1].len * axis[2].len))/axis[2].len;
          i3 = (i%(axis[1].len * axis[2].len))%(axis[2].len);
          ((float[][][])data)[i1][i2][i3] = fa[i];
          break;
        case 4:
          if (data == null) data = new float[axis[0].len][axis[1].len][axis[2].len][axis[3].len];
          i1 = i/(axis[1].len * axis[2].len * axis[3].len);
          i2 = (i%(axis[1].len * axis[2].len * axis[3].len))/(axis[2].len * axis[3].len);
          i3 = (i%(axis[1].len * axis[2].len * axis[3].len))%(axis[2].len * axis[3].len)/axis[3].len;
          i4 = (i%(axis[1].len * axis[2].len * axis[3].len))%(axis[2].len * axis[3].len)%(axis[3].len);
          ((float[][][][])data)[i1][i2][i3][i4] = fa[i];
          break;
        case 5:
          if (data == null) data = new float[axis[0].len][axis[1].len][axis[2].len][axis[3].len][axis[4].len];
          i1 = i/(axis[1].len * axis[2].len * axis[3].len * axis[4].len);
          i2 = (i%(axis[1].len * axis[2].len * axis[3].len * axis[4].len))/(axis[2].len * axis[3].len * axis[4].len);
          i3 = (i%(axis[1].len * axis[2].len * axis[3].len * axis[4].len))%(axis[2].len * axis[3].len * axis[4].len)/(axis[3].len * axis[4].len);
          i4 = (i%(axis[1].len * axis[2].len * axis[3].len * axis[4].len))%(axis[2].len * axis[3].len * axis[4].len)%(axis[3].len * axis[4].len)/(axis[4].len);
          i5 = (i%(axis[1].len * axis[2].len * axis[3].len * axis[4].len))%(axis[2].len * axis[3].len * axis[4].len)%(axis[3].len * axis[4].len)%(axis[4].len);
          ((float[][][][][])data)[i1][i2][i3][i4][i5] = fa[i];
          break;
        case 6:
          if (data == null) data = new float[axis[0].len][axis[1].len][axis[2].len][axis[3].len][axis[4].len][axis[5].len];
          i1 = i/(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len);
          i2 = (i%(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len))/(axis[2].len * axis[3].len * axis[4].len * axis[5].len);
          i3 = (i%(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len))%(axis[2].len * axis[3].len * axis[4].len * axis[5].len)/(axis[3].len * axis[4].len * axis[5].len);
          i4 = (i%(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len))%(axis[2].len * axis[3].len * axis[4].len * axis[5].len)%(axis[3].len * axis[4].len * axis[5].len)/(axis[4].len * axis[5].len);
          i5 = (i%(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len))%(axis[2].len * axis[3].len * axis[4].len * axis[5].len)%(axis[3].len * axis[4].len * axis[5].len)%(axis[4].len * axis[5].len)/(axis[5].len);
          i6 = (i%(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len))%(axis[2].len * axis[3].len * axis[4].len * axis[5].len)%(axis[3].len * axis[4].len * axis[5].len)%(axis[4].len * axis[5].len)%(axis[5].len);
          ((float[][][][][][])data)[i1][i2][i3][i4][i5][i6] = fa[i];
          break;
      }
    }
  }
  void assignDoubleValues(double[] fa)
  {
    int i1,i2,i3,i4,i5,i6;
    for (int i=0; i<fa.length; i++)
    {
      switch (ndim)
      {
        case 1:
          if (data == null) data = new double[axis[0].len];
          ((double [])data)[i] = fa[i];
          break;
        case 2:
          if (data == null) data = new double[axis[0].len][axis[1].len];
          i1 = i/(axis[1].len);
          i2 = i%(axis[1].len);
          ((double[][])data)[i1][i2] = fa[i];
          break;
        case 3:
          if (data == null) data = new double[axis[0].len][axis[1].len][axis[2].len];
          i1 = i/(axis[1].len * axis[2].len);
          i2 = (i%(axis[1].len * axis[2].len))/axis[2].len;
          i3 = (i%(axis[1].len * axis[2].len))%(axis[2].len);
          ((double[][][])data)[i1][i2][i3] = fa[i];
          break;
        case 4:
          if (data == null) data = new double[axis[0].len][axis[1].len][axis[2].len][axis[3].len];
          i1 = i/(axis[1].len * axis[2].len * axis[3].len);
          i2 = (i%(axis[1].len * axis[2].len * axis[3].len))/(axis[2].len * axis[3].len);
          i3 = (i%(axis[1].len * axis[2].len * axis[3].len))%(axis[2].len * axis[3].len)/axis[3].len;
          i4 = (i%(axis[1].len * axis[2].len * axis[3].len))%(axis[2].len * axis[3].len)%(axis[3].len);
          ((double[][][][])data)[i1][i2][i3][i4] = fa[i];
          break;
        case 5:
          if (data == null) data = new double[axis[0].len][axis[1].len][axis[2].len][axis[3].len][axis[4].len];
          i1 = i/(axis[1].len * axis[2].len * axis[3].len * axis[4].len);
          i2 = (i%(axis[1].len * axis[2].len * axis[3].len * axis[4].len))/(axis[2].len * axis[3].len * axis[4].len);
          i3 = (i%(axis[1].len * axis[2].len * axis[3].len * axis[4].len))%(axis[2].len * axis[3].len * axis[4].len)/(axis[3].len * axis[4].len);
          i4 = (i%(axis[1].len * axis[2].len * axis[3].len * axis[4].len))%(axis[2].len * axis[3].len * axis[4].len)%(axis[3].len * axis[4].len)/(axis[4].len);
          i5 = (i%(axis[1].len * axis[2].len * axis[3].len * axis[4].len))%(axis[2].len * axis[3].len * axis[4].len)%(axis[3].len * axis[4].len)%(axis[4].len);
          ((double[][][][][])data)[i1][i2][i3][i4][i5] = fa[i];
          break;
        case 6:
          if (data == null) data = new double[axis[0].len][axis[1].len][axis[2].len][axis[3].len][axis[4].len][axis[5].len];
          i1 = i/(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len);
          i2 = (i%(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len))/(axis[2].len * axis[3].len * axis[4].len * axis[5].len);
          i3 = (i%(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len))%(axis[2].len * axis[3].len * axis[4].len * axis[5].len)/(axis[3].len * axis[4].len * axis[5].len);
          i4 = (i%(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len))%(axis[2].len * axis[3].len * axis[4].len * axis[5].len)%(axis[3].len * axis[4].len * axis[5].len)/(axis[4].len * axis[5].len);
          i5 = (i%(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len))%(axis[2].len * axis[3].len * axis[4].len * axis[5].len)%(axis[3].len * axis[4].len * axis[5].len)%(axis[4].len * axis[5].len)/(axis[5].len);
          i6 = (i%(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len))%(axis[2].len * axis[3].len * axis[4].len * axis[5].len)%(axis[3].len * axis[4].len * axis[5].len)%(axis[4].len * axis[5].len)%(axis[5].len);
          ((double[][][][][][])data)[i1][i2][i3][i4][i5][i6] = fa[i];
          break;
      }
    }
  }
  void assignIntValues(int[] fa)
  {
    int i1,i2,i3,i4,i5,i6;
    for (int i=0; i<fa.length; i++)
    {
      switch (ndim)
      {
        case 1:
          if (data == null) data = new int[axis[0].len];
          ((int [])data)[i] = fa[i];
          break;
        case 2:
          if (data == null) data = new int[axis[0].len][axis[1].len];
          i1 = i/(axis[1].len);
          i2 = i%(axis[1].len);
          ((int[][])data)[i1][i2] = fa[i];
          break;
        case 3:
          if (data == null) data = new int[axis[0].len][axis[1].len][axis[2].len];
          i1 = i/(axis[1].len * axis[2].len);
          i2 = (i%(axis[1].len * axis[2].len))/axis[2].len;
          i3 = (i%(axis[1].len * axis[2].len))%(axis[2].len);
          ((int[][][])data)[i1][i2][i3] = fa[i];
          break;
        case 4:
          if (data == null) data = new int[axis[0].len][axis[1].len][axis[2].len][axis[3].len];
          i1 = i/(axis[1].len * axis[2].len * axis[3].len);
          i2 = (i%(axis[1].len * axis[2].len * axis[3].len))/(axis[2].len * axis[3].len);
          i3 = (i%(axis[1].len * axis[2].len * axis[3].len))%(axis[2].len * axis[3].len)/axis[3].len;
          i4 = (i%(axis[1].len * axis[2].len * axis[3].len))%(axis[2].len * axis[3].len)%(axis[3].len);
          ((int[][][][])data)[i1][i2][i3][i4] = fa[i];
          break;
        case 5:
          if (data == null) data = new int[axis[0].len][axis[1].len][axis[2].len][axis[3].len][axis[4].len];
          i1 = i/(axis[1].len * axis[2].len * axis[3].len * axis[4].len);
          i2 = (i%(axis[1].len * axis[2].len * axis[3].len * axis[4].len))/(axis[2].len * axis[3].len * axis[4].len);
          i3 = (i%(axis[1].len * axis[2].len * axis[3].len * axis[4].len))%(axis[2].len * axis[3].len * axis[4].len)/(axis[3].len * axis[4].len);
          i4 = (i%(axis[1].len * axis[2].len * axis[3].len * axis[4].len))%(axis[2].len * axis[3].len * axis[4].len)%(axis[3].len * axis[4].len)/(axis[4].len);
          i5 = (i%(axis[1].len * axis[2].len * axis[3].len * axis[4].len))%(axis[2].len * axis[3].len * axis[4].len)%(axis[3].len * axis[4].len)%(axis[4].len);
          ((int[][][][][])data)[i1][i2][i3][i4][i5] = fa[i];
          break;
        case 6:
          if (data == null) data = new int[axis[0].len][axis[1].len][axis[2].len][axis[3].len][axis[4].len][axis[5].len];
          i1 = i/(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len);
          i2 = (i%(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len))/(axis[2].len * axis[3].len * axis[4].len * axis[5].len);
          i3 = (i%(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len))%(axis[2].len * axis[3].len * axis[4].len * axis[5].len)/(axis[3].len * axis[4].len * axis[5].len);
          i4 = (i%(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len))%(axis[2].len * axis[3].len * axis[4].len * axis[5].len)%(axis[3].len * axis[4].len * axis[5].len)/(axis[4].len * axis[5].len);
          i5 = (i%(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len))%(axis[2].len * axis[3].len * axis[4].len * axis[5].len)%(axis[3].len * axis[4].len * axis[5].len)%(axis[4].len * axis[5].len)/(axis[5].len);
          i6 = (i%(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len))%(axis[2].len * axis[3].len * axis[4].len * axis[5].len)%(axis[3].len * axis[4].len * axis[5].len)%(axis[4].len * axis[5].len)%(axis[5].len);
          ((int[][][][][][])data)[i1][i2][i3][i4][i5][i6] = fa[i];
          break;
      }
    }
  }
  void assignShortValues(short[] fa)
  {
    int i1,i2,i3,i4,i5,i6;
    for (int i=0; i<fa.length; i++)
    {
      switch (ndim)
      {
        case 1:
          if (data == null) data = new short[axis[0].len];
          ((short [])data)[i] = fa[i];
          break;
        case 2:
          if (data == null) data = new short[axis[0].len][axis[1].len];
          i1 = i/(axis[1].len);
          i2 = i%(axis[1].len);
          ((short[][])data)[i1][i2] = fa[i];
          break;
        case 3:
          if (data == null) data = new short[axis[0].len][axis[1].len][axis[2].len];
          i1 = i/(axis[1].len * axis[2].len);
          i2 = (i%(axis[1].len * axis[2].len))/axis[2].len;
          i3 = (i%(axis[1].len * axis[2].len))%(axis[2].len);
          ((short[][][])data)[i1][i2][i3] = fa[i];
          break;
        case 4:
          if (data == null) data = new short[axis[0].len][axis[1].len][axis[2].len][axis[3].len];
          i1 = i/(axis[1].len * axis[2].len * axis[3].len);
          i2 = (i%(axis[1].len * axis[2].len * axis[3].len))/(axis[2].len * axis[3].len);
          i3 = (i%(axis[1].len * axis[2].len * axis[3].len))%(axis[2].len * axis[3].len)/axis[3].len;
          i4 = (i%(axis[1].len * axis[2].len * axis[3].len))%(axis[2].len * axis[3].len)%(axis[3].len);
          ((short[][][][])data)[i1][i2][i3][i4] = fa[i];
          break;
        case 5:
          if (data == null) data = new short[axis[0].len][axis[1].len][axis[2].len][axis[3].len][axis[4].len];
          i1 = i/(axis[1].len * axis[2].len * axis[3].len * axis[4].len);
          i2 = (i%(axis[1].len * axis[2].len * axis[3].len * axis[4].len))/(axis[2].len * axis[3].len * axis[4].len);
          i3 = (i%(axis[1].len * axis[2].len * axis[3].len * axis[4].len))%(axis[2].len * axis[3].len * axis[4].len)/(axis[3].len * axis[4].len);
          i4 = (i%(axis[1].len * axis[2].len * axis[3].len * axis[4].len))%(axis[2].len * axis[3].len * axis[4].len)%(axis[3].len * axis[4].len)/(axis[4].len);
          i5 = (i%(axis[1].len * axis[2].len * axis[3].len * axis[4].len))%(axis[2].len * axis[3].len * axis[4].len)%(axis[3].len * axis[4].len)%(axis[4].len);
          ((short[][][][][])data)[i1][i2][i3][i4][i5] = fa[i];
          break;
        case 6:
          if (data == null) data = new short[axis[0].len][axis[1].len][axis[2].len][axis[3].len][axis[4].len][axis[5].len];
          i1 = i/(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len);
          i2 = (i%(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len))/(axis[2].len * axis[3].len * axis[4].len * axis[5].len);
          i3 = (i%(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len))%(axis[2].len * axis[3].len * axis[4].len * axis[5].len)/(axis[3].len * axis[4].len * axis[5].len);
          i4 = (i%(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len))%(axis[2].len * axis[3].len * axis[4].len * axis[5].len)%(axis[3].len * axis[4].len * axis[5].len)/(axis[4].len * axis[5].len);
          i5 = (i%(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len))%(axis[2].len * axis[3].len * axis[4].len * axis[5].len)%(axis[3].len * axis[4].len * axis[5].len)%(axis[4].len * axis[5].len)/(axis[5].len);
          i6 = (i%(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len))%(axis[2].len * axis[3].len * axis[4].len * axis[5].len)%(axis[3].len * axis[4].len * axis[5].len)%(axis[4].len * axis[5].len)%(axis[5].len);
          ((short[][][][][][])data)[i1][i2][i3][i4][i5][i6] = fa[i];
          break;
      }
    }
  }
  int getRunningIndex(int i1,int i2,int i3,int i4,int i5,int i6)
  {
    switch (ndim)
    {
      case 1: return i1;
      case 2: return i1*axis[1].len + i2;
      case 3: return i1*(axis[1].len * axis[2].len) + 
                     i2*(axis[2].len) + i3;
      case 4: return i1*(axis[1].len * axis[2].len * axis[3].len) + 
                     i2*(axis[2].len * axis[3].len) + 
                     i3*(axis[3].len) + i4;
      case 5: return i1*(axis[1].len * axis[2].len * axis[3].len * axis[4].len) + 
                     i2*(axis[2].len * axis[3].len * axis[4].len) + 
                     i3*(axis[3].len * axis[4].len) + 
                     i4*(axis[4].len) + i5;
      case 6: return i1*(axis[1].len * axis[2].len * axis[3].len * axis[4].len * axis[5].len) + 
                     i2*(axis[2].len * axis[3].len * axis[4].len * axis[5].len) + 
                     i3*(axis[3].len * axis[4].len * axis[5].len) + 
                     i4*(axis[4].len * axis[5].len) + i5 * axis[5].len + i6;
    }
    return -1;
  }
  @Override
  public byte[] getByteArray()
  {
    return hByteArray;
  }
  @Override
  public void setByteArray(byte[] byteArray)
  {
    hByteArray = byteArray;
  }
  @Override
  public int getSizeInBytes()
  {
    return sizeInBytes;
  }
  public int getTotalLength()
  {
    return totalLength;
  }
  @Override
  public short getFormat()
  {
    return TFormat.CF_MDA;
  }
  @Override
  public String getTag()
  {
    return TFormat.toString((short)fmt);
  }
  public double getDoubleValue()
  {
    return 0;
  }
  public int getIntegerValue()
  {
    return 0;
  }
  public String getCsvString()
  {
    String s = toString();
    if (s.contains(",")) s = "\""+s+"\"";
    return s;
  }
  protected String separator = ", ";
  public void setFieldSeparator(String value)
  {
    if (value != null) separator = value;
  }
  public String toString()
  {
    String s = cmtStr.trim()+"\n[";
    for (int i=0; i<ndim; i++)
    {
      s += axis[i].len;
      if (axis[i].eguStr != null) s += " "+axis[i].eguStr.trim();
      if (axis[i].lblStr != null && axis[i].lblStr.length() > 0)
        s += " ("+axis[i].lblStr.trim()+")";
      if (i<ndim-1) s += " X ";
    }
    s += "]";
    if (totalLength <= 1000  && data != null)
    {
      s += "\n"+Arrays.deepToString((Object[]) data);
    }
    return s;    
  }
  public MDA clone()
  {
    return new MDA(this);
  }

}
