/*
 * File NAME32DBLDBL.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.types;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import de.desy.tine.definitions.TErrorList;
import de.desy.tine.definitions.TFormat;
import de.desy.tine.endianUtils.Swap;
import de.desy.tine.server.logger.MsgLog;

public final class NAME32DBLDBL implements TCompoundDataObject
{
  public String name;
  public double d1val;
  public double d2val;
  public double getDoubleValue() { return (double)d1val; }
  public int getIntegerValue() { return (int)d2val; }
  public static final int sizeInBytes = 48;
  public static final short dFormat = TFormat.CF_NAME32DBLDBL;
  public void setValues(String newName, double newD1Value, double newD2Value)
  {
    name = newName; d1val = newD1Value; d2val = newD2Value;
  }
  public void setName(String newName) { name = newName; }
  public String getName() { return name; }
  public int getSizeInBytes() { return sizeInBytes; }
  public short getFormat() { return dFormat; }    
  public String getTag() { return null; }
  ByteArrayOutputStream dBuffer = new ByteArrayOutputStream(sizeInBytes);
  public byte[] hByteArray;
  public void setByteArray(byte[] byteArray) { hByteArray = byteArray; }
  public byte[] getByteArray() 
  { 
    return hByteArray == null ? toByteArray() : hByteArray; 
  }
  public NAME32DBLDBL() {name = new String(new StringBuffer(32)); d1val = 0; d2val = 0;}  
  public NAME32DBLDBL(String Name) { name = Name; d1val = 0; d2val = 0;}  
  public NAME32DBLDBL(String Name,double d) { name = Name; d1val = d; d2val = 0;}  
  public NAME32DBLDBL(String Name,double d1,double d2) { name = Name; d1val = d1; d2val = d2;}  
  public NAME32DBLDBL(String Name,int d1,int d2) { name = Name; d1val = (double)d1; d2val = (double)d2;}  
  public NAME32DBLDBL(String Name,float d1,int d2) { name = Name; d1val = (double)d1; d2val = (double)d2;}  
  public NAME32DBLDBL(NAME32DBLDBL instance) 
  { 
    name = instance.name == null ? new String(new StringBuffer(32)) : new String(instance.name); 
    d1val = instance.d1val; d2val = instance.d2val; 
  }
  public void copy(TCompoundDataObject from)
  {
    if (from instanceof NAME32DBLDBL)
    {
      NAME32DBLDBL instance = (NAME32DBLDBL)from;
      name = instance.name;
      d1val = instance.d1val; d2val = instance.d2val; 
    }
  }
  protected String separator = ", ";
  public void setFieldSeparator(String value)
  {
    if (value != null) separator = value;
  }
  public String toString()
  {
      return "["+ name + separator + d1val + separator + d2val + "]";
  }
  public String getCsvString()
  {
    String tname = name.trim();
    if (tname.contains(",") || tname.contains(" ") ||
        tname.contains("\t") || tname.contains("\n"))
    {
      return "\""+tname+" "+d1val+" "+d2val+"\"";
    }
    return tname + " "+d1val+" "+d2val;
  }  
  public boolean equals(NAME32DBLDBL instance)
  {
    if (instance.name == null || name == null) return false;
    if (instance.name.compareTo(name) != 0) return false;
    if (instance.d1val != d1val) return false;
    if (instance.d2val != d2val) return false;
    return true;
  }
  public byte[] toByteArray()
  {
    if (name == null) name = new String(new StringBuffer(32));
    try
    {
      dBuffer.reset();
      DataOutputStream ds = new DataOutputStream(dBuffer);
      byte[] bname = name.getBytes();
      byte[] b = new byte[32];
      for (int i=0; i<32 && i<bname.length; i++) b[i] = bname[i];
      ds.write(b,0,b.length);
      byte[] db = Swap.Double(d1val);
      ds.write(db,0,8);
      db = Swap.Double(d2val);
      ds.write(db,0,8);
      ds.close();
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("NAME32DBLDBL","could not prepare byte stream",TErrorList.code_failure,e,0);
      return null;
    }
    hByteArray = dBuffer.toByteArray();
    return hByteArray;
  }  
  public void toStruct()
  {
    try
    {
      ByteArrayInputStream dis = new ByteArrayInputStream(hByteArray);
      DataInputStream ds = new DataInputStream(dis);
      byte[] b = new byte[sizeInBytes];
      ds.read(b,0,32);
      int len = 0; for (int i=0; i<32 && b[i] != 0; i++, len++);
      name = new String(b,0,len);
      byte[] db = new byte[8];
      ds.read(db,0,8);
      d1val = Swap.Double(db);
      ds.read(db,0,8);
      d2val = Swap.Double(db);
      ds.close(); dis.close();
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("NAME32DBLDBL","could not prepare byte stream",TErrorList.code_failure,e,0);
    }  
  }
  public NAME32DBLDBL newInstance()
  {
    return new NAME32DBLDBL();
  }
  public NAME32DBLDBL clone()
  {
    return new NAME32DBLDBL(this);
  }            
  @Override
  public int hashCode()
  {
    final int PRIME = 31;
    int result = 1;
    long temp;
    temp = Double.doubleToLongBits(d1val);
    result = PRIME * result + (int) (temp ^ (temp >>> 32));
    temp = Double.doubleToLongBits(d2val);
    result = PRIME * result + (int) (temp ^ (temp >>> 32));
    result = PRIME * result + ((name == null) ? 0 : name.hashCode());
    return result;
  }
  @Override
  public boolean equals(Object obj)
  {
    if (this == obj) return true;
    if (obj == null) return false;
    if (getClass() != obj.getClass()) return false;
    final NAME32DBLDBL other = (NAME32DBLDBL) obj;
    return equals(other);
  }  
}