/*
 * File SPECTRUM.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

package de.desy.tine.types;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Arrays;

import de.desy.tine.definitions.TErrorList;
import de.desy.tine.definitions.TFormat;
import de.desy.tine.endianUtils.Swap;
import de.desy.tine.server.logger.MsgLog;
public final class SPECTRUM implements TCompoundDataObject
{
  public static final int STD_LENGTH = 8192;
  public String comment;
  public int tm;
  public float s_start;
  public float s_inc;
  public int status;
  public double getDoubleValue() { return (double)d_spect_array[0]; }
  public int getIntegerValue() { return (int)status; }
  public float[] d_spect_array;
  public static int sizeInBytes;
  public static int hdrSizeInBytes = 96;
  public int length;
  public static final short dFormat = TFormat.CF_SPECTRUM;
  public int getSizeInBytes() { return sizeInBytes; }
  public short getFormat() { return dFormat; }    
  public String getTag() { return null; }
  ByteArrayOutputStream dBuffer = new ByteArrayOutputStream(sizeInBytes);
  public byte[] hByteArray;
  public void setByteArray(byte[] byteArray) { hByteArray = byteArray; }
  public byte[] getByteArray() 
  { 
    return hByteArray == null ? toByteArray() : hByteArray; 
  }
  protected String separator = ", ";
  public void setFieldSeparator(String value)
  {
    if (value != null) separator = value;
  }
  public String toString()
  {
    StringBuffer sb = new StringBuffer(length*4 + hdrSizeInBytes);
    sb.append("["+ comment + separator + tm + separator + s_start + separator + s_inc + separator + status + "]\n");
    for (int i=0; i<length; i++) sb.append(" " + d_spect_array[i]); 
    return sb.toString();
  }
  public String getCsvString()
  {
    return "\""+comment + ", " + tm + ", " + s_start + ", " + s_inc + ", " + status+"\"";
  }  
  public void setValues(int newTM, float newStart, float newInc, int newStatus)
  {
    tm = newTM; s_start = newStart; s_inc = newInc; status = newStatus; 
  }
  public SPECTRUM() 
  {
    s_start = 0; s_inc = 0; status = 0; tm  = 0; comment = ""; 
    d_spect_array = new float[STD_LENGTH];
    sizeInBytes = STD_LENGTH * 4 + 96;
    length = STD_LENGTH;
    hByteArray = new byte[sizeInBytes];
  }  
  public SPECTRUM(int spectrumLength) 
  {
    s_start = 0; s_inc = 0; status = 0; tm  = 0; comment = "";
    d_spect_array = new float[spectrumLength];
    sizeInBytes = spectrumLength * 4 + 96;
    length = spectrumLength;
    hByteArray = new byte[sizeInBytes];
  }  
  public SPECTRUM(SPECTRUM instance) 
  {
    if (instance == null) return;
    s_start = instance.s_start; s_inc = instance.s_inc; status = instance.status; tm  = instance.tm; 
    comment = new String(instance.comment); 
    length = instance.length;
    hByteArray = new byte[sizeInBytes];
    d_spect_array = new float[length];
    if (instance.d_spect_array != null)
      System.arraycopy(instance.d_spect_array, 0, d_spect_array, 0, length);
  }  
  @Override
  public void copy(TCompoundDataObject from)
  {
    if (from instanceof SPECTRUM)
    {
      SPECTRUM instance = (SPECTRUM)from;
      s_start = instance.s_start; s_inc = instance.s_inc; status = instance.status; tm  = instance.tm; 
      comment = new String(instance.comment); 
      length = Math.min(length, instance.length);
      if (instance.d_spect_array != null)
        System.arraycopy(instance.d_spect_array, 0, d_spect_array, 0, length);
    }
  }
  public boolean equals(SPECTRUM instance)
  {
    if (instance.s_start != s_start) return false;
    if (instance.s_inc != s_inc) return false;
    if (instance.status != status) return false;
    if (instance.length != length) return false;
    if (instance.comment == null || comment == null) return false;
    if (instance.comment.compareTo(comment) != 0) return false;
    if (!Arrays.equals(instance.d_spect_array, d_spect_array)) return false;
    return true;
  }
  public byte[] toByteArray()
  {
	try
	{
		dBuffer.reset();
		DataOutputStream ds = new DataOutputStream(dBuffer);
		byte[] b = new byte[80];
		ds.write(comment.getBytes());
		ds.write(b,0,80-comment.length());
		ds.writeInt(Swap.Int(tm));
		b = Swap.Float(s_start); ds.write(b,0,4);
		b = Swap.Float(s_inc); ds.write(b,0,4);
		ds.writeInt(Swap.Int(status));
		for (int i=0; i<d_spect_array.length; i++)
		{
		  b = Swap.Float(d_spect_array[i]); ds.write(b,0,4);
		}
		ds.close();
	}
	catch (IOException e)
	{
	  e.printStackTrace();
      MsgLog.log("SPECTRUM",e.getMessage(),TErrorList.code_failure,e,0);        
	}
	hByteArray = dBuffer.toByteArray();
	return hByteArray;
  }  
  public void toStruct()
  {
    try
    {
    	ByteArrayInputStream dis = new ByteArrayInputStream(hByteArray);
    	DataInputStream ds = new DataInputStream(dis);
    	byte[] b = new byte[96];
    	ds.read(b,0,80);
    	comment = (new String(b)).trim();
    	tm = Swap.Int(ds.readInt());
    	ds.read(b,0,4); s_start = Swap.Float(b);
    	ds.read(b,0,4); s_inc = Swap.Float(b);
    	status = Swap.Int(ds.readInt());
    	for (int i=0; i<d_spect_array.length && i<length; i++)
    	{
    	  if (ds.read(b,0,4) < 0) break;
    	  d_spect_array[i] = Swap.Float(b);
    	}
    	ds.close(); dis.close();
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("SPECTRUM.toStruct",e.getMessage(),TErrorList.code_failure,e,0);        
    }	 
  }
  public int hashCode()
  {
    final int PRIME = 31;
    int result = 1;
    result = PRIME * result + ((comment == null) ? 0 : comment.hashCode());
    result = PRIME * result + Arrays.hashCode(d_spect_array);
    result = PRIME * result + length;
    result = PRIME * result + Float.floatToIntBits(s_inc);
    result = PRIME * result + Float.floatToIntBits(s_start);
    result = PRIME * result + status;
    result = PRIME * result + tm;
    return result;
  }
  public SPECTRUM newInstance()
  {
    return new SPECTRUM();
  }
  public SPECTRUM clone()
  {
    return new SPECTRUM(this);
  }              
  public boolean equals(Object obj)
  {
    if (this == obj) return true;
    if (obj == null) return false;
    if (getClass() != obj.getClass()) return false;
    final SPECTRUM other = (SPECTRUM) obj;
    return equals(other);
  }
  public void setSpectrum(SPECTRUM newSpectrum)
  {
    if (newSpectrum == null) return;
    setValues(newSpectrum.getTime(), newSpectrum.getStart(), newSpectrum.getIncrement(), newSpectrum.getStatus());
    setComment(newSpectrum.getComment());
    setValues(newSpectrum.getValues());
  }
  public float[] getValues()
  {
    return d_spect_array;
  }
  public void setValues(float[] values)
  {
    if (values == null) return;
    int len = values.length;
    if (len > d_spect_array.length) len = d_spect_array.length;
    System.arraycopy(values, 0, d_spect_array, 0, len);
  }
  public void setComment(String newComment)
  {
    comment = newComment;
  }
  public String getComment()
  {
    return comment;
  }
  public int getLength()
  {
    return length;
  }
  public float getIncrement()
  {
    return s_inc;
  }
  public float getStart()
  {
    return s_start;
  }
  public int getTime()
  {
    return tm;
  }
  public int getStatus()
  {
    return status;
  }
}