/*
 * File TCompoundDataObject.java
 *  
 * Authors Philip Duval (DESY), Jaka Bobnar (CosyLab), 
 *    Jan Szczesny (DESY), Karol Bartkiewicz (DESY), 
 *    Marcus Walla (DESY), Tobias Tempel (DESY)  
 *  
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023  
 *  
 * The TINE .java codeset made available by open source license
 * inherits several conceptual ideas from the original ISOLDE 
 * control system written in the C language ((c)CERN, released under 
 * a permissive CERN licence akin to the later well known MIT licence) 
 * in use at CERN in 1992. We therefore owe a substantial debt to CERN 
 * for its inception, and would be remiss if we did not express our 
 * gratitude in the JAVA TINE open source file headers, although none 
 * of the current Java software is in anyway a redistribution of 
 * software from the ISOLDE system of that era.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");  
 * You may not use this file except in compliance with the License.  
 * You may obtain a copy of the License at  
 *  
 * http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" BASIS,  
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
 * See the License for the specific language governing permissions and  
 * limitations under the License.  
 */

/*
 * Created on Jan 31, 2006
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package de.desy.tine.types;

/**
 * interface for most TINE complex data types
 * 
 * @author duval
 *
 */
public interface TCompoundDataObject
{
  /**
   * Active method to convert the relevant data fields of this compound data 
   * type into the underlying byte stream (data being sent).  This primarily involves byte swapping 
   * from big-endian to little-endian.  Used systematically by TDataType() and other
   * routines.
   *
   * @return a reference to the underlying bytestream (as a courtesy) 
   */
  public byte[] toByteArray();
  /**
   * Active method to convert the underlying byte stream to the relevant data fields of
   * this compound data type. (data being received).  This primarily involved swapping from
   * little-endian to big-endian.  Used systematically by TDataType() and other routines.
   * @return
   */
  public void toStruct();
  /**
   * Passive method that simple returns the reference to the underlying bytestream
   * Used by TDataType(), TTaggedStruct, and THistoryRecord.
   * @return a reference to the underlying bytestream (if any)
   */
  public byte[] getByteArray();
  /**
   * Sets the underlying bytestream to that give
   * Used only by the HISTORY data type for manipulating compound data types.
   * @param byteArray a new reference to an underlying bytestream
   */
  public void setByteArray(byte[] byteArray);
  /**
   * Used by TDataType, TTaggedStruct and other routines.
   * @return the size in bytes of the of this compound data type
   */
  public int getSizeInBytes();
  /**
   * Used by TDataType, TTaggedStruct and other routines.
   * @return the format code of this compound data type
   */
  public short getFormat();
  /**
   * Used by TDataType, TTaggedStruct and other routines.
   * @return the data tag of this compound data type
   */
  public String getTag();
  /**
   * Used only by the THistoryRecord class in retrieving the primary double floating
   * point value for this compound data object (if available).  If the compound data
   * object has no double value, then a '0' is returned.
   * @return the canonical double value part of this compound data object.  
   */
  public double getDoubleValue();
  /**
   * Used only by the THistoryRecord class in retrieving the primary 32-bit integer
   * value for this compound data object (if available).  If the compound data
   * object has no integer value, then a '0' is returned.
   * @return the canonical integer value part of this compound data object.  
   */
  public int getIntegerValue();  
  /**
   * Used by TDataType.toCsvString() to get a valid parse-able string to 
   * of the objects data contents to supply to csv files.
   */
  public String getCsvString();
  public TCompoundDataObject clone();
  public void copy(TCompoundDataObject from);
  public TCompoundDataObject newInstance();
  /**
   * sets the field separator token used the the toString() method
   * @param value
   */
  public void setFieldSeparator(String value);
}
